/*
 * Copyright 2005 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/* 
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 * 
 * http://www.opensource.org/licenses/gpl-license.html 
 * http://www.gnu.org/copyleft/gpl.html
 */

/*!
 * @file clock.c
 * @brief API for setting up and retrieving clocks.
 *
 * This file contains API for setting up and retrieving clocks.
 *
 * @ingroup CLOCKS
 */

#include <linux/module.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/arch/clock.h>

/*!
 * Spinlock to protect CRM register accesses
 */
static spinlock_t mxc_crm_lock = SPIN_LOCK_UNLOCKED;

/*!
 * This function returns the PLL output value in Hz based on pll.
 * @param       pll     PLL as defined in enum plls
 * @return      PLL value in Hz.
 */
unsigned long mxc_pll_clock(enum plls pll) 
{
        unsigned long long mfi, mfn, mfd, pdf, ref_clk, temp;
        unsigned long pll_out;
        volatile unsigned long reg;
        
        if (pll == MCUPLL) {
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPCTL);
                mfi = (reg & MXC_CCM_MCUPCTL_MFI_MASK) >> 
                       MXC_CCM_MCUPCTL_MFI_OFFSET; 
                mfi = (mfi <= 5) ? 5: mfi;
                mfn = (reg & MXC_CCM_MCUPCTL_MFN_MASK); 
                mfn = (mfn <= 0x400) ? mfn: (mfn - 0x800);
        } else {
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_UPCTL);
                mfi = (reg & MXC_CCM_USBPCTL_MFI_MASK) >> 
                       MXC_CCM_USBPCTL_MFI_OFFSET; 
                mfi = (mfi <= 5) ? 5: mfi;
                mfn = reg & MXC_CCM_USBPCTL_MFN_MASK; 
        }

        pdf = (reg & MXC_CCM_PCTL_PDF_MASK) >> MXC_CCM_PCTL_PDF_OFFSET;
        mfd = (reg & MXC_CCM_PCTL_MFD_MASK) >> MXC_CCM_PCTL_MFD_OFFSET;
        ref_clk = MXC_PLL_REF_CLK;

        if (mfn < 1024) {
                temp = 2 * ref_clk * mfn;
                pll_out = (2 * ref_clk * mfi + (temp / (mfd + 1))) /
                          (pdf + 1);
        } else {
                temp = 2 * ref_clk * (2048 - mfn);
                pll_out = (2 * ref_clk * mfi - (temp / (mfd + 1))) /
                          (pdf + 1);
        }
        return pll_out;
}

/*!
 * This function returns the main clock values in Hz.
 *
 * @param       clk     as defined in enum mxc_clocks
 *
 * @return      clock value in Hz
 */
unsigned long mxc_get_clocks(enum mxc_clocks clk) 
{
        unsigned long pll, ret_val = 0, hclk;
        unsigned long brmm, max_pdf, ipg_pdf, nfc_pdf, csi_pdf, usb_pdf;
        unsigned long ssi1_sel, ssi2_sel, firi_sel, ssi1_prepdf, ssi1_pdf;
        unsigned long ssi2_prepdf, ssi2_pdf, firi_prepdf, firi_pdf;
        volatile unsigned long reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCR);
        volatile unsigned long reg1 = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR0);
        volatile unsigned long reg2 = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR1);

        brmm = reg1 & MXC_CCM_MPDR0_BRMM_MASK;
        max_pdf = (reg1 & MXC_CCM_MPDR0_MAX_PDF_MASK) >> MXC_CCM_MPDR0_MAX_PDF_OFFSET;
        ipg_pdf = (reg1 & MXC_CCM_MPDR0_IPG_PDF_MASK) >> MXC_CCM_MPDR0_IPG_PDF_OFFSET;
        nfc_pdf = (reg1 & MXC_CCM_MPDR0_NFC_PDF_MASK) >> MXC_CCM_MPDR0_NFC_PDF_OFFSET;
        csi_pdf = (reg1 & MXC_CCM_MPDR0_CSI_PDF_MASK) >> MXC_CCM_MPDR0_CSI_PDF_OFFSET;
        usb_pdf = (reg2 & MXC_CCM_MPDR1_USB_PDF_MASK) >> MXC_CCM_MPDR1_USB_PDF_OFFSET;

        switch (clk) {
        case CPU_CLK:
                pll = mxc_pll_clock(MCUPLL);
                if (brmm >= 5) {
                        printk("Wrong BRMM value in the CRM_AP, MPDR0 reg \n");
                        return 0;
                }
                hclk = pll / (max_pdf + 1);
                switch (brmm) {
                case 0:
                        ret_val = pll;
                        break;
                case 1:
                        ret_val = pll + (pll - hclk) / 4;
                        break;
                case 2:
                        ret_val = pll + (pll - hclk) / 2;
                        break;
                case 3:
                        ret_val = pll + (pll - hclk) * 3 / 4;
                        break;
                case 4:
                        ret_val = hclk;
                        break;
                default:
                        break;
                }
                break;
        case AHB_CLK:
        case IPU_CLK:
                pll = mxc_pll_clock(MCUPLL);
                ret_val = pll / (max_pdf + 1);
                break;
        case IPG_CLK:
        case UART1_BAUD:
        case UART2_BAUD:
        case UART3_BAUD:
        case UART4_BAUD:
        case I2C_CLK:
        case SDHC1_CLK:
        case SDHC2_CLK:
                pll = mxc_pll_clock(MCUPLL);
                ret_val = pll / ((max_pdf + 1) * (ipg_pdf + 1));
                break;
        case NFC_CLK:
                pll = mxc_pll_clock(MCUPLL);
                ret_val = pll / ((max_pdf + 1) * (nfc_pdf + 1));
                break;
        case USB_CLK:
                pll = mxc_pll_clock(USBPLL);
                ret_val = pll / (usb_pdf + 1);
                break;
        case SSI1_BAUD:
                ssi1_sel = reg & MXC_CCM_MCR_SSIS1;
                ssi1_prepdf = (reg2 & MXC_CCM_MPDR1_SSI1_PREPDF_MASK) >> 
                               MXC_CCM_MPDR1_SSI1_PREPDF_OFFSET;
                ssi1_pdf = (reg2 & MXC_CCM_MPDR1_SSI1_PDF_MASK) >> 
                            MXC_CCM_MPDR1_SSI1_PDF_OFFSET;
                pll = mxc_pll_clock(ssi1_sel);
                ret_val = pll / ((ssi1_prepdf + 1) * (ssi1_pdf + 1));
                break;
        case SSI2_BAUD:
                ssi2_sel = reg & MXC_CCM_MCR_SSIS2;
                ssi2_prepdf = (reg2 & MXC_CCM_MPDR1_SSI2_PREPDF_MASK) >> 
                               MXC_CCM_MPDR1_SSI2_PREPDF_OFFSET;
                ssi2_pdf = (reg2 & MXC_CCM_MPDR1_SSI2_PDF_MASK) >> 
                            MXC_CCM_MPDR1_SSI2_PDF_OFFSET;
                pll = mxc_pll_clock(ssi2_sel);
                ret_val = pll / ((ssi2_prepdf + 1) * (ssi2_pdf + 1));
                break;
        case CSI_BAUD:
                pll = mxc_pll_clock(MCUPLL);
                csi_pdf = (reg1 & MXC_CCM_MPDR0_CSI_PDF_MASK) >> 
                           MXC_CCM_MPDR0_CSI_PDF_OFFSET;
                ret_val = pll / (csi_pdf + 1);
                break;
        case FIRI_BAUD:
                firi_sel = reg & MXC_CCM_MCR_FIRS;
                firi_prepdf = (reg2 & MXC_CCM_MPDR1_FIRI_PREPDF_MASK) >> 
                               MXC_CCM_MPDR1_FIRI_PREPDF_OFFSET;
                firi_pdf = (reg2 & MXC_CCM_MPDR1_FIRI_PDF_MASK) >> 
                            MXC_CCM_MPDR1_FIRI_PDF_OFFSET;
                pll = mxc_pll_clock(firi_sel);
                ret_val = pll / ((firi_prepdf + 1) * (firi_pdf + 1));
                break;
        default:
                printk("This clock: %d not supported yet \n", clk);
                break;
        }

        return ret_val;
}

/*!
 * This function sets the PLL source for a clock.
 *
 * @param clk     as defined in enum mxc_clocks
 * @param pll_num the PLL that you wish to use as source for this clock
 */
void mxc_set_clocks_pll(enum mxc_clocks clk, enum plls pll_num)
{
        volatile unsigned long mcr;
        unsigned long flags;
                                                                                   
        spin_lock_irqsave(&mxc_crm_lock, flags);
        mcr = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCR);
        switch (clk) {
        case SSI1_BAUD:
                if (pll_num == 1) {
                        mcr |= MXC_CCM_MCR_SSIS1;
                } else {
                        mcr &= ~MXC_CCM_MCR_SSIS1;
                }
                break;
        case SSI2_BAUD:
                if (pll_num == 1) {
                        mcr |= MXC_CCM_MCR_SSIS2;
                } else {
                        mcr &= ~MXC_CCM_MCR_SSIS2;
                }
                break;
        case FIRI_BAUD:
                if (pll_num == 1) {
                        mcr |= MXC_CCM_MCR_FIRS;
                } else {
                        mcr &= ~MXC_CCM_MCR_FIRS;
                }
                break;
        default:
                printk("This clock does not have ability to choose the PLL\n");
                break;
        }
        writel(mcr, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCR);
        spin_unlock_irqrestore(&mxc_crm_lock, flags);
}

/*!
 * This function calculates the pre and post divider values for a clock
 *
 * @param div  divider value passed in
 * @param div1 returns the pre-divider value
 * @param div2 returns the post-divider value
 */
void mxc_clk_getdivs(unsigned int div, unsigned int *div1, 
                     unsigned int *div2)
{
        int i;

        if (div <= 8) {
                *div1 = div;
                *div2 = 1;
                return;
        }
        if (div <= 32) {
                *div1 = 1;
                *div2 = div;
                return;
        }
        for (i = 2; i < 9; i++) {
                if ((div % i) == 0) {
                        *div1 = i;
                        *div2 = div / i;
                        return;
                }
        }
        /* Set to default value */
        *div1 = 2;
        *div2 = 12; 
        return;
}

/*!
 * This function sets the division factor for a clock.
 *
 * @param clk as defined in enum mxc_clocks
 * @param div the division factor to be used for the clock (For SSI & CSI, pass in
 *            2 times the expected division value to account for FP vals)
 */
void mxc_set_clocks_div(enum mxc_clocks clk, unsigned int div)
{
        volatile unsigned long reg;
        unsigned long flags;
        unsigned int d, div1, div2;
                                                                                   
        spin_lock_irqsave(&mxc_crm_lock, flags);
                                                                                   
        switch (clk) {
        case AHB_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR0);
                reg = (reg & (~MXC_CCM_MPDR0_MAX_PDF_MASK)) |
                      ((div - 1) << MXC_CCM_MPDR0_MAX_PDF_OFFSET);
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR0);
                break;
        case IPG_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR0);
                reg = (reg & (~MXC_CCM_MPDR0_IPG_PDF_MASK)) |
                      ((div - 1) << MXC_CCM_MPDR0_IPG_PDF_OFFSET);
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR0);
                break;
        case NFC_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR0);
                reg = (reg & (~MXC_CCM_MPDR0_NFC_PDF_MASK)) |
                      ((div - 1) << MXC_CCM_MPDR0_NFC_PDF_OFFSET);
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR0);
                break;
        case CSI_BAUD:
                d = div / 2;
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR0);
                reg = (reg & (~MXC_CCM_MPDR0_CSI_PDF_MASK)) |
                      ((d - 1) << MXC_CCM_MPDR0_CSI_PDF_OFFSET);
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR0);
                break;
        case USB_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR1);
                reg = (reg & (~MXC_CCM_MPDR1_USB_PDF_MASK)) |
                      ((div - 1) << MXC_CCM_MPDR1_USB_PDF_OFFSET);
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR1);
                break;
        case SSI1_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR1);
                d = div / 2;
                mxc_clk_getdivs(d, &div1, &div2);
                reg = (reg & (~MXC_CCM_MPDR1_SSI1_PREPDF_MASK)) |
                      ((div1 - 1) << MXC_CCM_MPDR1_SSI1_PREPDF_OFFSET);
                reg = (reg & (~MXC_CCM_MPDR1_SSI1_PDF_MASK)) |
                      ((div2 - 1) << MXC_CCM_MPDR1_SSI1_PDF_OFFSET);
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR1);
                break;
        case SSI2_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR1);
                d = div / 2;
                mxc_clk_getdivs(d, &div1, &div2);
                reg = (reg & (~MXC_CCM_MPDR1_SSI2_PREPDF_MASK)) |
                      ((div1 - 1) << MXC_CCM_MPDR1_SSI2_PREPDF_OFFSET);
                reg = (reg & (~MXC_CCM_MPDR1_SSI2_PDF_MASK)) |
                      ((div2 - 1) << MXC_CCM_MPDR1_SSI2_PDF_OFFSET);
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR1);
                break;
        case FIRI_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR1);
                d = div / 2;
                mxc_clk_getdivs(d, &div1, &div2);
                reg = (reg & (~MXC_CCM_MPDR1_FIRI_PREPDF_MASK)) |
                      ((div1 - 1) << MXC_CCM_MPDR1_FIRI_PREPDF_OFFSET);
                reg = (reg & (~MXC_CCM_MPDR1_FIRI_PDF_MASK)) |
                      ((div2 - 1) << MXC_CCM_MPDR1_FIRI_PDF_OFFSET);
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MPDR1);
                break;
        default:
                break;
        }
}

/*!
 * This function is called to gate off the individual module clocks
 *
 * @param       clk     as defined in enum mxc_clocks
 */
void mxc_clks_disable(enum mxc_clocks clk) 
{
        unsigned long flags;
        volatile unsigned long reg;

        spin_lock_irqsave(&mxc_crm_lock, flags);

        switch(clk) {
        case UART1_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_UART1;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case UART2_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_UART2;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case UART3_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_UART3; 
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case UART4_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_UART4;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SSI1_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_SSI1;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0); 
                break;
        case CSPI1_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_CSPI1;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case FIRI_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_FIRI;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case GPT_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_GPTMCU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case RTC_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_RTC;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case EPIT1_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_EPIT1MCU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case EPIT2_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_EPIT2MCU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case EDIO_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_EDIO;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case WDOG_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_WDOGMCU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case PWM_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_PWMMCU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case I2C_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg &= ~MXC_CCM_MCGR0_I2C;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case IPU_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_IPU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case MPEG4_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_MPEG4;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SIM1_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_SIM1;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SIM2_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_SIM2;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case HAC_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_HAC;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case GEM_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_GEM;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case USB_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_USBOTG;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case CSPI2_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_CSPI2;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SDHC2_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_SDHC2;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SDHC1_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_SDHC1;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SSI2_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg &= ~MXC_CCM_MCGR1_SSI2;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SDMA_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                reg &= ~MXC_CCM_MCGR2_SDMA;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                break;
        case RNG_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                reg &= ~MXC_CCM_MCGR2_RNG;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                break;
        case KPP_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                reg &= ~MXC_CCM_MCGR2_KPP;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                break;
        case MU_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                reg &= ~MXC_CCM_MCGR2_MU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                break;
        case SPBA_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                reg &= ~MXC_CCM_MCGR2_SPBA;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                break;
        default:
                printk("The gateoff for this clock(%d) is not implemented\n", clk);
                break;
        }
        spin_unlock_irqrestore(&mxc_crm_lock, flags);
}

/*!
 * This function is called to enable the individual module clocks
 *
 * @param       clk     as defined in enum mxc_clocks
 */
void mxc_clks_enable(enum mxc_clocks clk)
{
        unsigned long flags;
        volatile unsigned long reg;

        spin_lock_irqsave(&mxc_crm_lock, flags);

        switch(clk) {
        case UART1_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_UART1;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case UART2_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_UART2;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case UART3_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_UART3; 
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case UART4_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_UART4;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SSI1_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_SSI1;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0); 
                break;
        case CSPI1_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_CSPI1;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case FIRI_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_FIRI;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case GPT_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_GPTMCU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case RTC_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_RTC;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case EPIT1_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_EPIT1MCU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case EPIT2_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_EPIT2MCU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case EDIO_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_EDIO;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case WDOG_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_WDOGMCU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case PWM_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_PWMMCU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case I2C_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                reg |= MXC_CCM_MCGR0_I2C;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR0);
                break;
        case IPU_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_IPU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case MPEG4_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_MPEG4;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SIM1_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_SIM1;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SIM2_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_SIM2;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case HAC_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_HAC;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case GEM_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_GEM;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case USB_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_USBOTG;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case CSPI2_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_CSPI2;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SDHC2_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_SDHC2;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SDHC1_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_SDHC1;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SSI2_BAUD:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                reg |= MXC_CCM_MCGR1_SSI2;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR1);
                break;
        case SDMA_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                reg |= MXC_CCM_MCGR2_SDMA;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                break;
        case RNG_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                reg |= MXC_CCM_MCGR2_RNG;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                break;
        case KPP_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                reg |= MXC_CCM_MCGR2_KPP;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                break;
        case MU_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                reg |= MXC_CCM_MCGR2_MU;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                break;
        case SPBA_CLK:
                reg = readl(IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                reg |= MXC_CCM_MCGR2_SPBA;
                writel(reg, IO_ADDRESS(CRM_MCU_BASE_ADDR) + MXC_CCM_MCGR2);
                break;
        default:
                printk("The gateoff for this clock(%d) is not implemented\n", clk);
                break;
        }
        spin_unlock_irqrestore(&mxc_crm_lock, flags);
}

EXPORT_SYMBOL(mxc_pll_clock);
EXPORT_SYMBOL(mxc_get_clocks);
EXPORT_SYMBOL(mxc_set_clocks_pll);
EXPORT_SYMBOL(mxc_set_clocks_div);
EXPORT_SYMBOL(mxc_clks_disable);
EXPORT_SYMBOL(mxc_clks_enable);
