/*
 *  File Name       : linux/arch/arm/mach-mp200/lcdc_intr.c
 *  Function        : LCDC Interrupt (VSYNC)
 *  Release Version : Ver 1.01
 *  Release Date    : 2006/02/20
 *
 *  Copyright (C) NEC Electronics Corporation 2005-2006
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <asm/arch/lcdc_intr.h>
#include "lcdc_intr.h"

static void run_vsync_timer(unsigned long data);

static driver_private_t lcdc_intr_driver;

static DECLARE_TASKLET(vsync_timer_tasklet, run_vsync_timer, 0);

static int lcdc_intr_init_flag = 1;

/* callback entries */
static lcdc_vsync_run_action_t act_ent[ACT_SIZE];

void (*set_lcdc_intr_handler(unsigned int bit,
			     void (*func) (unsigned long))) (unsigned long) {
	driver_private_t *drv = &lcdc_intr_driver;

	void (*old) (unsigned long) = (void (*)(unsigned long))-1;

	/* lcdc_intr initialize error check */
	if (lcdc_intr_init_flag != 0) {
		printk(KERN_INFO
		       "set_lcdc_intr_handler(): lcdc_intr initialize error\n");
		return (void (*)(unsigned long))-1;
	}

	if (func == (void (*)(unsigned long))-1) {
		return func;
	}

	switch (bit) {
	case LCDC_DMAERR:
		old = drv->dmaerr;
		drv->dmaerr = func;
		if (drv->dmaerr != 0) {
			LCDC_INTENSET = LCDC_DMAERR;
		} else {
			LCDC_INTENCLR = LCDC_DMAERR;
		}
		break;
	case LCDC_DMASTOP:
		old = drv->dmastop;
		drv->dmastop = func;
		if (drv->dmastop != 0) {
			LCDC_INTENSET = LCDC_DMASTOP;
		} else {
			LCDC_INTENCLR = LCDC_DMASTOP;
		}
		break;
	case LCDC_UNDERRUN:
		old = drv->underrun;
		drv->underrun = func;
		if (drv->underrun != 0) {
			LCDC_INTENSET = LCDC_UNDERRUN;
		} else {
			LCDC_INTENCLR = LCDC_UNDERRUN;
		}
		break;
	case LCDC_LCDSTOP:
		old = drv->lcdstop;
		drv->lcdstop = func;
		if (drv->lcdstop != 0) {
			LCDC_INTENSET = LCDC_LCDSTOP;
		} else {
			LCDC_INTENCLR = LCDC_LCDSTOP;
		}
		break;
	default:
		/* Not support */
		printk(KERN_INFO "set_lcdc_intr_handler(): Not support Bit\n");
		return (void (*)(unsigned long))-1;
	}

	return old;
}

static lcdc_vsync_run_action_t *get_lcdc_intr_act(void)
{
	driver_private_t *drv = &lcdc_intr_driver;
	unsigned long flags;
	lcdc_vsync_run_action_t *ract = 0;

	spin_lock_irqsave(&drv->free_act_ent_lock, flags);
	if (!list_empty(&drv->free_act_ent)) {
		ract =
		    list_entry(drv->free_act_ent.next, lcdc_vsync_run_action_t,
			       list);
		list_del(&ract->list);
	}
	spin_unlock_irqrestore(&drv->free_act_ent_lock, flags);

	return ract;
}

static void put_lcdc_intr_act(lcdc_vsync_run_action_t * ract)
{
	driver_private_t *drv = &lcdc_intr_driver;
	unsigned long flags;

	spin_lock_irqsave(&drv->free_act_ent_lock, flags);
	list_add(&ract->list, &drv->free_act_ent);
	spin_unlock_irqrestore(&drv->free_act_ent_lock, flags);
}

static lcdc_vsync_run_action_t *find_act_core(void (*func) (unsigned long),
					      unsigned long data)
{
	driver_private_t *drv = &lcdc_intr_driver;
	struct list_head *pos;
	lcdc_vsync_run_action_t *ract = 0;

	list_for_each(pos, &drv->list) {
		lcdc_vsync_run_action_t *tmp;
		tmp = list_entry(pos, lcdc_vsync_run_action_t, list);
		if ((tmp->action.function == func)
		    && (tmp->action.data == data)) {
			ract = tmp;
			break;
		}
	}

	return ract;
}

int add_vsync_action(lcdc_vsync_action_t * act)
{
	driver_private_t *drv = &lcdc_intr_driver;
	lcdc_vsync_run_action_t *ract;
	unsigned long flags;
	int ret = 0;
	int ena = 0;

	/* lcdc_intr initialize error check */
	if (lcdc_intr_init_flag != 0) {
		printk(KERN_INFO
		       "add_vsync_action(): lcdc_intr initialize error\n");
		return -EPERM;
	}

	if (act->function == 0) {
		ret = -EFAULT;	/* Bad address */
		goto out;
	}

	spin_lock_irqsave(&drv->lock, flags);

	ract = find_act_core(act->function, act->data);

	if (ract) {
		ret = -EUSERS;	/* Too many users */
		goto end;
	}

	ract = get_lcdc_intr_act();
	if (ract == 0) {
		ret = -ENOMEM;	/* Out of memory */
		goto end;
	}

	if (list_empty(&drv->list) != 0) {
		/* drv->list is 0 (first time) */
		ena = 1;
	}

	ract->action.count = act->count;
	ract->action.reschedule = act->reschedule;
	ract->action.immediately = act->immediately;
	ract->action.high_priority = act->high_priority;
	ract->action.function = act->function;
	ract->action.data = act->data;

	ract->dec_cnt = act->count;
	ract->enable = 1;

	if (ract->action.high_priority != 0) {
		list_add(&ract->list, &drv->list);
	} else {
		list_add_tail(&ract->list, &drv->list);
	}

	if (drv->enable && ena == 1) {
		LCDC_INTENSET = LCDC_LCDVS;
	}

      end:
	spin_unlock_irqrestore(&drv->lock, flags);

      out:
	return ret;
}

int del_vsync_action(lcdc_vsync_action_t * act)
{
	driver_private_t *drv = &lcdc_intr_driver;
	lcdc_vsync_run_action_t *ract;
	unsigned long flags;

	/* lcdc_intr initialize error check */
	if (lcdc_intr_init_flag != 0) {
		printk(KERN_INFO
		       "del_vsync_action(): lcdc_intr initialize error\n");
		return -EPERM;
	}

	spin_lock_irqsave(&drv->lock, flags);

	ract = find_act_core(act->function, act->data);
	if (ract) {
		list_del(&ract->list);
		put_lcdc_intr_act(ract);
	}

	if (list_empty(&drv->list) != 0) {
		/* drv->list is empty */
		LCDC_INTENCLR = LCDC_LCDVS;
		LCDC_INTFFCLR = LCDC_LCDVS;
	}

	spin_unlock_irqrestore(&drv->lock, flags);

	return 0;
}

int mod_vsync_action(lcdc_vsync_action_t * act)
{
	driver_private_t *drv = &lcdc_intr_driver;
	lcdc_vsync_run_action_t *ract;
	unsigned long flags;
	int ret = 0;

	/* lcdc_intr initialize error check */
	if (lcdc_intr_init_flag != 0) {
		printk(KERN_INFO
		       "mod_vsync_action(): lcdc_intr initialize error\n");
		return -EPERM;
	}

	spin_lock_irqsave(&drv->lock, flags);

	ract = find_act_core(act->function, act->data);
	if (ract != 0) {
		/* action found */
		list_del(&ract->list);

		ract->action.count = act->count;
		ract->action.reschedule = act->reschedule;
		ract->action.immediately = act->immediately;
		ract->action.high_priority = act->high_priority;

		ract->dec_cnt = act->count;
		ract->enable = 1;

		if (ract->action.high_priority != 0) {
			list_add(&ract->list, &drv->list);
		} else {
			list_add_tail(&ract->list, &drv->list);
		}

		spin_unlock_irqrestore(&drv->lock, flags);
	} else {
		/* action not found */
		spin_unlock_irqrestore(&drv->lock, flags);
		ret = add_vsync_action(act);
	}

	return ret;
}

int enable_vsync_interrupt(void)
{
	driver_private_t *drv = &lcdc_intr_driver;
	unsigned long ret = LCDC_INTENSET & LCDC_LCDVS;
	unsigned long flags;

	/* lcdc_intr initialize error check */
	if (lcdc_intr_init_flag != 0) {
		printk(KERN_INFO
		       "enable_vsync_interrupt(): lcdc_intr initialize error\n");
		return -EPERM;
	}

	spin_lock_irqsave(&drv->lock, flags);

	drv->enable = 1;

	if (ret == 0) {
		LCDC_INTENSET = LCDC_LCDVS;
	}

	spin_unlock_irqrestore(&drv->lock, flags);

	return ((ret) ? 1 : 0);
}

int disable_vsync_interrupt(void)
{
	driver_private_t *drv = &lcdc_intr_driver;
	unsigned long ret = LCDC_INTENSET & LCDC_LCDVS;
	unsigned long flags;

	/* lcdc_intr initialize error check */
	if (lcdc_intr_init_flag != 0) {
		printk(KERN_INFO
		       "disable_vsync_interrupt(): lcdc_intr initialize error\n");
		return -EPERM;
	}

	spin_lock_irqsave(&drv->lock, flags);

	drv->enable = 0;

	if (ret != 0) {
		LCDC_INTENCLR = LCDC_LCDVS;
		LCDC_INTFFCLR = LCDC_LCDVS;
	}

	spin_unlock_irqrestore(&drv->lock, flags);

	return ((ret) ? 1 : 0);
}

int enable_vsync_action(lcdc_vsync_action_t * act)
{
	driver_private_t *drv = &lcdc_intr_driver;
	lcdc_vsync_run_action_t *ract;
	unsigned long flags;
	int ret = 0;

	/* lcdc_intr initialize error check */
	if (lcdc_intr_init_flag != 0) {
		printk(KERN_INFO
		       "disable_vsync_interrupt(): lcdc_intr initialize error\n");
		return -EPERM;
	}

	spin_lock_irqsave(&drv->lock, flags);

	ract = find_act_core(act->function, act->data);
	if (ract != 0) {
		ract->enable = 1;
	}

	spin_unlock_irqrestore(&drv->lock, flags);

	return ret;
}

int disable_vsync_action(lcdc_vsync_action_t * act)
{
	driver_private_t *drv = &lcdc_intr_driver;
	lcdc_vsync_run_action_t *ract;
	unsigned long flags;
	int ret = 0;

	/* lcdc_intr initialize error check */
	if (lcdc_intr_init_flag != 0) {
		printk(KERN_INFO
		       "disable_vsync_interrupt(): lcdc_intr initialize error\n");
		return -EPERM;
	}

	spin_lock_irqsave(&drv->lock, flags);

	ract = find_act_core(act->function, act->data);
	if (ract != 0) {
		ract->enable = 0;
	}

	spin_unlock_irqrestore(&drv->lock, flags);

	return ret;
}

static void run_vsync_action(void)
{
	driver_private_t *drv = &lcdc_intr_driver;
	unsigned long flags;
	lcdc_vsync_run_action_t *ract;
	struct list_head *pos, *n;
	void (*func) (unsigned long);
	unsigned long data;

	spin_lock_irqsave(&drv->lock, flags);

	list_for_each_safe(pos, n, &drv->list) {
		ract = list_entry(pos, lcdc_vsync_run_action_t, list);
		if (ract->enable == 1 &&
		    (ract->dec_cnt == 0 || --ract->dec_cnt == 0)) {
			/* count == 0 */
			if (ract->action.immediately != 0) {
				func = ract->action.function;
				data = ract->action.data;

				(*func) (data);

				if (ract->action.reschedule != 0) {
					ract->dec_cnt = ract->action.count;
				} else {
					/* delete from action list */
					list_del(&ract->list);
					/* return to free entry */
					put_lcdc_intr_act(ract);
				}
			} else {
				ract->enable = 0;
				/* delete from action list */
				list_del(&ract->list);
				/* add to drv->queue */
				list_add_tail(&ract->list, &drv->queue);
			}
		}
	}

	if (list_empty(&drv->queue) == 0) {
		/* drv->queue is NOT empty */
		tasklet_schedule(&vsync_timer_tasklet);
	} else if (list_empty(&drv->list) != 0) {
		/* drv->list and drv->queue are empty */
		LCDC_INTENCLR = LCDC_LCDVS;
		LCDC_INTFFCLR = LCDC_LCDVS;
	}

	spin_unlock_irqrestore(&drv->lock, flags);
}

static void run_vsync_timer(unsigned long dummy)
{
	driver_private_t *drv = &lcdc_intr_driver;
	unsigned long flags;
	struct list_head *pos;
	lcdc_vsync_run_action_t *ract;
	void (*func) (unsigned long);
	unsigned long data;

	spin_lock_irqsave(&drv->lock, flags);

	while (list_empty(&drv->queue) == 0) {
		pos = drv->queue.next;
		ract = list_entry(pos, lcdc_vsync_run_action_t, list);
		list_del(&ract->list);
		if (ract->action.reschedule != 0) {
			/* reschedule */
			ract->dec_cnt = ract->action.count;
			ract->enable = 1;
			list_add_tail(&ract->list, &drv->list);
		} else {
			put_lcdc_intr_act(ract);
		}

		func = ract->action.function;
		data = ract->action.data;

		spin_unlock_irqrestore(&drv->lock, flags);

		(*func) (data);

		spin_lock_irqsave(&drv->lock, flags);
	}

	if (list_empty(&drv->list) != 0) {
		/* drv->list is empty */
		LCDC_INTENCLR = LCDC_LCDVS;
		LCDC_INTFFCLR = LCDC_LCDVS;
	}

	spin_unlock_irqrestore(&drv->lock, flags);
}

static irqreturn_t lcdc_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	driver_private_t *drv = &lcdc_intr_driver;
	unsigned int lcdc_intst = LCDC_INTSTATUS;

	if (lcdc_intst & LCDC_LCDVS) {
		LCDC_INTENCLR = LCDC_LCDVS;
		LCDC_INTFFCLR = LCDC_LCDVS;
		run_vsync_action();
		if (list_empty(&drv->list) == 0 || list_empty(&drv->queue) == 0) {
			LCDC_INTENSET = LCDC_LCDVS;
		}
	}

	if (lcdc_intst & LCDC_DMAERR) {
		LCDC_INTENCLR = LCDC_DMAERR;
		if (drv->dmaerr != 0) {
			(*drv->dmaerr) (LCDC_DMAERR);
		}
		LCDC_INTFFCLR = LCDC_DMAERR;
		LCDC_INTENSET = LCDC_DMAERR;
	}

	if (lcdc_intst & LCDC_DMASTOP) {
		LCDC_INTENCLR = LCDC_DMASTOP;
		if (drv->dmastop != 0) {
			(*drv->dmastop) (LCDC_DMASTOP);
		}
		LCDC_INTFFCLR = LCDC_DMASTOP;
		LCDC_INTENSET = LCDC_DMASTOP;
	}

	if (lcdc_intst & LCDC_UNDERRUN) {
		LCDC_INTENCLR = LCDC_UNDERRUN;
		if (drv->underrun != 0) {
			(*drv->underrun) (LCDC_UNDERRUN);
		}
		LCDC_INTFFCLR = LCDC_UNDERRUN;
		LCDC_INTENSET = LCDC_UNDERRUN;
	}

	if (lcdc_intst & LCDC_LCDSTOP) {
		LCDC_INTENCLR = LCDC_LCDSTOP;
		if (drv->lcdstop != 0) {
			(*drv->lcdstop) (LCDC_LCDSTOP);
		}
		LCDC_INTFFCLR = LCDC_LCDSTOP;
		LCDC_INTENSET = LCDC_LCDSTOP;
	}

	return IRQ_HANDLED;
}

static int __init lcdc_intr_init(void)
{
	driver_private_t *drv = &lcdc_intr_driver;
	int ret = 0;
	int i;

	printk(KERN_INFO "Starting lcdc_intr.\n");

	/* initialize driver struct */
	drv->lock = SPIN_LOCK_UNLOCKED;
	drv->free_act_ent_lock = SPIN_LOCK_UNLOCKED;
	INIT_LIST_HEAD(&drv->list);
	INIT_LIST_HEAD(&drv->queue);
	INIT_LIST_HEAD(&drv->free_act_ent);
	drv->enable = 1;
	drv->dmaerr = 0;
	drv->dmastop = 0;
	drv->underrun = 0;
	drv->lcdstop = 0;

	/* initialize free callback entry elements */
	for (i = 0; i < sizeof(act_ent) / sizeof(act_ent[0]); i++) {
		list_add(&act_ent[i].list, &drv->free_act_ent);
	}

	/* initialize interrupts */
	ret = request_irq(INT_LCD, lcdc_interrupt,
			  SA_INTERRUPT, "lcdc_intr", drv);
	if (ret != 0) {
		printk(KERN_INFO "lcdc_intr_init(): "
		       "lcdc_intr interrupts initial error %d\n", ret);
		return ret;
	}

	LCDC_INTENCLR = LCDC_INT_MASK;
	LCDC_INTFFCLR = LCDC_INT_MASK;

	lcdc_intr_init_flag = 0;

	return 0;
}

static void __exit lcdc_intr_exit(void)
{
	driver_private_t *drv = &lcdc_intr_driver;

	LCDC_INTENCLR = LCDC_INT_MASK;
	LCDC_INTFFCLR = LCDC_INT_MASK;

	free_irq(INT_LCD, drv);
}

module_init(lcdc_intr_init);
module_exit(lcdc_intr_exit);

EXPORT_SYMBOL(set_lcdc_intr_handler);
EXPORT_SYMBOL(add_vsync_action);
EXPORT_SYMBOL(del_vsync_action);
EXPORT_SYMBOL(mod_vsync_action);
EXPORT_SYMBOL(enable_vsync_interrupt);
EXPORT_SYMBOL(disable_vsync_interrupt);
EXPORT_SYMBOL(enable_vsync_action);
EXPORT_SYMBOL(disable_vsync_action);
