/*
 *  linux/arch/arm/mach-scma11/gpio_common.c
 *
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/module.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/setup.h>
#include <asm/mach-types.h>
#include <asm/mach/arch.h>

#include "gpio_priv.h"
#include <asm/arch/gpio.h>

/*!
 * @file gpio_common.c
 * 
 * @brief This file contains the GPIO and IOMUX implementation details. 
 * 
 * @ingroup GPIO
 */

static DEFINE_RAW_SPINLOCK(gpio_mux_lock);
static DEFINE_RAW_SPINLOCK(gpio_lock);
static DEFINE_RAW_SPINLOCK(edio_lock);
static DEFINE_RAW_SPINLOCK(gpio_request_lock);
static bool gpio_init_ok = false;

/*!
 * This is used to identify the different GPIO interrupt signals.
 */
const static int GPIO_PORT_ARRAY[GPIO_PORT_NUM] = {
        INT_GPIO1, 
        INT_GPIO2,
        INT_GPIO3,
        INT_GPIO4,
};

/*!
 * This is used to provide a name for each GPIO port in character string.
 */
const static char* GPIO_PORT_STR[GPIO_PORT_NUM] = {
        "GPIO1", 
        "GPIO2",
        "GPIO3",
        "GPIO4",
};

/*!
 * the base addresses for the GPIO modules
 */
static volatile __u32 *gpio_base[GPIO_PORT_NUM] = {
        (volatile __u32 *) IO_ADDRESS(GPIO1_AP_BASE_ADDR), 
        (volatile __u32 *) IO_ADDRESS(GPIO2_AP_BASE_ADDR),
        (volatile __u32 *) IO_ADDRESS(GPIO3_AP_BASE_ADDR),
        (volatile __u32 *) IO_ADDRESS(GPIO4_SH_BASE_ADDR),
};

/*!
 * the base addresses for the EDIO modules
 */
static volatile __u16 *edio_base =  (volatile __u16 *) IO_ADDRESS(EDIO_BASE_ADDR);

/*!
 * This function is used to configure a pin through the IOMUX module.
 *
 * @param  pin          a pin number as defined in \b enum \b iomux_pins
 * @param  out          an output function as defined in \b enum \b iomux_output_config
 * @param  in           an input function as defined in \b enum \b iomux_input_config 
 */
void iomux_config_mux(enum iomux_pins pin, enum iomux_output_config out, enum iomux_input_config in)
{
        unsigned long gpio_mux_flags;
        volatile unsigned int *base;
        int mux_index;
        int mux_field;
         
        if (pin <= AP_MAX_PIN) {
                base = (volatile unsigned int *)(IOMUXSW_AP_CTL);
        } else {
                base = (volatile unsigned int *)(IOMUXSW_SP_CTL);
                pin -= (AP_MAX_PIN + 1);
        }

        mux_index = pin / 4;
        mux_field = pin % 4;

        spin_lock_irqsave(&gpio_mux_lock, gpio_mux_flags);
        base[mux_index] = (base[mux_index] & (~(0xFF << (mux_field * MUX_CTL_BIT_LEN)))) | 
                          (((out << 4) | in) << (mux_field * MUX_CTL_BIT_LEN)); 
        spin_unlock_irqrestore(&gpio_mux_lock, gpio_mux_flags);
}

/*!
 * This function returns the pad value for a IOMUX group. 
 *
 * @param  grp          a group number as defined in \b enum \b #iopad_group
 *
 * @return A group pad value
 */
__u16 iomux_get_pad(enum iopad_group grp)
{
        volatile unsigned int *base = (volatile unsigned int *)(IOMUXSW_PAD_CTL);
        int pad_index = grp / 2;
        int pad_field = grp % 2;

        return (__u16) (base[pad_index] >> (pad_field * MUX_PAD_BIT_LEN));
}

/*!
 * This function configures the pad value for a IOMUX pin. 
 *
 * @param  grp          a pin number as defined in \b enum \b #iopad_group
 * @param  config       the ORed value of elements defined in \b enum \b #iomux_pad_config
 */
void iomux_set_pad(enum iopad_group grp, __u32 config)
{
        unsigned long gpio_mux_flags;
        volatile unsigned int *base = (volatile unsigned int *)(IOMUXSW_PAD_CTL);
        int pad_index = grp / 2;
        int pad_field = grp % 2;

        spin_lock_irqsave(&gpio_mux_lock, gpio_mux_flags);
        base[pad_index] = (base[pad_index] & (~(0xFFFF << (pad_field * MUX_PAD_BIT_LEN)))) | 
                          (config << (pad_field * MUX_PAD_BIT_LEN)); 
        spin_unlock_irqrestore(&gpio_mux_lock, gpio_mux_flags);
}

/*!
 * This function configures the GPIO signal to be either input or output. For 
 * input signals used for generating interrupts for the ARM core, how the 
 * interrupts being triggered is also passed in via \a icr. For output signals, 
 * the \a icr value doesn't matter.
 *
 * @param  port         specified port with 0 for GPIO port 1 and 1 for GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  out          #true for output, #false for input
 * @param  icr          one of the values defined in \b #gpio_int_cfg to indicate how to interrupt the ARM core 
 */
void gpio_config(__u32 port, __u32 sig_no, bool out, enum gpio_int_cfg icr)
{
        unsigned long gpio_flags;
        volatile __u32 temp, reg, sig;

        MXC_ERR_CHK((port >= GPIO_PORT_NUM) || (sig_no >= GPIO_NUM_PIN));

        spin_lock_irqsave(&gpio_lock, gpio_flags);
        if (out) {
                gpio_base[port][GPIO_GDIR >> 2] |= 1 << sig_no; 
        } else {
                gpio_base[port][GPIO_GDIR >> 2] &= ~(1 << sig_no);
                if (icr != GPIO_INT_NONE) {
                        reg = (sig_no <= 15) ? GPIO_ICR1 : GPIO_ICR2;
                        sig = (sig_no <= 15) ? sig_no : (sig_no - 16);

                        temp = gpio_base[port][reg >> 2];
                        temp = (temp & (~(0x3 << (sig * 2)))) | (icr << (sig * 2));
                        gpio_base[port][reg >> 2] = temp;
                }
        }
        spin_unlock_irqrestore(&gpio_lock, gpio_flags);
}

/*!
 * This function sets a GPIO signal value.
 *
 * @param  port         specified port with 0 for GPIO port 1 and 1 for GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  data         value to be set (only 0 or 1 is valid) 
 */
void gpio_set_data(__u32 port, __u32 sig_no, __u32 data)
{
        unsigned long gpio_flags;

        MXC_ERR_CHK((port >= GPIO_PORT_NUM) || (sig_no >= GPIO_NUM_PIN) || (data > 1));

        spin_lock_irqsave(&gpio_lock, gpio_flags);
        gpio_base[port][GPIO_DR >> 2] = (gpio_base[port][GPIO_DR >> 2] & (~(1 << sig_no))) 
                                            | (data << sig_no);
        spin_unlock_irqrestore(&gpio_lock, gpio_flags);
}

/*!
 * This function returns the value of the GPIO signal.
 *
 * @param  port         specified port with 0 for GPIO port 1 and 1 for GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 *
 * @return Value of the GPIO signal
 */
__u32 gpio_get_data(__u32 port, __u32 sig_no)
{
        MXC_ERR_CHK((port >= GPIO_PORT_NUM) || (sig_no >= GPIO_NUM_PIN));
        return ((gpio_base[port][GPIO_DR >> 2] >> sig_no) & 1);
}

/*!
 * This function returns the value of the GPIO signal pad value.
 *
 * @param  port         specified port with 0 for GPIO port 1 and 1 for GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 *
 * @return Pad value of the GPIO signal 
 */
__u32 gpio_get_pad(__u32 port, __u32 sig_no)
{
        MXC_ERR_CHK((port >= GPIO_PORT_NUM) || (sig_no >= GPIO_NUM_PIN));
        return ((gpio_base[port][GPIO_PSR >> 2] >> sig_no) & 1);
}

/*!
 * This function enables/disables a GPIO signal's interrupt.
 *
 * @param  port         specified port with 0 for GPIO port 1 and 1 for GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  en           \b #true for enabling the interrupt; \b #false for disabling the interrupt 
 */
void gpio_config_int_en(__u32 port, __u32 sig_no, bool en)
{
        unsigned long gpio_flags;
        __u32 mask = (!en) ? 0: 1;
        
        MXC_ERR_CHK((port >= GPIO_PORT_NUM) || (sig_no >= GPIO_NUM_PIN));

        spin_lock_irqsave(&gpio_lock, gpio_flags);
        gpio_base[port][GPIO_IMR >> 2] = (gpio_base[port][GPIO_IMR >> 2] & (~(1 << sig_no))) 
                                            | (mask << sig_no);
        spin_unlock_irqrestore(&gpio_lock, gpio_flags);
}

/*!
 * This function checks to see if the GPIO interrupt is enabled for a particular signal. 
 *
 * @param  port         specified port with 0 for GPIO port 1 and 1 for GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 *
 * @return \b #true if the interrupt is enabled for the GPIO signal; \b #false otherwise.
 */
bool gpio_int_enabled(__u32 port, __u32 sig_no)
{
        MXC_ERR_CHK((port >= GPIO_PORT_NUM) || (sig_no >= GPIO_NUM_PIN));

        if ((gpio_base[port][GPIO_ISR >> 2] >> sig_no) & 1) {
                return true;
        }
        return false;
}

/*!
 * This function clears the GPIO interrupt for a signal on a port.
 *
 * @param  port         specified port with 0 for GPIO port 1 and 1 for GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based) to clear
 */
void gpio_clear_int(__u32 port, __u32 sig_no)
{
        MXC_ERR_CHK((port >= GPIO_PORT_NUM) || (sig_no >= GPIO_NUM_PIN));

        gpio_base[port][GPIO_ISR >> 2] = 1 << sig_no;
}

/********************* EDIO related functions *********************/

/*!
 * This function configures the EDIO signal to be either input or output. For
 * input signals used for generating interrupts for the ARM core, how the
 * interrupts being triggered is also passed in via \a icr. For output signals
 * the \a icr value doesn't matter.
 *
 * @param  sig_no       EDIO signal (0 based) as defined in \b #edio_sig_no
 * @param  out          #true for output, #false for input
 * @param  icr          value defined in \b #edio_int_cfg
 */
void edio_config(enum edio_sig_no sig_no, bool out, enum edio_int_cfg icr)
{
        unsigned long edio_flags;
        volatile __u32 temp;

        MXC_ERR_CHK(sig_no > EDIO_MAX_SIG_NO);

        spin_lock_irqsave(&edio_lock, edio_flags);
        if (out) {
                edio_base[EDIO_EPDDR >> 1] |= 1 << sig_no; 
        } else {
                edio_base[EDIO_EPDDR >> 1] &= ~(1 << sig_no);
                if (icr != EDIO_INT_NONE) {
                        temp = edio_base[EDIO_EPPAR >> 1];
                        temp = (temp & (~(0x3 << (sig_no * 2)))) | (icr << (sig_no * 2));
                        edio_base[EDIO_EPPAR >> 1] = temp;
                }
        }
        spin_unlock_irqrestore(&edio_lock, edio_flags);
}

/*!
 * This function sets a EDIO signal value.
 *
 * @param  sig_no       EDIO signal (0 based) as defined in \b #edio_sig_no
 * @param  data         value to be set (only 0 or 1 is valid)
 */
void edio_set_data(enum edio_sig_no sig_no, __u32 data)
{
        unsigned long edio_flags;
        
        MXC_ERR_CHK((data > 1) || (sig_no > EDIO_MAX_SIG_NO));
        
        spin_lock_irqsave(&edio_lock, edio_flags);
        edio_base[EDIO_EPDR >> 1] = (edio_base[EDIO_EPDR >> 1] & (~(1 << sig_no))) 
                                            | (data << sig_no);
        spin_unlock_irqrestore(&edio_lock, edio_flags);
}

/*!
 * This function returns the value of the EDIO signal.
 *
 * @param  sig_no       EDIO signal (0 based) as defined in \b #edio_sig_no
 *
 * @return Value of the EDIO signal (either 0 or 1)
 */
__u32 edio_get_data(enum edio_sig_no sig_no)
{
        MXC_ERR_CHK(sig_no > EDIO_MAX_SIG_NO);
        return ((edio_base[EDIO_EPDR >> 1] >> sig_no) & 1);
}

/*!
 * This function returns the EDIO edge-trigged interrupt for a signal.
 *
 * @param  sig_no       EDIO signal (0 based) as defined in \b #edio_sig_no 
 */
__u32 edio_get_int(enum edio_sig_no sig_no)
{
        MXC_ERR_CHK(sig_no > EDIO_MAX_SIG_NO);
        return ((edio_base[EDIO_EPFR >> 1] >> sig_no) & 1);
}

/*!
 * This function clears the EDIO edge-trigged interrupt for a signal.
 *
 * @param  sig_no       EDIO signal (0 based) as defined in \b #edio_sig_no
 */
void edio_clear_int(enum edio_sig_no sig_no)
{
        MXC_ERR_CHK(sig_no > EDIO_MAX_SIG_NO);
        edio_base[EDIO_EPFR >> 1] = 1 << sig_no;
}

/********************** GPIO Interrupt related functions ************************/

/*!
 * (#GPIO_PORT_NUM x #GPIO_INT_PRIO_NUM) interrupt queues for the GPIO modules
 */
static t_gpio_int_queue gpio_int_q[GPIO_PORT_NUM][GPIO_INT_PRIO_NUM];

/*!
 * This table is used to indicate a registered ISR for a signal. 
 * -1 indicates no ISR has been registered to that signal.
 * A positive value (the actual signal number) indicates an 
 * registered ISR already.
 */
static  signed char register_status[GPIO_PORT_NUM][GPIO_NUM_PIN];

/*!
 * This function is responsible for registering a GPIO signal's ISR. 
 *
 * @param  port         specified port with 0 for GPIO port 1 and 1 for GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  prio         priority as defined in \b enum \b #gpio_prio
 * @param  handler      GPIO ISR function pointer for the GPIO signal
 * @param  irq_flags    irq flags (not used)
 * @param  devname      device name associated with the interrupt
 * @param  dev_id       some unique information for the ISR
 *
 * @return 0 if successful; non-zero otherwise. 
 */
int gpio_request_irq(__u32 port, __u32 sig_no, enum gpio_prio prio,
                     gpio_irq_handler handler, __u32 irq_flags, 
                     const char * devname, void *dev_id)
{
        unsigned long gpio_flags;
        __u32 i;

        MXC_ERR_CHK((port >= GPIO_PORT_NUM) || (sig_no >= GPIO_NUM_PIN));

        if (!gpio_init_ok) {
                printk("Can't bind irq. gpio_init() failed\n");
                return -100;     /* some random number */
        }
        spin_lock_irqsave(&gpio_request_lock, gpio_flags);
        if (register_status[port][sig_no] != -1) {
                spin_unlock_irqrestore(&gpio_request_lock, gpio_flags);
                printk("gpio_request_irq: port-%d, signal-%d already registered\n", port, sig_no);
                return -1;      /* already register */
        }
        
        i = gpio_int_q[port][prio].total_num;
        
        if (i == GPIO_NUM_PIN) {
                printk("gpio_request_irq: port-%d, signal-%d failed\n", port, sig_no);
                BUG(); /* oops */
        } else {
                register_status[port][sig_no] = i;
        }
        /* always add the new handler at the end of the queue since it is compact already,
         * i.e., no "holes" inside the queue */
        gpio_int_q[port][prio].node[i].handler = handler;
        gpio_int_q[port][prio].node[i].sig_num = sig_no;
        gpio_int_q[port][prio].node[i].flags = irq_flags;
        gpio_int_q[port][prio].node[i].dev_id = dev_id;
        gpio_int_q[port][prio].node[i].name = devname;
        
        gpio_int_q[port][prio].total_num ++;

        spin_unlock_irqrestore(&gpio_request_lock, gpio_flags);
        
        gpio_config_int_en(port, sig_no, true); /* enable the interrupt */

        return 0;
}

/*!
 * This function un-registers an ISR with the GPIO interrupt module.
 * 
 * @param  port         specified port with 0 for GPIO port 1 and 1 for GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  prio         priority as defined in \b enum \b #gpio_prio
 */
void gpio_free_irq(__u32 port, __u32 sig_no, enum gpio_prio prio)
{
        unsigned long gpio_flags;
        __u32 i, num;

        MXC_ERR_CHK((port >= GPIO_PORT_NUM) || (sig_no >= GPIO_NUM_PIN));

        spin_lock_irqsave(&gpio_request_lock, gpio_flags);
        if (register_status[port][sig_no] == -1) {
                spin_unlock_irqrestore(&gpio_request_lock, gpio_flags);
                printk("gpio_free_irq: port-%d, signal-%d not registered yet\n", port, sig_no);
                return;
        }
        
        gpio_config_int_en(port, sig_no, false); /* disable the signal interrupt */

        register_status[port][sig_no] = -1;
        num = gpio_int_q[port][prio].total_num;

        for (i = 0; i < GPIO_NUM_PIN; i++) {
                if (gpio_int_q[port][prio].node[i].sig_num == sig_no) {
                        if (i != (num - 1)) {
                                /* the deleted entry is in the middle of the table. Compact it.
                                 * Always use last entry to fill in the hole */
                                gpio_int_q[port][prio].node[i] = gpio_int_q[port][prio].node[num-1];
                        }
                        break;
                }
        }
        /* always mark the last entry as "free" */
        gpio_int_q[port][prio].node[num-1].handler = 0;
        gpio_int_q[port][prio].node[num-1].sig_num = -1;  
        gpio_int_q[port][prio].total_num --;
        
        spin_unlock_irqrestore(&gpio_request_lock, gpio_flags);
        if (i == GPIO_NUM_PIN) {
                BUG(); /* oops */
        }
}

/*!
 * There are not many GPIO signals used for interrupting the CPU normally. 
 * So instead of looping all the GPIO Interrupt Status register bits to find 
 * which GPIO interrupt is pending, we only check the valid service routines 
 * inside each queue. For each registered interrupt service routine, if the 
 * corresponding interrupt status register bit does show an interrupt occurred, 
 * the service routine gets called and the bit in status register is cleared. 
 * If for some reason after all the registered interrupt service routines are 
 * checked and the GPIO interrupt status register is still not zero, a spurious 
 * interrupt must have occurred.
 *
 * This function is the ISR for the two GPIO ports. Note that this ISR is shared 
 * by all the GPIO port interrupts. The proper interrupt is handled based on the 
 * passed-in \a irq number.
 *
 * @param  irq                  GPIO port IRQ number
 * @param  dev_id               device ID which is not used
 * @param  regs                 pointer to a structure containing the processor registers and state prior to
 *                              servicing the interrupt. It is not used in this function.
 *
 * @return Always \b IRQ_HANDLED as defined in the include/linux/interrupt.h file. 
 */
static irqreturn_t gpio_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
        __u32 i, j, k, port = -1;
        volatile __u32 int_status;

        for (i = 0; i < GPIO_PORT_NUM; i++) {
                if (irq == GPIO_PORT_ARRAY[i]) {
                        port = i;
                        break;
                }
        }
                
        if (-1 == port) {
                printk("Invalid GPIO interrupt with irq=%d\n", irq); 
                BUG(); /* oops */
        }
        
        /* get the valid interrupt status by ANDing with the IMR */
        int_status = gpio_base[port][GPIO_ISR >> 2] & gpio_base[port][GPIO_IMR >> 2];

        /* the first for loop start with 0 - the high priority queue */
        for (i = 0, j = 0; i < GPIO_INT_PRIO_NUM; i++) {
                for (k = 0; k < GPIO_NUM_PIN; k++) {
                        if (j == gpio_int_q[port][i].total_num) {
                                break;
                        }
                        if (gpio_int_q[port][i].node[k].sig_num != -1) { /* valid ISR */
                                j ++;
                                if (int_status & (1 << gpio_int_q[port][i].node[k].sig_num)) {
                                        gpio_int_q[port][i].node[k].handler(irq, gpio_int_q[port][i].node[k].dev_id, regs);
                                        /* clear interrupt status bit in the GPIO ISR reg */
                                        gpio_clear_int(port, gpio_int_q[port][i].node[k].sig_num);
                                        /* clear the bit in the int_status variable */
                                        int_status &= ~(1 << gpio_int_q[port][i].node[k].sig_num);
                                }
                        }
                }
        }
        if (int_status != 0) {
                printk("\nGPIO port(0-based): %d Spurious interrupt: 0x%0x\n\n",
                        port, int_status); 
                BUG(); /* oops */
        }
        
        return IRQ_HANDLED;
}

/*!
 * This function is called once the GPIO interrupt module is loaded during the kernel 
 * startup. 
 *
 * @return 0 if successful; non-zero otherwise 
 */
static int __init gpio_init(void)
{
        int i, j, k, ret;

        /* initialize the two queues */
        for (i = 0; i < GPIO_PORT_NUM; i++) {
                for (j = 0; j < GPIO_INT_PRIO_NUM; j++) {
                        gpio_int_q[i][j].total_num = 0;
                        for (k = 0; k < GPIO_NUM_PIN; k++) {
                                gpio_int_q[i][j].node[k].handler = 0;
                                gpio_int_q[i][j].node[k].sig_num = -1;
                                /* initialized twice. but who cares. no ISR registered yet */
                                register_status[i][k] = -1;
                        }
                }
        }

        for (i = 0; i < GPIO_PORT_NUM; i++) {
                ret = request_irq(GPIO_PORT_ARRAY[i], gpio_interrupt, 0, GPIO_PORT_STR[i], NULL);
                if (ret) {
                        printk(KERN_ERR "%s: IRQ%d already in use.\n", GPIO_PORT_STR[i], GPIO_PORT_ARRAY[i]);
                        return ret;
                }
        }

        /* disable the interrupt and clear the status */
        for (i = 0; i < GPIO_PORT_NUM; i++) {
                for (j = 0; j < 32; j++) {
                        gpio_config_int_en(i, j, false);
                        gpio_clear_int(i, j);
                }
        }
        gpio_init_ok = true;

        return 0;
}

/*!
 * This function is called once the GPIO interrupt module is unloaded. This function 
 * should never be called as it is statically linked into the kernel.
 */
static void __exit gpio_exit(void)
{
        __u32 i;
        for (i = 0; i < GPIO_PORT_NUM; i++) {
                free_irq(GPIO_PORT_ARRAY[i], NULL);
        }
        gpio_init_ok = false;
}

module_init(gpio_init);
module_exit(gpio_exit);

EXPORT_SYMBOL(iomux_config_mux);
EXPORT_SYMBOL(iomux_get_pad);
EXPORT_SYMBOL(iomux_set_pad);
EXPORT_SYMBOL(gpio_config);
EXPORT_SYMBOL(gpio_set_data);
EXPORT_SYMBOL(gpio_get_data);
EXPORT_SYMBOL(gpio_get_pad);
EXPORT_SYMBOL(gpio_config_int_en);
EXPORT_SYMBOL(gpio_int_enabled);
EXPORT_SYMBOL(gpio_clear_int);
EXPORT_SYMBOL(gpio_request_irq);
EXPORT_SYMBOL(gpio_free_irq);
EXPORT_SYMBOL(edio_config);
EXPORT_SYMBOL(edio_set_data);
EXPORT_SYMBOL(edio_get_data);
EXPORT_SYMBOL(edio_clear_int);
EXPORT_SYMBOL(edio_get_int);

MODULE_AUTHOR("Freescale Semiconductor");
MODULE_DESCRIPTION("GPIO driver");
MODULE_LICENSE("GPL");
