/*
 * linux/arch/arm/mach-zeus/gpio_priv.h
 *
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef __GPIO_PRIV_H
#define __GPIO_PRIV_H

#include <asm/sizes.h>
#include <asm/arch/platform.h>
#include <linux/interrupt.h>

/*!
 * @file gpio_priv.h
 * 
 * @brief This file contains the GPIO and IOMUX defines that are only intended
 * to be used internally by the GPIO module. 
 * 
 * @ingroup GPIO
 */

/*!
 * 8 bits for each MUX control field
 */                              
#define MUX_CTL_BIT_LEN         8

/*!
 * PAD control field
 */
#define MUX_PAD_BIT_LEN         16 

/*!
 * Zeus IOMUX register (base) addresses 
 */
enum iomux_reg_addr {
        AP_SW_MUX_CTL  = IO_ADDRESS(IOMUX_AP_BASE_ADDR)  + 0x000, /*!< MUX control for AP side */
        SP_SW_MUX_CTL  = IO_ADDRESS(IOMUX_COM_BASE_ADDR) + 0x000, /*!< MUX control for SP side */
        AP_SW_PAD_CTL = IO_ADDRESS(IOMUX_AP_BASE_ADDR) + 0x200, /*!< Pad control for AP side */
        SP_SW_PAD_CTL = IO_ADDRESS(IOMUX_COM_BASE_ADDR) + 0x200, /*!< Pad control for SP side */
        AP_INT_OBS1   = IO_ADDRESS(IOMUX_AP_BASE_ADDR)  + 0x600, /*!< Observe interrupts 0-3 */
        AP_INT_OBS2   = IO_ADDRESS(IOMUX_AP_BASE_ADDR)  + 0x604, /*!< Observe interrupt 4 */
};

// IOMUX_SW_MUX_CTL_REG converts from a pin number to an index for the SW_MUX_CTL
// register.  Zeus IOMUX packs 4 pins per SW_MUX_CTL register, so conversion is
// simple divide by 4.
#define IOMUX_SW_MUX_CTL_REG(pin)              ((pin) >> 2)

// IOMUX_SW_MUX_CTL_SHIFT converts from a pin number to a shift value used
// to reference the SW_MUX_CTL bits assigned to that pin.  Zeus
// IOMUX packs 4 pins per SW_MUX_CTL register (use % 4 to select pin for
// single register) and 8 bits per pin (use multiply by 8)
#define IOMUX_SW_MUX_CTL_SHIFT(pin)            (((pin) % 4) << 3)

// IOMUX_SW_MUX_CTL_MASK converts from a pin number to a mask used to isolate
// the SW_MUX_CTL register bits assigned to that pin.
#define IOMUX_SW_MUX_CTL_MASK(pin)  \
    (0x7F << IOMUX_SW_MUX_CTL_SHIFT(pin))

#define IOMUX_PIN_ASSIGN(pIOMUX, pin, in, out) \
    pIOMUX[IOMUX_SW_MUX_CTL_REG(pin)] &= (~(IOMUX_SW_MUX_CTL_MASK(pin))); \
    pIOMUX[IOMUX_SW_MUX_CTL_REG(pin)] |= ((in | out) << IOMUX_SW_MUX_CTL_SHIFT(pin));

/* GPIO related defines */

/*!
 * Number of GPIO port as defined in the IC Spec
 */
#define GPIO_PORT_NUM           3

/*!
 * Number of GPIO pins per port
 */
#define GPIO_NUM_PIN            32

/*!
 * Number of GPIO priorities. Two priorities are supported: high and low. If both
 * high and low priority GPIO interrupts are pending, the high priority one will be
 * serviced first.
 */
#define GPIO_INT_PRIO_NUM       2

/*!
 * This type definition is used to define a node in the GPIO interrupt queue for 
 * registered interrupts for GPIO pins. Each node contains the GPIO signal number 
 * associated with the ISR and the actual ISR function pointer.
 */
typedef struct gpio_sig_node {
        signed int sig_num;             /*!< the gpio signal number associated with ISR */
        irqreturn_t (*handler)(int, void *, struct pt_regs *); /*!< the ISR */
        const char *name;               /*!< device associated with the interrupt */
        void *dev_id;                   /*!< some unique information for the ISR */
        __u32 flags;                    /*!< not used */
}t_gpio_sig_node;

/*!
 * This type definition is used to define an interrupt queue for each priority. 
 * It contains data elements for the total number of registered ISRs and the 32 
 * GPIO signal nodes.
 */
typedef struct gpio_int_queue{
        __u32 total_num;                        /*!< total number of registered ISRs */
        t_gpio_sig_node node[GPIO_NUM_PIN];     /*!< GPIO node associated with a queue */
}t_gpio_int_queue;

enum gpio_reg {
        GPIO_DR         = 0x00,
        GPIO_GDIR       = 0x04,
        GPIO_PSR        = 0x08,
        GPIO_ICR1       = 0x0C,
        GPIO_ICR2       = 0x10,
        GPIO_IMR        = 0x14,
        GPIO_ISR        = 0x18,
};

enum edio_reg {
        EDIO_EPPAR      = 0x00,
        EDIO_EPDDR      = 0x02,
        EDIO_EPDR       = 0x04,
        EDIO_EPFR       = 0x06,
};

#define EDIO_MAX_SIG_NO         7       /* only 0-7 are valid */

#endif /* __GPIO_PRIV_H */
