/*
 * 2003 (c) MontaVista Software, Inc.  This file is licensed under the
 * terms of the GNU General Public License version 2. This program is
 * licensed "as is" without any warranty of any kind, whether express
 * or implied.
 */

#include <linux/config.h>
#include <linux/dpm.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/interrupt.h>

#include <asm/page.h>
#include <asm/uaccess.h>

struct dpm_bd dpm_bd;

static int
dpm_x86_init_opt(struct dpm_opt *opt)
{
	return dpm_bd.init_opt ? dpm_bd.init_opt(opt) : -1;
}

/* Fully determine the current machine-dependent operating point, and fill in a
   structure presented by the caller. This should only be called when the
   dpm_sem is held. This call can return an error if the system is currently at
   an operating point that could not be constructed by dpm_md_init_opt(). */

static int
dpm_x86_get_opt(struct dpm_opt *opt)
{
	return dpm_bd.get_opt ? dpm_bd.get_opt(opt) : -1;
}

int
dpm_x86_set_opt(struct dpm_opt *cur, struct dpm_opt *new)
{
	return dpm_bd.set_opt ? dpm_bd.set_opt(&new->md_opt) : -1;
}

/****************************************************************************
 * Machine-dependent /proc/driver/dpm/md entries
 ****************************************************************************/

static inline int 
p5d(char *buf, unsigned mhz)
{
	return sprintf(buf, "%5d", mhz ); /* Round */
}

static int
dpm_proc_print_opt(char *buf, struct dpm_opt *opt)
{
        int len = 0;
        struct dpm_md_opt *md_opt = &opt->md_opt;

        len += sprintf(buf + len, "%12s %9llu", 
                       opt->name, opt->stats.count);
        len += p5d(buf + len, md_opt->cpu);
        len += p5d(buf + len, md_opt->v);
        len += sprintf(buf + len, "\n");
        return len;
}

int
read_proc_dpm_md_opts(char *page, char **start, off_t offset,
		      int count, int *eof, void *data)
{
	int len = 0;
	int limit = offset + count;
	struct dpm_opt *opt;
	struct list_head *opt_list;
	
	/* FIXME: For now we assume that the complete table,
	 * formatted, fits within one page */
	if (offset >= PAGE_SIZE)
		return 0;

	if (dpm_lock_interruptible())
		return -ERESTARTSYS;

	if (!dpm_initialized)
		len += sprintf(page + len, "DPM is not initialized\n");
	else if (!dpm_enabled)
		len += sprintf(page + len, "DPM is disabled\n");
	else {
		len += sprintf(page + len,
			       "The active DPM policy is \"%s\"\n",
			       dpm_active_policy->name);
		len += sprintf(page + len, 
			       "The current operating point is \"%s\"\n",
			       dpm_active_opt->name);
	}

	if (dpm_initialized) {
		len += sprintf(page + len, 
			       "Table of all defined operating points\n");

		len += sprintf(page + len, 
				" Name           Count  CPU  VOLTAGE\n");

		list_for_each(opt_list, &dpm_opts) {
			opt = list_entry(opt_list, struct dpm_opt, list);
			if (len >= PAGE_SIZE)
				BUG();
			if (len >= limit)
				break;
			len += dpm_proc_print_opt(page + len, opt);
		}
	}
	dpm_unlock();
	*eof = 1;
	if (offset >= len)
		return 0;
	*start = page + offset;
	return min(count, len - (int)offset);
}

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * /proc/driver/dpm/md/cmd (Write-only)
 *
 *  This is a catch-all, simple command processor for the Innovator DPM
 *  implementation. These commands are for experimentation and development
 *  _only_, and may leave the system in an unstable state.
 *
 *  No commands defined now.
 *
 ****************************************************************************/

int 
write_proc_dpm_md_cmd (struct file *file, const char *buffer,
		       unsigned long count, void *data)
{
	char *buf, *tok, *s;
	char *whitespace = " \t\r\n";
	int ret = 0;

	if (current->uid != 0)
		return -EACCES;
	if (count == 0)
		return 0;
	if (!(buf = kmalloc(count + 1, GFP_KERNEL)))
		return -ENOMEM;
	if (copy_from_user(buf, buffer, count)) {
		kfree(buf);
		return -EFAULT;
	}
	buf[count] = '\0';
	s = buf + strspn(buf, whitespace);
	tok = strsep(&s, whitespace);
	
	if (strcmp(tok, "define-me") == 0) {
		;
	} else {
		ret = -EINVAL;
	}
		kfree(buf);
	if (ret == 0)
		return count;
	else 
		return ret;
}

/*
 * idle loop
 */

static void (*orig_idle)(void);

void dpm_x86_idle(void)
{
	extern void default_idle(void);

	if (orig_idle)
		orig_idle();
	else
		default_idle();
}

/****************************************************************************
 * Initialization/Exit
 ****************************************************************************/

void
dpm_x86_startup(void)
{
	orig_idle = pm_idle;
	pm_idle = dpm_idle;

	if (dpm_bd.startup)
		dpm_bd.startup();
}

void
dpm_x86_cleanup(void)
{
	pm_idle = orig_idle;

	if (dpm_bd.cleanup)
		dpm_bd.cleanup();
}

int __init
dpm_x86_init(void)
{
	printk("Dynamic Power Management for x86.\n");
	
	dpm_md.init		= NULL;
	dpm_md.init_opt		= dpm_x86_init_opt;
	dpm_md.set_opt		= dpm_x86_set_opt;
	dpm_md.get_opt		= dpm_x86_get_opt;
	dpm_md.idle		= dpm_x86_idle;
	dpm_md.startup		= dpm_x86_startup;
	dpm_md.cleanup		= dpm_x86_cleanup;
	return 0;
}
__initcall(dpm_x86_init);
