/*
 * power-dpm.c -- Dynamic Power Management LDM power hooks
 *
 * (c) 2003 MontaVista Software, Inc. This file is licensed under the
 * terms of the GNU General Public License version 2. This program is
 * licensed "as is" without any warranty of any kind, whether express or
 * implied.
 */

#include <linux/device.h>
#include <linux/pm.h>
#include <linux/dpm.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/slab.h>

#include "power.h"

#ifdef CONFIG_HOTPLUG
/*
 * power hotplug events
 */

#define BUFFER_SIZE	1024	/* should be enough memory for the env */
#define NUM_ENVP	32	/* number of env pointers */
static unsigned long sequence_num;
static spinlock_t sequence_lock = SPIN_LOCK_UNLOCKED;

void power_event(char *eventstr)
{
	char *argv [3];
	char **envp = NULL;
	char *buffer = NULL;
	char *scratch;
	int i = 0;
	int retval;
	unsigned long seq;

	if (!hotplug_path[0])
		return;

	envp = kmalloc(NUM_ENVP * sizeof (char *), GFP_KERNEL);
	if (!envp)
		return;
	memset (envp, 0x00, NUM_ENVP * sizeof (char *));

	buffer = kmalloc(BUFFER_SIZE, GFP_KERNEL);
	if (!buffer)
		goto exit;

	argv [0] = hotplug_path;
	argv [1] = "power";
	argv [2] = 0;

	/* minimal command environment */
	envp [i++] = "HOME=/";
	envp [i++] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";

	scratch = buffer;

	envp [i++] = scratch;
	scratch += sprintf(scratch, "ACTION=event") + 1;

	spin_lock(&sequence_lock);
	seq = sequence_num++;
	spin_unlock(&sequence_lock);

	envp [i++] = scratch;
	scratch += sprintf(scratch, "SEQNUM=%ld", seq) + 1;
	envp [i++] = scratch;
	scratch += sprintf(scratch, "EVENT=%s", eventstr) + 1;

	pr_debug ("%s: %s %s %s %s %s %s %s\n", __FUNCTION__, argv[0], argv[1],
		  envp[0], envp[1], envp[2], envp[3], envp[4]);
	retval = call_usermodehelper (argv[0], argv, envp, 0);
	if (retval)
		pr_debug ("%s - call_usermodehelper returned %d\n",
			  __FUNCTION__, retval);

exit:
	kfree(buffer);
	kfree(envp);
	return;
}

void device_power_event(struct device * dev, char *eventstr)
{
	char *argv [3];
	char **envp = NULL;
	char *buffer = NULL;
	char *scratch;
	int i = 0;
	int retval;
	unsigned long seq;

	if (!hotplug_path[0])
		return;

	envp = kmalloc(NUM_ENVP * sizeof (char *), GFP_KERNEL);
	if (!envp)
		return;
	memset (envp, 0x00, NUM_ENVP * sizeof (char *));

	buffer = kmalloc(BUFFER_SIZE, GFP_KERNEL);
	if (!buffer)
		goto exit;

	argv [0] = hotplug_path;
	argv [1] = "power";
	argv [2] = 0;

	/* minimal command environment */
	envp [i++] = "HOME=/";
	envp [i++] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";

	scratch = buffer;

	envp [i++] = scratch;
	scratch += sprintf(scratch, "ACTION=device-event") + 1;

	spin_lock(&sequence_lock);
	seq = sequence_num++;
	spin_unlock(&sequence_lock);

	envp [i++] = scratch;
	scratch += sprintf(scratch, "SEQNUM=%ld", seq) + 1;
	envp [i++] = scratch;
	scratch += sprintf(scratch, "DEVICE=%s", dev->bus_id) + 1;
	envp [i++] = scratch;
	scratch += sprintf(scratch, "EVENT=%s", eventstr) + 1;

	pr_debug ("%s: %s %s %s %s %s %s %s\n", __FUNCTION__, argv[0], argv[1],
		  envp[0], envp[1], envp[2], envp[3], envp[4]);
	retval = call_usermodehelper (argv[0], argv, envp, 0);
	if (retval)
		pr_debug ("%s - call_usermodehelper returned %d\n",
			  __FUNCTION__, retval);

exit:
	kfree(buffer);
	kfree(envp);
	return;
}
#else /* !defined(CONFIG_HOTPLUG) */
void power_event(char *eventstr)
{
}
void device_power_event(struct device * dev, char *eventstr)
{
}
#endif

EXPORT_SYMBOL(power_event);
EXPORT_SYMBOL(device_power_event);

/*
 * Device constraints
 */

#ifdef CONFIG_DPM
LIST_HEAD(dpm_constraints);
DECLARE_MUTEX(dpm_constraints_sem);

void assert_constraints(struct constraints *constraints)
{
	if (! constraints || constraints->asserted)
		return;

	down(&dpm_constraints_sem);
	constraints->asserted = 1;
	list_add_tail(&constraints->entry, &dpm_constraints);
	up(&dpm_constraints_sem);

	/* DPM-PM-TODO: Check against DPM state. */

}


void deassert_constraints(struct constraints *constraints)
{
	if (! constraints || ! constraints->asserted)
		return;

	down(&dpm_constraints_sem);
	constraints->asserted = 0;
	list_del_init(&constraints->entry);
	up(&dpm_constraints_sem);
}


static ssize_t 
constraints_show(struct device * dev, char * buf)
{
	int i, cnt = 0;
	
	if (dev->constraints) {
		for (i = 0; i < dev->constraints->count; i++) {
			cnt += sprintf(buf + cnt,"%s: min=%d max=%d\n",
				       dpm_param_names[dev->constraints->param[i].id],
				       dev->constraints->param[i].min,
				       dev->constraints->param[i].max);
		}

		cnt += sprintf(buf + cnt,"asserted=%s violations=%d\n",
			       dev->constraints->asserted ? 
			       "yes" : "no", dev->constraints->violations);
	} else {
		cnt += sprintf(buf + cnt,"none\n");
	}

	return cnt;
}

static ssize_t
constraints_store(struct device * dev, const char * buf, size_t count)
{
	int num_args, paramid, min, max;
	int cidx;
	const char *cp, *paramname;
	int paramnamelen;
	int provisional = 0;
	int ret = 0;

	if (!dev->constraints) {
		if (! (dev->constraints = kmalloc(sizeof(struct constraints),
						  GFP_KERNEL)))
			return -EINVAL;

		memset(dev->constraints, 0, 
		       sizeof(struct constraints));
		provisional = 1;
	}

	cp = buf;
	while((cp - buf < count) && *cp && (*cp == ' '))
		cp++;

	paramname = cp;

	while((cp - buf < count) && *cp && (*cp != ' '))
		cp++;

	paramnamelen = cp - paramname;
	num_args = sscanf(cp, "%d %d", &min, &max);

	if (num_args != 2) {
		printk("DPM: Need 2 integer parameters for constraint min/max.\n");
		ret = -EINVAL;
		goto out;
	}

	for (paramid = 0; paramid < DPM_PP_NBR; paramid++) {
		if (strncmp(paramname, dpm_param_names[paramid], paramnamelen) == 0)
			break;
	}

	if (paramid >= DPM_PP_NBR) {
		printk("DPM: Unknown power parameter name in device constraints\n");
		ret = -EINVAL;
		goto out;
	}
	
	for (cidx = 0; cidx < dev->constraints->count; cidx++)
		if (dev->constraints->param[cidx].id == paramid)
			break;

	if (cidx >= DPM_CONSTRAINT_PARAMS_MAX) {
		ret = -EINVAL;
		goto out;
	}
	
	dev->constraints->param[cidx].id = paramid; 
	dev->constraints->param[cidx].max = max; 
	dev->constraints->param[cidx].min = min; 
	
	if (cidx == dev->constraints->count)
		dev->constraints->count++;

	/* New constraints should start off with same state as power
	   state */
	if (provisional && (dev->power.power_state == 0))
		assert_constraints(dev->constraints);

out:

	if (provisional && (ret < 0)) {
		kfree(dev->constraints);
		dev->constraints = NULL;
	}

	return ret < 0 ? ret : count;
}

DEVICE_ATTR(constraints,S_IWUSR | S_IRUGO,
            constraints_show,constraints_store);

#else /* CONFIG_DPM */
void assert_constraints(struct constraints *constraints)
{
}

void deassert_constraints(struct constraints *constraints)
{
}
#endif /* CONFIG_DPM */

EXPORT_SYMBOL(assert_constraints);
EXPORT_SYMBOL(deassert_constraints);

#ifdef CONFIG_DPM

/*
 * Driver scale callbacks
 */

LIST_HEAD(dpm_scale_list);
DECLARE_MUTEX(dpm_scale_sem);

/* This function may be called by the platform frequency scaler before
   or after a frequency change, in order to let drivers adjust any
   clocks or calculations for the new frequency. */

void dpm_driver_scale(int level)
{
	struct list_head * entry;

	/* Go through the list of drivers with scale functions. */

	list_for_each(entry,&dpm_scale_list) {
		struct device_driver *drv = 
			container_of(entry, struct device_driver,
				     scale_entry);

		if (drv->scale)
			drv->scale(level);
	}
}

void dpm_register_scale(struct device_driver *drv)
{
	if (drv->scale) {
		down(&dpm_scale_sem);
		list_add_tail(&drv->scale_entry, &dpm_scale_list);
		up(&dpm_scale_sem);
	}
}

void dpm_unregister_scale(struct device_driver *drv)
{
	if (drv->scale) {
		down(&dpm_scale_sem);
		list_del_init(&drv->scale_entry);
		up(&dpm_scale_sem);
	}
}

EXPORT_SYMBOL(dpm_driver_scale);
EXPORT_SYMBOL(dpm_register_scale);
EXPORT_SYMBOL(dpm_unregister_scale);

int dpm_constraint_rejects = 0;


int
dpm_default_check_constraint(struct constraint_param *param, 
			     struct dpm_opt *opt)
{
	return (opt->pp[param->id] == -1) ||
		((param->min == -1 || opt->pp[param->id] >= param->min) && 
		 (param->max == -1 || opt->pp[param->id] <= param->max));
}

static int
dpm_check_a_constraint(struct constraints *constraints, struct dpm_opt *opt)
{
	int i;
	
	if (! constraints || !constraints->asserted) {
		return 1;
	}

	for (i = 0; i < constraints->count; i++) {
		struct constraint_param *param = &constraints->param[i];

		if (! dpm_md_check_constraint(param, opt)) {
			dpm_trace(DPM_TRACE_CONSTRAINT_ASSERTED,
				  param->id, param->min, param->max,
				  opt);
			return 0;
		}
	}

	/* if we got this far, the opt is valid */
	return 1;
}

int dpm_check_constraints(struct dpm_opt *opt)
{
	struct list_head * entry;
	int valid = 1;

	list_for_each(entry,&dpm_constraints) {
		struct constraints *constraints = 
			list_entry(entry, struct constraints, entry);
		if (!dpm_check_a_constraint(constraints, opt)) {
			constraints->violations++;
			dpm_constraint_rejects++;
			valid = 0;
		}
	}

	return valid;
}

EXPORT_SYMBOL(dpm_default_check_constraint);
EXPORT_SYMBOL(dpm_check_constraints);

int dpm_show_opconstraints(struct dpm_opt *opt, char * buf)
{
#ifdef CONFIG_PM
	struct list_head * entry;
	int len = 0;

	list_for_each_prev(entry,&dpm_active) {
		struct device * dev = to_device(entry);

		if (!dpm_check_a_constraint(dev->constraints, opt)) {
			len += sprintf(buf + len, "%s/%s\n", dev->bus->name,
				       dev->bus_id);
		}
	}

	return len;
#else /* CONFIG_PM */
	return 0;
#endif /* CONFIG_PM */
}
#endif /* CONFIG_DPM */

