/*
 *
 * Copyright 2006 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/device.h>
#include <linux/proc_fs.h>
#include <linux/devfs_fs.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/cdev.h>

#include <linux/serial_8250.h>

#include <asm/uaccess.h>
#include <linux/poll.h>
#include <linux/delay.h>

#include <linux/pm.h>
#include <asm/arch/pmu.h>
#include <asm/arch/smu.h>
#include <asm/arch/mp200_ad.h>
#include <asm/arch/gpiocore.h>
#include <asm/arch/icx1087_cradle.h>

/*******************************************************/
/* definition                                          */
/*******************************************************/

#define ICX1087_CRADLE_TIME_10MS       (HZ/100)   /* 10ms tick */
#define ICX1087_CHATTERING_CHECK_TIMES 5          /* chattering check times */
#define ICX1087_CRADLE_GPIO_MASK       0x00000080 /* Power IC GPIO7 */
#define ICX1087_CRADLE_AD_CHAN         6

/* device structure */
struct icx1087_cradle_dev {

    /* work queue */
    struct work_struct      work;
    struct workqueue_struct *workqueue;
    int                     wq_disable;   /* for suspend/resume/rmmod */
    int                     wq_done;      /* for serial open notifier */
    wait_queue_head_t       wqh_poll;     /* for poll notify */
    wait_queue_head_t       wqh_serial;   /* for serial_open notify */

    /* state */
    struct semaphore        sem;          /* semaphore */
    icx1087_cradle_state_t  state;        /* current */
    int                     report_count; /* reported count */
    int                     pwic_ver;     /* power IC version */

    atomic_t                open_flag;    /* open flag */

    /* char dev structure */
    struct cdev             cdev;

#ifdef CONFIG_PROC_FS
    struct proc_dir_entry   *proc_dir;
    struct proc_dir_entry   *proc_entry;
    struct proc_dir_entry   *proc_entry_resume;
#endif
};


/* device structure */
static int icx1087_cradle_major=0;
static int icx1087_cradle_minor=0;
static struct icx1087_cradle_dev icx1087_cradle_device;

/* etc */
static void icx1087_cradle_enable_irq( struct icx1087_cradle_dev *dev, int enable );
static int  icx1087_cradle_get_rxd( struct icx1087_cradle_dev *dev );
static int  icx1087_cradle_get_id( struct icx1087_cradle_dev *dev, int connect );
static void icx1087_cradle_sleep( struct icx1087_cradle_dev *dev, int timeout );

/* interrupt */
static irqreturn_t icx1087_cradle_interrupt( int irq, void *dev_id, struct pt_regs *regs );
static void icx1087_cradle_serial_event_notifier( int type );

/* work queue */
static int  icx1087_cradle_queue_work( struct icx1087_cradle_dev *dev );
static void icx1087_cradle_do_work( void *data );
static int  icx1087_cradle_do_work_check( struct icx1087_cradle_dev *dev );

/* user api */
static int icx1087_cradle_open( struct inode *inode, struct file *file );
static int icx1087_cradle_release( struct inode *inode, struct file *file );
static int icx1087_cradle_ioctl( struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg );
static u_int icx1087_cradle_poll (struct file *file, struct poll_table_struct *wait );

/* initialize/cleanup */
static int icx1087_cradle_probe( struct device *non );
static int icx1087_cradle_remove( struct device *non );


/* file operation */
struct file_operations icx1087_cradle_fops = {
    .owner   =  THIS_MODULE,
    .open    =  icx1087_cradle_open,
    .release =  icx1087_cradle_release,
    .ioctl   =  icx1087_cradle_ioctl,
    .poll    =  icx1087_cradle_poll,
};


/*  module param */
static int icx1087_cradle_param_resume = 1; /* 1: enable */
module_param_named( resume, icx1087_cradle_param_resume, int, 0 );


/*******************************************************/
/* etc..                                               */
/*******************************************************/

/*
 * enable/disable interrupt
 */
static void icx1087_cradle_enable_irq( struct icx1087_cradle_dev *dev, int enable )
{
    if ( enable ) {
        pwc_write( PWC_BMASK1, ICX1087_CRADLE_GPIO_MASK, ICX1087_CRADLE_GPIO_MASK );
        pwc_write( PWC_FMASK1, ICX1087_CRADLE_GPIO_MASK, ICX1087_CRADLE_GPIO_MASK );
    } else {
        pwc_write( PWC_BMASK1, 0, ICX1087_CRADLE_GPIO_MASK );
        pwc_write( PWC_FMASK1, 0, ICX1087_CRADLE_GPIO_MASK );
    }
}


/*
 * get RxD value
 */
static int icx1087_cradle_get_rxd( struct icx1087_cradle_dev *dev )
{
    int data;

    pwc_read( PWC_IOIN1, &data );
    if ( data & ICX1087_CRADLE_GPIO_MASK ) return 1;
    else return 0;
}

/*
 * get AD value
 */
static int icx1087_cradle_get_id( struct icx1087_cradle_dev *dev, int connect )
{
    int val;

    if ( !connect ) return 0; /* disconnect default */

    val = mp200_ad_get_val( ICX1087_CRADLE_AD_CHAN );

    if      ( val < 12  ) return ICX1087_CRADLE_ID_0; /* about 0   */
    else if ( val < 34  ) return ICX1087_CRADLE_ID_1; /* about 23  */
    else if ( val < 53  ) return ICX1087_CRADLE_ID_2; /* about 45  */
    else if ( val < 72  ) return ICX1087_CRADLE_ID_3; /* about 62  */
    else if ( val < 94  ) return ICX1087_CRADLE_ID_4; /* about 82  */
    else if ( val < 117 ) return ICX1087_CRADLE_ID_5; /* about 106 */
    else if ( val < 140 ) return ICX1087_CRADLE_ID_6; /* about 128 */
    else if ( val < 163 ) return ICX1087_CRADLE_ID_7; /* about 151 */
    else if ( val < 193 ) return ICX1087_CRADLE_ID_8; /* about 174 */
    else if ( val < 222 ) return ICX1087_CRADLE_ID_9; /* about 211 */
    else if ( val < 244 ) return ICX1087_CRADLE_ID_A; /* about 233 */
    else                  return ICX1087_CRADLE_ID_B; /* about 255 */

}


/*
 * sleep
 */
static void icx1087_cradle_sleep( struct icx1087_cradle_dev *dev, int timeout )
{
    set_current_state( TASK_INTERRUPTIBLE );
    schedule_timeout( timeout );
}


/*******************************************************/
/* interrupt                                           */
/*******************************************************/

/*
 * interrupt handler of RxD
 */
static irqreturn_t icx1087_cradle_interrupt( int irq, void *dev_id, struct pt_regs *regs )
{
    struct icx1087_cradle_dev *dev = dev_id;

    icx1087_cradle_queue_work( dev );
    return IRQ_HANDLED;
}


/*
 * notifier from serial
 */
static void icx1087_cradle_serial_event_notifier( int type )
{
    struct icx1087_cradle_dev *dev = &icx1087_cradle_device;

    switch ( type ) {

      case SERIAL8250_EVENT_TYPE_RD_ERR:
        /* framing error, break interrupt */

        /* check disconnect */
        icx1087_cradle_queue_work( dev );
        break;

      case SERIAL8250_EVENT_TYPE_OPEN:
        /* user open serial */

        /* disconnect state */
        if ( !dev->state.connect ) return;
        /* connect(now) and same id */
        if ( icx1087_cradle_get_rxd( dev ) && ( dev->state.id == icx1087_cradle_get_id( dev, 1 ) ) ) return;

        /* re-check state */

        /* enable interrupt with the best of care */
        icx1087_cradle_enable_irq( dev, 1 );
        /* clear wait flag */
        dev->wq_done = 0;
        /* check state */
        icx1087_cradle_queue_work( dev );
        /* wait done */
        wait_event_interruptible_timeout( dev->wqh_serial, dev->wq_done, ICX1087_CRADLE_TIME_10MS * 30 ); /* 300ms */

        break;

      default:
        break;
    }
}

/*******************************************************/
/* work queue                                          */
/*******************************************************/

/*
 * queuing work queue
 */
static int icx1087_cradle_queue_work( struct icx1087_cradle_dev *dev )
{
    if ( dev->wq_disable ) return -1;
    return queue_delayed_work( dev->workqueue, &dev->work, ICX1087_CRADLE_TIME_10MS );
}


/*
 * work queue thread
 */
static void icx1087_cradle_do_work( void *data )
{
    struct icx1087_cradle_dev *dev = data;

    if ( icx1087_cradle_do_work_check( dev ) > 0 ) {
        /* retry */
        icx1087_cradle_queue_work( dev );
    }
}


/*
 * work queue thread
 */
static int icx1087_cradle_do_work_check( struct icx1087_cradle_dev *dev )
{
    int connect, i, id;

    if ( dev->wq_disable ) return -1;

    /* first check */
    connect = icx1087_cradle_get_rxd( dev );
    id      = icx1087_cradle_get_id ( dev, connect );

    /* same current state ? */
    if ( ( dev->state.connect == connect ) && ( dev->state.id == id ) ) {
        return 0;
    }

    /* rest 4 times check */
    for ( i = 0; i < ICX1087_CHATTERING_CHECK_TIMES - 1 ; i++ ) { /* first sleep is in queue_delayed_work(10m) */

        /* sleep 10ms */
        icx1087_cradle_sleep( dev, ICX1087_CRADLE_TIME_10MS );

        if ( dev->wq_disable ) return -1;

        /* same value ? */
        if ( connect != icx1087_cradle_get_rxd( dev ) )    return 0;
        if ( id != icx1087_cradle_get_id( dev, connect ) ) return 1; /* retry */
    }

    /* if connect then disable interrupt */
    icx1087_cradle_enable_irq( dev, connect ? 0 : 1 );

    if ( connect ) {

        /* sleep 100ms */
        icx1087_cradle_sleep( dev, ICX1087_CRADLE_TIME_10MS*10 );

        if ( dev->wq_disable ) return -1;

        if ( id != icx1087_cradle_get_id( dev, connect ) ) {
            icx1087_cradle_enable_irq( dev, 1 ); /* enable interrupt */
            return 1;                            /* retry */
        }
    }

    /* update info */
    down( &dev->sem );
    dev->state.connect = connect;
    dev->state.id      = id;
    dev->state.count++;
    printk( KERN_INFO "%s: connect %d %02x, %d\n", __FUNCTION__, dev->state.connect, dev->state.id, dev->state.count );
    up( &dev->sem );

    /* notify poll */
    wake_up_interruptible( &dev->wqh_poll );

    /* notify open_serial */
    wake_up_interruptible( &dev->wqh_serial );
    dev->wq_done = 1;

    return 0;
}


/*******************************************************/
/* User API                                            */
/*******************************************************/

/*
 * open
 */
static int icx1087_cradle_open( struct inode *inode, struct file *file )
{
    struct icx1087_cradle_dev *dev = container_of( inode->i_cdev, struct icx1087_cradle_dev, cdev );
    file->private_data = dev;

    if ( !atomic_dec_and_test( &dev->open_flag ) ) {
        atomic_inc( &dev->open_flag );
        return -EBUSY;
    }

    if ( down_interruptible( &dev->sem ) )
      return -ERESTARTSYS;

    if ( dev->state.connect ) {
        dev->report_count = dev->state.count - 1;
    } else {
        /* not report first disconect state */
        dev->report_count = dev->state.count;
    }

    up( &dev->sem );

    return 0;
}


/*
 * close
 */
static int icx1087_cradle_release( struct inode *inode, struct file *file )
{
    struct icx1087_cradle_dev *dev = container_of( inode->i_cdev, struct icx1087_cradle_dev, cdev );

    atomic_inc( &dev->open_flag );
    return 0;
}


/*
 * close
 */
static int icx1087_cradle_ioctl( struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg )
{
    struct icx1087_cradle_dev *dev = file->private_data;
    int ret = 0;

    if ( down_interruptible( &dev->sem ) )
      return -ERESTARTSYS;

    switch ( cmd ) {

      case ICX1087_CRADLE_GET_STATE:
        ret = copy_to_user( (void*)arg, (void*)&dev->state, _IOC_SIZE( cmd ) );
        if ( ret ) {
            printk( KERN_ERR "%s: copy_to_user error %d\n", __FUNCTION__, ret );
            ret = -EFAULT;
        } else {
            dev->report_count = dev->state.count;
        }
        break;

      default:
        printk( KERN_WARNING "%s: illegal request\n", __FUNCTION__ );
        ret = -ENOTTY;
    }

    up( &dev->sem );
    return ret;
}


/*
 * poll
 */
static u_int icx1087_cradle_poll (struct file *file, struct poll_table_struct *wait )
{
    struct icx1087_cradle_dev *dev = file->private_data;
    int ret = 0;

    down( &dev->sem );
    poll_wait( file, &dev->wqh_poll, wait );
    if ( dev->report_count != dev->state.count ) {
        ret = POLLIN | POLLRDNORM;
    }
    up( &dev->sem );

    return ret;
}


/*******************************************************/
/* proc I/F                                            */
/*******************************************************/

#ifdef CONFIG_PROC_FS

static int  icx1087_cradle_proc_state( char *page, char **start, off_t off, int count, int *eof, void *data )
{
    struct icx1087_cradle_dev *dev = data;
    int len;

    len = sprintf( page, "%d %02x %d\n", dev->state.connect, dev->state.id, dev->state.count );
    *eof=1;
    return len;
}


static int  icx1087_cradle_proc_read_resume( char *page, char **start, off_t off, int count, int *eof, void *data )
{
    int len;

    len = sprintf( page, "%d\n", icx1087_cradle_param_resume );
    *eof=1;

    return len;
}


static int  icx1087_cradle_proc_write_resume( struct file *file, const char *buffer, unsigned long count, void *data )
{
    if ( buffer[0] == '0' ) {
        icx1087_cradle_param_resume = 0;
    } else {
        icx1087_cradle_param_resume = 1;
    }
    return count;
}


static int icx1087_cradle_create_proc( struct icx1087_cradle_dev *dev )
{

    dev->proc_dir = proc_mkdir( "driver/icx1087_cradle", NULL );
    if ( !dev->proc_dir ) return -1;
    dev->proc_dir->owner = THIS_MODULE;

    dev->proc_entry = create_proc_entry( "state", 0400, dev->proc_dir );
    if ( !dev->proc_entry ) return -1;

    dev->proc_entry->read_proc = icx1087_cradle_proc_state;
    dev->proc_entry->owner = THIS_MODULE;
    dev->proc_entry->data  = dev;

    dev->proc_entry_resume = create_proc_entry( "resume", 0600, dev->proc_dir );
    if ( !dev->proc_entry_resume ) return -1;

    dev->proc_entry_resume->read_proc  = icx1087_cradle_proc_read_resume;
    dev->proc_entry_resume->write_proc = icx1087_cradle_proc_write_resume;
    dev->proc_entry_resume->owner = THIS_MODULE;
    dev->proc_entry_resume->data  = dev;


    return 0;
}

static void icx1087_cradle_remove_proc( struct icx1087_cradle_dev *dev )
{

    if ( !dev->proc_dir ) return;
    if ( dev->proc_entry ) remove_proc_entry( "state", dev->proc_dir );
    if ( dev->proc_entry_resume ) remove_proc_entry( "resume", dev->proc_dir );
    remove_proc_entry( "driver/icx1087_cradle", NULL );
}

#endif


/*******************************************************/
/* driver initialize/cleanup                           */
/*******************************************************/

/*
 * driver cleanup
 */
static int __devexit icx1087_cradle_remove( struct device *non )
{
    struct icx1087_cradle_dev *dev = &icx1087_cradle_device;
    dev_t devno = MKDEV( icx1087_cradle_major, icx1087_cradle_minor);

    /* disable notifier from serial */
    serial8250_register_event_notifier( NULL );

    /* disable interrupt */
    icx1087_cradle_enable_irq( dev, 0 );
    free_irq( INT_PWC_IO07, dev );

    /* wait work queue empty. then destroy. */
    dev->wq_disable = 1;
    if ( dev->workqueue ) {
        cancel_delayed_work( &dev->work );
        flush_workqueue( dev->workqueue );
        destroy_workqueue( dev->workqueue );
    }

#ifdef CONFIG_PROC_FS
    icx1087_cradle_remove_proc( dev );
#endif

#ifdef CONFIG_DEVFS_FS
    devfs_remove( ICX1087_CRADLE_NAME );
#endif

    /* unregister driver */
    cdev_del( &dev->cdev );
    unregister_chrdev_region( devno, 1 );

    return 0;
}


/*
 * driver initialization
 */
static int icx1087_cradle_probe( struct device *non )
{
    struct icx1087_cradle_dev *dev = &icx1087_cradle_device;
    int ret;
    dev_t devno = 0;

    /* initialize dev structure */
    memset( dev, 0, sizeof( struct icx1087_cradle_dev ) );
    init_MUTEX( &dev->sem );
    init_waitqueue_head( &dev->wqh_poll );
    init_waitqueue_head( &dev->wqh_serial );
    dev->open_flag = (atomic_t)ATOMIC_INIT(1);
    pwc_read( PWC_LSIVER, &dev->pwic_ver );

    /* assigne devno */
    if ( icx1087_cradle_major ) {
        devno = MKDEV( icx1087_cradle_major, icx1087_cradle_minor );
        ret   = register_chrdev_region( devno, 1, ICX1087_CRADLE_NAME );
    } else {
        ret = alloc_chrdev_region( &devno, icx1087_cradle_minor, 1, ICX1087_CRADLE_NAME );
        icx1087_cradle_major = MAJOR( devno );
    }
    if ( ret < 0 ) {
        printk( KERN_ERR "%s: can't get major %d\n", __FUNCTION__, icx1087_cradle_major );
        return ret;
    }

    /* initialize cdev */
    cdev_init( &dev->cdev, &icx1087_cradle_fops );
    dev->cdev.owner = THIS_MODULE;
    dev->cdev.ops   = &icx1087_cradle_fops;

    ret = cdev_add( &dev->cdev, devno, 1 );
    if ( ret ) {
        printk( KERN_ERR "%s: cdev_add failed\n", __FUNCTION__ );
        unregister_chrdev_region( devno, 1 );
        return ret;
    }

#ifdef CONFIG_DEVFS_FS
    ret = devfs_mk_cdev( devno, S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP, ICX1087_CRADLE_NAME );
    if ( ret ) {
        printk( KERN_ERR "%s: devfs_mk_cdev failed\n", __FUNCTION__ );
        goto failed;
    }
#endif


#ifdef CONFIG_PROC_FS
    icx1087_cradle_create_proc( dev );
#endif

    /* initialize work queue */
    INIT_WORK( &dev->work, icx1087_cradle_do_work, dev );
    dev->workqueue = create_singlethread_workqueue( "cradle_wq" );
    if ( !dev->workqueue ) {
        printk( KERN_ERR "%s: create_singlethread_workqueue failed\n", __FUNCTION__ );
        ret = -ENOMEM;
        goto failed;
    }

    /* setup interrupt handler */
    set_irq_type( INT_PWC_IO07, IRQT_RISING );
    ret = request_irq( INT_PWC_IO07, icx1087_cradle_interrupt, 0, ICX1087_CRADLE_NAME, dev );
    if ( ret ) {
        printk( KERN_ERR "%s: request_irq failed\n", __FUNCTION__ );
        goto failed;
    }

    /* setup notifier from serial */
    serial8250_register_event_notifier( icx1087_cradle_serial_event_notifier );

    /* check current state */
    icx1087_cradle_queue_work( dev );

    printk( KERN_INFO "starting %s OK.\n", ICX1087_CRADLE_NAME );
    return 0;

  failed:

    icx1087_cradle_remove( NULL );
    printk( KERN_ERR "starting %s Failed!!\n", ICX1087_CRADLE_NAME );
    return ret;
}


#ifdef CONFIG_DPM

/*******************************************************/
/* suspend/resume                                      */
/*******************************************************/

extern int mp200_pm_get_usb_suspend_state( void );
/*
 * suspend
 */
static int icx1087_cradle_suspend( struct device *non, u32 state, u32 level )
{
    struct icx1087_cradle_dev *dev = &icx1087_cradle_device;

    switch( level ) {

      case SUSPEND_POWER_DOWN:

        /* disable do_work and new queuing (no wait) */
        dev->wq_disable = 1;

        /* anyway notify disconnect if connected */
        down( &dev->sem );
        if ( dev->state.connect ) {
            dev->state.connect = 0;
            dev->state.id      = 0;
            dev->state.count++;
            printk( KERN_INFO "%s: connect %d %02x, %d\n", __FUNCTION__, dev->state.connect, dev->state.id, dev->state.count );
            wake_up_interruptible( &dev->wqh_poll );
        }
        up( &dev->sem ); 

        /* enable interrupt for wakeup by RxD HIGH */
        if ( mp200_pm_get_usb_suspend_state() ) {
            icx1087_cradle_enable_irq( dev, 0 );
        } else if ( dev->pwic_ver >= 0x02 ) {
            icx1087_cradle_enable_irq( dev, 1 );
        } else {
            icx1087_cradle_enable_irq( dev, icx1087_cradle_param_resume );
        }

        break;

    default:
        break;
    }
    return 0;
}


/*
 * resume
 */
static int icx1087_cradle_resume( struct device *non, u32 level )
{
    struct icx1087_cradle_dev *dev = &icx1087_cradle_device;

    switch ( level ) {

      case RESUME_POWER_ON:

        /* wait workqueu empty */
        cancel_delayed_work( &dev->work );
        flush_workqueue( dev->workqueue );

        /* enable new queuing */
        dev->wq_disable = 0;

        /* current state check */
        icx1087_cradle_queue_work( dev );

        break;

    default:
        break;
    }
    return 0;
}

#endif

static void icx1087_cradle_shutdown( struct device *non )
{
    struct icx1087_cradle_dev *dev = &icx1087_cradle_device;

    if ( dev->pwic_ver == 0x01 ) return; /* not support */

    /* disable do_work and new queuing (no wait) */
    dev->wq_disable = 1;

    /* do not call interrupt handler(include kernel handler) */
    pwc_write( PWC_BMASK1, 0, ICX1087_CRADLE_GPIO_MASK );

    /* enable interrupt factor */
    pwc_write( PWC_FMASK1, ICX1087_CRADLE_GPIO_MASK, ICX1087_CRADLE_GPIO_MASK );

    /* delay (wait factor on if already RX-HIGH) */
    udelay( 200 );

    /* clear factor */
    pwc_write( PWC_CLRFACT1, ICX1087_CRADLE_GPIO_MASK, ICX1087_CRADLE_GPIO_MASK );

    /* set reset mask */
    pwc_write( PWC_REGPWONMSK, ICX1087_CRADLE_GPIO_MASK, ICX1087_CRADLE_GPIO_MASK );
}


/*******************************************************/
/* driver entry                                        */
/*******************************************************/

/*
 * platform device structure
 */
void icx1087_cradle_pdev_release( struct device * dev ){}
static struct platform_device icx1087_cradle_pdev = {
    .name    = "icx1087_cradle",
    .dev = {
        .release = icx1087_cradle_pdev_release,
    },
};


/*
 * platform driver structure
 */
static struct device_driver icx1087_cradle_pdrv = {
    .name    = "icx1087_cradle",
    .bus     = &platform_bus_type,
    .probe   = icx1087_cradle_probe,
    .remove  = __devexit_p( icx1087_cradle_remove ),
#ifdef CONFIG_DPM
    .suspend = icx1087_cradle_suspend,
    .resume  = icx1087_cradle_resume,
#endif
    .shutdown  = icx1087_cradle_shutdown,
};


/*
 * driver entry
 */
static int __init icx1087_cradle_init( void )
{
    if ( platform_device_register( &icx1087_cradle_pdev ) ) {
        return -1;
    }
    if ( driver_register( &icx1087_cradle_pdrv ) ) {
        platform_device_unregister( &icx1087_cradle_pdev );
        return -1;
    }
    return 0;
}


/*
 * driver cleanup
 */
static void __exit icx1087_cradle_cleanup( void )
{
    driver_unregister( &icx1087_cradle_pdrv );
    platform_device_unregister( &icx1087_cradle_pdev );
}


/*
 * kernel parameter
 */
static int __init icx1087_cradle_setup_no_resume( char *s )
{
    icx1087_cradle_param_resume = 0;
    return 1;
}
__setup( "no_cradle_resume", icx1087_cradle_setup_no_resume );

module_init( icx1087_cradle_init );
module_exit( icx1087_cradle_cleanup);

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("Sony Cradle device driver");
MODULE_LICENSE("GPL");
