/*
 * ksymbol.c -- Set/Get a value to/form kernel symbol.
 *
 * Copyright 2006 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/config.h>			/* kernel configuration */
#include <linux/kernel.h>			/* often-used kernel functions */
#include <linux/fs.h>               /* file system */
#include <linux/devfs_fs_kernel.h>  /* devfs file system */
#include <linux/cdev.h>             /* character device driver */
#include <linux/init.h>				/* __init macro */
#include <linux/module.h>			/* kernel modules */
#include <linux/version.h>			/* kernel version */
#include <linux/device.h>			/* device and driver structure */
#include <linux/proc_fs.h>			/* proc file system */
#include <asm/uaccess.h>			/* put_user(), get_user() */
#include <linux/ksymbol.h>
//#define DEBUG
#define PrHead	"ksymbol: "
#include <linux/prdbg.h>

#define KSYMBOL_DEVNAME			"ksymbol"	/* device name */
#define KSYMBOL_DEV_ENTRY_NAME	"ksymbol"	/* /dev entry name */

/*
 *  module parameters
 */
static int  s_Major = 0; module_param(s_Major, int, S_IRUGO);
static int  s_Minor = 0; module_param(s_Minor, int, S_IRUGO);

/*
 *  file scope structures & variables
 */
typedef struct {
    struct cdev         cdev;           /* character device structure */
    struct class_simple *class;         /* */
} ksymbol_ctrl_t;

static ksymbol_ctrl_t	s_Ctrl;

static int DriverIoctl(struct inode *inode, struct file *file,
					unsigned int cmd, unsigned long arg)
{
	ksymbol_io_t	param;
	void	*addr;
	int 	result = 0;

	switch (cmd) {
	case KSYMBOL_SET:
	case KSYMBOL_GET:
		/* get user parameters and check its */
		if (copy_from_user(&param, (void *)arg, sizeof(param))) {
			return -EFAULT;
		}
		if (param.symbol == NULL || *param.symbol == '\0' ||
			param.buf == NULL || param.bufsize < 1) {
			return -EINVAL;
		}

		/* get a symbol value */
		addr = __symbol_get(param.symbol);
		PrDebug("symbol_get(%s): addr=%p, value=%x\n",
			param.symbol, addr, (addr != NULL) ? *(unsigned *)addr: 0);
		if (addr == NULL) {
			return -ENOENT;
		}

		if (cmd == KSYMBOL_SET) {
			if (copy_from_user(addr + param.offset, param.buf, param.bufsize)) {
				return -EFAULT;
			}
		} else {
			if (copy_to_user(param.buf, addr + param.offset, param.bufsize)) {
				return -EFAULT;
			}
		}
		result = 0;
		break;
	case KSYMBOL_GET_JIFFIES:
		result = put_user(jiffies, (unsigned *)arg);
		break;
	default:
		PrInfo("Unsupport ioctl command. cmd=%x, arg=%lx\n", cmd, arg);
		result = -EINVAL;
		break;
	}

	return result;
}

static int DriverOpen(struct inode *inode, struct file *file)
{
	PrDebug("\n");
	return 0;
}

static int DriverRelease(struct inode *inode, struct file *file)
{
	PrDebug("\n");
	return 0;
}

static struct file_operations s_FileOps = {
	.open		= DriverOpen,
	.release	= DriverRelease,
	.ioctl		= DriverIoctl,
	.owner		= THIS_MODULE
};

static int DriverProbe(struct device *dev)
{
	int	result = 0;

	PrDebug("\n");
	return result;
}

static int DriverRemove(struct device *dev)
{
	PrDebug("\n");
	return 0;
}

static struct device_driver s_DeviceDriver = {
	.name		= "ksymbol",
//	.bus		= &platform_bus_type,
	.probe		= DriverProbe,
	.remove		= DriverRemove,
};

static inline int AllocChrDevNum(int *major, int minor, int count,
                                const char *dev_name)
{
    dev_t   dev = 0;
    int result;

    /* static or dynamic allocate character device number */
    if (*major != 0) {
        dev = MKDEV(*major, minor);
        result = register_chrdev_region(dev, count, dev_name);
    } else {
        result = alloc_chrdev_region(&dev, minor, count, dev_name);
        *major = MAJOR(dev);
    }
    return result;
}

int __init ksymbol_init(void)
{
	dev_t	devno;	/* device number */
	int		result;

	PrInfo("Starting ksymbol driver v0.2\n");

	/* static or dynamic allocate device number */
	result = AllocChrDevNum(&s_Major, s_Minor, 1, KSYMBOL_DEVNAME);
	if (result < 0) {
		PrErr("Can't allocate device major number '%d)\n", s_Major);
		return result;
	}
	devno = MKDEV(s_Major, s_Minor);
	PrDebug("major device number is %d\n", s_Major);

	/* initialize character device object */
	cdev_init(&s_Ctrl.cdev, &s_FileOps);
	s_Ctrl.cdev.owner = THIS_MODULE;

	/* add character device to kernel */
	result = cdev_add(&s_Ctrl.cdev, devno, 1);
	if (result < 0) {
		PrErr("cdev_add failed. (result=%d)\n", result);
		goto err3;
	}

	/* create /sys/class/ksymbol directory */
	s_Ctrl.class = class_simple_create(THIS_MODULE, KSYMBOL_DEVNAME);
	if (IS_ERR(s_Ctrl.class)) {
		result = PTR_ERR(s_Ctrl.class);
		PrErr("class_simple_create failed. (result=%d)\n", result);
		goto err2;
	}

	/* create /sys/class/ksymbol/ksymbol/dev file */
	class_simple_device_add(s_Ctrl.class, devno,
			(struct device *)NULL, KSYMBOL_DEVNAME);

	driver_register(&s_DeviceDriver);

#ifdef CONFIG_DEVFS_FS
    result = devfs_mk_cdev(devno, S_IFCHR | S_IRUSR | S_IWUSR,
                KSYMBOL_DEV_ENTRY_NAME);
    if (result != 0) {
        PrErr("devfs_mk_cdev failed. (result=%d)\n", result);
        goto err1;
    }
#endif
    return 0;

err1:
    driver_unregister(&s_DeviceDriver);

    /* remove /sys entry */
    class_simple_device_remove(devno);
    class_simple_destroy(s_Ctrl.class);
err2:
    /* delete character device object */
    cdev_del(&s_Ctrl.cdev);
err3:
    /* free character device number */
    unregister_chrdev_region(devno, 1);

	return result;
}

void __exit ksymbol_exit(void)
{
	dev_t	devno = MKDEV(s_Major, s_Minor);

	PrDebug("\n");

	driver_unregister(&s_DeviceDriver);

    /* remove /sys entry */
    class_simple_device_remove(devno);
    class_simple_destroy(s_Ctrl.class);

#ifdef CONFIG_DEVFS_FS
    devfs_remove(KSYMBOL_DEV_ENTRY_NAME);  /* remove /dev entry */
#endif

    /* delete character device object */
    cdev_del(&s_Ctrl.cdev);

    /* free character device number */
    unregister_chrdev_region(devno, 1);
}

module_init(ksymbol_init);
module_exit(ksymbol_exit);

MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL");

/* end of ksymbol.c */
