/*
 *  File Name       : linux/drivers/char/mp200-rtc.c
 *  Function        : MP200 Real Time Clock interface
 *  Release Version : Ver 1.01
 *  Release Date    : 2006/03/08
 *
 *  Copyright (C) NEC Electronics Corporation 2005-2006
 *
 *
 *  This program is free software;you can redistribute it and/or modify itunder the terms of
 *  the GNU General Public License as published by Free Softwere Foundation;either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

/*
 * Copyright 2006,2007 Sony Corporation. 
 */

#include <linux/kernel.h>   /* printk() */
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/rtc.h>
#include <linux/interrupt.h>

#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/arch/gpiocore.h>

#ifdef CONFIG_SONY_ICX1087
#include <asm/arch/mp200_rtc.h>

static int _n_xstp = 0;

//#define ALARM_W_ENABLE

#endif /* CONFIG_SONY_ICX1087 */

extern raw_spinlock_t rtc_lock;

/* RTC register access macros: */
#define RTC_READ(addr, val) gpiocore_pwc_read((addr), (unsigned int *)(&(val)));
#define RTC_WRITE(val, addr) gpiocore_pwc_write((addr), (unsigned int)(val), 0xff);

/* Local BCD/BIN conversion macros: */
#ifdef BCD_TO_BIN
#undef BCD_TO_BIN
#endif
#define BCD_TO_BIN(val) ((val) = ((val) & 15) + ((val) >> 4) * 10)

#ifdef BIN_TO_BCD
#undef BIN_TO_BCD
#endif
#define BIN_TO_BCD(val) ((val) = (((val) / 10) << 4) + (val) % 10)

static unsigned long epoch = 1900;	/* year corresponding to 0x00 */

static const unsigned char days_in_mo[] =
    { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

#ifdef CONFIG_SONY_ICX1087
static struct mgscc_func_table g_mgscc_func = { NULL };
#endif /* CONFIG_SONY_ICX1087 */

void get_rtc_time(struct rtc_time *rtc_tm)
{
	spin_lock_irq(&rtc_lock);
	RTC_READ(GPIOCORE_PWC_SEC_RD, rtc_tm->tm_sec);
	RTC_READ(GPIOCORE_PWC_MIN_RD, rtc_tm->tm_min);
	RTC_READ(GPIOCORE_PWC_HOUR_RD, rtc_tm->tm_hour);
	RTC_READ(GPIOCORE_PWC_DAY_RD, rtc_tm->tm_mday);
	RTC_READ(GPIOCORE_PWC_MONTH_RD, rtc_tm->tm_mon);
	RTC_READ(GPIOCORE_PWC_YEAR_RD, rtc_tm->tm_year);
	RTC_READ(GPIOCORE_PWC_WEEK_RD, rtc_tm->tm_wday);
	spin_unlock_irq(&rtc_lock);

	rtc_tm->tm_mon &= 0x1f;	/* Mask of D7(19/20) */

	BCD_TO_BIN(rtc_tm->tm_sec);
	BCD_TO_BIN(rtc_tm->tm_min);
	BCD_TO_BIN(rtc_tm->tm_hour);
	BCD_TO_BIN(rtc_tm->tm_mday);
	BCD_TO_BIN(rtc_tm->tm_mon);
	BCD_TO_BIN(rtc_tm->tm_year);

	if ((rtc_tm->tm_year += (epoch - 1900)) <= 69) {
		rtc_tm->tm_year += 100;
	}

	rtc_tm->tm_mon--;
}

int set_rtc_time(struct rtc_time *rtc_tm)
{
	unsigned char mon, day, hrs, min, sec, leap_yr;
	unsigned char wday;
	unsigned int yrs;

	yrs = rtc_tm->tm_year + 1900;
	mon = rtc_tm->tm_mon + 1;	/* tm_mon starts at zero */
	day = rtc_tm->tm_mday;
	hrs = rtc_tm->tm_hour;
	min = rtc_tm->tm_min;
	sec = rtc_tm->tm_sec;
	wday = rtc_tm->tm_wday;

	if (yrs < 1970) {
		return -EINVAL;
	}

	leap_yr = ((!(yrs % 4) && (yrs % 100)) || !(yrs % 400));

	if ((mon > 12) || (day == 0)) {
		return -EINVAL;
	}

	if (day > (days_in_mo[mon] + ((mon == 2) && leap_yr))) {
		return -EINVAL;
	}

	if ((hrs >= 24) || (min >= 60) || (sec >= 60)) {
		return -EINVAL;
	}

	if ((yrs -= epoch) > 255) {
		return -EINVAL;
	}

	if (yrs > 169) {
		return -EINVAL;
	}

	if (yrs >= 100) {
		yrs -= 100;
	}

	BIN_TO_BCD(sec);
	BIN_TO_BCD(min);
	BIN_TO_BCD(hrs);
	BIN_TO_BCD(day);
	BIN_TO_BCD(mon);
	BIN_TO_BCD(yrs);

	spin_lock_irq(&rtc_lock);
	RTC_WRITE(yrs, GPIOCORE_PWC_YEAR);
	RTC_WRITE(mon, GPIOCORE_PWC_MONTH);
	RTC_WRITE(day, GPIOCORE_PWC_DAY);
	RTC_WRITE(hrs, GPIOCORE_PWC_HOUR);
	RTC_WRITE(min, GPIOCORE_PWC_MIN);
	RTC_WRITE(sec, GPIOCORE_PWC_SEC);
	RTC_WRITE(wday, GPIOCORE_PWC_WEEK);
	spin_unlock_irq(&rtc_lock);

	return 0;
}

#ifdef CONFIG_SONY_ICX1087
void get_rtc_alarm_time(struct rtc_time *rtc_tm)
{
#ifdef ALARM_W_ENABLE
    unsigned char wday = 0;

    if (rtc_tm->tm_wday > 0) {
        spin_lock_irq(&rtc_lock);
        RTC_READ(GPIOCORE_PWC_ALM_WM_RD, rtc_tm->tm_min);
        RTC_READ(GPIOCORE_PWC_ALM_WH_RD, rtc_tm->tm_hour);
        RTC_READ(GPIOCORE_PWC_ALM_WW_RD, wday);
        spin_unlock_irq(&rtc_lock);
 
        for(rtc_tm->tm_wday = 0; wday > 0; wday = wday >> 1)
            rtc_tm->tm_wday ++;
    }
    else{
        spin_lock_irq(&rtc_lock);
        RTC_READ(GPIOCORE_PWC_ALM_DM_RD, rtc_tm->tm_min);
        RTC_READ(GPIOCORE_PWC_ALM_DH_RD, rtc_tm->tm_hour);
        spin_unlock_irq(&rtc_lock);
        rtc_tm->tm_wday = 0;
    }
#else	// !ALARM_W_ENABLE
    spin_lock_irq(&rtc_lock);
    RTC_READ(GPIOCORE_PWC_ALM_DM_RD, rtc_tm->tm_min);
    RTC_READ(GPIOCORE_PWC_ALM_DH_RD, rtc_tm->tm_hour);
    spin_unlock_irq(&rtc_lock);
    rtc_tm->tm_wday = 0;
#endif	// ALARM_W_ENABLE

	BCD_TO_BIN(rtc_tm->tm_min);
	BCD_TO_BIN(rtc_tm->tm_hour);
}

/*Set alarm's time no matter that alarm is on or off*/
static int set_rtc_alarm_time(struct rtc_time *rtc_tm)
{
#ifdef ALARM_W_ENABLE
	unsigned char wday, hrs, min;
#else	// !ALARM_W_ENABLE
	unsigned char hrs, min;
#endif	// ALARM_W_ENABLE

	hrs  = rtc_tm->tm_hour;
	min  = rtc_tm->tm_min;

	if ((hrs >= 24) || (min >= 60)) {
		return -EINVAL;
	}

    BIN_TO_BCD(min);
    BIN_TO_BCD(hrs);
		
#ifdef ALARM_W_ENABLE
	wday = rtc_tm->tm_wday;

    spin_lock_irq(&rtc_lock);
    if (wday > 0) {
        wday = 1 << (wday-1);
        RTC_WRITE(hrs, GPIOCORE_PWC_ALM_WH);
        RTC_WRITE(min, GPIOCORE_PWC_ALM_WM);
        RTC_WRITE(wday, GPIOCORE_PWC_ALM_WW);
    }
    else {
        RTC_WRITE(hrs, GPIOCORE_PWC_ALM_DH);
        RTC_WRITE(min, GPIOCORE_PWC_ALM_DM);
    }
    spin_unlock_irq(&rtc_lock);
#else	// !ALARM_W_ENABLE
    spin_lock_irq(&rtc_lock);
    RTC_WRITE(hrs, GPIOCORE_PWC_ALM_DH);
    RTC_WRITE(min, GPIOCORE_PWC_ALM_DM);
    spin_unlock_irq(&rtc_lock);
#endif	// ALARM_W_ENABLE

	return 0;
}

static int set_rtc_alarm(struct rtc_time *rtc_tm)
{
#ifdef ALARM_W_ENABLE
	unsigned char wday, hrs, min;
#else	// !ALARM_W_ENABLE
	unsigned char hrs, min;
#endif	// ALARM_W_ENABLE

	hrs  = rtc_tm->tm_hour;
	min  = rtc_tm->tm_min;

	if ((hrs >= 24) || (min >= 60)) {
		return -EINVAL;
	}

    BIN_TO_BCD(min);
    BIN_TO_BCD(hrs);

#ifdef ALARM_W_ENABLE
	wday = rtc_tm->tm_wday;

    if (wday > 0) { /*Alarm_W*/
        wday = 1 << (wday-1);

        spin_lock_irq(&rtc_lock);
        /*Activate Alarm_W*/
        gpiocore_pwc_write(GPIOCORE_PWC_RTC1, 0x80, 0x80);

        RTC_WRITE(wday, GPIOCORE_PWC_ALM_WW);
        RTC_WRITE(hrs, GPIOCORE_PWC_ALM_WH);
        RTC_WRITE(min, GPIOCORE_PWC_ALM_WM);
        spin_unlock_irq(&rtc_lock);
        printk(KERN_INFO "The alarm_W is set\n");
	}
    else{  /*Alarm_D*/
        spin_lock_irq(&rtc_lock);
        /*Activate Alarm_D*/
        gpiocore_pwc_write(GPIOCORE_PWC_RTC1, 0x40, 0x40);

        RTC_WRITE(hrs, GPIOCORE_PWC_ALM_DH);
        RTC_WRITE(min, GPIOCORE_PWC_ALM_DM);
        spin_unlock_irq(&rtc_lock);

        printk(KERN_INFO "The alarm_D is set\n");
    }
#else	// !ALARM_W_ENABLE
    spin_lock_irq(&rtc_lock);
    /*Activate Alarm_D*/
    gpiocore_pwc_write(GPIOCORE_PWC_RTC1, 0x40, 0x40);

    RTC_WRITE(hrs, GPIOCORE_PWC_ALM_DH);
    RTC_WRITE(min, GPIOCORE_PWC_ALM_DM);
    spin_unlock_irq(&rtc_lock);
#endif	// ALARM_W_ENABLE

	return 0;
}

static void close_rtc_alarm(struct rtc_time *rtc_tm)
{
#ifdef ALARM_W_ENABLE
	if (rtc_tm->tm_wday > 0) { /*Alarm-W*/
        /*Deactivate Alarm_W*/
        spin_lock_irq(&rtc_lock);
        gpiocore_pwc_write(GPIOCORE_PWC_RTC1, 0x0, 0x80);
        spin_unlock_irq(&rtc_lock);
	}
    else{  /*Alarm_D*/
        /*Deactivate Alarm_D*/
        spin_lock_irq(&rtc_lock);
        gpiocore_pwc_write(GPIOCORE_PWC_RTC1, 0x0, 0x40);
        spin_unlock_irq(&rtc_lock);
    }
#else	// !ALARM_W_ENABLE
    /*Deactivate Alarm_D*/
    spin_lock_irq(&rtc_lock);
    gpiocore_pwc_write(GPIOCORE_PWC_RTC1, 0x0, 0x40);
    spin_unlock_irq(&rtc_lock);
#endif	// ALARM_W_ENABLE

}

/*
 * The interrupt handler of alarm_w/alarm_d
 */
static irqreturn_t alarm_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    /*Write 0 into alarm flag(WAFG/DAFG) to turn off alarm interrupt(INTRAB/INTRBB)*/
    gpiocore_pwc_write(GPIOCORE_PWC_RTC2, 0x0, 0x3); 
    printk(KERN_INFO "\t\t%s():\t The RTC alarm is triggered \n", __FUNCTION__);

	return IRQ_HANDLED;
}
#endif /*CONFIG_SONY_ICX1087*/

static int rtc_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		     unsigned long arg)
{
	struct rtc_time wtime, rtc_tm;
    struct rtc_wkalrm alarm;
    int ret = 0;

	switch (cmd) {
	case RTC_AIE_ON:	/* Allow alarm interrupts.      */
	case RTC_AIE_OFF:	/* Mask alarm int. enab. bit    */
	case RTC_UIE_ON:	/* Allow ints for RTC updates.  */
	case RTC_UIE_OFF:	/* Mask ints from RTC updates.  */
	case RTC_PIE_ON:	/* Allow periodic ints          */
	case RTC_PIE_OFF:	/* Mask periodic int. enab. bit */
	case RTC_WIE_ON:	/* Watchdog int. enable on      */
	case RTC_WIE_OFF:	/* Mask watchdog int. enab. bit */
	case RTC_IRQP_READ:	/* Read the periodic IRQ rate.  */
	case RTC_IRQP_SET:	/* Set periodic IRQ rate.       */
	case RTC_EPOCH_READ:	/* Read the epoch.              */
	case RTC_EPOCH_SET:	/* Set the epoch.               */
		return -EINVAL;
 
#ifdef CONFIG_SONY_ICX1087
   /*Alarm_W or Alarm_D is decided by arg's tw_wday field*/
	case RTC_ALM_SET:	/* Store a time into the alarm  */
        if (!capable(CAP_SYS_TIME)) 
            return -EACCES;

		if (copy_from_user(&rtc_tm, (struct rtc_time *)arg,
                           sizeof(struct rtc_time))) 
			return -EFAULT;

		if( g_mgscc_func.Process_SccCommand ) {
			g_mgscc_func.Process_SccCommand( cmd, &rtc_tm ) ;
		}

		return set_rtc_alarm_time(&rtc_tm);

	case RTC_ALM_READ:	/* Read the present alarm time  */
		if (copy_from_user(&wtime, (struct rtc_time *)arg,
                           sizeof(struct rtc_time))) 
			return -EFAULT;

		get_rtc_alarm_time(&wtime);

		if( g_mgscc_func.Process_SccCommand ) {
			g_mgscc_func.Process_SccCommand( cmd, &wtime ) ;
		}

		break;

	case RTC_WKALM_SET:	/* Set wakeup alarm             */
        if (!capable(CAP_SYS_TIME)) 
            return -EACCES;

		if (copy_from_user(&alarm, (struct rtc_wkalrm *)arg,
                           sizeof(struct rtc_wkalrm))) 
			return -EFAULT;

		if( g_mgscc_func.Process_SccCommand ) {
			g_mgscc_func.Process_SccCommand( cmd, &(alarm.time) ) ;
		}

        if (alarm.enabled == 0){
            close_rtc_alarm(&(alarm.time));
        }
        else if(alarm.enabled == 1){
            ret = set_rtc_alarm(&(alarm.time));
        }
        return ret;

	case RTC_WKALM_RD:	/* Get wakeup alarm             */
		if (copy_from_user(&alarm, (struct rtc_wkalrm *)arg,
                           sizeof(struct rtc_wkalrm))) 
			return -EFAULT;

		get_rtc_alarm_time(&(alarm.time));

        RTC_READ(GPIOCORE_PWC_RTC1_RD, alarm.enabled);

#ifdef ALARM_W_ENABLE
        if (alarm.time.tm_wday > 0)
            alarm.enabled = (alarm.enabled & 0x80) ? 1 : 0;
        else 
            alarm.enabled = (alarm.enabled & 0x40) ? 1 : 0;
#else	// !ALARM_W_ENABLE
        alarm.enabled = (alarm.enabled & 0x40) ? 1 : 0;
#endif	// ALARM_W_ENABLE
        
		if( g_mgscc_func.Process_SccCommand ) {
			g_mgscc_func.Process_SccCommand( cmd, &(alarm.time) ) ;
		}

        return copy_to_user((void *)arg, &alarm, sizeof(alarm))? -EFAULT : 0;
#endif  /* CONFIG_ICX1087_SONY*/
        
	case RTC_RD_TIME:	/* Read the time/date from RTC  */
		memset(&wtime, 0, sizeof(struct rtc_time));

#ifdef CONFIG_SONY_ICX1087
		if( g_mgscc_func.Process_SccCommand ) {
			g_mgscc_func.Process_SccCommand( cmd, &wtime ) ;
		}
		else {
			get_rtc_time(&wtime);
		}
#else
		get_rtc_time(&wtime);
#endif
		break;

	case RTC_SET_TIME:	/* Set the RTC                  */
		if (!capable(CAP_SYS_TIME)) {
			return -EACCES;
		}

		if (copy_from_user(&rtc_tm, (struct rtc_time *)arg,
                           sizeof(struct rtc_time))) {
			return -EFAULT;
		}

#ifdef CONFIG_SONY_ICX1087
        printk(KERN_INFO "RTC_SET_TIME before Process_SccCommand\n");
		if( g_mgscc_func.Process_SccCommand ) {
            printk(KERN_INFO "RTC_SET_TIME Process_SccCommand start\n");
			return g_mgscc_func.Process_SccCommand( cmd, &rtc_tm ) ;
		}
		else {
			return set_rtc_time(&rtc_tm);
		}
#else
		return set_rtc_time(&rtc_tm);
#endif

	default:
		return -EINVAL;
	}

    return copy_to_user((void *)arg, &wtime, sizeof(wtime)) ? -EFAULT : 0;
}

static int rtc_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int rtc_release(struct inode *inode, struct file *file)
{
	return 0;
}

static struct file_operations rtc_fops = {
	.owner = THIS_MODULE,
	.ioctl = rtc_ioctl,
	.open = rtc_open,
	.release = rtc_release,
};

static struct miscdevice rtc_dev = {
	RTC_MINOR,
	"rtc",
	&rtc_fops
};

static int __init rtc_init(void)
{
    unsigned int nRTC2RD = 0;
    struct rtc_time rtc_tm;

	printk(KERN_INFO "Starting rtc.\n");
/* 2006.06.07 ish */  
#ifdef CONFIG_SONY_ICX1087
	RTC_READ(GPIOCORE_PWC_RTC2_RD, nRTC2RD);

    //printk(KERN_INFO "RTC2RD=0x%02x\n", nRTC2RD);

    if ((nRTC2RD & 0x10)) {
        /* check XSTP bit */
        printk(KERN_INFO "  RTC stop detected! Reset RTC to 2006/01/01 00:00:00. \n");

        rtc_tm.tm_sec  = 0;
        rtc_tm.tm_min  = 0;
        rtc_tm.tm_hour = 0;
        rtc_tm.tm_mday = 1;
        rtc_tm.tm_mon  = 1 - 1;			/* 0-11 */
        rtc_tm.tm_year = 2006 - epoch;	/* offset from 1900 */
        rtc_tm.tm_wday = 0; 			/* Sunday */
        
        set_rtc_time(&rtc_tm);

        _n_xstp = 1;
    }

    /*2006.06.09 zha */
    request_irq(INT_PWC_RTCINT, alarm_interrupt, 0, "ICX1087_Alarm", 0);
#endif  /* CONFIG_SONY_ICX1087*/

	RTC_WRITE(0x00, GPIOCORE_PWC_TRIM);
	RTC_WRITE(0x00, GPIOCORE_PWC_RTC1);
	RTC_WRITE(0x20, GPIOCORE_PWC_RTC2);
    
	misc_register(&rtc_dev);

	return 0;
}

static void __exit rtc_exit(void)
{
	misc_deregister(&rtc_dev);

    free_irq(INT_PWC_RTCINT, 0);
}

module_init(rtc_init);
module_exit(rtc_exit);

#ifdef CONFIG_SONY_ICX1087
/* callback function register (called from mgscc.c) */
int register_mgscc_func( struct mgscc_func_table *mgscc_func )
{
    int ret = 0 ;

    if( NULL == mgscc_func ) {
        return -1 ;
    }
    
    if( NULL != mgscc_func->Process_SccCommand ) {
        g_mgscc_func.Process_SccCommand = mgscc_func->Process_SccCommand ;
    }
    return ret;
}

/* callback function unregister (called from mgscc.c) */
int unregister_mgscc_func( void )
{
    g_mgscc_func.Process_SccCommand = NULL ;
    return 0 ;
}

/*
 * Returns XSTP bit in startup
 * NOTE: XSTP shows if RTC has stopped or not.
 */
int get_xstp_in_startup(void)
{
    return _n_xstp;
}


EXPORT_SYMBOL( register_mgscc_func ) ;
EXPORT_SYMBOL( unregister_mgscc_func ) ;
EXPORT_SYMBOL( get_rtc_time ) ;
EXPORT_SYMBOL( set_rtc_time ) ;
EXPORT_SYMBOL( get_xstp_in_startup );
EXPORT_SYMBOL( get_rtc_alarm_time );

#endif /* CONFIG_SONY_ICX1087 */
