/*
 *  File Name       : mp200_ricoh.c
 *  Function        : adv7174 video encoder for NTSC I/F Driver
 *  Release Version : Ver 1.00
 *  Release Date    : 2006/03/24
 *
 *  Copyright (C) NEC Electronics Corporation 2005
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/device.h>
#include <linux/videodev.h>
#include <linux/video_encoder.h>

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/uaccess.h>

#include <asm/arch/gpiocore.h>
#include "mp200_ntsc.h"

#define I2C_NAME(x) (x)->name

/* GPIOCORE_PWC_PSCONT */
#define PWC_PSCNTVIDEO 0x01
/* GPIOCORE_PWC_MODESET */
#define PWC_VVIDEOPSRR 0x40
/* GPIOCORE_PWC_ENCCTRL0 */
#define PWC_VDACSLP 0x20
#define PWC_VDIGSLP 0x10
#define PWC_NPSEL   0x40

/* ----------------------------------------------------------------------- */
struct adv7174 {
	int norm;
	int enable;
};
static char adv7174_name[] = "Ricoh RC5T731 video encoder";
/* ----------------------------------------------------------------------- */
static struct ntsc_data *mp200_ntsc;
static struct encoder_reg* encoder_adv7174;

static int mp200_adv7174_command (unsigned int       cmd,
				  void *             arg)
{
	int ret = 0;
	struct adv7174* encoder = encoder_adv7174->private_data;

	if(encoder == NULL) {
		return -1;
	}
	switch (cmd) {
	case ENCODER_GET_CAPABILITIES:
	{
		struct video_encoder_capability *cap = arg;

		cap->flags = VIDEO_ENCODER_PAL |
			     VIDEO_ENCODER_NTSC;
		cap->inputs = 1;
		cap->outputs = 1;
	}
		break;
	case ENCODER_SET_NORM:
	{
		int iarg = *(int *) arg;
		switch (iarg) {
		case VIDEO_MODE_NTSC:
			ret = gpiocore_pwc_write(GPIOCORE_PWC_ENCCTRL0,
							0x0, PWC_NPSEL);
			break;
		case VIDEO_MODE_PAL:
			ret = gpiocore_pwc_write(GPIOCORE_PWC_ENCCTRL0,
							0xff, PWC_NPSEL);
			break;
		default:
			return -EINVAL;
		}
	}
		break;
	case ENCODER_SET_INPUT:
	{
		int *iarg = arg;
		/* not much choice of inputs */
		if (*iarg != 0) {
			return -EINVAL;
		}
	}
		break;
	case ENCODER_SET_OUTPUT:
	{
		int *iarg = arg;
		/* not much choice of outputs */
		if (*iarg != 0) {
			return -EINVAL;
		}
	}
		break;
	case ENCODER_ENABLE_OUTPUT:
	{
		int *iarg = arg;
		encoder->enable = !!*iarg;
	}
		break;
	default:
		return -EINVAL;
	}
	if (ret != 0) {
		return ret;
	}
	return 0;
}

static void mp200_adv7174_shutdown(void *dummy)
{
	struct adv7174* encoder = encoder_adv7174->private_data;

	if(encoder)
		kfree(encoder);
	encoder_adv7174->private_data = NULL;
	gpiocore_pwc_write(GPIOCORE_PWC_ENCCTRL0, 0x0, PWC_VDACSLP);
	gpiocore_pwc_write(GPIOCORE_PWC_MODESET, 0x0, PWC_VVIDEOPSRR);
	gpiocore_pwc_write(GPIOCORE_PWC_PSCONT, 0x0, PWC_PSCNTVIDEO);
	gpiocore_pwc_write(GPIOCORE_PWC_ENCCTRL0, 0x0, PWC_VDIGSLP);
	return;
}

static int mp200_adv7174_init(void *dummy)
{
	int ret = 0;
	struct adv7174 *encoder;

	encoder = kmalloc(sizeof(struct adv7174), GFP_KERNEL);
	if (encoder == NULL) {
		return -ENOMEM;
	}
	memset(encoder, 0, sizeof(struct adv7174));
	encoder->norm = VIDEO_MODE_NTSC;
	encoder->enable = 1;
	/*
	 * PSCNTVIDEO :40h D0='1'
	 * 500us : REGVIDEO up/down 500us/5ms
	 * VVIDEOPSRR :26h D6='1'
	 * 200us : twak : wake up 0.2ms
	 * VDACSLP    :D0h D5='1'
	 * VDIGSLP    :D0h D4='1'
	 */
	ret |= gpiocore_pwc_write(GPIOCORE_PWC_PSCONT, 0xff, PWC_PSCNTVIDEO);
	udelay(500);
	ret |= gpiocore_pwc_write(GPIOCORE_PWC_MODESET, 0xff, PWC_VVIDEOPSRR);
	udelay(200);
	ret |= gpiocore_pwc_write(GPIOCORE_PWC_ENCCTRL0, 0xff, PWC_VDACSLP);
	ret |= gpiocore_pwc_write(GPIOCORE_PWC_ENCCTRL0, 0xff, PWC_VDIGSLP);
	if (ret != 0) {
		kfree(encoder);
		return ret;
	}
	encoder_adv7174->private_data = encoder;

	return 0;
}

int adv7174_init (void)
{
	encoder_adv7174 = kmalloc(sizeof(struct encoder_reg), GFP_KERNEL);
	if (encoder_adv7174 == NULL) {
		return -ENOMEM;
	}
	memset(encoder_adv7174, 0, sizeof(struct encoder_reg));
	strcpy(encoder_adv7174->name, adv7174_name);
	encoder_adv7174->hw_init	= mp200_adv7174_init;
	encoder_adv7174->hw_shutdown	= mp200_adv7174_shutdown;
	encoder_adv7174->hw_command	= mp200_adv7174_command;
	/* register mp200 ntsc I/F */
	if((mp200_ntsc = mp200_ntsc_register(encoder_adv7174)) == NULL) {
		kfree(encoder_adv7174);
		return -ENXIO;
	}
	return 0;
}

void adv7174_exit (void)
{
	if(encoder_adv7174) {
		if(encoder_adv7174->private_data) {
			void* dummy = 0;
			mp200_adv7174_shutdown(dummy);
		}
		kfree(encoder_adv7174);
		encoder_adv7174 = NULL;
	}
	/* unregister devices */
	mp200_ntsc_unregister(mp200_ntsc);
}
