/*
 *  File Name       : mp200_ntsc.c
 *  Function        : NTSC I/F Driver
 *  Release Version : Ver 1.00
 *  Release Date    : 2006/03/24
 *
 *  Copyright (C) NEC Electronics Corporation 2005
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/ctype.h>
#include <linux/pagemap.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/videodev.h>
#include <linux/video_encoder.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/moduleparam.h>
#include <linux/cdev.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <asm/arch/gpiocore.h>
#include <asm/arch/pmu.h>
#include <asm/arch/ntsc_ioctl.h>
#include "mp200_ntsc.h"

#if	defined(MP200_NTSC_DEBUG)
#define FNC_ENTRY \
	if (debug == 1 || debug >= 9) {	\
		printk(KERN_DEBUG "entry:%s\n", __FUNCTION__); \
	}
#define FNC_EXIT \
	if (debug == 1 || debug >= 9) {	\
		printk(KERN_DEBUG "exit:%s:%d\n", __FUNCTION__ , __LINE__); \
	}
#define d0b(fmt, args...)	\
		printk(KERN_DEBUG "%s:%d: " fmt , __FUNCTION__ , __LINE__ , \
		       ## args);
static int debug = 0;
#else
#define FNC_ENTRY
#define FNC_EXIT
#define d0b(fmt, args...)
#endif
#define err(format, arg...) printk(KERN_ERR format , ## arg)
#define info(format, arg...) printk(KERN_INFO format , ## arg)
#define warn(format, arg...) printk(KERN_WARNING format , ## arg)
#define emerg(format, arg...) printk(KERN_EMERG format , ## arg)

#define NTSC_MODULE_NAME	"NTSC I/F Driver"
#define NTSC_NAME		"mp200_ntsc"
#define NTSCDEV_MAJOR		247 /* 240-254 LOCAL/EXPERIMENTAL USE */
#define NTSC_OUTPUT_OFF		(-1)

#define MP200_NTS_LINE_NTSC	(243*2)
#define MP200_NTS_LINE_PAL	(288*2)
#define MP200_NTS_HOFFSET	(720)
#define MP200_NTS_DEPTHBYTE	(2)
#define MP200_NTS_FRAGSIZE	(4*1024*203)
/* 4 * 1024(kbyte) * 203(page) = 831488 */;
/* 720(dot) * 576(line) * 2(byte) = 829440 */

int adv7174_init(void);
void adv7174_exit(void);

static irqreturn_t mp200_ntsc_callback(int, void*, struct pt_regs*);
static int mp200_ntsc_open(struct inode*, struct file*);
static int mp200_ntsc_close(struct inode*, struct file*);
static int mp200_ntsc_write(struct file*, const char*, size_t, loff_t *);
static int mp200_ntsc_mmap(struct file*, struct vm_area_struct*);
static int mp200_ntsc_ioctl(
	struct inode*, struct file*, unsigned int, unsigned long);
static int __init mp200_ntsc_probe(struct device*);
static int mp200_ntsc_suspend(struct device*, u32, u32);
static int mp200_ntsc_resume(struct device*, u32);

static int nbfrags = 2;
static int clkpol = 0;
static int endian = 0;
static unsigned long fragsize = MP200_NTS_FRAGSIZE;
static int irq = INT_NTS;

static struct class_simple *ntsc_class;
static struct cdev ntsc_cdev;
static struct device *mp200_ntsc_device;
static struct device_driver mp200_ntsc_driver = {
	.name		= "mp200_ntsc",
	.bus		= &platform_bus_type,
	.probe		= mp200_ntsc_probe,
	.remove		= NULL,
	.suspend	= mp200_ntsc_suspend,
	.resume		= mp200_ntsc_resume,
};
static struct file_operations ntscdev_fops = {
	.owner		= THIS_MODULE,
	.open		= mp200_ntsc_open,
	.release       	= mp200_ntsc_close,
	.write		= mp200_ntsc_write,
	.mmap		= mp200_ntsc_mmap,
	.ioctl          = mp200_ntsc_ioctl,
	.llseek         = no_llseek,
};
static struct ntsc_data ntsc_priv;
static struct mp200_ntsc_frame ntsc_frame__priv[NTSC_FRAME_MAX];

/* ------------------------ Queue routines --------------------- */

/* Inits the queue */
static inline void mp200_ntsc_selfrm_initq(struct mp200_ntsc_queue *queue)
{
	FNC_ENTRY

	queue->head = queue->tail = 0;
	queue->len = 0;
	spin_lock_init(&queue->s_lock);

	FNC_EXIT
}
/* Pulls an element from the queue */
static inline int mp200_ntsc_selfrm_pullq(struct mp200_ntsc_queue *queue)
{
	int result;
	unsigned long flags;
	FNC_ENTRY

	spin_lock_irqsave(&queue->s_lock, flags);
	if (!queue->len) {
		spin_unlock_irqrestore(&queue->s_lock, flags);
		FNC_EXIT return -1;
	}
	result = queue->buf[queue->head];
	queue->head++;
	queue->head &= (NTSC_QUEUE_SIZE - 1);
	queue->len--;
	spin_unlock_irqrestore(&queue->s_lock, flags);

	FNC_EXIT return result;
}

/* Pushes an element into the queue */
static inline void mp200_ntsc_selfrm_pushq(
	struct mp200_ntsc_queue *queue, int element)
{
	unsigned long flags;
	FNC_ENTRY

	spin_lock_irqsave(&queue->s_lock, flags);
	if (queue->len == NTSC_QUEUE_SIZE) {
		/* remove the first element */
		queue->head++;
		queue->head &= (NTSC_QUEUE_SIZE - 1);
		queue->len--;
	}
	queue->buf[queue->tail] = element;
	queue->tail++;
	queue->tail &= (NTSC_QUEUE_SIZE - 1);
	queue->len++;

	spin_unlock_irqrestore(&queue->s_lock, flags);

	FNC_EXIT
}

/* ------------------------ register routines --------------------- */

static void mp200_ntsc_output_mode(int mode, struct ntsc_data *ntsc)
{
	unsigned int val = 0;
	FNC_ENTRY

	switch (mode) {
	case 0:
		if (ntsc->background) {
			val = NTS_OUT_NTSOUT_BLACK;
		}
		else {
			val = NTS_OUT_NTSOUT_BLUE;
		}
		break;
	case 1:
		val = NTS_OUT_NTSOUT_ON;
		break;
	default:
		val = NTS_OUT_NTSOUT_OFF;
	}

	REG_NTS_OUT = val;

	FNC_EXIT return;
}

static inline int mp200_ntsc_index2frame(int frame_index)
{
	return frame_index + 1;
}

static inline int mp200_ntsc_frame2index(int frame)
{
	return frame - 1;
}

static inline int mp200_ntsc_select_frame(
	int frame_index, struct ntsc_data *ntsc)
{
	FNC_ENTRY

	ntsc->selectframe = mp200_ntsc_index2frame(frame_index);
	REG_NTS_FRAMESEL = ntsc->selectframe;

	FNC_EXIT return 0;
}

static int mp200_ntsc_encoder_command(
	struct ntsc_data *ntsc, int cmd, void* data)
{
	FNC_ENTRY

	if (ntsc->encoder->hw_command == NULL) {
		FNC_EXIT return -1;
	}

	FNC_EXIT return ntsc->encoder->hw_command(cmd, data);
}

static int mp200_ntsc_setup(struct ntsc_data *ntsc)
{
	int i, zero = 0;
	unsigned int val = 0;
	unsigned int line;
	unsigned int hoffset;
	unsigned int fragshalf;
	struct mp200_ntsc_frame* ntscframe;
	FNC_ENTRY

	if (!ntsc) {
		FNC_EXIT return -ENODEV;
	}
	mp200_ntsc_output_mode(NTSC_OUTPUT_OFF, ntsc);

	mp200_ntsc_encoder_command(ntsc, ENCODER_SET_NORM, &ntsc->norm);
	mp200_ntsc_encoder_command(ntsc, ENCODER_SET_INPUT, &zero);

	if (VIDEO_MODE_PAL == ntsc->norm) {
		val |= NTS_CONTROL_OUTMODE;
		line = MP200_NTS_LINE_PAL;
	}
	else {
		line = MP200_NTS_LINE_NTSC;
	}

	hoffset = MP200_NTS_HOFFSET;

	if (ntsc->upscale) {
		val |= NTS_CONTROL_UPSCALE;
		line /= 2;
		hoffset /= 2;
	}

	if (ntsc->clkpol) {
		val |= NTS_CONTROL_CLKPOL;
	}
	if (ntsc->endian) {
		val |= NTS_CONTROL_ENDIAN;
	}
	REG_NTS_SWRESET = NTS_SWRESET_MASK;
	REG_NTS_CONTROL = val;
	REG_NTS_SWRESET = 0;

	REG_NTS_HOFFSET = hoffset;

	i = ntsc->nbfrags;
	fragshalf = line * hoffset;

	ntscframe = ntsc->frame;

	REG_NTS_YAREAAD_A  = ntscframe[0].padr;
	REG_NTS_UVAREAAD_A = ntscframe[0].padr + fragshalf;

	if (ntsc->nbfrags > 1) {
		REG_NTS_YAREAAD_B  = ntscframe[1].padr;
		REG_NTS_UVAREAAD_B = ntscframe[1].padr + fragshalf;
	}
	if (ntsc->nbfrags > 2) {
		REG_NTS_YAREAAD_C  = ntscframe[2].padr;
		REG_NTS_UVAREAAD_C = ntscframe[2].padr + fragshalf;
	}

	ntsc->width	= hoffset;
	ntsc->height	= line;
	ntsc->depthbyte  = MP200_NTS_DEPTHBYTE;

	FNC_EXIT return 0;
}

static int mp200_ntsc_hardware_release(struct ntsc_data *ntsc)
{
	FNC_ENTRY

	mp200_ntsc_output_mode(NTSC_OUTPUT_OFF, ntsc);

	REG_NTS_INTENCLR = NTS_INT_MASK;
	REG_NTS_INTFFCLR = NTS_INT_MASK;

	free_irq(irq, (void *)ntsc);

	mp200_pmu_clkctrl_off(MP200_CLKCTRL_NTS);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_NTSPCLK);
	mp200_pmu_reset_device(MP200_RESETDEVICE_NTS);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_NTS_CLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_NTS_PCLK);

	if (ntsc->encoder->hw_shutdown) {
		int dummy;
		ntsc->encoder->hw_shutdown(&dummy);
	}

	FNC_EXIT return 0;
}

static int mp200_ntsc_hardware_init(struct ntsc_data *ntsc)
{
	int ret = 0;
	FNC_ENTRY

	mp200_pmu_clkctrl_off(MP200_CLKCTRL_NTS);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_NTSPCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_NTS_PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_NTS_CLK);
	mp200_pmu_unreset_device(MP200_RESETDEVICE_NTS);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_NTS);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_NTSPCLK);

	REG_NTS_INTENCLR = NTS_INT_MASK;
	REG_NTS_INTFFCLR = NTS_INT_MASK;

	if ((ret = request_irq(
		irq, mp200_ntsc_callback, SA_INTERRUPT, "NTSC", (void *)ntsc))
		 < 0) {
		d0b("request_irq ret FAULT : 0x%08x\n",ret);
		FNC_EXIT return ret;
	}

	REG_NTS_INTENSET = NTS_INT_MASK;

	if (ntsc->encoder->hw_init) {
		int dummy;
		ret = ntsc->encoder->hw_init(&dummy);
		if (ret) {
			FNC_EXIT return ret;
		}
	}
	mp200_ntsc_output_mode(NTSC_OUTPUT_OFF, ntsc);

	FNC_EXIT return ret;
}

/* ---------------------- local functions --------------------- */

static void* mp200_ntsc_dma_alloc(size_t size, dma_addr_t * dma_handle)
{
	int i = 0;
	int sz;
	void *ret;
	unsigned long adr;
	int gfp = GFP_KERNEL;
	FNC_ENTRY

	sz = get_order(size);
	ret = (void *) __get_free_pages(gfp, sz);
	size = PAGE_SIZE * (1 << sz);
	if (ret != NULL) {
		memset(ret, 0, size);
		*dma_handle = virt_to_phys(ret);
		adr = (unsigned long) ret;
		while (size > 0) {
			i++;
			SetPageReserved(virt_to_page((void *)adr));
			adr += PAGE_SIZE;
			size -= PAGE_SIZE;
		}
	}

	FNC_EXIT return ret;
}


static void mp200_ntsc_dma_free(void *vaddr, size_t size)
{
	int i = 0;
	int sz;
	unsigned long adr;
	FNC_ENTRY

	sz = get_order(size);
	size = PAGE_SIZE * (1 << sz);
	if (vaddr) {
		adr = (unsigned long) vaddr;
		while ((long) size > 0) {
			i++;
			ClearPageReserved(virt_to_page((void *)adr));
			adr += PAGE_SIZE;
			size -= PAGE_SIZE;
		}
		free_pages((unsigned long) vaddr, sz);
	}
	FNC_EXIT
}

static int mp200_ntsc_free_frame_buff(struct ntsc_data *ntsc)
{
	size_t size;
	FNC_ENTRY

	size = ntsc->fragsize * ntsc->nbfrags;
	mp200_ntsc_dma_free(ntsc->frame[0].vadr, size);
	ntsc->frame[0].vadr = NULL;
	ntsc->frame[0].padr = 0;

	FNC_EXIT return 0;
}

static int mp200_ntsc_allocate_frame_buff(struct ntsc_data *ntsc)
{
	int i;
	size_t size;
	FNC_ENTRY

	size = ntsc->fragsize * ntsc->nbfrags;
	if (!ntsc->frame) {
		FNC_EXIT return -ENOMEM;
	}
	while (!ntsc->frame[0].vadr) {
		i = 0;
		ntsc->frame[i].vadr = mp200_ntsc_dma_alloc(size,
			 (dma_addr_t*)&ntsc->frame[i].padr);
		if (!ntsc->frame[i].vadr) {
			err(NTSC_MODULE_NAME ": unable to allocate memory\n ");
			FNC_EXIT return -ENOMEM;
		}
		memzero(ntsc->frame[i].vadr, size);
		i++;
		for (; i < ntsc->nbfrags; i++) {
			ntsc->frame[i].vadr =
			ntsc->frame[0].vadr + i * ntsc->fragsize;
			ntsc->frame[i].padr =
			ntsc->frame[0].padr + i * ntsc->fragsize;
		}
		break;
	}

	FNC_EXIT return 0;
}

static int mp200_ntsc_queue_frame(int frame_index, struct ntsc_data *ntsc)
{
	unsigned long flags;
	struct mp200_ntsc_frame* ntscframe;
	FNC_ENTRY

	if (!ntsc->active) {
		FNC_EXIT return -EBUSY;
	}
	ntscframe = &ntsc->frame[frame_index];
	ntscframe->state = NTSC_BUF_READY;
	ntscframe->redisplay = -1;

	local_irq_save(flags);
	if (NTSC_FRAME_EMPTY != ntsc->selectframe) {
		local_irq_restore(flags);
		mp200_ntsc_selfrm_pushq(&ntsc->outputq, frame_index);
	}
	else {
		mp200_ntsc_select_frame(frame_index, ntsc);
		if (NTSC_FRAME_EMPTY == ntsc->outputframe) {
			local_irq_restore(flags);
			ntsc->videoout = 1;
			mp200_ntsc_output_mode((ntsc->videoout)?1:0, ntsc);
		}
		else {
			local_irq_restore(flags);
		}
	}

	FNC_EXIT return 0;
}

static int mp200_ntsc_frame_sync(
	int frame_index, struct ntsc_data *ntsc, unsigned int f_flags,
	int mode)
{
	int res = 0;
	struct mp200_ntsc_frame* frame;
	DECLARE_WAITQUEUE(wait, current);
	FNC_ENTRY

	frame = &ntsc->frame[frame_index];
	switch (frame->state) {
	case NTSC_BUF_UNUSED:
		FNC_EXIT return -EINVAL;
	case NTSC_BUF_READY:
	case NTSC_BUF_OUTPUT:
		add_wait_queue(&frame->proc_list, &wait);
		do {
			set_current_state(TASK_INTERRUPTIBLE);
			if (ntsc->ntscstatus) {
				res = -EFAULT;
				err("%s hardware fault.\n", NTSC_MODULE_NAME);
				break;
			}
			if (f_flags & (O_NONBLOCK|O_NDELAY)) {
				res = -EAGAIN;
				break;
			}
			if (NTSC_BUF_DONE == frame->state) {
				break;
			}
			if (signal_pending (current)) {
				res = -ERESTARTSYS;
				break;
			}
			schedule();
		} while (1);
		remove_wait_queue (&frame->proc_list, &wait);
		set_current_state(TASK_RUNNING);
	case NTSC_BUF_DONE:
		if(mode) { frame->state = NTSC_BUF_UNUSED; }
		if (res) { break; }
		else if (ntsc->endless) { break; }
		else if (ntsc->timeout <= frame->redisplay) {
			res = -ETIME;
		}
	}

	FNC_EXIT return res;
}

/* ---------------------- Interrupt handling ------------------- */

static irqreturn_t mp200_ntsc_callback(int irq, void *dev_id,
				    struct pt_regs *regs)
{
	struct ntsc_data *ntsc = (struct ntsc_data *)dev_id;
	struct mp200_ntsc_frame* ntscframe;
	int frame_index;
	unsigned int status;
	FNC_ENTRY

	status = REG_NTS_INTSTATUS;
	ntsc->ntscstatus =
		status & (NTS_INT_UNDERRUN | NTS_INT_DMASTOP | NTS_INT_DMAERR);
	if (ntsc->ntscstatus) {
		REG_NTS_INTENCLR = NTS_INT_MASK;
	}
	REG_NTS_INTFFCLR = status;

	if (NTSC_FRAME_EMPTY != ntsc->outputframe) {
		frame_index = mp200_ntsc_frame2index(ntsc->outputframe);
		ntscframe = &ntsc->frame[frame_index];
		if (
		!ntsc->videoout || (NTSC_FRAME_EMPTY != ntsc->selectframe)) {
			ntscframe->state = NTSC_BUF_DONE;
			ntsc->outputframe = NTSC_FRAME_EMPTY;
		}
		else if (ntsc->timeout > ntscframe->redisplay) {
			ntscframe->redisplay++;
		}
		else if (!ntsc->endless) {
			ntscframe->state = NTSC_BUF_DONE;
			ntsc->outputframe = NTSC_FRAME_EMPTY;
		}
		wake_up_interruptible(&ntscframe->proc_list);
	}
	if (NTSC_FRAME_EMPTY != ntsc->selectframe) {
		frame_index = mp200_ntsc_frame2index(ntsc->selectframe);
		ntscframe = &ntsc->frame[frame_index];
		ntscframe->state = NTSC_BUF_OUTPUT;
		ntsc->outputframe = ntsc->selectframe;
		ntsc->selectframe = NTSC_FRAME_EMPTY;
		wake_up_interruptible(&ntscframe->proc_list);
	}
	if (0 < (frame_index = mp200_ntsc_selfrm_pullq(&ntsc->outputq))) {
		mp200_ntsc_select_frame(frame_index, ntsc);
	}

	FNC_EXIT return IRQ_HANDLED;
}

static inline void mp200_ntsc_frame_init(
	int frame_index, struct ntsc_data *ntsc)
{
	struct mp200_ntsc_frame* ntscframe;
	FNC_ENTRY

	ntscframe = &ntsc->frame[frame_index];
	ntscframe->redisplay   = -1;
	ntscframe->state   = NTSC_BUF_UNUSED;
	init_waitqueue_head(&ntscframe->proc_list);

	FNC_EXIT return;
}

/* --------------------- external functions ------------------- */

static int mp200_ntsc_open(struct inode *inode, struct file *file)
{
	int res;
	struct ntsc_data *ntsc = &ntsc_priv;
	FNC_ENTRY

	if (ntsc->open_count > 0) {
		d0b("mp200_ntsc already open\n");
		FNC_EXIT return -EBUSY;
	}
	down(&ntsc->busy_lock);

	if(1 > nbfrags) nbfrags = 1;
	if(3 < nbfrags) nbfrags = 3;
	if(0 == fragsize) fragsize = MP200_NTS_FRAGSIZE;

	mp200_ntsc_selfrm_initq(&ntsc->outputq);
	ntsc->outputframe = NTSC_FRAME_EMPTY;
	ntsc->selectframe = NTSC_FRAME_EMPTY;
	ntsc->frame_index   = 0;
	ntsc->ntscstatus  = 0;
	ntsc->mapped	  = 0;
	ntsc->active	  = 0;
	ntsc->videoout	  = 0;

	ntsc->nbfrags = nbfrags;
	ntsc->clkpol = clkpol;
	ntsc->endian = endian;
	ntsc->fragsize = fragsize;

	if ((res = mp200_ntsc_allocate_frame_buff(ntsc))) {
	}
	else if ((res = mp200_ntsc_hardware_init(ntsc))) {
		mp200_ntsc_free_frame_buff(ntsc);
	}
	else if ((res = mp200_ntsc_setup(ntsc))) {
		mp200_ntsc_free_frame_buff(ntsc);
		mp200_ntsc_hardware_release(ntsc);
	}
	else {
		int i;
		ntsc->active = 1;
		for (i=0; ntsc->nbfrags > i; i++)
			mp200_ntsc_frame_init(i, ntsc);
		mp200_ntsc_output_mode((ntsc->videoout)?1:0, ntsc);
		file->private_data = ntsc;
		ntsc->open_count = 1;
	}

	up(&ntsc->busy_lock);

	FNC_EXIT return res;
}

static int mp200_ntsc_close(struct inode *inode, struct file *file)
{
	int i;
	struct ntsc_data *ntsc = file->private_data;
	FNC_ENTRY

	ntsc->videoout = 0;
	for (i=0; ntsc->nbfrags > i; i++) {
		mp200_ntsc_frame_sync(i, ntsc, 0, 1);
	}
	ntsc->open_count = 0;

	/* clean up capture-buffers */
	mp200_ntsc_free_frame_buff(ntsc);
	mp200_ntsc_hardware_release(ntsc);

	FNC_EXIT return 0;
}

static int mp200_ntsc_write(struct file *file, const char *buf,
				       size_t count, loff_t * ppos)
{
	struct ntsc_data *ntsc = file->private_data;
	int res = 0;
	char *vadr;
	FNC_ENTRY

	if (!buf) {
		d0b("buf NULL\n");
		FNC_EXIT return -EINVAL;
	}
	if (1 == ntsc->nbfrags && ntsc->videoout) {
		FNC_EXIT return 0;
	}
	if (count != ntsc->width * ntsc->height * ntsc->depthbyte) {
		FNC_EXIT return -EINVAL;
	}
	/* make this _really_ smp and multithread-safe */
	if (down_interruptible(&ntsc->busy_lock)) {
		FNC_EXIT return -EINTR;
	}
	vadr = ntsc->frame[ntsc->frame_index].vadr;

	if(NTSC_BUF_UNUSED != ntsc->frame[ntsc->frame_index].state) {
		res = mp200_ntsc_frame_sync(ntsc->frame_index, ntsc ,
							file->f_flags, 1);
	}
	if (-ETIME == res) res = 0;
	if (res)  count = res;
	else if (copy_from_user(vadr, buf, count)) {
		d0b("copy_from_user failed\n");
		count = -EFAULT;
	}
	else {
		res = mp200_ntsc_queue_frame(ntsc->frame_index, ntsc);
		if (res) {
			count = res;
		}
		else {
			ntsc->frame_index += 1;
			if (ntsc->frame_index >= ntsc->nbfrags)
				ntsc->frame_index = 0;
		}
	}
	up(&ntsc->busy_lock);

	FNC_EXIT return count;
}

static int mp200_ntsc_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	struct ntsc_data *ntsc = file->private_data;
	unsigned long flags;
	int setting = 0;
	int res = 0;
	FNC_ENTRY

	/* make this _really_ smp and multithread-safe */
	if (down_interruptible(&ntsc->busy_lock)) {
		FNC_EXIT return -EINTR;
	}
	switch (cmd) {
	case NTSC_SET_OUTMODE:
	{
		int i;

		if (get_user(i, (int *)arg)) {
			d0b("get_user FAULT\n");
			res = -EFAULT;
			break;
		}
		switch(i) {
		case 0: /* NTSC 525/60 size:720x486 */
			ntsc->upscale = 0;
			ntsc->norm    = VIDEO_MODE_NTSC;
			break;
		case 1: /* PAL  625/60 size:720x576 */
			ntsc->upscale = 0;
			ntsc->norm    = VIDEO_MODE_PAL;
			break;
		case 2: /* NTSC 525/60 size:360x243 */
			ntsc->upscale = 1;
			ntsc->norm    = VIDEO_MODE_NTSC;
			break;
		case 3: /* PAL   625/60 size:360x288 */
			ntsc->upscale = 1;
			ntsc->norm    = VIDEO_MODE_PAL;
			break;
		default:
			res = -EINVAL;
		}
		setting = 2;
		break;
	}
	case NTSC_GET_OUTMODE:
	{
		int i = 0;

		if (VIDEO_MODE_NTSC == ntsc->norm) {
			i = 0; /* NTSC 525/60 size:720x486 */
			if (ntsc->upscale) {
				i = 2; /* NTSC 525/60 size:360x243 */
			}
		}
		else {
			i = 1; /* PAL 625/60 size:720x576 */
			if (ntsc->upscale) {
				i = 3; /* PAL  625/60 size:360x288 */
			}
		}
		if (put_user(i,(int*)arg)) {
			d0b("put_user FAULT\n");
			res = -EFAULT;
		}
		break;
	}
	case NTSC_SET_TIMEOUT:
	{
		int i;

		if (get_user(i, (int *)arg)) {
			d0b("get_user FAULT\n");
			res = -EFAULT;
			break;
		}
		local_irq_save(flags);
		if (0 > i) {
			ntsc->endless = 1;
			ntsc->timeout = 0;
		}
		else {
			ntsc->endless = 0;
			ntsc->timeout = i;
		}
		local_irq_restore(flags);
		break;
	}
	case NTSC_GET_TIMEOUT:
	{
		int i = 0;

		if(ntsc->endless) {
			i = -1;
		}
		else {
			i = ntsc->timeout;
		}
		if (put_user(i,(int*)arg)) {
			d0b("put_user FAULT\n");
			res = -EFAULT;
		}
		break;
	}
	case NTSC_BLUEBACK:
	{
		ntsc->background = 0;
		setting = 1;
		break;
	}
	case NTSC_BLACKBACK:
	{
		ntsc->background = 1;
		setting = 1;
		break;
	}
	case NTSC_GMBUF:
	{
		struct ntsc_mbuf vm;
		int i;

		memset(&vm, 0 , sizeof(vm));
		vm.vsize = ntsc->fragsize * ntsc->nbfrags;
		vm.nframe = ntsc->nbfrags;
		for (i = 0; i < ntsc->nbfrags; i++) {
			vm.frame[i].voffset = i * ntsc->fragsize;
			vm.frame[i].paddr = ntsc->frame[0].padr
						+ (i * ntsc->fragsize);
		}
		if (copy_to_user((void*)arg, &vm, sizeof(vm))) {
			d0b("copy to user FAULT\n");
			res = -EFAULT;
		}
		break;
	}
	case NTSC_SYNC:
	{
		int i;

		if (get_user(i, (int *)arg)) {
			d0b("get_user FAULT\n");
			FNC_EXIT return -EFAULT;
		}
		if (i < 0 || i >= ntsc->nbfrags) {
			res = -EINVAL;
			break;
		}
		res = mp200_ntsc_frame_sync(i, ntsc ,file->f_flags, 1);
		break;
	}
	case NTSC_QBUF:
	{
		int i;
		struct mp200_ntsc_frame* frame;

		if (get_user(i, (int *)arg)) {
			d0b("get_user FAULT\n");
			res = -EFAULT;
			break;
		}
		if (i < 0 || i >= ntsc->nbfrags) {
			res = -EINVAL;
			break;
		}
		frame = &ntsc->frame[i];

		if (frame->state != NTSC_BUF_UNUSED) {
			res = -EBUSY;
			break;
		}
		res = mp200_ntsc_queue_frame(i, ntsc);
		break;
	}

	default:
		res = -ENOIOCTLCMD;
	} /* switch */

	if (!res && setting) {
		int i;

		ntsc->active = 0;
		ntsc->videoout = 0;
		for (i=0; ntsc->nbfrags > i; i++) {
			mp200_ntsc_frame_sync(i, ntsc, 0, 0);
		}
		if (2 == setting)
			mp200_ntsc_setup(ntsc);
		ntsc->active = 1;
		mp200_ntsc_output_mode((ntsc->videoout)?1:0, ntsc);
	}
	up(&ntsc->busy_lock);

	FNC_EXIT return res;
}

static int mp200_ntsc_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct ntsc_data *ntsc = file->private_data;
	unsigned long start = vma->vm_start;
	unsigned long size  = vma->vm_end - vma->vm_start;
	unsigned long pos;
	int retval;
	FNC_ENTRY

	if (size > ntsc->fragsize * ntsc->nbfrags) {
		FNC_EXIT return -EINVAL;
	}
	/* make this _really_ smp-safe */
	if (down_interruptible(&ntsc->busy_lock)) {
		FNC_EXIT return -EINTR;
	}
	pos = (unsigned long)(ntsc->frame[0].padr);
	retval =
	remap_pfn_range(
		vma, start, pos >> PAGE_SHIFT, size, vma->vm_page_prot);
	if (retval) {
		up(&ntsc->busy_lock);
		FNC_EXIT return retval;
	}
	ntsc->mapped = 1;

	up(&ntsc->busy_lock);

	FNC_EXIT return 0;
}

static int __init mp200_ntsc_probe(struct device *dev)
{
	FNC_ENTRY

	mp200_ntsc_device = dev;

	FNC_EXIT return 0;
}
static int mp200_ntsc_suspend(struct device * dev, u32 state, u32 level)
{
	struct ntsc_data *ntsc = &ntsc_priv;
	FNC_ENTRY

	if (ntsc->open_count > 0) {
		FNC_EXIT return -EBUSY;
	}

	FNC_EXIT return 0;
}
static int mp200_ntsc_resume(struct device * dev, u32 level)
{
	FNC_ENTRY
	FNC_EXIT return 0;
}

struct ntsc_data* mp200_ntsc_register(struct encoder_reg* encoder)
{
	struct ntsc_data *ntsc = &ntsc_priv;
	FNC_ENTRY

	if (!encoder) {
		FNC_EXIT return NULL;
	}
	memset(ntsc, 0, sizeof(struct ntsc_data));

	ntsc->open_count = 0;

	init_MUTEX(&ntsc->busy_lock);
	ntsc->frame = ntsc_frame__priv;
	memset(ntsc->frame, 0, (sizeof(MP200_NTSC_FRAME) * NTSC_FRAME_MAX));

	ntsc->outputframe	= NTSC_FRAME_EMPTY;
	ntsc->selectframe	= NTSC_FRAME_EMPTY;
	ntsc->frame_index	= 0;
	ntsc->ntscstatus	= 0;
	ntsc->timeout		= 0;

	ntsc->norm		= VIDEO_MODE_NTSC;
	ntsc->encoder		= encoder;
	ntsc->width		= MP200_NTS_LINE_NTSC;
	ntsc->height		= MP200_NTS_HOFFSET;
	ntsc->depthbyte		= MP200_NTS_DEPTHBYTE;

	ntsc->background	= 0;	/* 1: brackback 0: buleback */
	ntsc->upscale		= 0;	/* 0: nomal 1: upscale */
	ntsc->mapped		= 0;	/* mmap()'ed */
	ntsc->active		= 0;	/* actually in progressi */
	ntsc->endless		= 0;
	ntsc->videoout		= 0;

	FNC_EXIT return ntsc;
}

void mp200_ntsc_unregister(struct ntsc_data *ntsc)
{
	FNC_ENTRY
	FNC_EXIT
}

static int __init mp200_init_ntsc(void)
{
	FNC_ENTRY

	info("Starting %s.\n", "ADV7174 video encoder driver");

	adv7174_init();

	register_chrdev_region(MKDEV(NTSCDEV_MAJOR, 0), 1, "ntsc");
	cdev_init(&ntsc_cdev, &ntscdev_fops);
	ntsc_cdev.owner = THIS_MODULE;
	kobject_set_name(&ntsc_cdev.kobj, "ntsc");
	if (cdev_add(&ntsc_cdev, MKDEV(NTSCDEV_MAJOR, 0), 1)) {
		d0b("ntsc: unable to add char device");
		unregister_chrdev_region(MKDEV(NTSCDEV_MAJOR, 0), 1);
		FNC_EXIT return -EIO;
	}
	devfs_mk_cdev(
		MKDEV(NTSCDEV_MAJOR, 0), S_IFCHR|S_IRUGO|S_IWUGO, "ntsc");
	driver_register(&mp200_ntsc_driver);
	ntsc_class = class_simple_create(THIS_MODULE, "ntsc");
	if (IS_ERR(ntsc_class)) {
		driver_unregister(&mp200_ntsc_driver);
		devfs_remove("ntsc");
		cdev_del(&ntsc_cdev);
		unregister_chrdev_region(MKDEV(NTSCDEV_MAJOR, 0), 1);
		FNC_EXIT return PTR_ERR(ntsc_class);
	}
	class_simple_device_add(
		ntsc_class, MKDEV(NTSCDEV_MAJOR, 0), mp200_ntsc_device,
		"ntsc");

	FNC_EXIT return 0;
}

static void __exit mp200_close_ntsc(void)
{
	FNC_ENTRY

	class_simple_device_remove(MKDEV(NTSCDEV_MAJOR, 0));
	class_simple_destroy(ntsc_class);
	driver_unregister(&mp200_ntsc_driver);
	devfs_remove("ntsc");
	cdev_del(&ntsc_cdev);
	unregister_chrdev_region(MKDEV(NTSCDEV_MAJOR, 0), 1);

	adv7174_exit();

	FNC_EXIT
}

module_init(mp200_init_ntsc);
module_exit(mp200_close_ntsc);

#ifdef MODULE
#ifdef MP200_NTSC_DEBUG
module_param(debug, int, 0644);
#endif
module_param(nbfrags, int, 0644);
module_param(clkpol, int, 0644);
module_param(endian, int, 0644);
module_param(fragsize, long, 0644);
MODULE_AUTHOR("2006 NEC ELECTRONICS CORPORATION");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION(NTSC_MODULE_NAME);
MODULE_SUPPORTED_DEVICE("");
#endif /* MODULE */

