/*
 *  File Name       : linux/drivers/char/mp200_spi_kern.c
 *  Function        : MP200 SPI interface
 *  Release Version : Ver 0.01
 *  Release Date    : 2006/06/02
 *
 *  Copyright (C) NEC Electronics Corporation 2005-2006
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

/*
 * Copyright 2006,2007 Sony Corporation. 
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>
/* 2006.10.10 has */
#ifdef CONFIG_SONY_ICX1087
#include <linux/device.h>
#endif

#include <asm/io.h>
#include <asm/arch/smu.h>
#include <asm/arch/pmu.h>
/* 2006.10.10 has */
#ifdef CONFIG_SONY_ICX1087
#include <asm/arch/pwc.h>
#endif

#define SPI_DEBUG_PARAM

#include "mp200_spi_kern.h"

#ifdef SPI_DEBUG
#define DPRINTK(fmt, args...) printk(KERN_INFO "\n  %s: %s: " fmt,__FILE__,__FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif

#define RETRY_COUNT(x)  ((loops_per_jiffy * x)/(1000000/HZ)) /* x us */

static unsigned int spi_clock_table[] = {
    400,            /* 400KHz */
	1500,			/* 1500KHz */
	3000,			/* 3MHz */
	6000,			/* 6MHz */
	12000,			/* 12MHz */
	24000,			/* 24MHz */
	48000,			/* 48MHz */
};

/* spi register info (SPI0) */
static spi_regs_t regs_spi0 = {
	.mode = &SPx_MODE(MP200_SP0_BASE),
	.pol = &SPx_POL(MP200_SP0_BASE),
	.control = &SPx_CONTROL(MP200_SP0_BASE),
	.tx_tim = &SPx_TX_TIM(MP200_SP0_BASE),
	.tx_data = &SPx_TX_DATA(MP200_SP0_BASE),
	.rx_data = &SPx_RX_DATA(MP200_SP0_BASE),
	.status = &SPx_STATUS(MP200_SP0_BASE),
	.raw_status = &SPx_RAW_STATUS(MP200_SP0_BASE),
	.enset = &SPx_ENSET(MP200_SP0_BASE),
	.enclr = &SPx_ENCLR(MP200_SP0_BASE),
	.ffclr = &SPx_FFCLR(MP200_SP0_BASE),
	.tx_fifo_p = &SPx_TX_FIFO_P(MP200_SP0_BASE),
	.rx_fifo_p = &SPx_RX_FIFO_P(MP200_SP0_BASE),
	.control2 = &SPx_CONTROL2(MP200_SP0_BASE),
};

/* spi register info (SPI1) */
static spi_regs_t regs_spi1 = {
	.mode = &SPx_MODE(MP200_SP1_BASE),
	.pol = &SPx_POL(MP200_SP1_BASE),
	.control = &SPx_CONTROL(MP200_SP1_BASE),
	.tx_tim = &SPx_TX_TIM(MP200_SP1_BASE),
	.tx_data = &SPx_TX_DATA(MP200_SP1_BASE),
	.rx_data = &SPx_RX_DATA(MP200_SP1_BASE),
	.status = &SPx_STATUS(MP200_SP1_BASE),
	.raw_status = &SPx_RAW_STATUS(MP200_SP1_BASE),
	.enset = &SPx_ENSET(MP200_SP1_BASE),
	.enclr = &SPx_ENCLR(MP200_SP1_BASE),
	.ffclr = &SPx_FFCLR(MP200_SP1_BASE),
	.tx_fifo_p = &SPx_TX_FIFO_P(MP200_SP1_BASE),
	.rx_fifo_p = &SPx_RX_FIFO_P(MP200_SP1_BASE),
	.control2 = &SPx_CONTROL2(MP200_SP1_BASE),
};

static spi_data_t spi_private[] = {
	{
	 .regs = &regs_spi0,
	 .pol  = SPI_POL_VAL_SPI0,
	 .pctl_spi = SPI_SMU_PCTL_SPI0_SK,
	 .div_sclk = SMU_SPI0_SCLK_DIV,
	 .lock = SPIN_LOCK_UNLOCKED,
	 },
	{
	 .regs = &regs_spi1,
	 .pol  = SPI_POL_VAL_SPI1,
	 .pctl_spi = SPI_SMU_PCTL_SPI1_SK,
	 .div_sclk = SMU_SPI1_SCLK_DIV,
	 .lock = SPIN_LOCK_UNLOCKED,
	 },
};

static inline void spi_sftrst_wait(void)
{
	unsigned int x = (inl(SMU_SPI0_SCLK_DIV) * 4);
	while (x--) { __asm__ __volatile__ (""); }
}

static inline int spi_get_clock_source(void)
{
	unsigned long plln, pllm, pllp;
	unsigned long pllclk, nrm_div;
	static const unsigned long smu_table_lplln[] = {
	  122880, 126976, 131072, 135168, 139264, 143360,
	  147456, 151552, 155648, 159744, 163840
	};
	int clock;

	switch (inl(SMU_PLLSEL) & MASK_SMU_PLLSEL) {
	case SMU_PLLSEL_OSC:
		/* 12MHz */
		clock = 12000;
		break;

	case SMU_PLLSEL_RTC:
		/* 32KHz */
		clock = 32;
		break;

	case SMU_PLLSEL_PLL:
		/* PLL */
		if(inl(SMU_PLLSEL2) & SMU_NRM_CK_SEL){ /* LPLL */
		  plln = inl(SMU_LPLL_CTRL1);
		  pllclk = smu_table_lplln[plln];
		}
		else /* HPLL */
		  {
		    plln = inl(SMU_HPLL_CTRL1) + 1;
		    pllm = inl(SMU_HPLL_CTRL2) + 1;
		    pllp = inl(SMU_HPLL_CTRL3);

		    pllclk = (12000 * plln) / pllm;
		    if(pllp != 0){
		      pllclk >>= 1;
		    }
		  }
		nrm_div = (inl(SMU_AUTO_FRQ_CHANGE) & 0xF000000) >> 24;
		clock = pllclk / (nrm_div + 1);

		break;

	default:
		/* 12MHz */
		clock = 12000;
		break;
	}
#ifdef SPI_DEBUG
	printk(KERN_INFO "%s(): clock = %d[KHz]\n", __FUNCTION__, clock);
#endif

	return clock;

}

static inline unsigned int spi_set_sclk(spi_data_t * spi, unsigned int offset)
{
	unsigned int sclk, source;
	int divsclk, i = 0;

	if (spi == NULL) {
		return -EINVAL;
	}

	if (offset > SPI_SCLK_48MHZ) {
		offset = SPI_SCLK_48MHZ;
	}
	sclk = spi_clock_table[offset];     /* chg 2006.05.29 */
	source = spi_get_clock_source();
	divsclk = (source / sclk) - 1;
	if (divsclk < 1) {
		divsclk = 1;
		sclk = source / (divsclk + 1);
	} else if (divsclk > 255) {
		divsclk = 255;
		sclk = source / (divsclk + 1);
        DPRINTK("The required frequency is too low(out of range). Set to lowest %d[KHz]\n", sclk);
	}
	outl(divsclk, spi->div_sclk);

	DPRINTK("sclk(1/%d) = %d[KHz]\n", (divsclk + 1), sclk);

	return i;
}

static inline int send_data(spi_regs_t *spi, unsigned int data, unsigned int cnt)
{
	*spi->tx_data = data;
	*spi->control = (SPx_CONTROL_WRT | SPx_CONTROL_START);
	while ((*spi->control & SPx_CONTROL_START) && (--cnt > 0)){}
		;
	if(cnt == 0) {
	  DPRINTK(" timeout! \n");
	  return -1; /* timeout */
	}
	return 0;
}

int
spi_write_without_intr(const SPI_CONFIG *config, const char *buf, int rst_flag)
{
	spi_data_t *s;
	unsigned long flag;
	unsigned int count, time, nb = 0;
	int i, ret = 0;

	if (config == 0) return -EINVAL;

#ifdef SPI_DEBUG_PARAM
	/* dev check*/
	if (config->dev != SPI_DEV_SPI0 && config->dev != SPI_DEV_SPI1) {
		return -EINVAL;
	}
#endif

	s = &spi_private[config->dev];
	nb = ((config->mode & SPx_MODE_NB_MASK) >> 8) + 1;

#ifdef SPI_DEBUG_PARAM
	/* m_s check */
	if (config->mode & SPx_MODE_M_S_MASK) { /* slave */
		return -EINVAL;
	}
	
	/* dma check */
	if (config->mode & SPx_MODE_DMA_MASK) { /* dma on */
		return -EINVAL;
	}
	
	/* nb check */
	if ((nb < SPI_NB_8BIT) || (nb > SPI_NB_32BIT)) {
		return -EINVAL;
	}
	
	/* sclk check */
	switch (config->sclk) {
    case SPI_SCLK_400KHZ:
	case SPI_SCLK_1500KHZ:
	case SPI_SCLK_3MHZ:
	case SPI_SCLK_6MHZ:
	case SPI_SCLK_12MHZ:
	case SPI_SCLK_24MHZ:
	case SPI_SCLK_48MHZ:
		break;
	default:
		return -EINVAL;
	}
	
	/* reset flag check */
	if ((rst_flag != SPI_RESET_OFF) && (rst_flag != SPI_RESET_ON)) {
		return -EINVAL;
	}
#endif

	count = config->size;
	time = RETRY_COUNT(
		(((100000 / spi_clock_table[config->sclk]) * nb) / 10) + 1);

	spin_lock_irqsave(&s->lock, flag); /* lock */

	spi_set_sclk(s, config->sclk);
	/* clear */
	*s->regs->ffclr = SPx_FFCLR_ALL_CLR;
	*s->regs->enclr = SPx_ENCLR_ALL_MASK;
	/* parameters */
	*s->regs->mode	= config->mode;
	*s->regs->pol	= config->pol;

	/* spi soft reset */
	if (rst_flag == SPI_RESET_ON) { /* reset on */
		*s->regs->control |= SPx_CONTROL_RST;
		spi_sftrst_wait();
		*s->regs->control &= ~SPx_CONTROL_RST;
	}
	if (buf == 0) {
		goto out;
	}

	/* send */
	if (nb > (SPI_NB_8BIT + SPI_NB_16BIT)) { /* 25~32 */
		unsigned int *data = (unsigned int *)buf;
		count = count/4;
		for (i = 0; i < count; i++) {
			if(send_data(s->regs, data[i], time)) {
				ret = -EBUSY;
				goto out;
			}
		}
	}
	else if (nb > SPI_NB_16BIT) { /* 17~24*/
		unsigned char *data = (unsigned char *)buf;
		count = (count / 3) * 3;
		for (i = 0; i < count; i+=3) {
			if(send_data(s->regs, ((data[i] << 16) 
				| (data[i+1] << 8) | (data[i+2])), time)) {
				ret = -EBUSY;
				goto out;
			}
		}
	}
	else if ((nb > SPI_NB_8BIT) && (nb <= SPI_NB_16BIT)) { /* 9~16*/
		unsigned short *data = (unsigned short *)buf;
		count = count/2;
		for (i = 0; i < count; i++) {
			if(send_data(s->regs, data[i], time)) {
				ret = -EBUSY;
				goto out;
			}
		}
	}
	else if (nb == SPI_NB_8BIT) {  /* 8 */
		unsigned char *data = (unsigned char *)buf;
		for (i = 0; i < count; i++) {
			if(send_data(s->regs, data[i], time)) {
				ret = -EBUSY;
				goto out;
			}
		}
	}
out:
	/* end */
	spin_unlock_irqrestore(&s->lock, flag); /* unlock */
	
	return ret;
}

/* 2006.10.05 has */
#ifdef CONFIG_SONY_ICX1087

static void spi_shutdown(struct device *dev)
{
	spi_data_t *s;

	/* set pin to float */
	s = &spi_private[SPI_DEV_SPI0];
	outl(inl(SMU_PCTL_SPI) | (s->pctl_spi), SMU_PCTL_SPI);

	/* pwc_regulator_off(PWC_PSCNTGP5); */
	pwc_regulator_off(PWC_PSCNTHP);

	return;
}

static int spi_suspend(struct device *dev, u32 state, u32 level)
{
	spi_data_t *s;

	if(level==SUSPEND_POWER_DOWN){

		/* set pin to float */
		s = &spi_private[SPI_DEV_SPI0];
		outl(inl(SMU_PCTL_SPI) | (s->pctl_spi), SMU_PCTL_SPI);

		/* set SPI0 all pin to pull down */
		mp200_pmu_write_reg(
			SMU_PULLCTL_SPI,
			SPI_SMU_PULLCTL_SPI0_DSLEEP,
			SPI_SMU_PULLCTL_SPI0_MASK
		);

		/* pwc_regulator_off(PWC_PSCNTGP5); */
		pwc_regulator_off(PWC_PSCNTHP);
	}

	return(0);
}

static int spi_resume(struct device *dev, u32 level)
{
	spi_data_t *s;

	if(level==RESUME_POWER_ON){

		/* pwc_regulator_on(PWC_PSCNTGP5); */
		pwc_regulator_on(PWC_PSCNTHP);

		/* set SPI0 pin(except SI) to none pull control */
		mp200_pmu_write_reg(
			SMU_PULLCTL_SPI,
			SPI_SMU_PULLCTL_SPI0_NORMAL,
			SPI_SMU_PULLCTL_SPI0_MASK
		);

		/* set pin to function */
		s = &spi_private[SPI_DEV_SPI0];
		outl(inl(SMU_PCTL_SPI) & ~(s->pctl_spi), SMU_PCTL_SPI);
	}

	return(0);
}

struct device_driver spi_driver = {
	.name =     "spi",
	.bus =      &platform_bus_type,
	.shutdown = spi_shutdown,
	.suspend =  spi_suspend,
	.resume =   spi_resume,
};

#endif

static int __init spi_init(void)
{
	spi_data_t *s;

	printk(KERN_INFO "Starting %s.\n", SPI_NAME);

/* 2006.10.05 has */
#ifdef CONFIG_SONY_ICX1087
	driver_register(&spi_driver);
#endif

	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SPI0SCLKLP);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SPI1SCLKLP);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SPI0PCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SPI1PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SPI0_PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SPI1_PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SPI0_SCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SPI1_SCLK);
	mp200_pmu_unreset_device(MP200_RESETDEVICE_SPI0);
	mp200_pmu_unreset_device(MP200_RESETDEVICE_SPI1);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SPI0PCLK);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SPI1PCLK);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SPI0SCLKLP);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_SPI1SCLKLP);

/* 2006.10.05 has */
#ifdef CONFIG_SONY_ICX1087
	/* pwc_regulator_on(PWC_PSCNTGP5); */
	pwc_regulator_on(PWC_PSCNTHP);
#endif

	s = &spi_private[SPI_DEV_SPI0];
	/* SPI_SMU_PCTL_SPIn_SK clear */
	outl(inl(SMU_PCTL_SPI) & ~(s->pctl_spi), SMU_PCTL_SPI);
	/* POL setting */
	*s->regs->pol = s->pol;
	*s->regs->control |= SPx_CONTROL_RST;
	spi_sftrst_wait();
	*s->regs->control &= ~SPx_CONTROL_RST;

	s = &spi_private[SPI_DEV_SPI1];
	/* SPI_SMU_PCTL_SPIn_SK clear */
	outl(inl(SMU_PCTL_SPI) & ~(s->pctl_spi), SMU_PCTL_SPI);
	/* POL setting */
	*s->regs->pol = s->pol;
	*s->regs->control |= SPx_CONTROL_RST;
	spi_sftrst_wait();
	*s->regs->control &= ~SPx_CONTROL_RST;

	return 0;
}

/*
 * exit
 */
static void __exit spi_exit(void)
{
/* 2006.10.05 has */
#ifdef CONFIG_SONY_ICX1087

	spi_data_t *s;

	/* set pin to float */
	s = &spi_private[SPI_DEV_SPI0];
	outl(inl(SMU_PCTL_SPI) | (s->pctl_spi), SMU_PCTL_SPI);

	/* pwc_regulator_off(PWC_PSCNTGP5); */
	pwc_regulator_off(PWC_PSCNTHP);
#endif

	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SPI0SCLKLP);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SPI1SCLKLP);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SPI0PCLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_SPI1PCLK);
	mp200_pmu_reset_device(MP200_RESETDEVICE_SPI0);
	mp200_pmu_reset_device(MP200_RESETDEVICE_SPI1);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SPI0_SCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SPI1_SCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SPI0_PCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SPI1_PCLK);

/* 2006.10.05 has */
#ifdef CONFIG_SONY_ICX1087
	driver_unregister(&spi_driver);
#endif

	return;
}

module_init(spi_init);
module_exit(spi_exit);
MODULE_LICENSE("GPL");
EXPORT_SYMBOL(spi_write_without_intr);

