/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 *
 * http://www.opensource.org/licenses/gpl-license.html
 * http://www.gnu.org/copyleft/gpl.html
 */

/*!
 * @file mxc_ipc.c
 *
 * @brief This file provides all the kernel level and user level API
 * definitions for all kind of transfers between MCU and DSP core.
 *
 * The Interfaces are with respect to the MCU core. Any driver or user-space
 * application on the MCU side can transfer messages or data to the DSP side 
 * using the interfaces implemented here.
 *
 * @ingroup IPC
 */

/*
 * Include Files
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/kdev_t.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/device.h>
#include <linux/kthread.h>
#include <linux/circ_buf.h>
#include <linux/uio.h>
#include <linux/poll.h>

#include <asm/arch/platform.h>
#include <asm/arch/mxc_mu.h>
#include <asm/arch/mxc_ipc.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/semaphore.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>

#include "mxc_mu_reg.h"

//#define DEBUG 1

#if DEBUG
#define DPRINTK(fmt, args...) printk("%s: " fmt, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif

/*!
 * Automatic major assignment
 */
#define MXC_IPC_MAJOR         			0x0

/*!
 * Buffer size for MU and SDMA channels 
 */
#define MAX_BUFFER_SIZE         		512
#define MAX_SDMA_BUF_SIZE       		MAX_BUFFER_SIZE
#define MAX_MU_BUF_SIZE         		MAX_BUFFER_SIZE

/*!
 * Max virtual channels available for this implementation
 */
#define IPC_MAX_VIRTUAL_CHANNELS      		32

/*!
 * Max ipc channels available for this implementation
 */
#define IPC_MAX_IPC_CHANNEL       		4

/*!
 * This defines an unused virtual channel
 */
#define IPC_DUMMY_CHANNEL       		\
IPC_MAX_VIRTUAL_CHANNELS + 1      

/*!
 * Max index allowed for IPC channels. There are three types:
 *  read/write channel for data
 *  read only for data
 *  read/write channel for messages
 */
#define IPC_MAX_MU_CHANNEL_INDEX      		2
#define IPC_MAX_SDMA_BIDI_CHANNEL_INDEX     	2
#define IPC_MAX_SDMA_MONO_CHANNEL_INDEX     	4

/*!
 * States allowed for an IPC channel
 */ 
#define CHANNEL_CLOSED          		0x0
#define CHANNEL_OPEN          			0x1
#define CHANNEL_READ_ONGOING         		0x2
#define CHANNEL_WRITE_ONGOING         		0x4

/*!
 * Physical channels used to implement IPC channels.
 * This defines depends on channels used on DSP side
 * i.e. if the DSP uses SDMA channel 1 to do writes, 
 * we must use SDMA channel 1 to receive data.
 *
 * Currently we use 5 physical channels to implement the
 * IPC abstraction.
 *
 * - MU channel 0   --> IPC channel 0
 * - MU channel 1   --> IPC channel 1
 * - SDMA channel 1 --> IPC channel 2
 * - SDMA channel 2 --> IPC channel 2
 * - SDMA channel 3 --> IPC channel 3
 *
 * Note that IPC channel 2 uses two physical channels. This
 * is because SDMA channels are unidirectional and IPC 
 * links must be bidirectional.
 * 
 */
#define SHORT_MESSAGE_CHANNEL0        		0
#define SHORT_MESSAGE_CHANNEL1        		1
#define PACKET_DATA_CHANNEL       		2
#define LOGGING_CHANNEL         		5

/*!
 * Operations allowed on IPC channels. These macros are 
 * used to tell the IPC driver-interface level which
 * operation to execute.
*/
#define IPC_READ          			0x1
#define IPC_WRITE         			0x2
#define IPC_RDWR          			IPC_READ | IPC_WRITE

typedef void (*read_callback_t)(HW_CTRL_IPC_READ_STATUS_T*);
typedef void (*write_callback_t)(HW_CTRL_IPC_WRITE_STATUS_T*);
typedef void (*notify_callback_t)(HW_CTRL_IPC_NOTIFY_STATUS_T*);

/*! 
 * This structure represents the SDMA channels used 
 * by the IPC abstraction.
 */
struct sdma_channel{
        /*!
         * SDMA read channel number
         */
        int read_channel;
        
        /*!
         * SDMA write channel number
         */
        int write_channel;
        
        /*!
         * Physical address of the SDMA write transfer buffer 
         */
        char* wpaddr;
        
        /*! 
         * Physical address of the SDMA read transfer buffer 
         */
        char* rpaddr;
        
        /*!
         * SDMA transfer write buffer 
         */
        char* wbuf;
        
        /*! 
         * SDMA transfer read buffer 
         */
        char* rbuf;
};

/*! 
 * This structure represents the MU channels used 
 * by the IPC abstraction. 
 */
struct mu_channel{
        /*!
         * MU channel number
         */
        int index;
        
        /*! 
         * MU buffer
         */
        struct circ_buf cbuf;
};

/*! 
 * This structure represents an IPC channel. It is represented
 * by one or even two physical channels, plus a control structure
 */
struct ipc_channel{
        /*! 
         * Physical channel(s) used to implement an IPC channel
         */
        union{
                struct sdma_channel sdma;
                struct mu_channel   mu;
        }ch;
        
        /*! 
         * control structure. It handles access and transfers on this IPC channel
         */
         struct ipc_priv_data* priv;
};

/*! 
 * This structure represents a virtual channel. Multiple virtual channels
 * could be connected to an IPC channel. IPC abstraction supports up to 32
 * virtual channels.
 */
struct virtual_channel{
        /*!
         * controls access to the handle of this channel
         */
        struct semaphore sem;
        
        /*!
         * current state of the channel
         */
        atomic_t state;
                
        /*! 
         *
         */
        char* data;
        
        /*! 
         * pointer to the control structure of the IPC channel mapped by
         * this virtual channel
         */
        struct ipc_channel* ipc;
};

/*!
 * Private data. Each time an IPC channel is opened, a private
 * structure is allocated and stored in the priv
 * field of the file descriptor. This data is
 * used for further operations on the recently opened
 * channel
 */
struct ipc_priv_data{
        /*!
         * This indicates the termination of a transfer (MCU->DSP)
         */
        unsigned char write_done;
        
        /*!
         * This indicates the termination of a transfer (DSP->MCU)
         */
        unsigned char read_done;
        
        /*!
         * this variable signals if a process is waiting
         * for the end of a write operation on this channel
         */
        unsigned char write_wait;
        
        /*!
         * this variable signals if a process is waiting
         * for the end of a read operation on this channel
         */
        unsigned char read_wait;
        
        /*!
         * blocking mode (non-blocking, blocking)
         */
         atomic_t blockmode;
         
        /*!
         * physical channel to which the
         * virtual channel is mapped to
         */
        unsigned short pchannel;
        
        /*!
         * index of virtual channel
         */
        unsigned short vchannel;
        
        /*!
         * This stores number of bytes transferred
         */
        unsigned int count;
        
        /*!
         * type of transfer being executed (read, write)
         */
        unsigned int transfer_type; 
        
        /*! 
         * write waitqueue
         */
        wait_queue_head_t wq;
        
        /*! 
         * read waitqueue
         */
        wait_queue_head_t rq;
                
        /*!
         * store the task_struct structure of
         * the calling process.
         */
        struct task_struct* owner;
        
        /*!
         * virtual channel assigned to the
         * IPC channel storing this structure.
         */
        struct virtual_channel* vc;
        
        /*!
         * callbacks used to signal the end of a operation
         * on an IPC channel. Only for kernel modules.
         */
        struct kernel_callbacks* k_callbacks;
};

/*!
 * This struct stores the callbacks pointers
 * used to communicate the end of an operation
 * on an IPC channel. These callbacks must be
 * provided by the caller. This feature is available
 * only for kernel modules.
 */
struct kernel_callbacks{
	read_callback_t read;
        write_callback_t write;
        notify_callback_t notify;
};

/*! 
 * Array used to store the physical channels.
 */
static struct ipc_channel ipc_channels[IPC_MAX_IPC_CHANNEL];

/*!
 * Array used to store the pointers to the virtual channels.
 * These pointers will be allocated when trying to open a virtual channel.
 */
static struct virtual_channel virtual_channels[IPC_MAX_VIRTUAL_CHANNELS];

/*!
 * Status registers of the MU channels
 */
static unsigned int wbits[2]     = {AS_MUMSR_MTE0, AS_MUMSR_MTE1};
static unsigned int rbits[2]     = {AS_MUMSR_MRF0, AS_MUMSR_MRF1};

/*!
 * Major number
 */
static int major_num = 0;

static void sdma_user_readcb(void* args);
static void sdma_user_writecb(void* args);

/*!
 * Locks the virtual channel located on the index channel.
 * @param channel index of the virtual channel to lock
 */
static inline void lock_virtual_channel(unsigned short channel)
{
        down_interruptible(&virtual_channels[channel].sem); 
}

/*!
 * Try to locks the virtual channel located on the index channel.
 * @param channel index of the virtual channel to (try)lock
 */
static inline int trylock_virtual_channel(unsigned short channel)
{
        return down_trylock(&virtual_channels[channel].sem);  
}

/*!
 * Unlocks the virtual channel located on the index channel
 * @param channel index of the virtual channel to unlock
 */
static inline void unlock_virtual_channel(unsigned short channel)
{
        up(&virtual_channels[channel].sem); 
}

/*!
 * Initializes the array of virtual channels at loading time
 * This operation is executed only when Kernel is starting up.
 */
inline void initialize_virtual_channels(void)
{
        struct virtual_channel* vch;
        int i;
        
        for (i = 0; i < IPC_MAX_VIRTUAL_CHANNELS; i++){
        	vch = &virtual_channels[i];
          
          	sema_init(&vch->sem, 1);
          	atomic_set(&vch->state, CHANNEL_CLOSED);
  
          	//vch->callbacks = NULL;
          	vch->ipc = NULL;
        }
}

/*!
 * Finds an unused channel into the virtual channel array
 *
 * @return  The function returns teh index of the first unused 
 *    channel found, IPC_DUMMY_CHANNEL otherwise.
 */
inline int get_free_virtual_channel(void)
{
        struct virtual_channel* vch;
        int i;
        
        
        for (i = 0; i < IPC_MAX_VIRTUAL_CHANNELS; i++){
        	if(trylock_virtual_channel(i) == 1){
            		continue;
          	}
          
          	vch = &virtual_channels[i];
            
          	if(atomic_read(&vch->state) == CHANNEL_CLOSED){
            		return i;
          	}
          
          	unlock_virtual_channel(i);
        }
        
        return IPC_DUMMY_CHANNEL;
}

/*!
 * The MU interrupt handler schedules this tasklet for 
 * execution to start processing a write request.
 * This function writes data to chnum MU register and signals 
 * the end of the transfer to the servicing thread attached 
 * to the IPC channel.
 *
 * @param chnum    MU channel number where the operation will occur
 */
static void mu_write_tasklet(int chnum)
{
        struct ipc_priv_data* priv;
        struct mu_channel* mu;
        char* p = NULL;
        int bytes;
                
        priv = ipc_channels[chnum].priv;
        mu = &ipc_channels[chnum].ch.mu;
        
        bytes = CIRC_CNT(mu->cbuf.head, mu->cbuf.tail, MAX_MU_BUF_SIZE);
        
        while ((bytes != 0) && (readl(AS_MUMSR) & wbits[chnum])) {
        	
        	DPRINTK("remaining bytes = %d\n", bytes);
        	
                p = &mu->cbuf.buf[mu->cbuf.tail];
                mu->cbuf.tail = (mu->cbuf.tail + 4) & (MAX_MU_BUF_SIZE - 1);
                
                mxc_mu_mcuwrite(mu->index, p);
                                
                bytes = CIRC_CNT(mu->cbuf.head, 
                       mu->cbuf.tail, MAX_MU_BUF_SIZE);
                priv->count += 4;
        }

        bytes = CIRC_CNT(mu->cbuf.head, mu->cbuf.tail, MAX_MU_BUF_SIZE);

        if(bytes != 0) {
        	DPRINTK("(2)remaining bytes = %d\n", bytes);
                mxc_mu_intenable(mu->index, TX);
        }
        
        bytes = CIRC_SPACE(mu->cbuf.head, mu->cbuf.tail, MAX_MU_BUF_SIZE);
        
        if( (atomic_read(&priv->blockmode) == 1) && (bytes >= 4) && (priv->write_wait > 0)){
                priv->write_done = 1;
                priv->write_wait--;
          	wake_up_interruptible(&priv->wq);
          
          	DPRINTK("waiting processes woken up\n");
        }
}

/*!
 * The MU interrupt handler schedules this tasklet for 
 * execution to start processing a read request.
 * This function reads data from chnum MU register and signals 
 * the end of the transfer to the servicing thread attached 
 * to the IPC channel.
 *
 * @param   chnum    MU channel number where the operation will occur
 */
static void mu_read_tasklet(int chnum)
{
        struct ipc_priv_data* priv;
        struct mu_channel* mu;
        int bytes;
        char p[4];
                
        priv = (struct ipc_priv_data*)ipc_channels[chnum].priv;
        mu = &ipc_channels[chnum].ch.mu;
        
        bytes = CIRC_SPACE(mu->cbuf.head, mu->cbuf.tail, MAX_MU_BUF_SIZE);
        
        while ((bytes >= 4) && (readl(AS_MUMSR) & rbits[chnum])){
        	
        	DPRINTK("space available = %d\n", bytes);
        	
                mxc_mu_mcuread(mu->index, p);
                
                *((int*)(&(mu->cbuf.buf[mu->cbuf.head]))) = *((int*)p);
                mu->cbuf.head = (mu->cbuf.head + 4) & (MAX_MU_BUF_SIZE - 1);
                
                bytes = CIRC_SPACE(mu->cbuf.head, mu->cbuf.tail, MAX_MU_BUF_SIZE);
                priv->count += 4;
        }
                
        bytes = CIRC_SPACE(mu->cbuf.head, mu->cbuf.tail, MAX_MU_BUF_SIZE);
        
        if(bytes < 4){
        	mxc_mu_intdisable(mu->index, RX);
        }else{
        	DPRINTK("(2)space available = %d\n", bytes);
        	mxc_mu_intenable(mu->index, RX);
        }
                       
        bytes = CIRC_CNT(mu->cbuf.head, mu->cbuf.tail, MAX_MU_BUF_SIZE);
                
        if((atomic_read(&priv->blockmode) == 1) && (bytes != 0) && (priv->read_wait > 0)){
                priv->read_done = 1;
                priv->read_wait--;
        	wake_up_interruptible(&priv->rq);
        	DPRINTK("waiting processes woken up\n");
        }        
}

static void mu_read_tasklet_kernel(int chnum)
{
        struct ipc_priv_data* priv;
        struct mu_channel* mu;
        char p[4];
        int bytes = 0;
        
        priv = (struct ipc_priv_data*)ipc_channels[chnum].priv;
        mu = &ipc_channels[chnum].ch.mu;

	while ((bytes < priv->count) && (readl(AS_MUMSR) & rbits[chnum])){
        	DPRINTK("space available = %d\n", bytes);
        	
                mxc_mu_mcuread(mu->index, p);
                
                *((int*)(&(priv->vc->data[bytes]))) = *((int*)p);
                bytes += 4;
        }
                
        mxc_mu_intdisable(mu->index, RX);
                
        if(priv->k_callbacks->read != NULL){
        	HW_CTRL_IPC_READ_STATUS_T status;
        	
        	status.channel_nb = priv->vchannel;
        	status.nb_bytes = bytes;
        	priv->k_callbacks->read(&status);
        }
        
        mxc_mu_unbind(mu->index, RX);
        atomic_set(&priv->vc->state, CHANNEL_OPEN);
}

/*!
 * The MU interrupt handler schedules this tasklet for 
 * execution to start processing a read request.
 * This function reads data from chnum MU register and signals 
 * the end of the transfer to the servicing thread attached 
 * to the IPC channel.
 *
 * This function is used when the caller is a kernel module
 *
 * @param   chnum    MU channel number where the operation will occur
 */
static void mu_write_tasklet_kernel(int chnum)
{
        struct ipc_priv_data* priv;
        struct mu_channel* mu;
        char* p = NULL;
        int bytes = 0;
        
        priv = ipc_channels[chnum].priv;
        mu = &ipc_channels[chnum].ch.mu;
                
        while ((bytes < priv->count) && (readl(AS_MUMSR) & wbits[chnum])){
		DPRINTK("remaining bytes = %d\n", bytes);
        	
                p = &priv->vc->data[bytes];
                mxc_mu_mcuwrite(mu->index, p);
                bytes += 4;
        }
        	
        if(priv->k_callbacks->write != NULL){
        		HW_CTRL_IPC_WRITE_STATUS_T status;
        	
        		status.channel_nb = priv->vchannel;
        		status.nb_bytes = bytes;
       	        	priv->k_callbacks->write(&status);
        }
        
        mxc_mu_unbind(mu->index, TX);
        atomic_set(&priv->vc->state, CHANNEL_OPEN);
}

/*! 
 * Initializes a MU channel in order to support IPC abstraction
 *
 * @param   chnum    MU channel to initialize
 * @return  returns 0 on success, -1 otherwise
 */
int initialize_mu_channel(int chnum)
{
        struct mu_channel* mu;
        
        mu = &ipc_channels[chnum].ch.mu;
  
        mu->index = mxc_mu_alloc_channel(chnum);
        if(mu->index == -1){
        	DPRINTK("Allocation of MU channel %d failed \n", chnum);
        	return -1;
        }
                
        mxc_mu_intdisable(mu->index, TX);
                
        mu->cbuf.buf = kmalloc(MAX_MU_BUF_SIZE, GFP_KERNEL);
        if(mu->cbuf.buf == NULL){
        	goto cleanup;
        }
        
        mu->cbuf.tail = mu->cbuf.head = 0;
        return 0;
        
cleanup:
        mxc_mu_dealloc_channel(mu->index);
        
        return -1;
}

/*! 
 * allocate SDMA channels 1, 2 and 5 to support IPC channels
 * 2 and 3.
 *
 * @param   chnum    SDMA channel number to initialize
 * @param   mode     open mode for this channel (read, write, read/write)
 *
 * @return  returns 0 on success, -1 otherwise.
 */
int initialize_sdma_channel(int chnum, int mode)
{
        dma_channel_params sdma_params;
        struct sdma_channel* sdma;
        int result = 0;
        
        if( (mode & IPC_READ) && (chnum == LOGGING_CHANNEL) ){
        	sdma = &ipc_channels[3].ch.sdma;
        	sdma->read_channel = chnum;
        	sdma->write_channel = -1;
        }else{
        	sdma = &ipc_channels[2].ch.sdma;
        	sdma->read_channel = chnum - 1;
        	sdma->write_channel = chnum;
        }
        	
        if(mode & IPC_WRITE){
	      	sdma->wbuf = sdma_malloc(MAX_SDMA_BUF_SIZE);
	          	
	       	if(sdma->wbuf == NULL){
	       		return -1;
	       	}
	          	        
	       	sdma->wpaddr = (char*)sdma_virt_to_phys(sdma->wbuf);
	       	result = mxc_request_dma(&sdma->write_channel, "IPC WRITE SDMA");
	          	
	       	if(result < 0){
	       		DPRINTK("Failed Opening SDMA channel %d\n", sdma->write_channel);
	       		return -1;
	       	}
	        
	      	memset(&sdma_params, 0, sizeof(dma_channel_params));
	       	sdma_params.peripheral_type = DSP;
	       	sdma_params.transfer_type = emi_2_dsp;
	       	sdma_params.event_id = 0;
	       	sdma_params.callback = NULL;
	       	sdma_params.arg = 0;
	       	sdma_params.bd_number = 1;
	        
	       	mxc_dma_setup_channel(sdma->write_channel, &sdma_params);
        	
	}
        
        if(mode & IPC_READ){
	       	sdma->rbuf = sdma_malloc(MAX_SDMA_BUF_SIZE);
	        	
	       	if(sdma->rbuf == NULL){
	       		return -1;
	       	}
	          	        
	       	sdma->rpaddr = (char*)sdma_virt_to_phys(sdma->rbuf);
	       	result = mxc_request_dma(&sdma->read_channel, "IPC READ SDMA");
	       	if(result < 0){
	       		DPRINTK("Failed Opening SDMA channel %d\n", sdma->read_channel);
	       		return -1;
	       	}
	        
	       	memset(&sdma_params, 0, sizeof(dma_channel_params));
	       	sdma_params.peripheral_type = DSP;
	       	sdma_params.transfer_type = dsp_2_emi;
	       	sdma_params.event_id = 0;
	       	sdma_params.callback = NULL;
	      	sdma_params.arg = 0;
	       	sdma_params.bd_number = 1;
	        	
	       	mxc_dma_setup_channel(sdma->read_channel, &sdma_params);
        }
        
        return 0;        
}

/*!
 * This function deallocates a MU channel.
 *
 * @param   chnum    MU channel number to deallocate
 */
inline void free_mu_channel(int chnum)
{
        struct mu_channel* mu = &ipc_channels[chnum].ch.mu;

        mxc_mu_unbind(mu->index, TX);
        mxc_mu_unbind(mu->index, RX);
        mxc_mu_dealloc_channel(mu->index);
        
        kfree(mu->cbuf.buf);
}

/*!
 * This function deallocates a SDMA channel.
 *
 * @param   chnum    SDMA channel number to deallocate
 */
inline void free_sdma_channel(int chnum)
{
        struct sdma_channel* sdma;
        
        sdma = &ipc_channels[chnum].ch.sdma;
        if(sdma->rbuf != NULL){
        	sdma_free(sdma->rbuf);
        	mxc_free_dma(sdma->read_channel);
        }

        if(sdma->wbuf != NULL){
        	sdma_free(sdma->wbuf);
        	mxc_free_dma(sdma->write_channel);
        }
}

/*!
 * Read bytes from the circular buffer of the current mu channel
 *
 * @param   mu    	struct pointer. This contains the description of the 
 * 		  	mu channel.
 * @param   count    	number of bytes to read
 * @param   buf		buffer to store data 
 */
static int read_bytes_from_cbuf(struct mu_channel* mu, int count, char* buf)
{
        struct circ_buf* cbuf;
        int bytes, result;
        
       	cbuf = &mu->cbuf;
        
        bytes = CIRC_CNT(cbuf->head, cbuf->tail, MAX_MU_BUF_SIZE);
        
        DPRINTK("bytes available into circular buffer = %d\n", bytes);
        if(bytes > count){
                bytes = count;
        }
        
        if(bytes != 0){
        	DPRINTK("copying %d bytes into user's buffer\n", bytes);
        	result = copy_to_user(buf, &cbuf->buf[cbuf->tail], bytes);
                if(result){
                        return -1;
                }
                
                cbuf->tail = (cbuf->tail + bytes) & (MAX_MU_BUF_SIZE - 1);
        }

        return bytes;
}

/*!
 * This function starts a DSP->MCU transfer. This transfer type uses
 * a MU channel
 *
 * @param    priv	Private data. Used to synchronize the top half
 * 			and the botton half as well as to keep track of 
 *			the transfer
 * @param    count	number of bytes to be transferred
 * @param    buf	buffer where received bytes are saved
 *
 * @return    		number of transferred bytes
 */
static int mu_start_dsp_transfer(struct ipc_priv_data* priv, 
					int count, char* buf)
{
	struct mu_channel* mu = &priv->vc->ipc->ch.mu;
	int bytes;
	
        bytes = read_bytes_from_cbuf(mu, count, buf);
        DPRINTK("bytes read = %d\n", count);
        
        priv->read_done = 0;
        
        if( (bytes == 0) && (atomic_read(&priv->blockmode) == 1) ){
        	priv->read_wait++;
        }

	mxc_mu_bind(mu->index, &mu_read_tasklet, RX);
	mxc_mu_intenable(mu->index, RX);
        
        if(bytes != 0){
        	return bytes;
        }
        /*
        //Non-blocking operation requested?
        if(atomic_read(&priv->blockmode) == 0){
        	return -EAGAIN;
        }
        */
        
        atomic_set(&priv->vc->state, CHANNEL_READ_ONGOING);
          
        //wait end of transfer
        wait_event_interruptible(priv->rq, (priv->read_done != 0));
        
        mxc_mu_unbind(mu->index, RX);                
        return read_bytes_from_cbuf(mu, count, buf);
}

/*!
 * Write bytes to the circular buffer of the current mu channel
 *
 * @param   mu    	struct pointer. This contains the description of the 
 * 		  	mu channel.
 * @param   count    	number of bytes to write
 * @param   buf		buffer containing the bytes to transfer.
 */
static int write_bytes_to_cbuf(struct mu_channel* mu, int count, const char* buf)
{
        struct circ_buf* cbuf;
        int result;
        int bytes = 0;
        
       	cbuf = &mu->cbuf;
          
       	bytes = CIRC_SPACE(cbuf->head, cbuf->tail, MAX_MU_BUF_SIZE);
          
       	//no space available to copy data to the MU buffer
       	if(bytes < 4){
       		DPRINTK("no space\n");
       		return 0;
       	}	 
        
        DPRINTK("space available into circular buffer = %d\n", bytes);
        
        if(bytes > count){
                bytes = count;
        }
	
        DPRINTK("copying %d bytes into circular buffer\n", bytes);
        result = copy_from_user(&cbuf->buf[cbuf->head], buf, bytes);
        if(result){
        	DPRINTK("EINVAL error\n");
        	return -1;
        }
                
        cbuf->head = (cbuf->head + bytes) & (MAX_MU_BUF_SIZE - 1);
        DPRINTK("buf head = %d\n", cbuf->head);
        
        return bytes;
}

/*!
 * This function starts a MCU->DSP transfer. This transfer uses
 * a MU channel
 *
 * @param    priv	Private data. Used to synchronize the top half
 * 			and the botton half as well as to keep track of 
 *			the transfer
 * @param    count	number of bytes to be transferred
 * @param    buf	buffer containing the bytes to be transferred
 *
 * @return    		number of transferred bytes.
 */
static int mu_start_mcu_transfer(struct ipc_priv_data* priv, 
					int count, const char* buf)
{
        struct mu_channel* mu = &priv->vc->ipc->ch.mu;
	int bytes;
	
        bytes = write_bytes_to_cbuf(mu, count, buf);
        DPRINTK("bytes written = %d\n", count);
        
        priv->write_done = 0;
        
        if( (bytes == 0) && (atomic_read(&priv->blockmode) == 1) ){
        	priv->write_wait++;
        }
                
        mxc_mu_bind(mu->index, &mu_write_tasklet, TX);
        mxc_mu_intenable(mu->index, TX);
        
        if(bytes != 0){
        	return bytes;
        }
                
        //wait end of transfer
        wait_event_interruptible(priv->wq, (priv->write_done != 0));
        
        mxc_mu_unbind(mu->index, TX);                
        return write_bytes_to_cbuf(mu, count, buf);
}

/*!
 * This function starts a MCU->DSP transfer requested by a kernel module. 
 * This transfer uses a MU channel. 
 *
 * @param    priv	Private data. Used to synchronize and to keep track of 
 *			the transfer
 * @param    count	number of bytes to be transferred
 * @param    buf	buffer containing the bytes to be transferred
 *
 * @return    		Zero
 */
static int mu_start_mcu_transfer_kernel(struct ipc_priv_data* priv, 
					int count, const char* buf)
{
	struct mu_channel* mu = &priv->vc->ipc->ch.mu;
	
	DPRINTK("writing %d bytes to DSP\n", count);
	
	if(count > MAX_MU_BUF_SIZE){
		count = MAX_MU_BUF_SIZE;
	}
		
	priv->count = count;
	priv->vc->data = (char*)buf;
                
        mxc_mu_bind(mu->index, &mu_write_tasklet_kernel, TX);
        mxc_mu_intenable(mu->index, TX);
                        
        return 0;
}

/*!
 * This function starts a DSP->MCU transfer requested by a kernel module. 
 * This transfer uses a MU channel. 
 *
 * @param    priv	Private data. Used to synchronize and to keep track of 
 *			the transfer
 * @param    count	number of bytes to be transferred
 * @param    buf	buffer containing the bytes to be transferred
 *
 * @return    		Zero
 */
static int mu_start_dsp_transfer_kernel(struct ipc_priv_data* priv, 
					int count, char* buf)
{
	struct mu_channel* mu = &priv->vc->ipc->ch.mu;
	
	if(count > MAX_MU_BUF_SIZE){
		count = MAX_MU_BUF_SIZE;
	}
	
	priv->count = count;
	priv->vc->data = buf;
	
	mxc_mu_bind(mu->index, &mu_read_tasklet_kernel, RX);
	mxc_mu_intenable(mu->index, RX);
                                
        return 0;
}

/*!
 * This function is called by the SDMA's ISR whenever a MCU->DSP 
 * transfer has finished. This callback is used whenever a
 * kernel module requests a transfer.
 *
 * @param    args	points to a priv structure. This structure contains
 *			fields to synchronize the IPC channel and the top half
 *
 * @return		None
 */
static void sdma_kernel_writecb(void* args)
{
        dma_request_t sdma_request;
        struct ipc_priv_data* priv;
        struct sdma_channel* sdma;
        
        priv = (struct ipc_priv_data*)args;
        sdma = &priv->vc->ipc->ch.sdma;
        
        mxc_dma_get_config(sdma->write_channel, &sdma_request, 0);
        
        priv->count = sdma_request.count;
        
        if(priv->k_callbacks->write != NULL){
        	HW_CTRL_IPC_WRITE_STATUS_T status;
        	
        	status.channel_nb = priv->vchannel;
        	status.nb_bytes = priv->count;
        	
       	        priv->k_callbacks->write(&status);
        }
        

        atomic_set(&priv->vc->state, CHANNEL_OPEN);
        DPRINTK("Called SDMA write callback %d\n", priv->count);
}

/*!
 * This function starts a MCU->DSP transfer requested by a kernel module. 
 * This transfer uses a SDMA channel. 
 *
 * @param    priv	Private data. Used to synchronize and to keep track of 
 *			the transfer
 * @param    count	number of bytes to be transferred
 * @param    buf	buffer containing the bytes to be transferred
 *
 * @return    		Zero
 */
static int sdma_start_mcu_transfer_kernel(struct ipc_priv_data* priv, 
						int count, const char* buf)
{
        dma_request_t sdma_request;
        struct sdma_channel* sdma;
        
        sdma = &priv->vc->ipc->ch.sdma;
         	        
        if(count > MAX_SDMA_BUF_SIZE){
                count = MAX_SDMA_BUF_SIZE;
        }
        
        memcpy(sdma->wbuf, buf, count);
        memset(&sdma_request, 0, sizeof(dma_request_t));
        
        DPRINTK("writing %d bytes to DSP\n", count);
        
        sdma_request.sourceAddr = (__u8*)sdma_virt_to_phys(sdma->wbuf);
        sdma_request.count = count;
        
        mxc_dma_set_callback(sdma->write_channel, sdma_kernel_writecb, priv);          	          
	mxc_dma_set_config(sdma->write_channel, &sdma_request, 0);
	mxc_dma_start(sdma->write_channel);
	          
        return 0;
}

/*!
 * This function is called by the SDMA's ISR whenever a MCU->DSP 
 * transfer has finished. This callback is used whenever a
 * user space program requests a transfer.
 *
 * @param    args	points to a priv structure. This structure contains
 *			fields to synchronize the IPC channel and the top half
 *
 * @return		None
 */
static void sdma_user_writecb(void* args)
{
        dma_request_t sdma_request;
        struct ipc_priv_data* priv;
        struct sdma_channel* sdma;
        
        priv = (struct ipc_priv_data*)args;
        sdma = &priv->vc->ipc->ch.sdma;
        
        mxc_dma_get_config(sdma->write_channel, &sdma_request, 0);
        priv->count = sdma_request.count;
        
        if( (atomic_read(&priv->blockmode) == 1) && (priv->write_wait > 0)){
                priv->write_done = 1;
                priv->write_wait--;
          	wake_up_interruptible(&priv->wq);
          
          	DPRINTK("waiting processes woken up\n");
        }
        
        DPRINTK("Called SDMA write callback %d\n", priv->count);
}

/*!
 * This function starts a MCU->DSP transfer requested by a user space
 * program. This transfer uses a SDMA channel. 
 *
 * @param    priv	Private data. Useful to synchronize and to keep track of 
 *			the transfer
 * @param    count	number of bytes to be transferred
 * @param    buf	buffer containing the bytes to be transferred
 *
 * @return    		Zero
 */
static int sdma_start_mcu_transfer(struct ipc_priv_data* priv, 
					int count, const char* buf)
{	
        dma_request_t sdma_request;
        struct sdma_channel* sdma;
        int result;
        
        sdma = &priv->vc->ipc->ch.sdma;
         	        
        if(count > MAX_SDMA_BUF_SIZE){
                count = MAX_SDMA_BUF_SIZE;
        }
        
        result = copy_from_user(sdma->wbuf, buf, count);
        if(result != 0){
        	return -EFAULT;
        }
        
        memset(&sdma_request, 0, sizeof(dma_request_t));
                          
        DPRINTK("writing %d bytes to SDMA channel %d\n", count, sdma->write_channel);
                  
        sdma_request.sourceAddr = (__u8*)sdma_virt_to_phys(sdma->wbuf);
        sdma_request.count = count;
        
        priv->write_done = 0;
        priv->write_wait++;
        
        mxc_dma_set_callback(sdma->write_channel, sdma_user_writecb, priv);  	          
	mxc_dma_set_config(sdma->write_channel, &sdma_request, 0);
	mxc_dma_start(sdma->write_channel);
	                
        /*wait end of transfer*/
        wait_event_interruptible(priv->wq, (priv->write_done != 0));
                  
        return priv->count;
}

/*!
 * This function is called by the SDMA ISR whenever a DSP->MCU
 * transfer has finished. This callback is used whenever a
 * kernel module requests a transfer.
 *
 * @param    args	points to a priv structure. This structure contains
 *			fields to synchronize the IPC channel and the top half
 *
 * @return		None
 */
static void sdma_kernel_readcb(void* args)
{
        dma_request_t sdma_request;
        struct ipc_priv_data* priv;
        struct sdma_channel* sdma;
        
        priv = (struct ipc_priv_data*)args;
        sdma = &priv->vc->ipc->ch.sdma;
        
        mxc_dma_get_config(sdma->read_channel, &sdma_request, 0);
        
        priv->count = sdma_request.count;
        
        memcpy(priv->vc->data, sdma->rbuf, priv->count);
        
        if(priv->k_callbacks->read != NULL){
        	HW_CTRL_IPC_READ_STATUS_T status;
        	
        	status.channel_nb = priv->vchannel;
        	status.nb_bytes = priv->count;
        	
       	        priv->k_callbacks->read(&status);
        }
        
        atomic_set(&priv->vc->state, CHANNEL_OPEN);        
        DPRINTK("Called SDMA read callback %d\n", priv->count);
}

/*!
 * This function starts a DSP->MCU transfer requested by a kernel module. 
 * This transfer uses a SDMA channel. 
 *
 * @param    priv	Private data. Used to synchronize and to keep track of 
 *			the transfer
 * @param    count	number of bytes to be transferred
 * @param    buf	buffer containing the bytes to be transferred
 *
 * @return    		Zero
 */
static int sdma_start_dsp_transfer_kernel(struct ipc_priv_data* priv, 
						int count, char* buf)
{
        dma_request_t sdma_request;
        struct sdma_channel* sdma;
        
        sdma = &priv->vc->ipc->ch.sdma;
         	        
        if(count > MAX_SDMA_BUF_SIZE){
                count = MAX_SDMA_BUF_SIZE;
        }
        
        priv->vc->data = buf;
        
        memset(&sdma_request, 0, sizeof(dma_request_t));
       
        DPRINTK("writing %d bytes to DSP\n", count);
        
        sdma_request.destAddr = (__u8*)sdma_virt_to_phys(sdma->rbuf);
        sdma_request.count = count;
        
        mxc_dma_set_callback(sdma->read_channel, sdma_kernel_readcb, priv);          	          
	mxc_dma_set_config(sdma->read_channel, &sdma_request, 0);
	mxc_dma_start(sdma->read_channel);
	          
        return 0;
}

/*!
 * This function is called by the SDMA's ISR whenever a DSP->MCU
 * transfer has finished. This callback is used whenever a
 * user space program requests a transfer.
 *
 * @param    args	points to a priv structure. This structure contains
 *			fields to synchronize the IPC channel and the top half
 *
 * @return		None
 */
static void sdma_user_readcb(void* args)
{
	
        dma_request_t sdma_request;
        struct ipc_priv_data* priv;
        struct sdma_channel* sdma;
        
        priv = (struct ipc_priv_data*)args;
        sdma = &priv->vc->ipc->ch.sdma;
        
        mxc_dma_get_config(sdma->write_channel, &sdma_request, 0);
        priv->count = sdma_request.count;
        
        if( (atomic_read(&priv->blockmode) == 1) && (priv->read_wait > 0)){
                priv->read_done = 1;
                priv->read_wait--;
          	wake_up_interruptible(&priv->wq);
          
          	DPRINTK("waiting processes woken up\n");
        }
        
        DPRINTK("Called SDMA read callback %d\n", priv->count);
}

/*!
 * This function starts a DSP->MCU transfer requested by a user
 * space program. This transfer uses a SDMA channel. 
 *
 * @param    priv	Private data. Used to synchronize and to keep track of 
 *			the transfer
 * @param    count	number of bytes to be transferred
 * @param    buf	buffer containing the bytes to be transferred
 *
 * @return    		Zero
 */
static int sdma_start_dsp_transfer(struct ipc_priv_data* priv, 
					int count, char* buf)
{	
        dma_request_t sdma_request;
        struct sdma_channel* sdma;
        int result;
        
        sdma = &priv->vc->ipc->ch.sdma;
         	        
        if(count > MAX_SDMA_BUF_SIZE){
                count = MAX_SDMA_BUF_SIZE;
        }
	        
        memset(&sdma_request, 0, sizeof(dma_request_t));
                          
        DPRINTK("reading %d bytes from DSP from channel %d\n", count, sdma->read_channel);
                  
        sdma_request.count = count;
        sdma_request.destAddr = (__u8*)sdma_virt_to_phys(sdma->rbuf);
        
        priv->read_done = 0;
        priv->read_wait++;
        
        mxc_dma_set_callback(sdma->read_channel, sdma_user_readcb, priv);
	mxc_dma_set_config(sdma->read_channel, &sdma_request, 0);
	mxc_dma_start(sdma->read_channel);
	    
        /*wait end of transfer*/
        wait_event_interruptible(priv->wq, (priv->read_done != 0));
        
        result = copy_to_user((void*)buf, sdma->rbuf, count);
        if(result != 0){
        	DPRINTK("EINVAL error\n");
        	return -1;
        }
                
        return priv->count;
}

/*!
 * This function writes a chunk of memory to the specified SDMA channel.
 * It is used by the writev implementation of the driver.
 *
 * @param    priv	Private data. Used to synchronize and to keep track of 
 *			the transfer
 * @param    vector	buffer containing the bytes to be transferred
 * @param    bytes	number of bytes to be transferred
 *
 * @return    		Number of bytes successfully transferred
 */
inline int linked_list_write(struct ipc_priv_data* priv, const char* vector, int bytes)
{
	int remainder = 0;
	int tr_bytes = 0;
	int count = 0;
	
	remainder = bytes;
	
	do{
		count = (remainder > MAX_SDMA_BUF_SIZE) ? 
				MAX_SDMA_BUF_SIZE : remainder;
		
		tr_bytes = sdma_start_mcu_transfer(priv, count, vector);
		if(tr_bytes < 0){
			return tr_bytes;	
		}
		
		remainder -= tr_bytes;
	}while(remainder != 0);
	
	return bytes;
}

/*!
 * This function allocates a virtual channel.
 * 
 * @param index		index of virtual channel to be open
 * @param config	pointer to a structure containing the type
 *			of channel to open.
 *
 * @return		0 on success, -1 otherwise
 */
int allocate_virtual_channel(unsigned short index, const HW_CTRL_IPC_OPEN_T* config)
{
	struct kernel_callbacks* k_callbacks;
        write_callback_t wcallback = NULL;
        struct ipc_priv_data* priv;
        unsigned short pchannel;
        
        
        wcallback = config->write_callback;
        	
        switch(config->type){
        case HW_CTRL_IPC_SHORT_MSG:
        	if( config->index >= IPC_MAX_MU_CHANNEL_INDEX ){
        		return -1;
        	}
        	
        	pchannel = config->index;
        	break;
        	
        case HW_CTRL_IPC_PACKET_DATA:
        	if( config->index >= IPC_MAX_SDMA_BIDI_CHANNEL_INDEX ){
        		return -1;
        	}
        	
        	pchannel = 2;        	
        	break;
        	
        case HW_CTRL_IPC_CHANNEL_LOG:
        	if( config->index >= IPC_MAX_SDMA_MONO_CHANNEL_INDEX ){
        		return -1;
        	}
        	
        	pchannel = 3;
        	wcallback  = NULL;
        	break;
        	
        default:
        	return -1;
        }
        
        priv = (struct ipc_priv_data*)
                         kmalloc(sizeof(struct ipc_priv_data), GFP_KERNEL);
        if(priv == NULL){
          	return -1;
        }
        
        k_callbacks = (struct kernel_callbacks*)
                         kmalloc(sizeof(struct kernel_callbacks), GFP_KERNEL);
        if(k_callbacks == NULL){
        	kfree(priv);
          	return -1;
        }
        
        priv->vc = &virtual_channels[index];
        
        atomic_set(&priv->vc->state, CHANNEL_OPEN);
        
        k_callbacks->write = wcallback;
        k_callbacks->read = config->read_callback;
        k_callbacks->notify = config->notify_callback;
        
        /* always non-blocking mode for kernel modules */
        atomic_set(&priv->blockmode, 0);
        
        priv->vchannel = index;
        priv->pchannel = pchannel;
        priv->write_done = 0;
        priv->read_done = 0;
        priv->count = 0;
        priv->read_wait = 0;
        priv->write_wait = 0;
        
        priv->vc->ipc = &ipc_channels[pchannel];
        priv->vc->ipc->priv = (void*)priv;
        priv->k_callbacks = k_callbacks;
        
        init_waitqueue_head(&priv->wq);
        init_waitqueue_head(&priv->rq);
        
        return 0;
}

/*!
 * Opens an IPC link. This functions can be called directly by kernel
 * modules. POSIX implementation of the IPC Driver also calls it.
 * 
 * @param channel_nb	virtual channel requested. It can go from 0 to 31
 * @param config	Pointer to a struct containing configuration para
 * 			meters for the channel to open (type of channel, 
 *			callbacks, etc) 
 *
 * @return		returns HW_CTRL_IPC_STATUS_OK on success, an error code
 *			otherwise.
 */
HW_CTRL_IPC_STATUS_T hw_ctrl_ipc_open(unsigned short channel_nb,
                                        const HW_CTRL_IPC_OPEN_T* config)
{
	struct virtual_channel* v;
        int status = 0;
        
        if(channel_nb > IPC_MAX_VIRTUAL_CHANNELS){
        	return HW_CTRL_IPC_STATUS_ERROR;
        }
        
        if(config == NULL){
        	return HW_CTRL_IPC_STATUS_ERROR;
        }
        
        v = &virtual_channels[channel_nb];
        
        if(atomic_read(&v->state) != CHANNEL_CLOSED){
        	DPRINTK(KERN_INFO "IPC: Virtual channel already opened\n");
        	return HW_CTRL_IPC_STATUS_CHANNEL_ALREADY_OPEN;
        }
        
        status = trylock_virtual_channel(channel_nb);
        if(status == 1){
        	DPRINTK(KERN_INFO "IPC Driver: Virtual channel already opened\n");
        	return HW_CTRL_IPC_STATUS_CHANNEL_ALREADY_OPEN;
        }
       
        status = allocate_virtual_channel(channel_nb, config);
        if(status == -1){
        	goto cleanup;
        	status = HW_CTRL_IPC_STATUS_ERROR;
        }
        
        atomic_set(&v->state, CHANNEL_OPEN);
        
        status = HW_CTRL_IPC_STATUS_OK;
        DPRINTK("Virtual channel %d opened\n", channel_nb);
        
cleanup:
	unlock_virtual_channel(channel_nb);
        return status;
}

/*!
 * Close an IPC link. This functions can be called directly by kernel
 * modules. 
 * 
 * @param channel_nb	virtual channel to close. It can go from 0 to 31
 *
 * @return		returns HW_CTRL_IPC_STATUS_OK on success, an error code
 *			otherwise.
 */
HW_CTRL_IPC_STATUS_T hw_ctrl_ipc_close(unsigned short channel_nb)
{
       	struct virtual_channel* v;
       	
        if(channel_nb > IPC_MAX_VIRTUAL_CHANNELS){
        	return HW_CTRL_IPC_STATUS_ERROR;
        }
        
        v = &virtual_channels[channel_nb];
        
        if(atomic_read(&v->state) == CHANNEL_CLOSED){
               	return HW_CTRL_IPC_STATUS_OK;
        }
        
        lock_virtual_channel(channel_nb);
                
        kfree(v->ipc->priv->k_callbacks);
        v->ipc->priv->k_callbacks = NULL;
        
        kfree(v->ipc->priv);
        v->ipc->priv = NULL;
        
        v->ipc = NULL;
        
        unlock_virtual_channel(channel_nb);
        
        atomic_set(&v->state, CHANNEL_CLOSED);
                	
	DPRINTK("Virtual channel %d closed\n", channel_nb);
	
        return HW_CTRL_IPC_STATUS_OK;
}

/*!
 * Reads data from an IPC link. This functions can be called directly by kernel
 * modules. POSIX implementation of the IPC Driver also calls it.
 * 
 * @param channel_nb	virtual channel where read has been requested. It can go from 0 to 31
 * @param buf		buffer to store data read from the channel.
 * @param nb_bytes	size of the buffer
 *
 * @return		returns HW_CTRL_IPC_STATUS_OK on success, an error code
 *			otherwise.
 */
HW_CTRL_IPC_STATUS_T hw_ctrl_ipc_read(unsigned short channel_nb, 
                                        unsigned char* buf,
                                        unsigned short nb_bytes)
{
        struct virtual_channel* v;
        int status = 0;
       	
        if(channel_nb > IPC_MAX_VIRTUAL_CHANNELS){
        	return HW_CTRL_IPC_STATUS_ERROR;
        }
        
        if(buf == NULL){
        	return HW_CTRL_IPC_STATUS_ERROR;
        }
        
        v = &virtual_channels[channel_nb];
        
        status = atomic_read(&v->state);
        if(status == CHANNEL_READ_ONGOING){
        	return HW_CTRL_IPC_STATUS_WRITE_ON_GOING;
        }
        
        if(status == CHANNEL_CLOSED){
        	return HW_CTRL_IPC_STATUS_CHANNEL_UNAVAILABLE;
        }
        
        atomic_set(&v->state, CHANNEL_READ_ONGOING);
        
        lock_virtual_channel(channel_nb);
        
        v->ipc->priv->count = 0;
	
	if(v->ipc->priv->pchannel > 1){
		status = sdma_start_dsp_transfer_kernel(v->ipc->priv, nb_bytes, buf);
	}else{
		status = mu_start_dsp_transfer_kernel(v->ipc->priv, nb_bytes, buf);
	}

	unlock_virtual_channel(channel_nb);
	return (status == 0) ? HW_CTRL_IPC_STATUS_OK : HW_CTRL_IPC_STATUS_ERROR;
}

/*!
 * Writes data to an IPC link. This functions can be called directly by kernel
 * modules. POSIX implementation of the IPC Driver also calls it.
 * 
 * @param channel_nb	virtual channel where read has been requested. It can go from 0 to 31
 * @param buf		buffer containing data t be written on the channel.
 * @param nb_bytes	size of the buffer
 *
 * @return		returns HW_CTRL_IPC_STATUS_OK on success, an error code
 *			otherwise.
 */
HW_CTRL_IPC_STATUS_T hw_ctrl_ipc_write(unsigned short channel_nb,
                                         unsigned char* buf, 
                                         unsigned short nb_bytes)
{
        struct virtual_channel* v;
        int status = 0;
       	
        if(channel_nb > IPC_MAX_VIRTUAL_CHANNELS){
        	return HW_CTRL_IPC_STATUS_ERROR;
        }
        
        if(buf == NULL){
        	return HW_CTRL_IPC_STATUS_ERROR;
        }
        
        v = &virtual_channels[channel_nb];
        
        status = atomic_read(&v->state);
        
        if(status == CHANNEL_WRITE_ONGOING){
        	return HW_CTRL_IPC_STATUS_WRITE_ON_GOING;
        }
        
        if(status == CHANNEL_CLOSED){
        	return HW_CTRL_IPC_STATUS_CHANNEL_UNAVAILABLE;
        }
        
        atomic_set(&v->state, CHANNEL_WRITE_ONGOING);               
        lock_virtual_channel(channel_nb);
        
        v->ipc->priv->count = 0;

	if(v->ipc->priv->pchannel > 1){
		status = sdma_start_mcu_transfer_kernel(v->ipc->priv, nb_bytes, buf);
	}else{
		status = mu_start_mcu_transfer_kernel(v->ipc->priv, nb_bytes, buf);
	}
	
	DPRINTK("transferred %d bytes on virtual channel %d\n", nb_bytes, channel_nb);
	
	unlock_virtual_channel(channel_nb);
	return (status == 0) ? HW_CTRL_IPC_STATUS_OK : HW_CTRL_IPC_STATUS_ERROR;
}

/*!
 * This function is called when an ioctl call is made from user space.
 *
 * @param    inode    Pointer to device inode
 * @param    file     Pointer to device file structure
 * @param    cmd      Ioctl command
 * @param    arg      Ioctl argument
 *
 * @return    The function returns 0 on success and a non-zero value on
 *            failure when cmd is other than what is specified
 */
static int mxc_ipc_ioctl(struct inode *inode, struct file * file, 
                           unsigned int cmd, unsigned long arg)
{
        //struct inode *i_node = file->f_dentry->d_inode;
        //unsigned int minor = MINOR(i_node->i_rdev);

        return -1;
}

/*!
 * This function is called when an IPC channel is opened. This function does
 * the initialization of the device corresponding to the channel user want to
 * open.
 *
 * @param    inode    Pointer to device inode
 * @param    file     Pointer to device file structure
 *
 * @return    The function returns 0 on success, -1 otherwise
 */
static int mxc_ipc_open(struct inode *inode, struct file *file)
{
        unsigned int minor = MINOR(inode->i_rdev);
        struct ipc_priv_data* priv = NULL;
        unsigned short vch_index;
        
                        
        if( (minor == 3) && (file->f_flags & (O_WRONLY | O_RDWR)) ){
        	return -ENXIO;
        }
          
        /*look for an empty virtual channel, if found, lock it*/
        vch_index = get_free_virtual_channel();
        
        /*if free virtual channel not found, return ENODEV*/
        if(vch_index == IPC_DUMMY_CHANNEL){
        	return -ENODEV;
        }
                
        priv = (struct ipc_priv_data*)
                         kmalloc(sizeof(struct ipc_priv_data), GFP_KERNEL);
        if(priv == NULL){
        	unlock_virtual_channel(vch_index);
        	return -ENOMEM;
        }
        
        atomic_set(&priv->blockmode, ( (file->f_flags & (O_NONBLOCK)) ? 0 : 1 ) );
        
        priv->vc = &virtual_channels[vch_index];
        priv->vchannel = vch_index;
        priv->pchannel = minor;
        priv->owner = current;
        priv->write_done = 0;
        priv->read_done = 0;
        priv->count = 0;
        priv->read_wait = 0;
        priv->write_wait = 0;
        
        priv->vc->ipc = &ipc_channels[priv->pchannel];
        priv->vc->ipc->priv = (void*)priv;
        
        init_waitqueue_head(&priv->wq);
        init_waitqueue_head(&priv->rq);
        atomic_set(&priv->vc->state, CHANNEL_OPEN);
        
        file->private_data = (void*)priv;
        
        unlock_virtual_channel(priv->vchannel);
        
        DPRINTK("Channel successfully opened\n");
        
        return 0;
}

/*!
 * This function is called to close an IPC channel. 
 * Any allocated buffers are freed
 *
 * @param    inode    Pointer to device inode
 * @param    file     Pointer to device file structure
 *
 * @return    The function returns 0 on success or
 *            returns NO_ACCESS if incorrect channel is accessed.
 */
static int mxc_ipc_close(struct inode *inode, struct file *file)
{
        struct ipc_priv_data* priv = NULL;
        unsigned int minor;
        
        minor = MINOR(file->f_dentry->d_inode->i_rdev);
        priv = (struct ipc_priv_data*)file->private_data;
                
        lock_virtual_channel(priv->vchannel);
        
        atomic_set(&priv->vc->state, CHANNEL_CLOSED);
        priv->vc->ipc = NULL;
        priv->vc = NULL;
        
        unlock_virtual_channel(priv->vchannel);
                
        kfree(file->private_data);
        file->private_data = NULL;
        
        DPRINTK("Channel %d closed\n", minor);
        
        return 0; 
}

/*!
 * The write function is available to the user-space to perform
 * a write operation
 *
 *  @param  file        Pointer to device file structure
 *  @param  buf         User buffer, where write data is placed.
 *  @param  bytes       Size of the requested data transfer
 *  @param  off         File position where the user is accessing.
 *
 *  @return             Returns number of bytes written or
 *                      Returns COUNT_ESIZE if the number of bytes
 *                      requested is not a multiple of channel
 *                      receive or transmit size or
 *                      Returns -EAGAIN for a non-block write when there is
 *      		no data in the buffer or
 *                      Returns -EFAULT if copy_to_user failed
 */
static ssize_t mxc_ipc_write(struct file* file, const char* buf, 
                               size_t bytes, loff_t* off)
{
        struct ipc_priv_data* priv;
        unsigned int minor;
        int blockmode;
        int status;
        int count = 0;

        minor = MINOR(file->f_dentry->d_inode->i_rdev);
                
        if( (minor <= 1) && ((bytes % 4) != 0) ){
        	DPRINTK("EINVAL error\n");
        	return -EINVAL;
        }
        
        if( (minor == 3) && (file->f_flags & (O_WRONLY | O_RDWR)) ){
        	DPRINTK("ENXIO error\n");
        	return -ENXIO; /*log channel is read-only*/
        }
        
        priv = (struct ipc_priv_data*)file->private_data;
        
        status = atomic_read(&priv->vc->state);
        blockmode = atomic_read(&priv->blockmode);
        
        /* If blocking mode has been requested and
         * channel is busy, then return -EAGAIN.
         * Otherwise, block the process.
         */
        if(( status == CHANNEL_WRITE_ONGOING) && (blockmode == 0)){
        	return -EAGAIN;
        }
        
        atomic_set(&priv->vc->state, CHANNEL_WRITE_ONGOING);        
        lock_virtual_channel(priv->vchannel);
        
        priv->count = 0;
        priv->transfer_type = IPC_WRITE;
        
        DPRINTK("bytes to write = %d\n", bytes);
        
        if(minor > 1){
        	count = sdma_start_mcu_transfer(priv, bytes, buf);
        }else{
        	count = mu_start_mcu_transfer(priv, bytes, buf);
        }
        	
        DPRINTK("bytes written = %d\n", count);
        
        atomic_set(&priv->vc->state, CHANNEL_OPEN);
        unlock_virtual_channel(priv->vchannel);
        
        return count;
}

/*!
 * The read function is available to the user-space to perform
 * a read operation
 *
 *  @param  file        Pointer to device file structure
 *  @param  buf		User buffer, where read data to be placed.
 *  @param  bytes       Size of the requested data transfer
 *  @param  off         File position where the user is accessing.
 *
 *  @return             Returns number of bytes read or
 *                      Returns COUNT_ESIZE if the number of bytes
 *                      requested is not a multiple of channel
 *                      receive or transmit size or
 *                      Returns -EAGAIN for a non-block read when no data
 *                      is ready in the buffer or in the register or
 *                      Returns -EFAULT if copy_to_user failed
 */
static ssize_t mxc_ipc_read(struct file* file, char* buf, 
                              size_t bytes, loff_t* off)
{
        struct ipc_priv_data* priv;
        unsigned int minor;
        int blockmode;
        int status;
        int count = 0;
        
        minor = MINOR(file->f_dentry->d_inode->i_rdev);
                        
        if( (minor <= 1) && ((bytes % 4) != 0) ){
        	return -EINVAL;
        }
        
        priv = (struct ipc_priv_data*)file->private_data;
        
        status = atomic_read(&priv->vc->state);
        blockmode = atomic_read(&priv->blockmode);
        
        /* If blocking mode has been requested and
         * channel is busy, then return -EAGAIN.
         * Otherwise, block the process.
         */
        if(( status == CHANNEL_READ_ONGOING) && (blockmode == 0)){
        	return -EAGAIN;
        }
        
        atomic_set(&priv->vc->state, CHANNEL_READ_ONGOING);
        
        lock_virtual_channel(priv->vchannel);
        
        priv->count = 0;
        priv->transfer_type = IPC_READ;
        
        DPRINTK("bytes to read = %d\n", bytes);
        if(minor > 1){
        	count = sdma_start_dsp_transfer(priv, bytes, buf);
        }else{
        	count = mu_start_dsp_transfer(priv, bytes, buf);
        }
        	
        DPRINTK("bytes read = %d\n", count);
        
        atomic_set(&priv->vc->state, CHANNEL_OPEN);
        unlock_virtual_channel(priv->vchannel);
        
        return count;
}

/*!
 * The writev function is available to the user-space to perform
 * a write operation on multiple (possibly non-contiguous) buffers
 *
 *  @param  file        Pointer to device file structure
 *  @param  iov      	list of iovec structs containing information about
 *			the buffers to be transferred.
 *  @param  segs      	number of iovec structs
 *  @param  off         File position where the user is accessing.
 *
 *  @return             Returns number of bytes written or
 *                      Returns ENXIO if the channel does not support
 *			this operation or
 *                      Returns -EAGAIN for a write that would block or
 *                      Returns -EFAULT if it is not possible to 
 *			access to any buffer to be transferred or
 *			Returns EINVAL if a negative size for any buffer
 *			is found.
 */
ssize_t mxc_ipc_writev(struct file* file, const struct iovec* iov, 
				unsigned long segs, loff_t* off)
{
        struct ipc_priv_data* priv;
        unsigned int minor;
        int blockmode;
        int nr_segs;
        int tr_bytes;
        int total;
        int status;

        minor = MINOR(file->f_dentry->d_inode->i_rdev);
         
        if(minor != 2){
        	return -ENXIO;
        }
        
        priv = (struct ipc_priv_data*)file->private_data;
        
        status = atomic_read(&priv->vc->state);
        blockmode = atomic_read(&priv->blockmode);
        
        if(( status == CHANNEL_WRITE_ONGOING) && (blockmode == 0)){
        	return -EAGAIN;
        }
        
        atomic_set(&priv->vc->state, CHANNEL_WRITE_ONGOING);
        lock_virtual_channel(priv->vchannel);
        
        priv->count = 0;
        priv->transfer_type = IPC_WRITE;
        
        tr_bytes = 0;
        total = 0;

	/*check validity of all segments*/
        for (nr_segs = 0; nr_segs < segs; nr_segs++){
		const struct iovec* iv = &iov[nr_segs];
		
		if(iv->iov_len == 0){
			continue;
		}
			
		if(iv->iov_len < 0){
			status = -EINVAL;
			goto output;
		}
		
		/*is this memory block a valid one?*/	
		if (access_ok(VERIFY_READ, iv->iov_base, iv->iov_len) == 0){
			status = -EFAULT;
			goto output;
		}

		/*send a memory segment*/				
		tr_bytes = linked_list_write(priv, iv->iov_base, iv->iov_len);
		if(tr_bytes < 0){
			status = tr_bytes;
			goto output;
		}
		
		DPRINTK("bytes written = %d\n", tr_bytes);
		total += tr_bytes;
        }

output:
        atomic_set(&priv->vc->state, CHANNEL_OPEN);
        unlock_virtual_channel(priv->vchannel);
        return total;
}

/*!
 * This function allows the caller to determine whether it can 
 * read from or write to one or more open files without blocking.
 *
 * @param    filp	Pointer to device file structure
 * @param    wait	used by poll_wait to indicate a change in the
 *			poll status.
 *		
 * @return		the bit mask describing which operations could 
 *			be completed immediately.
 */
unsigned int mxc_ipc_poll(struct file* filp, poll_table* wait)
{
        struct ipc_priv_data* priv;
        unsigned int minor;
        unsigned int mask = 0;
        int state = 0;
        
        minor = MINOR(filp->f_dentry->d_inode->i_rdev);               
        priv = (struct ipc_priv_data*)filp->private_data;
        
	poll_wait(filp, &priv->wq, wait);
	poll_wait(filp, &priv->rq, wait);
		
	state = atomic_read(&priv->vc->state);
	
	if(state == CHANNEL_OPEN){
		mask |= POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;
		return mask;
	}
	
	if(state == CHANNEL_WRITE_ONGOING){
		mask |= POLLIN | POLLRDNORM;
	}
	
	if(state == CHANNEL_READ_ONGOING){
		mask |= POLLOUT | POLLWRNORM;
	}
		
	return mask;
}
 
/*!
 * This function is called to resume the IPC link from a low power state.
 * Not yet implemented
 *
 * @param   dev   the device structure used to give information on which IPC
 *                link (0 through 3 channels) to suspend
 * @param   level the stage in device suspension process that we want the
 *                link to be put in
 *
 * @return  The function always returns 0.
 */
static int mxc_ipc_resume(struct device *dev, u32 level)
{
        return 0; 
}

/*!
 * This function is called to put the IPC link in a low power state.
 * Not yet implemented.
 *
 * @param   dev   the device structure used to give information on which IPC
 *                link (0 through 3 channels) to suspend
 * @param   state the power state the device is entering
 * @param   level the stage in device suspension process that we want the
 *                link to be put in
 *
 * @return  The function always returns 0.
 */
static int mxc_ipc_suspend(struct device *dev, u32 state, u32 level)
{
        return 0;
}

static struct file_operations mxc_ipc_fops = {
        .owner     = THIS_MODULE,
        .read      = mxc_ipc_read,
        .write     = mxc_ipc_write,
        .writev	   = mxc_ipc_writev,
        .poll 	   = mxc_ipc_poll,
        .ioctl     = mxc_ipc_ioctl,
        .open      = mxc_ipc_open,
        .release   = mxc_ipc_close,
};

/*!
 * This is platform device structure for adding IPC
 */
static struct platform_device mxc_ipc_device = {
        .name           = "mxc_ipc",
        .id             = 0,
};

/*!
 * This structure contains pointers to the power management callback functions.
 */
static struct device_driver mxc_ipc_driver = {
        .name           = "mxc_ipc",
        .bus            = &platform_bus_type,
        .suspend        = mxc_ipc_suspend,
        .resume         = mxc_ipc_resume,
};

/*!
 * This function is used to unload the module.
 */
static void ipc_cleanup_module(void)
{   
        free_sdma_channel(LOGGING_CHANNEL);
        free_sdma_channel(PACKET_DATA_CHANNEL);
        free_mu_channel(SHORT_MESSAGE_CHANNEL1);
        free_mu_channel(SHORT_MESSAGE_CHANNEL0);
        
        driver_unregister(&mxc_ipc_driver);
        platform_device_unregister(&mxc_ipc_device);
        unregister_chrdev(major_num, "mxc_ipc");
        
        DPRINTK("IPC Driver Module Unloaded\n");
}

/*!
 * This function is used to load the module. All initializations and 
 * resources requesting is done here
 *
 * @return   Returns 0 on success, -1 otherwise
 */
int __init ipc_init_module(void)
{
        int res = 0;
        int i;
        
        res = register_chrdev(MXC_IPC_MAJOR, "mxc_ipc", &mxc_ipc_fops);
        if(res < 0){
                DPRINTK("IPC Driver Module was not Loaded successfully\n");
                return -1;
        }
        
        major_num = res;
        res = driver_register(&mxc_ipc_driver);
        if(res == 0){
                res = platform_device_register(&mxc_ipc_device);
                if(res != 0){
                        driver_unregister(&mxc_ipc_driver);
                }
        }
        
        devfs_mk_dir("mxc_ipc");
        devfs_mk_cdev(MKDEV(major_num, 0), S_IFCHR | S_IRUSR | S_IWUSR, 
                        "mxc_ipc/0");
        devfs_mk_cdev(MKDEV(major_num, 1), S_IFCHR | S_IRUSR | S_IWUSR, 
                      "mxc_ipc/1");
        devfs_mk_cdev(MKDEV(major_num, 2), S_IFCHR | S_IRUSR | S_IWUSR, 
                        "mxc_ipc/2");
        devfs_mk_cdev(MKDEV(major_num, 3), S_IFCHR | S_IRUSR | S_IWUSR, 
                        "mxc_ipc/3");
                        
        if(initialize_mu_channel(SHORT_MESSAGE_CHANNEL0) == -1){
        	return -1;
        }
        
        if(initialize_mu_channel(SHORT_MESSAGE_CHANNEL1) == -1){
        	goto cleanup1;
        }
        
        if(initialize_sdma_channel(PACKET_DATA_CHANNEL, IPC_RDWR) == -1){
        	goto cleanup2;
        }
        
        if(initialize_sdma_channel(LOGGING_CHANNEL, IPC_READ) == -1){
        	goto cleanup3;
        }
        
        for(i = 0; i < IPC_MAX_VIRTUAL_CHANNELS; i++){
          	virtual_channels[i].ipc = NULL;
          
          	sema_init(&virtual_channels[i].sem, 1);
          	atomic_set(&virtual_channels[i].state, CHANNEL_CLOSED);
        }

        printk("IPC driver successfully loaded.\n");
        return major_num;
        
cleanup3:
        free_sdma_channel(PACKET_DATA_CHANNEL);
cleanup2:
        free_mu_channel(SHORT_MESSAGE_CHANNEL1);
cleanup1:
        free_mu_channel(SHORT_MESSAGE_CHANNEL0);
  
        return -1;
}

module_init(ipc_init_module);
module_exit(ipc_cleanup_module);

EXPORT_SYMBOL(hw_ctrl_ipc_open);
EXPORT_SYMBOL(hw_ctrl_ipc_close);
EXPORT_SYMBOL(hw_ctrl_ipc_read);
EXPORT_SYMBOL(hw_ctrl_ipc_write);

MODULE_AUTHOR("Freescale Semiconductor");
MODULE_DESCRIPTION("ArgonPlus/SCMA11 IPC driver");
MODULE_LICENSE("GPL");
