/*
 * icx1087_key_base.c -- ICX1087 key base driver
 *
 * Copyright 2006,2007 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


#include <linux/device.h>


static int  icx1087_key_base_suspend(struct device *dev, u32 state, u32 level);
static int  icx1087_key_base_resume(struct device *dev, u32 level);
static void icx1087_key_base_shutdown(struct device *dev);


static struct platform_device _icx1087_key_device = {
    .name      = "icx1087_key",
    .id        = 0,
};


static struct device_driver _icx1087_key_driver = {
    .name      = "icx1087_key",
    .bus       = &platform_bus_type,
    .suspend   = icx1087_key_base_suspend,
    .resume    = icx1087_key_base_resume,
    .shutdown  = icx1087_key_base_shutdown
};


static int  (*_suspend_handler)(struct device *dev, u32 state, u32 level) = 0;
static int  (*_resume_handler)(struct device *dev, u32 level) = 0;
static void (*_shutdown_handler)(struct device *dev) = 0;
static int  (*_get_code_handler)(int) = 0;

#ifdef DEBUG_ICX1087_KEY_BASE
#define DEBUG_MSG(format, arg...) \
    printk(KERN_DEBUG format , ## arg)
#else /* DEBUG */
#define DEBUG_MSG(format, args...) 
#endif /* DEBUG */


/**
 *
 */
void icx1087_key_base_register_suspend_handler(int (*func)(struct device *dev, u32 state, u32 level))
{
    _suspend_handler = func;
}

/**
 *
 */
void icx1087_key_base_register_resume_handler(int (*func)(struct device *dev, u32 level))
{
    _resume_handler = func;
}

/**
 *
 */
void icx1087_key_base_register_shutdown_handler(void (*func)(struct device *dev))
{
    _shutdown_handler = func;
}

/**
 *
 */
void icx1087_key_base_register_get_code_handler(int (*func)(int))
{
    _get_code_handler = func;
}



/**
 *
 */
int icx1087_key_base_init(void)
{
    int nRet = 0;
    
    printk(KERN_INFO "Starting icx1087_key_base.\n");
    
    nRet = platform_device_register(&_icx1087_key_device);

	if (0 != nRet) {
		printk(KERN_ERR "%s() platform_device_register() failed =%d\n", __FUNCTION__, nRet);
		nRet = -ENODEV;
		goto register_out;
	}
    
    nRet = driver_register(&_icx1087_key_driver);

	if (0 != nRet) {
		printk(KERN_ERR "%s() driver_register() failed =%d\n", __FUNCTION__, nRet);
        platform_device_unregister(&_icx1087_key_device);
		nRet = -ENODEV;
		goto register_out;
	}
    

  register_out:
    return nRet;
}



/**
 *
 */
void icx1087_key_base_exit(void)
{
    driver_unregister(&_icx1087_key_driver);
    platform_device_unregister(&_icx1087_key_device);
    return;
}


/*
 *
 */
static int icx1087_key_base_suspend(struct device *dev, u32 state, u32 level)
{
    DEBUG_MSG("%s() \n", __FUNCTION__);
    
    if (_suspend_handler) {
        return _suspend_handler(dev, state, level);
    } else {
        return 0;
    }

}


/**
 *
 */
static int icx1087_key_base_resume(struct device *dev, u32 level)
{
    DEBUG_MSG("%s() \n", __FUNCTION__);

    if (_resume_handler) {
        return _resume_handler(dev, level);
    } else {
        return 0;
    }

}



/**
 *
 */
static void icx1087_key_base_shutdown(struct device *dev)
{
    DEBUG_MSG("%s() \n", __FUNCTION__);

    if (_shutdown_handler) {
        _shutdown_handler(dev);
    }
}


/**
 *  Called from pm
 *  nFactor1 : PWC_FACTOR1
 */
int sonykpd_get_code(int nFactor1)
{
    int nRet = 0x0f; /* default : OPEN */

    if (_get_code_handler) {
        return _get_code_handler(nFactor1);
    } else {
        return nRet;
    }
    
}


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("Sony ICX1087 key base driver");

EXPORT_SYMBOL(icx1087_key_base_register_suspend_handler);
EXPORT_SYMBOL(icx1087_key_base_register_resume_handler);
EXPORT_SYMBOL(icx1087_key_base_register_shutdown_handler);
EXPORT_SYMBOL(icx1087_key_base_register_get_code_handler);
EXPORT_SYMBOL(sonykpd_get_code);

module_init(icx1087_key_base_init);
module_exit(icx1087_key_base_exit);


