/*
 *  File Name       : linux/drivers/input/touchscreen/mp200_ts.c
 *  Function        : ADS7846 touchscreen controller
 *  Release Version : Ver 1.01
 *  Release Date    : 2006/04/05
 *
 *  Copyright (C) NEC Electronics Corporation 2005-2006
 *
 *
 *  This program is free software;you can redistribute it and/or modify itunder the terms of
 *  the GNU General Public License as published by Free Softwere Foundation;either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/input.h>
#include <linux/device.h>

#include <asm/irq.h>

#include <asm/arch/gpiocore.h>
#include <asm/arch/uwire.h>
#include <asm/arch/lcdc_intr.h>

#include "mp200_ts.h"

/* global parameters */
static unsigned int gard_time = GARD_TIME;
static unsigned int scan_time = SCAN_TIME;
static unsigned int adjust_vsync = ADJUST_VSYNC;
static unsigned int gen_pen_event = GEN_PEN_EVENT;
static unsigned int post_jitter_filter = POST_JITTER_FILTER;
static unsigned int filter = FILTER_FIR;
static unsigned int ts_irq_mask_flag = 0;

/*
 * get absolute value
 */
static inline int mp200_ts_abs(int x)
{
	return (x < 0) ? -x : x;
}

/*
 * ADS7846 PENIRQ interrupt enable
 */
static inline void mp200_ts_irq_enable(int irq)
{
	ts_irq_mask_flag = 0;
	enable_irq(irq);
}

/*
 * ADS7846 PENIRQ interrupt disable
 */
static inline void mp200_ts_irq_disable(int irq)
{
	disable_irq(irq);
	ts_irq_mask_flag = 1;
}

/*
 * get ADS7846 PENIRQ status
 */
static inline int mp200_ts_pen_status(void)
{
	return ((gpiocore_read(GPIOCORE_GPIO_I, GPIOCORE_GPIO_TS_INT)) ? 0 : 1);
}

/*
 * get touch screen parameter
 */
static void mp200_ts_io(unsigned long data)
{
	struct mp200_ts_driver *ts = (struct mp200_ts_driver *)data;
	int err, i;
	int Xout, Yout, Pout;
	int x, y;
	int dx, dy;
	int pen_status;
	unsigned long t = 0;

	pen_status = mp200_ts_pen_status();

	/* Pen Down/Up event */
	if ((ts->pendown == 0) && (pen_status == 1)) {
#ifdef MP200_TS_DEBUG
		printk(KERN_INFO "%s(): PenDown\n", __FUNCTION__);
#endif
		ts->pendown = 1;
		if (ts->generate_pen_event == GEN_PEN_EVENT) {
			input_report_key(&ts->mp200_ts_dev, BTN_TOUCH, 1);
		}
	} else if ((ts->pendown == 1) && (pen_status == 0)) {
#ifdef MP200_TS_DEBUG
		printk(KERN_INFO "%s(): PenUp\n", __FUNCTION__);
#endif
		ts->pendown = 0;
		ts->nsample = 0;
		/* init filter parameter */
		if (ts->filter_type == FILTER_IIR) {
			ts->x_iir = 0;
			ts->y_iir = 0;
		} else if (ts->filter_type == FILTER_FIR) {
			ts->fir_index = 0;
		}
		/* add event */
		input_report_abs(&ts->mp200_ts_dev, ABS_X, 0);
		input_report_abs(&ts->mp200_ts_dev, ABS_Y, 0);
		input_report_abs(&ts->mp200_ts_dev, ABS_PRESSURE, 0);
		input_sync(&ts->mp200_ts_dev);
		if (ts->generate_pen_event == GEN_PEN_EVENT) {
			input_report_key(&ts->mp200_ts_dev, BTN_TOUCH, 0);
		}
		return;
	}

	/* Pen Move event */
	if (ts->pendown != 0) {
		/* get touch screen parameter Y */
		err = uwire_vl_read(ADS7846_UWIRE_CS, ADS7846_UWIRE_CLK,
				    MEASURE_Y, ADS7846_DLEN, &t);
		if (err != 0) {
			printk(KERN_INFO "%s(): get MEASURE_Y faild (%d)\n",
			       __FUNCTION__, err);
			return;
		}
		ts->y_raw = t & ADS7846_DATA_MASK;

		/* get touch screen parameter X */
		err = uwire_vl_read(ADS7846_UWIRE_CS, ADS7846_UWIRE_CLK,
				    MEASURE_X, ADS7846_DLEN, &t);
		if (err != 0) {
			printk(KERN_INFO "%s(): get MEASURE_X faild (%d)\n",
			       __FUNCTION__, err);
			return;
		}
		ts->x_raw = t & ADS7846_DATA_MASK;

		/* get touch screen parameter Z1 */
		err = uwire_vl_read(ADS7846_UWIRE_CS, ADS7846_UWIRE_CLK,
				    MEASURE_Z1, ADS7846_DLEN, &t);
		if (err != 0) {
			printk(KERN_INFO "%s(): get MEASURE_Z1 faild (%d)\n",
			       __FUNCTION__, err);
			return;
		}
		ts->z1_raw = t & ADS7846_DATA_MASK;

#ifdef MP200_TS_DEBUG
		printk(KERN_INFO "%s(): X=%d,Y=%d,Z1=%d\n",
		       __FUNCTION__, ts->x_raw, ts->y_raw, ts->z1_raw);
#endif

		if (ts->filter_type == FILTER_IIR) {
			if (ts->nsample == ts->ntotal) {
				ts->x_iir -= ts->x_iir >> ts->iir_shift;
				ts->y_iir -= ts->y_iir >> ts->iir_shift;
			} else {
				ts->nsample++;
			}
			ts->x_iir += ts->x_raw;
			ts->y_iir += ts->y_raw;

			if (ts->nsample <= ts->iir_output_threshold) {
				return;
			}

			x = ts->x_iir / ts->nsample;
			y = ts->y_iir / ts->nsample;
		} else if (ts->filter_type == FILTER_FIR) {
			ts->fir_x[ts->fir_index] = ts->x_raw;
			ts->fir_y[ts->fir_index] = ts->y_raw;

			if (++ts->fir_index == FIR_LEN) {
				ts->fir_index = 0;
			}
			if (ts->nsample < FIR_LEN) {
				ts->nsample++;
				return;
			}
			x = y = 0;
			for (i = 0; i < FIR_LEN; i++) {
				x += ts->fir_x[i];
				y += ts->fir_y[i];
			}
			x /= FIR_LEN;
			y /= FIR_LEN;
		} else {
			x = ts->x_raw;
			y = ts->y_raw;
		}

		Xout = x;
		Yout = y;
		Pout = ts->z1_raw;

		/* jitter filter */
		dx = mp200_ts_abs(Xout - ts->x_out);
		dy = mp200_ts_abs(Yout - ts->y_out);
		if ((dx > ts->post_jitter_filter)
		    || (dy > ts->post_jitter_filter)) {
			ts->x_out = Xout;
			ts->y_out = Yout;
		} else {
			Xout = ts->x_out;
			Yout = ts->y_out;
		}

#ifdef MP200_TS_DEBUG
		printk(KERN_INFO "%s(): X=%d,Y=%d,PRESSURE=%d,time=%u\n",
		       __FUNCTION__, Xout, Yout, Pout, (unsigned int)jiffies);
#endif
		/* add event */
		input_report_abs(&ts->mp200_ts_dev, ABS_X, Xout);
		input_report_abs(&ts->mp200_ts_dev, ABS_Y, Yout);
		input_report_abs(&ts->mp200_ts_dev, ABS_PRESSURE, Pout);
		input_sync(&ts->mp200_ts_dev);
	}
}

/* scan timer */
static void mp200_ts_scan_time(unsigned long data)
{
	struct mp200_ts_driver *ts = (struct mp200_ts_driver *)data;

	mp200_ts_io(data);

	if (ts->adjust_vsync == ADJUST_VSYNC) {
		if (ts->pendown != 0) {
			;
		} else {
			disable_vsync_action(&ts->vsync_action);
			mp200_ts_irq_enable(ts->irq);
		}
	} else {
		if (ts->pendown != 0) {
			mod_timer(&ts->acq_timer, jiffies + ts->scan_time);
		} else {
			mp200_ts_irq_enable(ts->irq);
		}
	}
}

/* first scan (chattering gard) */
static void mp200_ts_scan_first(unsigned long data)
{
	struct mp200_ts_driver *ts = (struct mp200_ts_driver *)data;

	if (mp200_ts_pen_status() != 0) {

		mp200_ts_io(data);

		if (ts->pendown != 0) {
			if (ts->adjust_vsync == ADJUST_VSYNC) {
				enable_vsync_action(&ts->vsync_action);
			} else {
				mod_timer(&ts->acq_timer,
					  jiffies + ts->scan_time);
			}
		} else {
			mp200_ts_irq_enable(ts->irq);
		}
	} else {
		mp200_ts_irq_enable(ts->irq);
	}
}

/* ADS7846 PENIRQ interrupt handler */
static irqreturn_t
mp200_ts_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct mp200_ts_driver *ts = dev_id;

	mp200_ts_irq_disable(ts->irq);

	mod_timer(&ts->chatter_timer, jiffies + ts->gard_time);

	return IRQ_HANDLED;
}

/*
 * mp200 touch screen driver initialize
 */
static int __init mp200_ts_probe(struct device *dev)
{
	int err;
	unsigned long tmp;
	struct mp200_ts_driver *ts;

	printk(KERN_INFO "Starting touchscreen.\n");

	ts = kmalloc(sizeof(struct mp200_ts_driver), GFP_KERNEL);
	if (ts == 0) {
		return -ENOMEM;
	}

	memset(ts, 0, sizeof(struct mp200_ts_driver));

	/* set LCD VSYNC parameters */
	if (adjust_vsync == 0) {
		ts->adjust_vsync = 0;
	} else {
		ts->adjust_vsync = ADJUST_VSYNC;
	}

	/* set filter parameters */
	ts->nsample = 0;
	ts->filter_type = filter;
	/* FIR */
	ts->fir_index = 0;
	/* IIR */
	ts->x_iir = 0;
	ts->y_iir = 0;
	ts->iir_shift = IIR_SHIFT_BITS;	/* 2**IIR_SHIFT_BITS */
	ts->iir_output_threshold = (1 << ts->iir_shift) - 1;
	ts->ntotal = (1 << ts->iir_shift);
	/* post jitter filter */
	ts->post_jitter_filter = post_jitter_filter;
	ts->x_out = 0;
	ts->y_out = 0;

	if (gen_pen_event == 0) {
		ts->generate_pen_event = 0;
	} else {
		ts->generate_pen_event = GEN_PEN_EVENT;
	}

	ts->pendown = 0;

	/* gard time initialize */
	ts->gard_time = gard_time * HZ / 1000;
	if (ts->gard_time < 1) {
		ts->gard_time = GARD_TIME * HZ / 1000;
	}

	/* scan time initialize */
	ts->scan_time = scan_time * HZ / 1000;
	if (ts->scan_time < 1) {
		ts->scan_time = SCAN_TIME * HZ / 1000;
	}

	/* chattering gard timer */
	init_timer(&ts->chatter_timer);
	ts->chatter_timer.function = mp200_ts_scan_first;
	ts->chatter_timer.data = (unsigned long)ts;

	if (ts->adjust_vsync == ADJUST_VSYNC) {
		/* use VSYNC scan */
		ts->vsync_action.count = VSYNC_COUNT;	/* VSYNC x 3 */
		ts->vsync_action.reschedule = 1;
		ts->vsync_action.immediately = 0;
		ts->vsync_action.high_priority = 0;
		ts->vsync_action.function = mp200_ts_scan_time;
		ts->vsync_action.data = (unsigned long)ts;
		err = add_vsync_action(&ts->vsync_action);
		if (err != 0) {
			printk(KERN_INFO
			       "%s(): add_vsync_action() failed (%d)\n",
			       __FUNCTION__, err);
			err = -EFAULT;
			goto out_kfree;
		}
		disable_vsync_action(&ts->vsync_action);
	} else {
		/* internal scan timer */
		init_timer(&ts->acq_timer);
		ts->acq_timer.function = mp200_ts_scan_time;
		ts->acq_timer.data = (unsigned long)ts;
	}

	/* ADS7846 get X/PowerDown/enable interrupt */
	err = uwire_vl_read(ADS7846_UWIRE_CS, ADS7846_UWIRE_CLK,
			    MEASURE_X, ADS7846_DLEN, &tmp);
	if (err != 0) {
		printk(KERN_INFO "%s(): uwire_vl_read() failed (%d)\n",
		       __FUNCTION__, err);
		goto out_del_vsync_action;
	}

	init_input_dev(&ts->mp200_ts_dev);
	ts->mp200_ts_dev.name = MP200_TS_NAME;
	ts->mp200_ts_dev.evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
	ts->mp200_ts_dev.keybit[LONG(BTN_TOUCH)] = BIT(BTN_TOUCH);
	ts->mp200_ts_dev.absbit[0] =
	    BIT(ABS_X) | BIT(ABS_Y) | BIT(ABS_PRESSURE);

	input_register_device(&ts->mp200_ts_dev);

	ts->irq = INT_GPIO_TS_INT;

	/* set interrupt low level */
	set_irq_type(ts->irq, IRQT_LOW);

	/* interrupt initialize */
	err = request_irq(ts->irq, mp200_ts_interrupt, SA_INTERRUPT,
			  MP200_TS_NAME, ts);
	if (err != 0) {
		printk(KERN_INFO "%s(): request_irq error\n", __FUNCTION__);
		goto out_input_unregister_device;
	}

	dev_set_drvdata(dev, ts);

	return 0;

      out_input_unregister_device:
	input_unregister_device(&ts->mp200_ts_dev);
      out_del_vsync_action:
	if (ts->adjust_vsync == ADJUST_VSYNC) {
		del_vsync_action(&ts->vsync_action);
	}
      out_kfree:
	kfree(ts);

	return err;
}

static int __exit mp200_ts_remove(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct mp200_ts_driver *ts = dev_get_drvdata(&pdev->dev);

	dev_set_drvdata(&pdev->dev, NULL);

	free_irq(ts->irq, ts);

	del_timer_sync(&ts->chatter_timer);
	if (ts->adjust_vsync == ADJUST_VSYNC) {
		del_vsync_action(&ts->vsync_action);
	} else {
		del_timer_sync(&ts->acq_timer);
	}

	input_unregister_device(&ts->mp200_ts_dev);

	kfree(ts);

	return 0;
}

static int mp200_ts_suspend(struct device *dev, u32 state, u32 level)
{
#ifdef MP200_TS_DEBUG
	printk(KERN_INFO "%s(): bus_id=%s state=%d level=%d\n",
	       __FUNCTION__, dev->bus_id, state, level);
#endif

	switch (level) {
	case SUSPEND_DISABLE:
		if (ts_irq_mask_flag != 0) {
			return -EAGAIN;
		}
		break;
	default:
		break;
	}

	return 0;
}

static struct device_driver mp200_ts_driver = {
	.name = MP200_TS_NAME,
	.bus = &platform_bus_type,
	.probe = mp200_ts_probe,
	.remove = __exit_p(mp200_ts_remove),
	.suspend = mp200_ts_suspend,
};

static int __init mp200_ts_init(void)
{
	return driver_register(&mp200_ts_driver);
}

static void __exit mp200_ts_exit(void)
{
	driver_unregister(&mp200_ts_driver);
}

#ifndef MODULE
/*
 * boot command line parameter
 * touchscreen=gardtime[ms]
 */
static int __init mp200_ts_setup(char *str)
{
	int inkb[2];

	get_options(str, ARRAY_SIZE(inkb), inkb);

	if (inkb[0] >= 1 && inkb[1] > 0) {
		gard_time = (unsigned int)inkb[1];
	}
	return 0;
}

__setup("touchscreen=", mp200_ts_setup);
#endif

module_init(mp200_ts_init);
module_exit(mp200_ts_exit);

MODULE_PARM(gard_time, "l");
MODULE_PARM_DESC(gard_time, "chattering gard time [ms]");

MODULE_PARM(scan_time, "l");
MODULE_PARM_DESC(scan_time, "scan interval time [ms]");

MODULE_PARM(gen_pen_event, "l");
MODULE_PARM_DESC(gen_pen_event, "generate pen up/down event");

MODULE_PARM(adjust_vsync, "l");
MODULE_PARM_DESC(adjust_vsync, "scan adjust vsync");

MODULE_PARM(post_jitter_filter, "l");
MODULE_PARM_DESC(post_jitter_filter, "Post XY-Point jitter filer threshold");

MODULE_PARM(filter, "l");
MODULE_PARM_DESC(filter, "filter type: 0-FIR/1-IIR/other-NONE");

MODULE_LICENSE("GPL");
