/*
 * Copyright 2005 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Freescale Semiconductor MXC Video4Linux2 Video Output driver
 *
 */
/*!
 * @file mxc_v4l2_output.c
 * 
 * @brief MXC V4L2 Video Output Driver  
 *
 * Video4Linux2 Output Device using MXC IPU Post-processing functionality.
 *
 * @ingroup MXC_V4L2_OUTPUT
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/semaphore.h>
#include "asm/arch/board.h"
#include "mxc_v4l2_output.h"


#define SDC_FG_FB_FORMAT        IPU_PIX_FMT_RGB565

#define MXC_V4L2_OUT_NUM_OUTPUTS        2
struct v4l2_output mxc_outputs[MXC_V4L2_OUT_NUM_OUTPUTS] = {
        {
                .index  = 0,
                .name   = "SDC vout",
                .type   = V4L2_OUTPUT_TYPE_ANALOG, /* not really correct, but no other choice */
                .audioset = 0,
                .modulator = 0,
                .std    = V4L2_STD_UNKNOWN
        },
        {
                .index  = 1,
                .name   = "ADC vout",
                .type   = V4L2_OUTPUT_TYPE_ANALOG, /* not really correct, but no other choice */
                .audioset = 0,
                .modulator = 0,
                .std    = V4L2_STD_UNKNOWN
        }
};

//#define MXC_V4L2OUT_DEBUG
#ifdef MXC_V4L2OUT_DEBUG

#  define DDPRINTK(fmt, args...) printk(KERN_ERR"%s :: %d :: %s - " fmt, __FILE__,__LINE__,__FUNCTION__ , ## args)
#  define DPRINTK(fmt, args...) printk("%s: " fmt, __FUNCTION__ , ## args)

#  define FUNC_START	DDPRINTK(" func start\n")
#  define FUNC_END	DDPRINTK(" func end\n")

#  define FUNC_ERR printk(KERN_ERR"%s :: %d :: %s  err= %d \n", __FILE__,__LINE__,__FUNCTION__ ,err)

#else //MXC_V4L2OUT_DEBUG

#define DDPRINTK(fmt, args...)  do {} while(0)
#define DPRINTK(fmt, args...)   do {} while(0)

#define FUNC_START
#define FUNC_END

#endif //MXC_V4L2OUT_DEBUG


static int video_nr = 16;
static spinlock_t g_lock = SPIN_LOCK_UNLOCKED;

// debug counters
uint32_t g_irq_cnt;
uint32_t g_buf_output_cnt;
uint32_t g_buf_q_cnt;
uint32_t g_buf_dq_cnt;

#define QUEUE_SIZE (MAX_FRAME_NUM + 1)
static __inline int queue_size(v4l_queue *q)
{
        if (q->tail >= q->head)
                return (q->tail - q->head);
        else
                return ((q->tail + QUEUE_SIZE) - q->head);
}

static __inline int queue_buf(v4l_queue *q, int idx)
{
        if (((q->tail + 1) % QUEUE_SIZE) == q->head)
                return -1; // queue full
        q->list[q->tail] = idx;
        q->tail = (q->tail + 1) % QUEUE_SIZE;
        return 0;
}

static __inline int dequeue_buf(v4l_queue *q)
{
        int ret;
        if (q->tail == q->head)
                return -1; // queue empty
        ret = q->list[q->head];
        q->head = (q->head + 1) % QUEUE_SIZE;
        return ret;
}

/*!
 * Private function to allocate IPU buffers
 *
 * @param bufs          Output array of pointers to buffers allocated
 *
 * @param num_buf       Input number of buffers to allocate
 *
 * @param size          Input size for each buffer to allocate
 *
 * @return status  -0 Successfully allocated a buffer, -ENOBUFS	failed.
 */
static int 
mxc_allocate_buffers(void * bufs[], int num_buf, int size)
{
        int i;

        for (i = 0; i < num_buf; i++)
        {  
                bufs[i] = (void*)IPU_Malloc(size);
                if (bufs[i] == 0)
                {
                        DPRINTK("IPU_Malloc failed.\n");
                        return -ENOBUFS;
                }
                DPRINTK("allocated @ paddr=0x%08X, size=%d.\n", (u32)bufs[i], size);
        } 

        return 0;
}

/*!
 * Private function to free IPU buffers
 *
 * @param bufs      void *  
 *
 * @param num_buf   int  
 *
 * @return status  0 success.
 */
static int 
mxc_free_buffers(void * bufs[], int num_buf)
{
        int i;

        for (i = 0; i < num_buf; i++)
        {
                if (bufs[i] != 0)
                {
                        IPU_Free((u32)bufs[i]);
                        DPRINTK("freed @ paddr=0x%08X\n", (u32)bufs[i]);
                        bufs[i] = NULL;
                }
        }
        return 0;
}

static irqreturn_t mxc_v4l2out_pp_in_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
        u32 index;
        int last_buf;
        vout_data * vout = dev_id;

        g_irq_cnt++;

//        printk("pp_irq %d, IPU_CUR_BUF=0x%08X, IPU_TASK_STAT=0x%08X, IPU_INT_STAT5=0x%08X\n", vout->next_done_ipu_buf, *((u32*)0xD42C0010), *((u32*)0xD42C001C), *((u32*)0xD42C004C));
        
        /* Process previous buffer */
        last_buf = vout->ipu_buf[vout->next_done_ipu_buf];
        if (last_buf != -1)
        {
                g_buf_output_cnt++;
                vout->v4l2_bufs[last_buf].flags = V4L2_BUF_FLAG_DONE;
                queue_buf(&vout->done_q, last_buf);
                vout->ipu_buf[vout->next_done_ipu_buf] = -1;
                wake_up_interruptible(&vout->v4l_bufq);
//                printk("pp_irq: buf %d done\n", vout->next_done_ipu_buf);
                vout->next_done_ipu_buf = !vout->next_done_ipu_buf;
        }

        if (vout->state == STATE_STREAM_STOPPING) {
                if ((vout->ipu_buf[0] == -1) && (vout->ipu_buf[1] == -1)) {
                        vout->state = STATE_STREAM_OFF;
                }
                goto exit;
        }
        
        /* Setup next buffer */
        index = dequeue_buf(&vout->ready_q);
        if (index == -1)        // no buffers ready
        {
                //printk("pp_irq: no queued buffers ready, buf = %d\n", vout->next_ipu_buf);
                // Check if both buffers are idle/empty
                if ((vout->ipu_buf[0] == -1) && (vout->ipu_buf[1] == -1)) {
                        DPRINTK("pp_irq: ipu stopped done = %d, rdy = %d\n", vout->next_done_ipu_buf, vout->next_rdy_ipu_buf);
                        vout->state = STATE_STREAM_PAUSED;
                }
                goto exit;
        }

        g_buf_dq_cnt++;
        vout->ipu_buf[vout->next_rdy_ipu_buf] = index;
        if (ipu_update_channel_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, 
                                      vout->next_rdy_ipu_buf, 
                                      vout->queue_buf_paddr[index]) < 0) {
                DPRINTK("pp_irq: unable to update buffer %d\n", vout->next_rdy_ipu_buf);
                goto exit;
        }
        ipu_select_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, vout->next_rdy_ipu_buf);
//        printk("pp_irq: setup buffer index=%d in ipu buffer %d\n", index, vout->next_rdy_ipu_buf);
        vout->next_rdy_ipu_buf = !vout->next_rdy_ipu_buf;


        if (vout->ipu_buf[vout->next_rdy_ipu_buf] == -1) {
                index = dequeue_buf(&vout->ready_q);
                if (index != -1)        // another buffer ready
                {
                        g_buf_dq_cnt++;
                        vout->ipu_buf[vout->next_rdy_ipu_buf] = index;
                        if (ipu_update_channel_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, 
                                                      vout->next_rdy_ipu_buf, 
                                                      vout->queue_buf_paddr[index]) < 0) {
                                printk("pp_irq: unable to update buffer %d\n", vout->next_rdy_ipu_buf);
                        }
                        ipu_select_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, vout->next_rdy_ipu_buf);
//                        printk("pp_irq: setup buffer index=%d in ipu buffer %d\n", index, vout->next_rdy_ipu_buf);
                        vout->next_rdy_ipu_buf = !vout->next_rdy_ipu_buf;
                }
        }        
                
exit:
        return IRQ_HANDLED;
}





/*!
 * Kick off the voutera
 *
 * @param vout      structure vout_data * 
 *
 * @return status  0 Success
 */
static int 
mxc_v4l2out_streamon_sdc(vout_data *vout)
{
        ipu_channel_params_t pp_params;
        int pp_in_buf[2];
        u16 out_width;
        u16 out_height;
        ipu_channel_t sdc_input_ch = MEM_PP_MEM;

        if (!vout)
                return -EINVAL;

        if (queue_size(&vout->ready_q) < 2) {
                DPRINTK("2 buffers not been queued yet!\n");
                return -EINVAL;
        }

        if (vout->rotate >= IPU_ROTATE_90_RIGHT) {
                out_width = vout->crop_current.height;
                out_height = vout->crop_current.width;
        }
        else {
                out_width = vout->crop_current.width;
                out_height = vout->crop_current.height;
        }

        vout->next_done_ipu_buf = vout->next_rdy_ipu_buf = 0;
        vout->ipu_buf[0] = pp_in_buf[0] = dequeue_buf(&vout->ready_q);
        vout->ipu_buf[1] = pp_in_buf[1] = dequeue_buf(&vout->ready_q);

        memset(&pp_params, 0, sizeof(pp_params));
        pp_params.mem_pp_mem.in_width = vout->v2f.fmt.pix.width;
        pp_params.mem_pp_mem.in_height = vout->v2f.fmt.pix.height;
        pp_params.mem_pp_mem.in_pixel_fmt = vout->v2f.fmt.pix.pixelformat;
        pp_params.mem_pp_mem.out_width = out_width;
        pp_params.mem_pp_mem.out_height = out_height;
        pp_params.mem_pp_mem.out_pixel_fmt = SDC_FG_FB_FORMAT;
        if (ipu_init_channel(MEM_PP_MEM, &pp_params) != 0) {
                DPRINTK(KERN_ERR"Error initializing PP channel\n");
                return -EINVAL;
        }

        if (ipu_init_channel_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, 
                                    pp_params.mem_pp_mem.in_pixel_fmt,
                                    pp_params.mem_pp_mem.in_width, pp_params.mem_pp_mem.in_height,
                                    pp_params.mem_pp_mem.in_width, IPU_ROTATE_NONE,
                                    vout->queue_buf_paddr[pp_in_buf[0]], 
                                    vout->queue_buf_paddr[pp_in_buf[1]]) != 0) {
                DPRINTK(KERN_ERR"Error initializing PP input buffer\n");
                return -EINVAL;
        }

        if (vout->rotate >= IPU_ROTATE_HORIZ_FLIP) {
                if (vout->rot_pp_bufs[0]) {
                        mxc_free_buffers(vout->rot_pp_bufs, 2);
                }
                if (mxc_allocate_buffers(vout->rot_pp_bufs, 2, vout->sdc_fg_buf_size) < 0) {
                        return -ENOBUFS;
                }
                
                        
                if (ipu_init_channel_buffer(MEM_PP_MEM, IPU_OUTPUT_BUFFER, 
                                            pp_params.mem_pp_mem.out_pixel_fmt,
                                            out_width, out_height,
                                            out_width, IPU_ROTATE_NONE,
                                            vout->rot_pp_bufs[0], 
                                            vout->rot_pp_bufs[1]) != 0) {
                        DPRINTK(KERN_ERR"Error initializing PP output buffer\n");
                        return -EINVAL;
                }

                if (ipu_init_channel(MEM_ROT_PP_MEM, NULL) != 0) {
                        DPRINTK(KERN_ERR"Error initializing PP ROT channel\n");
                        return -EINVAL;
                }


                if (ipu_init_channel_buffer(MEM_ROT_PP_MEM, IPU_INPUT_BUFFER, 
                                            pp_params.mem_pp_mem.out_pixel_fmt,
                                            out_width, out_height,
                                            out_width, vout->rotate,
                                            vout->rot_pp_bufs[0], 
                                            vout->rot_pp_bufs[1]) != 0) {
                        DPRINTK(KERN_ERR"Error initializing PP ROT input buffer\n");
                        return -EINVAL;
                }

                if (vout->rotate >= IPU_ROTATE_90_RIGHT) {
                        out_width = vout->crop_current.width;
                        out_height = vout->crop_current.height;
                }
                
                if (ipu_init_channel_buffer(MEM_ROT_PP_MEM, IPU_OUTPUT_BUFFER, 
                                            pp_params.mem_pp_mem.out_pixel_fmt,
                                            out_width, out_height,
                                            out_width, IPU_ROTATE_NONE,
                                            vout->sdc_fg_bufs[0],
                                            vout->sdc_fg_bufs[1]) != 0) {
                        DPRINTK(KERN_ERR"Error initializing PP output buffer\n");
                        return -EINVAL;
                }

                if (ipu_link_channels(MEM_PP_MEM, MEM_ROT_PP_MEM) < 0) {
                        return -EINVAL;
                }
                ipu_select_buffer(MEM_ROT_PP_MEM, IPU_INPUT_BUFFER, 0);
                ipu_select_buffer(MEM_ROT_PP_MEM, IPU_INPUT_BUFFER, 1);
                ipu_select_buffer(MEM_ROT_PP_MEM, IPU_OUTPUT_BUFFER, 0);
                ipu_select_buffer(MEM_ROT_PP_MEM, IPU_OUTPUT_BUFFER, 1);

                ipu_enable_channel(MEM_ROT_PP_MEM);

                sdc_input_ch = MEM_ROT_PP_MEM;
        } else {
                if (ipu_init_channel_buffer(MEM_PP_MEM, IPU_OUTPUT_BUFFER, 
                                            pp_params.mem_pp_mem.out_pixel_fmt,
                                            out_width, out_height,
                                            out_width, IPU_ROTATE_NONE,
                                            vout->sdc_fg_bufs[0],
                                            vout->sdc_fg_bufs[1]) != 0) {
                        DPRINTK(KERN_ERR"Error initializing PP output buffer\n");
                        return -EINVAL;
                }
        }

        ipu_init_channel(MEM_SDC_FG, NULL);
        ipu_sdc_set_window_pos(MEM_SDC_FG, vout->crop_current.left, vout->crop_current.top);
        if (ipu_init_channel_buffer(MEM_SDC_FG, IPU_INPUT_BUFFER, 
                                    pp_params.mem_pp_mem.out_pixel_fmt,
                                    out_width, out_height, out_width, 
                                    (vout->rotate < IPU_ROTATE_HORIZ_FLIP) ? vout->rotate : IPU_ROTATE_NONE,
                                    vout->sdc_fg_bufs[0],
                                    vout->sdc_fg_bufs[1]) != 0) {
                DPRINTK(KERN_ERR"Error initializing SDC FG buffer\n");
                return -EINVAL;
        }

        if (ipu_link_channels(sdc_input_ch, MEM_SDC_FG) < 0) {
                DPRINTK(KERN_ERR"Error linking ipu channels\n");
                return -EINVAL;
        }

        vout->state = STATE_STREAM_ON;

        ipu_select_buffer(MEM_PP_MEM, IPU_OUTPUT_BUFFER, 0);
        ipu_select_buffer(MEM_PP_MEM, IPU_OUTPUT_BUFFER, 1);
              
        ipu_select_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, 0);
        ipu_select_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, 1);

        ipu_select_buffer(MEM_SDC_FG, IPU_INPUT_BUFFER, 0);
        ipu_select_buffer(MEM_SDC_FG, IPU_INPUT_BUFFER, 1);

        ipu_enable_channel(MEM_PP_MEM);
        ipu_enable_channel(MEM_SDC_FG);

        return 0;
}

/*!
 * Kick off the voutera
 *
 * @param vout      structure vout_data * 
 *
 * @return status  0 Success
 */
static int 
mxc_v4l2out_streamon_adc(vout_data *vout)
{
        ipu_channel_params_t pp_params;
        int pp_in_buf[2];
        u16 out_width;
        u16 out_height;

        if (!vout)
                return -EINVAL;

        if (queue_size(&vout->ready_q) < 2) {
                DPRINTK("2 buffers not been queued yet!\n");
                return -EINVAL;
        }

        if (vout->rotate >= IPU_ROTATE_90_RIGHT) {
                out_width = vout->crop_current.height;
                out_height = vout->crop_current.width;
        }
        else {
                out_width = vout->crop_current.width;
                out_height = vout->crop_current.height;
        }

        vout->next_done_ipu_buf = vout->next_rdy_ipu_buf = 0;
        vout->ipu_buf[0] = pp_in_buf[0] = dequeue_buf(&vout->ready_q);
        vout->ipu_buf[1] = pp_in_buf[1] = dequeue_buf(&vout->ready_q);

        memset(&pp_params, 0, sizeof(pp_params));
        pp_params.mem_pp_adc.in_width = vout->v2f.fmt.pix.width;
        pp_params.mem_pp_adc.in_height = vout->v2f.fmt.pix.height;
        pp_params.mem_pp_adc.in_pixel_fmt = vout->v2f.fmt.pix.pixelformat;
        pp_params.mem_pp_adc.out_width = out_width;
        pp_params.mem_pp_adc.out_height = out_height;
        pp_params.mem_pp_adc.out_pixel_fmt = IPU_PIX_FMT_BGR32;
        pp_params.mem_pp_adc.out_top = vout->crop_current.top;
        pp_params.mem_pp_adc.out_left = vout->crop_current.left;
        if (ipu_init_channel(MEM_PP_ADC, &pp_params) != 0) {
                DPRINTK(KERN_ERR"Error initializing PP channel\n");
                return -EINVAL;
        }

        if (ipu_init_channel_buffer(MEM_PP_ADC, IPU_INPUT_BUFFER, 
                                    pp_params.mem_pp_adc.in_pixel_fmt,
                                    pp_params.mem_pp_adc.in_width, 
                                    pp_params.mem_pp_adc.in_height,
                                    pp_params.mem_pp_adc.in_width, IPU_ROTATE_NONE,
                                    vout->queue_buf_paddr[pp_in_buf[0]], 
                                    vout->queue_buf_paddr[pp_in_buf[1]]) != 0) {
                DPRINTK(KERN_ERR"Error initializing PP input buffer\n");
                return -EINVAL;
        }

        if (ipu_init_channel_buffer(MEM_PP_ADC, IPU_OUTPUT_BUFFER, 
                                    IPU_PIX_FMT_BGR32,
                                    out_width, out_height,
                                    out_height, IPU_ROTATE_NONE,
                                    NULL, NULL) != 0) {
                DPRINTK(KERN_ERR"Error initializing PP output buffer\n");
                return -EINVAL;
        }

        vout->state = STATE_STREAM_ON;

        msleep(100);
        ipu_select_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, 0);
        ipu_select_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, 1);

        ipu_enable_channel(MEM_PP_MEM);

        return 0;
}



/*!
 * Shut down the voutera
 *
 * @param vout      structure vout_data * 
 *
 * @return status  0 Success
 */
static int 
mxc_v4l2out_streamoff(vout_data *vout)
{
        int retval = 0;
        u32 lockflag = 0;
        
        if (!vout)
               return -EINVAL;

        if (vout->state == STATE_STREAM_OFF) {
                return 0;
        }
        
        spin_lock_irqsave(&g_lock, lockflag);

        if (vout->state == STATE_STREAM_ON) {
                vout->state = STATE_STREAM_STOPPING;
        }

        ipu_disable_irq(IPU_IRQ_PP_IN_EOF);
        
        spin_unlock_irqrestore(&g_lock, lockflag);


        if (vout->output == 0) { // SDC
                ipu_disable_channel(MEM_PP_MEM, true);
                if (vout->rotate >= IPU_ROTATE_HORIZ_FLIP) {
                        ipu_disable_channel(MEM_ROT_PP_MEM, true);
                        ipu_unlink_channels(MEM_PP_MEM, MEM_ROT_PP_MEM);
                        ipu_unlink_channels(MEM_ROT_PP_MEM, MEM_SDC_FG);
                }
                else {
                        ipu_unlink_channels(MEM_PP_MEM, MEM_SDC_FG);
                }
                ipu_disable_channel(MEM_SDC_FG, true);
                ipu_uninit_channel(MEM_SDC_FG);
                ipu_uninit_channel(MEM_PP_MEM);
                ipu_uninit_channel(MEM_ROT_PP_MEM);
        }
        else { // ADC
                ipu_disable_channel(MEM_PP_ADC, true);
                ipu_uninit_channel(MEM_PP_ADC);
        }
        vout->ready_q.head = vout->ready_q.tail = 0;

        vout->state = STATE_STREAM_OFF;
        return retval;
}

/*
 * Valid whether the palette is supported
 *
 * @param palette  V4L2_PIX_FMT_RGB565, V4L2_PIX_FMT_BGR24 or V4L2_PIX_FMT_BGR32 
 *
 * @return 1 if supported, 0 if failed
 */
static inline int 
valid_mode(u32 palette)
{
        return ((palette == V4L2_PIX_FMT_RGB565) || 
                (palette == V4L2_PIX_FMT_BGR24)  ||
                (palette == V4L2_PIX_FMT_RGB24)  ||
                (palette == V4L2_PIX_FMT_BGR32)  ||
                (palette == V4L2_PIX_FMT_RGB32)  ||
                (palette == V4L2_PIX_FMT_YUV422P)||
                (palette == V4L2_PIX_FMT_YUV420) );
}

/*
 * Returns bits per pixel for given pixel format
 *
 * @param pixelformat  V4L2_PIX_FMT_RGB565, V4L2_PIX_FMT_BGR24 or V4L2_PIX_FMT_BGR32 
 *
 * @return bits per pixel of pixelformat 
 */
static u32 
fmt_to_bpp(u32 pixelformat)
{
        u32 bpp;
        
        switch(pixelformat)
        {
                case V4L2_PIX_FMT_RGB565:
                        bpp = 16;
                        break;
                case V4L2_PIX_FMT_BGR24:
                case V4L2_PIX_FMT_RGB24:
                        bpp = 24;
                        break;
                case V4L2_PIX_FMT_BGR32:
                case V4L2_PIX_FMT_RGB32:
                        bpp = 32;
                        break;
                default:
                        bpp = 8;
                        break;
        }
        return bpp;
}

/*
 * V4L2 - Handles VIDIOC_G_FMT Ioctl
 *
 * @param vout         structure vout_data * 
 *
 * @param v4l2_format structure v4l2_format * 
 *
 * @return  status    0 success, EINVAL failed
 */
static int 
mxc_v4l2out_g_fmt(vout_data *vout, struct v4l2_format *f)
{
        if (f->type != V4L2_BUF_TYPE_VIDEO_OUTPUT) {
                return -EINVAL;
        }
        *f = vout->v2f;
        return 0;
}

/*
 * V4L2 - Handles VIDIOC_S_FMT Ioctl
 *
 * @param vout         structure vout_data * 
 *
 * @param v4l2_format structure v4l2_format * 
 *
 * @return  status    0 success, EINVAL failed
 */
static int 
mxc_v4l2out_s_fmt(vout_data *vout, struct v4l2_format *f)
{
        int retval = 0;
        u32 size = 0;
        u32 bytesperline;

        if (f->type != V4L2_BUF_TYPE_VIDEO_OUTPUT) {
                retval = -EINVAL;
                goto err0;
        }
        if (!valid_mode(f->fmt.pix.pixelformat)) {
                DPRINTK("pixel format not supported\n");
                retval = -EINVAL;
                goto err0;
        }  

        bytesperline = (f->fmt.pix.width * fmt_to_bpp(f->fmt.pix.pixelformat)) / 8;
        if (f->fmt.pix.bytesperline < bytesperline) {
                f->fmt.pix.bytesperline = bytesperline;
        }
        else
        {
                bytesperline = f->fmt.pix.bytesperline;
        }
        
        switch(f->fmt.pix.pixelformat)
        {
                case V4L2_PIX_FMT_YUV422P:
                        // byteperline for YUV planar formats is for Y plane only
                        size = bytesperline * f->fmt.pix.height * 2;
                        break;
                case V4L2_PIX_FMT_YUV420:
                        size = (bytesperline * f->fmt.pix.height * 3) / 2;
                        break;
                default:
                        size = bytesperline * f->fmt.pix.height;
                        break; 
        }        

        // Return the actual size of the image to the app
        f->fmt.pix.sizeimage = size;
                
        vout->v2f.fmt.pix.sizeimage = size;
        vout->v2f.fmt.pix.width = f->fmt.pix.width;
        vout->v2f.fmt.pix.height = f->fmt.pix.height;
        vout->v2f.fmt.pix.pixelformat = f->fmt.pix.pixelformat;
        vout->v2f.fmt.pix.bytesperline = f->fmt.pix.bytesperline;

        retval = 0;        
err0:        
        return retval;
}

/*
 * V4L2 - Handles VIDIOC_G_CTRL Ioctl
 *
 * @param vout         structure vout_data * 
 *
 * @param c           structure v4l2_control * 
 *
 * @return  status    0 success, EINVAL failed
 */
static int 
mxc_get_v42lout_control(vout_data *vout, struct v4l2_control *c)
{
        switch (c->id) 
        {
                case V4L2_CID_HFLIP:
                       	return vout->hflip;
                case V4L2_CID_VFLIP:
                       	return vout->vflip;
                case (V4L2_CID_PRIVATE_BASE+1):
                        return vout->rotate;
                default:
                        return -EINVAL;
        }
        return 0;
}

/*
 * V4L2 - Handles VIDIOC_S_CTRL Ioctl
 *
 * @param vout         structure vout_data * 
 *
 * @param c           structure v4l2_control * 
 *
 * @return  status    0 success, EINVAL failed
 */
static int 
mxc_set_v42lout_control(vout_data *vout, struct v4l2_control *c)
{
        switch (c->id) 
        {
                case V4L2_CID_HFLIP:
                        vout->hflip = (bool) c->value;
                        break;
                case V4L2_CID_VFLIP:
                        vout->vflip = (bool) c->value;
                        break;
                case V4L2_CID_MXC_ROTATE:
                        vout->rotate = c->value;
                        break;
                default:
                        return -EINVAL;
        }
        return 0;
}

/*!
 * V4L2 interface - open function
 *
 * @param inode        structure inode *
 *
 * @param file         structure file * 
 *
 * @return  status    0 success, ENODEV invalid device instance, 
 *                    ENODEV timeout, ERESTARTSYS interrupted by user 
 */
static int 
mxc_v4l2out_open(struct inode *inode, struct file *file)
{
        struct video_device *dev = video_devdata(file);
        vout_data *vout = video_get_drvdata(dev);
        int err;

        FUNC_START;

        if (!vout) 
        {
                DPRINTK("Internal error, vout_data not found!\n");
                return -ENODEV;
        }

        down(&vout->busy_lock);

        err = -EINTR;
        if(signal_pending(current))
                goto oops;

        if (vout->open_count++ == 0)
        {

                ipu_request_irq(IPU_IRQ_PP_IN_EOF, mxc_v4l2out_pp_in_irq_handler,
                                0, dev->name, vout);

//                ipu_request_irq(IPU_IRQ_SDC_FG_EOF, mxc_v4l2out_sdcfg_irq_handler,
//                                0, dev->name, vout);
                
                init_waitqueue_head(&vout->v4l_bufq);
                vout->state = STATE_STREAM_OFF;
                g_irq_cnt = g_buf_output_cnt = g_buf_q_cnt = g_buf_dq_cnt = 0;

        } 

        file->private_data = dev;

        up(&vout->busy_lock);

        FUNC_END;
        return 0;

oops:
        up(&vout->busy_lock);
        return err;
}

/*!
 * V4L2 interface - close function
 *
 * @param inode    struct inode * 
 *
 * @param file     struct file *
 *
 * @return         0 success 
 */
static int 
mxc_v4l2out_close(struct inode *inode, struct file *file)
{
        struct  video_device *dev = file->private_data;
        vout_data *vout = video_get_drvdata(dev);

        if (--vout->open_count == 0) 
        {
                DPRINTK("release resource\n");

                mxc_v4l2out_streamoff(vout);

                printk("g_irq_cnt = %d, g_buf_output_cnt = %d, g_buf_q_cnt = %d, g_buf_dq_cnt = %d\n",
                        g_irq_cnt, g_buf_output_cnt, g_buf_q_cnt, g_buf_dq_cnt);

                ipu_free_irq(IPU_IRQ_PP_IN_EOF, vout);
                ipu_free_irq(IPU_IRQ_SDC_FG_EOF, vout);

                file->private_data = NULL;

                mxc_free_buffers(vout->queue_buf_paddr, vout->buffer_cnt);
                vout->buffer_cnt = 0;
                mxc_free_buffers(vout->sdc_fg_bufs, 2);
                mxc_free_buffers(vout->rot_pp_bufs, 2);

                /* capture off */
                wake_up_interruptible(&vout->v4l_bufq);
//                gCondition++; 
        }

        //printk("state %x %x %x\n", vout->frame[0].state, vout->frame[1].state, vout->frame[2].state);  
        return 0;
}


/*!
 * V4L2 interface - ioctl function
 *
 * @param inode      struct inode * 
 *
 * @param file       struct file * 
 *
 * @param ioctlnr    unsigned int 
 *
 * @param arg        void * 
 *
 * @return           0 success, ENODEV for invalid device instance,
 *                   -1 for other errors. 
 */
static int 
mxc_v4l2out_do_ioctl(struct inode *inode, struct file *file, 
                 unsigned int ioctlnr, void *arg)
{
        struct video_device *dev = file->private_data;
        vout_data *vout = video_get_drvdata(dev);
        int retval = 0;
        int i = 0;

        if (!vout)
                return -ENODEV;

        /* make this _really_ smp-safe */
        if (down_interruptible(&vout->busy_lock))
                return -EINTR;

        switch (ioctlnr) 
        {
                /*!
                 * V4l2 VIDIOC_QUERYCAP ioctl 
                 */
                case VIDIOC_QUERYCAP:
                {        
                        struct v4l2_capability *cap = arg;
                        strcpy(cap->driver,"mxc_v4l2_output");
                        cap->version = 0;
                        cap->capabilities = V4L2_CAP_VIDEO_OUTPUT | V4L2_CAP_STREAMING;
                        retval = 0;
                        break;
                }

                case VIDIOC_G_FMT:
                {        
                        struct v4l2_format *gf = arg;
                        retval = mxc_v4l2out_g_fmt(vout, gf);
                        break;
                }

                case VIDIOC_S_FMT:
                {        
                        struct v4l2_format *sf = arg;
                        if (vout->state != STATE_STREAM_OFF) {
                                retval = -EBUSY;
                                break;
                        }
                        retval = mxc_v4l2out_s_fmt(vout, sf);
                        break;
                }

                /*!
                 * V4l2 VIDIOC_REQBUFS ioctl 
                 */
                case VIDIOC_REQBUFS: 
                {        
                        struct v4l2_requestbuffers *req = arg;
                        if ((req->type != V4L2_BUF_TYPE_VIDEO_OUTPUT) ||
                            (req->memory != V4L2_MEMORY_MMAP))
                        {
                                DPRINTK("VIDIOC_REQBUFS: incorrect buffer type\n");
                                retval = -EINVAL;
                                break;
                        }

                        if (req->count == 0) {
                                mxc_v4l2out_streamoff(vout);
                                if (vout->queue_buf_paddr[0] != 0) {
                                        mxc_free_buffers(vout->queue_buf_paddr, vout->buffer_cnt);
                                        DPRINTK("VIDIOC_REQBUFS: freed buffers\n");  
                                }
                                vout->buffer_cnt = 0;
                                break;
                        }

                        if (vout->queue_buf_paddr[0] != 0) {
                                DPRINTK("VIDIOC_REQBUFS: Cannot allocate buffers, free old buffers first\n");  
                                retval = -EBUSY;
                                break;
                        }

                        if (req->count < MIN_FRAME_NUM) {
                                req->count = MIN_FRAME_NUM;
                        } 
                        else if (req->count > MAX_FRAME_NUM) {
                                req->count = MAX_FRAME_NUM;
                        }
                        vout->buffer_cnt = req->count;
                        
                        retval = mxc_allocate_buffers(vout->queue_buf_paddr, 
                                vout->buffer_cnt, 
                                PAGE_ALIGN(vout->v2f.fmt.pix.sizeimage));
                        if (retval < 0)
                                break;

                        // Init buffer queues
                        vout->done_q.head = 0;
                        vout->done_q.tail = 0;
                        vout->ready_q.head = 0;
                        vout->ready_q.tail = 0;

                        for (i = 0; i < vout->buffer_cnt; i++) {
                                memset(&(vout->v4l2_bufs[i]), 0, sizeof(vout->v4l2_bufs[i]));
                                vout->v4l2_bufs[i].flags = 0;
                                vout->v4l2_bufs[i].memory = V4L2_MEMORY_MMAP;
                                vout->v4l2_bufs[i].index = i;
                                vout->v4l2_bufs[i].type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
                                vout->v4l2_bufs[i].length = PAGE_ALIGN(vout->v2f.fmt.pix.sizeimage);
                                vout->v4l2_bufs[i].m.offset = (unsigned long)vout->queue_buf_paddr[i];
                        }
                        break;
                }

                /*!
                 * V4l2 VIDIOC_QUERYBUF ioctl 
                 */
                case VIDIOC_QUERYBUF: 
                {        
                        struct v4l2_buffer *buf = arg;
                        u32 type = buf->type;
                        int index = buf->index;

                        if ((type != V4L2_BUF_TYPE_VIDEO_OUTPUT) || (index >= vout->buffer_cnt)) {
                                DPRINTK("VIDIOC_QUERYBUFS: incorrect buffer type\n");
                                retval = -EINVAL;
                                break;
                        }
                        down(&vout->param_lock);
                        memcpy(buf, &(vout->v4l2_bufs[index]), sizeof(*buf));
                        up(&vout->param_lock);
                        break;
                }
                        
                /*!
                 * V4l2 VIDIOC_QBUF ioctl 
                 */
                case VIDIOC_QBUF: 
                {        
                        struct v4l2_buffer *buf = arg;
                        int index = buf->index;
                        u32 lock_flags;

                        if ((buf->type != V4L2_BUF_TYPE_VIDEO_OUTPUT) || 
                            (index >= vout->buffer_cnt) ||
                            (buf->flags != 0)) {
                                retval = -EINVAL;
                                break;
                        }

                        spin_lock_irqsave(&g_lock, lock_flags);

                        DPRINTK("VIDIOC_QBUF: %d\n", buf->index);

                        memcpy(&(vout->v4l2_bufs[index]), buf, sizeof(*buf));
                        vout->v4l2_bufs[index].flags |= V4L2_BUF_FLAG_QUEUED;

                        g_buf_q_cnt++;
                        if (vout->state == STATE_STREAM_PAUSED) {
                                if (queue_size(&vout->ready_q) != 0)
                                        printk("VIDIOC_QBUF: IPU stopped and queue not empty!\n");
                                DPRINTK("VIDIOC_QBUF: restarting pp buf=%d, IPU_CUR_BUF=0x%08X\n", vout->next_rdy_ipu_buf, *((u32*)0xD42C0010));
                                if (ipu_update_channel_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, 
                                                          vout->next_rdy_ipu_buf, vout->queue_buf_paddr[index]) < 0) {
                                        DPRINTK("VIDIOC_QBUF: unable to update IPU buffer %d\n", vout->next_rdy_ipu_buf);
                                        retval = -EIO;
                                        spin_unlock_irqrestore(&g_lock, lock_flags);
                                        break;
                                }

                                g_buf_dq_cnt++;
                                
                                vout->ipu_buf[vout->next_rdy_ipu_buf] = index;
                                vout->state = STATE_STREAM_ON;
                                ipu_select_buffer(MEM_PP_MEM, IPU_INPUT_BUFFER, vout->next_rdy_ipu_buf);
                                vout->next_rdy_ipu_buf = !vout->next_rdy_ipu_buf;
                        }                        
                        else {
                                queue_buf(&vout->ready_q, index);
                        }

                        spin_unlock_irqrestore(&g_lock, lock_flags);
                        break;
                }

                /*!
                 * V4l2 VIDIOC_DQBUF ioctl 
                 */
                case VIDIOC_DQBUF: 
                {        
                        struct v4l2_buffer *buf = arg;
                        int idx;

                        //DPRINTK("VIDIOC_DQBUF: queue size = %d\n", queue_size(&vout->done_q));
                       
                        if ((queue_size(&vout->done_q) == 0) && (file->f_flags & O_NONBLOCK)) {
                                retval= -EAGAIN;
                                break;
                        }

                        if (!wait_event_interruptible_timeout(vout->v4l_bufq, queue_size(&vout->done_q) != 0, 2 * HZ))
                        {
                                printk("VIDIOC_DQBUF: timeout\n");
                                retval = -ETIME;
                                break;
                        }
                        else if (signal_pending(current))
                        {
                                printk("VIDIOC_DQBUF: interrupt received\n");
                                vout->state = STATE_STREAM_STOPPING;
                                //msleep_interruptible(70);
                                retval = -ERESTARTSYS;
                                break;
                        }
                        idx = dequeue_buf(&vout->done_q);
                        if (idx == -1) { // No frame free, 
                                printk("VIDIOC_DQBUF: ERROR: no free buffers, returning\n");
                                retval= -EAGAIN;
                                break;
                        }
                        if ((vout->v4l2_bufs[idx].flags & V4L2_BUF_FLAG_DONE) == 0)
                                printk("VIDIOC_DQBUF: buffer in done q, but not flagged as done\n");
                        
                        vout->v4l2_bufs[idx].flags = 0;
                        memcpy(buf, &(vout->v4l2_bufs[idx]), sizeof(*buf));
                        DPRINTK("VIDIOC_DQBUF: %d\n", buf->index);
                        break;
                }

                /*!
                 * V4l2 VIDIOC_STREAMON ioctl 
                 */
                case VIDIOC_STREAMON: 
                {        
                        if (vout->output == 0) { // SDC
                                retval = mxc_v4l2out_streamon_sdc(vout);
                        }
                        else {
                                retval = mxc_v4l2out_streamon_adc(vout);
                        }
                        break;
                }

                /*!
                 * V4l2 VIDIOC_STREAMOFF ioctl 
                 */
                case VIDIOC_STREAMOFF: 
                {        
                        retval = mxc_v4l2out_streamoff(vout);
                        break;
                }

                case VIDIOC_G_CTRL:
                {        
                        retval = mxc_get_v42lout_control(vout,arg);
                        break;
                }
                case VIDIOC_S_CTRL:
                {        
                        retval =  mxc_set_v42lout_control(vout,arg);
                        break;
                }
                case VIDIOC_CROPCAP:
                {
                        struct v4l2_cropcap *cap = arg;
                
                        if (cap->type != V4L2_BUF_TYPE_VIDEO_OUTPUT) {
                                retval = -EINVAL;
                                break;
                        }
                        cap->bounds  = vout->crop_bounds;
                        cap->defrect = vout->crop_defrect;
                        retval = 0;
                        break;
                }
                case VIDIOC_G_CROP:
                {
                        struct v4l2_crop * crop = arg;
                
                        if (crop->type != V4L2_BUF_TYPE_VIDEO_OUTPUT) {
                                retval = -EINVAL;
                                break;
                        }
                        crop->c = vout->crop_current;
                        break;
                }

                case VIDIOC_S_CROP:
                {
                        struct v4l2_crop *crop = arg;
                        struct v4l2_rect *b = &vout->crop_bounds;
                
                        if (crop->type != V4L2_BUF_TYPE_VIDEO_OUTPUT) {
                                retval = -EINVAL;
                                break;
                        }
                        if (crop->c.height < 0) {
                                retval = -EINVAL;
                                break;
                        }
                        if (crop->c.width < 0) {
                                retval = -EINVAL;
                                break;
                        }
                
                        if (crop->c.top < b->top)
                                crop->c.top = b->top;
                        if (crop->c.top > b->top + b->height)
                                crop->c.top = b->top + b->height;
                        if (crop->c.height > b->top - crop->c.top + b->height)
                                crop->c.height = b->top - crop->c.top + b->height;
                
                        if (crop->c.left < b->left)
                                crop->c.top = b->left;
                        if (crop->c.left > b->left + b->width)
                                crop->c.top = b->left + b->width;
                        if (crop->c.width > b->left - crop->c.left + b->width)
                                crop->c.width = b->left - crop->c.left + b->width;
                
                        vout->crop_current = crop->c;

                        vout->sdc_fg_buf_size = vout->crop_current.width * 
                                vout->crop_current.height;
                        vout->sdc_fg_buf_size *= fmt_to_bpp(SDC_FG_FB_FORMAT) / 8;

                        /* Free previously allocated buffer */
                        if (vout->sdc_fg_bufs[0] != NULL) {
                                mxc_free_buffers(vout->sdc_fg_bufs, 2);
                        }
                        if ((retval = mxc_allocate_buffers(vout->sdc_fg_bufs, 2, vout->sdc_fg_buf_size)) < 0) {
                                printk("mxc_v4l2out: unable to allocate SDC FG buffers\n");
                                return -ENOMEM;
                        }

                        ipu_sdc_set_window_pos(MEM_SDC_FG, 
                                               vout->crop_current.left,
                                               vout->crop_current.top);

                        break;
                }
                case VIDIOC_ENUMOUTPUT:
                {
                        struct v4l2_output * output = arg;

                        if (output->index >= MXC_V4L2_OUT_NUM_OUTPUTS) {
                                retval = -EINVAL;
                                break;
                        }

                        *output = mxc_outputs[output->index];

                        break;
                }
                case VIDIOC_G_OUTPUT:
                {
                        int * p_output_num = arg;

                        *p_output_num = vout->output;
                        break;
                }
                case VIDIOC_S_OUTPUT:
                {
                        int * p_output_num = arg;

                        if (*p_output_num >= MXC_V4L2_OUT_NUM_OUTPUTS) {
                                retval = -EINVAL;
                                break;
                        }

                        if (vout->state != STATE_STREAM_OFF) {
                                retval = -EBUSY;
                                break;
                        }

                        vout->output = *p_output_num;
                        break;
                }

                case VIDIOC_ENUM_FMT:
                case VIDIOC_TRY_FMT:
                case VIDIOC_QUERYCTRL:
                case VIDIOC_G_PARM:
                case VIDIOC_ENUMSTD:
                case VIDIOC_G_STD:
                case VIDIOC_S_STD:
                case VIDIOC_G_TUNER:
                case VIDIOC_S_TUNER:
                case VIDIOC_G_FREQUENCY:
                case VIDIOC_S_FREQUENCY:
                default:
                        retval = -EINVAL;
                        break;
                }
        
        up(&vout->busy_lock);
        return retval;
} 
        
/*
 * V4L2 interface - ioctl function
 *
 * @return  None
 */
static int 
mxc_v4l2out_ioctl(struct inode *inode, struct file *file, 
              unsigned int cmd, unsigned long arg)
{
        return video_usercopy(inode, file, cmd, arg, mxc_v4l2out_do_ioctl);
}
        
        
/*!
 * V4L2 interface - mmap function
 *
 * @param file        structure file *
 *
 * @param vma         structure vm_area_struct *
 *
 * @return status     0 Success, EINTR busy lock error, ENOBUFS remap_page error. 
 */
static int 
mxc_v4l2out_mmap(struct file *file, struct vm_area_struct *vma)
{
        struct video_device *dev = file->private_data;
        unsigned long start = vma->vm_start;
        unsigned long size  = vma->vm_end - vma->vm_start;
        int res = 0;
        vout_data *vout = video_get_drvdata(dev);

        /* make this _really_ smp-safe */
        if (down_interruptible(&vout->busy_lock))
                return -EINTR;
                
        if (remap_pfn_range(vma, start, vma->vm_pgoff, size, pgprot_noncached(PAGE_SHARED))) 
        {
                printk("mxc_mmap(V4L)i - remap_pfn_range failed\n");
                res = -ENOBUFS;
                goto mxc_mmap_exit;
        }

mxc_mmap_exit:
        up(&vout->busy_lock);
        return res;
}

/*!
 * V4L2 interface - poll function
 *
 * @param file       structure file *
 *
 * @param wait       structure poll_table * 
 *
 * @return  status   POLLIN | POLLRDNORM
 */
static unsigned int
mxc_v4l2out_poll(struct file *file, poll_table *wait)
{
        struct video_device *dev = file->private_data;
        vout_data *vout = video_get_drvdata(dev);

        wait_queue_head_t *queue = NULL;
        int res = POLLIN | POLLRDNORM;

        if (down_interruptible(&vout->busy_lock))
                return -EINTR;

        queue = &vout->v4l_bufq;
        poll_wait(file, queue, wait);

        up(&vout->busy_lock);
        return res;
}

static struct 
file_operations mxc_v4l2out_fops = {
        .owner      = THIS_MODULE,
        .open       = mxc_v4l2out_open,
        .release    = mxc_v4l2out_close,
        .ioctl      = mxc_v4l2out_ioctl,
        .mmap       = mxc_v4l2out_mmap,
        .poll       = mxc_v4l2out_poll,
};


static struct video_device mxc_v4l2out_template = {
        .owner      = THIS_MODULE,
        .name       = "MXC Video Output",
        .type       = 0,
	.type2      = V4L2_CAP_VIDEO_OUTPUT | V4L2_CAP_STREAMING,
        .hardware   = 39,
        .fops       = &mxc_v4l2out_fops,
        .release    = video_device_release,
};


/*!
 * Probe routine for the framebuffer driver. It is called during the
 * driver binding process.      The following functions are performed in
 * this routine: Framebuffer initialization, Memory allocation and
 * mapping, Framebuffer registration, IPU initialization.
 *
 * @return      Appropriate error code to the kernel common code
 */
static int mxc_v4l2out_probe(struct device *dev)
{
        vout_data * vout;

        FUNC_START;
        /*
         * Allocate sufficient memory for the fb structure
         */
        vout = kmalloc(sizeof(vout_data) + sizeof(u32) * 16, GFP_KERNEL);

        if (!vout)
                return 0;

        memset(vout, 0, sizeof(vout_data));

        vout->video_dev = video_device_alloc();
        if (vout->video_dev == NULL)
                return -1;
        vout->video_dev->dev = dev;
	vout->video_dev->minor   = -1;
        
        *(vout->video_dev) = mxc_v4l2out_template;

        /* register v4l device */
        if (video_register_device(vout->video_dev, VFL_TYPE_GRABBER, video_nr) == -1) 
        {
                printk(KERN_DEBUG "video_register_device failed\n");
                return 0;
        }
        printk(KERN_INFO "mxc_v4l2out: registered device video%d\n",
               vout->video_dev->minor & 0x1f);

        video_set_drvdata(vout->video_dev, vout);

        init_MUTEX(&vout->param_lock);
        init_MUTEX(&vout->busy_lock);

        /* setup cropping */
        vout->crop_bounds.left    = 0;
        vout->crop_bounds.top     = 0;
        vout->crop_bounds.width   = MXCFB_SCREEN_WIDTH;
        vout->crop_bounds.height  = MXCFB_SCREEN_HEIGHT;

        vout->crop_defrect = vout->crop_current = vout->crop_bounds;

        FUNC_END;
        return 0;
}

static int mxc_v4l2out_remove(struct device *dev)
{
        vout_data *vout = dev_get_drvdata(dev);

        FUNC_START;

        if (vout->video_dev) {
                if (-1 != vout->video_dev->minor)
                        video_unregister_device(vout->video_dev);
                else
                        video_device_release(vout->video_dev);
                vout->video_dev = NULL;
        }
       
        dev_set_drvdata(dev, NULL);

        kfree(vout);

        FUNC_END;
        return 0;
}

/*!
 * This structure contains pointers to the power management callback functions.
 */
static struct device_driver mxc_v4l2out_driver = {
        .name                   = "MXC Video Output",
        .bus                    = &platform_bus_type,
        .probe                  = mxc_v4l2out_probe,
        .remove                 = mxc_v4l2out_remove,
};

static struct platform_device mxc_v4l2out_device = {
        .name                   = "MXC Video Output",
        .id                     = 0,
};

/*!
 * mxc v4l2 init function
 *
 */
static int
mxc_v4l2out_init(void)
{
        u8 err = 0;
        FUNC_START;

        err = driver_register(&mxc_v4l2out_driver);
        if (err == 0) {
                platform_device_register(&mxc_v4l2out_device);
        }
        FUNC_END;
        return err;
}

/*!
 * mxc v4l2 cleanup function
 *
 */
static void mxc_v4l2out_clean(void)
{
        FUNC_START;

}

module_init(mxc_v4l2out_init);
module_exit(mxc_v4l2out_clean);


MODULE_PARM(video_nr,"i");
MODULE_AUTHOR("Freescale Semiconductor");
MODULE_DESCRIPTION("V4L2-driver for MXC video output");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE("video");
