/* 
 * Direct MTD block device access
 *
 * $Id: mtdblock.c,v 1.65 2004/11/16 18:28:59 dwmw2 Exp $
 *
 * (C) 2000-2003 Nicolas Pitre <nico@cam.org>
 * (C) 1999-2003 David Woodhouse <dwmw2@infradead.org>
 */

/*
 * Copyright 2006,2007 Sony Corporation. 
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/blktrans.h>
#include <linux/pagemap.h>

static struct mtdblk_dev {
	struct mtd_info *mtd;
	int count;
	struct semaphore cache_sem;
	unsigned char *cache_data;
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	loff_t cache_offset;
#else
	unsigned long cache_offset;
#endif
	unsigned int cache_size;
	enum { STATE_EMPTY, STATE_CLEAN, STATE_DIRTY } cache_state;
} *mtdblks[MAX_MTD_DEVICES];

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.23 yam */
static __inline int MTD_READ_NO_ERROR ( struct mtd_info *mtd, loff_t from, size_t len, 
                                        size_t *retlen, u_char *buf)
{
    int ret;

    ret = MTD_READ( mtd, from, len, retlen, buf );

    if ( ret || ( len != *retlen)  ) {
        if ( *retlen < 0 ) *retlen = 0;
        if ( *retlen < len ) {
            printk( KERN_WARNING "%s: read error, return dummy data. at %8.9llx size %08x\n", __FUNCTION__, from + *retlen , len - *retlen );
            memset( buf + *retlen , 0xff, len - *retlen );
        }
        *retlen = len;
    }

    return 0;
}
#endif
/*
 * Cache stuff...
 * 
 * Since typical flash erasable sectors are much larger than what Linux's
 * buffer cache can handle, we must implement read-modify-write on flash
 * sectors for each block write requests.  To avoid over-erasing flash sectors
 * and to speed things up, we locally cache a whole flash sector while it is
 * being written to until a different sector is required.
 */

static void erase_callback(struct erase_info *done)
{
	wait_queue_head_t *wait_q = (wait_queue_head_t *)done->priv;
	wake_up(wait_q);
}

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
static int erase_write (struct mtd_info *mtd, loff_t pos, 
			int len, const char *buf)
#else
static int erase_write (struct mtd_info *mtd, unsigned long pos, 
			int len, const char *buf)
#endif
{
	struct erase_info erase;
	DECLARE_WAITQUEUE(wait, current);
	wait_queue_head_t wait_q;
	size_t retlen;
	int ret;

	/*
	 * First, let's erase the flash block.
	 */

	init_waitqueue_head(&wait_q);
	erase.mtd = mtd;
	erase.callback = erase_callback;
	erase.addr = pos;
	erase.len = len;
	erase.priv = (u_long)&wait_q;

	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&wait_q, &wait);

	ret = MTD_ERASE(mtd, &erase);
	if (ret) {
		set_current_state(TASK_RUNNING);
		remove_wait_queue(&wait_q, &wait);
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.23 yam */
		printk (KERN_WARNING "mtdblock: erase of region [0x%8.9llx, 0x%x] "
				     "on \"%s\" failed\n",
			pos, len, mtd->name);
#else
		printk (KERN_WARNING "mtdblock: erase of region [0x%lx, 0x%x] "
				     "on \"%s\" failed\n",
			pos, len, mtd->name);
#endif
		return ret;
	}

	schedule();  /* Wait for erase to finish. */
	remove_wait_queue(&wait_q, &wait);

	/*
	 * Next, writhe data to flash.
	 */

	ret = MTD_WRITE (mtd, pos, len, &retlen, buf);
	if (ret)
		return ret;
	if (retlen != len)
		return -EIO;
	return 0;
}


static int write_cached_data (struct mtdblk_dev *mtdblk)
{
	struct mtd_info *mtd = mtdblk->mtd;
	int ret;

	if (mtdblk->cache_state != STATE_DIRTY)
		return 0;

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.23 yam */
	DEBUG(MTD_DEBUG_LEVEL2, "mtdblock: writing cached data for \"%s\" "
			"at 0x%8.9llx, size 0x%x\n", mtd->name, 
			mtdblk->cache_offset, mtdblk->cache_size);
#else
	DEBUG(MTD_DEBUG_LEVEL2, "mtdblock: writing cached data for \"%s\" "
			"at 0x%lx, size 0x%x\n", mtd->name, 
			mtdblk->cache_offset, mtdblk->cache_size);
#endif
	
	ret = erase_write (mtd, mtdblk->cache_offset, 
			   mtdblk->cache_size, mtdblk->cache_data);
	if (ret)
		return ret;

	/*
	 * Here we could argubly set the cache state to STATE_CLEAN.
	 * However this could lead to inconsistency since we will not 
	 * be notified if this content is altered on the flash by other 
	 * means.  Let's declare it empty and leave buffering tasks to
	 * the buffer cache instead.
	 */
	mtdblk->cache_state = STATE_EMPTY;
	return 0;
}


#ifdef CONFIG_SONY_ICX1087 /* 2006.09.06 yam */
static inline int _do_cached_write (struct mtdblk_dev *mtdblk, loff_t pos, 
			    int len, const char *buf)
#else
static int do_cached_write (struct mtdblk_dev *mtdblk, unsigned long pos, 
			    int len, const char *buf)
#endif
{
	struct mtd_info *mtd = mtdblk->mtd;
	unsigned int sect_size = mtdblk->cache_size;
	size_t retlen;
	int ret;

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.23 yam */
	DEBUG(MTD_DEBUG_LEVEL2, "mtdblock: write on \"%s\" at 0x%8.9llx, size 0x%x\n",
		mtd->name, pos, len);
#else
	DEBUG(MTD_DEBUG_LEVEL2, "mtdblock: write on \"%s\" at 0x%lx, size 0x%x\n",
		mtd->name, pos, len);
#endif
	
	if (!sect_size)
		return MTD_WRITE (mtd, pos, len, &retlen, buf);

	while (len > 0) {
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
		loff_t sect_start = pos & ~(loff_t)(sect_size-1);
		loff_t offset = pos - sect_start;
#else
		unsigned long sect_start = (pos/sect_size)*sect_size;
		unsigned int offset = pos - sect_start;
#endif
		unsigned int size = sect_size - offset;
		if( size > len ) 
			size = len;

		if (size == sect_size) {
			/* 
			 * We are covering a whole sector.  Thus there is no
			 * need to bother with the cache while it may still be
			 * useful for other partial writes.
			 */
#ifdef CONFIG_SONY_ICX1087 /* 2006.09.29 yam */
			if ( mtdblk->cache_state == STATE_DIRTY ) {
				if ( mtdblk->cache_offset == sect_start )
				  mtdblk->cache_state = STATE_EMPTY;
				else {
					ret = write_cached_data(mtdblk);
					if (ret) 
					  return ret;
				}
			}
#endif
			ret = erase_write (mtd, pos, size, buf);
			if (ret)
				return ret;
		} else {
			/* Partial sector: need to use the cache */

			if (mtdblk->cache_state == STATE_DIRTY &&
			    mtdblk->cache_offset != sect_start) {
				ret = write_cached_data(mtdblk);
				if (ret) 
					return ret;
			}

			if (mtdblk->cache_state == STATE_EMPTY ||
			    mtdblk->cache_offset != sect_start) {
				/* fill the cache with the current sector */
				mtdblk->cache_state = STATE_EMPTY;
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.23 yam */
				ret = MTD_READ_NO_ERROR(mtd, sect_start, sect_size, &retlen, mtdblk->cache_data);
#else
				ret = MTD_READ(mtd, sect_start, sect_size, &retlen, mtdblk->cache_data);
#endif
				if (ret)
					return ret;
				if (retlen != sect_size)
					return -EIO;

				mtdblk->cache_offset = sect_start;
				mtdblk->cache_size = sect_size;
				mtdblk->cache_state = STATE_CLEAN;
			}

			/* write data to our local cache */
			memcpy (mtdblk->cache_data + offset, buf, size);
			mtdblk->cache_state = STATE_DIRTY;
		}

		buf += size;
		pos += size;
		len -= size;
	}

	return 0;
}


#ifdef CONFIG_SONY_ICX1087 /* 2006.09.06 yam */
static inline int _do_cached_read (struct mtdblk_dev *mtdblk, loff_t pos, 
			   int len, char *buf)
#else
static int do_cached_read (struct mtdblk_dev *mtdblk, unsigned long pos, 
			   int len, char *buf)
#endif
{
	struct mtd_info *mtd = mtdblk->mtd;
	unsigned int sect_size = mtdblk->cache_size;
	size_t retlen;
	int ret;

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.23 yam */
	DEBUG(MTD_DEBUG_LEVEL2, "mtdblock: read on \"%s\" at 0x%8.9llx, size 0x%x\n", 
			mtd->name, pos, len);
#else
	DEBUG(MTD_DEBUG_LEVEL2, "mtdblock: read on \"%s\" at 0x%lx, size 0x%x\n", 
			mtd->name, pos, len);
#endif
	
	if (!sect_size)
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.23 yam */
		return MTD_READ_NO_ERROR (mtd, pos, len, &retlen, buf);
#else
		return MTD_READ (mtd, pos, len, &retlen, buf);
#endif

	while (len > 0) {
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
		loff_t sect_start = pos & ~(loff_t)(sect_size-1);
		loff_t offset = pos - sect_start;
#else
		unsigned long sect_start = (pos/sect_size)*sect_size;
		unsigned int offset = pos - sect_start;
#endif
		unsigned int size = sect_size - offset;
		if (size > len) 
			size = len;

		/*
		 * Check if the requested data is already cached
		 * Read the requested amount of data from our internal cache if it
		 * contains what we want, otherwise we read the data directly
		 * from flash.
		 */
		if (mtdblk->cache_state != STATE_EMPTY &&
		    mtdblk->cache_offset == sect_start) {
			memcpy (buf, mtdblk->cache_data + offset, size);
		} else {
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.23 yam */
			ret = MTD_READ_NO_ERROR (mtd, pos, size, &retlen, buf);
#else
			ret = MTD_READ (mtd, pos, size, &retlen, buf);
#endif
			if (ret)
				return ret;
			if (retlen != size)
				return -EIO;
		}

		buf += size;
		pos += size;
		len -= size;
	}

	return 0;
}


#ifdef CONFIG_SONY_ICX1087 /* 2006.09.06 yam */
static int do_cached_read (struct mtdblk_dev *mtdblk, loff_t pos, 
			   int len, char *buf)
{
	int rc;
	down(&mtdblk->cache_sem);
	rc = _do_cached_read( mtdblk, pos, len, buf );
	up(&mtdblk->cache_sem);
	return rc;
}


static int do_cached_write (struct mtdblk_dev *mtdblk, loff_t pos, 
				int len, const char *buf)
{
	int rc;
	down(&mtdblk->cache_sem);
	rc = _do_cached_write( mtdblk, pos, len, buf );
	up(&mtdblk->cache_sem);
	return rc;
}
#endif


static int mtdblock_readsect(struct mtd_blktrans_dev *dev,
			      unsigned long block, char *buf)
{
	struct mtdblk_dev *mtdblk = mtdblks[dev->devnum];
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	return do_cached_read(mtdblk, ((loff_t)block)<<9, 512, buf);
#else
	return do_cached_read(mtdblk, block<<9, 512, buf);
#endif
}

#ifdef CONFIG_SONY_ICX1087 /* 2006.10.13 yam */
static int mtdblock_readsectX(struct mtd_blktrans_dev *dev,
			      unsigned long block, unsigned long nsect, char *buf)
{
	struct mtdblk_dev *mtdblk = mtdblks[dev->devnum];
	unsigned long nsectX, mask, inc_block;

	if ( ( mtdblk->mtd->type == MTD_NANDFLASH ) && !( mtdblk->mtd->erasesize & ~0x00001800 ) && virt_addr_valid( buf ) ) {
		/* cache-less */

		if ( mtdblk->mtd->erasesize & 0x00001000 ) {
			/* 4096 */
			nsectX	  = nsect >> 3;
			mask	  = 0x07;
			inc_block = 8;
		} else {
			/* 2048 */
			nsectX	  = nsect >> 2;
			mask	  = 0x03;
			inc_block = 4;
		}

		if ( !( block & mask ) && !( (int)buf & 0x03 ) ) {
			for ( ; nsectX > 0; nsectX--, block+=inc_block, buf+=mtdblk->mtd->erasesize )
			  if ( do_cached_read( mtdblk, ((loff_t)block)<<9, mtdblk->mtd->erasesize, buf ) )
				return -EIO;
			nsect &= mask;
		}
	}

	for ( ; nsect > 0; nsect--, block++, buf += 512 )
	  if ( do_cached_read( mtdblk, ((loff_t)block)<<9, 512, buf ) )
		return -EIO;

	return 0;
}
#endif

static int mtdblock_writesect(struct mtd_blktrans_dev *dev,
			      unsigned long block, char *buf)
{
	struct mtdblk_dev *mtdblk = mtdblks[dev->devnum];
	if (unlikely(!mtdblk->cache_data)) {
#ifdef CONFIG_SONY_ICX1087 /* 2006.09.29 yam */
		if ( mtdblk->mtd->type == MTD_NANDFLASH ) {
			int order = ffs(mtdblk->mtd->erasesize) - 1 - PAGE_SHIFT;
			if ( order < 0 ) order = 0;
			mtdblk->cache_data = (unsigned char *)__get_free_pages(GFP_KERNEL, (unsigned int )order );
		} else {
			mtdblk->cache_data = vmalloc(mtdblk->mtd->erasesize);
		}
#else
#ifdef CONFIG_MP201
		mtdblk->cache_data = kmalloc(mtdblk->mtd->erasesize, GFP_KERNEL);
#else
		mtdblk->cache_data = vmalloc(mtdblk->mtd->erasesize);
#endif
#endif

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
        if (!mtdblk->cache_data) {
            printk(KERN_WARNING "%s: alloc cache_date failed!\n", __FUNCTION__ );
        }
#endif
		if (!mtdblk->cache_data)
			return -EINTR;
		/* -EINTR is not really correct, but it is the best match
		 * documented in man 2 write for all cases.  We could also
		 * return -EAGAIN sometimes, but why bother?
		 */
	}
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	return do_cached_write(mtdblk, ((loff_t)block)<<9, 512, buf);
#else
	return do_cached_write(mtdblk, block<<9, 512, buf);
#endif
}

#ifdef CONFIG_SONY_ICX1087 /* 2006.10.13 yam */
static int mtdblock_writesectX(struct mtd_blktrans_dev *dev,
			      unsigned long block, unsigned long nsect, char *buf)
{
	struct mtdblk_dev *mtdblk = mtdblks[dev->devnum];
	unsigned long nsectX, mask, inc_block;

	if (unlikely(!mtdblk->cache_data)) {
		if ( mtdblk->mtd->type == MTD_NANDFLASH ) {
			int order = ffs(mtdblk->mtd->erasesize) - 1 - PAGE_SHIFT;
			if ( order < 0 ) order = 0;
			mtdblk->cache_data = (unsigned char *)__get_free_pages(GFP_KERNEL, (unsigned int )order );
		} else {
			mtdblk->cache_data = vmalloc(mtdblk->mtd->erasesize);
		}

		if (!mtdblk->cache_data) {
			printk(KERN_WARNING "%s: alloc cache_date failed!\n", __FUNCTION__ );
			return -EINTR;
		}

		/* -EINTR is not really correct, but it is the best match
		 * documented in man 2 write for all cases.	 We could also
		 * return -EAGAIN sometimes, but why bother?
		 */
	}

	if ( ( mtdblk->mtd->type == MTD_NANDFLASH ) && !( mtdblk->mtd->erasesize & ~0x00001800 ) && virt_addr_valid( buf ) ) {
		/* cache-less */

		if ( mtdblk->mtd->erasesize & 0x00001000 ) {
			/* 4096 */
			nsectX	  = nsect >> 3;
			mask	  = 0x07;
			inc_block = 8;
		} else {
			/* 2048 */
			nsectX	  = nsect >> 2;
			mask	  = 0x03;
			inc_block = 4;
		}

		if ( !( block & mask ) && !( (int)buf & 0x03 ) ) {
			for ( ; nsectX > 0; nsectX--, block+=inc_block, buf+=mtdblk->mtd->erasesize )
			  if ( do_cached_write( mtdblk, ((loff_t)block)<<9, mtdblk->mtd->erasesize, buf ) )
				return -EIO;
			nsect &= mask;
		}
    }

	for ( ; nsect > 0; nsect--, block++, buf += 512 )
	  if ( do_cached_write( mtdblk, ((loff_t)block)<<9, 512, buf ) )
		return -EIO;

	return 0;
}
#endif

static int mtdblock_open(struct mtd_blktrans_dev *mbd)
{
	struct mtdblk_dev *mtdblk;
	struct mtd_info *mtd = mbd->mtd;
	int dev = mbd->devnum;

	DEBUG(MTD_DEBUG_LEVEL1,"mtdblock_open\n");
	
	if (mtdblks[dev]) {
		mtdblks[dev]->count++;
		return 0;
	}
	
	/* OK, it's not open. Create cache info for it */
	mtdblk = kmalloc(sizeof(struct mtdblk_dev), GFP_KERNEL);
	if (!mtdblk)
		return -ENOMEM;

	memset(mtdblk, 0, sizeof(*mtdblk));
	mtdblk->count = 1;
	mtdblk->mtd = mtd;

	init_MUTEX (&mtdblk->cache_sem);
	mtdblk->cache_state = STATE_EMPTY;
	if ((mtdblk->mtd->flags & MTD_CAP_RAM) != MTD_CAP_RAM &&
	    mtdblk->mtd->erasesize) {
		mtdblk->cache_size = mtdblk->mtd->erasesize;
		mtdblk->cache_data = NULL;
	}

	mtdblks[dev] = mtdblk;
	
	DEBUG(MTD_DEBUG_LEVEL1, "ok\n");

	return 0;
}

static int mtdblock_release(struct mtd_blktrans_dev *mbd)
{
	int dev = mbd->devnum;
	struct mtdblk_dev *mtdblk = mtdblks[dev];

   	DEBUG(MTD_DEBUG_LEVEL1, "mtdblock_release\n");

	down(&mtdblk->cache_sem);
	write_cached_data(mtdblk);
	up(&mtdblk->cache_sem);

	if (!--mtdblk->count) {
		/* It was the last usage. Free the device */
		mtdblks[dev] = NULL;
		if (mtdblk->mtd->sync)
			mtdblk->mtd->sync(mtdblk->mtd);
#ifdef CONFIG_SONY_ICX1087 /* 2006.09.29 yam */
		if ( mtdblk->mtd->type == MTD_NANDFLASH ) {
			int order = ffs(mtdblk->mtd->erasesize) - 1 - PAGE_SHIFT;
			if ( order < 0 ) order = 0;
			free_pages((unsigned long)mtdblk->cache_data, (unsigned int )order );
		} else {
			vfree(mtdblk->cache_data);
		}
#else
#ifdef CONFIG_MP201
		kfree(mtdblk->cache_data);
#else
		vfree(mtdblk->cache_data);
#endif
#endif
		kfree(mtdblk);
	}
	DEBUG(MTD_DEBUG_LEVEL1, "ok\n");

	return 0;
}  

static int mtdblock_flush(struct mtd_blktrans_dev *dev)
{
	struct mtdblk_dev *mtdblk = mtdblks[dev->devnum];

	down(&mtdblk->cache_sem);
	write_cached_data(mtdblk);
	up(&mtdblk->cache_sem);

	if (mtdblk->mtd->sync)
		mtdblk->mtd->sync(mtdblk->mtd);
	return 0;
}

static void mtdblock_add_mtd(struct mtd_blktrans_ops *tr, struct mtd_info *mtd)
{
	struct mtd_blktrans_dev *dev = kmalloc(sizeof(*dev), GFP_KERNEL);

	if (!dev)
		return;

	memset(dev, 0, sizeof(*dev));

	dev->mtd = mtd;
	dev->devnum = mtd->index;
	dev->blksize = 512;
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	dev->size = (unsigned long)(mtd->size >> 9);
#else
	dev->size = mtd->size >> 9;
#endif
	dev->tr = tr;

	if (!(mtd->flags & MTD_WRITEABLE))
		dev->readonly = 1;

	add_mtd_blktrans_dev(dev);
}

static void mtdblock_remove_dev(struct mtd_blktrans_dev *dev)
{
	del_mtd_blktrans_dev(dev);
	kfree(dev);
}

static struct mtd_blktrans_ops mtdblock_tr = {
	.name		= "mtdblock",
	.major		= 31,
	.part_bits	= 0,
	.open		= mtdblock_open,
	.flush		= mtdblock_flush,
	.release	= mtdblock_release,
	.readsect	= mtdblock_readsect,
	.writesect	= mtdblock_writesect,
#ifdef CONFIG_SONY_ICX1087 /* 2006.10.13 yam */
	.readsectX	= mtdblock_readsectX,
	.writesectX	= mtdblock_writesectX,
#endif
	.add_mtd	= mtdblock_add_mtd,
	.remove_dev	= mtdblock_remove_dev,
	.owner		= THIS_MODULE,
};

static int __init init_mtdblock(void)
{
	return register_mtd_blktrans(&mtdblock_tr);
}

static void __exit cleanup_mtdblock(void)
{
	deregister_mtd_blktrans(&mtdblock_tr);
}

module_init(init_mtdblock);
module_exit(cleanup_mtdblock);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Nicolas Pitre <nico@cam.org> et al.");
MODULE_DESCRIPTION("Caching read/erase/writeback block device emulation access to MTD devices");

void mtdblock_flush_all(void)
{
	int i;

	for (i = 0; i < MAX_MTD_DEVICES; i++) {
		struct mtdblk_dev *mtdblk = mtdblks[i];

		if (mtdblk) {
			down(&mtdblk->cache_sem);
			write_cached_data(mtdblk);
			up(&mtdblk->cache_sem);

			if (mtdblk->mtd->sync)
				mtdblk->mtd->sync(mtdblk->mtd);
		}
	}
}
