/*
 *
 *  Copyright 2006,2007 Sony Corporation. 
 *
 *  Based on drivers/mtd/nand_base.c
 *
 *  Overview:
 *   This is the generic MTD driver for NAND flash devices. It should be
 *   capable of working with almost all NAND chips currently available.
 *   Basic support for AG-AND chips is provided.
 *
 *	Additional technical information is available on
 *	http://www.linux-mtd.infradead.org/tech/nand.html
 *
 *  Copyright (C) 2000 Steven J. Hill (sjhill@realitydiluted.com)
 * 		  2002 Thomas Gleixner (tglx@linutronix.de)
 *
 *  02-08-2004  tglx: support for strange chips, which cannot auto increment
 *		pages on read / read_oob
 *
 *  03-17-2004  tglx: Check ready before auto increment check. Simon Bayes
 *		pointed this out, as he marked an auto increment capable chip
 *		as NOAUTOINCR in the board driver.
 *		Make reads over block boundaries work too
 *
 *  04-14-2004	tglx: first working version for 2k page size chips
 *
 *  05-19-2004  tglx: Basic support for Renesas AG-AND chips
 *
 *  09-24-2004  tglx: add support for hardware controllers (e.g. ECC) shared
 *		among multiple independend devices. Suggestions and initial patch
 *		from Ben Dooks <ben-mtd@fluff.org>
 *
 *  12-05-2004	dmarlin: add workaround for Renesas AG-AND chips "disturb" issue.
 *		Basically, any block not rewritten may lose data when surrounding blocks
 *		are rewritten many times.  JFFS2 ensures this doesn't happen for blocks
 *		it uses, but the Bad Block Table(s) may not be rewritten.  To ensure they
 *		do not lose data, force them to be rewritten when some of the surrounding
 *		blocks are erased.  Rather than tracking a specific nearby block (which
 *		could itself go bad), use a page address 'mask' to select several blocks
 *		in the same area, and rewrite the BBT when any of them are erased.
 *
 *  01-03-2005	dmarlin: added support for the device recovery command sequence for Renesas
 *		AG-AND chips.  If there was a sudden loss of power during an erase operation,
 * 		a "device recovery" operation must be performed when power is restored
 * 		to ensure correct operation.
 *
 *  01-20-2005	dmarlin: added support for optional hardware specific callback routine to
 *		perform extra error status checks on erase and write failures.  This required
 *		adding a wrapper function for nand_read_ecc.
 *
 *  04-27-2005  support for mp201 ((c) NEC Electronics Corporation )
 *
 *
 * Credits:
 *	David Woodhouse for adding multichip support
 *
 *	Aleph One Ltd. and Toby Churchill Ltd. for supporting the
 *	rework for 2K page size chips
 *
 * TODO:
 *	Enable cached programming for 2k page size chips
 *	Check, if mtd->ecctype should be set to MTD_ECC_HW
 *	if we have HW ecc support.
 *	The AG-AND chips have nice features for speed improvement,
 *	which are not supported yet. Read / program 4 pages in one go.
 *
 * $Id: nand_base.c,v 1.145 2005/05/31 20:32:53 gleixner Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/config.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/compatmac.h>
#include <linux/interrupt.h>
#include <linux/bitops.h>
#include <asm/io.h>
#include <asm/cacheflush.h>

#ifdef CONFIG_MTD_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#include "icx1087_nand_base.h"

/*
 * NAND low-level MTD interface functions
 */
static int  nand_read  (struct mtd_info *mtd, loff_t from, size_t len, size_t * retlen, u_char * buf);
static int  nand_write (struct mtd_info *mtd, loff_t to, size_t len, size_t * retlen, const u_char * buf);
static int  nand_erase (struct mtd_info *mtd, struct erase_info *instr);
static void nand_sync  (struct mtd_info *mtd);
static void nand_get_device     ( struct mtd_info *mtd, int new_state );
static void nand_release_device ( struct mtd_info *mtd );

/**
 * nand_get_device - [GENERIC] Get chip for selected access
 * @mtd:	MTD device structure
 * @new_state:	the state which is requested
 *
 * Get the device and lock it for exclusive access
 */
static void nand_get_device ( struct mtd_info *mtd, int new_state )
{
    struct icx1087_nand_dev *this = mtd->priv;
    struct nand_chip *active;
    spinlock_t *lock;
    wait_queue_head_t *wq;
    DECLARE_WAITQUEUE ( wait, current );

    lock = (this->controller) ? &this->controller->lock : &this->chip_lock;
    wq = (this->controller) ? &this->controller->wq : &this->wq;
retry:
    active = (struct nand_chip *)this;
    spin_lock( lock );

    /* Hardware controller shared among independend devices */
    if ( this->controller ) {
        if ( this->controller->active )
            active = this->controller->active;
        else
            this->controller->active = (struct nand_chip *)this;
    }
    if ( active == (struct nand_chip *)this && this->state == FL_READY ) {
        this->state = new_state;
        spin_unlock(lock);
        return;
    }
    set_current_state( TASK_UNINTERRUPTIBLE );
    add_wait_queue( wq, &wait );
    spin_unlock( lock );
    schedule();
    remove_wait_queue( wq, &wait );
    goto retry;
}

/**
 * nand_release_device - [GENERIC] release chip
 * @mtd:	MTD device structure
 *
 * Deselect, release chip lock and wake up anyone waiting on the device
 */
static void nand_release_device (struct mtd_info *mtd)
{
    struct icx1087_nand_dev *this = mtd->priv;

    /* De-select the NAND device */
    this->select_chip( mtd, -1 );

    if ( this->controller ) {
        /* Release the controller and the chip */
        spin_lock( &this->controller->lock );
        this->controller->active = NULL;
        this->state = FL_READY;
        wake_up( &this->controller->wq );
        spin_unlock( &this->controller->lock );
    } else {
        /* Release the chip */
        spin_lock( &this->chip_lock );
        this->state = FL_READY;
        wake_up( &this->wq );
        spin_unlock( &this->chip_lock );
    }
}

/**
 * nand_block_checkbad - [GENERIC] Check if a block is marked bad
 * @mtd:	MTD device structure
 * @ofs:	offset from device start
 * @getchip:	0, if the chip is already selected
 * @allowbbt:	1, if its allowed to access the bbt area
 *
 * Check, if the block is bad. Either by reading the bad block table or
 * calling of the scan function.
 */
static int nand_default_block_isbad ( struct mtd_info *mtd, loff_t ofs )
{
    struct icx1087_nand_dev *this = mtd->priv;
    int block;
    uint8_t	res;

    if ( !this->bbt )
      return 0; /*  */

    /* Return info from the table */

    /* Get block number */
    block = (int) ( ofs >> this->bbt_erase_shift );
    res = ( this->bbt[block >> 3] >> ( block & 0x07 ) ) & 0x01;

    DEBUG (MTD_DEBUG_LEVEL2, "nand_isbad_bbt(): bbt info for ofs 0x%08x: (block %d) 0x%02x\n", 
           (unsigned int)ofs, block, res );

    return res;
}

/**
 * nand_default_block_markbad - [DEFAULT] mark a block bad
 * @mtd:	MTD device structure
 * @ofs:	offset from device start
 *
 * This is the default implementation, which can be overridden by
 * a hardware specific driver.
*/
static int nand_default_block_markbad(struct mtd_info *mtd, loff_t ofs)
{
    struct icx1087_nand_dev *this = mtd->priv;
    int block;

    printk("%s\n", __FUNCTION__);

    if ( !this->bbt ) 
      return 0;

    /* Get block number */

    block = (int) ( ofs >> this->bbt_erase_shift );
    this->bbt[block >> 3] |= 0x01 << ( block & 0x03 );

    return 0;
}

/**
 * nand_scan_bbt - [NAND Interface] scan, find, read and maybe create bad block table(s)
 * @mtd:	MTD device structure
 * @bd:		descriptor for the good/bad block search pattern
 *
 * The function checks, if a bad block table(s) is/are already 
 * available. If not it scans the device for manufacturer
 * marked good / bad blocks and writes the bad block table(s) to
 * the selected place.
 *
 * The bad block table memory is allocated here. It must be freed
 * by calling the nand_free_bbt function.
 *
*/
int nand_scan_default_bbt (struct mtd_info *mtd )
{
    struct icx1087_nand_dev *this = mtd->priv;
    int i, numblocks, len;
    int startblock;
    uint8_t *buf = this->data_buf;
    loff_t from, from2;
    size_t retlen;
    int rc;

    len = ( this->total_size + 7 ) >> (this->bbt_erase_shift + 3);

    /* Allocate memory (1bit per block) */
    this->bbt = (uint8_t *) kmalloc ( len, GFP_KERNEL );
    if ( !this->bbt ) {
        printk (KERN_ERR "nand_scan_bbt: Out of memory\n");
        return -ENOMEM;
    }

    /* Clear the memory bad block table */
    memset ( this->bbt, 0x00, len );

    numblocks  = this->total_size >> ( this->bbt_erase_shift );
    startblock = 0;
    from       = 0;

    printk (KERN_INFO "Scanning device for bad blocks(%d)\n", numblocks);

    for ( i = startblock; i < numblocks; i++, from += (loff_t)this->block_size ) {

        rc = nand_read( mtd, from, this->page_size, &retlen, buf );

        if ( !rc && ( buf[0] == 0xff ) && ( buf[2048-16*3] == 0xff ) ) {

            from2 = from + (loff_t)( this->block_size - this->page_size );
            rc = nand_read( mtd, from2, this->page_size, &retlen, buf );

            if ( !rc && ( buf[0] == 0xff ) && ( buf[2048-16*3] == 0xff ) ) continue;
        }

        this->bbt[i >> 3] |= 0x01 << ( i & 0x07 );
        printk ( KERN_WARNING "Bad eraseblock %d at 0x%08x\n", i, (unsigned int) from );

    }
    return 0;
}


/**
 * nand_check_wp - [GENERIC] check if the chip is write protected
 * @mtd:	MTD device structure
 * Check, if the device is write protected
 *
 * The function expects, that the device is already selected
 */
static int nand_check_wp ( struct mtd_info *mtd )
{
    struct icx1087_nand_dev *this = mtd->priv;
    /* Check the WP bit */
    return ( this->read_status( mtd ) & NAND_STATUS_WP ) ? 0 : 1;
}

/**
 * nand_wait - [DEFAULT]  wait until the command is done
 * @mtd:	MTD device structure
 * @this:	NAND chip structure
 * @state:	state to select the max. timeout value
 *
 * Wait for command done. This applies to erase and program only
 * Erase can take up to 400ms and program up to 20ms according to
 * general NAND and SmartMedia specs
 *
*/
static int nand_wait(struct mtd_info *mtd, int state)
{
    struct icx1087_nand_dev *this = mtd->priv;
    unsigned long   timeo = jiffies;
    int status;

    if (state == FL_ERASING)
      timeo += (HZ * 400) / 1000;
    else
      timeo += (HZ * 20) / 1000;

    /* Apply this short delay always to ensure that we do wait tWB in
     * any case on any machine. */
    ndelay (100);

    status = this->read_status( mtd );

    while (time_before(jiffies, timeo)) {
        /* Check, if we were interrupted */
        if (this->state != state)
          return 0;
        if (this->dev_ready) {
            if (this->dev_ready(mtd))
              break;
        }
        if ( this->read_status( mtd ) & NAND_STATUS_READY )
          break;
        cond_resched();
    }
    status = (int) this->read_status( mtd );
 
    return status;
}

#if 0
/*
 * Wait for the ready pin, after a command
 * The timeout is catched later.
 */
static void nand_wait_ready( struct mtd_info *mtd )
{
    struct icx1087_nand_dev *this = mtd->priv;
    unsigned long   timeo = jiffies + 2;

    /* wait until command is processed or timeout occures */
    do {
        if ( this->dev_ready( mtd ) )
          return;
    } while ( time_before( jiffies, timeo ) );
}
#endif

/**
 * nand_read - [MTD Interface] MTD compability function for nand_read_ecc
 * @mtd:	MTD device structure
 * @from:	offset to read from
 * @len:	number of bytes to read
 * @retlen:	pointer to variable to store the number of read bytes
 * @buf:	the databuffer to put data
 *
 * This function simply calls nand_read_ecc with oob buffer and oobsel = NULL
*/
static int nand_read (struct mtd_info *mtd, loff_t from, size_t len, size_t * retlen, u_char * buf)
{
    int j, col, realpage, page, end, chipnr, read = 0, rc = 0;
    struct icx1087_nand_dev *this = mtd->priv;
    u_char *data_poi;

    DEBUG ( MTD_DEBUG_LEVEL3, "%s: from = 0x%08x, len = %i\n", __FUNCTION__, (unsigned int) from, (int) len );

    /* Do not allow reads past end of device */
    if ( ( from + len ) > this->total_size ) {
        DEBUG ( MTD_DEBUG_LEVEL0, "%s: Attempt read beyond end of device\n", __FUNCTION__ );
        *retlen = 0;
        return -EINVAL;
    }

    /* Grab the lock and see if the device is available */
    this->get_device ( mtd , FL_READING );

    /* Select the NAND device */
    chipnr = (int)( from >> this->chip_shift );
    this->select_chip( mtd, chipnr );

    /* First we calculate the starting page */
    realpage = (int) ( from >> this->page_shift );
    page = realpage & this->pagemask;

    /* Get raw starting column */
    col = from & ( this->page_size - 1 );

    end = this->page_size;

    /* Loop until all data read */
    while ( read < len ) {

        int aligned = ( !col && ( len - read ) >= end );

        /*
         * If the read is not page aligned, we have to read into data buffer
         * due to ecc, else we read into return buffer direct
         */
        if ( aligned ) {
            data_poi = &buf[read];
        } else {
            data_poi = this->data_buf;
        }

        /* Check, if we have this page in the buffer
         *
         * FIXME: Make it work when we must provide oob data too,
         * check the usage of data_buf oob field
         */
        if ( realpage == this->pagebuf ) {
            /* aligned read ? */
            if ( aligned )
                memcpy ( data_poi, this->data_buf, end );
        } else {
            rc = this->read_page( mtd, page, data_poi, NULL );
            if ( rc ) break;
        }

        /* Partial page read, transfer data into fs buffer */
        if ( !aligned ) {
#if 1 /* Speed Up: 2006.08.31 */
            if ( end - col < len - read )
              j = end - col;
            else
              j = len - read;
            memcpy( &buf[read], &data_poi[col], j );
            read += j;
#else
            for ( j = col; j < end && read < len; j++ )
              buf[read++] = data_poi[j];
#endif
            this->pagebuf = realpage;
        } else
          read += this->page_size;

        if ( read == len )
          break;

        /* For subsequent reads align to page boundary. */
        col = 0;
        /* Increment page address */
        realpage++;

        page = realpage & this->pagemask;
        /* Check, if we cross a chip boundary */
        if (!page) {
            chipnr++;
            this->select_chip(mtd, -1);
            this->select_chip(mtd, chipnr);
        }
    }

    /* Deselect and wake up anyone waiting on the device */
    this->release_device(mtd);

    *retlen = read;

    if ( rc ) {
        /* Invalidate the pagebuffer reference */
        this->pagebuf = -1;
        return -EBADMSG;
    }
    return 0;
}



#define NOTALIGNED(x) (x & (this->page_size-1)) != 0

/**
 * nand_write - [MTD Interface] compability function for nand_write_ecc
 * @mtd:	MTD device structure
 * @to:		offset to write to
 * @len:	number of bytes to write
 * @retlen:	pointer to variable to store the number of written bytes
 * @buf:	the data to write
 *
 * This function simply calls nand_write_ecc with oob buffer and oobsel = NULL
 *
*/
static int nand_write (struct mtd_info *mtd, loff_t to, size_t len, size_t * retlen, const u_char * buf)
{
    int startpage, page, ret = -EIO, written = 0, chipnr, totalpages;
    struct icx1087_nand_dev *this = mtd->priv;
    int ppblock = (1 << (this->block_shift - this->page_shift));
    u_char *bufstart;

    DEBUG ( MTD_DEBUG_LEVEL3, "%s: to = 0x%08x, len = %i\n", __FUNCTION__, (unsigned int) to, (int) len );

    /* Initialize retlen, in case of early exit */
    *retlen = 0;

    /* Do not allow write past end of device */
    if ( (to + len) > this->total_size ) {
        DEBUG (MTD_DEBUG_LEVEL0, "%s: Attempt to write past end of page\n", __FUNCTION__ );
        return -EINVAL;
    }

    /* reject writes, which are not page aligned */
    if ( NOTALIGNED (to) || NOTALIGNED(len) ) {
        printk ( KERN_NOTICE "%s: Attempt to write not page aligned data\n", __FUNCTION__ );
        return -EINVAL;
    }

    /* Grab the lock and see if the device is available */
    this->get_device ( mtd, FL_WRITING );

    /* Calculate chipnr */
    chipnr = (int)( to >> this->chip_shift );
    /* Select the NAND device */
    this->select_chip( mtd, chipnr );

    /* Check, if it is write protected */
    if (nand_check_wp(mtd))
      goto nand_write_out;

    /* Setup variables and oob buffer */
    totalpages = len >> this->page_shift;
    page = (int) ( to >> this->page_shift );
    /* Invalidate the page cache, if we write to the cached page */
    if ( page <= this->pagebuf && this->pagebuf < (page + totalpages) )
      this->pagebuf = -1;

    /* Set it relative to chip */
    page &= this->pagemask;
    startpage = page;
    bufstart = (u_char *)buf;

    /* Loop until all data is written */
    while ( written < len ) {

        this->data_poi = (u_char*) &buf[written];

        /* Write one page. If this is the last page to write
         * or the last page in this block, then use the
         * real pageprogram command, else select cached programming
         * if supported by the chip.
         */
        ret = this->write_page ( mtd, page, this->data_poi, NULL );
        /* See if device thinks it succeeded */
        if ( ret ) {
            DEBUG ( MTD_DEBUG_LEVEL0, "%s: " "Failed write, page 0x%08x, ", __FUNCTION__, page );
            ret = -EIO;
            goto nand_write_out;
        }

        /* Update written bytes count */
        written += this->page_size;
        if (written == len)
          goto nand_write_cmp;

        /* Increment page address */
        page++;

        /* Have we hit a block boundary ? Then we have to verify and
         * if verify is ok, we have to setup the oob buffer for
         * the next pages.
         */
        if ( !(page & (ppblock - 1)) ){
            this->data_poi = bufstart;
            *retlen     = written;
            totalpages -= page - startpage;
            page       &= this->pagemask;
            startpage   = page;

            /* Check, if we cross a chip boundary */
            if ( !page ) {
                chipnr++;
                this->select_chip( mtd, -1 );
                this->select_chip( mtd, chipnr );
            }
        }
    }

  nand_write_cmp:
    *retlen = written;

  nand_write_out:
    /* Deselect and wake up anyone waiting on the device */
    this->release_device( mtd );

    return ret;
}

/**
 * nand_erase - [MTD Interface] erase block(s)
 * @mtd:	MTD device structure
 * @instr:	erase instruction
 *
 * Erase one ore more blocks
 */
static int nand_erase ( struct mtd_info *mtd, struct erase_info *instr )
{
    return nand_erase_nand ( mtd, instr, 0 );
}

/**
 * nand_erase_intern - [NAND Interface] erase block(s)
 * @mtd:        MTD device structure
 * @instr:      erase instruction
 * @allowbbt:   allow erasing the bbt area
 *
 * Erase one ore more blocks
 */
int nand_erase_nand (struct mtd_info *mtd, struct erase_info *instr, int allowbbt)
{
    int page, len, pages_per_block, ret, chipnr;
    struct icx1087_nand_dev *this = mtd->priv;

    DEBUG ( MTD_DEBUG_LEVEL3,
            "nand_erase: start = 0x%08x, len = %i\n", (unsigned int) instr->addr, (unsigned int) instr->len );

    /* Start address must align on block boundary */
    if ( instr->addr & ((1 << this->phys_erase_shift) - 1) ) {
        DEBUG ( MTD_DEBUG_LEVEL0, "nand_erase: Unaligned address\n" );
        return -EINVAL;
    }

    /* Length must align on block boundary */
    if ( instr->len & ((1 << this->phys_erase_shift) - 1) ) {
        DEBUG ( MTD_DEBUG_LEVEL0, "nand_erase: Length not block aligned\n" );
        return -EINVAL;
    }

    /* Do not allow erase past end of device */
    if ((instr->len + instr->addr) > this->total_size) {
        DEBUG ( MTD_DEBUG_LEVEL0, "nand_erase: Erase past end of device\n" );
        return -EINVAL;
    }

    instr->fail_addr = 0xffffffff;

    /* Grab the lock and see if the device is available */
    this->get_device ( mtd, FL_ERASING );

    /* Shift to get first page */
    page = (int) ( instr->addr >> this->page_shift );
    chipnr = (int) ( instr->addr >> this->chip_shift );

    /* Calculate pages in each block */
    pages_per_block = 1 << ( this->phys_erase_shift - this->page_shift );

    /* Select the NAND device */
    this->select_chip( mtd, chipnr );

    /* Check the WP bit */
    /* Check, if it is write protected */
    if ( nand_check_wp( mtd ) ) {
        DEBUG ( MTD_DEBUG_LEVEL0, "nand_erase: Device is write protected!!!\n" );
        instr->state = MTD_ERASE_FAILED;
        goto erase_exit;
    }

    /* Loop through the pages */
    len = instr->len;

    instr->state = MTD_ERASING;

    while (len) {
        /* Check if we have a bad block, we do not erase bad blocks ! */
        if ( this->block_isbad( mtd, ((loff_t) page) << this->page_shift) ) {
            printk ( KERN_WARNING "nand_erase: attempt to erase a bad block at page 0x%08x\n", page );
            instr->state = MTD_ERASE_FAILED;
            goto erase_exit;
        }

        /* Invalidate the page cache, if we erase the block which contains
           the current cached page */
        if ( page <= this->pagebuf && this->pagebuf < (page + pages_per_block) )
          this->pagebuf = -1;

        ret = this->erase_block ( mtd, page & this->pagemask );
        /* See if block erase succeeded */
        if ( ret ) {
            DEBUG ( MTD_DEBUG_LEVEL0, "nand_erase: " "Failed erase, page 0x%08x\n", page );
            instr->state = MTD_ERASE_FAILED;
            instr->fail_addr = ( (loff_t)page << this->page_shift );
            goto erase_exit;
        }

        /* Increment page address and decrement length */
        len -= ( 1 << this->phys_erase_shift );
        page += pages_per_block;

        /* Check, if we cross a chip boundary */
        if ( len && !(page & this->pagemask) ) {
            chipnr++;
            this->select_chip(mtd, -1);
            this->select_chip(mtd, chipnr);
        }
    }
    instr->state = MTD_ERASE_DONE;

  erase_exit:

    ret = instr->state == MTD_ERASE_DONE ? 0 : -EIO;
    /* Do call back function */
    if ( !ret )
      mtd_erase_callback( instr );

    /* Deselect and wake up anyone waiting on the device */
    this->release_device( mtd );

    /* Return more or less happy */
    return ret;
}

/**
 * nand_sync - [MTD Interface] sync
 * @mtd:	MTD device structure
 *
 * Sync is actually a wait for chip ready function
 */
static void nand_sync ( struct mtd_info *mtd )
{
    struct icx1087_nand_dev *this = mtd->priv;

    DEBUG (MTD_DEBUG_LEVEL3, "nand_sync: called\n");

    /* Grab the lock and see if the device is available */
    this->get_device ( mtd, FL_SYNCING );
    /* Release it and go back */
    this->release_device ( mtd );
}


/**
 * nand_block_isbad - [MTD Interface] Check whether the block at the given offset is bad
 * @mtd:	MTD device structure
 * @ofs:	offset relative to mtd start
 */
static int nand_block_isbad ( struct mtd_info *mtd, loff_t ofs )
{
    struct icx1087_nand_dev *this = mtd->priv;

    /* Check for invalid offset */
    if ( ofs > mtd->size )
      return -EINVAL;

    return this->block_isbad( mtd, ofs );
}

/**
 * nand_block_markbad - [MTD Interface] Mark the block at the given offset as bad
 * @mtd:	MTD device structure
 * @ofs:	offset relative to mtd start
 */
static int nand_block_markbad ( struct mtd_info *mtd, loff_t ofs )
{
    struct icx1087_nand_dev *this = mtd->priv;

    if ( nand_block_isbad(mtd, ofs) ) {
        /* If it was bad already, return success and do nothing. */
        return 0;
    }

    return this->block_markbad( mtd, ofs );
}

/**
 * nand_scan - [NAND Interface] Scan for the NAND device
 * @mtd:	MTD device structure
 * @maxchips:	Number of chips to scan for
 *
 * This fills out all the not initialized function pointers
 * with the defaults.
 * The flash ID is read and the mtd/chip structures are
 * filled with the appropriate values. Buffers are allocated if
 * they are not provided by the board driver
 *
 */
int icx1087_nand_scan (struct mtd_info *mtd, int maxchips)
{
    int i, maf_id;
    struct icx1087_nand_dev *this = mtd->priv;
    u_char id[8];

    /* check for proper chip_delay setup, set 20us if not */
    if (!this->chip_delay)
      this->chip_delay = 20;

    if ( !this->select_chip ||
         !this->read_id     ||
         !this->read_page   ||
         !this->write_page  ||
         !this->erase_block ||
         0 ) {
        return -EINVAL;
    }

    if ( !this->waitfunc )       this->waitfunc       = nand_wait;
    if ( !this->block_isbad )    this->block_isbad    = nand_default_block_isbad;
    if ( !this->block_markbad )  this->block_markbad  = nand_default_block_markbad;
    if ( !this->get_device)      this->get_device     = nand_get_device;
    if ( !this->release_device ) this->release_device = nand_release_device;

    /* Select the device */
    this->select_chip( mtd, 0 );

    /* Send the command for reading device ID */
    this->read_id( mtd, id, 4 );

    /* Read manufacturer and device IDs */
    this->maf_id = id[0];
    this->dev_id = id[1];

    /* Print and store flash device information */
    for ( i = 0; nand_flash_ids[i].name != NULL; i++ ) {

        if ( this->dev_id != nand_flash_ids[i].id )
          continue;

        if ( !this->name ) this->name = nand_flash_ids[i].name;
        this->chip_size = ((loff_t)nand_flash_ids[i].chipsize) << 20;

        /* New devices have all the information in additional id bytes */
        if (!nand_flash_ids[i].pagesize) {
            int extid = id[3]; /* The 4th id byte */
            /* Calc pagesize */
            this->page_size = 1024 << (extid & 0x3);
            extid >>= 2;
            /* Calc oobsize */
            this->oob_size = (8 << (extid & 0x03)) * (this->page_size / 512);
            extid >>= 2;
            /* Calc blocksize. Blocksize is multiples of 64KiB */
            this->block_size = this->erase_size = (64 * 1024)  << (extid & 0x03);
            extid >>= 2;
        } else {
            /* Old devices have this data hardcoded in the
             * device id table */
            this->block_size = this->erase_size = nand_flash_ids[i].erasesize;
            this->page_size  = nand_flash_ids[i].pagesize;
            this->oob_size   = this->page_size / 32;
        }

        /* Try to identify manufacturer */
        for ( maf_id = 0; nand_manuf_ids[maf_id].id != 0x0; maf_id++ ) {
            if ( nand_manuf_ids[maf_id].id == this->maf_id )
              break;
        }

        /* Calculate the address shift from the page size */
        this->page_shift  = ffs(this->page_size) - 1;
        this->block_shift = this->bbt_erase_shift = this->phys_erase_shift = ffs(this->block_size) - 1;
        this->chip_shift  = ffs(this->chip_size) - 1;

        printk (KERN_INFO "NAND device: Manufacturer ID:"
                " 0x%02x, Chip ID: 0x%02x (%s %s)\n", this->maf_id, this->dev_id,
                nand_manuf_ids[maf_id].name , nand_flash_ids[i].name);

        break;
    }

    if ( !nand_flash_ids[i].name ) {
        printk (KERN_WARNING "No NAND device found!!!\n");
        this->select_chip( mtd, -1 );
        return 1;
    }

    for ( i=1; i < maxchips; i++ ) {
        this->select_chip( mtd, i );
        /* Send the command for reading device ID */
        this->read_id( mtd, id, 2 );
        /* Read manufacturer and device IDs */
        if ( this->maf_id != id[0] || this->dev_id != id[1] )
          break;
    }
    if (i > 1)
      printk(KERN_INFO "%d NAND chips detected\n", i);


    /* Allocate buffers, if neccecary */
    if (!this->data_buf) {
        size_t len;
        len = this->page_size + this->oob_size;
        this->data_buf = kmalloc (len, GFP_KERNEL );
        if ( !this->data_buf ) {
            printk (KERN_ERR "nand_scan(): Cannot allocate data_buf\n");
            return -ENOMEM;
        }
        this->options |= NAND_DATABUF_ALLOC;
    }

    /* Store the number of chips and calc total size for mtd */
    this->chip_num = i;
    this->total_size = mtd->size = i * this->chip_size;
    /* Convert chipsize to number of pages per chip -1. */
    this->pagemask = ( this->chip_size >> this->page_shift ) - 1;


    /* Initialize state, waitqueue and spinlock */
    this->state = FL_READY;
    init_waitqueue_head ( &this->wq );
    spin_lock_init ( &this->chip_lock );
    if ( this->controller ) {
        init_waitqueue_head ( &this->controller->wq );
        spin_lock_init ( &this->controller->lock );
    }


    /* De-select the device */
    this->select_chip( mtd, -1 );

    /* Invalidate the pagebuffer reference */
    this->pagebuf = -1;

    return 0;
}

int icx1087_nand_register( struct mtd_info *mtd, const struct mtd_partition *parts, int nbparts )
{
    struct icx1087_nand_dev *this = mtd->priv;

    /* Fill in remaining MTD driver data */
    mtd->type    = MTD_NANDFLASH;
    mtd->flags   = MTD_CAP_NANDFLASH | MTD_ECC;
    mtd->erase   = nand_erase;
    mtd->point   = NULL;
    mtd->unpoint = NULL;
    mtd->read    = nand_read;
    mtd->write   = nand_write;
    mtd->sync    = nand_sync;
    mtd->lock    = NULL;
    mtd->unlock  = NULL;
    mtd->suspend = NULL;
    mtd->resume  = NULL;
    mtd->block_isbad   = nand_block_isbad;
    mtd->block_markbad = nand_block_markbad;

    mtd->oobblock  = this->page_size;
    mtd->oobsize   = this->oob_size;
    mtd->oobavail  = this->oob_avail;
    mtd->erasesize = this->erase_size;
    mtd->size      = this->total_size;
    mtd->name      = this->name;

    mtd->owner = THIS_MODULE;

    return add_mtd_partitions( mtd, parts, nbparts );
}


/**
 * nand_release - [NAND Interface] Free resources held by the NAND device
 * @mtd:	MTD device structure
*/
void icx1087_nand_release (struct mtd_info *mtd)
{
    struct icx1087_nand_dev *this = mtd->priv;

    /* Deregister partitions */
    del_mtd_partitions (mtd);
    /* Deregister the device */
    del_mtd_device (mtd);

    /* Free bad block table memory, if allocated */
    if (this->bbt)
        kfree (this->bbt);
    /* Buffer allocated by nand_scan ? */
    if (this->options & NAND_OOBBUF_ALLOC)
        kfree (this->oob_buf);
    /* Buffer allocated by nand_scan ? */
    if (this->options & NAND_DATABUF_ALLOC)
        kfree (this->data_buf);
}


EXPORT_SYMBOL_GPL (icx1087_nand_scan);
EXPORT_SYMBOL_GPL (icx1087_nand_release );

MODULE_LICENSE ("GPL");
MODULE_AUTHOR ("Steven J. Hill <sjhill@realitydiluted.com>, Thomas Gleixner <tglx@linutronix.de>");
MODULE_DESCRIPTION ("Generic NAND flash driver code");


EXPORT_SYMBOL (icx1087_nand_register );
