/*
 *
 *  Copyright 2006,2007 Sony Corporation. 
 *
 *  Based on linux/include/linux/mtd/nand.h
 *
 *  Copyright (c) 2000 David Woodhouse <dwmw2@mvhi.com>
 *                     Steven J. Hill <sjhill@realitydiluted.com>
 *                     Thomas Gleixner <tglx@linutronix.de>
 *
 * $Id: nand.h,v 1.63 2004/07/07 16:29:43 gleixner Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Info:
 *   Contains standard defines and IDs for NAND flash devices
 *
 *  Changelog:
 *   01-31-2000 DMW     Created
 *   09-18-2000 SJH     Moved structure out of the Disk-On-Chip drivers
 *                      so it can be used by other NAND flash device
 *                      drivers. I also changed the copyright since none
 *                      of the original contents of this file are specific
 *                      to DoC devices. David can whack me with a baseball
 *                      bat later if I did something naughty.
 *   10-11-2000 SJH     Added private NAND flash structure for driver
 *   10-24-2000 SJH     Added prototype for 'nand_scan' function
 *   10-29-2001 TG      changed nand_chip structure to support 
 *                      hardwarespecific function for accessing control lines
 *   02-21-2002 TG      added support for different read/write adress and
 *                      ready/busy line access function
 *   02-26-2002 TG      added chip_delay to nand_chip structure to optimize
 *                      command delay times for different chips
 *   04-28-2002 TG      OOB config defines moved from nand.c to avoid duplicate
 *                      defines in jffs2/wbuf.c
 *   08-07-2002 TG      forced bad block location to byte 5 of OOB, even if
 *                      CONFIG_MTD_NAND_ECC_JFFS2 is not set
 *   08-10-2002 TG      extensions to nand_chip structure to support HW-ECC
 *
 *   08-29-2002 tglx    nand_chip structure: data_poi for selecting 
 *                      internal / fs-driver buffer
 *                      support for 6byte/512byte hardware ECC
 *                      read_ecc, write_ecc extended for different oob-layout
 *                      oob layout selections: NAND_NONE_OOB, NAND_JFFS2_OOB,
 *                      NAND_YAFFS_OOB
 *  11-25-2002 tglx     Added Manufacturer code FUJITSU, NATIONAL
 *                      Split manufacturer and device ID structures 
 *
 *  02-08-2004 tglx     added option field to nand structure for chip anomalities
 *  05-25-2004 tglx     added bad block table support, ST-MICRO manufacturer id
 *                      update of nand_chip structure description
 */

#ifndef __LINUX_MTD_ICX1087_NAND_BASE_H
#define __LINUX_MTD_ICX1087_NAND_BASE_H

#include <linux/config.h>
#include <linux/wait.h>
#include <linux/spinlock.h>
#include <linux/mtd/mtd.h>

/**
 * struct nand_chip - NAND Private Flash Chip Data
 * @write_page:         [REPLACEABLE] write page data from the buffer to the chip
 * @read_page:          [REPLACEABLE] read page data from the chip into the buffer
 * @verify_buf:         [REPLACEABLE] verify buffer contents against the chip data
 * @select_chip:        [REPLACEABLE] select chip nr
 * @block_bad:          [REPLACEABLE] check, if the block is bad
 * @block_markbad:      [REPLACEABLE] mark the block bad
 * @dev_ready:          [BOARDSPECIFIC] hardwarespecific function for accesing device ready/busy line
 *                      If set to NULL no access to ready/busy is available and the ready/busy information
 *                      is read from the chip status register
 * @cmdfunc:            [REPLACEABLE] hardwarespecific function for writing commands to the chip
 * @waitfunc:           [REPLACEABLE] hardwarespecific function for wait on ready
 * @erase_cmd:          [INTERN] erase command write function, selectable due to AND support
 * @scan_bbt:           [REPLACEABLE] function to scan bad block table
 * @chip_delay:         [BOARDSPECIFIC] chip dependent delay for transfering data from array to read regs (tR)
 * @chip_lock:          [INTERN] spinlock used to protect access to this structure and the chip
 * @wq:                 [INTERN] wait queue to sleep on if a NAND operation is in progress
 * @state:              [INTERN] the current state of the NAND device
 * @page_shift:         [INTERN] number of address bits in a page (column address bits)
 * @phys_erase_shift:   [INTERN] number of address bits in a physical eraseblock
 * @bbt_erase_shift:    [INTERN] number of address bits in a bbt entry
 * @chip_shift:         [INTERN] number of address bits in one chip
 * @data_buf:           [INTERN] internal buffer for one page + oob 
 * @oob_buf:            [INTERN] oob buffer for one eraseblock
 * @oobdirty:           [INTERN] indicates that oob_buf must be reinitialized
 * @data_poi:           [INTERN] pointer to a data buffer
 * @options:            [BOARDSPECIFIC] various chip options. They can partly be set to inform nand_scan about
 *                      special functionality. See the defines for further explanation
 * @badblockpos:        [INTERN] position of the bad block marker in the oob area
 * @numchips:           [INTERN] number of physical chips
 * @chipsize:           [INTERN] the size of one chip for multichip arrays
 * @pagemask:           [INTERN] page number mask = number of (pages / chip) - 1
 * @pagebuf:            [INTERN] holds the pagenumber which is currently in data_buf
 * @autooob:            [REPLACEABLE] the default (auto)placement scheme
 * @bbt:                [INTERN] bad block table pointer
 * @bbt_td:             [REPLACEABLE] bad block table descriptor for flash lookup
 * @bbt_md:             [REPLACEABLE] bad block table mirror descriptor
 * @priv:               [OPTIONAL] pointer to private chip date
 */


extern int  icx1087_nand_scan         ( struct mtd_info *mtd, int max_chips );
extern int  icx1087_nand_register     ( struct mtd_info *mtd, const struct mtd_partition *parts, int nbparts );
extern void icx1087_nand_release      ( struct mtd_info *mtd );
extern void icx1087_nand_get_chip     ( struct mtd_info *mtd, int new_state );
extern void icx1087_nand_release_chip ( struct mtd_info *mtd );

struct icx1087_nand_dev {

    void        (*select_chip) ( struct mtd_info *mtd, int ce );
    int         (*read_id)     ( struct mtd_info *mtd, u_char *buf, int len );
    int         (*reset)       ( struct mtd_info *mtd );
    int         (*read_status) ( struct mtd_info *mtd );

    int         (*read_page)   ( struct mtd_info *mtd, int page, u_char *buf, u_char *ext );
    int         (*write_page)  ( struct mtd_info *mtd, int page, const u_char *buf, const u_char *ext );
    int         (*erase_block) ( struct mtd_info *mtd, int page);
    int         (*verify_buf)  ( struct mtd_info *mtd, const u_char *buf, int len);

    int         (*scan_bbt)       ( struct mtd_info *mtd );
    int         (*block_isbad)    ( struct mtd_info *mtd, loff_t ofs );
    int         (*block_markbad)  ( struct mtd_info *mtd, loff_t ofs );

    void        (*get_device)     ( struct mtd_info *mtd, int new_state );
    void        (*release_device) ( struct mtd_info *mtd );

    int         (*dev_ready)      ( struct mtd_info *mtd );
    int         (*waitfunc)       ( struct mtd_info *mtd, int state) ;


    char        *name;
    int         maf_id;
    int         dev_id;

    int         chip_delay;

    int         ce;

    loff_t      total_size;
    loff_t      chip_size;
    int         chip_num;
    int         max_chip;

    int         page_size;
    int         block_size;
    int         oob_size;
    int         oob_avail;

    int         chip_shift;
    int         block_shift;
    int         page_shift;
    int         phys_erase_shift;
    int         bbt_erase_shift;

    u_char      *data_buf;
    u_char      *oob_buf;
    u_char      *data_poi;
    u_int       options;

    int         pagemask;
    int         pagebuf;

    uint8_t     *bbt;
    void        *priv;

    spinlock_t              chip_lock;
    wait_queue_head_t       wq;
    nand_state_t            state;
    struct nand_hw_control  *controller;

    int         erase_size;

    int         (*read_page2)   ( struct mtd_info *mtd, int cn[], int page[], u_char *buf[], unsigned char *ext[], int rc[] );
    int         (*write_page2)  ( struct mtd_info *mtd, int cn[], int page[], const u_char *buf[], const u_char *ext[], int rc[] );
    int         (*erase_block2) ( struct mtd_info *mtd, int cn[], int page[], int rc[] );
};

extern struct icx1087_nand_dev *icx1087_nand_bios_get( void );

#endif /* __LINUX_MTD_ICX1087_NAND_BASE_H */
