/*
 *
 *  Copyright 2006,2007 Sony Corporation. 
 *
 *  Based on drivers/mtd/nand/mp200-nand.c
 *
 *  Copyright (C) NEC Electronics Corporation 2006
 *
 *  This program is free software;you can redistribute it and/or modify itunder the terms of
 *  the GNU General Public License as published by Free Softwere Foundation;either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/device.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>

#include <asm/io.h>
#include <asm/delay.h>
#include <asm/memory.h>
#include <asm/arch/pmu.h>
#include <asm/arch/smu.h>
#include <asm/arch/gpiocore.h>

#include "icx1087_nand_bios.h"
#include "icx1087_nand_base.h"


/********************************************************************
  Partition Table 
********************************************************************/
static struct mtd_partition icx1087_nand_partition_info[] = {
    {
        .name = "ICX1087 NAND FS",
        .offset = 0,
        .size = MTDPART_SIZ_FULL
    }
};

/********************************************************************
  Device Driver Structure
********************************************************************/
static struct mtd_info         icx1087_nand_mtd;
static struct icx1087_nand_dev icx1087_nand_bios;
static struct nand_hw_control  icx1087_nand_controller;
static int    icx1087_nand_initialized = 0;
static struct semaphore        icx1087_nand_susp_sem;
static int    icx1087_nand_susp_ce;
#ifdef ICX1087_NAND_USE_INTERRUPT
static wait_queue_head_t icx1087_wqh_interrupt;
volatile static int icx1087_interrupt_done = 1;
#endif

/********************************************************************
  Internal Function
********************************************************************/
static u_int nand_bios_read( void );
static int   nand_bios_init_hw( void );
static int   nand_bios_init_hw_post( void );
static void  nand_bios_select_chip( int ce );
static int   nand_bios_reset( void );
static int   nand_bios_read_status( void );
static int   nand_bios_wait( void );
static int   nand_bios_read_id( u_char *buf, int len );
static int   nand_correct_data( unsigned long dest, int offset );
#ifdef ICX1087_NAND_USE_INTERRUPT
static irqreturn_t nand_bios_interrupt( int irq, void *dev_id, struct pt_regs *regs );
#endif

/********************************************************************
  External Function
********************************************************************/
static void  icx1087_nand_bios_select_chip(struct mtd_info *mtd, int ce );
static int   icx1087_nand_bios_reset( struct mtd_info *mtd );
static int   icx1087_nand_bios_read_status( struct mtd_info *mtd );
static int   icx1087_nand_bios_read_id( struct mtd_info *mtd, u_char *buf, int len );
static int   icx1087_nand_bios_read_page( struct mtd_info *mtd, int  page, u_char *buf, unsigned char *ext);
static int   icx1087_nand_bios_write_page( struct mtd_info *mtd, int page, const u_char *buf, const u_char *ext );
static int   icx1087_nand_bios_erase_block( struct mtd_info *mtd, int page );
     

/**
 *  and_bios_read - read byte data
 */
static unsigned int nand_bios_read( void )
{
    outl( 0x00, NAND_BDRR );    /* ByteDataRead trigger Register */
    udelay( NAND_DELAY_US );
    NAND_WAIT_NOT_CACT();       /* CPU Bus Busy */
    return( inl( NAND_FRD ) & NAND_FRD_DATA_MASK );
}


/**
  * nand_bios_init_hw - initialize hardwaere
  */
static int nand_bios_init_hw( void )
{
    unsigned int i;


    /* --- SMU, PMU --- */

    /* Select Pin Function */
    mp200_pmu_write_reg( SMU_PCTL_NAND, SMU_PCTL_NAND_VALUE, SMU_PCTL_NAND_MASK );
    /* GPIO (XWP) Output Mode */
    mp200_gio_set_mode_output_high32( 1 << ( GPIOCORE_NAND_XWP - 32 ) );
    /* Pull Control */
    mp200_pmu_write_reg( SMU_PULLCTL_NAND, SMU_PULLCTL_NAND_VALUE, SMU_PULLCTL_NAND_MASK );

    /* Regulator ON */
    pwc_regulator_on( PWC_PSCNTGP3 );
    /* Clock ON */
    mp200_pmu_clkctrl_off( MP200_CLKCTRL_NAND );
    mp200_pmu_clkctrl_off( MP200_CLKCTRL_NANDPCLK );
    mp200_pmu_open_clockgate( MP200_CLOCKGATE_NAND_CLK );
    mp200_pmu_open_clockgate( MP200_CLOCKGATE_NAND_PCLK );
    /* Unreset */
    mp200_pmu_unreset_device( MP200_RESETDEVICE_NAND );
    /* Auto Clock ON */
    mp200_pmu_clkctrl_on( MP200_CLKCTRL_NAND );
    mp200_pmu_clkctrl_on( MP200_CLKCTRL_NANDPCLK );


    /* --- NAND Controller --- */

    /* --- BBC --- */
    outl(NAND_MSR_INIT, NAND_MSR);       /* 2048B/page  8bit width */

    outl( BBC_INT_MASK, NAND_INTENCLR ); /* Diable Interrupt */
    outl( BBC_INT_MASK, NAND_INTFFCLR ); /* Interrupt Clear */

    /* --- EMMU --- */
    outl( 0x0000,         NAND_EICR );   /* Diable Interrupt */
    outl( EMMU_INT_MASK,  NAND_EICLR );  /* Interrupt Clear */
    outl( NAND_CECR_MASK, NAND_CECR );   /* Disable All CE */
    outl( 0x0000ff00,     NAND_CDR );    /* Compare Data: no compare.. */
    outl( 0x00000870,     NAND_SCCR );   /* Status Check Cont: default(0x0370) */
    outl( 0x0000,         NAND_CR );     /* NANDǡΥåߥ(0=REZ_OUT1=) */

    /* --- LDMA --- */
    outl( 0x00000000,    NAND_LCR );     /* Control */
    outl( 0x00000000,    NAND_LICR );    /* Disable Interrupt */
    outl( LDMA_INT_MASK, NAND_LICLR );   /* Interrupt Clear */


    /* --- NAND Chip --- */

    /* Send Reset Command */
    for( i = 0; i < NUM_CHIP; i++ ) {
        nand_bios_select_chip( i );
        nand_bios_reset(); 
    }
    nand_bios_select_chip( -1 );

    return 0;
}


/**
  * nand_bios_init_hw_post - initialize hardwaere after chip ditect
  */
static int nand_bios_init_hw_post( void )
{
    struct icx1087_nand_dev *this = &icx1087_nand_bios;

    if ( this->chip_num == 2 ) {
        /* Pull UP RB2 */
        mp200_pmu_write_reg( SMU_PULLCTL_NAND, SMU_PULLCTL_NAND_RB2_PU, SMU_PULLCTL_NAND_RB2_MASK );
    }

    if ( ( this->maf_id == 0x98 ) && ( this->dev_id == 0xd5 ) ) {

        /* TOSHIBA 2G */
        outl(NAND_MSR_TOSHIBA, NAND_MSR);
    } else if ( ( this->maf_id == 0xec ) && ( this->dev_id == 0xd5 ) ) {

        /* SAMSUNG 2G */
        outl(NAND_MSR_SAMSUNG, NAND_MSR);
    } else {

        printk( KERN_WARNING "%s: unknown chip access timing.\n", __FUNCTION__ );
    }

    return 0;
}


/**
  * nand_bios_select_chip - set chip enable
  * @ce: chip no.
  */
static void nand_bios_select_chip( int ce )
{
    if ( ( ce < 0 ) || ( ce >= NUM_CHIP ) ) {
        outl( NAND_CECR_MASK, NAND_CECR );
        gpiocore_write( GPIOCORE_GP_DO0, GPIOCORE_NAND_XWP, 0x00 );
    } else {
        outl( NAND_CECR_MASK & ~( NAND_CECR_CEZ << ce ), NAND_CECR );
        gpiocore_write( GPIOCORE_GP_DO0, GPIOCORE_NAND_XWP, 0x01 );
    }
}


/**
  * nand_bios_reset - send reset command and wait ready
  */
static int nand_bios_reset( void )
{
    NAND_MACRO_CWR( NAND_CMD_RESET );
    return nand_bios_wait();
}


/**
  * nand_bios_read_status - send status_read command and get status data
  */
static int nand_bios_read_status( void )
{
    int status;
    NAND_MACRO_CWR( NAND_CMD_STATUS );
    status = nand_bios_read();
    return status;
}


/**
  * nand_bios_wait - wait ready
  */
static int nand_bios_wait( void )
{
    int status = 0;
    unsigned long timeo = jiffies + 2;

    do {
        status = nand_bios_read_status(); 
    } while ( ( !( status & NAND_STATUS_READY ) ) && time_before( jiffies, timeo ) );

    return status;
}


/**
  * nand_bios_read_id - send id_read comamnd and get ID data
  */
static int nand_bios_read_id( u_char *buf, int len )
{
    int i;
    NAND_MACRO_CWR( NAND_CMD_READID );
    NAND_MACRO_LAWR( 0x00 );
    for ( i = 0; i < len; i++ ) {
        buf[i] = nand_bios_read();
    }
    return 0;
}


/**
  * icx1087_nand_bios_select_chip - set chip enable
  */
static void icx1087_nand_bios_select_chip(struct mtd_info *mtd, int ce )
{
    struct icx1087_nand_dev *this = mtd->priv;

    down( &icx1087_nand_susp_sem );

    this->ce = icx1087_nand_susp_ce = ce;

    nand_bios_select_chip( ce );

    up( &icx1087_nand_susp_sem );
}


/**
  * icx1087_nand_bios_reset - nand chip reset
  */
static int icx1087_nand_bios_reset( struct mtd_info *mtd )
{
    int ret;

    down( &icx1087_nand_susp_sem );

    ret = nand_bios_reset();

    up( &icx1087_nand_susp_sem );

    return ret;
}


/**
  * icx1087_nand_bios_read_status - get status data
  */
static int icx1087_nand_bios_read_status( struct mtd_info *mtd )
{
    int status;

    down( &icx1087_nand_susp_sem );

    status = nand_bios_read_status();

    up( &icx1087_nand_susp_sem );

    return status;
}


/**
  * icx1087_nand_bios_read_id - get ID data
  */
static int icx1087_nand_bios_read_id( struct mtd_info *mtd, u_char *buf, int len )
{
    int ret;

    down( &icx1087_nand_susp_sem );

    ret = nand_bios_read_id( buf, len );

    up( &icx1087_nand_susp_sem );

    return ret;
}


/**
  * nand_correct_data - correct data
  */
static int nand_correct_data( unsigned long dest, int offset )
{
    unsigned long ERR_VAL = NAND_RSEVxR;
    unsigned long ERR_PLA = NAND_RSEPxR;
    unsigned long rs_status;
    unsigned int  pla, val;
    unsigned long cor_adr;
    unsigned char cor_val;
    int i;

    rs_status = inl( NAND_RSSR ) & NAND_RSSR_MASK;

    if ( rs_status == RS_NO_ERROR ) {
        return 0;   /* no error */
    }
    if ( rs_status == RS_UNCORRECT_ERROR ) {
        return -1;  /* uncorrectable error */
    }
    for ( i = 0; i <= 3; i++ ) {
        pla = inl( ERR_PLA );
        val = inl( ERR_VAL );
        if ( pla > 0x0d ) {
            /* data field error. */
            if ( val > 0xff ) {
                return -1;
            } else if( offset >= 0 ) {
                cor_adr = ( (dest + offset) + 0x20d - pla );
                cor_val = ( *(volatile unsigned char *)(cor_adr) );
                cor_val ^= val;
                (*(volatile unsigned char *)(cor_adr)) = cor_val;
            }
        } else if ( pla > 0x07 ) {
            /* spare field error. */
            if ( val > 0xff ) {
                return -1;
            } else if( offset < 0 ){
                cor_adr = ( (dest + offset) + 0x00d - pla );
                cor_val = ( *(volatile unsigned char *)(cor_adr) );
                cor_val ^= val;
                (*(volatile unsigned char *)(cor_adr)) = cor_val;
            }
        } else if ( !((pla == 0x0) && (i != 0)) ) {
            /* RS field error. ignored. */
            cor_adr = pla;
        } else {  /* pla == 0 && i != 0 */
            /* no more error */
            break;
        }
        ERR_PLA += 0x4;
        ERR_VAL += 0x4;
    }
    return 0;
}


#ifdef ICX1087_NAND_USE_INTERRUPT
/**
  * nand_bios_interrupt - interrupt handler
  */
static irqreturn_t nand_bios_interrupt( int irq, void *dev_id, struct pt_regs *regs )
{
    if ( icx1087_interrupt_done ) {
        printk(KERN_INFO "%s: illegal interrupt!!\n", __FUNCTION__ );
        /* BBC INT disable */
        outl( BBC_INT_MASK, NAND_INTENCLR );
        /* EMMU interrupt disable */
        outl( 0x0000, NAND_EICR );
        /* LDMA interrupt disable */
        outl( 0x00, NAND_LICR );
        return IRQ_HANDLED;
    }

    if ( inl(NAND_INTSTATUS) & BBC_INT_MASK ) {
        /* BBC INT disable */
        outl( BBC_INT_MASK, NAND_INTENCLR );
        /* BBC INT status all clear */
        outl( BBC_INT_MASK, NAND_INTFFCLR );
    }

    if ( inl(NAND_EIRSR) & EMMU_INT_COMPEN ) {
        /* EMMU interrupt disable */
        outl( 0x0000, NAND_EICR );
        /* wake up */
        wake_up( &icx1087_wqh_interrupt );
        icx1087_interrupt_done = 1;
    }

    if ( inl(NAND_LIRSR) & LDMA_INT_PSC ) {
        /* LDMA interrupt disable */
        outl( 0x00, NAND_LICR );
        /* wake up */
        wake_up( &icx1087_wqh_interrupt );
        icx1087_interrupt_done = 1;
    }

    return IRQ_HANDLED;
}
#endif


/**
  * icx1087_nand_bios_read_page - read page
  */
static int icx1087_nand_bios_read_page( struct mtd_info *mtd, int  page, u_char *buf, unsigned char *ext)
{
    struct   icx1087_nand_dev *this = mtd->priv;
    unsigned long  pointer;
    unsigned int   offset, psc_step, psc_ldma_start_step;
    unsigned long  msr, ipr, status;
    unsigned short etmp, blank;
    unsigned int   phys;
    int uncorrect, i;
    int needSpareChk = 0;
    int ce = this->ce;

    if ( ( ce < 0 ) || ( ce > 3 ) ) return BIOS_ERR_PARAM;
    ce = 1 << ce;

    down( &icx1087_nand_susp_sem );

    /* --- BBC --- */

    /* set destination address(SRAM address) */
    phys = dma_map_single( NULL, (void*)buf, ICX1087_NAND_PAGE_SIZE, DMA_FROM_DEVICE );
    outl( (unsigned long)phys, NAND_ATBAR );
    /* AHB Interrupt Clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );


    /* --- EMMU --- */

    /* Compare Data: no compare.. */
    outl( 0x0000ff00,     NAND_CDR );
    /* EMMU interrupt clear */
    outl( EMMU_INT_COMPEN, NAND_EICLR );


    /* --- PSC --- */

    PSC_CMD_SET_BEGIN();

    PSC_CMD_SET( PSC_CMD_CESEL,  ce ^ NAND_CECR_MASK ); /* CEZ_OUT0 = '0' */
    PSC_CMD_SET( PSC_CMD_OUTCMD, NAND_CMD_READ0 );      /* Read command(00h) */

    /* send address command */
    PSC_CMD_SET( PSC_CMD_OUTADRS, 0 );
    PSC_CMD_SET( PSC_CMD_OUTADRS, 0 );
    PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page)       & 0x00ff) );
    PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page >> 8)  & 0x00ff) );
    PSC_CMD_SET( PSC_CMD_OUTLTADRS, ( (page >> 16) & 0x00ff) );

    PSC_CMD_SET( PSC_CMD_OUTCMD, NAND_CMD_READSTART ); /* read start command */
    PSC_CMD_SET( PSC_CMD_WAITBUSY, ce );               /* wait busy          */
    PSC_CMD_SET( PSC_CMD_WAITRDY,  ce );               /* wait ready         */

    /* Local DMA start command */
    psc_ldma_start_step = PSC_SEQUENCE_CUR();
    for ( i = 0; i < 4 ; i++ ) {
        PSC_CMD_SET( PSC_CMD_DMAST,  PSC_DMA_START );  /* LDMA start    */
        PSC_CMD_SET( PSC_CMD_WAITTC, PSC_DMA_RSOK  );  /* wait rs check */
    }

    PSC_CMD_SET( PSC_CMD_STOPINT, PSC_IIDR_DMA_END );  /* PSC stop command */

    PSC_CMD_SET_END();


    /* --- LDMA --- */

    /* LDMA Interrupt Clear */
    outl( LDMA_INT_MASK, NAND_LICLR );


    /* --- BBC --- */

    /* AHB Master trans start */
    msr = inl( NAND_MSR );
    msr |= ( NAND_MSR_AHBDIR | NAND_MSR_AHBMEN );      /* Master write and enable */
    outl( msr, NAND_MSR );                             /* DMAenableˤ顢ǸޤǼ¹ԤʤȤʤ */


    /* --- PSC --- */

    /* start PSC sequence */
    outl( PSC_START, NAND_PSCR );
    uncorrect = 0;
    i = NAND_READ_TIMEOUT;

    while (1) {
        status = inl( NAND_LIRSR );

        if ( status & LDMA_INT_PSC ) {
            /* PSC complete */

            /* LDMA INT status clear */
            outl( ( LDMA_INT_RSOK | LDMA_INT_PSC | LDMA_INT_LDMACTC ), NAND_LICLR );

            /* wait AHB trans */
            while (!(inl(NAND_INTRAWSTATUS) & BBC_INT_PAGEINT)) {}

            /* BBC INT status clear */
            outl( BBC_INT_MASK, NAND_INTFFCLR );

            /* trans OK */
            break;

        } else if ( status & LDMA_INT_RSERR ) {
            /* PSC abort(RS error) */

            /* wait RS calculate */
            while ( !( inl(NAND_RSSR) & NAND_RSSR_CALC_END ) ) {}

            /* LDMA INT status all clear */
            outl( LDMA_INT_MASK, NAND_LICLR );

            /* wait AHB trans */
            while ( !(inl(NAND_INTRAWSTATUS) & ( BBC_INT_PAGEINT | BBC_INT_512INT ) ) ) {}

            /* BBC INT status all clear */
            outl( BBC_INT_MASK, NAND_INTFFCLR );

            /* correct data */
            ipr = inl( NAND_IPR );
            offset = RS_CHECK_SIZE * (((ipr-psc_ldma_start_step) / 2) - 1);
            if ( nand_correct_data( (unsigned long)buf, offset ) ) {
                uncorrect = 1; /* uncorrectable error */
            }
            /* flush cache */
            consistent_sync( (void*)(buf + offset), RS_CHECK_SIZE, DMA_TO_DEVICE );

            /* last sequence ? */
            if ( ipr == PSC_SEQUENCE_END() ) {
                needSpareChk = 1;
                break;
            } else {
                /* restart PSC sequence */
                outl( PSC_START, NAND_PSCR );
            }
        }
        if(--i <= 0){
            goto timeout;
        }
    }

    /* Ĺ٤0xFFξECC顼ʤäȤˤ */
    /* set spare data */
    blank = 0xffff;
    for( i = 0; i < 12; i += 4 ){
        etmp = inl( NAND_SDH0R + i );
        blank &= etmp;
        if( ext ) {
            *ext++ = etmp;
            *ext++ = etmp >> 8;
        }
    }

    /* RS error correct */
    if( needSpareChk && ext ) nand_correct_data( (unsigned long)ext, -6 );

    /* for RS ECC read */
    for( i = 0; i < 20; i+=4 ){
        etmp = inl( NAND_SDH3R + i );
        blank &= etmp;
    }
    blank &= etmp;
    if( blank == 0xFFFF ) uncorrect = 0;

    up( &icx1087_nand_susp_sem );

    return uncorrect;

timeout:

    printk( KERN_NOTICE "%s: timeout!!\n", __FUNCTION__ );
    /* STOP PSC */
    outl( 0, NAND_PSCR );
    /* LDMA INT status clear */
    outl( LDMA_INT_MASK, NAND_LICLR );
    /* BBC INT status clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );

    up( &icx1087_nand_susp_sem );

    return BIOS_ERR_TIMEOUT;
}


/**
  * icx1087_nand_bios_write_page - write page
  */
static int icx1087_nand_bios_write_page( struct mtd_info *mtd, int page, const u_char *buf, const u_char *ext )
{
    struct icx1087_nand_dev *this = mtd->priv;
    unsigned long pointer;
    unsigned int  psc_step;
    unsigned long msr;
    unsigned int  phys;
    int ce = this->ce;
    int i;

    if ( ( ce < 0 ) || ( ce > 3 ) ) return BIOS_ERR_PARAM;
    ce = 1 << ce;

    down( &icx1087_nand_susp_sem );

    /* --- BBC --- */

    /* set destination address(SRAM address) */
    phys = dma_map_single( NULL, (void*)buf, ICX1087_NAND_PAGE_SIZE, DMA_TO_DEVICE );
    outl( (unsigned long)phys, NAND_ATBAR );
    /* AHB Interrupt Clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );


    /* --- EMMU --- */

    /* Compare Data: no compare.. */
    outl( 0x0000ff00,     NAND_CDR );
    /* EMMU interrupt clear */
    outl( EMMU_INT_COMPEN, NAND_EICLR );

    /* set spare data */
    if ( ext ) {
        outl( ext[0] | ext[1] << 8, NAND_SDH0R );
        outl( ext[2] | ext[3] << 8, NAND_SDH1R );
        outl( ext[4] | ext[5] << 8, NAND_SDH2R );
    } else {
        outl( 0, NAND_SDH0R );
        outl( 0, NAND_SDH1R );
        outl( 0, NAND_SDH2R );
    }


    /* --- PSC --- */

    PSC_CMD_SET_BEGIN();

    PSC_CMD_SET( PSC_CMD_CESEL,  ce ^ NAND_CECR_MASK); /* CEZ_OUT0 = '0' */
    PSC_CMD_SET( PSC_CMD_OUTCMD, NAND_CMD_SEQIN);      /* DATAIN command(80h) */

    /* send address command */
    PSC_CMD_SET( PSC_CMD_OUTADRS, 0 );
    PSC_CMD_SET( PSC_CMD_OUTADRS, 0 );
    PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page)       & 0xff ) );
    PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page >> 8)  & 0xff ) );
    PSC_CMD_SET( PSC_CMD_OUTLTADRS, ( (page >> 16) & 0xff ) );

    /* Local DMA start command */
    for (i = 4; i > 0; i--) {
        PSC_CMD_SET(PSC_CMD_DMAST,  PSC_DMA_START);    /* LDMA start    */
        PSC_CMD_SET(PSC_CMD_WAITTC, PSC_DMA_TCEA);     /* wait TC event active */
    }

    PSC_CMD_SET(PSC_CMD_OUTCMD, NAND_CMD_PAGEPROG);    /* program command() */
    PSC_CMD_SET(PSC_CMD_WAITBUSY, ce);                 /* wait busy */
    PSC_CMD_SET(PSC_CMD_WAITRDY, ce);                  /* wait ready */
    PSC_CMD_SET(PSC_CMD_CHKST, NAND_CMD_STATUS);       /* status read */
    PSC_CMD_SET(PSC_CMD_STOPINT, PSC_IIDR_ADR_SEND);   /* PSC stop command */

    PSC_CMD_SET_END();


    /* --- LDMA --- */

    /* LDMA Interrupt Clear */
    outl( LDMA_INT_MASK, NAND_LICLR );


    /* --- BBC --- */

    /* AHB Master trans start */
    msr = inl( NAND_MSR );
    msr |= NAND_MSR_AHBMEN;
    msr &= ~NAND_MSR_AHBDIR;
    outl( msr, NAND_MSR );


#ifdef ICX1087_NAND_USE_INTERRUPT

    /* --- INT Enable --- */

    icx1087_interrupt_done = 0;
    /* AHB PAGE INT enable */
    outl( BBC_INT_PAGEINT, NAND_INTENSET );
    /* LDMA INT enable */
    outl( LDMA_INT_PSC, NAND_LICR );

#endif


    /* --- PSC --- */

    /* start PSC sequence */
    outl( PSC_START, NAND_PSCR );


#ifdef ICX1087_NAND_USE_INTERRUPT

    wait_event_timeout( icx1087_wqh_interrupt, icx1087_interrupt_done, NAND_WRITE_WAIT_TICK );
    icx1087_interrupt_done = 1;

    /* LDMA INT disable */
    outl( 0x00, NAND_LICR );
    /* BBC INT disable */
    outl( BBC_INT_MASK, NAND_INTENCLR );

    /* wait PSC sequence end */
    i = 1;

#else

    /* wait PSC sequence end */
    i = NAND_WRITE_TIMEOUT;

#endif

    while ( !( inl(NAND_LIRSR) & LDMA_INT_PSC ) ) {
        if(--i <= 0){
            goto timeout;
        }
    }

    /* LDMA INT status clear */
    outl( LDMA_INT_MASK, NAND_LICLR );
    /* BBC INT status clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );

    up( &icx1087_nand_susp_sem );

    return inl( NAND_FWSR ) & 1;

timeout:

    printk( KERN_NOTICE "%s: timeout!!\n", __FUNCTION__ );
    /* STOP PSC */
    outl( 0, NAND_PSCR );
    /* LDMA INT status clear */
    outl( LDMA_INT_MASK, NAND_LICLR );
    /* BBC INT status clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );

    up( &icx1087_nand_susp_sem );

    return BIOS_ERR_TIMEOUT;
}


/**
  * icx1087_nand_bios_erase_block - erase block
  */
static int icx1087_nand_bios_erase_block( struct mtd_info *mtd, int page )
{
    struct icx1087_nand_dev *this = mtd->priv;
    unsigned long pointer;
    unsigned int  psc_step;
    int ce = this->ce;
    int i;


    if ( ( ce < 0 ) || ( ce > 3 ) ) return BIOS_ERR_PARAM;
    ce = 1 << ce;

    down( &icx1087_nand_susp_sem );

    /* --- LDMA --- */

    /* LDMA Interrupt Clear */
    outl( LDMA_INT_MASK, NAND_LICLR );


    /* --- BBC --- */

    /* AHB Interrupt Clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );


    /* --- EMMU --- */

    /* Compare Data: no compare.. */
    outl( 0x0000ff00,     NAND_CDR );
    /* EMMU interrupt clear */
    outl( EMMU_INT_COMPEN, NAND_EICLR );


    /* --- PSC --- */

    PSC_CMD_SET_BEGIN();

    PSC_CMD_SET( PSC_CMD_CESEL,  ce ^ NAND_CECR_MASK ); /* CEZ_OUT0 = '0' */
    PSC_CMD_SET( PSC_CMD_OUTCMD, NAND_CMD_ERASE1 );     /* erase1 command() */

    /* send address command */
    PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page)       & 0xff ) );
    PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page >> 8)  & 0xff ) );
    PSC_CMD_SET( PSC_CMD_OUTLTADRS, ( (page >> 16) & 0xff ) );

    PSC_CMD_SET( PSC_CMD_OUTCMD, NAND_CMD_ERASE2 );    /* erase1 command() */
    PSC_CMD_SET( PSC_CMD_WAITBUSY, ce );               /* wait busy */
    PSC_CMD_SET( PSC_CMD_WAITRDY,  ce );               /* wait ready */
    PSC_CMD_SET( PSC_CMD_CHKST, NAND_CMD_STATUS );     /* status read */
    PSC_CMD_SET( PSC_CMD_STOPINT, PSC_IIDR_ADR_SEND);  /* PSC stop command */

    PSC_CMD_SET_END();


#ifdef ICX1087_NAND_USE_INTERRUPT

    /* --- INT Enable --- */

    icx1087_interrupt_done = 0;
    /* interrupt enable */
    outl( LDMA_INT_PSC, NAND_LICR );

#endif


    /* --- PSC --- */

    /* start PSC sequence */
    outl( PSC_START, NAND_PSCR );


#ifdef ICX1087_NAND_USE_INTERRUPT

    wait_event_timeout( icx1087_wqh_interrupt, icx1087_interrupt_done, NAND_ERASE_WAIT_TICK );
    icx1087_interrupt_done = 1;

    /* LDMA INT disable */
    outl( 0x00, NAND_LICR );

    /* wait PSC sequence end */
    i = 1;

#else

    /* wait PSC sequence end */
    i = NAND_ERASE_TIMEOUT;

#endif

    while ( !( inl(NAND_LIRSR) & LDMA_INT_PSC ) ) {
        if(--i <= 0){
            goto timeout;
        }
    }

    /* LDMA Interrupt Clear */
    outl(LDMA_INT_MASK, NAND_LICLR);
    /* BBC INT status clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );

    up( &icx1087_nand_susp_sem );

    return inl( NAND_FWSR ) & 1;

timeout:

    printk( KERN_NOTICE "%s: timeout!!\n", __FUNCTION__ );
    /* STOP PSC */
    outl( 0, NAND_PSCR );
    /* LDMA Interrupt Clear */
    outl( LDMA_INT_MASK, NAND_LICLR );
    /* BBC INT status clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );

    up( &icx1087_nand_susp_sem );

    return BIOS_ERR_TIMEOUT;
}


/**
  * icx1087_nand_bios_read_page2 - read page ( 2 chips interleave )
  */
static int icx1087_nand_bios_read_page2( struct mtd_info *mtd, int cn[], int page[], u_char *buf[], unsigned char *ext[], int rc[] )
{
    unsigned long  pointer;
    unsigned int   offset, psc_step, psc_ldma_start_step;
    unsigned long  msr, ipr, status;
    unsigned short etmp, blank;
    unsigned int   phys;
    int uncorrect, i, j, ce[NUM_CHIP], ces=0, ret=0, n=2;
    int needSpareChk = 0;
    u_char *pext;

    for ( i = 0; !ret && ( i < n ); i++ ) {
        ce[i] = 1 << cn[i];
        if ( ( cn[i] < 0 ) || ( cn[i] >= NUM_CHIP ) || ( ces & ce[i] ) ) {
            ret = BIOS_ERR_PARAM;
        } else {
            ces |= ce[i];
        }
    }

    if ( ret ) {
        for ( i = 0; i < n; i++ ) rc[i] = ret;
        return ret;
    }

    for ( i = 0; i < n; i++ ) rc[i] = 0;

    down( &icx1087_nand_susp_sem );


    /* --- PSC --- */

    PSC_CMD_SET_BEGIN();

    for ( i = 0; i < n; i++ ) {

        PSC_CMD_SET( PSC_CMD_CESEL,  ce[i] ^ NAND_CECR_MASK ); /* CEZ_OUT0 = '0' */
        PSC_CMD_SET( PSC_CMD_OUTCMD, NAND_CMD_READ0 );         /* Read command(00h) */

        /* send address command */
        PSC_CMD_SET( PSC_CMD_OUTADRS, 0 );
        PSC_CMD_SET( PSC_CMD_OUTADRS, 0 );
        PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page[i])       & 0x00ff) );
        PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page[i] >> 8)  & 0x00ff) );
        PSC_CMD_SET( PSC_CMD_OUTLTADRS, ( (page[i] >> 16) & 0x00ff) );

        PSC_CMD_SET( PSC_CMD_OUTCMD, NAND_CMD_READSTART );     /* read start command */
    }

    PSC_CMD_SET( PSC_CMD_STOPINT, PSC_IIDR_DMA_END );          /* PSC stop command */

    PSC_CMD_SET_END();

    /* --- BBC --- */

    /* AHB Interrupt Clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );

    /* --- EMMU --- */

    /* Compare Data: no compare.. */
    outl( 0x0000ff00,     NAND_CDR );
    /* EMMU interrupt clear */
    outl( EMMU_INT_COMPEN, NAND_EICLR );

    /* --- LDMA --- */

    /* LDMA Interrupt Clear */
    outl( LDMA_INT_MASK, NAND_LICLR );

    /* --- PSC --- */

    /* start PSC sequence */
    outl( PSC_START, NAND_PSCR );

    /* wait PSC sequence end */
    i = NAND_READ_TIMEOUT;

    while ( !( inl(NAND_LIRSR) & LDMA_INT_PSC ) ) {
        if(--i <= 0){
            goto timeout;
        }
    }

    for ( i = 0; i < n; i++ ) {

        /* --- BBC --- */

        /* set destination address(SRAM address) */
        phys = dma_map_single( NULL, (void*)buf[i], ICX1087_NAND_PAGE_SIZE, DMA_FROM_DEVICE );
        outl( (unsigned long)phys, NAND_ATBAR );
        /* AHB Interrupt Clear */
        outl( BBC_INT_MASK, NAND_INTFFCLR );

        /* --- PSC --- */

        PSC_CMD_SET_BEGIN();

        PSC_CMD_SET( PSC_CMD_CESEL,  ce[i] ^ NAND_CECR_MASK ); /* CEZ_OUT0 = '0' */
        PSC_CMD_SET( PSC_CMD_WAITRDY,  ce[i] );                /* wait ready     */

        /* Local DMA start command */
        psc_ldma_start_step = PSC_SEQUENCE_CUR();
        for ( j = 0; j < 4 ; j++ ) {
            PSC_CMD_SET( PSC_CMD_DMAST,  PSC_DMA_START );  /* LDMA start    */
            PSC_CMD_SET( PSC_CMD_WAITTC, PSC_DMA_RSOK  );  /* wait rs check */
        }

        PSC_CMD_SET( PSC_CMD_STOPINT, PSC_IIDR_DMA_END );  /* PSC stop command */

        PSC_CMD_SET_END();


        /* --- LDMA --- */

        /* LDMA Interrupt Clear */
        outl( LDMA_INT_MASK, NAND_LICLR );


        /* --- BBC --- */

        /* AHB Master trans start */
        msr = inl( NAND_MSR );
        msr |= ( NAND_MSR_AHBDIR | NAND_MSR_AHBMEN );      /* Master write and enable */
        outl( msr, NAND_MSR );                             /* DMAenableˤ顢ǸޤǼ¹ԤʤȤʤ */


        /* --- PSC --- */

        /* start PSC sequence */
        outl( PSC_START, NAND_PSCR );
        uncorrect = 0;
        j = NAND_READ_TIMEOUT;

        while (1) {
            status = inl( NAND_LIRSR );

            if ( status & LDMA_INT_PSC ) {
                /* PSC complete */

                /* LDMA INT status clear */
                outl( ( LDMA_INT_RSOK | LDMA_INT_PSC | LDMA_INT_LDMACTC ), NAND_LICLR );

                /* wait AHB trans */
                while (!(inl(NAND_INTRAWSTATUS) & BBC_INT_PAGEINT)) {}

                /* BBC INT status clear */
                outl( BBC_INT_MASK, NAND_INTFFCLR );

                /* trans OK */
                break;

            } else if ( status & LDMA_INT_RSERR ) {
                /* PSC abort(RS error) */

                /* wait RS calculate */
                while ( !( inl(NAND_RSSR) & NAND_RSSR_CALC_END ) ) {}

                /* LDMA INT status all clear */
                outl( LDMA_INT_MASK, NAND_LICLR );

                /* wait AHB trans */
                while ( !(inl(NAND_INTRAWSTATUS) & ( BBC_INT_PAGEINT | BBC_INT_512INT ) ) ) {}

                /* BBC INT status all clear */
                outl( BBC_INT_MASK, NAND_INTFFCLR );

                /* correct data */
                ipr = inl( NAND_IPR );
                offset = RS_CHECK_SIZE * (((ipr-psc_ldma_start_step) / 2) - 1);
                if ( nand_correct_data( (unsigned long)buf[i], offset ) ) {
                    uncorrect = 1; /* uncorrectable error */
                }

                /* last sequence ? */
                if ( ipr == PSC_SEQUENCE_END() ) {
                    needSpareChk = 1;
                    break;
                } else {
                    /* restart PSC sequence */
                    outl( PSC_START, NAND_PSCR );
                }
            }
            if(--j <= 0){
                goto timeout;
            }
        }

        /* Ĺ٤0xFFξECC顼ʤäȤˤ */
        /* set spare data */
        blank = 0xffff;
        if( ext && ext[i] ) pext = ext[i]; else pext = NULL;
        for( j = 0; j < 12; j += 4 ){
            etmp = inl( NAND_SDH0R + j );
            blank &= etmp;
            if( pext ) {
                *pext++ = etmp;
                *pext++ = etmp >> 8;
            }
        }

        /* RS error correct */
        if( needSpareChk && ext && ext[i] ) nand_correct_data( (unsigned long)ext[i], -6 );

        /* for RS ECC read */
        for( j = 0; j < 20; j+=4 ){
            etmp = inl( NAND_SDH3R + j );
            blank &= etmp;
        }
        blank &= etmp;
        if( blank == 0xFFFF ) uncorrect = 0;

        rc[i] = uncorrect;
    }

    for ( i = 0, ret = 0; i < n; i++ ) ret |= rc[i];

    up( &icx1087_nand_susp_sem );

    return ret;

timeout:

    printk( KERN_NOTICE "%s: timeout!!\n", __FUNCTION__ );
    /* STOP PSC */
    outl( 0, NAND_PSCR );
    /* LDMA INT status clear */
    outl( LDMA_INT_MASK, NAND_LICLR );
    /* BBC INT status clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );

    for ( i = 0, ret = 0; i < n; i++ ) rc[i] = BIOS_ERR_TIMEOUT;

    up( &icx1087_nand_susp_sem );

    return BIOS_ERR_TIMEOUT;
}


/**
  * icx1087_nand_bios_write_page - write page ( 2 chips interleave )
  */
static int icx1087_nand_bios_write_page2( struct mtd_info *mtd, int cn[], int page[], const u_char *buf[], const u_char *ext[], int rc[] )
{
    unsigned long pointer;
    unsigned int  psc_step, psc_chkst_step;
    unsigned long msr, ipr;
    unsigned int  phys;
    int i, j, ce[NUM_CHIP], ces=0, ret=0, n=2;

    for ( i = 0; !ret && ( i < n ); i++ ) {
        ce[i] = 1 << cn[i];
        if ( ( cn[i] < 0 ) || ( cn[i] >= NUM_CHIP ) || ( ces & ce[i] ) ) {
            ret = BIOS_ERR_PARAM;
        } else {
            ces |= ce[i];
        }
    }

    if ( ret ) {
        for ( i = 0; i < n; i++ ) rc[i] = ret;
        return ret;
    }

    for ( i = 0; i < n; i++ ) rc[i] = 0;

    down( &icx1087_nand_susp_sem );

    gpiocore_write( GPIOCORE_GP_DO0, GPIOCORE_NAND_XWP, 0x01 );

    /* --- EMMU --- */

    /* Compare Data: no compare.. */
    outl( 0x0000ff00,     NAND_CDR );
    /* EMMU interrupt clear */
    outl( EMMU_INT_COMPEN, NAND_EICLR );

    for ( i = 0; i < n; i++ ) {

        /* --- BBC --- */

        /* set destination address(SRAM address) */
        phys = dma_map_single( NULL, (void*)buf[i], ICX1087_NAND_PAGE_SIZE, DMA_TO_DEVICE );
        outl( (unsigned long)phys, NAND_ATBAR );
        /* AHB Interrupt Clear */
        outl( BBC_INT_MASK, NAND_INTFFCLR );

        /* --- EMMU --- */

        /* set spare data */
        if ( ext && ext[i] ) {
            outl( ext[i][0] | ext[i][1] << 8, NAND_SDH0R );
            outl( ext[i][2] | ext[i][3] << 8, NAND_SDH1R );
            outl( ext[i][4] | ext[i][5] << 8, NAND_SDH2R );
        } else {
            outl( 0, NAND_SDH0R );
            outl( 0, NAND_SDH1R );
            outl( 0, NAND_SDH2R );
        }

        /* --- PSC --- */

        PSC_CMD_SET_BEGIN();

        PSC_CMD_SET( PSC_CMD_CESEL,  ce[i] ^ NAND_CECR_MASK); /* CEZ_OUT0 = '0' */
        PSC_CMD_SET( PSC_CMD_OUTCMD, NAND_CMD_SEQIN);         /* DATAIN command(80h) */

        /* send address command */
        PSC_CMD_SET( PSC_CMD_OUTADRS, 0 );
        PSC_CMD_SET( PSC_CMD_OUTADRS, 0 );
        PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page[i])       & 0xff ) );
        PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page[i] >> 8)  & 0xff ) );
        PSC_CMD_SET( PSC_CMD_OUTLTADRS, ( (page[i] >> 16) & 0xff ) );

        /* Local DMA start command */
        for (j = 4; j > 0; j--) {
            PSC_CMD_SET(PSC_CMD_DMAST,  PSC_DMA_START);    /* LDMA start    */
            PSC_CMD_SET(PSC_CMD_WAITTC, PSC_DMA_TCEA);     /* wait TC event active */
        }

        PSC_CMD_SET(PSC_CMD_OUTCMD, NAND_CMD_PAGEPROG);    /* program command() */
        PSC_CMD_SET(PSC_CMD_STOPINT, PSC_IIDR_ADR_SEND);   /* PSC stop command */

        PSC_CMD_SET_END();

        /* --- LDMA --- */

        /* LDMA Interrupt Clear */
        outl( LDMA_INT_MASK, NAND_LICLR );

        /* --- BBC --- */

        /* AHB Master trans start */
        msr = inl( NAND_MSR );
        msr |= NAND_MSR_AHBMEN;
        msr &= ~NAND_MSR_AHBDIR;
        outl( msr, NAND_MSR );


#ifdef ICX1087_NAND_USE_INTERRUPT

        /* --- INT Enable --- */

        icx1087_interrupt_done = 0;
        /* AHB PAGE INT enable */
        outl( BBC_INT_PAGEINT, NAND_INTENSET );
        /* LDMA INT enable */
        outl( LDMA_INT_PSC, NAND_LICR );

#endif

        /* --- PSC --- */

        /* start PSC sequence */
        outl( PSC_START, NAND_PSCR );

#ifdef ICX1087_NAND_USE_INTERRUPT

        wait_event_timeout( icx1087_wqh_interrupt, icx1087_interrupt_done, NAND_WRITE_WAIT_TICK );
        icx1087_interrupt_done = 1;

        /* LDMA INT disable */
        outl( 0x00, NAND_LICR );
        /* BBC INT disable */
        outl( BBC_INT_MASK, NAND_INTENCLR );

        /* wait PSC sequence end */
        j = 1;

#else

        /* wait PSC sequence end */
        j = NAND_WRITE_TIMEOUT;

#endif

        while ( !( inl(NAND_LIRSR) & LDMA_INT_PSC ) ) {
            if(--j <= 0){
                goto timeout;
            }
        }
    }

    /* --- LDMA --- */

    /* LDMA INT status clear */
    outl( LDMA_INT_MASK, NAND_LICLR );


    /* --- BBC --- */

    /* AHB Interrupt Clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );

    /* --- EMMU --- */

    /* Compare Data: bit0 is 0 */
    outl( 0x0000fe00,     NAND_CDR );
    /* EMMU interrupt clear */
    outl( EMMU_INT_COMPEN, NAND_EICLR );

    /* --- PSC --- */

    PSC_CMD_SET_BEGIN();

    psc_chkst_step = PSC_SEQUENCE_CUR();

    for ( i = 0; i < n; i++ ) {

        PSC_CMD_SET( PSC_CMD_CESEL,  ce[i] ^ NAND_CECR_MASK); /* CEZ_OUT0 = '0' */
        PSC_CMD_SET(PSC_CMD_WAITRDY, ce[i]);                  /* wait ready */
        PSC_CMD_SET(PSC_CMD_CHKST, NAND_CMD_STATUS);          /* status read */

    }

    PSC_CMD_SET(PSC_CMD_STOPINT, PSC_IIDR_ADR_SEND);          /* PSC stop command */

    PSC_CMD_SET_END();

    while ( 1 ) {

#ifdef ICX1087_NAND_USE_INTERRUPT

        /* --- INT Enable --- */

        icx1087_interrupt_done = 0;
        /* interrupt enable */
        outl( LDMA_INT_PSC, NAND_LICR );
        /* EMMU interrupt enable */
        outl( EMMU_INT_COMPEN, NAND_EICR );

#endif

        /* --- PSC --- */

        /* start PSC sequence */
        outl( PSC_START, NAND_PSCR );

#ifdef ICX1087_NAND_USE_INTERRUPT

        wait_event_timeout( icx1087_wqh_interrupt, icx1087_interrupt_done, NAND_WRITE_WAIT_TICK );
        icx1087_interrupt_done = 1;

        /* LDMA INT disable */
        outl( 0x00, NAND_LICR );
        /* EMMU INT disable */
        outl( 0x0000, NAND_EICR );

        /* wait PSC sequence end */
        j = 1;

#else

        /* wait PSC sequence end */
        j = NAND_WRITE_TIMEOUT;

#endif

        while ( ( !( inl(NAND_LIRSR) & LDMA_INT_PSC ) ) && ( !( inl(NAND_EIRSR) & EMMU_INT_COMPEN ) ) ) {
            if(--j <= 0){
                goto timeout;
            }
        }

        /* LDMA Interrupt Clear */
        outl(LDMA_INT_MASK, NAND_LICLR);
        /* EMMU interrupt clear */
        outl( EMMU_INT_COMPEN, NAND_EICLR );

        /* current instruction pointer */
        ipr = inl( NAND_IPR );
        i = ( ipr - psc_chkst_step - 1) / 3;

        if ( i < n ) {
            rc[i] = inl( NAND_FWSR ) & 1;
            /* restart PSC */
            if ( i < n - 1 ) continue;
            /* STOP PSC */
            outl( 0, NAND_PSCR );
        }

        break;
    }

    for ( i = 0, ret = 0; i < n; i++ ) ret |= rc[i];

    up( &icx1087_nand_susp_sem );

    return ret;

timeout:

    printk( KERN_NOTICE "%s: timeout!!\n", __FUNCTION__ );
    /* STOP PSC */
    outl( 0, NAND_PSCR );
    /* LDMA Interrupt Clear */
    outl(LDMA_INT_MASK, NAND_LICLR);
    /* BBC INT status clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );
    /* EMMU interrupt clear */
    outl( EMMU_INT_COMPEN, NAND_EICLR );

    for ( i = 0, ret = 0; i < n; i++ ) rc[i] = BIOS_ERR_TIMEOUT;

    up( &icx1087_nand_susp_sem );

    return BIOS_ERR_TIMEOUT;
}


/**
  * icx1087_nand_bios_erase_block - erase block ( 2 chips interleave )
  */
static int icx1087_nand_bios_erase_block2( struct mtd_info *mtd, int cn[],int page[], int rc[] )
{
    unsigned long pointer;
    unsigned int  psc_step, psc_chkst_step;
    unsigned long ipr;
    int i, j, ce[NUM_CHIP], ces=0, ret=0, n=2;

    for ( i = 0; !ret && ( i < n ); i++ ) {
        ce[i] = 1 << cn[i];
        if ( ( cn[i] < 0 ) || ( cn[i] >= NUM_CHIP ) || ( ces & ce[i] ) ) {
            ret = BIOS_ERR_PARAM;
        } else {
            ces |= ce[i];
        }
    }

    if ( ret ) {
        for ( i = 0; i < n; i++ ) rc[i] = ret;
        return ret;
    }

    for ( i = 0; i < n; i++ ) rc[i] = 0;

    down( &icx1087_nand_susp_sem );

    gpiocore_write( GPIOCORE_GP_DO0, GPIOCORE_NAND_XWP, 0x01 );

    /* --- LDMA --- */

    /* LDMA Interrupt Clear */
    outl( LDMA_INT_MASK, NAND_LICLR );

    /* --- BBC --- */

    /* AHB Interrupt Clear */
    outl( BBC_INT_MASK, NAND_INTFFCLR );

    /* --- EMMU --- */

    /* Compare Data: bit0 is 0 */
    outl( 0x0000fe00,     NAND_CDR );
    /* EMMU interrupt clear */
    outl( EMMU_INT_COMPEN, NAND_EICLR );

    /* --- PSC --- */

    PSC_CMD_SET_BEGIN();

    for ( i = 0; i < n; i++ ) {

        PSC_CMD_SET( PSC_CMD_CESEL,  ce[i] ^ NAND_CECR_MASK ); /* CEZ_OUT0 = '0' */
        PSC_CMD_SET( PSC_CMD_OUTCMD, NAND_CMD_ERASE1 );        /* erase1 command() */

        /* send address command */
        PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page[i])       & 0xff ) );
        PSC_CMD_SET( PSC_CMD_OUTADRS,   ( (page[i] >> 8)  & 0xff ) );
        PSC_CMD_SET( PSC_CMD_OUTLTADRS, ( (page[i] >> 16) & 0xff ) );

        PSC_CMD_SET( PSC_CMD_OUTCMD, NAND_CMD_ERASE2 );        /* erase1 command() */

    }

    psc_chkst_step = PSC_SEQUENCE_CUR();

    for ( i = 0; i < n; i++ ) {

        PSC_CMD_SET( PSC_CMD_CESEL,  ce[i] ^ NAND_CECR_MASK);  /* CEZ_OUT0 = '0' */
        PSC_CMD_SET(PSC_CMD_WAITRDY, ce[i]);                   /* wait ready */
        PSC_CMD_SET(PSC_CMD_CHKST, NAND_CMD_STATUS);           /* status read */

    }

    PSC_CMD_SET(PSC_CMD_STOPINT, PSC_IIDR_ADR_SEND);           /* PSC stop command */

    PSC_CMD_SET_END();

    while ( 1 ) {

#ifdef ICX1087_NAND_USE_INTERRUPT

        /* --- INT Enable --- */

        icx1087_interrupt_done = 0;
        /* interrupt enable */
        outl( LDMA_INT_PSC, NAND_LICR );
        /* EMMU interrupt enable */
        outl( EMMU_INT_COMPEN, NAND_EICR );

#endif

        /* --- PSC --- */

        /* start PSC sequence */
        outl( PSC_START, NAND_PSCR );

#ifdef ICX1087_NAND_USE_INTERRUPT

        wait_event_timeout( icx1087_wqh_interrupt, icx1087_interrupt_done, NAND_WRITE_WAIT_TICK );
        icx1087_interrupt_done = 1;

        /* LDMA INT disable */
        outl( 0x00, NAND_LICR );
        /* EMMU INT disable */
        outl( 0x0000, NAND_EICR );

        /* wait PSC sequence end */
        j = 1;

#else

        /* wait PSC sequence end */
        j = NAND_ERASE_TIMEOUT;

#endif

        while ( ( !( inl(NAND_LIRSR) & LDMA_INT_PSC ) ) && ( !( inl(NAND_EIRSR) & EMMU_INT_COMPEN ) ) ) {
            if(--j <= 0){
                goto timeout;
            }
        }

        /* LDMA Interrupt Clear */
        outl(LDMA_INT_MASK, NAND_LICLR);
        /* EMMU interrupt clear */
        outl( EMMU_INT_COMPEN, NAND_EICLR );

        /* current instruction pointer */
        ipr = inl( NAND_IPR );
        i = ( ipr - psc_chkst_step - 1) / 3;

        if ( i < n ) {
            rc[i] = inl( NAND_FWSR ) & 1;
            /* restart PSC sequence */
            if ( i < n - 1 ) continue;
            /* STOP PSC */
            outl( 0, NAND_PSCR );
        }

        break;
    }

    for ( i = 0, ret = 0; i < n; i++ ) ret |= rc[i];

    up( &icx1087_nand_susp_sem );

    return ret;

timeout:

    printk( KERN_NOTICE "%s: timeout!!\n", __FUNCTION__ );
    /* STOP PSC */
    outl( 0, NAND_PSCR );
    /* LDMA Interrupt Clear */
    outl(LDMA_INT_MASK, NAND_LICLR);
    /* EMMU interrupt clear */
    outl( EMMU_INT_COMPEN, NAND_EICLR );

    for ( i = 0, ret = 0; i < n; i++ ) rc[i] = BIOS_ERR_TIMEOUT;

    up( &icx1087_nand_susp_sem );

    return BIOS_ERR_TIMEOUT;
}


/**
  * icx1087_nand_bios_probe - driver probe
  */
static int icx1087_nand_bios_probe( struct device *non )
{
    struct mtd_info         *mtd  = &icx1087_nand_mtd;
    struct icx1087_nand_dev *this = &icx1087_nand_bios;

    printk(KERN_INFO "Starting %s.\n", NAND_NAME);

    memset( mtd,  0, sizeof(mtd)  );
    memset( this, 0, sizeof(this) );
    init_MUTEX( &icx1087_nand_susp_sem );
    icx1087_nand_initialized = 0;

   /*  */
    nand_bios_init_hw();


    /* Link the private data with the MTD structure */
    mtd->priv = this;
    this->controller = &icx1087_nand_controller;


    /* initialize I/O function pointers */

    this->chip_delay  = NAND_DELAY_US;
    this->max_chip    = NUM_CHIP;

    this->select_chip = icx1087_nand_bios_select_chip;
    this->reset       = icx1087_nand_bios_reset;
    this->read_status = icx1087_nand_bios_read_status;
    this->read_id     = icx1087_nand_bios_read_id;

    this->read_page   = icx1087_nand_bios_read_page;
    this->write_page  = icx1087_nand_bios_write_page;
    this->erase_block = icx1087_nand_bios_erase_block;

    this->read_page2   = icx1087_nand_bios_read_page2;
    this->write_page2  = icx1087_nand_bios_write_page2;
    this->erase_block2 = icx1087_nand_bios_erase_block2;

#ifdef ICX1087_NAND_USE_INTERRUPT
    init_waitqueue_head( &icx1087_wqh_interrupt );
    if ( request_irq( INT_NAND, nand_bios_interrupt, 0, "nand_bios", non ) ) {
        return -EIO;
    }
#endif

    if ( icx1087_nand_scan( mtd, this->max_chip ) ) {
        return -EIO;
    }

    /* NANDåν */
    nand_bios_init_hw_post();

    if ( icx1087_nand_register( mtd, icx1087_nand_partition_info, ARRAY_SIZE( icx1087_nand_partition_info ) ) ) {
        icx1087_nand_release ( mtd );
        return -EIO;
    }

    icx1087_nand_initialized = 1;
    return 0;
}


/**
  * icx1087_nand_bios_remove - driver remove
  */
static int __devexit icx1087_nand_bios_remove( struct device *dev )
{
    icx1087_nand_release( &icx1087_nand_mtd );

#ifdef ICX1087_NAND_USE_INTERRUPT
    free_irq( INT_NAND, dev );
#endif

    /* Auto Clock OFF */
    mp200_pmu_clkctrl_off( MP200_CLKCTRL_NAND );
    mp200_pmu_clkctrl_off( MP200_CLKCTRL_NANDPCLK );
    /* Reset */
    mp200_pmu_reset_device( MP200_RESETDEVICE_NAND );
    mp200_pmu_close_clockgate( MP200_CLOCKGATE_NAND_CLK );
    /* Clock OFF */
    mp200_pmu_close_clockgate( MP200_CLOCKGATE_NAND_PCLK );
    /* Regulator OFF */
    pwc_regulator_off( PWC_PSCNTGP3 );

    return 0;
}


#ifdef CONFIG_DPM
/**
  * icx1087_nand_bios_suspend - driver suspend
  */
static int icx1087_nand_bios_suspend( struct device *dev, u32 state, u32 level )
{
    switch( level ) {
    case SUSPEND_POWER_DOWN:
        down( &icx1087_nand_susp_sem );
        /* Regulator OFF */
        pwc_regulator_off( PWC_PSCNTGP3 );
        /* set NAND all pin to float */
        mp200_pmu_write_reg(SMU_PCTL_NAND,SMU_PCTL_NAND_VALUE_DS,SMU_PCTL_NAND_MASK);
        /* set NAND all pin to pull down */
        mp200_pmu_write_reg(SMU_PULLCTL_NAND,SMU_PULLCTL_NAND_VALUE_DS,SMU_PULLCTL_NAND_MASK);
    default:
        break;
    }
    return 0;
}


/**
  * icx1087_nand_bios_resume - driver resume
  */
static int icx1087_nand_bios_resume( struct device *dev, u32 level )
{
    int ret = 0;

    switch ( level ) {
    case RESUME_POWER_ON:
        nand_bios_init_hw();
        nand_bios_init_hw_post();
        nand_bios_select_chip( icx1087_nand_susp_ce );
        up( &icx1087_nand_susp_sem );
        break;
    default:
        break;
    }
    return ret;
}
#endif


/**
  * icx1087_nand_bios_shutdown - driver shutdown
  */
static void icx1087_nand_bios_shutdown( struct device *dev )
{
    /* Regulator OFF */
    pwc_regulator_off( PWC_PSCNTGP3 );
}


/* platform device */
void icx1087_nand_bios_release( struct device * dev ){}
static struct platform_device icx1087_nand_bios_pdev = {
    .name    = "icx1087_nand_bios",
    .dev = {
        .release = icx1087_nand_bios_release,
    },
};

/* platform driver */
static struct device_driver icx1087_nand_bios_pdrv = {
    .name    = "icx1087_nand_bios",
    .bus     = &platform_bus_type,
    .probe   = icx1087_nand_bios_probe,
    .remove  = __devexit_p( icx1087_nand_bios_remove ),
#ifdef CONFIG_DPM
    .suspend = icx1087_nand_bios_suspend,
    .resume  = icx1087_nand_bios_resume,
#endif
    .shutdown = icx1087_nand_bios_shutdown,
};


/**
  * init_icx1087_nand_bios - driver entry
  */
static int __init init_icx1087_nand_bios( void )
{
    if ( platform_device_register( &icx1087_nand_bios_pdev ) ) {
        return -1;
    }
    if ( driver_register( &icx1087_nand_bios_pdrv ) ) {
        platform_device_unregister( &icx1087_nand_bios_pdev );
        return -1;
    }
    return 0;
}


/**
  * exit_icx1087_nand_bios - driver cleanup
  */
static void __exit exit_icx1087_nand_bios( void )
{
    driver_unregister( &icx1087_nand_bios_pdrv );
    platform_device_unregister( &icx1087_nand_bios_pdev );
}


/**
  * icx1087_nand_dev *icx1087_nand_bios_get - get device structure
  */
struct icx1087_nand_dev *icx1087_nand_bios_get( void )
{
    if ( !icx1087_nand_initialized ) return NULL;
    return &icx1087_nand_bios;
}

EXPORT_SYMBOL (icx1087_nand_bios_get );

module_init( init_icx1087_nand_bios );
module_exit( exit_icx1087_nand_bios );

MODULE_LICENSE("GPL");

