/*
 *  File Name       : drivers/mtd/nand/mp200-nand-flash.c
 *  Function        : NAND Flash memory access on mp201 based devices
 *  Release Version : Ver 1.02
 *  Release Date    : 2006/06/02
 *
 *  Copyright (C) NEC Electronics Corporation 2006
 *
 *  This program is free software;you can redistribute it and/or modify itunder the terms of
 *  the GNU General Public License as published by Free Softwere Foundation;either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

/*
 * Copyright 2006,2007 Sony Corporation. 
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/device.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>

#include <asm/io.h>
#include <asm/delay.h>
#include <asm/memory.h>
#include <asm/arch/pmu.h>
#include <asm/arch/smu.h>
#include <asm/arch/gpiocore.h>
#include <asm/mach/flash.h>
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
#include <linux/dma-mapping.h>
#endif

#include "mp200_nand.h"


#define NAND_DELAY_US		15		/* 15us < wait(toshiba) . failed :-( */
#define NAND_DELAY_CWR		24		/* CMD wait after Write */
#define NAND_DELAY_BDRR		8		/* ReadTriger wait */
						/* Case:Samsun(DELAY, CWR,BDRR)=(15,22,8) ok */

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
#define NUM_CHIP		2
#else
#define NUM_CHIP		1		/* value:1-4 */
#endif
#define NAND_MTD_NAME		"mp200-nand"

/* NAND pin function */
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
#define SMU_PULLCTL_NAND_VALUE	0x00000000
#else
#define SMU_PULLCTL_NAND_VALUE	0x020000ff	/* RBZ0 not pull up */
#endif

/* MASK */
#define MASK_NAND_CECR		0x0000000F
#define MASK_NAND_FRD		0x0000FFFF
#define MASK_NAND_CACT		0x00080000
#define MASK_NAND_AHB_INT	0x00000013	/* all clear */
#define MASK_NAND_PSIZE		0x000000C0
#define MASK_NAND_PSIZE_BIT	6

#define MASK_NAND_PAGE_INT	0x00000002
#define MASK_NAND_512B_INT	0x00000001

#define MASK_NAND_MODE		0x0000ffff
#define MODE_BBC		0x0001
#define MODE_PSC		0x0002


#define MASK_RS_CHA		0x0000000F
#define MASK_RS_CALC_ENDA	0x00000008
#define MASK_RS_ERRA		0x00000002
#define MASK_RS_NOERRA		0x00000001




#define NAND_MACRO_CWR(cmd)     outl((cmd), NAND_CHA_CWR);\
                                udelay(NAND_DELAY_CWR);\
                                while( inl(NAND_FRD) & MASK_NAND_CACT )		/* CPU Bus Busy */

#define NAND_MACRO_AWR(cmd)     outl((cmd), NAND_CHA_AWR);\
                                udelay(NAND_DELAY_US);\
                                while( inl(NAND_FRD) & MASK_NAND_CACT )		/* CPU Bus Busy */

#define NAND_MACRO_LAWR(cmd)    outl((cmd), NAND_CHA_LAWR);\
                                udelay(NAND_DELAY_US);\
                                while( inl(NAND_FRD) & MASK_NAND_CACT )		/* CPU Bus Busy */


static unsigned int gNandCommand = 0xff;
static unsigned int gITR_Offset;
static unsigned int g512Count;
static unsigned int gRsLast;
static unsigned int gPageSize;
static unsigned int gErrorCorrect;
static unsigned int gColumn;
static unsigned int gPageAddr;


/*
 * MTD structure for mp200
 */
static struct mtd_info *mp200_nand_mtd = NULL;

#ifdef CONFIG_MTD_PARTITIONS
const char *part_probes[] = { "cmdlinepart", NULL, };
#endif



static struct nand_oobinfo mp200_nand_oob = {
	.useecc = MTD_NANDECC_AUTOPLACE,
	.eccbytes = 24,
	.eccpos = {  0, 1, 2, 3, 4, 5,
	            16,17,18,19,20,21,
	            32,33,34,35,36,37,
	            48,49,50,51,52,53 },
	.oobfree = {{ 6, 10 }, { 22, 10 }, { 38, 10 }, { 54, 10 }}
};

#ifdef CONFIG_JFFS2_FS
static const unsigned char clean_marker[8] = {
	0x85, 0x19, 0x03, 0x20, 0x08, 0x00, 0x00, 0x00
};
#endif



#ifdef CONFIG_DPM
static int mp200_nand_suspend(struct device *dev, u32 state, u32 level);
static int mp200_nand_resume(struct device *dev, u32 level);
#endif



/*
 * Hardware specific access to control lines.
 */
static void mp200_nand_hwcontrol(struct mtd_info *mtd, int cmd)
{
	switch(cmd) {
	case NAND_CTL_SETCLE:
	case NAND_CTL_CLRCLE:
	case NAND_CTL_SETALE:
	case NAND_CTL_CLRALE:
	case NAND_CTL_SETNCE:
	case NAND_CTL_CLRNCE:
	default:
		break;
	}
}


/*
 * Set Enable 'CEZ_OUT[0-3]'
 *  CE Control Register : 214H
 *
 *  num: 0-3
 */
static __inline void nand_enable_ce( unsigned int num )
{
	unsigned int cez;
	unsigned int value;

	value = inl(NAND_MSR) & ~(MASK_NAND_PSIZE);
	value |= gPageSize << MASK_NAND_PSIZE_BIT;
	/* mode set */
	outl( value, NAND_MSR );


	cez = (inl(NAND_CECR) & MASK_NAND_CECR );	/* 1:Inactive(High), 0:Active(Low). */
	outl( cez & ~( 0x01 << num), NAND_CECR );

	/* write protect cleared.(change to write enable) */
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	gpiocore_write( GPIOCORE_GP_DO0, GPIOCORE_GPIO_NAND_XWP, 0x01 );
#else
	gpiocore_write( GPIOCORE_GP_DO0, GPIOCORE_GPIO_P37, 0x01 );
#endif

	return;
}

/*
 * Set Disable 'CEZ_OUT[0-3]'
 *  CE Control Register : 214H
 *
 *  num: 0-3
 */
static __inline void nand_disable_ce( unsigned int num )
{
	unsigned int cez;

	cez = inl(NAND_CECR);
	outl( cez | (0x01 << num), NAND_CECR );		/* 1:Inactive(High), 0:Active(Low).*/

	/* write protected.(change to write disable) */
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	gpiocore_write( GPIOCORE_GP_DO0, GPIOCORE_GPIO_NAND_XWP, 0x00 );
#else
	gpiocore_write( GPIOCORE_GP_DO0, GPIOCORE_GPIO_P37, 0x00 );
#endif
}

static __inline void nand_disable_ce_all( void )
{
	outl( MASK_NAND_CECR, NAND_CECR );

	/* write protected.(change to write disable) */
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	gpiocore_write( GPIOCORE_GP_DO0, GPIOCORE_GPIO_NAND_XWP, 0x00 );
#else
	gpiocore_write( GPIOCORE_GP_DO0, GPIOCORE_GPIO_P37, 0x00 );
#endif
}



/*
 * BBC: Flash Read Data
 */
static __inline unsigned int read_nand(void)
{
	unsigned int frd = 0;

	outl( 0x00, NAND_CHA_BDRR );		/* ByteDataRead trigger Register */
	udelay( NAND_DELAY_BDRR );
	do {
		frd = inl( NAND_FRD );
	} while( frd & MASK_NAND_CACT );	/* CPU Bus Busy */

	return( frd & MASK_NAND_FRD );
}


/*
 * Write Setting.
 */
static int set_dma_write(struct mtd_info *mtd, int column, int page_addr)
{
	unsigned int *pITR = (unsigned int *)NAND_ITR_00H;

	/* BBC */
	/* source buffer set to NAND_ATBAR is after(in other function). */
	outl( 0x0103, NAND_INTENCLR ); 			/* [01CH] */
	outl( 0x0103, NAND_AHB_INTFFCLR);		/* [020H]*/
	outl( MASK_NAND_PAGE_INT, NAND_INTENSET );	/* page int */

	outl( 0xffff, NAND_CHA_SDH0R );
	outl( 0xffff, NAND_CHA_SDH1R );
	outl( 0xffff, NAND_CHA_SDH2R );

	outl( 0x01ff, NAND_EICLR );
	outl( 0x0100, NAND_EICR );	/* compare error */

	/* set to ITR */
	outl( 0x000A | ((inl(NAND_CECR)&0x0f)<<8), pITR ); pITR++;
	outl( 0x8010, pITR ); pITR++;		/* command(0x80) */
	outl( ((column & 0x00ff) << 8) | 0x0012, pITR ); pITR++;	/* col adr */
	outl(  (column & 0xff00)       | 0x0012, pITR ); pITR++;	/* col adr */

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	outl( ((page_addr & 0x0000ff)<<8) | 0x0012, pITR ); pITR++;	/* row adr */
	outl(  (page_addr & 0x00ff00)     | 0x0012, pITR ); pITR++;	/* row adr */
	outl( ((page_addr & 0xff0000)>>8) | 0x0014, pITR ); pITR++;	/* row adr */
	gITR_Offset = 7 * 4;
#else
	if( (mtd->size/mtd->oobblock) > 0xffff ) {
		outl( ((page_addr & 0x0000ff)<<8) | 0x0012, pITR ); pITR++;	/* row adr */
		outl(  (page_addr & 0x00ff00)     | 0x0012, pITR ); pITR++;	/* row adr */
		outl( ((page_addr & 0xff0000)>>8) | 0x0014, pITR ); pITR++;	/* row adr */
		gITR_Offset = 7 * 4;
	}else{
		outl( ((page_addr & 0x00ff)<<8) | 0x0012, pITR ); pITR++;	/* row adr */
		outl(  (page_addr & 0xff00)     | 0x0014, pITR ); pITR++;	/* row adr */
		gITR_Offset = 6 * 4;
	}
#endif

	return( 0 );
}


/*
 * 2nd phase Write Setting
 */
static void set_dma_write2(void)
{
	unsigned int *pITR;
	unsigned int rez;

	/* set 2nd cycle */
	rez = (~inl(NAND_CECR) & 0x0f) << 8;

	pITR = (unsigned int *)(NAND_ITR_00H + gITR_Offset);
	outl( 0x1010, pITR ); pITR++;		/* command(0x10) */
	outl( 0x00F2 | rez, pITR ); pITR++;
	outl( 0x00F3 | rez, pITR ); pITR++;
	outl( 0x701E, pITR ); pITR++;
	outl( 0x01D1, pITR ); pITR++;		/* Stop*/

	outl( 0x0000, pITR ); pITR++;		/* Instruction Clear */
	outl( 0x0000, pITR ); pITR++;
}


/*
 * Read Setting.
 */
static int set_dma_read(struct mtd_info *mtd, int column, int page_addr)
{
	unsigned int *pITR = (unsigned int *)NAND_ITR_00H;

	/* BBC */
	/* source buffer set to NAND_ATBAR is after(in other function). */
	outl( 0x0003, NAND_AHB_INTFFCLR );	/* [020H] */
	outl( 0x0003, NAND_INTENCLR ); 		/* [01CH] */

	/* EMMU */
	outl( 0x0000, NAND_CONTROL );
	outl( 0x10, NAND_INTENSET );	/* page,512b interrupt enable not set.*/

	outl( 0xffff, NAND_CHA_SDH0R );
	outl( 0xffff, NAND_CHA_SDH1R );
	outl( 0xffff, NAND_CHA_SDH2R );

	outl( 0x01ff, NAND_EICLR );
	outl( 0x0100, NAND_EICR );	/* [208H] compare error enable */

	/* set to ITR */
	outl( 0x000A | ((inl(NAND_CECR)&0x0f)<<8), pITR ); pITR++;
	outl( 0x0010, pITR ); pITR++;		/* read command:1st 0x00 */
	outl( ((column & 0x00ff) << 8) | 0x0012, pITR ); pITR++;	/* col adr */
	outl(  (column & 0xff00)       | 0x0012, pITR ); pITR++;	/* col adr */

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	outl( ((page_addr & 0x0000ff)<<8) | 0x0012, pITR ); pITR++;	/* row adr */
	outl(  (page_addr & 0x00ff00)     | 0x0012, pITR ); pITR++;	/* row adr */
	outl( ((page_addr & 0xff0000)>>8) | 0x0014, pITR ); pITR++;	/* row adr */
	gITR_Offset = 7 * 4;
#else
	if( (mtd->size/mtd->oobblock) > 0xffff ) {
		outl( ((page_addr & 0x0000ff)<<8) | 0x0012, pITR ); pITR++;	/* row adr */
		outl(  (page_addr & 0x00ff00)     | 0x0012, pITR ); pITR++;	/* row adr */
		outl( ((page_addr & 0xff0000)>>8) | 0x0014, pITR ); pITR++;	/* row adr */
		gITR_Offset = 7 * 4;
	}else{
		outl( ((page_addr & 0x00ff)<<8) | 0x0012, pITR ); pITR++;	/* row adr */
		outl(  (page_addr & 0xff00)     | 0x0014, pITR ); pITR++;	/* row adr */
		gITR_Offset = 6 * 4;
	}
#endif

	return( 0 );
}


/*
 * wait interrupt for write
 */
static void mp200_wait_write_interrupt(void)
{
	unsigned int int_bbc;
	unsigned int int_emmu;
	unsigned int int_ldma;

	do {
		do {
			int_bbc = inl( NAND_INTSTATUS );
			int_emmu = inl( NAND_EICR ) & inl( NAND_EIRSR );
			int_ldma = inl( NAND_LICR ) & inl( NAND_LIRSR );
		} while( ( int_bbc | int_emmu | int_ldma ) == 0 );

		if( (int_bbc & 0x02) ) {	/* page int */
			outl( 0x0002, NAND_INTENCLR);
			outl( 0x0003, NAND_AHB_INTFFCLR );
			int_bbc &= (~0x0003);		/* interrupt clear */
		}else
		if( (int_emmu & 0x0100) != 0 ){			/* compare error */
			outl( 0x0000, NAND_EICR );
			outl( 0x0100, NAND_EICLR );
			int_emmu &= ~0x0100;
			int_bbc  &= ~0x0004;
			outl( 0x0002, NAND_INTENCLR );
			outl( 0x0000, NAND_LICR );

			printk(KERN_INFO "%s : write failed.\n", __FUNCTION__);	/* bad block */
			break;
		}else{
			/*  */
			int_ldma &= ~(0x0008);		/* ldma interrupt clear. */
			outl( 0x0000, NAND_LICR );
			outl( 0x000C, NAND_LICLR );
			int_ldma &= ~(0x000C);
			outl( 0x0000, NAND_EICR );
			outl( 0x0002, NAND_INTENCLR );
			break;
		}
	}while( 1 );
}



static unsigned int rs_error_correct(int count)
{
	int i;
	unsigned int value;
	int rs_error_place;
	unsigned char *address;
	unsigned char error_value;
	unsigned int err_flag = 0;

	value = inl( NAND_RSSR );		/* [2A0H] */
	switch(value) {
	case 0x09:
		/* no error */
		break;
	case 0x0A:
		for( i = 0; i < 4; i++ ) {
			rs_error_place = inl(NAND_CHA_RSEP0R + i*4);
			if( rs_error_place > 0x000D ) {
				/* Error byte address caluculate */
				address = __va(inl(NAND_ATBAR)) + 0x200*count + 0x20D - rs_error_place;

				error_value = inl(NAND_CHA_RSEV0R + i*4);
				*address ^= error_value;
			}else
			if( rs_error_place > 0x0007 ) {
				/* error in spare data0-2 */
				/* not need change.       */
			}else
			if( (rs_error_place == 0x0000) && (i != 0) ) {
				break;
			}else{
				/* can't correct error. */
				err_flag = 1;
				continue;
			}
		}
		break;
	case 0x0E:	/* Uncorrectable error */
		err_flag = 1;
		break;
	default:
		break;
	}
	return( err_flag );
}



/*
 * wait interrupt for read
 * (spare data read is not exec in this function. becase if no error then interrupt is once, and finished.)
 */
static void mp200_wait_read_interrupt(struct mtd_info *mtd)
{
	unsigned int int_bbc;
#ifndef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	unsigned int int_emmu;
#endif
	unsigned int int_ldma;
	unsigned int count = 0;		/* 512Byte trns counter */
	unsigned int intmask = MASK_NAND_512B_INT;

	while( 1 ) {
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
		do {
			int_ldma = inl( NAND_LICR ) & inl( NAND_LIRSR );	// [60CH][610H]
		} while( int_ldma == 0 );
#else
		do {
			int_bbc = inl( NAND_INTSTATUS );			/* [010H]       */
			int_emmu = inl( NAND_EICR ) & inl( NAND_EIRSR );	/* [208H][20CH] */
			int_ldma = inl( NAND_LICR ) & inl( NAND_LIRSR );	/* [60CH][610H] */
		} while( ( int_bbc | int_emmu | int_ldma ) == 0 );
#endif

		/* PSC int */
		if( int_ldma & 0x0008 ) {
			outl( 0, NAND_LICR );
			outl( 0x002C, NAND_LICLR );
			int_ldma &= (~0x002C);
			outl( intmask, NAND_INTENSET );
			do {
				int_bbc = inl( NAND_INTSTATUS );
			} while( (int_bbc & intmask) == 0 );		/* wait page/512b interrupt */
			outl( MASK_NAND_PAGE_INT | MASK_NAND_512B_INT, NAND_INTENCLR );
			outl( MASK_NAND_PAGE_INT | MASK_NAND_512B_INT, NAND_AHB_INTFFCLR );
			int_bbc &= ~(MASK_NAND_PAGE_INT | MASK_NAND_512B_INT);

			break;
		}else
		if( inl(NAND_IPR) >= gRsLast ) {	/* instruction is Last */
			outl( 0, NAND_LICR );

			/* wait calculate end. */
			while( (inl(NAND_RSSR) & MASK_RS_CALC_ENDA) == 0 ) {	/* [2A0H] */
				;
			}
			outl( 0x01ff, NAND_LICLR );	/* clear interrupt(after RS calc) */

			outl( intmask, NAND_INTENSET );	/* [018H]page interrupt enable */
			do {
				int_bbc = inl( NAND_INTSTATUS );
			} while( (int_bbc & intmask) == 0 );		/* wait page/512b interrupt */

			if( (int_ldma & 0x80) && (inl(NAND_RSSR) != 0)) {
				count = (mtd->oobblock/512) - 1;
				if( rs_error_correct(count) > 0 ) {
					gErrorCorrect++;
				}
			}

			/* mask */
			outl( int_bbc, NAND_INTENCLR );
			/* interrupt clear */
			outl( int_bbc, NAND_AHB_INTFFCLR );

			break;
		}else{
			outl( 0, NAND_LICR );

			/* wait calculate end. */
			while( (inl(NAND_RSSR) & MASK_RS_CALC_ENDA) == 0 ) {	/* [2A0H] */
				;
			}
			outl( 0x00ac, NAND_LICLR );		/* clear all interrupt(after RS calc) */

			outl( intmask, NAND_INTENSET );		/* [018H]512b/page interrupt enable */
			do {
				int_bbc = inl( NAND_INTSTATUS );
			} while( (int_bbc & intmask) == 0 );		/* wait page/512b interrupt */

			if( int_ldma & 0x80 ) {
				if( rs_error_correct(count) > 0 ) {
					gErrorCorrect++;
				}
			}
			count++;

			outl( 0x0088, NAND_LICR );		/* RSERR, PSC */
			outl( 0x0001, NAND_PSCR );		/* PSC start */
		}
	}
}





static void psc_write_exec(void)
{
	unsigned int value;

	/* after ITR table write. */
	outl( 0, NAND_IPR );			/* Instruction Pointer Reset */
	outl( 0x00AC, NAND_LICLR );
	outl( 0x0008, NAND_LICR );

	value = inl( NAND_MSR ) & (~0x03);
	outl( value | 0x0001, NAND_MSR );
	outl( 0x0000, NAND_PSCR );
	outl( 0x0001, NAND_PSCR );		/* PSC start */

	/* wait interrupt */
	mp200_wait_write_interrupt();
}


static void psc_read_exec(struct mtd_info *mtd)
{
	/* after ITR table write. */
	outl( 0, NAND_IPR );			/* Instruction Pointer Reset */
	outl( 0x00AC, NAND_LICLR );
	outl( 0x0088, NAND_LICR );		/* RSERR, PSC */

	outl( inl( NAND_MSR ) | 0x03, NAND_MSR );
	outl( 0x0000, NAND_PSCR );
	outl( 0x0001, NAND_PSCR );		/* PSC start */

	/* wait interrupt */
	mp200_wait_read_interrupt(mtd);
}

static unsigned int mp200_nand_simple_command(unsigned int command)
{
	unsigned int ret = 0;
	int value;
	int frd;

	switch (command) {
	case NAND_CMD_STATUS:	/* 0x70 */	/* status read in this driver */
		/* status not read in mp200_nand_read16() */
		do {
			NAND_MACRO_CWR(command);
			outl( 0x00, NAND_CHA_BDRR );		/* ByteDataRead trigger Register */
			do {
				frd = inl( NAND_FRD );
			} while( frd & MASK_NAND_CACT );	/* CPU Bus Busy */
			value = frd;
		}while( !(value & NAND_STATUS_READY) );
		ret = value;
		break;
	case NAND_CMD_READID:	/* 0x90 */
		NAND_MACRO_CWR(command);
		NAND_MACRO_LAWR(0x00);
		break;
	case NAND_CMD_RESET:	/* 0xff */
		NAND_MACRO_CWR(command);
		/* check status ready/busy */
		mp200_nand_simple_command( NAND_CMD_STATUS );	/* while bus-busy loop */
		break;
	default:
		/* unsupport command */
		printk(KERN_INFO "%s : unsupport command(0x%08X).\n", __FUNCTION__, command);
		break;
	}
	return( ret );
}



static void mp200_nand_command(struct mtd_info *mtd, unsigned int command, int column, int page_addr)
{
//	struct nand_chip *this = mtd->priv;
	int i;
	unsigned char *ptr;
	unsigned char value;

	gNandCommand = command;
	gColumn = column;
	gPageAddr = page_addr;

	switch (command) {
	case NAND_CMD_READ0:	/* 0x00 */
		g512Count = 0;
		gErrorCorrect = 0;

		if( page_addr != -1 ) {
			set_dma_read(mtd, column, page_addr);
		}else{
			NAND_MACRO_CWR(command);
		}
		break;
	case NAND_CMD_STATUS:	/* 0x70 */
		NAND_MACRO_CWR(command);
		/* parameter read is in mp200_nand_read16() */
		break;
	case NAND_CMD_READID:	/* 0x90 */
		NAND_MACRO_CWR(command);
		NAND_MACRO_LAWR(0x00);
		break;
	case NAND_CMD_RESET:	/* 0xff */
		NAND_MACRO_CWR(command);
		mp200_nand_simple_command( NAND_CMD_STATUS );
		break;
	case NAND_CMD_SEQIN:	/* 0x80 */
		g512Count = 0;
		if( column >= mtd->oobblock ) {
#ifdef CONFIG_JFFS2_FS /* CONFIG_SONY_ICX1087 */ /* 2006.05.18 yam */
			/* oob write */
			NAND_MACRO_CWR(NAND_CMD_SEQIN);
			NAND_MACRO_AWR(column & 0xff);
			NAND_MACRO_AWR((column>>8) & 0xff);
			NAND_MACRO_AWR(page_addr & 0xff);
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
			NAND_MACRO_AWR((page_addr >> 8) & 0xff);
			NAND_MACRO_LAWR((page_addr >>16) & 0xff);
#else
			if( (mtd->size/mtd->oobblock) > 0xffff ) {
				NAND_MACRO_AWR((page_addr >> 8) & 0xff);
				NAND_MACRO_LAWR((page_addr >>16) & 0xff);
			}else{
				NAND_MACRO_LAWR((page_addr >>8) & 0xff);
			}
#endif
#endif
		}else{
			set_dma_write(mtd, column, page_addr);
		}
		break;
	case NAND_CMD_PAGEPROG:	/* 0x10 : 2nd cycle(for SEQIN, )*/
				/* maybe command 1st:0x80, 2nd:0x15 ... no support now */
		if( g512Count != 0 ) {
			/* write data all 0xff then not write..for RS not change. */
			ptr = __va(inl(NAND_ATBAR));
			value = 0xff;
			for( i = 0; i < mtd->oobblock; i++ ) {
				value &= *ptr++;
			}
			if( value != 0xff ) {
				set_dma_write2();	/* set 2nd cycle */
				psc_write_exec();
			}
		}else{
			/* oobblock write                            */
			/* already wrote... called write_buf timing..*/
		}
		break;
	case NAND_CMD_READOOB:	/* 0x50 */
		;	/* no code */
		break;
	case NAND_CMD_ERASE1:	/* 0x60*/
		NAND_MACRO_CWR(command);
		NAND_MACRO_AWR(page_addr&0xff);
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
		NAND_MACRO_AWR((page_addr >> 8) & 0xff);
		NAND_MACRO_LAWR((page_addr >>16) & 0xff);
#else
		if( (mtd->size/mtd->oobblock) > 0xffff ) {
			NAND_MACRO_AWR((page_addr >> 8) & 0xff);
			NAND_MACRO_LAWR((page_addr >>16) & 0xff);
		}else{
			NAND_MACRO_LAWR((page_addr>>8)&0xff);
		}
#endif
		break;
	case NAND_CMD_ERASE2:	/* 0xD0*/
		NAND_MACRO_CWR(command);
		break;
	default:
		/* unsupport command */
		printk(KERN_INFO "%s : unsupport command(0x%08X).\n", __FUNCTION__, command);
		break;
	}
}


/**
 * mp200_nand_write16
 * @mtd:	MTD device structure
 * @byte:	pointer to data byte to write
 *
 * write function for 16bit buswidth
 *
 */
static void mp200_nand_write16(struct mtd_info *mtd, u_char byte)
{
	unsigned int uiData;

	uiData = (unsigned int)byte;
	outl( uiData, NAND_CHA_BDWR );

	return;
}

static u_char mp200_nand_read16( struct mtd_info *mtd )
{
	return( read_nand() );
}


static void mp200_nand_write_word(struct mtd_info *mtd, u16 word)
{
	unsigned int uiData;

	uiData = (unsigned int)word;
	outl( uiData, NAND_CHA_BDWR );

	return;
}

static u16 mp200_nand_read_word( struct mtd_info *mtd )
{
	return( read_nand() );
}


static void mp201_nand_hw_busywait(void)
{
	const int   c_iTimeOutUs    = 8000; /* timeout us (8000us = 8ms) */
	const int   c_iCheckTimeUs  = 500;  /* Busy Check timing (500us) */
	const int   c_iCheckMax     = 4;    /* Ready Check Count         */
	unsigned int cez;

	int iTimeUs     = 0;    /* time_us */
	int iRedyCount  = 0;    /* Ready count */
	int iLoop       = 0;    /* Checked count */

	cez = ~inl(NAND_CECR) & 0x0f;
	while(iTimeUs < c_iTimeOutUs){
		/* wait(500us) */
		udelay(c_iCheckTimeUs);
		iTimeUs += c_iCheckTimeUs;

		/* check Ready */
		iRedyCount = 0;

		for(iLoop = 0; iLoop < c_iCheckMax; iLoop++){
			if( (inl(NAND_ESR) & cez) == cez){	/* signal RYBYZ_IN[0] */
				iRedyCount++;
			}
		}

		/* if Ready OK then return */
		if(iRedyCount >= c_iCheckMax){
			return;
		}
	}
}



static void init_nand_hw(void)
{
	unsigned int value;
#if 0							/* Port Setting at GPIO */ 
	uint64_t value64;
#endif

	mp200_pmu_clkctrl_off(MP200_CLKCTRL_NAND);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_NANDPCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_NAND_CLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_NAND_PCLK);
	mp200_pmu_unreset_device(MP200_RESETDEVICE_NAND);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_NAND);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_NANDPCLK);

	value = inl(SMU_PCTL_NAND);
	value &= ~(MASK_PCTL_NAND | MASK_GCTL_NAND);
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	outl(value | PCTL_SEL_NAND | PCTL_NAND_DATA_BIT, SMU_PCTL_NAND);
	mp200_gio_set_mode_output_high32( 1 << ( GPIOCORE_GPIO_NAND_XWP - 32 ) );
#else
	outl(value | PCTL_SEL_NAND | PCTL_NAND_RBZ1_BIT, SMU_PCTL_NAND);
#endif

	/* data[15-8],RBZ3-1 pullup*/
	outl(SMU_PULLCTL_NAND_VALUE, SMU_PULLCTL_NAND );

	/* Set Port OutputMode */
#if 0							/* Port Setting at GPIO */
	value64 = mp200_gio_get_mode();
	value = (value64>>32) | (1<<(GPIOCORE_GPIO_P37 - 0x20));
	mp200_gio_set_mode_output_high32(value);
#endif
}


/*
 *
 */

static int init_nand_reg(void)
{
	unsigned int value;
	unsigned int read_data[5];

	value =   (0x00 << 31)	    //[31]      Flash Bus Drive Control (1=Idleɥ饤֤)
		  | (0x00 << 28)    //[30:28]   Flash Bus WEZtCLH/tALH(max:0x07)
		  | (0x00 << 27)    //[27]      reserved
		  | (0x00 << 24)    //[26:24]   Flash Bus WEZ(REZ)tCLS/tALS(max:0x07)
		  | (0x02 << 20)    //[23:20]   Flash Bus REZtREH(max:0x0F)
		  | (0x02 << 16)    //[19:16]   Flash Bus REZtRP(max:0x0F)
		  | (0x02 << 12)    //[15:12]   Flash Bus WEZtWH(max:0x0F)
		  | (0x02 <<  8)    //[11:8]    Flash Bus WEZtWP(max:0x0F)
		  | (0x02 <<  6)    //[7:6]     Flash DevicePage Size     (2=2048Bytes)
		  | (0x00 <<  4)    //[5:4]     Flash BusΥ           (0=8bit Single)
		  | (0x00 <<  2)    //[3:2]     AHB BusΥ-ž
		  | (0x00 <<  1)    //[1]       AHB Master DMAž    (1=Flash Read)
		  | (0x00 <<  0);   //[0]       AHB Master DMA Enable       (1=Enable)


	/* --- BBC --- */
	/* mode set */
	outl( value, NAND_MSR );


	/* AHB Interrupt Clear */
	outl( MASK_NAND_AHB_INT, NAND_AHB_INTFFCLR );
	/* Interrupt Enable Clear */
	outl( MASK_NAND_AHB_INT, NAND_INTENCLR );

	/* --- EMMU --- */
	outl( 0x0000, NAND_EICR );
	/* Interrupt Clear */
	outl( 0x01ff, NAND_INTSTATUS);
	/* Interrupt Control */
	outl( 0, NAND_ATACR );	/* all disable */
	/* CE Control by enable_ce(), disable_ce() function */
	outl( 0x0f, NAND_CECR );	/* 1:Inactive(High), 0:Active(Low). start is all disable */

	/* Compare Data */
	outl( 0x0000ff00, NAND_CDR );	/* no compare.. */
	/* Status Check Cont */
	outl( 0x00000870, NAND_SCCR );/* default(0x0370) */
	/* EMMU:Control Register [0x200] */
	outl( 0x0000, NAND_CONTROL );       /* NANDǡΥåߥ(0=REZ_OUT1=) */


	/* --- PSC --- */
	/* Programmable Sequencer Control*/
	outl( 0x00000000, NAND_PSCR );

	/* --- LDMA --- */
	outl( 0x00000000, NAND_LCR );		/* Control */
	outl( 0x00000000, NAND_LICR );	/* Interrupt All Disable */
	outl( 0x000000AC, NAND_LICLR );	/* Interrupt All Clear */

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	{
		int i;
		for( i = 0; i < NUM_CHIP; i++ ) {
			nand_enable_ce(i);
			mp200_nand_simple_command( NAND_CMD_RESET );
		}
		nand_enable_ce(0);
	}
#else
	nand_enable_ce(0);				/* need for use nand command */

	mp200_nand_simple_command( NAND_CMD_STATUS );	/* wait ready */
	mp200_nand_simple_command( NAND_CMD_RESET );
	mp201_nand_hw_busywait();
#endif

	mp200_nand_simple_command( NAND_CMD_READID );
	read_data[0] = read_nand();		/* maker code */
	read_data[1] = read_nand();		/* device code */
	read_data[2] = read_nand();		/* don't care */
	read_data[3] = read_nand();


	/* page size set 0:1K->01B, 1:2K->10B, 2:4K->11B , 3:8K(no support) */
	gPageSize = ((read_data[3] + 1)& 0x03);
	if( gPageSize == 0 ) {
		nand_disable_ce(0);
		return( -EINVAL );
	}

	if( (read_data[0] == NAND_MFR_TOSHIBA) && (read_data[3] == 0xC0) ) {
		NAND_MACRO_CWR(0x91);
		read_data[4] = read_nand();	/* no use */
	}

	nand_disable_ce(0);

	return( 0 );
}



/**
 * nand select chip
 * @mtd:	MTD device structure
 * @chip:	chipnumber to select, -1 for deselect
 *
 * mp201 nand I/F is chip[0-3]
 */
static void mp200_nand_select_chip(struct mtd_info *mtd, int chip)
{
	switch(chip) {
	case -1:
		nand_disable_ce_all();
		break;
	case 0:
	case 1:
	case 2:
	case 3:
		nand_enable_ce( chip );
		break;
	default:
		BUG();
	}

	return;
}


/*
 * not use command'0x15'
 */
#ifndef CONFIG_JFFS2_FS
static void mp200_nand_pio_writeoob(struct mtd_info *mtd, const u_char *buf, int len, int page_addr)
{
	int i, j;
	int column;
	unsigned char *rptr;

	for( j = 0; j < len/0x10; j++ ) {
		column = 512 + (j * (512 + 16));

		/* oob write */
		NAND_MACRO_CWR(NAND_CMD_SEQIN);
		NAND_MACRO_AWR(column & 0xff);
		NAND_MACRO_AWR((column>>8) & 0xff);
		NAND_MACRO_AWR(page_addr & 0xff);
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
		NAND_MACRO_AWR((page_addr >> 8) & 0xff);
		NAND_MACRO_LAWR((page_addr >>16) & 0xff);
#else
		if( (mtd->size/mtd->oobblock) > 0xffff ) {
			NAND_MACRO_AWR((page_addr >> 8) & 0xff);
			NAND_MACRO_LAWR((page_addr >>16) & 0xff);
		}else{
			NAND_MACRO_LAWR((page_addr >>8) & 0xff);
		}
#endif

		rptr = (unsigned char *)&buf[j*0x10];
		for( i = 0; i < 0x10; i++ ) {
			outl( *rptr, NAND_CHA_BDWR);
			rptr++;
			udelay( NAND_DELAY_US );		/* need wait */
		}
		NAND_MACRO_CWR(NAND_CMD_PAGEPROG);
		mp200_nand_simple_command(NAND_CMD_STATUS);
	}
}
#endif





/*
 * if Dma write, Rs auto write in oob.
 */
static void mp200_nand_write_buf(struct mtd_info *mtd, const u_char *buf, int len)
{
//	struct nand_chip *this = mtd->priv;
	int i;
	unsigned int *pITR;
	unsigned int phys = __pa(buf);

	switch( gNandCommand ) {
	case NAND_CMD_SEQIN:
		if( (len == 0x200)			/* for eccmode=NAND_ECC_HW6_512 */
			|| (len == mtd->oobblock) ) {	/* for eccmode=NAND_ECC_SOFT,(Samsun:0x800) */
			if( g512Count == 0 ) {
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
				phys = dma_map_single( NULL, (void*)buf, mtd->oobblock, DMA_TO_DEVICE );
#endif
				outl( phys & 0xfffffffc, NAND_ATBAR );
			}
			pITR = (unsigned int *)(NAND_ITR_00H + gITR_Offset);
			for( i = 0; i < len/512; i++ ) {
				outl( 0x05C0, pITR ); pITR++;		/* Dma 1st 512 */
				outl( 0x01F1, pITR ); pITR++;		/* wait */
				g512Count++;
				gITR_Offset += 2 * 4;
			}
		}
#ifndef CONFIG_JFFS2_FS
		else
		if( len == mtd->oobsize ) {
			/* already wrote "RS" by h/w.                 */
			/* if g512Count == 0 then command is WRITE OOB */
			if( g512Count == 0 ) {		/* if 0 use PIO write */
				mp200_nand_pio_writeoob( mtd, buf, len, gPageAddr );
			}
		}
#endif

		break;
	default:
		printk(KERN_INFO "%s : not exec write\n", __FUNCTION__);
	}
}


/*
 *
 */

static void mp200_nand_pio_read(struct mtd_info *mtd, u_char *buf, int len, int column, int page_addr)
{
	int i;

	NAND_MACRO_CWR(NAND_CMD_READ0);
	NAND_MACRO_AWR(column & 0xff);
	NAND_MACRO_AWR((column>>8) & 0xff);
	NAND_MACRO_AWR(page_addr & 0xff);
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	NAND_MACRO_AWR((page_addr>>8)&0xff);
	NAND_MACRO_LAWR((page_addr >>16) & 0xff);
#else
	if( (mtd->size/mtd->oobblock) > 0xffff ) {
		NAND_MACRO_AWR((page_addr>>8)&0xff);
		NAND_MACRO_LAWR((page_addr >>16) & 0xff);
	}else{
		NAND_MACRO_LAWR((page_addr >>8) & 0xff);
	}
#endif
	NAND_MACRO_CWR(0x30);

	for( i = 0; i < len; i++ ) {
		*(buf+i) = read_nand();
	}
}





/*
 *
 */

static void mp200_nand_pio_readoob(struct mtd_info *mtd, u_char *buf, int page_addr, int num)
{
	int i;
	int column;

	column = 512 + (num * (512 + 16));

	NAND_MACRO_CWR(NAND_CMD_READ0);
	NAND_MACRO_AWR(column & 0xff);
	NAND_MACRO_AWR((column>>8) & 0xff);
	NAND_MACRO_AWR(page_addr & 0xff);
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	NAND_MACRO_AWR((page_addr>>8)&0xff);
	NAND_MACRO_LAWR((page_addr >>16) & 0xff);
#else
	if( (mtd->size/mtd->oobblock) > 0xffff ) {
		NAND_MACRO_AWR((page_addr>>8)&0xff);
		NAND_MACRO_LAWR((page_addr >>16) & 0xff);
	}else{
		NAND_MACRO_LAWR((page_addr >>8) & 0xff);
	}
#endif
	NAND_MACRO_CWR(0x30);

	for( i = 0; i < 0x10; i++ ) {
		*(buf+i) = read_nand();
	}
}


/*
 * ret 0 : OK (fill by 0xff).
 *     1 : NG (fill by any data).
 */

static int mp200_nand_pio_rscheck(struct mtd_info *mtd, int page_addr, int num)
{
	int i;
	int column;
	unsigned int value;

	column = 512 + (num * (512 + 16));

	NAND_MACRO_CWR(NAND_CMD_READ0);
	NAND_MACRO_AWR(column & 0xff);
	NAND_MACRO_AWR((column>>8) & 0xff);
	NAND_MACRO_AWR(page_addr & 0xff);
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
	NAND_MACRO_AWR((page_addr>>8)&0xff);
	NAND_MACRO_LAWR((page_addr >>16) & 0xff);
#else
	if( (mtd->size/mtd->oobblock) > 0xffff ) {
		NAND_MACRO_AWR((page_addr>>8)&0xff);
		NAND_MACRO_LAWR((page_addr >>16) & 0xff);
	}else{
		NAND_MACRO_LAWR((page_addr >>8) & 0xff);
	}
#endif
	NAND_MACRO_CWR(0x30);

	value = 0xff;
	for( i = 0; i < 0x10; i++ ) {
		value &= read_nand();
	}

	return( value == 0xff ? 0 : 1);
}


static void rs_2nd_check(struct mtd_info *mtd)
{
	int ret = -1;
	int i;
	int page_addr = gPageAddr;

	for( i = 0; i < mtd->oobsize/0x10; i++ ) {
		ret = mp200_nand_pio_rscheck( mtd, page_addr, i );
		if( ret ) return;
	}
	if( ret == 0 ) {
		gErrorCorrect = 0;
	}
}



static void mp200_nand_read_buf(struct mtd_info *mtd, u_char *buf, int len)
{
	int i;
	unsigned int *pITR;
	unsigned int phys = __pa(buf);
	unsigned int value;
	unsigned int rez;
	int num;

	switch( gNandCommand ) {
	case NAND_CMD_READ0:	/* 0x00 */
		rez = (~inl(NAND_CECR) & 0x0f) << 8;
		if( len == (mtd->oobblock + mtd->oobsize) ) {
			/* read data including oob into buffer */

#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
			phys = dma_map_single( NULL, (void*)buf, mtd->oobblock, DMA_FROM_DEVICE );
#endif
			outl( phys & 0xfffffffc, NAND_ATBAR );
			pITR = (unsigned int *)(NAND_ITR_00H + gITR_Offset);

			outl( 0x3010, pITR ); pITR++;		/* 2nd : 0x30 */
			outl( 0x00F2 | rez, pITR ); pITR++;		/* wait */
			outl( 0x00F3 | rez, pITR ); pITR++;		/* wait */

			for( i = 0; i < (mtd->oobblock/512); i++ ) {
				outl( 0x05C0, pITR ); pITR++;		/* Dma 512 */
				outl( 0x02F1, pITR ); pITR++;		/* wait */
			}

			gRsLast = (pITR - (unsigned int *)NAND_ITR_00H);

			outl( 0x01D1, pITR ); pITR++;		/* Stop*/
			outl( 0x0000, pITR ); pITR++;		/* Instruction Clear */
			outl( 0x0000, pITR ); pITR++;

			psc_read_exec(mtd);

			/* maybe, scan block then read column from 2048 ? */
			for( i = 0; i < mtd->oobblock/512; i++ ) {
				mp200_nand_pio_readoob(mtd, buf+mtd->oobblock+(i*0x10), gPageAddr, i);
			}
			if( gErrorCorrect > 0 ) {		/* other is badblock, or wrote block */
				value = 0xff;			/* check ff */
				for( i = 0; i < mtd->oobsize; i++ ) {
					value &= buf[mtd->oobblock+i];
				}
				if( value == 0xff ) {
					gErrorCorrect = 0;
#ifdef CONFIG_JFFS2_FS
					if( (gPageAddr & ((mtd->erasesize/mtd->oobblock)-1)) == 0 ) {
						memcpy( &buf[mtd->oobblock+6], clean_marker, sizeof(clean_marker) );
					}
#endif
				}
			}
		}else
		if( len == 0x200 ) {			/* for eccmode=NAND_ECC_HW6_512 */
			pITR = (unsigned int *)(NAND_ITR_00H + gITR_Offset);
			if( g512Count == 0 ) {
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
				phys = dma_map_single( NULL, buf, mtd->oobblock, DMA_FROM_DEVICE );
#endif
				outl( phys & 0xfffffffc, NAND_ATBAR );

				outl( 0x3010, pITR ); pITR++;		/* 2nd Cycle: 0x30 */
				outl( 0x00F2 | rez, pITR ); pITR++;		/* wait */
				outl( 0x00F3 | rez, pITR ); pITR++;		/* wait */
				gITR_Offset += 3 * 4;
			}
			outl( 0x05C0, pITR ); pITR++;		/* Dma 512 */
			outl( 0x02F1, pITR ); pITR++;		/* wait */
			gITR_Offset += 2 * 4;

			if( ((g512Count+1)*512) == mtd->oobblock ) {
				gRsLast = (pITR - (unsigned int *)NAND_ITR_00H);

				outl( 0x01D1, pITR ); pITR++;		/* Stop*/

				outl( 0x0000, pITR ); pITR++;		/* Instruction Clear */
				outl( 0x0000, pITR ); pITR++;

				psc_read_exec(mtd);
				rs_2nd_check(mtd);
			}
			g512Count++;			/* count up */

#ifdef CONFIG_JFFS2_FS /* CONFIG_SONY_ICX1087 */ /* 2006.05.18 yam */
		}else
		if( ((g512Count*512) == mtd->oobblock) && (len == mtd->oobsize) ) {
			/* read by PIO. */
			for( i = 0; i < mtd->oobblock/512; i++ ) {
				mp200_nand_pio_readoob(mtd, buf+(i*0x10), gPageAddr, i);
			}
#ifdef CONFIG_JFFS2_FS
			if( (gPageAddr & ((mtd->erasesize/mtd->oobblock)-1)) == 0 ) {
				value = 0xff;			/* check ff */
				for( i = 0; i < mtd->oobsize; i++ ) {
					value &= buf[i];
				}
				if( value == 0xff ) {
					memcpy( buf+6, clean_marker, sizeof(clean_marker) );
				}
			}
#endif
		}else
		if(len == 0x06) {	/* option = NAND_HW_ECC_SYNDOROME. */
			if( g512Count > 0 ) {
				mp200_nand_pio_read(mtd, buf, len, 512+((g512Count-1)*(512+0x10)), gPageAddr);
			}
		}else{
			if( (g512Count*512) == mtd->oobblock ) {
				/* read remain is oob area. */
				if( (mtd->oobsize - (g512Count*0x06)) == len ) {
					num = mtd->oobblock/512;		/* loop */
					value = len/num;			/* read size */
					for( i = 0; i < num; i++ ) {
						mp200_nand_pio_read(mtd, buf+(i*value), value, 512+0x06+(i*(512+0x10)), gPageAddr );
					}
				}
			}else{
				/* normal read by PIO. */
				mp200_nand_pio_read(mtd, buf, len, gColumn, gPageAddr );
			}
#endif
		}
		break;
	case NAND_CMD_READOOB:	/* 0x50 */
#ifdef CONFIG_JFFS2_FS /* CONFIG_SONY_ICX1087 */ /* 2006.05.18 yam */
		for( i = 0; i < len/0x10; i++ ) {
			mp200_nand_pio_readoob(mtd, buf+(i*0x10), gPageAddr, i);
		}
#ifdef CONFIG_JFFS2_FS
		if( (gPageAddr & ((mtd->erasesize/mtd->oobblock)-1)) == 0 ) {
			value = 0xff;
			for( i = 0; i < len; i++ ) {
				value &= buf[i];
			}
			if( value == 0xff ) {
				memcpy( buf+6, clean_marker, sizeof(clean_marker) );
			}
		}
#endif
#endif
		break;
	default:
		printk(KERN_INFO "%s : not exec read\n", __FUNCTION__);
		break;
	}

}

static int mp200_nand_verify_buf(struct mtd_info *mtd, const u_char *buf, int len)
{
	int i;
	unsigned int value;
	unsigned int column;
	static int num=0;

	if( len == 0x200 ) {
		if( g512Count == 0 ) {
			num = 0;
			g512Count++;
		}
		column = (num * (512 + 16));

		NAND_MACRO_CWR(NAND_CMD_READ0);
		NAND_MACRO_AWR(column & 0xff);
		NAND_MACRO_AWR((column>>8) & 0xff);
		NAND_MACRO_AWR(gPageAddr & 0xff);
		if( (mtd->size/mtd->oobblock) > 0xffff ) {
			NAND_MACRO_AWR((gPageAddr>>8)&0xff);
			NAND_MACRO_LAWR((gPageAddr >>16) & 0xff);
		}else{
			NAND_MACRO_LAWR((gPageAddr >>8) & 0xff);
		}
		NAND_MACRO_CWR(0x30);

		for( i = 0; i < len; i++ ) {
			value = read_nand();
			if( buf[i] != value ) {
				return( -EFAULT );
			}
		}
		num++;
	}else{
		/* len = 0x06 */
		/* buf data is ECC. can't verify */
	}

	return( 0 );
}



/*
 *
 */

static void mp200_nand_enable_hwecc(struct mtd_info *mtd, int mode )
{
	;			/* no code */
}



/**
 * nand_calculate_ecc
 * @mtd:	MTD block structure
 * @dat:	raw data
 * @ecc_code:	buffer for ECC
 */
int mp200_nand_calculate_ecc(struct mtd_info *mtd, const u_char *dat, u_char *ecc_code)
{
	/* data no use. */
	ecc_code[0] = ecc_code[1] = ecc_code[2] = 0;
	return 0;
}


/*
 *
 */
static int mp200_nand_correct_data(struct mtd_info *mtd, u_char *dat, u_char *read_ecc, u_char *calc_ecc)
{
	/* if RS Error return -1. */
	if( (g512Count*512) == mtd->oobblock ) {
		if( gErrorCorrect > 0 ) {
			return( -1 );
		}
	}

	return( 0 );
}


#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
static int mp200_nand_scan_bbt(struct mtd_info *mtd)
{
	 return( 0 );
}
static int mp200_nand_block_bad(struct mtd_info *mtd, loff_t ofs, int getchip)
{
	 return( 0 );
}
#endif





static int mp200_nand_probe(struct device *dev)
{
	int err = 0;
	struct platform_device *pdev = to_platform_device(dev);
	struct flash_platform_data *pdata = pdev->dev.platform_data;
	struct nand_chip *this;
	struct mtd_partition *parts = 0;

	init_nand_hw();
	err = init_nand_reg();
	if( err ) {
		return( err );
	}

	/* Allocate memory for MTD device structure and private data */
	mp200_nand_mtd =
	    kmalloc(sizeof(struct mtd_info) + sizeof(struct nand_chip),
		    GFP_KERNEL);
	if (mp200_nand_mtd == 0) {
		printk(KERN_WARNING
		       "Unable to allocate NAND MTD device structure.\n");
		return -ENOMEM;
	}

	/* Initialize structures */
	memset((char *)mp200_nand_mtd, 0,
	       sizeof(struct mtd_info) + sizeof(struct nand_chip));

	/* Get pointer to private data */
	this = (struct nand_chip *)(&mp200_nand_mtd[1]);

	/* Link the private data with the MTD structure */
	mp200_nand_mtd->priv = this;

	/* initialize I/O function pointers */
	this->hwcontrol = mp200_nand_hwcontrol;
	this->chip_delay = NAND_DELAY_US;		/* 15 us command delay time */
	this->eccmode = NAND_ECC_HW6_512;
	this->options = NAND_HWECC_SYNDROME;

	this->cmdfunc = mp200_nand_command;
	this->select_chip = mp200_nand_select_chip;

	this->read_byte = mp200_nand_read16;
	this->write_byte = mp200_nand_write16;
	this->read_word = mp200_nand_read_word;
	this->write_word = mp200_nand_write_word;
	this->read_buf = mp200_nand_read_buf;
	this->write_buf = mp200_nand_write_buf;
	this->verify_buf = mp200_nand_verify_buf;
	this->enable_hwecc = mp200_nand_enable_hwecc;
	this->calculate_ecc = mp200_nand_calculate_ecc;
	this->correct_data = mp200_nand_correct_data;
	this->autooob = &mp200_nand_oob;
#ifdef CONFIG_SONY_ICX1087 /* 2006.05.18 yam */
this->scan_bbt = mp200_nand_scan_bbt;
this->block_bad = mp200_nand_block_bad;
#endif

	if (nand_scan(mp200_nand_mtd, NUM_CHIP) != 0) {
		kfree( mp200_nand_mtd );
		return( -EIO );
	}


#ifdef CONFIG_MTD_PARTITIONS
	err = parse_mtd_partitions(mp200_nand_mtd, part_probes, &parts, 0);
	if (err > 0) {
		add_mtd_partitions(mp200_nand_mtd, parts, err);
	} else if (pdata->parts != 0) {
		add_mtd_partitions(mp200_nand_mtd, pdata->parts,
				   pdata->nr_parts);
	} else {
		add_mtd_device(mp200_nand_mtd);
	}
#else
	add_mtd_device(mp200_nand_mtd);
#endif

	return( 0 );
}


static int __devexit mp200_nand_remove(struct device *dev)
{
	if (mp200_nand_mtd != 0) {
		nand_release(mp200_nand_mtd);
		kfree(mp200_nand_mtd);
		mp200_nand_mtd = NULL;
	}

	return 0;
}


static struct device_driver mp200_nand_driver = {
	.name = "mp200_nand",
	.bus = &platform_bus_type,
	.probe = mp200_nand_probe,
	.remove = __devexit_p(mp200_nand_remove),
#ifdef CONFIG_DPM
	.suspend = mp200_nand_suspend,
	.resume = mp200_nand_resume,
#endif
};



#ifdef CONFIG_DPM
static int mp200_nand_suspend(struct device *dev, u32 state, u32 level)
{
	switch( level ) {
	case SUSPEND_POWER_DOWN:
		mp200_pmu_clkctrl_off(MP200_CLKCTRL_NAND);
		mp200_pmu_clkctrl_off(MP200_CLKCTRL_NANDPCLK);
		mp200_pmu_reset_device(MP200_RESETDEVICE_NAND);
		mp200_pmu_close_clockgate(MP200_CLOCKGATE_NAND_CLK);
		mp200_pmu_close_clockgate(MP200_CLOCKGATE_NAND_PCLK);
	default:
		break;
	}
	return( 0 );
}


static int mp200_nand_resume(struct device *dev, u32 level)
{
	int err = 0;

	switch (level) {
	case RESUME_POWER_ON:
		mp200_pmu_clkctrl_off(MP200_CLKCTRL_NAND);
		mp200_pmu_clkctrl_off(MP200_CLKCTRL_NANDPCLK);
		mp200_pmu_open_clockgate(MP200_CLOCKGATE_NAND_CLK);
		mp200_pmu_open_clockgate(MP200_CLOCKGATE_NAND_PCLK);
		mp200_pmu_unreset_device(MP200_RESETDEVICE_NAND);
		mp200_pmu_clkctrl_on(MP200_CLKCTRL_NAND);
		mp200_pmu_clkctrl_on(MP200_CLKCTRL_NANDPCLK);

		err = init_nand_reg();
		break;
	default:
		break;
	}
	return(err);
}
#endif


static int __init mp200_nand_init(void)
{
	return( driver_register(&mp200_nand_driver) );
}


static void __exit mp200_nand_exit(void)
{
	driver_unregister(&mp200_nand_driver);
}


#ifdef MODULE
module_init(mp200_nand_init);
module_exit(mp200_nand_exit);
#else
__initcall(mp200_nand_init);
__exitcall(mp200_nand_exit);
#endif


MODULE_LICENSE("GPL");

