/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/* 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  
 * 02111-1307, USA
 *
 */

/*!
 * @file atlas_core.c
 * @brief This is the main file for the Atlas driver. 
 *
 * It contains devfs file system 
 * and initialization of the driver.
 *
 * @ingroup ATLAS
 */

/*
 * Includes
 */

#define __INIT_TAB_DEF 

#include <asm/arch/gpio.h>

#include "atlas_register.h"
#include "atlas_event.h"
#include "atlas_external.h"

static int atlas_major;

extern void gpio_atlas_active(void);
extern void gpio_atlas_clear_int(void);

/*
 * Global variables
 */
 
DECLARE_WORK(atlas_wq_task, (void*)atlas_wq_handler, (unsigned long)0);

/*!
 * This function is called when Atlas interrupt occurs on the processor.
 * It is the interrupt handler for the Atlas module.
 *
 * @param        irq        the irq number
 * @param        dev_id     the pointer on the device
 * @param        regs       the interrupt parameters
 *
 * @return       The function returns IRQ_RETVAL(1) when handled.
 */

static irqreturn_t atlas_irq_handler(int irq, void *dev_id, 
        struct pt_regs *regs)
{
        gpio_atlas_clear_int();
        /* prepare a task */
        TRACEMSG(_K_D("* atlas irq handler top half *"));
        schedule_work(&atlas_wq_task);
        return IRQ_RETVAL(1);
}

/*!
 * This function implements IOCTL controls on a ATLAS device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @param        cmd         the command
 * @param        arg         the parameter
 * @return       This function returns 0 if successful.
 */
static int atlas_ioctl(struct inode *inode, struct file *file,
        unsigned int cmd, unsigned long arg)
{
        register_info reg_info;
#ifdef __TEST_CODE_ENABLE__
        type_event_notification event_sub;
#endif /* __TEST_CODE_ENABLE__ */

        if (_IOC_TYPE(cmd) != 'P')
                return -ENOTTY;
        
        if (copy_from_user(&reg_info, (register_info*) arg, 
                sizeof(register_info))) {
                return -EFAULT;
        }

        switch (cmd) {
        case ATLAS_READ_REG :
                TRACEMSG(_K_D("read reg"));
                CHECK_ERROR(atlas_read_reg(0,
                        reg_info.reg, &(reg_info.reg_value)));
                break;
        case ATLAS_WRITE_REG :
                TRACEMSG(_K_D("write reg %d %x"),reg_info.reg,
                        reg_info.reg_value);
                CHECK_ERROR(atlas_write_reg(0,
                        reg_info.reg, &(reg_info.reg_value)));
                break;
#ifdef __TEST_CODE_ENABLE__
        case ATLAS_SUBSCRIBE :
                TRACEMSG(_K_I("event subscribe"));
                get_event(reg_info.event,&event_sub);
                TRACEMSG(_K_I(" *** event sub %d"),event_sub.event);
                CHECK_ERROR(atlas_event_subscribe(event_sub));
                TRACEMSG(_K_I("subscribe done"));
                break;
        case ATLAS_UNSUBSCRIBE :
                TRACEMSG(_K_D("event unsubscribe"));
                get_event(reg_info.event,&event_sub);
                TRACEMSG(_K_I(" *** event unsub %d"),event_sub.event);
                CHECK_ERROR(atlas_event_unsubscribe(event_sub));
                break;
#endif /* __TEST_CODE_ENABLE__ */
        default :
                TRACEMSG(_K_D("%d unsupported ioctl command"),(int) cmd);
                return -EINVAL;
        }
        
        if (copy_to_user((register_info*) arg, &reg_info, 
                sizeof(register_info))) {
                return -EFAULT;
        }
               
        return ERROR_NONE;
}

/*!
 * This function implements the open method on a ATLAS device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @return       This function returns 0.
 */
static int atlas_open(struct inode *inode, struct file *file) 
{
        TRACEMSG(_K_D("atlas : atlas_open()"));
        return 0;
}

/*!
 * This function implements the release method on a ATLAS device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @return       This function returns 0.
 */

static int atlas_free(struct inode *inode, struct file *file)
{
        TRACEMSG(_K_D("atlas : atlas_free()"));
        return 0;
}

/*!
 * This structure defines file operations for a ATLAS device.
 */
static struct file_operations atlas_fops = {
        /*!
        * the owner
        */
        .owner = 	THIS_MODULE,
        /*!
        * the ioctl operation
        */
        .ioctl = 	atlas_ioctl,
        /*!
        * the open operation
        */
        .open = 	atlas_open,
        /*!
        * the release operation
        */
        .release = 	atlas_free,
};

/*
 * Initialization and Exit
 */

/*!
 * This function implements the init function of the ATLAS device.
 * This function is called when the module is loaded.
 * It init the interrupt handler and write initialization value of Atlas.
 *
 * @return       This function returns 0.
 */
static int __init atlas_init(void)
{
        int ret=0,i=0,rev_id=0,rev1=0, rev2=0, finid=0;
        
        atlas_major = register_chrdev(0, "atlas", &atlas_fops);
        if (atlas_major < 0) {
                TRACEMSG(_K_D("unable to get a major for atlas"));
                return atlas_major;
        }
        
        devfs_mk_cdev(MKDEV(atlas_major,0), S_IFCHR | S_IRUGO | S_IWUSR, 
                      "atlas");
        
        init_register_access_wait();
        
        /* init all register */
        TRACEMSG(_K_D("init all register value"));
        for (i=0;tab_init_reg[i][0]!=-1;i++) {
                TRACEMSG(_K_D("reg %d, value %x"),tab_init_reg[i][0],
                        tab_init_reg[i][1]);
                CHECK_ERROR(atlas_write_reg(0,tab_init_reg[i][0],
                        &tab_init_reg[i][1]));
        }
        /* install our own atlas handler */
        atlas_init_event_and_it_gpio();

        /* IOMux configuration for MX3 */
        /* initialize GPIO for Interrupt of Atlas */
        gpio_atlas_active();

        /* bind IT */

#if defined(CONFIG_ARCH_MX3)
        ret = gpio_request_irq(IT_ATLAS_PORT, GPIO_INT3, 
                GPIO_HIGH_PRIO, atlas_irq_handler, 0, 0, 0);
#endif

#if defined(CONFIG_ARCH_ARGONPLUSEVB)
        ret = request_irq(INT_EXT_INT1, atlas_irq_handler, 0, 0, 0);
#endif

#if defined(CONFIG_ARCH_SCMA11BB) || defined(CONFIG_ARCH_SCMA11EVB)
        ret = request_irq(INT_EXT_INT5, atlas_irq_handler, 0, 0, 0);
#endif

        if (ret) {
                TRACEMSG(_K_I("gpio1: irq%d error."), INT_GPIO1);
                return ret;
        }
        CHECK_ERROR(atlas_read_reg(0,REG_REVISION,&rev_id));
        rev1 = (rev_id & 0x018) >> 3;
        rev2 = (rev_id & 0x007);
        finid = (rev_id & 0x01E00) >> 9 ;
        TRACEMSG_ALL_TIME(KERN_INFO "Atlas Rev %d.%d FinVer %X detected\n",
                rev1, rev2, finid);
        if (rev1 == 0 ) {
                TRACEMSG_ALL_TIME(KERN_INFO"Atlas Core: FAILED\t!!!\tAccess failed\t!!!\n");
        } else if (rev1 == 1 ) {
#ifdef __ATLAS_2_0_
                TRACEMSG_ALL_TIME(KERN_INFO"Atlas Core: FAILED\t!!!\tBad detected version (2.x needed)\t!!!\n");
#endif
        } else if (rev1 == 2 ) {
#ifndef __ATLAS_2_0_
                TRACEMSG_ALL_TIME(KERN_INFO"Atlas Core: FAILED\t!!!\tBad detected version (1.x needed)\t!!!\n");
#endif
        } else {
                TRACEMSG_ALL_TIME(KERN_INFO"Atlas Core: successfully loaded\n");
        }        
        return 0;
}

/*!
 * This function implements the exit function of the ATLAS device.
 * This function is called when the module is unloaded.
 *
 */
static void __exit atlas_exit(void)
{
        devfs_remove("atlas");
        unregister_chrdev(atlas_major, "atlas");
        gpio_free_irq(IT_ATLAS_PORT,IT_ATLAS_BIT, GPIO_HIGH_PRIO);

        TRACEMSG_ALL_TIME(_K_I("atlas device: successfully unloaded\n"));
}

/*
 * Module entry points
 */

subsys_initcall(atlas_init);
module_exit(atlas_exit);

MODULE_DESCRIPTION("Atlas char device driver");
MODULE_AUTHOR("FreeScale");
MODULE_LICENSE("GPL");
