/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/* 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  
 * 02111-1307, USA
 *
 */

/*!
 * @file atlas_register.c
 * @brief This file contains register function of Atlas driver. 
 *
 * @ingroup ATLAS
 */

/*
 * Includes
 */
#include "atlas_register.h"
#include "atlas_spi_inter.h"
#include "atlas_config.h"

static bool pending_client[MAX_PRIO_QUEUE];
static int waiting_pending_client_cpt[MAX_PRIO_QUEUE];
static bool pending_hs_client;
static int waiting_hs_pending_client_cpt;
static DECLARE_WAIT_QUEUE_HEAD(atlas_reg_hip);
static wait_queue_head_t atlas_reg_prio[MAX_PRIO_QUEUE];

static t_atlas_status blk_mode;

/*!
 * This function initializes register access. 
 *
 * @return       This function returns 0 if successful.
 */
int init_register_access_wait(void) 
{
        /* initialize all semaphore for all clients */
        int i=0;
        for (i=0;i<MAX_PRIO_QUEUE;i++) {
                pending_client[i]=0;
                waiting_pending_client_cpt[i]=0;
                init_waitqueue_head(&atlas_reg_prio[i]);
        }
        pending_hs_client=0;
        waiting_hs_pending_client_cpt=0;
        blk_mode=ATLAS_FREE;
        CHECK_ERROR(spi_init());
        return ERROR_NONE;
}

/*!
 * This function sleeps for R/W access if the spi is busy.
 *
 * @param        priority   priority of access
 *
 * @return       This function returns 0 if successful.
 */
int register_access_wait(t_prio priority) 
{
        DEFINE_WAIT(wait);
        if (priority==0) {
                if (waiting_hs_pending_client_cpt<MAX_HS_CLIENTS) {
                        TRACEMSG(_K_D("sleep hight speed register access"));
                        pending_hs_client = true;
                        waiting_hs_pending_client_cpt++;
                        prepare_to_wait(&atlas_reg_hip, &wait, TASK_UNINTERRUPTIBLE);
                        schedule();
                        finish_wait(&atlas_reg_hip,&wait);
                        TRACEMSG(_K_D("wake up hight speed register access"));
                } else {
                        TRACEMSG(_K_D("nb of hight speed s_hs overflow...."));
                        return ERROR_CLIENT_NBOVERFLOW;
                }
        } else {
                if (waiting_pending_client_cpt[priority]<MAX_CLIENTS) {
                        TRACEMSG(_K_D("sleep register access prio %d ...."),
                                priority);
                        pending_client[priority] = true;
                        waiting_pending_client_cpt[priority]++;
                        prepare_to_wait(&atlas_reg_prio[priority], &wait, TASK_UNINTERRUPTIBLE);
                        schedule();
                        finish_wait(&atlas_reg_prio[priority],&wait);
                        TRACEMSG(_K_D("wake up register access prio %d ...."),
                                priority);
                } else {
                        return ERROR_CLIENT_NBOVERFLOW;
                }
        }
        return ERROR_NONE;
}

/*!
 * This function reads a register in high speed mode.
 *
 * @param        reg   	    register.
 * @param        reg_value   register value.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_hs_read_reg(int reg, unsigned int *reg_value) 
{    /* hight speed read  register, use by atlas protocol driver */
        return atlas_reg_access(0,true,reg,reg_value);
}

/*!
 * This function writes a register in high speed mode.
 *
 * @param        reg   	    register.
 * @param        reg_value   register value.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_hs_write_reg(int reg, unsigned int *reg_value)
{   
        /* hight speed write register, use by atlas protocol driver */
        return atlas_reg_access(0,false,reg,reg_value);
}

/*!
 * This function R/W register on Atlas.
 *
 * @param        priority   priority of access
 * @param        rw   	    Select read or write operation
 * @param        reg   	    the register
 * @param        reg_value   the register value
 *
 * @return       This function returns 0 if successful.
 */
int atlas_reg_access(t_prio priority, bool rw,int reg, unsigned int *reg_value) 
{
        if (blk_mode == ATLAS_BUSY ) {
                TRACEMSG(_K_D("spi used, waite ...."));	
                if(register_access_wait(priority)==ERROR_CLIENT_NBOVERFLOW){
                        return ERROR_CLIENT_NBOVERFLOW;					
                }
        }
        blk_mode = ATLAS_BUSY;
        if (rw==true) {
                TRACEMSG(_K_D("read reg = %d"),reg);
                CHECK_ERROR(spi_read_reg(reg, reg_value));
                TRACEMSG(_K_D("read value = 0x%x"),*reg_value);
        } else {
                TRACEMSG(_K_D("write reg = %d, value 0x%x"),reg,*reg_value);
                CHECK_ERROR(spi_write_reg(reg, reg_value));
        }
        TRACEMSG(_K_D("r/w register done"));
        CHECK_ERROR(unlock_client());
        return ERROR_NONE;
}

/*!
 * This function unlocks on locked register access.
 *
 * @return       This function returns 0 if successful.
 */
int unlock_client(void) 
{
        int i=0;
        bool queue_date = false;
        
        if(pending_hs_client==true) { 
                queue_date = true;
                TRACEMSG(_K_D("unlock hight speed"));
                waiting_hs_pending_client_cpt--;
                wake_up(&atlas_reg_hip);
                if (waiting_hs_pending_client_cpt==0) {
                        pending_hs_client=false;
                }
        } else {
                for (i=0;i<MAX_PRIO_QUEUE;i++) { 
                        if(pending_client[i]==true) { 
                                queue_date = true;
                                TRACEMSG(_K_D("unlock register prio %d"),i);
                                waiting_pending_client_cpt[i]--;
                                wake_up(&atlas_reg_prio[i]);
                                if (waiting_pending_client_cpt[i]==0) {
                                        pending_client[i]=false;
                                }
                                break;
                        }
                }
        } 
        if ( queue_date == false) {
                TRACEMSG(_K_D("no pending client release SPI"));
                blk_mode = ATLAS_FREE;
        }
        return ERROR_NONE;
}

