/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/* 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  
 * 02111-1307, USA
 *
 */

/*!
 * @file atlas_battery.c
 * @brief This is the main file of Atlas Battery driver. 
 *
 * @ingroup ATLAS_BATTERY
 */

/*
 * Includes
 */

#include "../core/component_spec.h"
#include "../core/atlas_config.h"
#include "../core/atlas_event.h"
#include "../core/atlas_external.h"
#include "atlas_battery.h"
#include "atlas_battery_defs.h"
#include "linux/device.h"

static int atlas_battery_major;
static unsigned int save_reg_stat;

/* EXPORTED FUNCTIONS */
EXPORT_SYMBOL(atlas_battery_set_out_voltage);
EXPORT_SYMBOL(atlas_battery_get_out_voltage);
EXPORT_SYMBOL(atlas_battery_set_dac_current);
EXPORT_SYMBOL(atlas_battery_get_dac_current);
EXPORT_SYMBOL(atlas_battery_set_trickle_current);
EXPORT_SYMBOL(atlas_battery_get_trickle_current);
EXPORT_SYMBOL(atlas_battery_set_out_control);
EXPORT_SYMBOL(atlas_battery_get_out_control);
EXPORT_SYMBOL(atlas_battery_set_reverse);
EXPORT_SYMBOL(atlas_battery_get_reverse);
EXPORT_SYMBOL(atlas_battery_set_threshold);
EXPORT_SYMBOL(atlas_battery_get_threshold);
EXPORT_SYMBOL(atlas_battery_set_unregulator);
EXPORT_SYMBOL(atlas_battery_get_unregulator);
EXPORT_SYMBOL(atlas_battery_set_charge_led);
EXPORT_SYMBOL(atlas_battery_get_charge_led);
EXPORT_SYMBOL(atlas_battery_set_5k_pull);
EXPORT_SYMBOL(atlas_battery_get_5k_pull);
EXPORT_SYMBOL(atlas_battery_event_sub);
EXPORT_SYMBOL(atlas_battery_event_unsub);

/*
 * Internal function
 */
int atlas_battery_event(t_bat_int event, void* callback, bool sub);

#ifdef __TEST_CODE_ENABLE__
/*!
 * This is the callback function called on TSI Atlas event, used in synchronous
 * call.
 */ 
static void callback_battery_test(void) 
{
        TRACEMSG_BATTERY(_K_I("*** Battery IT TEST Atlas ***"));
}
#endif /* __TEST_CODE_ENABLE__ */

/*!
 * This is the suspend of power management for the atlas battery API.
 * It suports SAVE and POWER_DOWN state. 
 *
 * @param        dev            the device
 * @param        state          the state
 * @param        level          the level
 *
 * @return       This function returns 0 if successful.
 */
static int atlas_battery_suspend(struct device * dev, u32 state, u32 level)
{
        unsigned int reg_value=0;
        
        switch (level) {
        case SUSPEND_NOTIFY:
                /* not supported for Atlas Battery */
                return -1;
        break;
        case SUSPEND_SAVE_STATE:
                CHECK_ERROR(atlas_read_reg(PRIO_BATTERY,
                        REG_CHARGER , &save_reg_stat));
        break;
        case SUSPEND_POWER_DOWN:
                CHECK_ERROR(atlas_write_reg(PRIO_BATTERY,REG_CHARGER,&reg_value));
        break;
        }
        
        return 0;
};

/*!
 * This is the resume of power management for the atlas battery API.
 * It suports RESTORE state. 
 *
 * @param        dev            the device
 * @param        level          the level
 *
 * @return       This function returns 0 if successful.
 */
static int atlas_battery_resume(struct device * dev, u32 level)
{
        switch (level) {
        case RESUME_RESTORE_STATE:
                CHECK_ERROR(atlas_write_reg_value(PRIO_BATTERY,
                        REG_CHARGER, save_reg_stat));
        break;
        case RESUME_POWER_ON:
                /* nothing for Atlas battery */
                return -1;
        break;
        }

        return 0;
};

static struct device_driver atlas_battery_driver_ldm = {
        .name           = "Atlas_battery",
        .bus            = &platform_bus_type,
        .probe          = NULL,
        .remove         = NULL,
        .suspend        = atlas_battery_suspend,
        .resume         = atlas_battery_resume,
};

static struct platform_device atlas_battery_ldm = {
        .name       = "Atlas_battery",
        .id         = 1,
};


/*!
 * This function sets output voltage of charge regulator.
 *
 * @param        out_voltage  	output voltage
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_set_out_voltage(t_output_voltage out_voltage)
{
        return atlas_set_reg_value(PRIO_BATTERY,REG_CHARGER,
                BITS_OUT_VOLTAGE, (int) out_voltage,LONG_OUT_VOLTAGE);
}

/*!
 * This function gets output voltage of charge regulator.
 *
 * @param        out_voltage  	output voltage
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_get_out_voltage(t_output_voltage* out_voltage)
{
        return atlas_get_reg_value(PRIO_BATTERY, REG_CHARGER,
                BITS_OUT_VOLTAGE,(int*) out_voltage,LONG_OUT_VOLTAGE);
}

/*!
 * This function sets the current of the main charger DAC.
 *
 * @param        current_val  	current of main dac.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_set_dac_current(t_dac_current current_val)
{
        return atlas_set_reg_value(PRIO_BATTERY, REG_CHARGER,
                BITS_CURRENT_MAIN, (int) current_val, LONG_CURRENT_MAIN);
}

/*!
 * This function gets the current of the main charger DAC.
 *
 * @param        current_val  	current of main dac.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_get_dac_current(t_dac_current* current_val)
{
        return atlas_get_reg_value(PRIO_BATTERY, REG_CHARGER,
                BITS_CURRENT_MAIN, (int*) current_val, LONG_CURRENT_MAIN);
}

/*!
 * This function sets the current of the trickle charger.
 *
 * @param        current_val  	current of the trickle charger.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_set_trickle_current(t_trickle_current current_val)
{
        return atlas_set_reg_value(PRIO_BATTERY, REG_CHARGER,
                BITS_CURRENT_TRICKLE, (int) current_val, LONG_CURRENT_TRICKLE);
}

/*!
 * This function gets the current of the trickle charger.
 *
 * @param        current_val  	current of the trickle charger.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_get_trickle_current(t_trickle_current *current_val)
{
        return atlas_get_reg_value(PRIO_BATTERY,REG_CHARGER,
                BITS_CURRENT_TRICKLE,(int*) current_val, LONG_CURRENT_TRICKLE);
}

/*!
 * This function sets the output controls.
 * It sets the FETOVRD and FETCTRL bits of Atlas
 *
 * @param        control  	type of control.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_set_out_control(t_control control)
{
        switch (control) {
        case CONTROL_HARDWARE:
                CHECK_ERROR(atlas_set_reg_bit(PRIO_BATTERY,REG_CHARGER, 
                        BIT_FETOVRD, 0));
        break;
        case CONTROL_BPFET_LOW:
                CHECK_ERROR(atlas_set_reg_bit(PRIO_BATTERY,REG_CHARGER, 
                        BIT_FETOVRD, 1));
                CHECK_ERROR(atlas_set_reg_bit(PRIO_BATTERY,REG_CHARGER, 
                        BIT_FETCTRL, 0));
        break;
        case CONTROL_BPFET_HIGH:
                CHECK_ERROR(atlas_set_reg_bit(PRIO_BATTERY,REG_CHARGER, 
                        BIT_FETOVRD, 1));
                CHECK_ERROR(atlas_set_reg_bit(PRIO_BATTERY,REG_CHARGER, 
                        BIT_FETCTRL, 1));
        break;
        default:
                return ERROR_BAD_INPUT;
        break;
        }
        return ERROR_NONE;
}

/*!
 * This function gets the output controls.
 * It gets the FETOVRD and FETCTRL bits of Atlas
 *
 * @param        control  	type of control.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_get_out_control(t_control* control)
{
        int ovrd=0, ctrl=0;

        CHECK_ERROR(atlas_get_reg_bit(PRIO_BATTERY, REG_CHARGER, BIT_FETOVRD, 
                &ovrd));
        CHECK_ERROR(atlas_get_reg_bit(PRIO_BATTERY, REG_CHARGER, BIT_FETCTRL, 
                &ctrl));
        
        if (ovrd == 0) {
                *control = CONTROL_HARDWARE; 
        } else {
                if (ctrl == 0) {
                        *control = CONTROL_BPFET_LOW;
                } else {
                        *control = CONTROL_BPFET_HIGH;
                } 
        }
        return ERROR_NONE;
}

/*!
 * This function sets reverse mode.
 *
 * @param        mode  	if true, reverse mode is enable.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_set_reverse(bool mode)
{
        return atlas_set_reg_bit(PRIO_BATTERY, REG_CHARGER, BIT_RVRSMODE, 
                mode);
}

/*!
 * This function gets reverse mode.
 *
 * @param        mode  	if true, reverse mode is enable.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_get_reverse(bool* mode)
{
        return atlas_get_reg_bit(PRIO_BATTERY, REG_CHARGER, BIT_RVRSMODE,   
                mode);
}

/*!
 * This function sets over voltage threshold.
 *
 * @param        threshold  	value of over voltage threshold.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_set_threshold(int threshold)
{
        if (threshold > BAT_THRESHOLD_MAX) {
                return ERROR_BAD_INPUT;
        }
        return atlas_set_reg_value(PRIO_BATTERY,REG_CHARGER,
                BITS_OVERVOLTAGE, (int) threshold,LONG_OVERVOLTAGE);
}

/*!
 * This function gets over voltage threshold.
 *
 * @param        threshold  	value of over voltage threshold.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_get_threshold(int* threshold)
{
        return atlas_get_reg_value(PRIO_BATTERY, REG_CHARGER,
                BITS_OVERVOLTAGE, (int*) threshold,LONG_OVERVOLTAGE);
}

/*!
 * This function sets unregulated charge.
 *
 * @param        unregul  	if true, the regulator charge is disabled.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_set_unregulator(bool unregul)
{
        return atlas_set_reg_bit(PRIO_BATTERY, REG_CHARGER, 
                BIT_UNREGULATED, unregul);
}

/*!
 * This function gets unregulated charge state.
 *
 * @param        unregul  	if true, the regulator charge is disabled.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_get_unregulator(bool *unregul)
{
        return atlas_get_reg_bit(PRIO_BATTERY, REG_CHARGER, 
                BIT_UNREGULATED, unregul);
}

/*!
 * This function enables/disables charge led.
 *
 * @param        led  	Enable/disable led for charge.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_set_charge_led(bool led)
{
        return atlas_set_reg_bit(PRIO_BATTERY, REG_CHARGER, BIT_CHRG_LED, 
                led);
}

/*!
 * This function returns state of charge led.
 *
 * @param        led  	state of charge led.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_get_charge_led(bool* led)
{
        return atlas_get_reg_bit(PRIO_BATTERY, REG_CHARGER, BIT_CHRG_LED, 
                led);
}

/*!
 * This function enables/disables a 5K pull down.
 *
 * @param        en_dis  	Enable/disable pull down.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_set_5k_pull(bool en_dis)
{
        return atlas_set_reg_bit(PRIO_BATTERY, REG_CHARGER, BIT_CHRGRAWPDEN, 
                en_dis);
}

/*!
 * This function returns 5K pull down configuration.
 *
 * @param        en_dis  	Enable/disable pull down.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_get_5k_pull(bool* en_dis)
{
        return atlas_get_reg_bit(PRIO_BATTERY, REG_CHARGER, BIT_CHRGRAWPDEN, 
                en_dis);
}

/*!
 * This function is used to un/subscribe on battery event IT.
 *
 * @param        event  	type of event.
 * @param        callback  	event callback function.
 * @param        sub      	define if Un/subscribe event.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_event(t_bat_int event, void* callback, bool sub)
{
        type_event_notification bat_event;
       
        atlas_event_init(&bat_event);
        bat_event.callback = callback;
        switch (event) {
        case BAT_IT_CHG_DET:
                bat_event.event = EVENT_WLOWI;
        break;
        case BAT_IT_CHG_OVERVOLT:
                bat_event.event = EVENT_CHGOVI;
        break;
        case BAT_IT_CHG_REVERSE:
                bat_event.event = EVENT_CHGREVI;
        break;
        case BAT_IT_CHG_SHORT_CIRCUIT:
                bat_event.event = EVENT_CHGSHORTI;
        break;
        case BAT_IT_CCCV:
                bat_event.event = EVENT_CCCVI;
        break;
        case BAT_IT_BELOW_THRESHOLD:
                bat_event.event = EVENT_CHRGCURRI;
        break;
        default:
                return ERROR_BAD_INPUT;
        break;
        }
        if (sub == true) {
                CHECK_ERROR(atlas_event_subscribe(bat_event));
        } else {
                CHECK_ERROR(atlas_event_unsubscribe(bat_event));
        }
        return ERROR_NONE;
}

/*!
 * This function is used to subscribe on battery event IT.
 *
 * @param        event  	type of event.
 * @param        callback  	event callback function.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_event_sub(t_bat_int event, void* callback)
{
        return atlas_battery_event(event, callback, true);
}

/*!
 * This function is used to un subscribe on battery event IT.
 *
 * @param        event  	type of event.
 * @param        callback  	event callback function.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_battery_event_unsub(t_bat_int event, void* callback)
{
        return atlas_battery_event(event, callback, false);
}

/*!
 * This function implements IOCTL controls on a Atlas battery device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @param        cmd         the command
 * @param        arg         the parameter
 * @return       This function returns 0 if successful.
 */
static int atlas_battery_ioctl( struct inode *inode, struct file *file, 
        unsigned int cmd, unsigned long arg)
{       
        int i=0,ret_val=0;
        if (_IOC_TYPE(cmd) != 'B')
                return -ENOTTY;

        switch (cmd) {
        case ATLAS_BATTERY_INIT :
                TRACEMSG_BATTERY(_K_D("Init Battery"));
                CHECK_ERROR(atlas_set_reg_value(PRIO_BATTERY, REG_CHARGER,
                0, (int) DEF_VALUE, 24));
                break;
#ifdef __TEST_CODE_ENABLE__        
        case ATLAS_BATTERY_CHECK :
                TRACEMSG_BATTERY(_K_D("Check Atlas Battery API"));

                CHECK_ERROR(atlas_battery_set_charge_led(true));
                CHECK_ERROR(atlas_battery_get_charge_led(&ret_val));
                if (ret_val != true ){
                        TRACEMSG_BATTERY(_K_I("ERROR in led config"));
                }

                for (i=OUT_VOLT_405;i<=OUT_VOLT_410;i++) {
                        CHECK_ERROR(atlas_battery_set_out_voltage(i)); 
                        CHECK_ERROR(atlas_battery_get_out_voltage(
                                (t_output_voltage*) &ret_val));        
                        if (ret_val != i) {
                                TRACEMSG_BATTERY(_K_I(
                                        "ERROR in out voltage config"));
                                return -EINVAL;
                        }
                }
                CHECK_ERROR(atlas_battery_set_out_voltage(0)); 
                for (i=TRICKLE_CURRENT_0;i<=TRICKLE_CURRENT_12;i++) {
                        CHECK_ERROR(atlas_battery_set_trickle_current(i)); 
                        CHECK_ERROR(atlas_battery_get_trickle_current(
                                (t_trickle_current*) &ret_val));        
                        if (ret_val != i) {
                                TRACEMSG_BATTERY(_K_I(
                                        "ERROR in trickle current config"));
                                return -EINVAL;
                        }
                }
                CHECK_ERROR(atlas_battery_set_trickle_current(
                        TRICKLE_CURRENT_0)); 
                for (i=CONTROL_HARDWARE; i <= CONTROL_BPFET_HIGH; i++) {
                        CHECK_ERROR(atlas_battery_set_out_control(i)); 
                        CHECK_ERROR(atlas_battery_get_out_control(
                                (t_control*) &ret_val));        
                        if (ret_val != i) {
                                TRACEMSG_BATTERY(_K_I(
                                        "ERROR in contol config"));
                                return -EINVAL;
                        }
                }
                CHECK_ERROR(atlas_battery_set_out_control(CONTROL_HARDWARE)); 

                for (i=0; i <= 3; i++) {
                        CHECK_ERROR(atlas_battery_set_threshold(i)); 
                        CHECK_ERROR(atlas_battery_get_threshold(&ret_val));    
                        if (ret_val != i) {
                                TRACEMSG_BATTERY(_K_I(
                                        "ERROR in threshold config"));
                                return -EINVAL;
                        }
                }
                CHECK_ERROR(atlas_battery_set_threshold(0));
                
                CHECK_ERROR(atlas_battery_set_unregulator(true));
                CHECK_ERROR(atlas_battery_get_unregulator(&ret_val));
                if (ret_val != true ){
                        TRACEMSG_BATTERY(_K_I("ERROR in unregulator config"));
                }

                CHECK_ERROR(atlas_battery_set_unregulator(false));
                CHECK_ERROR(atlas_battery_get_unregulator(&ret_val));
                if (ret_val != false ){
                        TRACEMSG_BATTERY(_K_I("ERROR in unregulator config"));
                }

                CHECK_ERROR(atlas_battery_set_5k_pull(true));
                CHECK_ERROR(atlas_battery_get_5k_pull(&ret_val));
                if (ret_val != false ){
                        TRACEMSG_BATTERY(_K_I("ERROR in led config"));
                }
                
                for (i=BAT_IT_CHG_DET; i <= BAT_IT_BELOW_THRESHOLD; i++) {
                        CHECK_ERROR(atlas_battery_event_sub(i, 
                                callback_battery_test)); 
                        CHECK_ERROR(atlas_battery_event_unsub(i, 
                                callback_battery_test));        
                }

                CHECK_ERROR(atlas_set_reg_value(PRIO_BATTERY, REG_CHARGER,
                0, (int) DEF_VALUE, 24));
                break;		
#endif  /* __TEST_CODE_ENABLE__ */                
        default :
                TRACEMSG_BATTERY(_K_D("%d unsupported ioctl command"),
                        (int) cmd);
                return -EINVAL;
        }
        return ERROR_NONE;
}

/*!
 * This function implements the open method on a Atlas battery device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @return       This function returns 0.
 */
static int atlas_battery_open( struct inode *inode, struct file *file) 
{
        return ERROR_NONE;
}

/*!
 * This function implements the release method on a Atlas battery device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @return       This function returns 0.
 */
static int atlas_battery_release( struct inode *inode, struct file *file) 
{
        return ERROR_NONE;
}

static struct file_operations atlas_battery_fops = {
        .owner	 =	THIS_MODULE,
        .ioctl	 =	atlas_battery_ioctl,
        .open	 =	atlas_battery_open,
        .release =	atlas_battery_release,
};

/*
 * Init and Exit
 */

static int __init atlas_battery_init(void)
{
        atlas_battery_major = register_chrdev(0, ATLAS_BATTERY_STRING, 
                &atlas_battery_fops);
        
        if (atlas_battery_major < 0) {
                TRACEMSG_BATTERY(_K_D(
                        "Unable to get a major for atlas_battery"));
                return -1;
        }
        
        devfs_mk_cdev(MKDEV(atlas_battery_major,0), S_IFCHR | S_IRUGO | 
                S_IWUSR, ATLAS_BATTERY_STRING);
        
        driver_register(&atlas_battery_driver_ldm);

        platform_device_register(&atlas_battery_ldm);

        printk(KERN_INFO "Atlas Battery loading\n");
        
        return 0;
}

static void __exit atlas_battery_exit(void)
{
        devfs_remove(ATLAS_BATTERY_STRING);

        driver_unregister(&atlas_battery_driver_ldm);

        platform_device_unregister(&atlas_battery_ldm);

        unregister_chrdev(atlas_battery_major, ATLAS_BATTERY_STRING);
        TRACEMSG_BATTERY(_K_I("Atlas_battery : successfully unloaded"));
}

/*
 * Module entry points
 */

module_init(atlas_battery_init);
module_exit(atlas_battery_exit);

MODULE_DESCRIPTION("atlas_battery driver");
MODULE_AUTHOR("FreeScale");
MODULE_LICENSE("GPL");

