/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/* 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  
 * 02111-1307, USA
 *
 */

/*!
 * @file atlas_rtc.c
 * @brief This is the main file of Atlas rtc driver. 
 *
 * @ingroup ATLAS_RTC
 */

/*
 * Includes
 */
#include "../core/atlas_external.h"
#include "../core/atlas_config.h"
#include <linux/wait.h>
#include <linux/poll.h>
#include "atlas_rtc.h"

/*
 * Global variables
 */
static int atlas_rtc_major;
static void callback_alarm_asynchronous(void);
static void callback_alarm_synchronous(void);
static unsigned int atlas_rtc_poll(struct file *file, poll_table *wait);
static DECLARE_WAIT_QUEUE_HEAD(queue_alarm);
static DECLARE_WAIT_QUEUE_HEAD(atlas_rtc_wait);
static type_event_notification alarm_event;
static bool atlas_rtc_done = 0;

/* EXPORTED FUNCTIONS */
EXPORT_SYMBOL(atlas_rtc_set_time);
EXPORT_SYMBOL(atlas_rtc_get_time);
EXPORT_SYMBOL(atlas_rtc_set_time_alarm);
EXPORT_SYMBOL(atlas_rtc_get_time_alarm);
EXPORT_SYMBOL(atlas_rtc_wait_alarm);
EXPORT_SYMBOL(atlas_rtc_event_sub);
EXPORT_SYMBOL(atlas_rtc_event_unsub);

/*
 * Real Time Clock Atlas API
 */

/*!
 * This is the callback function called on TSI Atlas event, used in asynchronous
 * call.
 */ 
static void callback_alarm_asynchronous(void) 
{
        atlas_rtc_done=true;
}

/*!
 * This is the callback function is used in test code for (un)sub.
 */ 
static void callback_test_sub(void) 
{
        TRACEMSG_ALL_TIME(_K_I("*****************************************"));
        TRACEMSG_ALL_TIME(_K_I("***** Atlas RTC 'Alarm IT CallBack' *****"));
        TRACEMSG_ALL_TIME(_K_I("*****************************************"));
}

/*!
 * This is the callback function called on TSI Atlas event, used in synchronous
 * call.
 */ 
static void callback_alarm_synchronous(void) 
{
        TRACEMSG_ALL_TIME(_K_D("*** Alarm IT Atlas ***"));
        wake_up(&queue_alarm);
}

/*!
 * This function wait the Alarm event
 *
 * @return       This function returns 0 if successful.
 */
int atlas_rtc_wait_alarm(void)
{
        DEFINE_WAIT(wait);
        atlas_event_init(&alarm_event);
        alarm_event.event = EVENT_TODAI;
        alarm_event.callback = callback_alarm_synchronous;
        CHECK_ERROR(atlas_event_subscribe(alarm_event));
        prepare_to_wait(&queue_alarm, &wait, TASK_UNINTERRUPTIBLE);
        schedule();
        finish_wait(&queue_alarm,&wait);
        CHECK_ERROR(atlas_event_unsubscribe(alarm_event));
        return ERROR_NONE;
}

/*!
 * This function set the real time clock of atlas
 *
 * @param        atlas_time  	value of date and time
 *
 * @return       This function returns 0 if successful.
 */
int atlas_rtc_set_time(struct timeval *atlas_time)
{
        unsigned int tod_reg_val=0;
        unsigned int day_reg_val=0;
        tod_reg_val = atlas_time->tv_sec % 86400;
        day_reg_val = atlas_time->tv_sec / 86400;
        
        CHECK_ERROR(atlas_write_reg_value(PRIO_RTC,REG_RTC_TIME,tod_reg_val));
        CHECK_ERROR(atlas_write_reg_value(PRIO_RTC,REG_RTC_DAY,day_reg_val));
        
        return ERROR_NONE;
}

/*!
 * This function get the real time clock of atlas
 *
 * @param        atlas_time  	return value of date and time
 *
 * @return       This function returns 0 if successful.
 */
int atlas_rtc_get_time(struct timeval *atlas_time)
{
        unsigned int tod_reg_val=0;
        unsigned int day_reg_val=0;
        
        CHECK_ERROR(atlas_read_reg(PRIO_RTC,REG_RTC_TIME,&tod_reg_val));
        CHECK_ERROR(atlas_read_reg(PRIO_RTC,REG_RTC_DAY,&day_reg_val));
        atlas_time->tv_sec =(unsigned long) ( (unsigned long) (tod_reg_val & 
                        0x0001FFFF) + (unsigned long) (day_reg_val*86400));
        return ERROR_NONE;
}

/*!
 * This function set the real time clock alarm of atlas
 *
 * @param        atlas_time  	value of date and time
 *
 * @return       This function returns 0 if successful.
 */
int atlas_rtc_set_time_alarm(struct timeval *atlas_time)
{
        unsigned int tod_reg_val=0;
        unsigned int day_reg_val=0;
        tod_reg_val = atlas_time->tv_sec % 86400;
        day_reg_val = atlas_time->tv_sec / 86400;
        
        CHECK_ERROR(atlas_write_reg_value(PRIO_RTC,REG_RTC_ALARM,tod_reg_val));
        CHECK_ERROR(atlas_write_reg_value(PRIO_RTC,REG_RTC_DAY_ALARM,
                day_reg_val));
        
        return ERROR_NONE;
}

/*!
 * This function get the real time clock alarm of atlas
 *
 * @param        atlas_time  	return value of date and time
 *
 * @return       This function returns 0 if successful.
 */
int atlas_rtc_get_time_alarm(struct timeval *atlas_time)
{
        unsigned int tod_reg_val=0;
        unsigned int day_reg_val=0;
        
        CHECK_ERROR(atlas_read_reg(PRIO_RTC,REG_RTC_ALARM,&tod_reg_val));
        CHECK_ERROR(atlas_read_reg(PRIO_RTC,REG_RTC_DAY_ALARM,&day_reg_val));
        
        atlas_time->tv_sec = (unsigned long)  ((unsigned long) (tod_reg_val & 
                        0x0001FFFF) + (unsigned long) ( day_reg_val*86400));
             
        return ERROR_NONE;
}

/*!
 * This function is used to un/subscribe on rtc event IT.
 *
 * @param        event  	type of event.
 * @param        callback  	event callback function.
 * @param        sub      	define if Un/subscribe event.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_rtc_event(t_rtc_int event, void* callback, bool sub) 
{
        type_event_notification rtc_event;
    
        atlas_event_init(&rtc_event);
        rtc_event.callback = callback;
        switch (event) {
        case RTC_IT_ALARM:
                rtc_event.event = EVENT_TODAI;
        break;
        case RTC_IT_1HZ:
                rtc_event.event = EVENT_E1HZI;
        break;
        case RTC_IT_RST:
                rtc_event.event = EVENT_RTCRSTI;
        break;
        default:
                return ERROR_BAD_INPUT;
        break;
        }
        if (sub == true) {
                CHECK_ERROR(atlas_event_subscribe(rtc_event));
        } else {
                CHECK_ERROR(atlas_event_unsubscribe(rtc_event));
        }
        return ERROR_NONE;
}

/*!
 * This function is used to subscribe on rtc event IT.
 *
 * @param        event  	type of event.
 * @param        callback  	event callback function.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_rtc_event_sub(t_rtc_int event, void* callback) 
{
        CHECK_ERROR(atlas_rtc_event(event, callback, true));
        return ERROR_NONE;

}

/*!
 * This function is used to un subscribe on rtc event IT.
 *
 * @param        event  	type of event.
 * @param        callback  	event callback function.
 *
 * @return       This function returns 0 if successful.
 */
int atlas_rtc_event_unsub(t_rtc_int event, void* callback) 
{
        CHECK_ERROR(atlas_rtc_event(event, callback, false));
        return ERROR_NONE;

}

/* Called without the kernel lock - fine */
static unsigned int atlas_rtc_poll(struct file *file, poll_table *wait)
{
        /*poll_wait(file, &atlas_rtc_wait, wait);*/

        if (atlas_rtc_done)
                return POLLIN | POLLRDNORM;
        return 0;
}

/*!
 * This function implements IOCTL controls on a Atlas rtc device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @param        cmd         the command
 * @param        arg         the parameter
 * @return       This function returns 0 if successful.
 */
static int atlas_rtc_ioctl( struct inode *inode,  struct file *file, 
        unsigned int cmd, unsigned long arg)
{
        struct timeval *atlas_time=NULL;
        
        if (_IOC_TYPE(cmd) != 'R')
                return -ENOTTY;

        if (arg) {
                if ((atlas_time = kmalloc(sizeof(struct timeval), GFP_KERNEL))
                        == NULL) {
                        return -ENOMEM;
        	}
                if (copy_from_user(atlas_time, (struct timeval*) arg, 
                        sizeof(struct timeval))) {
                        return -EFAULT;
        	}
        }
        
        switch (cmd) {
        case ATLAS_RTC_SET_TIME :
                TRACEMSG_RTC(_K_D("SET RTC"));
                CHECK_ERROR(atlas_rtc_set_time(atlas_time));
                break;		
        case ATLAS_RTC_GET_TIME :
                TRACEMSG_RTC(_K_D("GET RTC"));
                CHECK_ERROR(atlas_rtc_get_time(atlas_time));
                break;		
        case ATLAS_RTC_SET_ALARM :
                TRACEMSG_RTC(_K_D("SET RTC ALARM"));
                CHECK_ERROR(atlas_rtc_set_time_alarm(atlas_time));
                break;		
        case ATLAS_RTC_GET_ALARM :
                TRACEMSG_RTC(_K_D("GET RTC ALARM"));
                CHECK_ERROR(atlas_rtc_get_time_alarm(atlas_time));
                break;	
        case ATLAS_RTC_WAIT_ALARM :
                TRACEMSG_RTC(_K_I("WAIT ALARM..."));
                CHECK_ERROR(atlas_rtc_event_sub(RTC_IT_ALARM, 
                        callback_test_sub));
                CHECK_ERROR(atlas_rtc_wait_alarm());
                TRACEMSG_RTC(_K_I("ALARM DONE"));
                CHECK_ERROR(atlas_rtc_event_unsub(RTC_IT_ALARM, 
                        callback_test_sub));
                break;	
        case ATLAS_RTC_ALARM_REGISTER :
                TRACEMSG_RTC(_K_I("ATLAS RTC ALARM REGISTER"));
                atlas_event_init(&alarm_event);
                alarm_event.event = EVENT_TODAI;
                alarm_event.callback = callback_alarm_asynchronous;
                CHECK_ERROR(atlas_event_subscribe(alarm_event));
                break;	
        case ATLAS_RTC_ALARM_UNREGISTER :
                TRACEMSG_RTC(_K_I("ATLAS RTC ALARM UNREGISTER"));
                atlas_event_init(&alarm_event);
                alarm_event.event = EVENT_TODAI;
                alarm_event.callback = callback_alarm_asynchronous;
                CHECK_ERROR(atlas_event_unsubscribe(alarm_event));
                atlas_rtc_done=false;
                break;	
        default :
                TRACEMSG_RTC(_K_D("%d unsupported ioctl command"),(int) cmd);
                return -EINVAL;
        }

        if (arg) {
                if (copy_to_user((struct timeval*) arg, atlas_time, 
                        sizeof(struct timeval))) {
                        return -EFAULT;
        	}
                kfree(atlas_time);
        }

        return ERROR_NONE;
}

/*!
 * This function implements the open method on a Atlas rtc device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @return       This function returns 0.
 */
static int atlas_rtc_open( struct inode *inode,  struct file *file) 
{
        return ERROR_NONE;
}

/*!
 * This function implements the release method on a Atlas rtc device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @return       This function returns 0.
 */
static int atlas_rtc_release( struct inode *inode, struct file *file) 
{
        return ERROR_NONE;
}

static struct file_operations atlas_rtc_fops = {
        .owner	 =	THIS_MODULE,
        .ioctl	 =	atlas_rtc_ioctl,
 	.poll	 =      atlas_rtc_poll,
        .open	 =	atlas_rtc_open,
        .release =	atlas_rtc_release,
};

/*
 * Initialization and Exit
 */
static int __init atlas_rtc_init(void)
{
        atlas_rtc_major = register_chrdev(0, "atlas_rtc", &atlas_rtc_fops);
        
        if (atlas_rtc_major < 0) {
                TRACEMSG_RTC(_K_D("Unable to get a major for atlas_rtc"));
                return -1;
        }
        
        devfs_mk_cdev(MKDEV(atlas_rtc_major,0), S_IFCHR | S_IRUGO | S_IWUSR, 
                "atlas_rtc");
        
        printk(KERN_INFO "Atlas RTC loading\n");
        
        return 0;
}

static void __exit atlas_rtc_exit(void)
{
        devfs_remove("atlas_rtc");
        unregister_chrdev(atlas_rtc_major, "atlas_rtc");
        TRACEMSG_RTC(_K_I("Atlas_rtc : successfully unloaded"));
}

/*
 * Module entry points
 */

module_init(atlas_rtc_init);
module_exit(atlas_rtc_exit);

MODULE_DESCRIPTION("Atlas_rtc driver");
MODULE_AUTHOR("FreeScale");
MODULE_LICENSE("GPL");

