/*
 * Copyright 2004 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  
 * 02111-1307, USA
 *
 */

/*!
 * @file dam.c
 * @brief This is the brief documentation for this dam.c file.
 *
 * This file contains the implementation of the DAM driver main services
 *
 * @ingroup DAM
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include "dam.h"
#include "registers.h"

/*!
 * This include to define bool type, false and true definitions.
 */
#include <asm/hardware.h>

#ifdef TEST_DAM
#include <linux/devfs_fs_kernel.h>
#endif

#define DAM_NAME   "dam"

EXPORT_SYMBOL(dam_select_mode);
EXPORT_SYMBOL(dam_select_RxClk_direction);
EXPORT_SYMBOL(dam_select_RxClk_source);
EXPORT_SYMBOL(dam_select_RxD_source);
EXPORT_SYMBOL(dam_select_RxFS_direction);
EXPORT_SYMBOL(dam_select_RxFS_source);
EXPORT_SYMBOL(dam_select_TxClk_direction);
EXPORT_SYMBOL(dam_select_TxClk_source);
EXPORT_SYMBOL(dam_select_TxFS_direction);
EXPORT_SYMBOL(dam_select_TxFS_source);
EXPORT_SYMBOL(dam_set_internal_network_mode_mask);
EXPORT_SYMBOL(dam_set_synchronous);
EXPORT_SYMBOL(dam_switch_Tx_Rx);
EXPORT_SYMBOL(dam_reset_register);

/*!
 * DAM major
 */
#ifdef TEST_DAM
static int major_dam;

typedef struct _mxc_cfg {
        int reg;
        int val;
} mxc_cfg;
#endif

/*!
 * This function selects the operation mode of the port.
 *
 * @param        port              the DAM port to configure
 * @param        the_mode          the operation mode of the port
 *
 * @return       This function returns the result of the operation 
 *               (0 if successful, -1 otherwise).
 */
int dam_select_mode(dam_port port, dam_mode the_mode)
{
        int result;
        result = 0;

        ModifyRegister32(dam_mode_masq << dam_mode_shift,
                         the_mode << dam_mode_shift,
                         _reg_DAM_PDCR(port));

        return result;
}

/*!
 * This function controls Receive clock signal direction for the port.
 *
 * @param        port              the DAM port to configure
 * @param        direction         the Rx clock signal direction
 */
void dam_select_RxClk_direction(dam_port port, signal_direction direction)
{
        ModifyRegister32(1  <<  dam_receive_clock_direction_shift,
                         direction  <<  dam_receive_clock_direction_shift,
                         _reg_DAM_PTCR(port));
}

/*!
 * This function controls Receive clock signal source for the port.
 *
 * @param        p_config          the DAM port to configure
 * @param        from_RxClk        the signal comes from RxClk or TxClk of 
 *                                 the source port
 * @param        p_source          the source port
 */
void dam_select_RxClk_source(dam_port p_config,
                             bool from_RxClk,
                             dam_port p_source)
{
        ModifyRegister32(dam_selection_mask << dam_receive_clock_select_shift,
                         ((from_RxClk << 3)|p_source) <<
                          dam_receive_clock_select_shift,
                         _reg_DAM_PTCR(p_config));
}

/*!
 * This function selects the source port for the RxD data.
 *
 * @param        p_config          the DAM port to configure
 * @param        p_source          the source port
 */
void dam_select_RxD_source(dam_port p_config, dam_port p_source)
{
        ModifyRegister32(dam_selection_mask << dam_receive_data_select_shift,
                         p_source << dam_receive_data_select_shift,
                         _reg_DAM_PDCR(p_config));
}

/*!
 * This function controls Receive Frame Sync signal direction for the port.
 *
 * @param        port              the DAM port to configure
 * @param        direction         the Rx Frame Sync signal direction
 */
void dam_select_RxFS_direction(dam_port port, signal_direction direction)
{
        ModifyRegister32(1 << dam_receive_frame_sync_direction_shift,
                         direction << dam_receive_frame_sync_direction_shift,
                         _reg_DAM_PTCR(port));
}

/*!
 * This function controls Receive Frame Sync signal source for the port.
 *
 * @param        p_config          the DAM port to configure
 * @param        from_RxFS         the signal comes from RxFS or TxFS of 
 *                                 the source port
 * @param        p_source          the source port
 */
void dam_select_RxFS_source(dam_port p_config,
                            bool from_RxFS,
                            dam_port p_source)
{
        ModifyRegister32(dam_selection_mask<<
                         dam_receive_frame_sync_select_shift,
                         ((from_RxFS << 3)|p_source) <<
                         dam_receive_frame_sync_select_shift,
                         _reg_DAM_PTCR(p_config));
}

/*!
 * This function controls Transmit clock signal direction for the port.
 *
 * @param        port              the DAM port to configure
 * @param        direction         the Tx clock signal direction
 */
void dam_select_TxClk_direction(dam_port port, signal_direction direction)
{
        ModifyRegister32(1 << dam_transmit_clock_direction_shift,
                         direction << dam_transmit_clock_direction_shift,
                         _reg_DAM_PTCR(port));
}

/*!
 * This function controls Transmit clock signal source for the port.
 *
 * @param        p_config          the DAM port to configure
 * @param        from_RxClk        the signal comes from RxClk or TxClk of 
 *                                 the source port
 * @param        p_source          the source port
 */
void dam_select_TxClk_source(dam_port p_config,
                             bool from_RxClk,
                             dam_port p_source)
{
        ModifyRegister32(dam_selection_mask<<dam_transmit_clock_select_shift,
                         ((from_RxClk << 3)|p_source) <<
                         dam_transmit_clock_select_shift,
                         _reg_DAM_PTCR(p_config));
}

/*!
 * This function controls Transmit Frame Sync signal direction for the port.
 *
 * @param        port              the DAM port to configure
 * @param        direction         the Tx Frame Sync signal direction
 */
void dam_select_TxFS_direction(dam_port port, signal_direction direction)
{
        ModifyRegister32(1 << dam_transmit_frame_sync_direction_shift,
                         direction << dam_transmit_frame_sync_direction_shift,
                         _reg_DAM_PTCR(port));
}

/*!
 * This function controls Transmit Frame Sync signal source for the port.
 *
 * @param        p_config          the DAM port to configure
 * @param        from_RxFS         the signal comes from RxFS or TxFS of 
 *                                 the source port
 * @param        p_source          the source port
 */
void dam_select_TxFS_source(dam_port p_config,
                            bool from_RxFS,
                            dam_port p_source)
{
        ModifyRegister32(dam_selection_mask <<
                         dam_transmit_frame_sync_select_shift,
                         ((from_RxFS << 3)|p_source) <<
                         dam_transmit_frame_sync_select_shift,
                         _reg_DAM_PTCR(p_config));
}

/*!
 * This function sets a bit mask that selects the port from which of the RxD
 * signals are to be ANDed together for internal network mode.
 * Bit 6 represents RxD from Port7 and bit0 represents RxD from Port1.
 * 1 excludes RxDn from ANDing. 0 includes RxDn for ANDing.
 *
 * @param        port              the DAM port to configure
 * @param        bit_mask          the bit mask
 *
 * @return       This function returns the result of the operation 
 *               (0 if successful, -1 otherwise).
 */
int dam_set_internal_network_mode_mask( dam_port port, unsigned char bit_mask)
{
        int result;
        result = 0;

        ModifyRegister32(dam_internal_network_mode_mask << 
                                 dam_internal_network_mode_shift,
                         bit_mask << dam_internal_network_mode_shift,
                         _reg_DAM_PDCR(port));

        return result;
}

/*!
 * This function controls whether or not the port is in synchronous mode.
 * When the synchronous mode is selected, the receive and the transmit sections
 * use common clock and frame sync signals.
 * When the synchronous mode is not selected, separate clock and frame sync 
 * signals are used for the transmit and the receive sections.
 * The defaut value is the synchronous mode selected.
 *
 * @param        port              the DAM port to configure
 * @param        synchronous       the state to assign
 */
void dam_set_synchronous(dam_port port, bool synchronous)
{
        ModifyRegister32(1 << dam_synchronous_mode_shift,
                         synchronous << dam_synchronous_mode_shift,
                         _reg_DAM_PTCR(port));
}

/*!
 * This function swaps the transmit and receive signals from (Da-TxD, Db-RxD) 
 * to (Da-RxD, Db-TxD).
 * This default signal configuration is Da-TxD, Db-RxD.
 *
 * @param        port              the DAM port to configure
 * @param        value             the switch state
 */
void dam_switch_Tx_Rx(dam_port port, bool value)
{
        ModifyRegister32(1 << dam_transmit_receive_switch_shift,
                         value << dam_transmit_receive_switch_shift,
                         _reg_DAM_PDCR(port));
}

/*!
 * This function resets the two registers of the selected port.
 *
 * @param        port              the DAM port to reset
 */
void dam_reset_register(dam_port port)
{
        ModifyRegister32(0xFFFFFFFF, 0x00000000,
                         _reg_DAM_PTCR(port));
        ModifyRegister32(0xFFFFFFFF, 0x00000000,
                         _reg_DAM_PDCR(port));
}

#ifdef TEST_DAM

#ifdef CONFIG_ARCH_SCMA11EVB
#define SCMA11_EVB_PASS1 1
/* See scma11evb_gpio.c to complete the workaround for SSI scma11 pass1 */
#endif
/*!
 * This function implements IOCTL controls on a DAM device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @param        cmd         the command
 * @param        arg         the parameter :\n
 * DAM_CONFIG_SSI1_ATLAS:\n
 * data from port 1 to port 4, clock and FS from port 1 (SSI1)\n
 * DAM_CONFIG_SSI2_ATLAS:\n
 * data from port 2 to port 5, clock and FS from port 2 (SSI2)\n
 * DAM_CONFIG_SSI_NETWORK_MODE_ATLAS:\n
 * network mode for mix digital with data from port 1 to port4,\n
 * data from port 2 to port 4, clock and FS from port 1 (SSI1)
 *
 * @return       This function returns 0 if successful.
 */
static int dam_ioctl(struct inode *inode,
                     struct file *file,
                     unsigned int cmd,
                     unsigned long arg)
{
        int result = 0;

        if (_IOC_TYPE(cmd) != DAM_IOCTL)
                return -ENOTTY;

        switch (cmd) {
        case DAM_CONFIG_SSI1_ATLAS:
                dam_reset_register(port_1);
                dam_reset_register(port_4);

                dam_select_mode(port_1, normal_mode);
                dam_select_mode(port_4, normal_mode);
                
		dam_set_synchronous(port_1, true);
                dam_set_synchronous(port_4, true);
                
		dam_select_RxD_source(port_4, port_1);
                dam_select_RxD_source(port_1, port_4);
                
		dam_select_TxFS_direction(port_4, signal_out);
                dam_select_TxFS_source(port_4, false, port_1);

#ifdef SCMA11_EVB_PASS1
                dam_reset_register(port_7);
                dam_select_mode(port_7, normal_mode);
		dam_set_synchronous(port_7, false);
        	dam_select_RxClk_source(port_7, false, port_1);
                dam_select_RxClk_direction(port_7, signal_out);
#else

                dam_select_TxClk_direction(port_4, signal_out);
                dam_select_TxClk_source(port_4, false, port_1);
#endif
                break;

        case DAM_CONFIG_SSI2_ATLAS:
                dam_reset_register(port_2);
                dam_reset_register(port_5);

                dam_set_synchronous(port_2, true);
		
                dam_select_RxD_source(port_2, port_5);
                dam_select_RxD_source(port_5, port_2);

                dam_select_TxFS_direction(port_5, signal_out);
                dam_select_TxFS_source(port_5, false, port_2);

#ifdef SCMA11_EVB_PASS1
		dam_set_synchronous(port_5, false);
        	dam_select_RxClk_source(port_5, false, port_2);
                dam_select_RxClk_direction(port_5, signal_out);
#else
                dam_set_synchronous(port_5, true);
                dam_select_TxClk_direction(port_5, signal_out);
                dam_select_TxClk_source(port_5, false, port_2);
#endif
                break;

        case DAM_CONFIG_SSI_NETWORK_MODE_ATLAS: 
                dam_reset_register(port_1);
                dam_reset_register(port_2);
                dam_reset_register(port_4);
                dam_reset_register(port_5);

                dam_set_synchronous(port_1, true);
                dam_select_RxD_source(port_1, port_4);

                dam_set_synchronous(port_2, true);
                dam_select_RxD_source(port_2, port_4);
                dam_select_TxClk_direction(port_2, signal_out);
                dam_select_TxFS_direction(port_2, signal_out);
                dam_select_TxClk_source(port_2, false, port_1);
                dam_select_TxFS_source(port_2, false, port_1);

                dam_set_synchronous(port_4, true);
                dam_select_TxClk_direction(port_4, signal_out);
                dam_select_TxFS_direction(port_4, signal_out);
                dam_select_TxClk_source(port_4, false, port_1);
                dam_select_TxFS_source(port_4, false, port_1);
                dam_select_mode(port_4, internal_network_mode);
                dam_set_internal_network_mode_mask(port_4, 0xFC);
                break;

        default:
                break;
        }

        return result;
}

/*!
 * This function implements the open method on a DAM device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 *
 * @return       This function returns 0.
 */
static int dam_open(struct inode *inode, struct file *file)
{
        /* DBG_PRINTK("ssi : dam_open()\n"); */
        return 0;
}

/*!
 * This function implements the release method on a DAM device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 *
 * @return       This function returns 0.
 */
static int dam_free(struct inode *inode, struct file *file)
{
        /* DBG_PRINTK("ssi : dam_free()\n"); */
        return 0;
}

/*!
 * This structure defines file operations for a DAM device.
 */
static struct file_operations dam_fops = {

      /*!
       * the owner
       */
        .owner =        THIS_MODULE,

      /*!
       * the ioctl operation
       */
        .ioctl =        dam_ioctl,

      /*!
       * the open operation
       */
        .open =         dam_open,

      /*!
       * the release operation
       */
        .release =      dam_free,
};

#endif

/*!
 * This function implements the init function of the DAM device.
 * This function is called when the module is loaded.
 *
 * @return       This function returns 0.
 */
static int __init dam_init(void)
{
#ifdef TEST_DAM
        printk(KERN_DEBUG "dam : dam_init(void) \n");

        major_dam = register_chrdev(0,DAM_NAME, &dam_fops);
        if (major_dam < 0)
        {
                printk(KERN_WARNING "Unable to get a major for dam");
                return major_dam;
        }

        printk("dam : creating devfs entry for DAM \n");
        devfs_mk_cdev(MKDEV(major_dam,0), 
                      S_IFCHR | S_IRUGO | S_IWUSR, DAM_NAME );
#endif
        return 0;
}

/*!
 * This function implements the exit function of the SPI device.
 * This function is called when the module is unloaded.
 *
 */
static void __exit dam_exit(void)
{
#ifdef TEST_DAM
        devfs_remove(DAM_NAME);
        unregister_chrdev(major_dam, DAM_NAME);
        printk(KERN_DEBUG "dam : successfully unloaded\n");
#endif
}

module_init(dam_init);
module_exit(dam_exit);

MODULE_DESCRIPTION("DAM char device driver");
MODULE_AUTHOR("Freescale Semiconductor, Inc.");
MODULE_LICENSE("GPL");

