/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Driver for the Freescale Semiconductor MXC IPU
 *
 */

/*!
 * @file mt9v111.c
 * 
 * @brief mt9v111 camera driver functions
 * 
 * @ingroup Camera
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/ctype.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/devfs_fs_kernel.h>

#include "mt9v111.h"
#include "asm/arch/mxc_i2c.h"
#include "asm/arch/gpio.h"
#include "../drivers/mxc/ipu/ipu.h"
#include "../drivers/mxc/ipu/ipu_prv.h"
#include "../drivers/media/video/mxc_v4l2_capture.h"

static mt9v111_conf mt9v111_device;
static int gMajor_mt9v111 = 0;
static u16 testpattern = 0;
static u32 p_mclk_freq = 0x1B00000; 

/*
 * Function definitions
 */

static u16
mt9v111_endian_swap16(u16 data)
{
        u16 temp;

        temp = data;
        temp = ((data >> 8) & 0xff) | ((data << 8) & 0xff00);

        return temp; 
}

/*!
 * Initialize mt9v111_sensor_lib
 * Libarary for Sensor configuration through I2C
 *
 * @param       coreReg       Core Registers
 * @param       ifpReg        IFP Register
 *
 * @return status
 */
static u8 
mt9v111_sensor_lib(mt9v111_coreReg *coreReg, mt9v111_IFPReg *ifpReg)
{
        u8 reg[1];
        u32 data;
        u8 error = 0;

        FUNC_START;
        /* 
         * setup to IFP registers
         */
        reg[0] = MT9V111I_ADDR_SPACE_SEL;
        data = ifpReg->addrSpaceSel;
        data = mt9v111_endian_swap16(data); 
        DPRINTK("mt9v111_sensor_lib: write MT9V111I_ADDR_SPACE_SEL %x.\n",data);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);
        data = 0; 
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111I_ADDR_SPACE_SEL %x.\n", data);

        // Operation Mode Control
        reg[0] = MT9V111I_MODE_CONTROL;
        data = ifpReg->modeControl;
        data = mt9v111_endian_swap16(data); 
        DPRINTK("mt9v111_sensor_lib: write MT9V111I_MODE_CONTROL %x.\n", data);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);
        data = 0; 
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111I_MODE_CONTROL %x.\n", data);

        // Output format
        reg[0] = MT9V111I_FORMAT_CONTROL;
        data = ifpReg->formatControl;     // Set bit 12
        data = mt9v111_endian_swap16(data); 
        DPRINTK("mt9v111_sensor_lib: write MT9V111I_FORMAT_CONTROL %x\n",data);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);
        data = 0; 
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111I_FORMAT_CONTROL %x.\n", data);

        // AE limit 4
        reg[0] = MT9V111I_SHUTTER_WIDTH_LIMIT_AE;
        data = ifpReg->gainLimitAE;
        data = mt9v111_endian_swap16(data); 
        DPRINTK("mt9v111_sensor_lib: write " 
                 "MT9V111I_SHUTTER_WIDTH_LIMIT_AE %x\n",data);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);
        data = 0; 
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111I_SHUTTER_WIDTH_LIMIT_AE %x.\n",
                 data);

        reg[0] = MT9V111I_OUTPUT_FORMAT_CTRL2;
        data = ifpReg->outputFormatCtrl2;
        data = mt9v111_endian_swap16(data); 
        DPRINTK("mt9v111_sensor_lib: write MT9V111I_FORMAT_CONTROL2 %x\n",data);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);
        data = 0; 
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111I_FORMAT_CONTROL2 %x.\n", data);

        reg[0] = MT9V111I_AE_SPEED;
        data = ifpReg->AESpeed;
        data = mt9v111_endian_swap16(data); 
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

        /* 
         * setup to sensor core registers
         */
        // switch to sensor core registers space
        reg[0] = MT9V111I_ADDR_SPACE_SEL;
        data = coreReg->addressSelect;
        data = mt9v111_endian_swap16(data); 
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

        // enable changes and put the Sync bit on
        reg[0] = MT9V111S_OUTPUT_CTRL;
        data = MT9V111S_OUTCTRL_SYNC | MT9V111S_OUTCTRL_CHIP_ENABLE | 0x3000;
        data = mt9v111_endian_swap16(data); 
        DPRINTK("mt9v111_sensor_lib: write MT9V111S_OUTPUT_CTRL %x.\n", data);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);
        data = 0;
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111S_OUTPUT_CTRL %x.\n", data);

        // min PIXCLK - Default
        reg[0] = MT9V111S_PIXEL_CLOCK_SPEED;
        data = coreReg->pixelClockSpeed;
        DPRINTK("mt9v111_sensor_lib: write MT9V111S_PIXEL_CLOCK_SPEED %x.\n", 
                data);
        data = mt9v111_endian_swap16(data); 
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

        //Setup image flipping / Dark rows / row/column skip
        reg[0] = MT9V111S_READ_MODE;
        data = coreReg->readMode;
        data = mt9v111_endian_swap16(data);
        DPRINTK("mt9v111_sensor_lib: write MT9V111S_READ_MODE %x.\n", data);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

     	//zoom 0
        reg[0] = MT9V111S_DIGITAL_ZOOM;
        data = coreReg->digitalZoom;
        data = mt9v111_endian_swap16(data); 
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

        // min H-blank
        reg[0] = MT9V111S_HOR_BLANKING;
        data = coreReg->horizontalBlanking;
        data = mt9v111_endian_swap16(data); 
        DPRINTK("mt9v111_sensor_lib: write MT9V111S_HOR_BLANKING %x.\n", data);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

        // min V-blank
        reg[0] = MT9V111S_VER_BLANKING;
        data = coreReg->verticalBlanking;
        data = mt9v111_endian_swap16(data); 
        DPRINTK("mt9v111_sensor_lib: write MT9V111S_VER_BLANKING %x.\n", data);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

        // MT9V111S_SHUTTER_WIDTH
        reg[0] = MT9V111S_VER_BLANKING;
        data = coreReg->shutterWidth;
        data = mt9v111_endian_swap16(data); 
        DPRINTK("mt9v111_sensor_lib: write MT9V111S_VER_BLANKING %x.\n", data);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

        // changes become effective 
        reg[0] = MT9V111S_OUTPUT_CTRL;
        data = MT9V111S_OUTCTRL_CHIP_ENABLE | 0x3000;
        data = mt9v111_endian_swap16(data); 
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

        msleep(1000);

        data = 0;
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111S_OUTPUT_CTRL %x.\n", data);
 
        reg[0] = MT9V111S_HOR_BLANKING;
        data = 0;
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111S_HOR_BLANKING %x.\n", data);

        // Debug
        reg[0] = MT9V111S_VER_BLANKING;
        data = 0;
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111S_VER_BLANKING %x.\n", data);

        // Debug
        reg[0] = MT9V111S_READ_MODE;
        data = 0;
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111S_READ_MODE %x.\n", data);

        reg[0] = MT9V111S_PIXEL_CLOCK_SPEED;
        data = 0;
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_sensor_lib: MT9V111S_PIXEL_CLOCK_SPEED %x.\n", data);

        FUNC_END;
        return error;
}

/*!
 * mt9v111_power function
 *    Turn Sensor power On/Off
 *
 * @param       cameraOn      true to turn camera on, otherwise shut down
 *
 * @return status
 */
u8 
mt9v111_power(bool cameraOn)
{
        u8 reg[1];
        u16 data;
        u8 error = 0;

        FUNC_START;

        ipu_csi_enable_mclk(true);

        // switch to sensor registers
        reg[0] = MT9V111I_ADDR_SPACE_SEL;
        data = (u16) MT9V111I_SEL_SCA;
        data = mt9v111_endian_swap16(data); 
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

        reg[0] = MT9V111S_OUTPUT_CTRL;
        data = (cameraOn == false) ? 0 : 
                MT9V111S_OUTCTRL_CHIP_ENABLE | testpattern  | 0x3000;
        data = mt9v111_endian_swap16(data); 
        printk("mt9v111_power :write MT9V111S_OUTPUT_CTRL %x cameraOn = %x \n",
                 data, cameraOn);
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);
        // Debug
        data = 0;
        mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2); 
        DPRINTK("mt9v111_power :MT9V111S_OUTPUT_CTRL %x.\n", data);

        FUNC_END;

        return error;
}

extern void gpio_sensor_active(void);

/*!
 * MT9V111 csi Configuration and Initialization
 *
 * @return  Error code indicating success or failure
 */
static int 
mt9v111_csi_config(void)
{
        u8 err = 0;
        ipu_csi_signal_cfg_t param;
        FUNC_START;

        gpio_sensor_active();  
      
        /*
         * Micron MT9M111 Init 
         * pixel format :YUV422 ,8 bit per component, color order:UYVY 
         * BT.656 with synchronization codes. 
         */
        param.sens_clksrc= 0x1; //gated  
        param.clk_mode   = 0x0; //gated  
        param.pixclk_pol = 0x0;
        param.data_width = 0x1; 
        param.data_pol   = 0x0;  
        param.ext_vsync  = 0x0; 
        param.Vsync_pol  = 0x0; 
        param.Hsync_pol  = 0x0; 
     
        ipu_csi_init_interface(640 -1, 480 -1, IPU_PIX_FMT_UYVY, &p_mclk_freq, 
                               param);        
        FUNC_END;
        return err;
}

/*!
 * MT9V111 sensor configuration
 *
 * @param frame_rate       int 
 * @return  Error code indicating success or failure
 */
int 
mt9v111_config(int frame_rate)
{
        u32 num_clock_per_row;
        u32 max_rate;
        u32 temp;

        FUNC_START;

        if (!mt9v111_device.coreReg)
                mt9v111_device.coreReg = (mt9v111_coreReg *) 
                        kmalloc(sizeof(mt9v111_coreReg), GFP_KERNEL);

        if (!mt9v111_device.coreReg) 
                return -1;

        memset(mt9v111_device.coreReg, 0, sizeof(mt9v111_coreReg)); 

        if (!mt9v111_device.ifpReg)
                mt9v111_device.ifpReg = (mt9v111_IFPReg *) 
                        kmalloc(sizeof(mt9v111_IFPReg), GFP_KERNEL);

        if (!mt9v111_device.ifpReg) 
                return -1;

        memset(mt9v111_device.ifpReg, 0, sizeof(mt9v111_IFPReg)); 

        mt9v111_device.coreReg->addressSelect = MT9V111I_SEL_SCA;  
        mt9v111_device.ifpReg->addrSpaceSel = MT9V111I_SEL_IFP;

        mt9v111_device.coreReg->windowHeight =	MT9V111_WINHEIGHT;
        mt9v111_device.coreReg->windowWidth = MT9V111_WINWIDTH; 
        mt9v111_device.coreReg->zoomColStart = 0;
        mt9v111_device.coreReg->zomRowStart = 0; 
        mt9v111_device.coreReg->digitalZoom = 0x0;

        mt9v111_device.coreReg->verticalBlanking = MT9V111_VERTBLANK_DEFAULT;
        mt9v111_device.coreReg->horizontalBlanking = MT9V111_HORZBLANK_MIN;
        mt9v111_device.coreReg->pixelClockSpeed = 0;   
        mt9v111_device.coreReg->readMode = 0xd0a1;

        mt9v111_device.ifpReg->outputFormatCtrl2 = 0;
        mt9v111_device.ifpReg->gainLimitAE = 0x300;
        mt9v111_device.ifpReg->AESpeed	= 0x80;

        if (frame_rate == 0) {
                mt9v111_device.ifpReg->formatControl =	0xc800;
                mt9v111_device.ifpReg->modeControl = 0x708e;
                mt9v111_device.coreReg->shutterWidth = 0xf8;
        }
        else {
                num_clock_per_row = (mt9v111_device.coreReg->windowWidth + 
                       mt9v111_device.coreReg->horizontalBlanking + 114) * 2;
                max_rate = (10 * p_mclk_freq) / (num_clock_per_row * 
                           (mt9v111_device.coreReg->windowHeight + 
                           mt9v111_device.coreReg->verticalBlanking - 5));
                temp = p_mclk_freq / 10 * (max_rate - frame_rate * 10);
                temp = temp / (max_rate * frame_rate * num_clock_per_row / 10);
                mt9v111_device.coreReg->shutterWidth = temp;

                mt9v111_device.ifpReg->formatControl =	0xc000;
                mt9v111_device.ifpReg->modeControl = 0x300c;
        }

        mt9v111_sensor_lib(mt9v111_device.coreReg, mt9v111_device.ifpReg);

        FUNC_END;

        return 0;
}

/*!
 * Exit and cleanup for the sensor configuration
 *
 * @return  Error code indicating success or failure
 */
static void 
mt9v111_config_clean(void)
{
        FUNC_START;

        if (mt9v111_device.coreReg)
        { 
                kfree(mt9v111_device.coreReg); 
                mt9v111_device.coreReg = 0;
        }

        if (mt9v111_device.ifpReg)
        { 
                kfree(mt9v111_device.ifpReg); 
                mt9v111_device.ifpReg = 0;
        }

        FUNC_END;
}

/*!
 * Set sensor to test mode, which will generate test pattern.
 *		 
 * @return none
 */
static void 
mt9v111_test_pattern(bool flag)
{
        u8 reg[1];
        u16 data;

        FUNC_START;

        // set the chess box.
        // switch to sensor registers
        reg[0] = MT9V111I_ADDR_SPACE_SEL;
        data = MT9V111I_SEL_SCA;
        data = mt9v111_endian_swap16(data); 
        mxc_i2c_write(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, (u8 *) &data, 2);

        if (flag == true) {
                testpattern = MT9V111S_OUTCTRL_TEST_MODE;

                reg[0] = MT9V111S_ROW_NOISE_CTRL;
                data = mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1, 
                                    (u8 *) &data, 2) & 0xBF;
                DPRINTK("MT9V111S_ROW_NOISE_CTRL:write %x.\n", data);
                mxc_i2c_write(I2C_BUS,MT9V111_I2C_ADDRESS,reg,1,(u8 *)&data,2);
                // Debug
                data = 0;
                mxc_i2c_read(I2C_BUS,MT9V111_I2C_ADDRESS,reg,1,(u8 *)&data,2); 
                DPRINTK("MT9V111S_ROW_NOISE_CTRL %x.\n", data);

                reg[0] = MT9V111S_TEST_DATA;
                data = 0;
                data = mt9v111_endian_swap16(data); 
                DPRINTK("mt9v111_test_pattern :write %x.\n", data);
                mxc_i2c_write(I2C_BUS,MT9V111_I2C_ADDRESS,reg,1,(u8 *)&data,2);
                // Debug
                data = 0;
                mxc_i2c_read(I2C_BUS,MT9V111_I2C_ADDRESS,reg,1,(u8 *)&data,2); 
                DPRINTK("mt9v111_test_pattern :MT9V111S_TEST_DATA %x.\n",data);

                reg[0] = MT9V111S_OUTPUT_CTRL;
                // changes take effect
                data = MT9V111S_OUTCTRL_CHIP_ENABLE | testpattern | 0x3000;
                data = mt9v111_endian_swap16(data); 
                DPRINTK("write MT9V111S_OUTPUT_CTRL %x \n", data);
                mxc_i2c_write(I2C_BUS,MT9V111_I2C_ADDRESS,reg,1,(u8 *)&data,2);
                // Debug
                data = 0;
                mxc_i2c_read(I2C_BUS,MT9V111_I2C_ADDRESS, reg,1,(u8 *)&data,2); 
                DPRINTK("MT9V111S_OUTPUT_CTRL %x.\n", data);
        }
        else {
                testpattern = 0;

                reg[0] = MT9V111S_ROW_NOISE_CTRL;
                data = mxc_i2c_read(I2C_BUS, MT9V111_I2C_ADDRESS, reg, 1,
                                    (u8 *) &data, 2) | 0x40;
                DPRINTK("MT9V111S_ROW_NOISE_CTRL :write %x.\n", data);
                mxc_i2c_write(I2C_BUS,MT9V111_I2C_ADDRESS,reg,1,(u8 *)&data,2);
                // Debug
                data = 0;
                mxc_i2c_read(I2C_BUS,MT9V111_I2C_ADDRESS,reg,1,(u8 *)&data,2); 
                DPRINTK("MT9V111S_ROW_NOISE_CTRL %x.\n", data);

                reg[0] = MT9V111S_OUTPUT_CTRL;
                // changes take effect
                data = MT9V111S_OUTCTRL_CHIP_ENABLE | testpattern | 0x3000;
                data = mt9v111_endian_swap16(data); 
                DPRINTK("write MT9V111S_OUTPUT_CTRL %x \n", data);
                mxc_i2c_write(I2C_BUS,MT9V111_I2C_ADDRESS,reg,1,(u8 *) &data,2);
                // Debug
                data = 0;
                mxc_i2c_read(I2C_BUS,MT9V111_I2C_ADDRESS,reg,1,(u8 *)&data,2); 
                DPRINTK("MT9V111S_OUTPUT_CTRL %x.\n",data);
        }

        FUNC_END;
}

/*!
 * open interface
 *
 * @param inode       struct inode * 
 * @param filp        struct file * 
 *
 * @return  None
 */
static int 
mt9v111_open(struct inode *inode, struct file *filp)
{
        return 0;
}

/*!
 * close interface
 *
 * @param inode       struct inode * 
 * @param filp        struct file * 
 *
 * @return  None
 */
static int 
mt9v111_release(struct inode *inode, struct file *filp)
{
        return 0;	
}

// The read method is for Debug only
static ssize_t 
mt9v111_read(struct file *file, char __user *buffer, size_t count, loff_t *ppos)
{
        wait_event_interruptible(mt9v111_device.power_queue, 
                                 mt9v111_device.low_power == false);

        mt9v111_test_pattern(true);
        while (1) {
                msleep(10); 
                if (signal_pending(current)) {
                        DPRINTK("mt9v111_read interrupt received\n");
                        return 0;
                }
        }
        return 0;
}

/*!
 * ioctl interface 
 *  MT9V111_IOCTL_POWER_ON: send I2C command to turn on sensor;
 *  MT9V111_IOCTL_POWER_OFF: send I2C command to turn off sensor;
 *  MT9V111_IOCTL_TEST_ON: turn on sensor test pattern;
 *  MT9V111_IOCTL_TEST_OFF: turn off sensor test pattern;
 * @param inode       struct inode * 
 * @param filp        struct file * 
 * @param cmd         unsigned int 
 * @param arg         unsigned long 
 *
 * @return  None
 */
static int 
mt9v111_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, 
              unsigned long arg)
{
        wait_event_interruptible(mt9v111_device.power_queue, 
                                 mt9v111_device.low_power == false);
        switch (cmd) {
                /*!
                 * MT9V111_IOCTL_POWER_ON: send I2C command to turn on sensor
                 */
                case MT9V111_IOCTL_POWER_ON:
                        mt9v111_power(true);
                        break;

                /*!
                 * MT9V111_IOCTL_POWER_OFF: send I2C command to turn off sensor
                 */
                case MT9V111_IOCTL_POWER_OFF:
                        mt9v111_power(false);
                        break;

                /*!
                 * MT9V111_IOCTL_TEST_ON: turn on sensor test pattern
                 */
                case MT9V111_IOCTL_TEST_ON:
                        mt9v111_test_pattern(true);
                        break;

                /*!
                 * MT9V111_IOCTL_TEST_OFF: turn off sensor test pattern
                 */
                case MT9V111_IOCTL_TEST_OFF:
                        mt9v111_test_pattern(false);
                        break;

                default:
                        DPRINTK("mt9v111_ioctl not supported ioctls\n");
                        return -1;
        }

        return 0;
}

static struct file_operations mt9v111_fops = {
        .owner =   THIS_MODULE,
        .open =    mt9v111_open,
        .release = mt9v111_release,
        .read =    mt9v111_read,
        .ioctl =   mt9v111_ioctl,
};

/*!
 * This function is called to put the sensor in a low power state. Refer to the
 * document driver-model/driver.txt in the kernel source tree for more
 * information.
 *
 * @param   dev   the device structure
 * @param   state the power state the device is entering
 * @param   level the stage in device suspension process that we want the
 *                device to be put in
 *
 * @return  The function returns 0 on success and -1 on failure.
 */
static int mt9v111_suspend(struct device *dev, u32 state, u32 level)
{
        mt9v111_conf *mt9v111_dev = dev_get_drvdata(dev);
        
        if (mt9v111_dev == NULL) {
                return -1;
        }

        if (level == SUSPEND_DISABLE) {
                mt9v111_dev->low_power = true;
        } else if (level == SUSPEND_POWER_DOWN) {
                mt9v111_power(false); 
                ipu_csi_enable_mclk(false);
        }
        return 0;
}

/*!
 * This function is called to bring the sensor back from a low power state. Refer
 * to the document driver-model/driver.txt in the kernel source tree for more
 * information.
 *
 * @param   dev   the device structure 
 * @param   level the stage in device resumption process that we want the
 *                device to be put in
 *
 * @return  The function returns 0 on success and -1 on failure
 */
static int mt9v111_resume(struct device *dev, u32 level)
{
        mt9v111_conf *mt9v111_dev = dev_get_drvdata(dev);

        if (mt9v111_dev == NULL) {
                return -1;
        }

        if (level == RESUME_ENABLE) {
                mt9v111_dev->low_power = false;
                wake_up_interruptible(&mt9v111_device.power_queue);
        } else if (level == RESUME_POWER_ON) {
                mt9v111_power(true);
        }
        return 0;
}

/*!
 * This function is called during the driver binding process.
 *
 * @param   dev   the device structure used to store device specific
 *                information that is used by the suspend, resume and remove
 *                functions
 *
 * @return  The function always returns 0.
 */
static int mt9v111_probe(struct device *dev)
{
        dev_set_drvdata(dev, &mt9v111_device);

        init_waitqueue_head(&mt9v111_device.power_queue);

        return 0;
}

/*!
 * Dissociates the driver.
 *
 * @param   dev   the device structure 
 *
 * @return  The function always returns 0.
 */
static int mt9v111_remove(struct device *dev)
{
        dev_set_drvdata(dev, NULL);
        return 0;
}

/*!
 * This structure contains pointers to the power management functions.
 */
static struct device_driver mt9v111_driver = {
        .name           = "mt9v111",
        .bus            = &platform_bus_type,
        .probe          = mt9v111_probe,
        .remove         = mt9v111_remove,
        .suspend        = mt9v111_suspend,
        .resume         = mt9v111_resume,
};

/*! 
 * Device Definition for Mt9v111 devices
 */ 
static struct platform_device mt9v111_devices = {
        .name           = "mt9v111",
        .id             = 0,
};

/*!
 * MT9V111 init function
 *
 * @return  Error code indicating success or failure
 */
static __init int 
mt9v111_init(void)
{
        u8 err = 0;
        FUNC_START;

        mt9v111_csi_config();

        camera_init();

        gMajor_mt9v111 = register_chrdev(0,"mt9v111", &mt9v111_fops);
        if (gMajor_mt9v111 < 0) {
                DPRINTK("mt9v111_init: unable to register driver\n");
                return gMajor_mt9v111;
        }

        devfs_mk_cdev(MKDEV(gMajor_mt9v111,0), S_IFCHR | S_IRUGO | S_IWUSR,
                            "mt9v111");

        /* Register the device driver structure. */
        err = driver_register(&mt9v111_driver);
        if (err != 0) {
                printk("mt9v111_init: driver_register failed.\n");
                return err;
        }

        /* Register the I2C device */
        err = platform_device_register(&mt9v111_devices);
        if (err != 0) {
                printk("mt9v111_init: platform_device_register failed.\n");
                driver_unregister(&mt9v111_driver);
        }

        FUNC_END;
        return err;
}

/*!
 * MT9V111 cleanup function
 *
 * @return  Error code indicating success or failure
 */
static void __exit 
mt9v111_clean(void)
{
        FUNC_START;

        camera_exit();

        if(gMajor_mt9v111 > 0) {
                devfs_remove("mt9v111");
                unregister_chrdev(gMajor_mt9v111, "mt9v111");
        }
               
        driver_unregister(&mt9v111_driver);
        platform_device_unregister(&mt9v111_devices);
        mt9v111_config_clean();

        FUNC_END;
}

module_init(mt9v111_init);
module_exit(mt9v111_clean);

MODULE_AUTHOR("Freescale Semiconductor");
MODULE_DESCRIPTION("Mt9v111 Driver");
MODULE_LICENSE("GPL");

EXPORT_SYMBOL(mt9v111_power);
