/*
 * Copyright 2005 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ADC related functions for the Freescale Semiconductor MXC IPU
 *
 */

/*
 * @file ipu_adc.c
 *
 * @brief IPU ADC functions
 *
 * @ingroup IPU 
 */

#include <linux/types.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/errno.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include "ipu.h"
#include "ipu_prv.h"
#include "ipu_regs.h"
#include "ipu_param_mem.h"
#include "asm-arm/io.h"
#include <asm/arch/board.h>
#include "asm/arch/clock.h"

/* M3IF Snooping Configuration Register 0 (M3IFSCFG0) READ/WRITE*/
#define M3IFSCFG0 IO_ADDRESS(0xB8003028)
/* M3IF Snooping Configuration Register 1 (M3IFSCFG1) READ/WRITE*/
#define M3IFSCFG1 IO_ADDRESS(0xB800302C)
/* M3IF Snooping Configuration Register 2 (M3IFSCFG2) READ/WRITE*/
#define M3IFSCFG2 IO_ADDRESS(0xB8003030)


#define ADC_CHAN1_SA_MASK 0xFF800000


static void _ipu_set_cmd_data_mappings(display_port_t disp, uint32_t pixel_fmt);

static const uint32_t DI_mappings[] =
{
	0x1600AAAA, 0x00E05555, 0x00070000, 3,	// RGB888
	0x1400AAAF, 0x00A0555F, 0x0005000F, 3,	// RGB666
	0x0004003F, 0x000A000F, 0x000F003F, 1	// RGB565
};

int32_t _ipu_adc_init_channel(ipu_channel_t chan, display_port_t disp, mcu_mode_t cmd,
                              int16_t x_pos, int16_t y_pos)
{
        uint32_t reg;
        uint32_t y_off, start_addr, stride;
        uint32_t lock_flags;
        uint32_t size;
        
        size = 0;//(bits_per_pixel == 32) ? 3 : ((bits_per_pixel == 16) ? 2 : 1);

        switch (disp)
        {
        case DISP0:
                reg = readl(ADC_DISP0_CONF);
                stride = reg & ADC_DISP_CONF_SL_MASK;
                break;
        case DISP1:
                reg = readl(ADC_DISP1_CONF);
                stride = reg & ADC_DISP_CONF_SL_MASK;
                break;
        case DISP2:
                reg = readl(ADC_DISP2_CONF);
                stride = reg & ADC_DISP_CONF_SL_MASK;
                break;
        }
        stride++;
        start_addr = (y_pos * stride) + x_pos;
        
        spin_lock_irqsave(&ipu_lock, lock_flags);
        reg = readl(ADC_CONF);

        switch (chan)
        {
        case ADC_SYS1:
                reg &= ~0x00FF4000;
                reg |= ((uint32_t)size<<21 | (uint32_t)disp<<19 | (uint32_t)cmd<<16);

                writel(start_addr, ADC_SYSCHA1_SA);
                break;

        case ADC_SYS2:
                reg &= ~0xFF008000;
                reg |= ((uint32_t)size<<29 | (uint32_t)disp<<27 | (uint32_t)cmd<<24);

                writel(start_addr, ADC_SYSCHA2_SA);
                break;

        case CSI_PRP_VF_ADC:
        case MEM_PRP_VF_ADC:
                reg &= ~0x000000F9;
                reg |= ((uint32_t)size<<5 | (uint32_t)disp<<3 | ADC_CONF_PRP_EN);

                writel(start_addr, ADC_PRPCHAN_SA);
                break;

        case MEM_PP_ADC:
                reg &= ~0x00003F02;
                reg |= ((uint32_t)size<<10 | (uint32_t)disp<<8 | ADC_CONF_PP_EN);

                writel(start_addr, ADC_PPCHAN_SA);
                break;
        default:
                spin_unlock_irqrestore(&ipu_lock, lock_flags);
                return -1;
                break;
        }
        writel(reg, ADC_CONF);
        spin_unlock_irqrestore(&ipu_lock, lock_flags);
        return 0;
}
                              
int32_t _ipu_adc_uninit_channel(ipu_channel_t chan)
{
        uint32_t reg;
        uint32_t y_off, start_addr, stride;
        uint32_t lock_flags;
        
        spin_lock_irqsave(&ipu_lock, lock_flags);
        reg = readl(ADC_CONF);

        switch (chan)
        {
        case ADC_SYS1:
                reg &= ~0x00FF4000;
                break;
        case ADC_SYS2:
                reg &= ~0xFF008000;
                break;
        case CSI_PRP_VF_ADC:
        case MEM_PRP_VF_ADC:
                reg &= ~0x000000F9;
                break;
        case MEM_PP_ADC:
                reg &= ~0x00003F02;
                break;
        default:
                spin_unlock_irqrestore(&ipu_lock, lock_flags);
                return -1;
                break;
        }
        writel(reg, ADC_CONF);
        spin_unlock_irqrestore(&ipu_lock, lock_flags);
        return 0;
}

#if 0
int32_t ipu_adc_set_display(display_port_t disp, sys_channel_t chan,
                            uint32_t width, uint32_t height, uint32_t stride,
                            mcu_mode_t cmd,  uint32_t bits_per_pixel, 
                            display_addressing_t addr, 
                            uint32_t vsync_width, vsync_t mode)
{
        uint32_t reg = 0;
        uint32_t size = 0;  // default is 8-bpp
        uint32_t lock_flags;
        
        spin_lock_irqsave(&ipu_lock, lock_flags);

        size = (bits_per_pixel == 32) ? 3 : ((bits_per_pixel == 16) ? 2 : 1);

        reg = readl(ADC_CONF);

        switch (chan)
        {
        case SYSCHAN1:
                reg &= ~0x00FF4000;
                reg |= ((uint32_t)size<<21 | (uint32_t)disp<<19 | (uint32_t)cmd<<16);
                        // hylim need to find out why??? 
                writel(0, ADC_SYSCHA1_SA);
                break;

        case SYSCHAN2:
                reg &= ~0xFF008000;
                reg |= ((uint32_t)size<<29 | (uint32_t)disp<<27 | (uint32_t)cmd<<24);
                // hylim need to find out why??? 
                writel(0, ADC_SYSCHA2_SA);
                break;

        case PRPCHAN:
                reg &= ~0x000000F9;
                reg |= ((uint32_t)size<<5 | (uint32_t)disp<<3 | 0x1 /*PRP_EN*/);
                // hylim need to find out why??? 
                writel(0, ADC_PRPCHAN_SA);
                break;

        case PPCHAN:
                reg &= ~0x00003F02;
                reg |= ((uint32_t)size<<10 | (uint32_t)disp<<8 | 0x2 /*PP_EN*/);
                // hylim need to find out why??? 
                        writel(0, ADC_PPCHAN_SA);
                break;
        }

        writel(reg, ADC_CONF);
        spin_unlock_irqrestore(&ipu_lock, lock_flags);
   
        reg = ((uint32_t)((((size == 3)||(size == 2))?1:0)<<14) | (uint32_t)addr<<12 |
                                        (stride-1));
        switch (disp)
        {
                case DISP0:
                        writel(reg, ADC_DISP0_CONF);
                        writel((((height-1)<<16) | (width-1)), ADC_DISP0_SS);
                        reg = ((uint32_t)vsync_width<<16 | (uint32_t)mode);
                        break;
                case DISP1:
                        writel(reg, ADC_DISP1_CONF);
                        reg = ((uint32_t)vsync_width<<24 | (uint32_t)mode<<2);
                        break;
                case DISP2:
                        writel(reg, ADC_DISP2_CONF);
                        reg = ((uint32_t)vsync_width<<24 | (uint32_t)mode<<2 | 0x10);
                        break;
                default :
                        return -EINVAL;
        }

        if (mode!=Disable)
        {
                spin_lock_irqsave(&ipu_lock, lock_flags);
                writel((reg |readl(ADC_DISP_VSYNC)), ADC_DISP_VSYNC);
                spin_unlock_irqrestore(&ipu_lock, lock_flags);

                if (disp!=DISP0)        
                        writel((((height-1)<<16) | (width-1)), ADC_DISP12_SS);
        }

        return 0;
}
#endif

int32_t    
ipu_adc_write_template(display_port_t disp, uint32_t *pCmd, bool write)
{
    uint32_t ima_addr = 0;
    uint32_t row_nu;
    int i;

    // Set IPU_IMA_ADDR (IPU Internal Memory Access Address)
    // MEM_NU = 0x0001 (CPM) 
    // ROW_NU = 2*N ( N is channel number)
    // WORD_NU = 0
    if (write)
    {
        row_nu = (uint32_t)disp * 2 * ATM_ADDR_RANGE;
    }
    else
    {
        row_nu = ((uint32_t)disp * 2 + 1) * ATM_ADDR_RANGE;
    }

	// form template addr for IPU_IMA_ADDR
    ima_addr = (0x3<<16 /*Template memory*/ | row_nu<<3);

    writel(ima_addr, IPU_IMA_ADDR);

	// write template data for IPU_IMA_DATA
    for (i = 0; i < TEMPLATE_BUF_SIZE; i++)
		// only DATA field are needed
        writel(pCmd[i], IPU_IMA_DATA);

    return 0;
}

int32_t	
ipu_adc_write_cmd(display_port_t disp, cmddata_t type,
					uint32_t cmd, const uint32_t * params, 
					uint16_t numParams)
{
    uint16_t i;

    writel((uint32_t)((type ? 0x0 : 0x1) | disp<<1 | 0x10), DI_DISP_LLA_CONF);
	writel(cmd, DI_DISP_LLA_DATA);

	writel((uint32_t)(0x10 | disp<<1 | 0x1), DI_DISP_LLA_CONF);	
	for (i = 0; i < numParams; i++)
		writel(params[i], DI_DISP_LLA_DATA);	
	return 0;
}

int32_t ipu_adc_set_update_mode(ipu_channel_t channel, ipu_adc_update_mode_t mode,
                                uint32_t refresh_rate, 
                                unsigned long addr, uint32_t * size)
{
        uint32_t ref_per, reg, src=0;
        uint32_t lock_flags;
        
        spin_lock_irqsave(&ipu_lock, lock_flags);
        
        reg = readl(IPU_FS_DISP_FLOW);
        reg &= ~FS_AUTO_REF_PER_MASK;
        switch (mode)
        {
                case IPU_ADC_REFRESH_NONE:
                        src = 0;
                        break;
                case IPU_ADC_AUTO_REFRESH:
                        ref_per = (refresh_rate * g_ipu_clk) / 217;
                        ref_per--;
                        reg |= ref_per << FS_AUTO_REF_PER_OFFSET;

                        src = FS_SRC_AUTOREF;
                        break;
                case IPU_ADC_AUTO_REFRESH_SNOOP:
                        ref_per = (refresh_rate * g_ipu_clk) / 217;
                        ref_per--;
                        reg |= ref_per << FS_AUTO_REF_PER_OFFSET;

                        src = FS_SRC_AUTOREF_SNOOP;
                        break;
                case IPU_ADC_SNOOPING:
                        src = FS_SRC_SNOOP;
                        break;
        }

        switch (channel)
        {
                case ADC_SYS1:
                        reg &= ~FS_ADC1_SRC_SEL_MASK;
                        reg |= src << FS_ADC1_SRC_SEL_OFFSET;
                        break;
                case ADC_SYS2:
                        reg &= ~FS_ADC2_SRC_SEL_MASK;
                        reg |= src << FS_ADC2_SRC_SEL_OFFSET;
                        break;
                default:
                        spin_unlock_irqrestore(&ipu_lock, lock_flags);
                        return -EINVAL;
        }
        writel(reg, IPU_FS_DISP_FLOW);

        // Setup M3IF for snooping
        if ((mode == IPU_ADC_AUTO_REFRESH_SNOOP) ||
            (mode == IPU_ADC_SNOOPING)) {
                uint32_t msb;
                uint32_t seg_size;
                uint32_t window_size;
                int i;

                msb = fls(*size);
                if (!(*size & ((1UL << msb) - 1)))
                        msb--;       // Already aligned to power 2
                if (msb < 11)
                        msb = 11;

                window_size = (1UL << msb);
                seg_size = window_size / 64;
                
                msb -= 11;

                reg = addr & ~((1UL << msb) - 1);
                reg |= msb << 1;
                reg |= 1; // enable snooping
                writel(reg, M3IFSCFG0);

                reg = 0;
                for (i = 0; i < 32; i++) {
                        if (i*seg_size >= *size)
                                break;
                        reg |= 1UL << i;
                }
                writel(reg, M3IFSCFG1);


                reg = 0;
                for (i = 32; i < 64; i++) {
                        if (i*seg_size >= *size)
                                break;
                        reg |= 1UL << (i - 32);
                }
                writel(reg, M3IFSCFG2); 

                *size = window_size;

                DPRINTK("Snooping enabled: window size = 0x%X, M3IFSCFG0=0x%08X, M3IFSCFG1=0x%08X, M3IFSCFG2=0x%08X\n",
                        window_size, readl(M3IFSCFG0), readl(M3IFSCFG1), readl(M3IFSCFG2));
        }

        spin_unlock_irqrestore(&ipu_lock, lock_flags);
        return 0;
}

// Parallel
void ipu_adc_init_parallel_ifc(display_port_t disp, 
                               uint16_t width, uint16_t height,
                               uint32_t pixel_fmt,
                               uint32_t stride,
                               ipu_adc_sig_cfg_t sig,
                               display_addressing_t addr,
                               uint32_t vsync_width, vsync_t mode)
{
	uint32_t temp;
        uint32_t lock_flags;
        uint32_t old_conf;
        uint32_t adc_disp_conf;
        uint32_t adc_disp_vsync;
        uint32_t old_pol;
        
//        adc_disp_conf = ((uint32_t)((((size == 3)||(size == 2))?1:0)<<14) | 
//                         (uint32_t)addr<<12 | (stride-1));
        adc_disp_conf = (uint32_t)addr<<12 | (stride - 1);
        
        _ipu_set_cmd_data_mappings(disp, pixel_fmt);

        spin_lock_irqsave(&ipu_lock, lock_flags);
        old_conf = readl(DI_DISP_IF_CONF);
        old_pol = readl(DI_DISP_SIG_POL);
        adc_disp_vsync = readl(ADC_DISP_VSYNC);
    
        switch (disp)
        {
        case DISP0:
                writel(adc_disp_conf, ADC_DISP0_CONF);
                writel((((height-1)<<16) | (width-1)), ADC_DISP0_SS);

                adc_disp_vsync &= ~(ADC_DISP_VSYNC_D0_MODE_MASK | 
                                    ADC_DISP_VSYNC_D0_WIDTH_MASK);
                adc_disp_vsync |= (vsync_width<<16) | (uint32_t)mode;

                old_pol &= ~0x2000003FL;
                temp = (uint32_t)(sig.data_pol | sig.cs_pol<<1 | sig.addr_pol<<2 | 
                                sig.read_pol <<3 | sig.write_pol<<4 | sig.Vsync_pol<<5 | sig.burst_pol<<29);
                writel(temp | old_pol, DI_DISP_SIG_POL);

                old_conf &= ~0x0000001FL;
                writel(old_conf | (uint32_t)(sig.burst_mode<<3 | sig.ifc_mode<<1) | 0x1, DI_DISP_IF_CONF);
                break;
        case DISP1:
                writel(adc_disp_conf, ADC_DISP1_CONF);
                writel((((height-1)<<16) | (width-1)), ADC_DISP12_SS);

                adc_disp_vsync &= ~(ADC_DISP_VSYNC_D12_MODE_MASK | 
                                    ADC_DISP_VSYNC_D12_WIDTH_MASK);
                adc_disp_vsync |= (vsync_width<<16) | (uint32_t)mode;

                old_pol &= ~0x4000FF00L;
                temp = (uint32_t)(sig.data_pol<<8 | sig.cs_pol<<9 | sig.addr_pol<<10 | 
                                sig.read_pol <<11 | sig.write_pol<<12 | sig.Vsync_pol<<6 | sig.burst_pol<<30);
                writel(temp | old_pol, DI_DISP_SIG_POL);
                old_conf &= ~0x00003F00L;
                writel(old_conf | (uint32_t)(sig.burst_mode<<12 | sig.ifc_mode<<9) | 0x100, DI_DISP_IF_CONF);
                break;
        case DISP2:                     
                writel(adc_disp_conf, ADC_DISP2_CONF);
                writel((((height-1)<<16) | (width-1)), ADC_DISP12_SS);

                adc_disp_vsync &= ~(ADC_DISP_VSYNC_D12_MODE_MASK | 
                                    ADC_DISP_VSYNC_D12_WIDTH_MASK);
                adc_disp_vsync |= (vsync_width<<16) | (uint32_t)mode;

                old_pol &= ~0x80FF0000L;
                temp = (uint32_t)(sig.data_pol << 16 | sig.cs_pol << 17 | 
                                  sig.addr_pol << 18 | sig.read_pol <<19 | 
                                  sig.write_pol<<20 | sig.Vsync_pol<<6 | 
                                  sig.burst_pol<<31);
                writel(temp | old_pol, DI_DISP_SIG_POL);

                old_conf &= ~0x003F0000L;
                writel(old_conf | (uint32_t)(sig.burst_mode<<20 | sig.ifc_mode<<17) | 0x10000, DI_DISP_IF_CONF);
                break;
        default:
                break;
        }

        writel(adc_disp_vsync, ADC_DISP_VSYNC);        

    spin_unlock_irqrestore(&ipu_lock, lock_flags);

    //FIXME
    writel(((0x25 << 22) | (0x9 << 12) | 0x105),DI_DISP0_TIME_CONF_1);
    writel(((0x38 << 22) | (0x9 << 12) | 0x195),DI_DISP0_TIME_CONF_2);
    writel(((0x66 << 16) | 0x328),DI_DISP0_TIME_CONF_2);
}

//Serial
void ipu_adc_init_serial_ifc(display_port_t disp, ipu_adc_sig_cfg_t sig, 
                             bool premable_en, bool burst, uint8_t premableLen, 
                             uint8_t premable, uint8_t serialBits,
                             uint32_t pixel_fmt)
{
        uint32_t temp;
        uint32_t old_conf = readl(DI_DISP_IF_CONF);
        uint32_t old_pol = readl(DI_DISP_SIG_POL);
        uint32_t serial_reg = (uint32_t)((burst ? 1 << 24 : 0) | 
                          ((serialBits & 0x1F) << 16) | (premable << 8) |
                                                  ((premableLen & 0x7) << 4) | (sig.RW_mode << 1) |
                                                  (premable_en ? 1 : 0));

        _ipu_set_cmd_data_mappings(disp, pixel_fmt);

        switch (disp)
        {
                case DISP1:
                        temp = (uint32_t)(sig.data_pol<<13 | sig.clk_pol<<14 | sig.addr_pol<<15);
                        writel(old_conf | (uint32_t)(sig.ifc_mode << 9) | 0x100, DI_DISP_IF_CONF);
                        writel(temp | old_pol, DI_DISP_SIG_POL);
                        writel(serial_reg, DI_SER_DISP1_CONF);
                        break;
                case DISP2:
                        temp = (uint32_t)(sig.data_pol<<21 | sig.clk_pol<<22 | sig.addr_pol<<23);
                        writel(old_conf | (uint32_t)(sig.ifc_mode << 17) | 0x10000, DI_DISP_IF_CONF);
                        writel(temp | old_pol, DI_DISP_SIG_POL);
                        writel(serial_reg, DI_SER_DISP2_CONF);
                        break;
                default:
                        break;
        }               

}

#if 0
void 
BMXxIPUADC::SetDITimingConfig(display_port_t disp, time_conf_t type, 
							uint16_t period, uint16_t up, uint16_t down)
{
	uint32_t reg = (uint32_t)((down&0x3FF)<<22 | (up&0x3FF)<<12 | (period&0xFFF));
	switch (disp)
	{
		case DISP0:
			if (type == WRITE)
				writel(DI_DISP0_TIME_CONF_1, reg);
			else
				writel(DI_DISP0_TIME_CONF_2, reg);
			break;
		case DISP1:
			if (type == WRITE)
				writel(DI_DISP1_TIME_CONF_1, reg);
			else
				writel(DI_DISP1_TIME_CONF_2, reg);
			break;
		case DISP2:
			if (type == WRITE)
				writel(DI_DISP2_TIME_CONF_1, reg);
			else
				writel(DI_DISP2_TIME_CONF_2, reg);
			break;
			break;
		default:
			break;
	}		
	
}

void 
BMXxIPUADC::SetDITimingConfig(display_port_t disp, uint16_t pixClkPeriod, 
						uint16_t read_pos, uint8_t read_ws, uint8_t deadclk)
{
	uint32_t reg  = readl(DI_DISP_IF_CONF) & 0x7FFFFFF;
	writel(DI_DISP_IF_CONF, reg | (uint32_t)((deadclk&0xF)<<26));

	reg = (uint32_t)((read_ws&0x3)<<28 | (read_pos&0x3FF)<<16 | (pixClkPeriod&0xFFF));

	switch (disp)
	{
		case DISP0:
			writel(DI_DISP0_TIME_CONF_3, reg);
			break;
		case DISP1:
			writel(DI_DISP1_TIME_CONF_3, reg);
			break;
		case DISP2:
			writel(DI_DISP2_TIME_CONF_3, reg);
			break;
		default:
			break;
	}		
}
#endif

// Private methods
static void _ipu_set_cmd_data_mappings(display_port_t disp, uint32_t pixel_fmt) 
{
	uint32_t map0, map1, map2, cycles;

	switch (pixel_fmt)
	{
		case IPU_PIX_FMT_BGR24:
			map0 = DI_mappings[0];
			map1 = DI_mappings[1];
			map2 = DI_mappings[2];
			cycles = DI_mappings[3];
			break;
		case IPU_PIX_FMT_RGB666:
			map0 = DI_mappings[4];
			map1 = DI_mappings[5];
			map2 = DI_mappings[6];
			cycles = DI_mappings[7];
			break;
		default:
			map0 = DI_mappings[8];
			map1 = DI_mappings[9];
			map2 = DI_mappings[10];
			cycles = DI_mappings[11];
			break;
	}
	switch (disp)
	{
		case 0:
			writel(0x00070000, DI_DISP0_CB0_MAP);
			writel(0x0000FFFF, DI_DISP0_CB1_MAP);
			writel(0x0000FFFF, DI_DISP0_CB2_MAP);
			writel(map0, DI_DISP0_DB0_MAP);
			writel(map1, DI_DISP0_DB1_MAP);
			writel(map2, DI_DISP0_DB2_MAP);
			writel(readl(DI_DISP_ACC_CC) | (cycles-1), DI_DISP_ACC_CC);
			break;
		case 1:
			writel(0x00070000, DI_DISP1_CB0_MAP);
			writel(0x0000FFFF, DI_DISP1_CB1_MAP);
			writel(0x0000FFFF, DI_DISP1_CB2_MAP);
			writel(map0, DI_DISP1_DB0_MAP);
			writel(map1, DI_DISP1_DB1_MAP);
			writel(map2, DI_DISP1_DB2_MAP);
			writel(readl(DI_DISP_ACC_CC) | ((uint32_t)(cycles-1)<<4), DI_DISP_ACC_CC);
			break;
		case 2:
			writel(0x00070000, DI_DISP2_CB0_MAP);
			writel(0x0000FFFF, DI_DISP2_CB1_MAP);
			writel(0x0000FFFF, DI_DISP2_CB2_MAP);
			writel(map0, DI_DISP2_DB0_MAP);
			writel(map1, DI_DISP2_DB1_MAP);
			writel(map2, DI_DISP2_DB2_MAP);
			writel(readl(DI_DISP_ACC_CC) | ((uint32_t)(cycles-1)<<8), DI_DISP_ACC_CC);
			break;
		default:
			break;
	}
}

//EXPORT_SYMBOL(ipu_adc_set_display);
EXPORT_SYMBOL(ipu_adc_write_template);
EXPORT_SYMBOL(ipu_adc_write_cmd);
EXPORT_SYMBOL(ipu_adc_set_update_mode);
EXPORT_SYMBOL(ipu_adc_init_parallel_ifc);
EXPORT_SYMBOL(ipu_adc_init_serial_ifc);

