/*
 * Copyright 2005 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * helper functions for IPU parameter memory access.
 *
 */
/*!
 * @file ipu_common.c
 * 
 * @brief This file contains the IPU driver common API functions. 
 * 
 * @ingroup IPU
 */

#include <linux/types.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/errno.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include "ipu.h"
#include "ipu_prv.h"
#include "ipu_regs.h"
#include "ipu_param_mem.h"
#include "asm-arm/io.h"
#include <asm/arch/board.h>
#include "asm/arch/clock.h"

/*
 * This type definition is used to define a node in the GPIO interrupt queue for 
 * registered interrupts for GPIO pins. Each node contains the GPIO signal number 
 * associated with the ISR and the actual ISR function pointer.
 */
struct ipu_irq_node {
        irqreturn_t (*handler)(int, void *, struct pt_regs *); /*!< the ISR */
        const char *name;               /*!< device associated with the interrupt */
        void *dev_id;                   /*!< some unique information for the ISR */
        __u32 flags;                    /*!< not used */
};


/* Globals */
int32_t         gOpenCount;        // Counter to track opens.
uint32_t        g_ipu_clk;
uint32_t        g_ipu_csi_clk;
bool            gSecChanEn[21];
uint32_t        gChannelInitMask;
uint32_t        gfs_disp_flow;
DEFINE_SPINLOCK(ipu_lock);

static struct ipu_irq_node ipu_irq_list[IPU_IRQ_COUNT];


/* Static functions */
static irqreturn_t ipu_irq_handler(int irq, void * desc,
			     struct pt_regs *regs);
static void _ipu_pf_init(ipu_channel_params_t * params);
static void _ipu_pf_uninit(void);


inline static uint32_t channel_2_dma(ipu_channel_t ch, ipu_buffer_t type) {
    return ((type==IPU_INPUT_BUFFER) ? ((uint32_t)ch & 0xFF) : \
           ((type==IPU_OUTPUT_BUFFER) ? (((uint32_t)ch>>8) & 0xFF) : (((uint32_t)ch>>16) & 0xFF)));
};

inline static uint32_t DMAParamAddr(uint32_t dma_ch) {
    return ( 0x10000 | (dma_ch << 4) );
};

/*!
 * This function is called by the driver framework to initialize the IPU 
 * hardware.
 * 
 * @param       dev       The device structure for the IPU passed in by the framework.
 *
 * @return      This function returns 0 on success or negative error code on error
 */
static  
int ipu_probe(struct device *dev)
{
    FUNC_START;

    spin_lock_init(&ipu_lock);

    if (IPU_PoolInitialize(MXCIPU_MEM_ADDRESS, MXCIPU_MEM_SIZE, SZ_4K)) {
        return -EBUSY;
    }
    
    // Enable IPU and CSI clocks

    // Get IPU clock freq
    mxc_clks_enable(IPU_CLK);
    g_ipu_clk = mxc_get_clocks(IPU_CLK);
    printk("g_ipu_clk = %d\n", g_ipu_clk);

#if 0
    mxc_clks_enable(CSI_BAUD);
    g_ipu_csi_clk = mxc_get_clocks(CSI_BAUD);
    printk("g_ipu_csi_clk = %d\n", g_ipu_csi_clk);
#endif

#ifdef CONFIG_ARCH_ARGONPLUS
    if (request_irq(INT_IPU, ipu_irq_handler, 0, 
        "IPU SYNC IRQ", 0) != 0) 
    {
        printk("request SYNC interrupt failed\n");
        return -EBUSY;
    }
#else
    if (request_irq(INT_IPU_SYN, ipu_irq_handler, 0, 
        "IPU SYNC IRQ", 0) != 0) 
    {
        printk("request SYNC interrupt failed\n");
        return -EBUSY;
    }

    if (request_irq(INT_IPU_ERR, ipu_irq_handler, 0, 
        "IPU ERR IRQ", 0) != 0) 
    {
        printk("request ERR interrupt failed\n");
        return -EBUSY;
    }
#endif
    writel(0x00100010L, DI_HSP_CLK_PER);

#if 0 //debug as Michael suggested
        int counter;
        writel(0x00010000, IPU_IMA_ADDR);
        for(counter = 0; counter <64*5; counter++)
                writel(0, IPU_IMA_DATA);
        writel(0x0000C001L, IDMAC_CHA_PRI);
#else
    // Set SDC refresh channels as high priority
    writel(0x0000C000L, IDMAC_CHA_PRI);
#endif

    /* Set to max back to back burst requests */
    writel(0x00000070L, IDMAC_CONF);

    FUNC_END;
    return 0;
}


/*!
 * This structure contains pointers to the power management callback functions.
 */
static struct device_driver mxcipu_driver = {
        .name           = "mxc_ipu",
        .bus            = &platform_bus_type,
        .probe          = ipu_probe,
};

/*!
 * Device definition for the IPU platform
 */
static struct platform_device mxcipu_device = {
        .name           = "mxc_ipu",
        .id             = 0,
};

/*!
 * This function is called to initialize a logical IPU channel.
 * 
 * @param       channel Input parameter for the logical channel ID to initalize.
 *
 * @param       params  Input parameter containing union of channel initialization 
 *                      parameters.
 *
 * @return      This function returns 0 on success or negative error code on fail
 */
int32_t ipu_init_channel(ipu_channel_t channel, ipu_channel_params_t * params )
{
    uint32_t ipu_conf;
    uint32_t reg;
    uint32_t lock_flags;

    FUNC_START;
    
    if (gChannelInitMask & (1L << IPU_CHAN_ID(channel)))
    {
        DPRINTK("Warning: channel already initialized\n");
    }
    
    spin_lock_irqsave(&ipu_lock, lock_flags);

    gChannelInitMask |= 1L << IPU_CHAN_ID(channel);
    
    // Enable IPU sub module
    ipu_conf = readl(IPU_CONF);
    if (gChannelInitMask & 0x00000066L) //CSI
    {
        ipu_conf |= IPU_CONF_CSI_EN;
    }
    if (gChannelInitMask & 0x00001FFFL) //IC
    {
        ipu_conf |= IPU_CONF_IC_EN;
    }
    if (gChannelInitMask & 0x00000A10L) //ROT
    {
        ipu_conf |= IPU_CONF_ROT_EN;
    }
    if (gChannelInitMask & 0x0001C000L) //SDC
    {
        ipu_conf |= IPU_CONF_SDC_EN | IPU_CONF_DI_EN;
    }
    if (gChannelInitMask & 0x00061140L) //ADC
    {
        ipu_conf |= IPU_CONF_ADC_EN | IPU_CONF_DI_EN;
    }
    if (gChannelInitMask & 0x00380000L) //PF
    {
        ipu_conf |= IPU_CONF_PF_EN;
    }
    writel(ipu_conf, IPU_CONF);
    
    switch(channel)
    {
        case CSI_PRP_VF_MEM:
            reg = readl(IPU_FS_PROC_FLOW);
            writel(reg & ~FS_VF_IN_VALID, IPU_FS_PROC_FLOW);

            if (params->mem_prp_vf_mem.graphics_combine_en)
                gSecChanEn[IPU_CHAN_ID(channel)] = true;

            _ipu_ic_init_prpvf(params, true);
            break;
        case CSI_PRP_VF_ADC:
            reg = readl(IPU_FS_PROC_FLOW);
            writel(reg | (FS_DEST_ADC << FS_PRPVF_DEST_SEL_OFFSET), 
                   IPU_FS_PROC_FLOW);

            reg = params->csi_prp_vf_adc.out_left | 
                  params->csi_prp_vf_adc.out_top << 8;           
            writel(reg, ADC_PRPCHAN_SA);
            writel((readl(ADC_CONF) | 0x1), ADC_CONF);

            _ipu_ic_init_prpvf(params, true);
            break;
        case MEM_PRP_VF_MEM:
            reg = readl(IPU_FS_PROC_FLOW);
            writel(reg | FS_VF_IN_VALID, IPU_FS_PROC_FLOW);

            if (params->mem_prp_vf_mem.graphics_combine_en)
                gSecChanEn[IPU_CHAN_ID(channel)] = true;

            _ipu_ic_init_prpvf(params, false);
            break;
        case MEM_ROT_VF_MEM:
            _ipu_ic_init_rotate_vf(params);
            break;
        case CSI_PRP_ENC_MEM:
            reg = readl(IPU_FS_PROC_FLOW);
            writel(reg & ~FS_ENC_IN_VALID, IPU_FS_PROC_FLOW);
            _ipu_ic_init_prpenc(params, true);
            break;
        case MEM_PRP_ENC_MEM:
            reg = readl(IPU_FS_PROC_FLOW);
            writel(reg | FS_ENC_IN_VALID, IPU_FS_PROC_FLOW);
            _ipu_ic_init_prpenc(params, false);
            break;
        case MEM_ROT_ENC_MEM:
            _ipu_ic_init_rotate_enc(params);
            break;
        case MEM_PP_ADC:
            reg = readl(IPU_FS_PROC_FLOW);
            writel(reg | (FS_DEST_ADC << FS_PP_DEST_SEL_OFFSET), 
                   IPU_FS_PROC_FLOW);

            gfs_disp_flow = readl(IPU_FS_DISP_FLOW);
            writel(gfs_disp_flow & ~(FS_ADC1_SRC_SEL_MASK|FS_ADC2_SRC_SEL_MASK), 
                   IPU_FS_DISP_FLOW);

            // FIXME: position calc depends on screen size
            reg = params->mem_pp_adc.out_left | 
                  params->mem_pp_adc.out_top << 8;           
            writel(reg, ADC_PPCHAN_SA);
            writel((readl(ADC_CONF) | 0x2), ADC_CONF);

            if (params->mem_pp_adc.graphics_combine_en)
                gSecChanEn[IPU_CHAN_ID(channel)] = true;

            _ipu_ic_init_pp(params);
            break;
        case MEM_PP_MEM:
            if (params->mem_pp_mem.graphics_combine_en)
                gSecChanEn[IPU_CHAN_ID(channel)] = true;

            _ipu_ic_init_pp(params);
            break;
        case MEM_ROT_PP_MEM:
            _ipu_ic_init_rotate_pp(params);
            break;
        case CSI_MEM:
            _ipu_ic_init_csi(params);
            break;

        case MEM_PF_Y_MEM:
        case MEM_PF_U_MEM:
        case MEM_PF_V_MEM:
            // Enable PF block
            _ipu_pf_init(params);
            break;
        
        case MEM_SDC_BG:
            DPRINTK("Initializing SDC BG\n");

            //_ipu_sdc_bg_init(params);
            break;
        case MEM_SDC_FG:
            DPRINTK("Initializing SDC FG\n");

            //_ipu_sdc_fg_init(params);
            break;
        case ADC_SYS1:
            _ipu_adc_init_channel(ADC_SYS1, params->adc_sys1.disp,
                                  params->adc_sys1.ch_mode,
                                  params->adc_sys1.x_pos,
                                  params->adc_sys1.y_pos);
            break;
        case ADC_SYS2:    
            _ipu_adc_init_channel(ADC_SYS2, params->adc_sys2.disp,
                                  params->adc_sys2.ch_mode,
                                  params->adc_sys2.x_pos,
                                  params->adc_sys2.y_pos);
            break;
        default:
            printk("Missing channel initialization\n");
            break;
    }
    
    spin_unlock_irqrestore(&ipu_lock, lock_flags);

    FUNC_END;
    return 0;
}
    
/*!
 * This function is called to uninitialize a logical IPU channel.
 * 
 * @param       channel Input parameter for the logical channel ID to uninitalize.
 */
void ipu_uninit_channel(ipu_channel_t channel)
{
    uint32_t ipu_conf;
    uint32_t lock_flags;
    uint32_t reg;

    FUNC_START;

    if ((gChannelInitMask & (1L << IPU_CHAN_ID(channel))) == 0)
    {
        DPRINTK("Channel already uninitialized\n");
        return;
    }

    spin_lock_irqsave(&ipu_lock, lock_flags);

    gChannelInitMask &= ~(1L << IPU_CHAN_ID(channel));

    gSecChanEn[IPU_CHAN_ID(channel)] = false;

    // Make sure channel is disabled
    //ipu_disable_channel(channel, false);

    switch(channel)
    {
        case CSI_MEM:
            _ipu_ic_uninit_csi();
            break;
        case CSI_PRP_VF_ADC:
            reg = readl(IPU_FS_PROC_FLOW);
            writel(reg & ~FS_PRPVF_DEST_SEL_MASK, IPU_FS_PROC_FLOW);

            writel((readl(ADC_CONF) & (~0x1)), ADC_CONF);

            // Fall thru
        case CSI_PRP_VF_MEM:
        case MEM_PRP_VF_MEM:
            _ipu_ic_uninit_prpvf();
            break;
        case MEM_PRP_VF_ADC:
            break;
        case MEM_ROT_VF_MEM:
            _ipu_ic_uninit_rotate_vf();
            break;
        case CSI_PRP_ENC_MEM:
        case MEM_PRP_ENC_MEM:
            _ipu_ic_uninit_prpenc();
            break;
        case MEM_ROT_ENC_MEM:
            _ipu_ic_uninit_rotate_enc();
            break;
        case MEM_PP_ADC:
            reg = readl(IPU_FS_PROC_FLOW);
            writel(reg & ~FS_PP_DEST_SEL_MASK, IPU_FS_PROC_FLOW);

            writel((readl(ADC_CONF) & (~0x2)), ADC_CONF);
            writel(gfs_disp_flow, IPU_FS_DISP_FLOW);

            // Fall thru
        case MEM_PP_MEM:
            _ipu_ic_uninit_pp();
            break;
        case MEM_ROT_PP_MEM:
            _ipu_ic_uninit_rotate_pp();
            break;

        case MEM_PF_Y_MEM:
            _ipu_pf_uninit();
            break;
        case MEM_PF_U_MEM:
        case MEM_PF_V_MEM:
            break;

        case MEM_SDC_BG:
            //_ipu_sdc_bg_uninit();
            break;
        case MEM_SDC_FG:
            //_ipu_sdc_fg_uninit();
            break;
        case ADC_SYS1:
            _ipu_adc_uninit_channel(ADC_SYS1);
            break;
        case ADC_SYS2:
            _ipu_adc_uninit_channel(ADC_SYS2);
            break;
        case MEM_SDC_MASK:
        case CHAN_NONE:
            break;
    }

    ipu_conf = readl(IPU_CONF);
    if ((gChannelInitMask & 0x00000063L) == 0) //CSI
    {
        //ipu_conf &= ~IPU_CONF_CSI_EN;
    }
    if ((gChannelInitMask & 0x00001FFFL) == 0) //IC
    {
        //ipu_conf &= ~IPU_CONF_IC_EN;
    }
    if ((gChannelInitMask & 0x00000A10L) == 0) //ROT
    {
        ipu_conf &= ~IPU_CONF_ROT_EN;
    }
    if ((gChannelInitMask & 0x0001C000L) == 0) //SDC
    {
        ipu_conf &= ~IPU_CONF_SDC_EN;
    }
    if ((gChannelInitMask & 0x00061140L) == 0) //ADC
    {
        ipu_conf &= ~IPU_CONF_ADC_EN;
    }
    if ((gChannelInitMask & 0x0007D140L) == 0) //DI
    {
        ipu_conf &= ~IPU_CONF_DI_EN;
    }
    if ((gChannelInitMask & 0x00380000L) == 0) //PF
    {
        ipu_conf &= ~IPU_CONF_PF_EN;
    }
    writel(ipu_conf, IPU_CONF);

    spin_unlock_irqrestore(&ipu_lock, lock_flags);

    FUNC_END;
}


/*!
 * This function is called to initialize a buffer for logical IPU channel.
 * 
 * @param       channel         Input parameter for the logical channel ID.
 *
 * @param       type            Input parameter which buffer to initialize.
 *
 * @param       pixel_fmt       Input parameter for pixel format of buffer. Pixel 
 *                              format is a FOURCC ASCII code.
 *
 * @param       width           Input parameter for width of buffer in pixels.
 *
 * @param       height          Input parameter for height of buffer in pixels.
 *
 * @param       stride          Input parameter for stride length of buffer 
 *                              in pixels.
 *
 * @param       rot_mode        Input parameter for rotation setting of buffer.
 *                              A rotation setting other than \b IPU_ROTATE_VERT_FLIP 
 *                              should only be used for input buffers of rotation 
 *                              channels. 
 *
 * @param       phyaddr_0       Input parameter buffer 0 physical address.
 *
 * @param       phyaddr_1       Input parameter buffer 1 physical address.
 *                              Setting this to a value other than NULL enables
 *                              double buffering mode.
 *
 * @return      This function returns 0 on success or negative error code on fail
 */
int32_t ipu_init_channel_buffer(ipu_channel_t channel, ipu_buffer_t type,
                                uint32_t pixel_fmt,
                                uint16_t width, uint16_t height,
                                uint32_t stride,
                                ipu_rotate_mode_t rot_mode,
                                void * phyaddr_0, void * phyaddr_1)
{
    uint32_t params[10];
    uint32_t lock_flags;
    uint32_t reg;
    uint32_t dma_chan;
    uint32_t stride_bytes;

    FUNC_START;

    dma_chan = channel_2_dma(channel, type);
    stride_bytes = stride * bytes_per_pixel(pixel_fmt);

    if (dma_chan == IDMA_CHAN_INVALID)
        return -EINVAL;

    if (stride_bytes % 4)
    {
        printk("Error: Stride length must be 32-bit aligned, stride = %d, bytes = %d\n",
               stride, stride_bytes);
        return -EINVAL;
    }

    // IC channels' stride must be multiple of 8 pixels
    if ((dma_chan <= 13) && (stride % 8))
    {
        printk("Error: Stride must be 8 pixel multiple\n");
        return -EINVAL;
    }
    
    // Build parameter memory data for DMA channel
    _ipu_ch_param_set_size(params, pixel_fmt, width, height, stride_bytes);
    _ipu_ch_param_set_buffer(params, phyaddr_0, phyaddr_1);
    _ipu_ch_param_set_rotation(params, rot_mode);
    // Some channels (rotation) have restriction on burst length
    if ((dma_chan == 10) || (dma_chan == 11) || (dma_chan == 13))
    {
        _ipu_ch_param_set_burst_size(params, 8);
    }
    else if (dma_chan == 24)        // PF QP channel
    {
        _ipu_ch_param_set_burst_size(params, 4);
    }
    else if (((dma_chan == 14) || (dma_chan == 15)) &&
             pixel_fmt == IPU_PIX_FMT_RGB565)
    {
        _ipu_ch_param_set_burst_size(params, 16);
    }

    spin_lock_irqsave(&ipu_lock, lock_flags);

    _ipu_write_param_mem(DMAParamAddr(dma_chan), params, 10);

    reg = readl(IPU_CHA_DB_MODE_SEL);
    if (phyaddr_1 != NULL) {
        reg |= 1UL << dma_chan;
    }
    else {
        reg &= ~(1UL << dma_chan);
    }
    writel(reg, IPU_CHA_DB_MODE_SEL);

    // Reset to buffer 0
    writel(1UL << dma_chan, IPU_CHA_CUR_BUF);

    spin_unlock_irqrestore(&ipu_lock, lock_flags);

    FUNC_END;
    return 0;
}


/*!
 * This function is called to update the physical address of a buffer for 
 * a logical IPU channel.
 * 
 * @param       channel         Input parameter for the logical channel ID.
 *
 * @param       type            Input parameter which buffer to initialize.
 *
 * @param       bufNum          Input parameter for which buffer number to update.
 *                              0 or 1 are the only valid values.
 *
 * @param       phyaddr         Input parameter buffer physical address.
 *
 * @return      This function returns 0 on success or negative error code on 
 *              fail. This function will fail if the buffer is set to ready.
 */
int32_t ipu_update_channel_buffer(ipu_channel_t channel, ipu_buffer_t type,
                             uint32_t bufNum, void * phyaddr)
{
    uint32_t reg;
    uint32_t lock_flags;
    uint32_t dma_chan = channel_2_dma(channel, type);

    FUNC_START;

    if (dma_chan == IDMA_CHAN_INVALID)
        return -EINVAL;

    spin_lock_irqsave(&ipu_lock, lock_flags);

    if (bufNum == 0)
    {
        reg = readl(IPU_CHA_BUF0_RDY);
        if (reg & (1UL << dma_chan))
        {
            spin_unlock_irqrestore(&ipu_lock, lock_flags);
            return -EACCES;
        }
        writel(DMAParamAddr(dma_chan) + 0x0008UL, IPU_IMA_ADDR);
        writel((uint32_t)phyaddr, IPU_IMA_DATA);
    }
    else
    {
        reg = readl(IPU_CHA_BUF1_RDY);
        if (reg & (1UL << dma_chan))
        {
            spin_unlock_irqrestore(&ipu_lock, lock_flags);
            return -EACCES;
        }
        writel(DMAParamAddr(dma_chan) + 0x0009UL, IPU_IMA_ADDR);
        writel((uint32_t)phyaddr, IPU_IMA_DATA);
    }

    spin_unlock_irqrestore(&ipu_lock, lock_flags);
    DPRINTK("IPU: update IDMA ch %d buf %d = 0x%08X\n", dma_chan, bufNum, phyaddr);
    return 0;
}


/*!
 * This function is called to set a channel's buffer as ready.
 * 
 * @param       channel         Input parameter for the logical channel ID.
 *
 * @param       type            Input parameter which buffer to initialize.
 *
 * @param       bufNum          Input parameter for which buffer number set to
 *                              ready state.
 *
 * @return      This function returns 0 on success or negative error code on fail
 */
int32_t ipu_select_buffer(ipu_channel_t channel, ipu_buffer_t type, uint32_t bufNum)
{
    uint32_t dma_chan = channel_2_dma(channel, type);

    FUNC_START;

    if (dma_chan == IDMA_CHAN_INVALID)
        return -EINVAL;

    if (bufNum == 0)
    {
        //Mark buffer 0 as ready.
        writel(1UL << dma_chan, IPU_CHA_BUF0_RDY);
    }
    else
    {
        //Mark buffer 1 as ready.
        writel(1UL << dma_chan, IPU_CHA_BUF1_RDY);
    }

    FUNC_END;
    return 0;
}


/*!
 * This function links 2 channels together for automatic frame
 * synchronization. The output of the source channel is linked to the input of 
 * the destination channel.
 * 
 * @param       src_ch          Input parameter for the logical channel ID of
 *                              the source channel.
 *
 * @param       dest_ch         Input parameter for the logical channel ID of
 *                              the destination channel.
 *
 * @return      This function returns 0 on success or negative error code on 
 *              fail. 
 */
int32_t ipu_link_channels(ipu_channel_t src_ch, ipu_channel_t dest_ch)
{
    uint32_t lock_flags;
    uint32_t out_dma;
    uint32_t in_dma;
    bool     isProc;
    uint32_t value;
    uint32_t mask;
    uint32_t offset;
    uint32_t fs_proc_flow;
    uint32_t fs_disp_flow;

    FUNC_START;

    spin_lock_irqsave(&ipu_lock, lock_flags);

    fs_proc_flow = readl(IPU_FS_PROC_FLOW);
    fs_disp_flow = readl(IPU_FS_DISP_FLOW);

    out_dma = (1UL << channel_2_dma(src_ch, IPU_OUTPUT_BUFFER));
    in_dma  = (1UL << channel_2_dma(dest_ch, IPU_INPUT_BUFFER));

    /* PROCESS THE OUTPUT DMA CH*/
    switch (out_dma)
    {
        /*VF->*/
        case IDMA_IC_1:
            printk("Link VF->");
            isProc = true;
            mask   = FS_PRPVF_DEST_SEL_MASK;
            offset = FS_PRPVF_DEST_SEL_OFFSET;
            value  = (in_dma == IDMA_IC_11)  ? FS_DEST_ROT    : //->VF_ROT
                     (in_dma == IDMA_SDC_BG) ? FS_DEST_SDC_BG : //->SDC_BG
                     (in_dma == IDMA_SDC_FG) ? FS_DEST_SDC_FG : //->SDC_FG
                     /* ->ADCDirect */
                     0;
            break;

        /*VF_ROT->*/
        case IDMA_IC_9:
            printk("Link VF_ROT->");
            isProc = true;
            mask   = FS_PRPVF_ROT_DEST_SEL_MASK;
            offset = FS_PRPVF_ROT_DEST_SEL_OFFSET;
            value  = (in_dma == IDMA_ADC_SYS1_WR) ? FS_DEST_ADC1: //->ADC1
                     /* ->ADC2 */
                     (in_dma == IDMA_SDC_BG) ? FS_DEST_SDC_BG : //->SDC_BG
                     (in_dma == IDMA_SDC_FG) ? FS_DEST_SDC_FG : //->SDC_FG
                     0;
            break;
            

        /*ENC->*/
        case IDMA_IC_0:
            printk("Link ENC->");
            isProc = true;
            mask   = 0;
            offset = 0;
            value  = (in_dma == IDMA_IC_10) ? FS_PRPENC_DEST_SEL: //->ENC_ROT
                     0;
            break;

        /*PP->*/
        case IDMA_IC_2:
            printk("Link PP->");
            isProc = true;
            mask   = FS_PP_DEST_SEL_MASK;
            offset = FS_PP_DEST_SEL_OFFSET;
            value  = (in_dma == IDMA_IC_13)  ? FS_DEST_ROT    : //->PP_ROT
                     /* ->ADC1 */
                     /* ->ADC2 */
                     (in_dma == IDMA_SDC_BG) ? FS_DEST_SDC_BG : //->SDC_BG
                     (in_dma == IDMA_SDC_FG) ? FS_DEST_SDC_FG : //->SDC_FG
                     /* ->ADCDirect */
                     0;
            break;          

        /*PP_ROT->*/
        case IDMA_IC_12:
            printk("Link PP_ROT->");
            isProc = true;
            mask   = FS_PP_ROT_DEST_SEL_MASK;
            offset = FS_PP_ROT_DEST_SEL_OFFSET;
            value  = (in_dma == IDMA_IC_5)  ? FS_DEST_ROT    : //->PP
                     /* ->ADC1 */
                     /* ->ADC2 */
                     (in_dma == IDMA_SDC_BG) ? FS_DEST_SDC_BG : //->SDC_BG
                     (in_dma == IDMA_SDC_FG) ? FS_DEST_SDC_FG : //->SDC_FG
                     0;
            break;          

        /*PF->*/
        case IDMA_PF_Y_OUT:
        case IDMA_PF_U_OUT:
        case IDMA_PF_V_OUT:
            printk("Link PF->");
            isProc = true;
            mask   = FS_PF_DEST_SEL_MASK;
            offset = FS_PF_DEST_SEL_OFFSET;
            value  = (in_dma == IDMA_IC_5)  ? FS_PF_DEST_PP :
                     (in_dma == IDMA_IC_13) ? FS_PF_DEST_ROT :
                     0;
            break;

        /* Invalid Chainings: ENC_ROT-> */
        default:
            printk("Link Invalid->");
            value      = 0;
            break;

    }

    if(value)
    {
        if(isProc)
        {
            fs_proc_flow &= ~mask;
            fs_proc_flow |= (value << offset);
        }
        else
        {
            fs_disp_flow &= ~mask;
            fs_disp_flow |= (value << offset);
        }
    }
    else
    {
        printk("Invalid channel chaining %d -> %d\n",
                  out_dma, in_dma);
        return -EINVAL;
    }

    

    /* PROCESS THE INPUT DMA CH*/
    switch(in_dma)
    {
        /* ->VF_ROT */
        case IDMA_IC_11:
            printk("VF_ROT\n");
            isProc = true;
            mask   = 0;
            offset = 0;
            value  = (out_dma == IDMA_IC_1) ? FS_PRPVF_ROT_SRC_SEL: //VF->
                     0;
            break;

        /* ->ENC_ROT */
        case IDMA_IC_10:
            printk("ENC_ROT\n");
            isProc = true;
            mask   = 0;
            offset = 0;
            value  = (out_dma == IDMA_IC_0) ? FS_PRPENC_ROT_SRC_SEL: //ENC->
                     0;
            break;

        /* ->PP */
        case IDMA_IC_5:
            printk("PP\n");
            isProc = true;
            mask   = FS_PP_SRC_SEL_MASK;
            offset = FS_PP_SRC_SEL_OFFSET;
            value  = (out_dma == IDMA_PF_Y_OUT) ? FS_PP_SRC_PF:  //PF->
                     (out_dma == IDMA_PF_U_OUT) ? FS_PP_SRC_PF:  //PF->
                     (out_dma == IDMA_PF_V_OUT) ? FS_PP_SRC_PF:  //PF->
                     (out_dma == IDMA_IC_12)    ? FS_PP_SRC_ROT: //PP_ROT->
                     0;
            break;
            
        /* ->PP_ROT */
        case IDMA_IC_13:
            printk("PP_ROT\n");
            isProc = true;
            mask   = FS_PP_ROT_SRC_SEL_MASK;
            offset = FS_PP_ROT_SRC_SEL_OFFSET;
            value  = (out_dma == IDMA_PF_Y_OUT) ? FS_PP_SRC_PF:  //PF->
                     (out_dma == IDMA_PF_U_OUT) ? FS_PP_SRC_PF:  //PF->
                     (out_dma == IDMA_PF_V_OUT) ? FS_PP_SRC_PF:  //PF->
                     (out_dma == IDMA_IC_2)     ? FS_ROT_SRC_PP: //PP->
                     0;
            break;

    
        /* ->SDC_BG */
        case IDMA_SDC_BG:
            printk("SDC_BG\n");
            isProc = false;
            mask   = FS_SDC_BG_SRC_SEL_MASK;
            offset = FS_SDC_BG_SRC_SEL_OFFSET;
            value  = (out_dma == IDMA_IC_9)  ? FS_SRC_ROT_VF: //VF_ROT->
                     (out_dma == IDMA_IC_12) ? FS_SRC_ROT_PP: //PP_ROT->
                     (out_dma == IDMA_IC_1)  ? FS_SRC_VF:     //VF->
                     (out_dma == IDMA_IC_2)  ? FS_SRC_PP:     //PP->
                     0;
            break;


        /* ->SDC_FG */
        case IDMA_SDC_FG:
            printk("SDC_FG\n");
            isProc = false;
            mask   = FS_SDC_FG_SRC_SEL_MASK;
            offset = FS_SDC_FG_SRC_SEL_OFFSET;
            value  = (out_dma == IDMA_IC_9)  ? FS_SRC_ROT_VF: //VF_ROT->
                     (out_dma == IDMA_IC_12) ? FS_SRC_ROT_PP: //PP_ROT->
                     (out_dma == IDMA_IC_1)  ? FS_SRC_VF:     //VF->
                     (out_dma == IDMA_IC_2)  ? FS_SRC_PP:     //PP->
                     0;
            break;

        /* ->ADC1 */
        case IDMA_ADC_SYS1_WR:
            printk("ADC_SYS1\n");
            isProc = false;
            mask   = FS_ADC1_SRC_SEL_MASK;
            offset = FS_ADC1_SRC_SEL_OFFSET;
            value  = (out_dma == IDMA_IC_9)  ? FS_SRC_ROT_VF: //VF_ROT->
                     (out_dma == IDMA_IC_12) ? FS_SRC_ROT_PP: //PP_ROT->
                     (out_dma == IDMA_IC_1)  ? FS_SRC_VF:     //VF->
                     (out_dma == IDMA_IC_2)  ? FS_SRC_PP:     //PP->
                     0;
            break;

        /* ->ADC2 */
        //TBD

        //Invalid chains:
        // ->ENC, ->VF, ->PF, ->VF_COMBINE, ->PP_COMBINE
        default:
            printk("Invalid\n");
            value  = 0;
            break;

    }

    if(value)
    {
        if(isProc)
        {
            fs_proc_flow &= ~mask;
            fs_proc_flow |= (value << offset);
        }
        else
        {
            fs_disp_flow &= ~mask;
            fs_disp_flow |= (value << offset);
        }
    }
    else
    {
        DPRINTK("Invalid channel chaining %d -> %d\n",
                  out_dma, in_dma);
        return -EINVAL;
    }

    writel(fs_proc_flow, IPU_FS_PROC_FLOW);
    writel(fs_disp_flow, IPU_FS_DISP_FLOW);
    
    spin_unlock_irqrestore(&ipu_lock, lock_flags);
    FUNC_END;
    return 0;
}


/*!
 * This function unlinks 2 channels and disables automatic frame
 * synchronization.
 * 
 * @param       src_ch          Input parameter for the logical channel ID of
 *                              the source channel.
 *
 * @param       dest_ch         Input parameter for the logical channel ID of
 *                              the destination channel.
 *
 * @return      This function returns 0 on success or negative error code on 
 *              fail. 
 */
int32_t ipu_unlink_channels(ipu_channel_t src_ch, ipu_channel_t dest_ch)
{
    uint32_t lock_flags;
    uint32_t out_dma;
    uint32_t in_dma;
    uint32_t fs_proc_flow;
    uint32_t fs_disp_flow;

    FUNC_START;

    spin_lock_irqsave(&ipu_lock, lock_flags);

    fs_proc_flow = readl(IPU_FS_PROC_FLOW);
    fs_disp_flow = readl(IPU_FS_DISP_FLOW);

    out_dma = (1UL << channel_2_dma(src_ch, IPU_OUTPUT_BUFFER));
    in_dma  = (1UL << channel_2_dma(dest_ch, IPU_INPUT_BUFFER));

    //clear the src_ch's output destination
    switch(out_dma)
    {
        /*VF->*/
        case IDMA_IC_1:
            printk("Unlink VF->");
            fs_proc_flow &= ~FS_PRPVF_DEST_SEL_MASK;
            break;

        /*VF_ROT->*/
        case IDMA_IC_9:
            printk("Unlink VF_Rot->");
            fs_proc_flow &= ~FS_PRPVF_ROT_DEST_SEL_MASK;
            break;

        /*ENC->*/
        case IDMA_IC_0:
            printk("Unlink ENC->");
            fs_proc_flow &= ~FS_PRPENC_DEST_SEL;
            break;

        /*PP->*/
        case IDMA_IC_2:
            printk("Unlink PP->");
            fs_proc_flow &= ~FS_PP_DEST_SEL_MASK;
            break;

        /*PP_ROT->*/
        case IDMA_IC_12:
            printk("Unlink PP_ROT->");
            fs_proc_flow &= ~FS_PP_ROT_DEST_SEL_MASK;
            break;

        /*PF->*/
        case IDMA_PF_Y_OUT:
        case IDMA_PF_U_OUT:
        case IDMA_PF_V_OUT:
            printk("Unlink PF->");
            fs_proc_flow &= ~FS_PF_DEST_SEL_MASK;
            break;

        default: //ENC_ROT->
            printk("Unlink Invalid->");
            break;
    }

    //clear the dest_ch's input source
    switch(in_dma)
    {
        /*->VF_ROT*/
        case IDMA_IC_11:
            printk("VF_ROT\n");
            fs_proc_flow &= ~FS_PRPVF_ROT_SRC_SEL;
            break;

        /*->Enc_ROT*/
        case IDMA_IC_10:
            printk("ENC_ROT\n");
            fs_proc_flow &= ~FS_PRPENC_ROT_SRC_SEL;
            break;

        /*->PP*/
        case IDMA_IC_5:
            printk("PP\n");
            fs_proc_flow &= ~FS_PP_SRC_SEL_MASK;
            break;

        /*->PP_ROT*/
        case IDMA_IC_13:
            printk("PP_ROT\n");
            fs_proc_flow &= ~FS_PP_ROT_SRC_SEL_MASK;
            break;

        /*->SDC_FG*/
        case IDMA_SDC_FG:
            printk("SDC_FG\n");
            fs_disp_flow &= ~FS_SDC_FG_SRC_SEL_MASK;
            break;
            
        /*->SDC_BG*/
        case IDMA_SDC_BG:
            printk("SDC_BG\n");
            fs_disp_flow &= ~FS_SDC_BG_SRC_SEL_MASK;
            break;

        /*->ADC1*/
        case IDMA_ADC_SYS1_WR:
            printk("ADC_SYS1\n");
            fs_disp_flow &= ~FS_ADC1_SRC_SEL_MASK;
            break;

        /*->ADC2*/
        //TBD

        default: //->VF, ->ENC, ->VF_COMBINE, ->PP_COMBINE, ->PF*
            printk("Invalid\n");
            break;
    }

    writel(fs_proc_flow, IPU_FS_PROC_FLOW);
    writel(fs_disp_flow, IPU_FS_DISP_FLOW);
    
    spin_unlock_irqrestore(&ipu_lock, lock_flags);
    FUNC_END;
    return 0;
}


/*!
 * This function enables a logical channel.
 * 
 * @param       channel         Input parameter for the logical channel ID.
 *
 * @return      This function returns 0 on success or negative error code on 
 *              fail. 
 */
int32_t ipu_enable_channel( ipu_channel_t channel )
{
    uint32_t reg;
    uint32_t lock_flags;
    uint32_t in_dma;
    uint32_t sec_dma;
    uint32_t out_dma;
    uint32_t chan_mask = 0;

    FUNC_START;

    spin_lock_irqsave(&ipu_lock, lock_flags);

    reg = readl(IDMAC_CHA_EN);

    // Get input and output dma channels
    out_dma = channel_2_dma(channel, IPU_OUTPUT_BUFFER);
    if (out_dma != IDMA_CHAN_INVALID)
        reg |= 1UL << out_dma;
    in_dma = channel_2_dma(channel, IPU_INPUT_BUFFER);
    if (in_dma != IDMA_CHAN_INVALID)
        reg |= 1UL << in_dma;

    // Get secondary input dma channel
    if (gSecChanEn[IPU_CHAN_ID(channel)])
    {
        sec_dma = channel_2_dma(channel, IPU_SEC_INPUT_BUFFER);
        if (sec_dma != IDMA_CHAN_INVALID)
        {
            reg |= 1UL << sec_dma;
        }
    }

    writel(reg | chan_mask, IDMAC_CHA_EN);

    switch(channel)
    {
        case CSI_PRP_VF_ADC:
        case CSI_PRP_VF_MEM:
            writel(readl(IC_CONF) | IC_CONF_PRPVF_EN, IC_CONF);
            break;
        case MEM_PRP_VF_MEM:
            writel(readl(IC_CONF) | IC_CONF_PRPVF_EN, IC_CONF);
            break;
        case MEM_ROT_VF_MEM:
            writel(readl(IC_CONF) | IC_CONF_PRPVF_ROT_EN, IC_CONF);
            break;
        case CSI_PRP_ENC_MEM:
            writel(readl(IC_CONF) | IC_CONF_PRPENC_EN, IC_CONF);
            break;
        case MEM_PRP_ENC_MEM:
            writel(readl(IC_CONF) | IC_CONF_PRPENC_EN, IC_CONF);
            break;
        case MEM_ROT_ENC_MEM:
            writel(readl(IC_CONF) | IC_CONF_PRPENC_ROT_EN, IC_CONF);
            break;
        case MEM_PP_MEM:
            writel(readl(IC_CONF) | IC_CONF_PP_EN, IC_CONF);
            break;
        case MEM_ROT_PP_MEM:
            writel(readl(IC_CONF) | IC_CONF_PP_ROT_EN, IC_CONF);
            break;
        case CSI_MEM:
            writel(readl(IC_CONF) | IC_CONF_RWS_EN | IC_CONF_PRPENC_EN, IC_CONF);
            break;

        case MEM_PF_Y_MEM:
        case MEM_PF_U_MEM:
        case MEM_PF_V_MEM:
            break;
        
        case MEM_SDC_BG:
            DPRINTK("Initializing SDC BG\n");
             _ipu_sdc_bg_init(NULL);
            break;
        case MEM_SDC_FG:
            DPRINTK("Initializing SDC FG\n");
            _ipu_sdc_fg_init(NULL);
            break;
        case ADC_SYS1:
//            writel(readl(ADC_CONF) | 0x00030000, ADC_CONF);     
            break;
        case ADC_SYS2:   
//            writel(readl(ADC_CONF) | 0x03000000, ADC_CONF);     
            break;
        default:
            printk("Missing channel initialization\n");
            break;
    }

    spin_unlock_irqrestore(&ipu_lock, lock_flags);
    FUNC_END;
    return 0;
}


/*!
 * This function disables a logical channel.
 * 
 * @param       channel         Input parameter for the logical channel ID.
 *
 * @param       wait_for_stop   Flag to set whether to wait for channel end
 *                              of frame or return immediately.
 *
 * @return      This function returns 0 on success or negative error code on 
 *              fail. 
 */
int32_t ipu_disable_channel( ipu_channel_t channel, bool wait_for_stop )
{
    uint32_t reg;
    uint32_t lock_flags;
    uint32_t sec_dma;
    uint32_t in_dma;
    uint32_t out_dma;
    uint32_t chan_mask = 0;

    FUNC_START;

    spin_lock_irqsave(&ipu_lock, lock_flags);

    // Get input and output dma channels
    out_dma = channel_2_dma(channel, IPU_OUTPUT_BUFFER);
    if (out_dma != IDMA_CHAN_INVALID)
        chan_mask = 1UL << out_dma;
    in_dma = channel_2_dma(channel, IPU_INPUT_BUFFER);
    if (in_dma != IDMA_CHAN_INVALID)
        chan_mask |= 1UL << in_dma;
    sec_dma = channel_2_dma(channel, IPU_SEC_INPUT_BUFFER);
    if (sec_dma != IDMA_CHAN_INVALID)
        chan_mask |= 1UL << sec_dma;

    spin_unlock_irqrestore(&ipu_lock, lock_flags);

    if (wait_for_stop)
    {
        uint32_t timeout = 8; 
        while ((readl(IDMAC_CHA_BUSY) & chan_mask) ||
               (readl(IPU_CHA_BUF0_RDY) & chan_mask) ||
               (readl(IPU_CHA_BUF1_RDY) & chan_mask))
        {
            timeout--;
            msleep(10);
            if (timeout == 0)
            {
                printk("Timeout waiting for channel to stop,\n" \
                       "\tbuf0_rdy = 0x%08X, buf1_rdy = 0x%08X\n" \
                       "\tbusy = 0x%08X, mask = 0x%08X\n", 
                    readl(IPU_CHA_BUF0_RDY), readl(IPU_CHA_BUF1_RDY),
                    readl(IDMAC_CHA_BUSY), chan_mask);
                break;
            }
        }
        DPRINTK("timeout = %d * 10ms\n", 8 - timeout);
    }

    // SDC BG and FG must be disabled before DMA is disabled
    if (channel == MEM_SDC_BG)
    {
        _ipu_sdc_bg_uninit();
    }
    else if (channel == MEM_SDC_FG)
    {
        _ipu_sdc_fg_uninit();
    }
    
    // Disable DMA channel(s)
    reg = readl(IDMAC_CHA_EN);
    writel(reg & ~chan_mask, IDMAC_CHA_EN);

    // Clear DMA related interrupts
    writel(chan_mask, IPU_INT_STAT_1);
    writel(chan_mask, IPU_INT_STAT_2);
    writel(chan_mask, IPU_INT_STAT_4);

    FUNC_END;
    return 0;
}

static
irqreturn_t ipu_irq_handler(int irq, void * desc,
			     struct pt_regs *regs)
{
    uint32_t line = 0;
    irqreturn_t result = IRQ_NONE;
    uint32_t int_stat;

    int_stat = readl(IPU_INT_STAT_1);
    int_stat &= readl(IPU_INT_CTRL_1);
    writel(int_stat, IPU_INT_STAT_1);
    while (int_stat)
    {
        if (int_stat & 0x1UL)
        {
            DPRINTK("intr on line %d\n", line);
            result |= ipu_irq_list[line].handler(line, ipu_irq_list[line].dev_id, regs);
        }
        int_stat = int_stat >> 1UL;
        line++;
    }

    line = 32;
    int_stat = readl(IPU_INT_STAT_2);
    int_stat &= readl(IPU_INT_CTRL_2);
    writel(int_stat, IPU_INT_STAT_2);
    while (int_stat)
    {
        if (int_stat & 0x1UL)
        {
            DPRINTK("intr on line %d\n", line);
            result |= ipu_irq_list[line].handler(line, ipu_irq_list[line].dev_id, regs);
        }
        int_stat = int_stat >> 1UL;
        line++;
    }
            
    line = 64;
    int_stat = readl(IPU_INT_STAT_3);
    int_stat &= readl(IPU_INT_CTRL_3);
    writel(int_stat, IPU_INT_STAT_3);
    while (int_stat)
    {
            if (int_stat & 0x1UL)
            {
                    DPRINTK("intr on line %d\n", line);
                    result |= ipu_irq_list[line].handler(line, ipu_irq_list[line].dev_id, regs);
            }
        int_stat = int_stat >> 1UL;
        line++;
    }

    line = 96;
    int_stat = readl(IPU_INT_STAT_4);
    int_stat &= readl(IPU_INT_CTRL_4);
    writel(int_stat, IPU_INT_STAT_4);
    while (int_stat)
    {
        if (int_stat & 0x1UL)
        {
            DPRINTK("intr on line %d\n", line);
            result |= ipu_irq_list[line].handler(line, ipu_irq_list[line].dev_id, regs);
        }
        int_stat = int_stat >> 1UL;
        line++;
    }
    
    line = 128;
    int_stat = readl(IPU_INT_STAT_5);
    int_stat &= readl(IPU_INT_CTRL_5);
    writel(int_stat, IPU_INT_STAT_5);
    while (int_stat)
    {
        if (int_stat & 0x1UL)
        {
            DPRINTK("intr on line %d\n", line);
            result |= ipu_irq_list[line].handler(line, ipu_irq_list[line].dev_id, regs);
        }
        int_stat = int_stat >> 1UL;
        line++;
    }
    
    return result;
}


/*!
 * This function enables the interrupt for the specified interrupt line.
 * The interrupt lines are defined in \b ipu_irq_line enum.
 * 
 * @param       irq             Interrupt line to enable interrupt for.
 *
 */
void ipu_enable_irq(uint32_t irq)
{
    uint32_t reg;
    uint32_t lock_flags;

    FUNC_START;

    spin_lock_irqsave(&ipu_lock, lock_flags);

    reg = readl(IPUIRQ_2_CTRLREG(irq));
    reg |= IPUIRQ_2_MASK(irq);
    writel(reg, IPUIRQ_2_CTRLREG(irq));

    spin_unlock_irqrestore(&ipu_lock, lock_flags);
    FUNC_END;
}


/*!
 * This function disables the interrupt for the specified interrupt line.
 * The interrupt lines are defined in \b ipu_irq_line enum.
 * 
 * @param       irq             Interrupt line to disable interrupt for.
 *
 */
void ipu_disable_irq(uint32_t irq)
{
    uint32_t reg;
    uint32_t lock_flags;

    FUNC_START;

    spin_lock_irqsave(&ipu_lock, lock_flags);

    reg = readl(IPUIRQ_2_CTRLREG(irq));
    reg &= ~IPUIRQ_2_MASK(irq);
    writel(reg, IPUIRQ_2_CTRLREG(irq));

    spin_unlock_irqrestore(&ipu_lock, lock_flags);
    FUNC_END;
}


/*!
 * This function returns the current interrupt status for the specified interrupt 
 * line. The interrupt lines are defined in \b ipu_irq_line enum.
 * 
 * @param       irq             Interrupt line to get status for.
 *
 * @return      Returns true if the interrupt is pending/asserted or false if  
 *              the interrupt is not pending.
 */
bool ipu_get_irq_status(uint32_t irq)
{
    uint32_t reg = readl(IPUIRQ_2_STATREG(irq));

    FUNC_START;

    if (reg & IPUIRQ_2_MASK(irq))
        return true;
    else
        return false;
}


/*!
 * This function registers an interrupt handler function for the specified 
 * interrupt line. The interrupt lines are defined in \b ipu_irq_line enum.
 * 
 * @param       irq             Interrupt line to get status for.
 *
 * @param       handler         Input parameter for address of the handler 
 *                              function.
 *
 * @param       irq_flags       Flags for interrupt mode. Currently not used.
 *
 * @param       devname         Input parameter for string name of driver 
 *                              registering the handler.
 *
 * @param       dev_id          Input parameter for pointer of data to be passed  
 *                              to the handler.
 *
 * @return      This function returns 0 on success or negative error code on 
 *              fail. 
 */
int ipu_request_irq(uint32_t irq, irqreturn_t (*handler)(int, void *, struct pt_regs *),
                 uint32_t irq_flags, const char * devname, void *dev_id)
{
    uint32_t lock_flags;

    FUNC_START;

    MXC_ERR_CHK(irq >= IPU_IRQ_COUNT);

    spin_lock_irqsave(&ipu_lock, lock_flags);

    if (ipu_irq_list[irq].handler != NULL)
    {
        printk("ipu_request_irq - handler already installed on irq %d\n", irq);
        spin_unlock_irqrestore(&ipu_lock, lock_flags);
        return -EINVAL;
    }

    ipu_irq_list[irq].handler = handler;
    ipu_irq_list[irq].flags = irq_flags;
    ipu_irq_list[irq].dev_id = dev_id;
    ipu_irq_list[irq].name = devname;
    
    spin_unlock_irqrestore(&ipu_lock, lock_flags);
    
    ipu_enable_irq(irq); /* enable the interrupt */

    FUNC_END;
    return 0;
}


/*!
 * This function unregisters an interrupt handler for the specified interrupt 
 * line. The interrupt lines are defined in \b ipu_irq_line enum.
 * 
 * @param       irq             Interrupt line to get status for.
 *
 * @param       dev_id          Input parameter for pointer of data to be passed  
 *                              to the handler. This must match value passed to
 *                              ipu_request_irq().
 *
 */
void ipu_free_irq(uint32_t irq, void *dev_id)
{
    uint32_t lock_flags;
    FUNC_START;

    spin_lock_irqsave(&ipu_lock, lock_flags);

    if (ipu_irq_list[irq].dev_id != dev_id)
    {
        spin_unlock_irqrestore(&ipu_lock, lock_flags);
        return;
    }

    ipu_disable_irq(irq); /* disable the interrupt */

    ipu_irq_list[irq].handler = NULL;
    spin_unlock_irqrestore(&ipu_lock, lock_flags);
    FUNC_END;
}


// Private functions

void _ipu_write_param_mem(uint32_t addr, uint32_t * data, uint32_t numWords)
{
    for ( ; numWords > 0; numWords--)
    {
        DPRINTK("IPU WriteParamMem() - addr = 0x%08X, data = 0x%08X\n", addr, *data);
        writel(addr, IPU_IMA_ADDR);
        writel(*data++, IPU_IMA_DATA);
        addr++;
        if ((addr & 0x7) == 5)
        {
            addr &= ~0x7;       // set to word 0
            addr += 8;          // increment to next row
        }
    }
}

static void _ipu_pf_init(ipu_channel_params_t * params)
{
        uint32_t reg;
        reg = readl(PF_CONF);

        //Setup the type of filtering required
        reg = (reg & ~PF_CONF_TYPE_MASK) | params->mem_pf_mem.operation;

        switch(params->mem_pf_mem.operation)
        {
                case PF_MPEG4_DEBLOCK:
                case PF_MPEG4_DERING:
                case PF_MPEG4_DEBLOCK_DERING: 
                        gSecChanEn[IPU_CHAN_ID(MEM_PF_Y_MEM)] = true;
                        gSecChanEn[IPU_CHAN_ID(MEM_PF_U_MEM)] = false;
                        break;
                case PF_H264_DEBLOCK:
                        gSecChanEn[IPU_CHAN_ID(MEM_PF_Y_MEM)] = true;
                        gSecChanEn[IPU_CHAN_ID(MEM_PF_U_MEM)] = true;
                        break;
                default:
                        gSecChanEn[IPU_CHAN_ID(MEM_PF_Y_MEM)] = false;
                        gSecChanEn[IPU_CHAN_ID(MEM_PF_U_MEM)] = false;
                        break;
        }

        //If h.264, then configure pause row setting
        if(params->mem_pf_mem.operation == PF_H264_DEBLOCK)
        {
                if(params->mem_pf_mem.h264_pause_row != 0)
                {
                        reg &= ~PF_CONF_PAUSE_EN;
                }
                else
                {
                        reg |= PF_CONF_PAUSE_EN;
                        reg |= (params->mem_pf_mem.h264_pause_row << PF_CONF_PAUSE_ROW_SHIFT) & PF_CONF_PAUSE_ROW_MASK;
                }
        }

        writel(reg, PF_CONF);
}

static void _ipu_pf_uninit(void)
{
        writel(0x0L, PF_CONF);
        gSecChanEn[IPU_CHAN_ID(MEM_PF_Y_MEM)] = false;
        gSecChanEn[IPU_CHAN_ID(MEM_PF_U_MEM)] = false;
}

uint32_t
bytes_per_pixel(uint32_t fmt)
{
    switch (fmt)
    {
        case IPU_PIX_FMT_GENERIC: //generic data
            return 1;
            break;
        case IPU_PIX_FMT_RGB565:
        case IPU_PIX_FMT_YUYV:
        case IPU_PIX_FMT_UYVY:
            return 2;
            break;
        case IPU_PIX_FMT_BGR24:
        case IPU_PIX_FMT_RGB24:
            return 3;
            break;
        case IPU_PIX_FMT_BGR32:
        case IPU_PIX_FMT_RGB32:
            return 4;
            break;
        default:
            return 1;
            break;
    }
    return 0;
}

ipu_color_space_t format_to_colorspace(uint32_t fmt)
{
    switch (fmt)
    {
        case IPU_PIX_FMT_RGB565:
        case IPU_PIX_FMT_BGR24:
        case IPU_PIX_FMT_RGB24:
        case IPU_PIX_FMT_BGR32:
        case IPU_PIX_FMT_RGB32:
            return RGB;
            break;

        default:
            return YCbCr;
            break;
    }
    return RGB;

}

int32_t __init ipu_gen_init(void)
{
    int32_t ret;

    FUNC_START;

    ret = driver_register(&mxcipu_driver);
    if (ret == 0) {
            ret = platform_device_register(&mxcipu_device);
            if (ret != 0) {
                    driver_unregister(&mxcipu_driver);
            }
    }

    FUNC_END;
    return 0;
}
subsys_initcall(ipu_gen_init);

static void __exit ipu_gen_uninit(void)
{
    FUNC_START;
#ifdef CONFIG_ARCH_ARGONPLUS
        free_irq(INT_IPU, 0);
#else
        free_irq(INT_IPU_SYN, 0);
        free_irq(INT_IPU_ERR, 0);
#endif
    FUNC_END;
}
module_exit(ipu_gen_uninit);

EXPORT_SYMBOL(ipu_init_channel);
EXPORT_SYMBOL(ipu_uninit_channel);
EXPORT_SYMBOL(ipu_init_channel_buffer);
EXPORT_SYMBOL(ipu_unlink_channels);
EXPORT_SYMBOL(ipu_update_channel_buffer);
EXPORT_SYMBOL(ipu_select_buffer);
EXPORT_SYMBOL(ipu_link_channels);
EXPORT_SYMBOL(ipu_enable_channel);
EXPORT_SYMBOL(ipu_disable_channel);
EXPORT_SYMBOL(ipu_enable_irq);
EXPORT_SYMBOL(ipu_disable_irq);
EXPORT_SYMBOL(ipu_get_irq_status);
EXPORT_SYMBOL(ipu_request_irq);
EXPORT_SYMBOL(ipu_free_irq);

