/*
 * Copyright 2005 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Image Converter related functions for the Freescale Semiconductor MXC IPU
 *
 */

/*
 * @file ipu_ic.c
 *
 * @brief IPU IC functions
 *
 * @ingroup IPU 
 */
#include <linux/types.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/errno.h>
#include <linux/spinlock.h>
#include "ipu.h"
#include "ipu_prv.h"
#include "ipu_regs.h"
#include "ipu_param_mem.h"
#include "asm-arm/io.h"

enum {
    IC_TASK_VIEWFINDER,
    IC_TASK_ENCODER,
    IC_TASK_POST_PROCESSOR
};

static void _init_csc(uint8_t ic_task, ipu_color_space_t in_format, ipu_color_space_t out_format);
static bool _calc_resize_coeffs(uint32_t inSize, uint32_t outSize, uint32_t * resizeCoeff, uint32_t * downsizeCoeff);

void _ipu_ic_init_prpvf(ipu_channel_params_t * params, bool src_is_csi)
{
    uint32_t reg, ic_conf;
    uint32_t downsizeCoeff, resizeCoeff;
    ipu_color_space_t in_fmt, out_fmt;

    // Setup vertical resizing
    _calc_resize_coeffs(params->mem_prp_vf_mem.in_height,
                     params->mem_prp_vf_mem.out_height,
                     &resizeCoeff,
                     &downsizeCoeff);
    reg = (downsizeCoeff << 30) | (resizeCoeff << 16);

    // Setup horizontal resizing
    _calc_resize_coeffs(params->mem_prp_vf_mem.in_width,
                     params->mem_prp_vf_mem.out_width,
                     &resizeCoeff,
                     &downsizeCoeff);
    reg |= (downsizeCoeff << 14) | resizeCoeff;

    writel(reg, IC_PRP_VF_RSC);

    ic_conf = readl(IC_CONF);

    // Setup color space conversion
    in_fmt = format_to_colorspace(params->mem_prp_vf_mem.in_pixel_fmt);
    out_fmt = format_to_colorspace(params->mem_prp_vf_mem.out_pixel_fmt);
    if ( in_fmt == RGB )
    {
        if ( (out_fmt == YCbCr) || (out_fmt == YUV) )
        {
            _init_csc(IC_TASK_VIEWFINDER, RGB, out_fmt);
            ic_conf |= IC_CONF_PRPVF_CSC1;    // Enable RGB->YCBCR CSC
        }
    }
    if ( (in_fmt == YCbCr) || (in_fmt == YUV) )
    {
        if ( out_fmt == RGB )
        {
            _init_csc(IC_TASK_VIEWFINDER, YCbCr, RGB);
            ic_conf |= IC_CONF_PRPVF_CSC1;    // Enable YCBCR->RGB CSC
        }
        else
        {
            // TODO: Support YUV<->YCbCr conversion?
        }
    }

    if (params->mem_prp_vf_mem.graphics_combine_en)
    {
        ic_conf |= IC_CONF_PRPVF_CMB;
    }
    
#ifndef CONFIG_VIRTIO_SUPPORT // Setting RWS_EN doesn't work in Virtio
    if (src_is_csi)
    {
        ic_conf &= ~IC_CONF_RWS_EN;
    }
    else
    {
        ic_conf |= IC_CONF_RWS_EN;
    }
#endif    
    writel(ic_conf, IC_CONF);
}

void _ipu_ic_uninit_prpvf(void)
{
    uint32_t reg;

    reg = readl(IC_CONF);
    reg &= ~(IC_CONF_PRPVF_EN | IC_CONF_PRPVF_CMB| IC_CONF_PRPVF_CSC2 | IC_CONF_PRPVF_CSC1);
    writel(reg, IC_CONF);
}


void _ipu_ic_init_rotate_vf(ipu_channel_params_t * params)
{
}

void _ipu_ic_uninit_rotate_vf(void)
{
    uint32_t reg;
    reg = readl(IC_CONF);
    reg &= ~IC_CONF_PRPVF_ROT_EN;
    writel(reg, IC_CONF);
}

void _ipu_ic_init_csi(ipu_channel_params_t * params)
{
    uint32_t reg;
    reg = readl(IC_CONF);
    reg &= ~IC_CONF_CSI_MEM_WR_EN;
    writel(reg, IC_CONF);
}

void _ipu_ic_uninit_csi(void)
{
    uint32_t reg;
    reg = readl(IC_CONF);
    reg &= ~(IC_CONF_RWS_EN | IC_CONF_PRPENC_EN);
    writel(reg, IC_CONF);
}

void _ipu_ic_init_prpenc(ipu_channel_params_t * params, bool src_is_csi)
{
    uint32_t reg, ic_conf;
    uint32_t downsizeCoeff, resizeCoeff;
    ipu_color_space_t in_fmt, out_fmt;

    // Setup vertical resizing
    _calc_resize_coeffs(params->mem_prp_enc_mem.in_height,
                     params->mem_prp_enc_mem.out_height,
                     &resizeCoeff,
                     &downsizeCoeff);
    reg = (downsizeCoeff << 30) | (resizeCoeff << 16);

    // Setup horizontal resizing
    _calc_resize_coeffs(params->mem_prp_enc_mem.in_width,
                     params->mem_prp_enc_mem.out_width,
                     &resizeCoeff,
                     &downsizeCoeff);
    reg |= (downsizeCoeff << 14) | resizeCoeff;

    writel(reg, IC_PRP_ENC_RSC);

    ic_conf = readl(IC_CONF);

    // Setup color space conversion
    in_fmt = format_to_colorspace(params->mem_prp_enc_mem.in_pixel_fmt);
    out_fmt = format_to_colorspace(params->mem_prp_enc_mem.out_pixel_fmt);
    if ( in_fmt == RGB )
    {
        if ( (out_fmt == YCbCr) || (out_fmt == YUV) )
        {
            // TODO: ERROR!
        }
    }
    if ( (in_fmt == YCbCr) || (in_fmt == YUV) )
    {
        if ( out_fmt == RGB )
        {
            _init_csc(IC_TASK_ENCODER, YCbCr, RGB);
            ic_conf |= IC_CONF_PRPENC_CSC1;    // Enable YCBCR->RGB CSC
        }
        else
        {
            // TODO: Support YUV<->YCbCr conversion?
        }
    }
    
    if (src_is_csi)
    {
        ic_conf &= ~IC_CONF_RWS_EN;
    }
    else
    {
        ic_conf |= IC_CONF_RWS_EN;
    }

    writel(ic_conf, IC_CONF);
}

void _ipu_ic_uninit_prpenc(void)
{
    uint32_t reg;

    reg = readl(IC_CONF);
    reg &= ~(IC_CONF_PRPENC_EN | IC_CONF_PRPENC_CSC1);
    writel(reg, IC_CONF);
}

void _ipu_ic_init_rotate_enc(ipu_channel_params_t * params)
{
}


void _ipu_ic_uninit_rotate_enc(void)
{
    uint32_t reg;

    reg = readl(IC_CONF);
    reg &= ~(IC_CONF_PRPENC_ROT_EN);
    writel(reg, IC_CONF);
}


void _ipu_ic_init_pp(ipu_channel_params_t * params)
{
    uint32_t reg, ic_conf;
    uint32_t downsizeCoeff, resizeCoeff;
    ipu_color_space_t in_fmt, out_fmt;

    // Setup vertical resizing
    _calc_resize_coeffs(params->mem_pp_mem.in_height,
                     params->mem_pp_mem.out_height,
                     &resizeCoeff,
                     &downsizeCoeff);
    reg = (downsizeCoeff << 30) | (resizeCoeff << 16);

    // Setup horizontal resizing
    _calc_resize_coeffs(params->mem_pp_mem.in_width,
                     params->mem_pp_mem.out_width,
                     &resizeCoeff,
                     &downsizeCoeff);
    reg |= (downsizeCoeff << 14) | resizeCoeff;

    writel(reg, IC_PP_RSC);

    ic_conf = readl(IC_CONF);

    // Setup color space conversion
    in_fmt = format_to_colorspace(params->mem_pp_mem.in_pixel_fmt);
    out_fmt = format_to_colorspace(params->mem_pp_mem.out_pixel_fmt);
    if ( in_fmt == RGB )
    {
        if ( (out_fmt == YCbCr) || (out_fmt == YUV) )
        {
            _init_csc(IC_TASK_POST_PROCESSOR, RGB, out_fmt);
            ic_conf |= IC_CONF_PP_CSC2;    // Enable RGB->YCBCR CSC
        }
    }
    if ( (in_fmt == YCbCr) || (in_fmt == YUV) )
    {
        if ( out_fmt == RGB )
        {
            _init_csc(IC_TASK_POST_PROCESSOR, YCbCr, RGB);
            ic_conf |= IC_CONF_PP_CSC1;    // Enable YCBCR->RGB CSC
        }
        else
        {
            // TODO: Support YUV<->YCbCr conversion?
        }
    }
    
    if (params->mem_pp_mem.graphics_combine_en)
    {
        ic_conf |= IC_CONF_PP_CMB;
    }
    
    writel(ic_conf, IC_CONF);
}

void _ipu_ic_uninit_pp(void)
{
    uint32_t reg;

    reg = readl(IC_CONF);
    reg &= ~(IC_CONF_PP_EN | IC_CONF_PP_CSC1 | IC_CONF_PP_CSC2);
    writel(reg, IC_CONF);
}

void _ipu_ic_init_rotate_pp(ipu_channel_params_t * params)
{
}

void _ipu_ic_uninit_rotate_pp(void)
{
    uint32_t reg;
    reg = readl(IC_CONF);
    reg &= ~IC_CONF_PP_ROT_EN;
    writel(reg, IC_CONF);
}


static void _init_csc(uint8_t ic_task, ipu_color_space_t in_format, ipu_color_space_t out_format)
{
/*     Y = R *  .299 + G *  .587 + B *  .114;
       U = R * -.169 + G * -.332 + B *  .500 + 128.;
       V = R *  .500 + G * -.419 + B * -.0813 + 128.;*/
    static const uint32_t rgb2ycbcr_coeff[4][3] = {
        { 0x004D, 0x0096, 0x001D },
        { 0x01D5, 0x01AB, 0x0080 },
        { 0x0080, 0x0195, 0x01EB },
        { 0x0000, 0x0200, 0x0200 },  // A0, A1, A2
    };            

/*     R = Y + (1.4075 * (V - 128));
       G = Y - (0.3455 * (U - 128) - (0.7169 * (V - 128));
       B = Y + (1.7790 * (U - 128); 
    static const uint32_t ycbcr2rgb_coeff_prp[4][3] = {
        { 0x0080, 0x0000, 0x00B4 },
        { 0x0080, 0x01D4, 0x01A5 },
        { 0x0080, 0x00E3, 0x0000 },
        { 0x1E99, 0x010F, 0x1E3A },  // A0, A1, A2
    };            
*/
/*     R = (1.164 * (Y - 16)) + (1.596 * (Cr - 128));
       G = (1.164 * (Y - 16)) - (0.392 * (Cb - 128)) - (0.813 * (Cr - 128));
       B = (1.164 * (Y - 16)) + (2.017 * (Cb - 128); */
    static const uint32_t ycbcr2rgb_coeff[4][3] = {
        { 149,   0, 204 },
        { 149, 462, 408 },
        { 149, 255,   0 },
        { 8192 - 446, 266, 8192 - 554 },  // A0, A1, A2
/* 
        { 128,   0, 175 },
        { 128, 468, 422 },
        { 128, 221,   0 },
        { 8192 - 443, 265, 8192 - 351 },  // A0, A1, A2*/
    };            

    uint32_t param[2];
    uint32_t address = 0;

    if ( ic_task == IC_TASK_VIEWFINDER )
    {
        address = 0x5A5 << 3;
    }
    else if ( ic_task == IC_TASK_ENCODER )
    {
        address = 0x2D1 << 3;
    }
    else if ( ic_task == IC_TASK_POST_PROCESSOR )
    {
        address = 0x87C << 3;
    }
    else
    {
        MXC_ERR_CHK(1);
        return;
    }

    if ( (in_format == YCbCr) && (out_format == RGB) )
    {
        // Init CSC1 (YCbCr->RGB)
        param[0] = (ycbcr2rgb_coeff[3][0] << 27) | (ycbcr2rgb_coeff[0][0] << 18) | 
                   (ycbcr2rgb_coeff[1][1] << 9) | ycbcr2rgb_coeff[2][2];
        param[1] = (ycbcr2rgb_coeff[3][0] >> 5) | (2L << (40-32)); // scale = 2, sat = 0
        _ipu_write_param_mem( address, param, 2 );
        DPRINTK("addr 0x%04X: word0 = 0x%08X, word1 = 0x%08X\n",
            address, param[0], param[1]);
        
        param[0] = (ycbcr2rgb_coeff[3][1] << 27) | (ycbcr2rgb_coeff[0][1] << 18) | 
                   (ycbcr2rgb_coeff[1][0] << 9) | ycbcr2rgb_coeff[2][0];
        param[1] = (ycbcr2rgb_coeff[3][1] >> 5);
        address += 1L << 3;
        _ipu_write_param_mem( address, param, 2 );
        DPRINTK("addr 0x%04X: word0 = 0x%08X, word1 = 0x%08X\n",
            address, param[0], param[1]);

        param[0] = (ycbcr2rgb_coeff[3][2] << 27) | (ycbcr2rgb_coeff[0][2] << 18) | 
                   (ycbcr2rgb_coeff[1][2] << 9) | ycbcr2rgb_coeff[2][1];
        param[1] = (ycbcr2rgb_coeff[3][2] >> 5);
        address += 1L << 3;
        _ipu_write_param_mem( address, param, 2 );
        DPRINTK("addr 0x%04X: word0 = 0x%08X, word1 = 0x%08X\n",
            address, param[0], param[1]);
    }
    else if ( (in_format == RGB) && (out_format == YCbCr) )
    {
        // Init CSC1 (RGB->YCbCr)
        param[0] = (rgb2ycbcr_coeff[3][0] << 27) | (rgb2ycbcr_coeff[0][0] << 18) | 
                   (rgb2ycbcr_coeff[1][1] << 9) | rgb2ycbcr_coeff[2][2];
        param[1] = (rgb2ycbcr_coeff[3][0] >> 5) | (1UL << 8); // scale = 1, sat = 0
        _ipu_write_param_mem( address, param, 2 );
        DPRINTK("addr 0x%04X: word0 = 0x%08X, word1 = 0x%08X\n",
            address, param[0], param[1]);

        param[0] = (rgb2ycbcr_coeff[3][1] << 27) | (rgb2ycbcr_coeff[0][1] << 18) | 
                   (rgb2ycbcr_coeff[1][0] << 9) | rgb2ycbcr_coeff[2][0];
        param[1] = (rgb2ycbcr_coeff[3][1] >> 5);
        address += 1L << 3;
        _ipu_write_param_mem( address, param, 2 );
        DPRINTK("addr 0x%04X: word0 = 0x%08X, word1 = 0x%08X\n",
            address, param[0], param[1]);

        param[0] = (rgb2ycbcr_coeff[3][2] << 27) | (rgb2ycbcr_coeff[0][2] << 18) | 
                   (rgb2ycbcr_coeff[1][2] << 9) | rgb2ycbcr_coeff[2][1];
        param[1] = (rgb2ycbcr_coeff[3][2] >> 5);
        address += 1L << 3;
        _ipu_write_param_mem( address, param, 2 );
        DPRINTK("addr 0x%04X: word0 = 0x%08X, word1 = 0x%08X\n",
            address, param[0], param[1]);
    }
    else
    {
        DPRINTK("TODO: Unsupported color space conversion\n");
    }

}

static bool _calc_resize_coeffs(uint32_t inSize, uint32_t outSize, uint32_t * resizeCoeff, uint32_t * downsizeCoeff)
{
    uint32_t tempSize;
    uint32_t tempDownsize;

    // Cannot downsize more than 8:1
    if ((outSize << 3) < inSize) 
    {
            return false;
    }

    // compute downsizing coefficient
    tempDownsize = 0;
    tempSize = inSize;
    while ((tempSize >= outSize*2) && (tempDownsize < 3)) {
            tempSize >>= 1;
            tempDownsize++;
    }
    *downsizeCoeff = tempDownsize;

    // compute resizing coefficient using the following equation:
    //      resizeCoeff = M*(SI -1)/(SO - 1)
    //      where M = 2^13, SI - input size, SO - output size
    *resizeCoeff = (8192L * (tempSize - 1)) / (outSize - 1);
    if (*resizeCoeff >= 16384L)
    {
        DPRINTK("Warning! Overflow on resize coeff.\n");
        *resizeCoeff = 0x3FFF;
    }

/*    DPRINTK("resizing from %d -> %d pixels, " \
              "downsize=%d, resize=%f\n", inSize, outSize, 
              *downsizeCoeff, 8192.0f / *resizeCoeff);
*/
    return true;
}

