/*
 * Copyright 2005 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * helper functions for IPU parameter memory access.
 *
 */

#ifndef _INCLUDE_IPU_PARAM_MEM_H_
#define _INCLUDE_IPU_PARAM_MEM_H_

#include <linux/types.h>

static __inline void _ipu_ch_param_set_size(uint32_t * params, uint32_t pixel_fmt,
                                      uint16_t width, uint16_t height, 
                                      uint16_t stride)
{
    uint32_t u_offset = 0;
    uint32_t v_offset = 0;
    memset(params, 0, 40);

    params[3] = (uint32_t)((width-1) << 12) | ((uint32_t)(height-1) << 24);
    params[4] = (uint32_t)(height-1) >> 8;
    params[7] = (uint32_t)(stride - 1) << 3;

    switch (pixel_fmt) {
        case IPU_PIX_FMT_GENERIC:
            //Represents Generic data
            params[7] |= 3 | (7UL << (81-64)) | (31L << (89-64));   // BPP & PFS
            params[8] = 2;                      // SAT = use 32-bit access
            break;
        case IPU_PIX_FMT_RGB565:
            params[7] |= 2L | (4UL << (81-64)) | (7L << (89-64));        // BPP & PFS
            params[8] = 2 |                     // SAT = 32-bit access
                        (0UL  <<(99-96)) |      // Red bit offset
                        (5UL  <<(104-96)) |     // Green bit offset
                        (11UL <<(109-96)) |     // Blue bit offset
                        (16UL <<(114-96)) |     // Alpha bit offset
                        (4UL  <<(119-96)) |     // Red bit width - 1
                        (5UL  <<(122-96)) |     // Green bit width - 1
                        (4UL  <<(125-96));      // Blue bit width - 1
            break;
        case IPU_PIX_FMT_BGR24:
            params[7] |= 1 | (4UL << (81-64)) | (7L << (89-64)); // 24 BPP & RGB PFS
            params[8] = 2 |                     // SAT = 32-bit access
                        (8UL  <<(104-96)) |     // Green bit offset
                        (16UL <<(109-96)) |     // Blue bit offset
                        (24UL << (114-96)) |    // Alpha bit offset
                        (7UL  <<(119-96)) |     // Red bit width - 1
                        (7UL  <<(122-96)) |     // Green bit width - 1
                        (uint32_t)(7UL  <<(125-96));    // Blue bit width - 1
            break;
        case IPU_PIX_FMT_BGRA32:
            params[9] = 7;                      // Alpha bit width - 1
            // Fall thru
        case IPU_PIX_FMT_BGR32:
            params[7] |= 0 | (4UL << (81-64)) | (7 << (89-64)); // BPP & pixel fmt
            params[8] = 2 |                     // SAT = 32-bit access
                        (8UL  <<(99-96))  |     // Red bit offset
                        (16UL <<(104-96)) |     // Green bit offset
                        (24UL <<(109-96)) |     // Blue bit offset
                        (0UL  <<(114-96)) |     // Alpha bit offset
                        (7UL  <<(119-96)) |     // Red bit width - 1
                        (7UL  <<(122-96)) |     // Green bit width - 1
                        (uint32_t)(7UL  <<(125-96));    // Blue bit width - 1
            break;
        case IPU_PIX_FMT_YUYV:
            params[7] |= 2 | (6UL << 17) | (7 << (89-64)); // BPP & pixel format
            params[8] = 2;                      // SAT = 32-bit access
            break;
        case IPU_PIX_FMT_YUV420P2:
        case IPU_PIX_FMT_YUV420P:
            params[7] |= 3 | (3UL << 17) | (7 << (89-64)); // BPP & pixel format
            params[8] = 2;                      // SAT = 32-bit access
            u_offset = stride * height;
            v_offset = u_offset + u_offset/4;
            break;
        case IPU_PIX_FMT_YVU422P:
            params[7] |= 3 | (2UL << 17) | (7 << (89-64)); // BPP & pixel format
            params[8] = 2;                      // SAT = 32-bit access
            v_offset = stride * height;
            u_offset = v_offset + v_offset/2;
            break;
        default:
            printk("mxc ipu: unimplemented pixel format\n");
            break;
    }

    params[1] = u_offset << (53-32);
    params[2] = u_offset >> (64-53);
    params[2] |= v_offset << (79-64);
    params[3] |= v_offset >> (96-79);
}

static __inline void _ipu_ch_param_set_burst_size(uint32_t * params, uint16_t burst_pixels)
{
    params[7] &= ~(0x3FL << (89-64));
    params[7] |= (uint32_t)(burst_pixels - 1) << (89-64);
};        

static __inline void _ipu_ch_param_set_buffer(uint32_t * params, void * buf0, void * buf1)
{
    params[5] = (uint32_t)buf0;
    params[6] = (uint32_t)buf1;
};

static __inline void _ipu_ch_param_set_rotation(uint32_t * params, ipu_rotate_mode_t rot)
{
    params[7] |= (uint32_t)rot << (84-64);
};

void _ipu_write_param_mem(uint32_t addr, uint32_t * data, uint32_t numWords);

#endif

