/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Driver for the Freescale Semiconductor MXC IPU
 *
 */

/*!
 * @file ipu_prp_enc.c
 * 
 * @brief IPU Use case for PRP-ENC
 * 
 * @ingroup IPU
 */

#include <linux/interrupt.h>
#include "../drivers/media/video/mxc_v4l2_capture.h"
#include "../ipu.h"
#include "../ipu_prv.h"
#include "asm/arch/board.h"

static ipu_rotate_mode_t grotation = IPU_ROTATE_NONE;

/*
 * Function definitions
 */

/*!
 * PrpENC interface - Set Format function
 *
 * @param cam       struct cam_data * mxc capture instance
 *
 * @return  None
 */
static void 
prp_enc_set_format(cam_data *cam)
{
        ipu_channel_params_t enc;
        void *dummy = (void *)0xdeadbeaf;

        FUNC_START;

        if (!cam) {
                DPRINTK("prp_enc_set_format private is NULL\n");
                return;
        }

        enc.csi_prp_enc_mem.in_width = MXC_SENSOR_SCREEN_WIDTH;
        enc.csi_prp_enc_mem.in_height = MXC_SENSOR_SCREEN_HEIGHT;
        enc.csi_prp_enc_mem.in_pixel_fmt = IPU_PIX_FMT_YUYV;
        enc.csi_prp_enc_mem.out_width = cam->v2f.fmt.pix.width;
        enc.csi_prp_enc_mem.out_height = cam->v2f.fmt.pix.height;
        if (cam->v2f.fmt.pix.pixelformat == V4L2_PIX_FMT_YUV420) {
                enc.csi_prp_enc_mem.out_pixel_fmt = IPU_PIX_FMT_YUV420P;
                DPRINTK("prp_enc_set_format YUV420\n");
        }  
        else if (cam->v2f.fmt.pix.pixelformat == V4L2_PIX_FMT_BGR24) {
                enc.csi_prp_enc_mem.out_pixel_fmt = IPU_PIX_FMT_BGR24;
                DPRINTK("prp_enc_set_format BGR24\n");
        }  
        else if (cam->v2f.fmt.pix.pixelformat == V4L2_PIX_FMT_RGB565) {
                enc.csi_prp_enc_mem.out_pixel_fmt = IPU_PIX_FMT_RGB565;
                DPRINTK("prp_enc_set_format RGB565\n");
        }  
        else if (cam->v2f.fmt.pix.pixelformat == V4L2_PIX_FMT_BGR32) {
                enc.csi_prp_enc_mem.out_pixel_fmt = IPU_PIX_FMT_BGR32;
                DPRINTK("prp_enc_set_format BGR32\n");
        }  
        else
                DPRINTK("prp_enc_set_format format not supported\n");

        ipu_uninit_channel(CSI_PRP_ENC_MEM);
        ipu_init_channel(CSI_PRP_ENC_MEM, &enc);

        grotation = cam->rotation;
        if (cam->rotation >= IPU_ROTATE_90_RIGHT) {
                if (cam->rot_enc_bufs[0]) {
                        IPU_Free((u32)(cam->rot_enc_bufs[0]));
                }
                if (cam->rot_enc_bufs[1]) {
                        IPU_Free((u32)(cam->rot_enc_bufs[1]));
                }
                cam->rot_enc_bufs[0] = (void *) 
                        IPU_Malloc(PAGE_ALIGN(cam->v2f.fmt.pix.sizeimage));
                if (!cam->rot_enc_bufs[0])	{
                        DPRINTK("prp_enc_set_format: alloc enc_bufs0\n");
                        return;
                }
                cam->rot_enc_bufs[1] = (void *) 
                        IPU_Malloc(PAGE_ALIGN(cam->v2f.fmt.pix.sizeimage));
                if (!cam->rot_enc_bufs[1])	{
                        IPU_Free((u32)(cam->rot_enc_bufs[0]));
                        DPRINTK("prp_enc_set_format: alloc enc_bufs1\n");
                        return;
                }

                if (ipu_init_channel_buffer(CSI_PRP_ENC_MEM, IPU_OUTPUT_BUFFER, 
                        enc.csi_prp_enc_mem.out_pixel_fmt, 
                        enc.csi_prp_enc_mem.out_height, 
                        enc.csi_prp_enc_mem.out_width, 
                        enc.csi_prp_enc_mem.out_height, 
                        IPU_ROTATE_NONE, cam->rot_enc_bufs[0], 
                        cam->rot_enc_bufs[1]) != 0) {
                        DPRINTK("prp_enc_set_format:" 
                                "CSI_PRP_ENC_MEM input buffer\n");
                        return;
                }

                ipu_uninit_channel(MEM_ROT_ENC_MEM);
                if (ipu_init_channel(MEM_ROT_ENC_MEM, NULL) != 0) {
                        DPRINTK("prp_enc_set_format: "
                                 "MEM_ROT_ENC_MEM channel\n");
                        return;
                }

                if (ipu_init_channel_buffer(MEM_ROT_ENC_MEM, IPU_INPUT_BUFFER, 
                        enc.csi_prp_enc_mem.out_pixel_fmt, 
                        enc.csi_prp_enc_mem.out_height, 
                        enc.csi_prp_enc_mem.out_width, 
                        enc.csi_prp_enc_mem.out_height, 
                        cam->rotation, cam->rot_enc_bufs[0],
                        cam->rot_enc_bufs[1]) != 0) {
                        DPRINTK("prp_enc_set_format: Error" 
                                "MEM_ROT_ENC_MEM input buffer\n");
                        return;
                }

                if (ipu_init_channel_buffer(MEM_ROT_ENC_MEM, IPU_OUTPUT_BUFFER, 
                        enc.csi_prp_enc_mem.out_pixel_fmt, 
                        enc.csi_prp_enc_mem.out_width, 
                        enc.csi_prp_enc_mem.out_height,	
                        enc.csi_prp_enc_mem.out_width, IPU_ROTATE_NONE, 
                        dummy, dummy) != 0) {
                        DPRINTK("prp_enc_set_format: Error " 
                                "MEM_ROT_ENC_MEM output buffer\n");
                        return;
                }
        }
        else {
                ipu_init_channel_buffer(CSI_PRP_ENC_MEM, IPU_OUTPUT_BUFFER, 
                        enc.csi_prp_enc_mem.out_pixel_fmt, 
                        enc.csi_prp_enc_mem.out_width, 
                        enc.csi_prp_enc_mem.out_height,	
                        enc.csi_prp_enc_mem.out_width, 
                        cam->rotation, dummy, dummy);
        }
                
        FUNC_END;
}

/*!
 * function to update physical buffer address for encorder IDMA channel
 *
 * @param eba         u32 physical buffer address for encorder IDMA channel
 * @param buffer_num  int buffer 0 or buffer 1
 *
 * @return  None
 */
static void 
prp_enc_eba_update(u32 eba, int *buffer_num)
{
        FUNC_START;

        DPRINTK("prp_enc_eba_update: eba %x\n", eba);
        if (grotation >= IPU_ROTATE_90_RIGHT) {
                ipu_update_channel_buffer(MEM_ROT_ENC_MEM, IPU_OUTPUT_BUFFER, 
                                          *buffer_num, (u8 *)(eba));
        }
        else {
                ipu_update_channel_buffer(CSI_PRP_ENC_MEM, IPU_OUTPUT_BUFFER, 
                                          *buffer_num, (u8 *)(eba));
        } 
        ipu_select_buffer(CSI_PRP_ENC_MEM, IPU_OUTPUT_BUFFER, *buffer_num);

        *buffer_num = (*buffer_num == 0) ? 1 : 0;
        FUNC_END;
}

/*!
 * Enable encoder task 
 * @param private       struct cam_data * mxc capture instance
 *
 * @return  None
 */
static void 
prp_enc_enabling_tasks(void *private)
{ 
        cam_data *cam = (cam_data *) private;
        FUNC_START;

        prp_enc_set_format(cam);

        ipu_enable_channel(CSI_PRP_ENC_MEM);
        if (cam->rotation >= IPU_ROTATE_90_RIGHT) {
                ipu_enable_channel(MEM_ROT_ENC_MEM);
        }
        
        FUNC_END;
}

/*!
 * Disable encoder task 
 * @param private       struct cam_data * mxc capture instance
 *
 * @return  None
 */
static void 
prp_enc_disabling_tasks(void *private)
{ 
        cam_data *cam = (cam_data *) private;
        FUNC_START;
        
        ipu_uninit_channel(CSI_PRP_ENC_MEM);
        if (cam->rotation >= IPU_ROTATE_90_RIGHT) {
                ipu_uninit_channel(MEM_ROT_ENC_MEM);
        }

        ipu_disable_channel(CSI_PRP_ENC_MEM, true);
        if (cam->rotation >= IPU_ROTATE_90_RIGHT) {
                ipu_disable_channel(MEM_ROT_ENC_MEM, true);
        }

        FUNC_END;
}

/*!
 * IPU ENC callback function.
 *		 
 * @param irq       int irq line
 * @param dev_id    void * device id
 * @param regs      struct pt_regs *
 *		 
 * @return status   IRQ_HANDLED for handled
 */
static irqreturn_t 
prp_enc_callback(int irq, void *dev_id, struct pt_regs *regs)
{
        cam_data *cam = (cam_data *) dev_id;
        FUNC_START;

        if (grotation >= IPU_ROTATE_90_RIGHT) {
                ipu_select_buffer(MEM_ROT_ENC_MEM, IPU_INPUT_BUFFER, 
                                  cam->ping_pong_rot);
                ipu_select_buffer(MEM_ROT_ENC_MEM, IPU_OUTPUT_BUFFER, 
                                  cam->ping_pong_rot);
                cam->ping_pong_rot = (cam->ping_pong_rot == 0) ? 1 : 0;
        } 
        else {
                camera_callback(irq, dev_id);
        } 
        FUNC_END;

        return IRQ_HANDLED;
}

/*!
 * IPU ENC rotationcallback function.
 *		 
 * @param irq       int irq line
 * @param dev_id    void * device id
 * @param regs      struct pt_regs *
 *		 
 * @return status   IRQ_HANDLED for handled
 */
static irqreturn_t 
prp_enc_rot_callback(int irq, void *dev_id, struct pt_regs *regs)
{
        FUNC_START;

        if (grotation >= IPU_ROTATE_90_RIGHT) {
                camera_callback(irq, dev_id);
        } 
        else {
                DPRINTK("prp_enc_rot_callback :Error\n"); 
        } 
        FUNC_END;

        return IRQ_HANDLED;
}

/*!
 * function to select PRP-ENC as the working path
 *
 * @param private       struct cam_data * mxc capture instance
 *
 * @return  None
 */
void 
prp_enc_select(void *private)
{
        cam_data *cam = (cam_data *) private;

        FUNC_START;
        if (cam) {
                cam->enc_update_eba = prp_enc_eba_update;
                cam->enc_enable = prp_enc_enabling_tasks;
                cam->enc_disable = prp_enc_disabling_tasks;
        }

        if (ipu_request_irq(IPU_IRQ_PRP_ENC_OUT_EOF, prp_enc_callback, 
                            0, "Mxc Camera", cam) != 0) {
                DPRINTK("prp_enc_select: Error registering irq.\n");
        }

        if (ipu_request_irq(IPU_IRQ_PRP_ENC_ROT_OUT_EOF, prp_enc_rot_callback, 
                            0, "Mxc Camera", cam) != 0) {
                DPRINTK("prp_enc_select: Error registering rot irq\n");
        }
 
        FUNC_END;
}

/*!
 * function to de-select PRP-ENC as the working path
 *
 * @param private       struct cam_data * mxc capture instance
 *
 * @return  None
 */
void 
prp_enc_deselect(void *private)
{
        cam_data *cam = (cam_data *) private;
        FUNC_START;

        prp_enc_disabling_tasks(cam);

        ipu_disable_irq(IPU_IRQ_PRP_ENC_OUT_EOF);
        ipu_free_irq(IPU_IRQ_PRP_ENC_OUT_EOF, cam);

        ipu_disable_irq(IPU_IRQ_PRP_ENC_ROT_OUT_EOF);
        ipu_free_irq(IPU_IRQ_PRP_ENC_ROT_OUT_EOF, cam);

        if (cam) {
                cam->enc_update_eba = NULL;
                cam->enc_enable = NULL;
                cam->enc_disable = NULL;
                if (cam->rot_enc_bufs[0]) {
                        IPU_Free((u32)(cam->rot_enc_bufs[0]));
                        cam->rot_enc_bufs[0] = NULL; 
                }
                if (cam->rot_enc_bufs[1]) {
                        IPU_Free((u32)(cam->rot_enc_bufs[1]));
                        cam->rot_enc_bufs[1] = NULL; 
                }
        }

        FUNC_END;
}

/*!
 * Init the Encorder channels
 *
 * @return  Error code indicating success or failure
 */
__init int 
prp_enc_init(void)
{
        FUNC_START;

        FUNC_END;
        return 0;
}

/*!
 * Deinit the Encorder channels
 *
 */
void __exit 
prp_enc_exit(void)
{
        FUNC_START;

        FUNC_END;
}

module_init(prp_enc_init);
module_exit(prp_enc_exit);

EXPORT_SYMBOL(prp_enc_select);
EXPORT_SYMBOL(prp_enc_deselect);

MODULE_AUTHOR("Freescale Semiconductor");
MODULE_DESCRIPTION("IPU PRP ENC Driver");
MODULE_LICENSE("GPL");
