/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Driver for the Freescale Semiconductor MXC IPU
 *
 */

/*!
 * @file ipu_still.c
 * 
 * @brief IPU Use case for still image capture
 * 
 * @ingroup IPU
 */

#include <asm/semaphore.h>
#include <linux/interrupt.h>
#include "../drivers/media/video/mxc_v4l2_capture.h"
#include "../ipu.h"
#include "../ipu_prv.h"

static int callback_flag;
/*
 * Function definitions
 */

/*!
 * CSI EOF callback function.
 *		 
 * @param irq       int irq line
 * @param dev_id    void * device id
 * @param regs      struct pt_regs *
 *		 
 * @return status   IRQ_HANDLED for handled
 */
static irqreturn_t 
prp_csi_eof_callback(int irq, void *dev_id, struct pt_regs *regs)
{
        FUNC_START; 
        if (callback_flag == 2) {
                ipu_select_buffer(CSI_MEM, IPU_OUTPUT_BUFFER, 0);
                ipu_enable_channel(CSI_MEM);
        }

        callback_flag++;
        FUNC_END;
        return IRQ_HANDLED;
}

/*!
 * start csi->mem task 
 * @param private       struct cam_data * mxc capture instance
 *
 * @return  None
 */
static void 
prp_still_start(void *private)
{ 
        cam_data *cam = (cam_data *) private;
        FUNC_START;

        ipu_uninit_channel(CSI_MEM);
        ipu_init_channel(CSI_MEM, NULL);

        ipu_init_channel_buffer(CSI_MEM, IPU_OUTPUT_BUFFER, 
                IPU_PIX_FMT_YUYV, cam->v2f.fmt.pix.width, 
                cam->v2f.fmt.pix.height, cam->v2f.fmt.pix.width, 
                IPU_ROTATE_NONE, cam->still_buf, NULL);
                     
        callback_flag = 0;
        if (ipu_request_irq(IPU_IRQ_SENSOR_EOF, prp_csi_eof_callback, 
                            0, "Mxc Camera", cam) != 0) {
                DPRINTK("prp_still_start: Error IPU_IRQ_SENSOR_EOF \n");
        }

        FUNC_END;
}

/*!
 * stop csi->mem encoder task 
 * @param private       struct cam_data * mxc capture instance
 *
 * @return  None
 */
static void 
prp_still_stop(void *private)
{ 
        FUNC_START;
        
        ipu_disable_irq(IPU_IRQ_SENSOR_EOF);
        ipu_free_irq(IPU_IRQ_SENSOR_EOF, private);

        ipu_uninit_channel(CSI_MEM);
        ipu_disable_channel(CSI_MEM, true);

        FUNC_END;
}

/*!
 * CSI callback function.
 *		 
 * @param irq       int irq line
 * @param dev_id    void * device id
 * @param regs      struct pt_regs *
 *		 
 * @return status   IRQ_HANDLED for handled
 */
static irqreturn_t 
prp_still_callback(int irq, void *dev_id, struct pt_regs *regs)
{
        cam_data *cam = (cam_data *) dev_id;
        FUNC_START;

        cam->still_counter++; 
        wake_up_interruptible(&cam->still_queue);
        //printk(" prp_idma7_eof_callback\n");

        FUNC_END;
        return IRQ_HANDLED;
}

/*!
 * function to select CSI_MEM as the working path
 *
 * @param private       struct cam_data * mxc capture instance
 *
 * @return  None
 */
void 
prp_still_select(void *private)
{
        cam_data *cam = (cam_data *) private;

        FUNC_START;
        if (cam) {
                cam->csi_start = prp_still_start;
                cam->csi_stop = prp_still_stop;
        }

        if (ipu_request_irq(IPU_IRQ_SENSOR_OUT_EOF, prp_still_callback, 
                            0, "Mxc Camera", cam) != 0) {
                DPRINTK("prp_still_select: Error registering irq.\n");
        }

        FUNC_END;
}

/*!
 * function to de-select CSI_MEM as the working path
 *
 * @param private       struct cam_data * mxc capture instance
 *
 * @return  None
 */
void 
prp_still_deselect(void *private)
{
        cam_data *cam = (cam_data *) private;
        FUNC_START;

        prp_still_stop(cam);

        ipu_disable_irq(IPU_IRQ_SENSOR_OUT_EOF);
        ipu_free_irq(IPU_IRQ_SENSOR_OUT_EOF, cam);

        if (cam) {
                cam->csi_start = NULL;
                cam->csi_stop = NULL;
        }

        FUNC_END;
}

/*!
 * Init the Encorder channels
 *
 * @return  Error code indicating success or failure
 */
__init int 
prp_still_init(void)
{
        FUNC_START;
        FUNC_END;
        return 0;
}

/*!
 * Deinit the Encorder channels
 *
 */
void __exit 
prp_still_exit(void)
{
        FUNC_START;
        FUNC_END;
}

module_init(prp_still_init);
module_exit(prp_still_exit);

EXPORT_SYMBOL(prp_still_select);
EXPORT_SYMBOL(prp_still_deselect);

MODULE_AUTHOR("Freescale Semiconductor");
MODULE_DESCRIPTION("IPU PRP STILL IMAGE Driver");
MODULE_LICENSE("GPL");
