/*
 * Copyright 2005 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*!
 * @file dvfs.c
 *
 * @brief Driver for the Freescale Semiconductor MXC DVFS module.  
 * 
 * The DVFS driver
 * driver is designed as a character driver which interacts with the MXC DVFS 
 * hardware. Upon initialization, the DVFS driver initializes the DVFS hardware
 * sets up driver nodes attaches to the DVFS interrupt and initializes internal 
 * data structures. When the DVFS interrupt occurs the driver checks the cause 
 * of the interrupt (lower frequency, increase frequency or emergency) and changes 
 * the CPU voltage according to tarnslation table that is loaded into the driver.
 * The driver read method is used to read the log buffer.
 * Driver ioctls are used to change driver parameters and enable/disable the 
 * DVFS operation.
 *
 * @ingroup PM
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/arch/platform.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/workqueue.h>
#include <linux/proc_fs.h>
#include <asm/semaphore.h>
#include <linux/jiffies.h>
#include <linux/device.h>
#include <asm/arch/clock.h>
#include <asm/arch/gpio.h>
#include <asm/arch/sdma.h>

#include "dvfs.h"


/*!
 * This function sets the weight of general purpose signals
 */ 
static void set_gp_weight(int gp_id, int weight)
{
        if (gp_id < 9){
                SET_LTR3_WSW(gp_id,weight);
        } else {
                SET_LTR2_WSW(gp_id,weight);
        }
}

/*!
 * This function is called for module initialization.
 * It initializes the driver data structures and sets up the DVFS hardware.
 *
 * @return   0 to indicate success else returns a negative number. 
 *
 */   
int __init init_dvfs_controller(dvfs_dptc_params_s *params)
{
        int i;
        
        iomux_config_mux(MX31_PIN_DVFS0, OUTPUTCONFIG_FUNC, INPUTCONFIG_NONE);
        iomux_config_mux(MX31_PIN_DVFS1, OUTPUTCONFIG_FUNC, INPUTCONFIG_NONE);
        iomux_config_mux(MX31_PIN_GPIO1_5, OUTPUTCONFIG_GPIO, INPUTCONFIG_FUNC);

        /* Mask DVFS interrupts */
        MASK_DVFS_INT();
        MASK_DVFS_LB_INT();

        params->dvfs_log_buffer = sdma_malloc(DVFS_LB_SIZE);
        
        /* Set general purpose weights to 0 */
        for (i=0 ; i < 16 ; i++) {
                set_gp_weight(i, 0);
        }
        
        /* Set the div_3_clk clock to maximum if DVFS
         * is controlled by ARM
         */
        SET_DIV3CK(3);
        
#ifndef CONFIG_MXC_DVFS_SDMA
        /*
         * ARM will get DVFS interrupt
         */
        SET_DVFIS();
#endif        
        /*
         * Frequency decrease threshold
         */
        SET_DNTHR(18);
        /*
         * Frequency increase threshold
         */
        SET_UPTHR(30);
        /*
         * Panic threshold
         */
        SET_PNCTHR(63);

        SET_UPCNT(0x33);
        SET_DNCNT(0x33);

        //SET_DVFEV();

        SET_EMAC(0x8);

        /* Set ARM clock as CLK0 output */
        _reg_CCM_COSR = 0x00000306;

        printk(KERN_INFO "DVFS controller initialized\n");
        
        return 0;
}

/*!
 * This function enables the DVFS module. this function updates the DVFS
 * thresholds, updates the Atlas, unmasks the DVFS interrupt and enables 
 * the DVFS module
 *
 * @return      0 if DVFS module was enabled else returns -EINVAL. 
 *
 */   
int start_dvfs(dvfs_dptc_params_s *params)
{
        int i;
        
        /* Check if DPTC module isn't already active */
        if (params->dvfs_is_active == FALSE) {

                for (i = 12; i <= 15 ; i++ ) {                        
                        set_gp_weight(i,7);
                }
                
                UNMASK_DVFS_INT();
                /* UNMASK_DVFS_LB_INT(); */
                
                /* SET_LTBRSH(); */
                SET_LTBRSR();
                
                ENABLE_DVFS();        
                
                params->dvfs_is_active = TRUE;

                return 0;
        }

        /* DVFS module already active return error */
        return -EINVAL;
}

/*!
 * This function disables the DVFS module.
 * @return      0 if DVFS module was enabled else returns -EINVAL. 
 *
 */   
int stop_dvfs(dvfs_dptc_params_s *params)
{
        /* Check if DPTC module isn't already unactive (disabled) */
        if (params->dvfs_is_active != FALSE) { 

                DISABLE_DVFS();
                
                MASK_DVFS_INT();
                MASK_DVFS_LB_INT();
                
                params->dvfs_is_active = FALSE;
                                
                return 0;
        }

        /* DVFS module already unactive return error */
        return -EINVAL;
}

int set_sw_gp(unsigned char arg)
{
        /* Clear upper 4 bits */
        arg &= 0xf;
        SET_DVGP(arg);

        return 0;
}

#ifndef CONFIG_MXC_DVFS_SDMA
void make_step(dvfs_dptc_params_s *params, int fsvai)
{
        unsigned long pmcr0;
        int dvsup;
        unsigned long pdr0;
        unsigned long pll;
        unsigned long ltbr0;
        unsigned long ltbr1;
        unsigned long lbcf;
        int curr_freq_index;
        int pll_switch;
        int dfsup;

        pmcr0 = _reg_CCM_PMCR0;

        if (pmcr0 & FSVAIM) {
                /* Do nothing. DVFS interrupt is masked */
                return;
        }

        /* Log buffer section */
        if (GET_LBFL(pmcr0) == 1) {                
                lbcf = GET_LBCF(pmcr0);
                switch (lbcf) {
                case 0:
                        /* Log buffer size is 4 */
                        ltbr0 = _reg_CCM_LTBR0;
                        break;
                case 1:
                        /* Log buffer size is 8 */
                        ltbr0 = _reg_CCM_LTBR0;
                        break;
                case 2:
                        /* Log buffer size is 12 */
                        ltbr0 = _reg_CCM_LTBR0;
                        ltbr1 = _reg_CCM_LTBR1;
                        break;
                case 4:
                        /* Log buffer size is 16 */
                        ltbr0 = _reg_CCM_LTBR0;
                        ltbr1 = _reg_CCM_LTBR1;
                        break;        
                }
        }

        /* End of log buffer section */

        if (fsvai == 0) {
               /* Do nothing. Freq change is not required */
                return; 
        }

        if (! (pmcr0 & UPDTEN)) {
                /* Do nothing. DVFS didn't finish previous flow update */
                return;
        }

        if (((GET_DVSUP() == 3) && (fsvai == DVFS_DECREASE)) ||
            ((GET_DVSUP() == 0) && ((fsvai == DVFS_INCREASE) ||
             (fsvai == DVFS_EMERG)))) {
                /* Do nothing. DVFS is already at lowest (highest) state */
                return;
        }

        dvsup = GET_DVSUP();
        curr_freq_index = dvsup;
        
        if (fsvai == DVFS_DECREASE) {
                SET_UDSC(pmcr0, 0);
                dvsup++;
                SET_VSCNT(pmcr0, 1);
                pll_switch = params->dvfs_dptc_tables_ptr->table[curr_freq_index].pll_sw_down;
                pdr0 = params->dvfs_dptc_tables_ptr->table[curr_freq_index].pdr0_down;
                pll = params->dvfs_dptc_tables_ptr->table[curr_freq_index].pll_down;
                
        } else {
                SET_UDSC(pmcr0, 1);
                dvsup = 0;
                SET_VSCNT(pmcr0, params->dvfs_dptc_tables_ptr->table[curr_freq_index].vscnt);
                pll_switch = params->dvfs_dptc_tables_ptr->table[curr_freq_index].pll_sw_up;
                pdr0 = params->dvfs_dptc_tables_ptr->table[curr_freq_index].pdr0_up;
                pll = params->dvfs_dptc_tables_ptr->table[curr_freq_index].pll_up;
        }

        /* DVSUP setup */
        SET_DVSUP(pmcr0, dvsup);

        dfsup = GET_DFSUP();

        if (pll_switch == 0) {
                /* Update only pdr0 */
                dfsup |= 1;
        }

        if (pll_switch == 1) {
                /* Update pll and pdr0 */
                dfsup = ~dfsup;
                dfsup &= 2;
        }

        SET_DFSUP(pmcr0, dfsup);

        _reg_CCM_PMCR0 = pmcr0;

        //_reg_CCM_PDR0 = pdr0;
        mxc_ccm_modify_reg(MXC_CCM_PDR0, 0xffffffff, pdr0);

        if (dfsup == 2) {
                /* MPCTL will be updated */
                mxc_ccm_modify_reg(MXC_CCM_MPCTL, 0xffffffff, pll);
        } else {
                /* SRPCTL will be updated */
                mxc_ccm_modify_reg(MXC_CCM_SRPCTL, 0xffffffff, pll);
        }

        /*
          printk("ARM frequency: %dMHz (%d)\n",
          (int)mxc_get_clocks(CPU_CLK)/1000000,
          (int)jiffies);
        */
}

/*!
 * This function is the DVFS Interrupt handler.
 */ 
void dvfs_irq(dvfs_dptc_params_s *params)
{
        int fsvai;

        fsvai = GET_FSVAI();

        make_step(params, fsvai);
}
#endif /* CONFIG_MXC_DVFS_SDMA */

/*!
 * This function sets DVFS to monitor WFI signal
 * @return      0 on success, error coad on fail 
 */
int set_wfi(unsigned char arg)
{
        int res;
        
        if (arg == 0) {
                /* Disable */
                _reg_CCM_PMCR0 |= WFIM;
                res = 0;
        } else if (arg == 1) {
                /* Enable */
                _reg_CCM_PMCR0 &= ~WFIM;
                res = 0;  
        } else {
                res = -EINVAL;
        }

        return res;
}
  

MODULE_AUTHOR("Freescale Semiconductor Inc");
MODULE_DESCRIPTION("DVFS driver");
MODULE_LICENSE("GPL");

