/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*!
 * @file dvfs_dptc_struct.h
 *
 * @brief MXC dvfs & dptc structure definitions file. 
 * 
 * @ingroup PM
 */ 
#ifndef __DVFS_DPTC_STRUCT_H__
#define __DVFS_DPTC_STRUCT_H__

#include <asm/arch/pm_api.h>

/*!
 * Number of entries in the DPTC log buffer
 */
#define LOG_ENTRIES 1024

/*!
 * Log buffer Structure.\n
 * This structure records the DPTC changes. \n
 * This structure can be read by the user using the proc file system DPTC read entry.
 */
typedef struct
{
        /*!
         * Index to the head of the log buffer
         */
        int head;

        /*!
         * Index to the tail of the log buffer
         */
        int tail;

        /*!
         * Mutex to allow access to the log buffer
         */
        struct semaphore	mutex;

        /*!
         * Array of log buffer entries
         */
        dptc_log_entry_s	entries[LOG_ENTRIES];
} dptc_log_s;

/*!
 * DPTC driver data structure.\n
 * Holds all driver parameters and data structures.
 */
typedef struct
{
        /*!
         * This variable holds the current frequency index
         */
        int current_freq_index;       

        /*!
         * Boolean variable, if TRUE the DPTC module is enabled
         * if FALSE the DPTC module is disabled
         */
        int dptc_is_active;

#ifdef CONFIG_MXC_DVFS        
        /*!
         * Boolean variable, if TRUE the DVFS module is enabled
         * if FALSE the DPTC module is disabled
         */
        int dvfs_is_active;
#endif        

        /*!
         * Boolean variable, if TRUE the DPTC module is suspended
         */
        int suspended;

        unsigned char rc_state;

        /*!
         * Pointer to the DVFS & DPTC translation table
         */
        dvfs_dptc_tables_s *dvfs_dptc_tables_ptr;	

        /*!
         * The DPTC log buffer
         */
        dptc_log_s dptc_log_buffer;

        /*!
         * The DVFS log buffer
         */
        unsigned char *dvfs_log_buffer;

#ifdef CONFIG_MXC_DVFS_SDMA
        /*!
         * SDMA channel number
         */
        int sdma_channel;

        /*!
         * This holds the previous working point
         */
        int prev_wp;

#endif
} dvfs_dptc_params_s;

/*
 * CCM Registers
 */ 
#ifdef CONFIG_ARCH_MX3

#define    _reg_CCM_PDR0   \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x04))))
#define    _reg_CCM_MPCTL  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x10))))
#define    _reg_CCM_SRPCTL  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x18))))
#define    _reg_CCM_COSR  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x1C))))
#define    _reg_CCM_DCVR0  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x34))))
#define    _reg_CCM_DCVR1  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x38))))
#define    _reg_CCM_DCVR2  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x3C))))
#define    _reg_CCM_DCVR3  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x40))))
#define    _reg_CCM_LTR0   \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x44))))
#define    _reg_CCM_LTR1   \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x48))))
#define    _reg_CCM_LTR2   \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x4C))))
#define    _reg_CCM_LTR3   \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x50))))
#define    _reg_CCM_LTBR0  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x54))))
#define    _reg_CCM_LTBR1  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x58))))
#define    _reg_CCM_PMCR0  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x5C))))
#define    _reg_CCM_PMCR1  \
(*((volatile unsigned long *)(IO_ADDRESS(CCM_BASE_ADDR + 0x60))))

#define DPTEN		(unsigned long)0x00000001
#define PTVAI		(unsigned long)0x00000006
#define PTVAIM		(unsigned long)0x00000008
#define DVFEN		(unsigned long)0x00000010
#define DCR             (unsigned long)0x00000020
#define DRCE0           (unsigned long)0x00000040
#define DRCE1           (unsigned long)0x00000080
#define DRCE2           (unsigned long)0x00000100
#define DRCE3           (unsigned long)0x00000200
#define WFIM            (unsigned long)0x00000400
#define DPVV            (unsigned long)0x00000800
#define DPVCR           (unsigned long)0x00001000
#define FSVAI           (unsigned long)0x00006000
#define FSVAIM          (unsigned long)0x00008000
#define UPDTEN          (unsigned long)0x00010000
#define PTVIS           (unsigned long)0x00020000
#define LBCF            (unsigned long)0x000c0000
#define LBFL            (unsigned long)0x00100000
#define LBMI            (unsigned long)0x00200000
#define DVFIS           (unsigned long)0x00400000
#define DVFEV           (unsigned long)0x00800000
#define VSCNT           (unsigned long)0x07000000
#define UDSC            (unsigned long)0x08000000
#define DVSUP           (unsigned long)0x30000000
#define DFSUP           (unsigned long)0xC0000000

#define DVFS_DVSUP_OFFSET 28

/*!
 * Read the DVSUP bits from the CCM
 */
#define GET_DVSUP()		(_reg_CCM_PMCR0 & DVSUP)>>DVFS_DVSUP_OFFSET


#else /* ARGON+ */
#define    _reg_CCM_PDR0   \
(*((volatile unsigned long *)(IO_ADDRESS(CRM_MCU_BASE_ADDR + 0x04))))
#define    _reg_CCM_MPCTL  \
(*((volatile unsigned long *)(IO_ADDRESS(CRM_MCU_BASE_ADDR + 0x10))))
#define    _reg_CCM_COSR   \
(*((volatile unsigned long *)(IO_ADDRESS(CRM_MCU_BASE_ADDR + 0x18))))
#define    _reg_CCM_DCVR0  \
(*((volatile unsigned long *)(IO_ADDRESS(CRM_MCU_BASE_ADDR + 0x28))))
#define    _reg_CCM_DCVR1  \
(*((volatile unsigned long *)(IO_ADDRESS(CRM_MCU_BASE_ADDR + 0x2C))))
#define    _reg_CCM_DCVR2  \
(*((volatile unsigned long *)(IO_ADDRESS(CRM_MCU_BASE_ADDR + 0x30))))
#define    _reg_CCM_DCVR3  \
(*((volatile unsigned long *)(IO_ADDRESS(CRM_MCU_BASE_ADDR + 0x34))))
#define    _reg_CCM_PMCR0   \
(*((volatile unsigned long *)(IO_ADDRESS(CRM_MCU_BASE_ADDR + 0x38))))

#define DPTEN		(unsigned long)0x00000001
#define PTVAI		(unsigned long)0x00000006
#define PTVAIM		(unsigned long)0x00000008
#define DRCE0           (unsigned long)0x00080000
#define DRCE1           (unsigned long)0x00100000
#define DRCE2           (unsigned long)0x00200000
#define DRCE3           (unsigned long)0x00400000
#define DCR             (unsigned long)0x00020000

#endif /* CONFIG_ARCH_MX3 */

/*
 * if not defined define TREU and FALSE values.
 */
#ifndef TRUE
        #define TRUE  1
        #define FALSE 0
#endif /* TRUE */


#endif
