/*
 * Copyright 2005 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA
 *
 */

/*!
 * @file rr_core.c
 * @brief This is the main file for the Roadrunner driver.
 *
 * It contains devfs file system and initialization of the driver.
 *
 * @ingroup PMIC_RR_CORE
 */

/*
 * Includes
 */

#include "rr_config.h"
#include "rr_register.h"
#include "rr_event.h"
#include "asm/arch/pmic_external.h"
#include "asm/arch/gpio.h"

/*
 * For test only
 */
#define RR_DEBUG

#ifdef RR_DEBUG
#define STATIC
#else
#define STATIC    static
#endif

/*
 * End of For test only
 */

static int rr_major;

extern void gpio_rr_active(void);

/*!
 * This structure contains initialized value of registers.
 */
static unsigned int tab_init_reg[][2] = {
        {-1, -1},
};

#define IT_RR_PORT   0
#define IT_RR_BIT    5


/*
 * Global variables
 */

DECLARE_WORK(rr_wq_task, (void*)rr_wq_handler, (unsigned long)0);

/*!
 * This function is called when Roadrunner interrupt occurs on the processor.
 * It is the interrupt handler for the Roadrunner module.
 *
 * @param        irq        the irq number
 * @param        dev_id     the pointer on the device
 * @param        regs       the interrupt parameters
 *
 * @return       The function returns IRQ_RETVAL(1) when handled.
 */

STATIC irqreturn_t rr_irq_handler(int irq, void *dev_id,
        struct pt_regs *regs)
{
        edio_clear_int(ED_INT1); 
        edio_get_int(ED_INT1);

        /* prepare a task */
        TRACEMSG(_K_D("* roadrunner irq handler top half *"));
        schedule_work(&rr_wq_task);
        return IRQ_RETVAL(1);
}

/*!
 * This function implements IOCTL controls on a ATLAS device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @param        cmd         the command
 * @param        arg         the parameter
 * @return       This function returns 0 if successful.
 */
STATIC int rr_ioctl(struct inode *inode, struct file *file,
        unsigned int cmd, unsigned long arg)
{
#ifdef __TEST_CODE_ENABLE__
        register_info reg_info;
        type_event_notification event_sub;
#endif /* __TEST_CODE_ENABLE__ */

        if (_IOC_TYPE(cmd) != 'P')
                return -ENOTTY;

        if (copy_from_user(&reg_info, (register_info*) arg,
                sizeof(register_info))) {
                return -EFAULT;
        }

        switch (cmd) {
#ifdef __TEST_CODE_ENABLE__
        case PMIC_READ_REG :
                TRACEMSG(_K_D("read reg"));
                CHECK_ERROR(rr_read_reg(0,
                        reg_info.reg, &(reg_info.reg_value)));
                break;
        case PMIC_WRITE_REG :
                TRACEMSG(_K_D("write reg %d %x %x"), reg_info.reg,
                        reg_info.reg_value, reg_info.reg_mask);
                CHECK_ERROR(rr_write_reg(0,
                        reg_info.reg, reg_info.reg_value, reg_info.reg_mask));
                break;
        case PMIC_SUBSCRIBE :
                TRACEMSG(_K_I(" *** event sub %d"), event_sub.event);
                CHECK_ERROR(rr_event_subscribe_internal(event_sub));
                TRACEMSG(_K_I("subscribe done"));
                break;
        case PMIC_UNSUBSCRIBE :
                TRACEMSG(_K_I(" *** event unsub %d"),event_sub.event);
                CHECK_ERROR(rr_event_unsubscribe_internal(event_sub));
                break;
#endif /* __TEST_CODE_ENABLE__ */
        default :
                TRACEMSG(_K_D("%d unsupported ioctl command"),(int) cmd);
                return -EINVAL;
        }

        if (copy_to_user((register_info*) arg, &reg_info,
                sizeof(register_info))) {
                return -EFAULT;
        }

        return 0;
}

/*!
 * This function implements the open method on a ATLAS device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @return       This function returns 0.
 */
STATIC int rr_open(struct inode *inode, struct file *file)
{
        TRACEMSG(_K_D("roadrunner : rr_open()"));
        return 0;
}

/*!
 * This function implements the release method on a ATLAS device.
 *
 * @param        inode       pointer on the node
 * @param        file        pointer on the file
 * @return       This function returns 0.
 */

STATIC int rr_free(struct inode *inode, struct file *file)
{
        TRACEMSG(_K_D("roadrunner : rr_free()"));
        return 0;
}

/*!
 * This structure defines file operations for a ATLAS device.
 */
STATIC struct file_operations rr_fops = {
        /*!
        * the owner
        */
        .owner =    THIS_MODULE,
        /*!
        * the ioctl operation
        */
        .ioctl =    rr_ioctl,
        /*!
        * the open operation
        */
        .open =     rr_open,
        /*!
        * the release operation
        */
        .release =  rr_free,
};

/*
 * Initialization and Exit
 */

/*!
 * This function implements the init function of the ATLAS device.
 * This function is called when the module is loaded.
 * It init the interrupt handler and write initialization value of Roadrunner.
 *
 * @return       This function returns 0.
 */
STATIC int __init rr_init(void)
{
#if CONFIG_RR_PMIC_FIXARB == y
        extern void rr_fix_arbitration(void);
#endif

        int ret=0, i=0;

        rr_major = register_chrdev(0, "roadrunner", &rr_fops);
        if (rr_major < 0) {
                TRACEMSG(_K_D("unable to get a major for roadrunner"));
                return rr_major;
        }

        devfs_mk_cdev(MKDEV(rr_major,0), S_IFCHR | S_IRUGO | S_IWUSR,
                      "roadrunner");

#if CONFIG_RR_PMIC_FIXARB == y
        rr_fix_arbitration();
#endif

        rr_init_register_access();

        /* init all register */
        TRACEMSG(_K_D("init all register value"));
        for (i=0; tab_init_reg[i][0]!=-1; i++) {
                TRACEMSG(_K_D("reg %d, value %x"),tab_init_reg[i][0],
                        tab_init_reg[i][1]);
                CHECK_ERROR(rr_write_reg(0, tab_init_reg[i][0],
                        tab_init_reg[i][1], PMIC_ALL_BITS));
        }

        /* install our own roadrunner handler */
        rr_init_event();

        /* initialize GPIO for Interrupt of Roadrunner */
        gpio_rr_active();

        /* bind IT */
        ret = request_irq(INT_EXT_INT1, rr_irq_handler, 0, 0, 0);

        if (ret) {
                TRACEMSG(_K_I("gpio1: irq%d error."), INT_GPIO1);
                return ret;
        }

        TRACEMSG_ALL_TIME(KERN_INFO"Roadrunner Core loaded\n");

        return 0;
}

/*!
 * This function implements the exit function of the ATLAS device.
 * This function is called when the module is unloaded.
 *
 */
static void __exit rr_exit(void)
{
        devfs_remove("roadrunner");
        unregister_chrdev(rr_major, "roadrunner");
        gpio_free_irq(IT_RR_PORT, IT_RR_BIT, GPIO_HIGH_PRIO);

        TRACEMSG_ALL_TIME(_K_I("roadrunner device: successfully unloaded\n"));
}

/*
 * Module entry points
 */

module_init(rr_init);
module_exit(rr_exit);

MODULE_DESCRIPTION("Roadrunner char device driver");
MODULE_AUTHOR("FreeScale");
MODULE_LICENSE("GPL");
