/*
 * Copyright 2005 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA
 *
 */

/*!
 * @file rr_event.c
 * @brief This file manage all event of Roadrunner component.
 *
 * It contains subscribe and unsubscribe function and the bottom half of IT.
 *
 * @ingroup PMIC_RR_CORE
 */

/*
 * Includes
 */
#include <linux/spinlock.h>
#include "rr_config.h"
#include "rr_register.h"
#include "rr_event.h"
#include "asm/arch/gpio.h"

static spinlock_t event_list_lock = SPIN_LOCK_UNLOCKED;
static notify_event  *notify_event_tab[EVENT_NB];

/*!
 * This function is used to subscribe on an event.
 *
 * @param        event_sub   structure of event, it contains type of event and callback
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS rr_event_subscribe_internal(type_event_notification event_sub)
{
        notify_event *new_notify_event;
        notify_event *tail;

        TRACEMSG(_K_D("event subscribe\n"));
        spin_lock(&event_list_lock);

        /* Insert the new node to the end of the queue */
        new_notify_event =
                (notify_event*)kmalloc(sizeof(notify_event),GFP_KERNEL);
        new_notify_event->callback = event_sub.callback;
        new_notify_event->param = event_sub.param;
        new_notify_event->next_notify_event = NULL;

        if (notify_event_tab[event_sub.event] == NULL) {
                notify_event_tab[event_sub.event] = new_notify_event;
                CHECK_ERROR(unmask(event_sub.event));
        } else {
                tail = notify_event_tab[event_sub.event];
                while (tail->next_notify_event != NULL) {
                        tail = tail->next_notify_event;
                }
                tail->next_notify_event = new_notify_event;
        }

        spin_unlock(&event_list_lock);
        /* test if event is not available */

        rr_wq_handler(NULL);
        return PMIC_SUCCESS;
}

/*!
 * This function is used to unsubscribe on an event.
 *
 * @param        event_unsub   structure of event, it contains type of event and callback
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS rr_event_unsubscribe_internal(type_event_notification event_unsub)
{
        notify_event *pne1;
        notify_event *pne2;

        spin_lock(&event_list_lock);
        TRACEMSG(_K_D("event unsubscribe\n"));

        pne1 = notify_event_tab[event_unsub.event];
        pne2 = NULL;

        while (pne1 != NULL) {
                if (pne1->callback == event_unsub.callback) {
                        if (pne2 != NULL) {
                                pne2->next_notify_event =
                                        pne1->next_notify_event;
                        }
                        else {
                            notify_event_tab[event_unsub.event] = NULL;
                            CHECK_ERROR(mask(event_unsub.event));
                        }
                        kfree(pne1);
                        break;
                }
                pne2 = pne1;
                pne1 = pne1->next_notify_event;
        }

        if (pne1 == NULL) {
                TRACEMSG(_K_I("unsubsribe error (event not found)"));
                return PMIC_UNSUBSCRIBE_ERROR;
        }

        spin_unlock(&event_list_lock);
        return PMIC_SUCCESS;
}

/*!
 * This function is the bottom half of the Roadrunner Interrupt.
 * It checks the IT and launch client callback.
 *
 * WARNING: Be very careful about putting debugging statements in
 *          this function. Using printk() calls here may result in
 *          strange and usually bad behavior due to possible side
 *          effects while running in the context of an interrupt
 *          handler.
 */
void rr_wq_handler(void *code)
{
        unsigned int status=0,j=0;
        unsigned int mask_value=0;

        /* read and clear the status bit */
        rr_read_reg(PRIO_CORE, REG_ISR, &status);
        status &= PMIC_INTS_TO_AP;
        rr_write_reg(PRIO_CORE, REG_ISR, status, status);

        if (status != 0) {
                for (j=0;j<EVENT_NB;j++) {
                        mask_value = 1<<j;
                        if (mask_value & status) {
                                launch_all_call_back_event((type_event)j);
                        }
                }
        }
}

/*!
 * This function initializes events for Roadrunner interrupt.
 *
 */
void rr_init_event(void)
{
        unsigned int status=0;
        int i=0;
        
        for (i=0;i<EVENT_NB;i++) {
                notify_event_tab[i]= NULL;
        }
        
        /* Clear ISR */
        rr_read_reg(PRIO_CORE, REG_ISR, &status);
        status &= PMIC_INTS_TO_AP;
        rr_write_reg(PRIO_CORE, REG_ISR, status, status);

        spin_lock_init(&event_list_lock);
}

/*!
 * This function sets a bit in mask register of Roadrunner to mask an event IT.
 *
 * @param        event          structure of event, it contains type of event
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS unmask(type_event event)
{
        return rr_write_reg(PRIO_CORE, REG_IMR, 0, 1<<event);
}

/*!
 * This function clears a bit in mask register of Roadrunner to enable an event IT.
 *
 * @param        event   structure of event, it contains type of event
 *
 * @return       This function returns PMIC_STATUS if successful.
 */
PMIC_STATUS mask(type_event event)
{
        return rr_write_reg(PRIO_CORE, REG_IMR, 1<<event, 1<<event);
}

/*!
 * This function calls all callback of a specific event.
 *
 * @param        event   structure of event, it contains type of event
 */
void launch_all_call_back_event(type_event event)
{
        notify_event  *pne;

        spin_lock(&event_list_lock);

        pne = notify_event_tab[event];

        if (pne == NULL) {
                TRACEMSG(_K_I("it roadrunner event %d detected but not subscribed"),
                        event);
        }

        /* call each callback function */
        while (pne != NULL) {
                if (pne->callback != NULL) {
                        TRACEMSG(_K_D(" Calling  callback"));
                        pne->callback(pne->param);
                        TRACEMSG(_K_D("Callback done"));
                }
                pne = pne->next_notify_event;
        }
        spin_unlock(&event_list_lock);
}

