/*
 * Copyright (c) 2005 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  
 * 02111-1307, USA
 *
 */


/*!
 * @file rr_fixarb.c
 * @brief Flips the Roadrunner arbitration bits to allow register read/write
 *        access for the secondary processor.
 *        
 * Note that the initialiation function defined below must be called prior
 * to any of the other Roadrunner modules. We guarantee this by putting a call
 * to rr_fix_arbitration() in the initialization routine of rr_core.c.
 *
 * Also note that the rr_fix_arbitration() routine requires that CSPI2 be
 * connected to the Roadrunner's primary processor SPI bus. This currently
 * requires a hardware board modification to the RF Deck of the Sphinx board.
 * Specifically, the following 4 resistors must be removed in order to
 * disconnect the MQSPI:
 *
 *     R151, R154, R156, and R157
 *
 * Next, the following connections must be made to connect CSPI2 to the
 * Roadrunner's primary processor SPI bus:
 *
 *     P6077 to R151 radio board side pad
 *     etc.
 *
 * @ingroup PMIC_RR_CORE
 *
 */


#include <linux/kernel.h>
#include "rr_register.h"
#include "rr_config.h"
#include "../../spi/spi.h"
#include "asm/arch/pmic_status.h"


static PMIC_STATUS init_cspi2_access(void);
static PMIC_STATUS fix_irqs(void);
static PMIC_STATUS fix_connectivity(void);
static PMIC_STATUS fix_audio(void);
static PMIC_STATUS fix_reg_access(void);

static PMIC_STATUS rr_cspi2_read_reg(t_prio priority, int reg, unsigned int *reg_value);
static PMIC_STATUS rr_cspi2_write_reg(t_prio priority, int reg, unsigned int reg_value,
                         unsigned int reg_mask);
static PMIC_STATUS register_access_wait(t_prio priority);
static PMIC_STATUS unlock_client(void);
static PMIC_STATUS rr_spi_send_frame(int num_reg, unsigned int* reg_value, bool _rw);

static int pending_client_count[MAX_PRIO_QUEUE];
static wait_queue_head_t rr_reg_queue[MAX_PRIO_QUEUE];

static t_rr_status rr_status;
static spi_config the_config;
static void* spi_id;


/**************************************************************************
 * Module initialization and termination functions.
 *
 * Note that if this code is compiled into the kernel, then the
 * module_init() function will be called within the device_initcall()
 * group.
 **************************************************************************
 */

/*
 * Initialization
 */
void rr_fix_arbitration(void)
{
        TRACEMSG_ALL_TIME(KERN_INFO "Roadrunner changing arbitration bits "
                          "to allow secondary processor write access.\n");

        /* First, configure access using CSPI2. */
        init_cspi2_access();

        /* Next, make sure that interrupts are routed to the secondary
         * processor.
         */
        fix_irqs();

        /* Now change the arbitration bits on various Roadrunner registers
         * as required to allow write access from the secondary processor.
         */
        fix_connectivity();
        fix_audio();
        fix_reg_access();
        
        TRACEMSG_ALL_TIME(KERN_INFO "Roadrunner completed reconfiguration "
                          "of arbitration bits\n");

}

static PMIC_STATUS fix_irqs(void)
{
        PMIC_STATUS rc = PMIC_SUCCESS;
        unsigned int reg_value;

        /* Check to make sure that the INT_SEL register has the expected
         * reset value of 0xffffff.
         */
        rr_cspi2_read_reg(PRIO_CORE, REG_INT_SEL, &reg_value);
        if (reg_value != 0xffffff)
        {
            /* The INT_SEL (interrupt select) register does not have the
             * expected reset value, so let's manually set it.
             */
            TRACEMSG_ALL_TIME(KERN_INFO "Roadrunner changing INT_SEL from "
                              "0x%06x to 0xffffff\n", reg_value);

            rr_cspi2_write_reg(PRIO_CORE, REG_INT_SEL, 0xffffff, 0xffffff);

            /* Now check that the INT_SEL register was successfully updated. */
            rr_cspi2_read_reg(PRIO_CORE, REG_INT_SEL, &reg_value);
            if (reg_value != 0xffffff)
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Roadrunner failed to set INT_SEL "
                                  "to 0xffffff, current value is 0x%06x\n",
                                  reg_value);
                rc = PMIC_ERROR;
            }
        }

        return rc;
}

static PMIC_STATUS fix_connectivity(void)
{
        PMIC_STATUS rc = PMIC_SUCCESS;
        unsigned int reg_value;

        rr_cspi2_read_reg(PRIO_CONN, REG_BUSCTRL, &reg_value);
        if (!(reg_value & 0x800000))
        {
            /* Need to set BUS_PRI_ADJ bit to allow secondary processor
             * write access.
             */
            rr_cspi2_write_reg(PRIO_CONN, REG_BUSCTRL, 0x800000, 0x800000);

            /* Confirm that change was successful. */
            rr_cspi2_read_reg(PRIO_CONN, REG_BUSCTRL, &reg_value);
            if (!(reg_value & 0x800000))
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Failed to change BUSCTRL "
                                  "arbitration bits (0x%06x)\n", reg_value);
                rc = PMIC_ERROR;
            }
            else
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Changed BUS_PRI_ADJ bit in "
                                  "BUSCTRL (0x%06x)\n", reg_value);
            }
        }
        return rc;
}

static PMIC_STATUS fix_audio(void)
{
        PMIC_STATUS rc = PMIC_SUCCESS;
        unsigned int reg_value;

        rr_cspi2_read_reg(PRIO_AUDIO, REG_AUD_CODEC, &reg_value);
        if (!(reg_value & 0x800000))
        {
            /* Need to set CDC_PRI_ADJ bit to allow secondary processor
             * write access.
             */
            rr_cspi2_write_reg(PRIO_AUDIO, REG_AUD_CODEC, 0x800000, 0x800000);

            /* Confirm that change was successful. */
            rr_cspi2_read_reg(PRIO_AUDIO, REG_AUD_CODEC, &reg_value);
            if (!(reg_value & 0x800000))
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Failed to change AUD_CODEC "
                                  "arbitration bits (0x%06x)\n", reg_value);
                rc = PMIC_ERROR;
            }
            else
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Changed CDC_PRI_ADJ bit in "
                                  "AUD_CODEC (0x%06x)\n", reg_value);
            }
        }

        rr_cspi2_read_reg(PRIO_AUDIO, REG_RX_AUD_AMPS, &reg_value);
        if ((reg_value & 0xb20000) != 0xb20000)
        {
            /* Need to set RX_PRI_ADJ, MONO_PRI_ADJ, AUDOG_PRI_ADJ, and
             * A2_ADJ_LOCK bits to allow secondary processor write access.
             */
            rr_cspi2_write_reg(PRIO_AUDIO, REG_RX_AUD_AMPS, 0xb20000, 0xb20000);

            /* Confirm that change was successful. */
            rr_cspi2_read_reg(PRIO_AUDIO, REG_RX_AUD_AMPS, &reg_value);
            if ((reg_value & 0xb20000) != 0xb20000)
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Failed to change RX_AUD_AMPS "
                                  "arbitration bits (0x%06x)\n", reg_value);
                rc = PMIC_ERROR;
            }
            else
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Changed arbitration bits in "
                                  "RX_AUD_AMPS (0x%06x)\n", reg_value);
            }
        }

        rr_cspi2_read_reg(PRIO_AUDIO, REG_ST_DAC, &reg_value);
        if ((reg_value & 0xe00000) != 0xe00000)
        {
            /* Need to set ST_DAC_PRI_ADJ, and AUD_MIX_PRI_ADJ bits to allow
             * secondary processor write access.
             */
            rr_cspi2_write_reg(PRIO_AUDIO, REG_ST_DAC, 0xe00000, 0xe00000);

            /* Confirm that change was successful. */
            rr_cspi2_read_reg(PRIO_AUDIO, REG_ST_DAC, &reg_value);
            if ((reg_value & 0xe00000) != 0xe00000)
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Failed to change ST_DAC "
                                  "arbitration bits (0x%06x)\n", reg_value);
                rc = PMIC_ERROR;
            }
            else
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Changed arbitration bits in "
                                  "ST_DAC (0x%06x)\n", reg_value);
            }
        }

        rr_cspi2_read_reg(PRIO_AUDIO, REG_TX_AUD_AMPS, &reg_value);
        if ((reg_value & 0xa82000) != 0xa82000)
        {
            /* Need to set TX_PRI_ADJ, AUDIG_PRI_ADJ, AUDIO_BIAS_ARB, and
             * A1_Config bits to allow secondary processor write access.
             */
            rr_cspi2_write_reg(PRIO_AUDIO, REG_TX_AUD_AMPS, 0xa82000, 0xa82000);

            /* Confirm that change was successful. */
            rr_cspi2_read_reg(PRIO_AUDIO, REG_TX_AUD_AMPS, &reg_value);
            if ((reg_value & 0xa82000) != 0xa82000)
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Failed to change TX_AUD_AMPS "
                                  "arbitration bits (0x%06x)\n", reg_value);
                rc = PMIC_ERROR;
            }
            else
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Changed arbitration bits in "
                                  "TX_AUD_AMPS (0x%06x)\n", reg_value);
            }
        }

        return rc;
}

static PMIC_STATUS fix_reg_access(void)
{
        PMIC_STATUS rc = PMIC_SUCCESS;
        unsigned int reg_value;

        rr_cspi2_read_reg(PRIO_CORE, REG_ARB_REG, &reg_value);

        if ((reg_value & 0x7067EC) != 0x7067EC)
        {
            /* Need to set bits in ARB_REG bit to allow secondary processor
             * write access.
             */
            rr_cspi2_write_reg(PRIO_CORE, REG_ARB_REG, 0x7067EC, 0x7067EC);

            /* Confirm that change was successful. */
            rr_cspi2_read_reg(PRIO_CORE, REG_ARB_REG, &reg_value);
            if (!(reg_value & 0x7067EC))
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Failed to change ARB_REG (0x%06x)\n", reg_value);
                rc = PMIC_ERROR;
            }
            else
            {
                TRACEMSG_ALL_TIME(KERN_INFO "Changed ARB_REG to (0x%06x)\n", reg_value);
            }
        } else {
            TRACEMSG_ALL_TIME(KERN_INFO "ARB_REG is (0x%06x)\n", reg_value);
        }
        return rc;
}

static PMIC_STATUS init_cspi2_access(void)
{
        /* initialize all semaphore for all clients */
        int i=0;
        
        for (i=0; i<MAX_PRIO_QUEUE; i++) {
                pending_client_count[i]=0;
                init_waitqueue_head(&(rr_reg_queue[i]));
        }

        rr_status = RR_FREE;

        /* initialize SPI */
        the_config.module_number = SPI2;
        the_config.priority = HIGH;
        the_config.master_mode = true;
        the_config.bit_rate = 4000000;
        the_config.bit_count = 32;
        the_config.active_high_polarity = true;
        the_config.active_high_ss_polarity = false;
        the_config.phase = false;
        the_config.ss_low_between_bursts = true;
        the_config.ss_asserted = SS_0;
        the_config.tx_delay = 10;
        spi_id = spi_get_device_id(&the_config);

        return PMIC_SUCCESS;
}


/*!
 * This function reads a register.
 *
 * @param        priority   priority of access
 * @param        reg        register.
 * @param        reg_value   register value.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
static PMIC_STATUS rr_cspi2_read_reg(t_prio priority, int reg, unsigned int *reg_value)
{
        if (rr_status == RR_BUSY) {
                CHECK_ERROR(register_access_wait(priority));
        }
        
        TRACEMSG(_K_D("read reg = %d"),reg);
        CHECK_ERROR(rr_spi_send_frame(reg, reg_value, false));
        TRACEMSG(_K_D("read register done, read value = 0x%x"), *reg_value);
        CHECK_ERROR(unlock_client());

        return PMIC_SUCCESS;
}


/*!
 * This function writes a register.
 *
 * @param        priority   priority of access
 * @param        reg        register.
 * @param        reg_value  register value.
 * @param        reg_mask   bitmap mask indicating which bits to be modified.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
static PMIC_STATUS rr_cspi2_write_reg(t_prio priority, int reg, unsigned int reg_value,
                         unsigned int reg_mask)
{
        unsigned int temp;

        if (rr_status == RR_BUSY) {
                CHECK_ERROR(register_access_wait(priority));
        }

        TRACEMSG(_K_D("write reg = %d"),reg);
        if ((reg_mask & PMIC_ALL_BITS) ^ PMIC_ALL_BITS)
        {
            CHECK_ERROR(rr_spi_send_frame(reg, &temp, false));
            temp &= ~reg_mask;
            temp |= (reg_value & reg_mask);
        }
        else
        {
            temp = reg_value;
        }

        CHECK_ERROR(rr_spi_send_frame(reg, &temp, true));

        TRACEMSG(_K_D("write register done"));
        CHECK_ERROR(unlock_client());

        return PMIC_SUCCESS;
}


/*!
 * This function sleeps for R/W access if the spi is busy.
 *
 * @param        priority   priority of access
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
static PMIC_STATUS register_access_wait(t_prio priority) 
{
        DEFINE_WAIT(wait);
        
        if (pending_client_count[priority]<MAX_CLIENTS) {
                TRACEMSG(_K_D("sleep register access prio %d ...."),
                        priority);

                pending_client_count[priority]++;
                prepare_to_wait(&(rr_reg_queue[priority]), &wait,
                                TASK_INTERRUPTIBLE);
                schedule();
                rr_status = RR_BUSY;
                finish_wait(&(rr_reg_queue[priority]), &wait);

                TRACEMSG(_K_D("wake up register access prio %d ...."),
                        priority);
        } else {
                return PMIC_CLIENT_NBOVERFLOW;
        }

        return PMIC_SUCCESS;
}


/*!
 * This function unlocks on locked register access.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
static PMIC_STATUS unlock_client()
{
        int i;
        bool queue_date = false;

        for (i=0; i< MAX_PRIO_QUEUE; i++) {
                if (pending_client_count[i] > 0) {
                        queue_date = true;
                        TRACEMSG(_K_D("unlock register prio %d"),i);
                        pending_client_count[i]--;
                        wake_up_interruptible(&(rr_reg_queue[i]));
                        break;
                }
        }

        if (queue_date == false) {
                TRACEMSG(_K_D("no pending client release SPI"));
                rr_status = RR_FREE;
        }

        return PMIC_SUCCESS;
}


/*!
 * This function is used to send a frame on SPI bus.
 *
 * @param        num_reg    Atlas register number
 * @param        reg_value  Register value
 * @param        _rw       read or write operation
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
static PMIC_STATUS rr_spi_send_frame(int num_reg, unsigned int* reg_value,
                                     bool _rw)
{
        unsigned int send_val;
        unsigned int frame=0;
        unsigned int frame_ret=0;
        unsigned int result=0;

        if(_rw==true) {
                frame |= 0x80000000;
                frame |= (*reg_value & 0x0ffffff);
        } else {
                frame = 0;
        }
        frame |= ((unsigned int) num_reg & 0x3f) << 0x19;

        TRACEMSG(_K_D("\t\tspi send frame : value=0x%.8x"), frame&0xffffffff);


        send_val = ( ((frame&0x000000ff)<<0x18) |
                ((frame&0x0000ff00)<<0x08) |
                ((frame&0x00ff0000)>>0x08) |
                ((frame&0xff000000)>>0x18) );
        /* use this to launch SPI operation.*/
        result = spi_send_frame( (unsigned char*)(&send_val),
                (unsigned long) 4, spi_id);

        frame_ret = ( ((send_val&0x000000ff)<<0x18) |
                ((send_val&0x0000ff00)<<0x08) |
                ((send_val&0x00ff0000)>>0x08) |
                ((send_val&0xff000000)>>0x18) );
        *reg_value = frame_ret & 0x00ffffff;

        TRACEMSG(_K_D("\t\tspi received frame : value=0x%.8x")
                , frame_ret&0xffffffff);


        return PMIC_SUCCESS;
};
