/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*!
 * @file mxc_hacc.c
 *
 * @brief APIs for HAC Module.
 *
 * This file provides the APIs for accessing Hash Acceleration (HAC)
 * module. HAC module accelerates the creation of a SHA-1 hash over
 * selected memory spaces. The SHA-1 algorithm is a one-way hash
 * algorithm that creates 160 bit hash of any length input.
 *
 * @ingroup MXC_Security
 */

#include "mxc_hacc.h"
#ifdef CONFIG_HAC_TEST_DEBUG
#include <linux/module.h>
#endif /* HAC_TEST_DEBUG */

/*!
 * This variable indicates whether HAC module is in suspend or in resume
 * mode.
 */
static unsigned short hac_suspend_state = 0;

/*!
 * These variables store the HAC Module status register values when HAC is
 * in suspend mode.
 */
#ifdef CONFIG_PM 
static unsigned long store_status;
#endif

/*!
 * This API configures the start address and block length of the data that 
 * needs to be hashed. Start address indicates starting location from where 
 * the data in the flash memory is to be hashed. The number of blocks that 
 * needs to be hashed is loaded in the block count register. This API does
 * the starting of Hashing process or continue with Hashing of next block of 
 * data configured in the START_ADDR, BLOCK_COUNT register depending on the
 * hash parameter passed.
 *
 * @param   start_address   Starting address of the flash memory to be hashed.
 *                          user has to pass physical address here.    
 * @param   blk_len         Number of blocks to be hashed.
 * @param   option          Mode of operation like Start or Continue hashing.\n
 *                          Following parameters are passed:
 *                          HAC_START       : Starts the Hashing process.
 *                          HAC_LAST_START  : Starts the Hashing process with 
 *                                            last block of data.
 *                          HAC_CONTINUE    : Continue the Hashing process.
 *                          HAC_LAST        : Continue the Hashing process with
 *                                            last block of data.            
 *                                   
 *
 * @return  HAC_SUCCESS    Successfully hashed the data.\n
 *          HAC_FAILURE    Error in the parameters passed.
 *          HAC_BUSY       HAC module is busy in Hashing process.     
 */
hac_ret hac_hash_data (ulong start_address, ulong blk_len, hac_hash option)
{
        hac_ret ret_val = HAC_SUCCESS;
        if (hac_suspend_state == 1)      {
                HAC_DEBUG("HAC Module: HAC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        HAC_DEBUG("Function %s. HAC Module: Start address: 0x%08lX, "\
                   "block length: 0x%08lX, hash option: 0x%08X\n",\
                   __FUNCTION__, start_address, blk_len, option);
        /* Validating the parameters. Checking for start address to be in
           512 bit boundary(64 byte) and block count value must not to be 
           zero. */
        if ((!start_address) || (blk_len > HAC_MAX_BLOCK_LENGTH) || \
             (blk_len == 0) || (!((start_address % 64) == 0))) {
                HAC_DEBUG("HAC Module: Invalid parameters passed. \n");
                return HAC_FAILURE;
        }
        if ((HAC_CTL & HAC_CTL_BUSY) == 0) {
                HAC_START_ADDR = start_address;
                HAC_BLK_CNT = blk_len ;
                HAC_DEBUG("HAC Module: Hashing start address 0x%08lX\n ", \
                           start_address);
                HAC_DEBUG("HAC Module: Hashing blk length 0x%08lX\n ", \
                           blk_len);       
        } else {
                HAC_DEBUG("HAC Module: HAC module is busy in Hashing "\
                           "process.\n");
                return HAC_HASH_BUSY;
        }                
                
        switch ( option ) {
        case HAC_START:
                /* 
                 * HAC_START will starts the Hashing of data in the memory. 
                 * Before starting the Hashing process, it checks for 'STOP'
                 * bit, 'DONE' bit and 'ERROR' bit is set in the HAC Control 
                 * register. If 'STOP' bit is set, it clears the 'STOP' bit in
                 * HAC Control register. If 'DONE' bit and  'ERROR' bit are 
                 * set, they are cleared.
                 */ 
                HAC_DEBUG("HAC Module: Starts the hashing process \n");
                /* Checking if the Stop bit is been set. */
                if ((HAC_CTL & HAC_CTL_STOP) == HAC_CTL_STOP) {
                        HAC_DEBUG("HAC Module: STOP bit is set while" \
                                   "starting the Hashing\n");  
                        HAC_CTL &= ~HAC_CTL_STOP;   
                }
                /* Checking if the 'DONE' bit and 'ERROR' bit is been set. 
                 If they are set write to clear those bits */
                if (((HAC_CTL & HAC_CTL_DONE) == HAC_CTL_DONE) || 
                     ((HAC_CTL & HAC_CTL_ERROR) == HAC_CTL_ERROR)) {
                        HAC_DEBUG("HAC Module: DONE and ERROR bit is set" \
                                   "while starting the Hashing\n");     
                        HAC_CTL |= HAC_CTL_DONE;
                        HAC_CTL |= HAC_CTL_ERROR;
                }     
                HAC_CTL |= HAC_CTL_START; 
                break;

        case HAC_START_LAST:
                /*
                 * HAC_START_LAST will starts the Hashing of last block of data
                 * in the memory. Since this is last block of data in the 
                 * memory 'PAD' bit in HAC control register is set to add
                 * appropriate padding to the end of the data structure.
                 * Before starting the Hashing process, it checks for 'STOP'
                 * bit, 'DONE' bit and 'ERROR' bit is set in the HAC Control
                 * register. If 'STOP' bit is set, it clears the 'STOP' bit in
                 * HAC Control register. If 'DONE' bit and  'ERROR' bit are
                 * set, they are cleared. 
                 */
                HAC_DEBUG("HAC Module: Starts with last block" \
                           "the hashing process \n");
                /* Checking if the Stop bit is been set. */
                if ((HAC_CTL & HAC_CTL_STOP) == HAC_CTL_STOP) {
                        HAC_DEBUG("HAC Module: STOP bit is set while" \
                                   "starting the Hashing\n");
                        HAC_CTL &= ~HAC_CTL_STOP;
                }
                /* Checking if the 'DONE' bit and 'ERROR' bit is been set.
                 If they are set write to clear those bits */
                if (((HAC_CTL & HAC_CTL_DONE) == HAC_CTL_DONE) ||
                     ((HAC_CTL & HAC_CTL_ERROR) == HAC_CTL_ERROR)) {
                        HAC_DEBUG(" HAC Module: DONE and ERROR bit is set" \
                                   "while  starting the Hashing\n");
                        HAC_CTL |= HAC_CTL_DONE;
                        HAC_CTL |= HAC_CTL_ERROR;
                }
                HAC_CTL |= HAC_CTL_START;
                /* Hash for the last block by padding it. */
                HAC_DEBUG("HAC Module: Setting the PAD bit while start" \
                           "Hashing the last block\n");      
                HAC_CTL |= HAC_CTL_PAD;
                break;
                
        case HAC_CONTINUE:
                /*
                 * HAC_CONTINUE will continue the Hashing of data in the memory.
                 * This will continue the hashing processing by taking into 
                 * consideration of the previous hash result and continues
                 * further hashing of the new data block. Before continue the
                 * Hashing process, it checks for 'STOP' bit, 'DONE' bit and 
                 * 'ERROR' bit is set in the HAC Control register. If 'STOP'
                 * bit is set, it clears the 'STOP' bit in Control register.
                 * If 'DONE' bit is set, it clears the 'DONE' bit in control 
                 * register. If 'ERROR' bit is set, then error message is 
                 * indicated to the user.
                 */
                HAC_DEBUG("HAC Module: Continue hashing process. \n");
                /* Checking if the Stop bit is been set. */
                if ((HAC_CTL & HAC_CTL_STOP) == HAC_CTL_STOP) {
                        HAC_CTL &= ~HAC_CTL_STOP;
                }
                /* Checking if the 'DONE' bit is been set. If it is set write 
                 one to clear the bit */
                if ((HAC_CTL & HAC_CTL_DONE) == HAC_CTL_DONE) {
                        HAC_CTL |= HAC_CTL_DONE; 
                }
                /* Checking if 'ERROR' bit is been set. If it is set resturn
                  return back indicating error in Hashing porcess. */
                if ((HAC_CTL & HAC_CTL_ERROR) == HAC_CTL_ERROR) {
                        return HAC_FAILURE;
                }
                HAC_CTL |= HAC_CTL_CONTINUE;
                break;

        case HAC_LAST:
                /*
                 * HAC_LAST will continue the Hashing of last block of data 
                 * in the memory. Since this is last block of data in the
                 * memory 'PAD' bit in HAC control register is set to add
                 * appropriate padding to the end of the data structure.
                 * This will continue the hashing processing by taking into
                 * consideration of the previous hash result and continues
                 * further hashing of the new data block. Before continue the
                 * Hashing process, it checks for 'STOP' bit, 'DONE' bit and
                 * 'ERROR' bit is set in the HAC Control register. If 'STOP'
                 * bit is set, it clears the 'STOP' bit in Control register.
                 * If 'DONE' bit is set, it clears the 'DONE' bit in control
                 * register. If 'ERROR' bit is set, then error message is
                 * indicated to the user.
                 */
                HAC_DEBUG("HAC Module: Last block to hash. \n");
                /* Checking if the Stop bit is been set. */
                if ((HAC_CTL & HAC_CTL_STOP) == HAC_CTL_STOP) {
                        HAC_CTL &= ~HAC_CTL_STOP;
                }
                /* Checking if the 'DONE' bit is been set. If it is set write
                 one to clear the bit */
                if ((HAC_CTL & HAC_CTL_DONE) == HAC_CTL_DONE) {
                        HAC_CTL |= HAC_CTL_DONE;
                }
                /* Checking if 'ERROR' bit is been set. If it is set resturn
                  return back indicating error in Hashing porcess. */
                if ((HAC_CTL & HAC_CTL_ERROR) == HAC_CTL_ERROR) {
                        return HAC_FAILURE;
                }
                HAC_CTL |= HAC_CTL_CONTINUE;
                /* Continuing the hash for the last block by padding it. */
                HAC_CTL |= HAC_CTL_PAD;
                break;                                                 

        default:
                ret_val = HAC_FAILURE;
                /* NOT RESPONDING */ 
                break;
        }
        }
        return ret_val;
}

/*!
 * This API returns the status of the Hashing.
 *
 * @return      HAC_BUSY : Indicated HAC Module is busy with Hashing.\n
 *              HAC_DONE : Indicates Hashing of data is done.\n
 *              HAC_ERR  : Indicates error has occurred during Hashing.\n
 *              HAC_UNKNOWN: Hashing status unknown. This may be when the
 *                           hashing process has not been initiated atleast 
 *                           once or 'ERROR' bit or 'DONE' bits were reset 
 *                           after the hashing process was completed.
 */
hac_hash_status hac_hashing_status(void)
{
        if ( (HAC_CTL & HAC_CTL_BUSY) != 0 ) {
                HAC_DEBUG("HAC Module: Hash module is in busy state \n");
                return HAC_BUSY;
        } else if ((HAC_CTL & HAC_CTL_DONE) != 0) {
                /* Clearing the done bit of the control register */
                HAC_DEBUG("HAC Module: Hashing of data is done \n");
                HAC_CTL |= HAC_CTL_DONE;
                return HAC_DONE;
        } else if ((HAC_CTL & HAC_CTL_ERROR) != 0) {
                /* Clearing the error bit of the control register */
                HAC_DEBUG("HAC Module: Error has occurred during hashing \n");
                HAC_CTL |= HAC_CTL_ERROR;
                return HAC_ERR;
        } else {
                return HAC_UNKNOWN;
        }
}

/*!
 * This API returns the status of the Hash module.
 *
 * @return      Value of the Hashing control register.
 */
ulong hac_get_status(void)
{
        HAC_DEBUG("HAC Module: Hashing status register value 0x%08lX\n ",
                   HAC_CTL);
        return HAC_CTL;
}

/*!
 * This API stops the Hashing of data when the Hashing is in progress.
 */
hac_ret hac_stop(void)
{
        if (hac_suspend_state == 1)      {
                HAC_DEBUG("HAC Module: HAC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        HAC_DEBUG("HAC Module: Stop hashing process. \n");
        HAC_CTL |= HAC_CTL_STOP;
        }
        return HAC_SUCCESS;
}

/*!
 * This API reads 160 bit hash result from Hash result register. The data is
 * copied to the memory pointed by the input pointer.
 *
 * @param    hash_result_reg    structure Pointer where the hash result is
 *                              copied.
 */
hac_ret hac_hash_result(hac_hash_rlt *hash_result_reg)
{
        if (hac_suspend_state == 1)      {
                HAC_DEBUG("HAC Module: HAC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        HAC_DEBUG("HAC Module: Read hash result \n");
        hash_result_reg->hash_result[0]   = HAC_HSH4;
        hash_result_reg->hash_result[1]   = HAC_HSH3;
        hash_result_reg->hash_result[2]   = HAC_HSH2;
        hash_result_reg->hash_result[3]   = HAC_HSH1;
        hash_result_reg->hash_result[4]   = HAC_HSH0;
        }
        return HAC_SUCCESS;
}

#ifdef CONFIG_PM
/*!
 * This function is called to put the HAC in a low power state. Refer to the
 * document driver-model/driver.txt in the kernel source tree for more
 * information.
 *
 * @param   dev   the device structure used to give information on HAC
 *                to suspend.
 * @param   state the power state the device is entering.
 * @param   level the stage in device suspension process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
hac_ret hac_suspend(struct device *dev, ulong state, ulong level)
{
        unsigned short reg_val;
        hac_ret ret_val = HAC_SUCCESS;
        HAC_DEBUG("HAC Module: HAC suspend function.%08lX\n", level);
        switch(level) {
        case SUSPEND_DISABLE:
                HAC_DEBUG("HAC Module: In suspend Disable\n");
                /* Setting suspend flag. */
                hac_suspend_state = 1;
                break;

        case SUSPEND_SAVE_STATE:
                HAC_DEBUG("HAC Module: In Suspend save state.\n");
                /* Store HAC Status register value. */
                reg_val = HAC_CTL;
                store_status = reg_val;
                break;

        case SUSPEND_POWER_DOWN:
                HAC_DEBUG("HAC Module: In suspend power down.\n");
                /* Setting suspend flag. */
                hac_suspend_state = 1;
                /* Disable start, stop, continue, pad bits in HAC Control
                   Register. */
                HAC_CTL |= HAC_CTL_STOP;
                break;

        default:
                ret_val = HAC_FAILURE;
                break;
        }
        return ret_val;
}

/*!
 * This function is called to bring the HAC back from a low power state.
 * Refer to the document driver-model/driver.txt in the kernel source tree
 * for more information.
 *
 * @param   dev   the device structure used to give information on HAC
 *                to resume.
 * @param   level the stage in device resumption process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
hac_ret hac_resume(struct device *dev, ulong level)
{
        hac_ret ret_val = HAC_SUCCESS;
        HAC_DEBUG("HAC Module: HAC resume function.%08lX\n", level);
        switch(level) {
        case RESUME_POWER_ON:
                HAC_DEBUG("HAC Module: Resume power on.\n");
                /* Clearing the suspend flag. */
                hac_suspend_state = 0;
                break;

        case RESUME_RESTORE_STATE:
                HAC_DEBUG("HAC Module: Resume restore state.\n");
                HAC_CTL = store_status;
                break;

        case RESUME_ENABLE:
                HAC_DEBUG("HAC Module: Resume Enable state called.\n");
                /* Clearing the suspend flag. */
                hac_suspend_state = 0;
                /* Enable stop bit in HAC Control register. */
                HAC_CTL &= ~HAC_CTL_STOP;
                break;

        default:
                ret_val = HAC_FAILURE;
                break;
        }
        return ret_val;
}
#else
#define mxc_hac_suspend NULL
#define mxc_hac_resume  NULL
#endif /* CONFIG_PM */
