/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*!
 * @file mxc_rnga.c
 *
 * @brief This file allows the user to use the API's for configuring
 * seeding and generation of random numbers.
 *
 * @ingroup MXC_Security
 */

/*
 * Module header file
 */
#include "mxc_rnga.h"
#include <linux/module.h> 
#include <linux/device.h>

/*!
 * This variable indicates whether RNGA driver is in suspend or in resume
 * mode.
 */
static unsigned short rnga_suspend_state = 0;

/*!
 * These variables store the RNGA Module status register values when RNGA is
 * in suspend mode.
 */
#ifdef CONFIG_PM 
static unsigned long store_status;
#endif

/*!
 * This API configures the RNGA module for interrupt enable or disable. It
 * configures the module to be configured for High assurance boot (HAB).
 * Enabling HAB will inform the SCC in case of any Security violation.
 *
 * @param   rnga_cfg   configuration value for RNGA.
 *
 * @return  RNGA_SUCCESS  Successfully configured the RNGA module.\n
 *          RNGA_FAILURE  Invalid parameter passed.
 */
rnga_ret rnga_configure(rnga_config rnga_cfg)
{
        rnga_config config_chk;

        if (rnga_suspend_state == 1)      {
                RNGA_DEBUG("RNGA Module: RNGA Module is in suspend mode.\n");
                return -EPERM;
        } else {
        RNGA_DEBUG("In function %s. RNGA CFG=0x%02X\n", __FUNCTION__, rnga_cfg);
        config_chk = RNGA_HIGH_ASSUR | RNGA_INTR_MSK;
        if ((~(config_chk)) & rnga_cfg) {
                return RNGA_FAILURE;
        } 
        
        RNGA_CONTROL |= rnga_cfg;
        }
        return RNGA_SUCCESS;
}

/*!
 * This API returns the status of the RNGA module.
 *
 * @return   Status of the RNGA module.
 */
ulong rnga_get_status(void)
{
        RNGA_DEBUG("In function %s. Status=0x%08lX\n", \
                    __FUNCTION__, RNGA_STATUS);

        return RNGA_STATUS;
}

/*!
 * This API returns the total count value of the random stored in the RNGA FIFO
 *
 * @return  Total count value of random numbers stored in the RNGA FIFO.
 */
unchar rnga_get_fifo_level(void)
{
        unchar rnga_fifo_lvl;
        ulong temp_rnga_sts;

        temp_rnga_sts = rnga_get_status();
        RNGA_DEBUG("In function %s. Status=0x%08lX\n", \
                    __FUNCTION__, temp_rnga_sts);
        rnga_fifo_lvl = temp_rnga_sts  >> RNGA_MSBYTE_2_LSBYTE;
        RNGA_DEBUG("In function %s. Fifo level=0x%02X\n", \
                    __FUNCTION__, rnga_fifo_lvl);

        return rnga_fifo_lvl;
}

/*!
 * Seeds the Random number generator with the new seed value.
 *
 * @param  entropy_val seed for the random number generator.
 */
rnga_ret rnga_seed_rand(ulong entropy_val)
{
        if (rnga_suspend_state == 1)      {
                RNGA_DEBUG("RNGA Module: RNGA Module is in suspend mode.\n");
                return -EPERM;
        } else {
        RNGA_DEBUG("In function %s. RNGA entropy=0x%08lX\n", \
                    __FUNCTION__, entropy_val);
        RNGA_ENTROPY = entropy_val;
        }
        return RNGA_SUCCESS;
}

/*!
 * This API gets the random number from the FIFO. If RNGA is in power down mode
 * (Sleep mode), no random numbers are generated. In such cases, this API checks
 * if any random numbers are available in the FIFO. It will return any available
 * random numbers from the FIFO or return zero indicating that the module is not
 * generating any random numbers.
 *
 * @return  Random number or zero (0x00) if there are no random number available
 *          in the FIFO.
 */
ulong rnga_get_rand(void)
{
        ulong rand_num;

        if (rnga_suspend_state == 1)      {
                RNGA_DEBUG("RNGA Module: RNGA Module is in suspend mode.\n");
                return -EPERM;
        } else {

        RNGA_DEBUG("In function %s. RNGA Control = 0x%08lX\n", \
                    __FUNCTION__, RNGA_CONTROL);

        if ((!(RNGA_STATUS & RNGA_SLEEP)) || (rnga_get_fifo_level() > 0x00)) {
                if (!(RNGA_STATUS & RNGA_SLEEP)) {
                        /* If not in sleep mode, random numbers are 
                         * generated. */
                        RNGA_DEBUG("Value of control reg sec 0x%08lX\n", 
                                    RNGA_CONTROL);
                        RNGA_DEBUG("Value of control reg trd 0x%08lX\n", 
                                    RNGA_CONTROL);
                        RNGA_DEBUG("Value of status reg trd 0x%08lX\n",
                                    RNGA_STATUS);
                        if ((RNGA_CONTROL & 0x01) != RNGA_START_RAND) {
                                RNGA_CONTROL |= RNGA_GEN_RAND;
                                RNGA_DEBUG("In function %s. RNGA Control "\
                                "after setting GO bit = 0x%08lX\n",\
                                __FUNCTION__, RNGA_CONTROL);
                        }
                }

                /* Checking if there are any bytes in the FIFO */
                while (rnga_get_fifo_level() == 0x00) {
                        /* Do Nothing */
                }
                rand_num = RNGA_FIFO;
                RNGA_DEBUG("In function %s. RNGA Fifo = 0x%08lX\n", \
                            __FUNCTION__, rand_num);
        } else {
                RNGA_DEBUG("In function %s. Currently in sleep mode. RNGA "\
                            "Control = 0x%08lX\n", __FUNCTION__, RNGA_CONTROL);
                /* RNGA in Sleep mode so returning zero. */
                rand_num = 0x00;
        }
        }
        return rand_num;
}

/*!
 * This API clears the RNGA interrupt
 */
rnga_ret rnga_clear_int(void)
{
        if (rnga_suspend_state == 1)      {
                RNGA_DEBUG("RNGA Module: RNGA Module is in suspend mode.\n");
                return -EPERM;
        } else {
        RNGA_CONTROL |= RNGA_CLR_INTR;
        RNGA_DEBUG("In function %s. RNGA Control=0x%08lX\n", \
                    __FUNCTION__, RNGA_CONTROL);
        }
        return RNGA_SUCCESS;
}

/*!
 * This API returns the FIFO size of the RNGA
 *
 * @return  FIFO size of the RNGA module.
 */
unchar rnga_get_fifo_size(void)
{
        unchar rnga_fifo_size;
        ulong temp_rnga_sts;
        if (rnga_suspend_state == 1)      {
                RNGA_DEBUG("RNGA Module: RNGA Module is in suspend mode.\n");
                return -EPERM;
        } else {

        RNGA_DEBUG("In function %s. Status=0x%08lX\n", \
                    __FUNCTION__, RNGA_STATUS);
        temp_rnga_sts = rnga_get_status();
        rnga_fifo_size = (temp_rnga_sts & 0xFF0000) >> 16;
        RNGA_DEBUG("In function %s. Fifo size=0x%02X\n", \
                    __FUNCTION__, rnga_fifo_size);
        }
        return rnga_fifo_size;
}

/*!
 * This API is used to set the mode of operation of RNGA module to either
 * Verification mode or oscillator frequency test mode.
 *
 * @param        rngamode  mode in which RNGA needs to be operated.
 *
 * @return   RNGA_SUCCESS  Successfully configured RNGA mode.
 *           RNGA_FAILURE  Invalid parameter passed.
 */
rnga_ret rnga_configure_mode(rnga_mode rngamode)
{
        rnga_mode mode_chk;
        if (rnga_suspend_state == 1)      {
                RNGA_DEBUG("RNGA Module: RNGA Module is in suspend mode.\n");
                return -EPERM;
        } else {
        RNGA_DEBUG("In function %s. rnga_mode=0x%02X\n", \
                    __FUNCTION__, rngamode);
        mode_chk = RNGA_VRFY_MODE | RNGA_OSC_FREQ_TEST;
        if ((~(mode_chk)) & rngamode) {
                return RNGA_FAILURE;
        }
         
        RNGA_MODE |= rngamode;
        RNGA_DEBUG("In function %s. Mode=0x%08lX\n", __FUNCTION__, \
                    RNGA_MODE);
        }
        return RNGA_SUCCESS;
}

/*!
 * This API brings the RNGA module to normal state from power down mode
 * (Sleep mode). This is used in scenario where the RNGA module is been put to
 * sleep mode and needs to be brought back to normal mode.
 *
 */
rnga_ret rnga_normal(void)
{
        if (rnga_suspend_state == 1)      {
                RNGA_DEBUG("RNGA Module: RNGA Module is in suspend mode.\n");
                return -EPERM;
        } else {        
        if (RNGA_STATUS & RNGA_SLEEP) {
                RNGA_CONTROL &= (~(RNGA_SLEEP));
        }
        RNGA_DEBUG("In function %s. RNGA Status = 0x%08lX\n", \
                    __FUNCTION__, RNGA_STATUS);
        }
        return RNGA_SUCCESS;
}

/*!
 * This API puts the RNGA module to sleep mode. After placing the RNGA module
 * to sleep mode (low power mode), it stops generating the random numbers.
 */
rnga_ret rnga_sleep(void)
{
        if (rnga_suspend_state == 1)      {
                RNGA_DEBUG("RNGA Module: RNGA Module is in suspend mode.\n");
                return -EPERM;
        } else {
        RNGA_CONTROL |= RNGA_SLEEP;
        RNGA_DEBUG("In function %s. RNGA Control=0x%08lX\n", \
                    __FUNCTION__, RNGA_CONTROL);
        RNGA_DEBUG("Value of control reg sleep 0x%08lX\n", RNGA_CONTROL);
        }
        return RNGA_SUCCESS;
}

/*!
 * This API is used to control verification clocks. It is used reset the shift
 * registers, use system clock and to turn on the shift clock.
 *
 * @param        rnga_vrfy_control configures the RNGA verification controller.
 *
 * @return   RNGA_SUCCESS  Successfully configured the RNGA verification
 *                         controller.
 *           RNGA_FAILURE  Parameter passed has invalid data.
 */
rnga_ret rnga_config_vrfy_ctrl(rnga_vrfy_ctrl rnga_vrfy_control)
{
        rnga_vrfy_ctrl vrfy_chk;

        RNGA_DEBUG("In function %s. rnga_vrfy_ctrl=0x%08X\n", \
                     __FUNCTION__, rnga_vrfy_control);
        vrfy_chk = RNGA_SHFT_CLK_ON | RNGA_SYS_CLK_ON | RNGA_RESET_SHFT_REG;
        /* Checking if any invalid parameter is been passed. */
        if ((~(vrfy_chk)) & rnga_vrfy_control) {
                return RNGA_FAILURE;
        }
         
        RNGA_VRFY_CTRL |= rnga_vrfy_control;
        RNGA_DEBUG("In function %s. RNGA Verify Control=0x%08lX\n", \
                    __FUNCTION__, RNGA_VRFY_CTRL);

        return RNGA_SUCCESS;
}

/*!
 * This API is used for the oscillator test logic to specify the length of
 * the count oscillator clock pulses.
 *
 * @param       osc_count_ptr   counter value of oscillator.
 * @param       rnga_rw         read/write data from/to oscillator counter.
 *
 * @return   RNGA_SUCCESS  Successfully preformed the operation.
 *           RNGA_FAILURE  Invalid parameter passed.
 */
rnga_ret rnga_config_osc_cntctrl(ulong *osc_count_ptr, rnga_readwrite rnga_rw)
{
        rnga_ret ret_val = RNGA_SUCCESS;
        
        RNGA_DEBUG("In function %s. osc_count_ptr=0x%08lX rnga_mode=%d\n", \
                   __FUNCTION__, (ulong)osc_count_ptr, rnga_rw);
        /* Checking for the validity of the parameteres if address is zero. */
        if ((!osc_count_ptr) || (*osc_count_ptr > RNGA_MAX_OSC_COUNT)) {
                ret_val = RNGA_FAILURE;
        } else {
                switch (rnga_rw) {
                case RNGA_WRITE:
                        RNGA_OSC_CTRL_CNT = *osc_count_ptr;
                        RNGA_DEBUG("In function %s. OSC_CTRL_CNT = 0x%08lX\n", \
                                   __FUNCTION__, RNGA_OSC_CTRL_CNT);
                        break;

                case RNGA_READ:
                        *osc_count_ptr = RNGA_OSC_CTRL_CNT;
                        RNGA_DEBUG("In function %s. *osc_count_ptr = " \
                                    "0x%08lX\n", __FUNCTION__, *osc_count_ptr);

                default:
                        ret_val = RNGA_FAILURE;
                        break;
                }
        }
        return ret_val;
}

/*!
 * This API is used to count the number of oscillator pulses received from
 * Oscillator #1/#2 starting from the time the Oscillator Counter Control
 * Register is written and ending at the time the Oscillator Counter Control
 * Register reaches zero. If the oscillator number is other than that specified,
 * it will return 0.
 *
 * @param       osc_num         oscillator number.
 *
 * @return      Count stored in the oscillator specified.
 */
ulong rnga_osc_cnt(rnga_osc_num osc_num)
{
        ulong osc_count;

        RNGA_DEBUG("In function %s. osc_num = 0x%02X\n",\
                    __FUNCTION__, osc_num);

        switch (osc_num) {
        case RNGA_OSCILLATOR1: 
                osc_count = RNGA_OSC1_CNT;
                break;
                
        case RNGA_OSCILLATOR2:
                osc_count = RNGA_OSC2_CNT;
                break;

        default:
                osc_count = 0;
                RNGA_DEBUG("In function %s. In default.\n", __FUNCTION__);
                break;
        }

        RNGA_DEBUG("In function %s. osc_count = 0x%08lX\n", \
                    __FUNCTION__, osc_count);

        return osc_count;
}

/*!
 * This API gives the oscillator counter which have toggled 0x400 times. If bit
 * 0 is set then oscillator #1 and if bit 1 is set the oscillator #2 has crossed
 * 0x400 times. If both the bits are set then both the oscillators have crossed
 * 0x400 times.
 *
 * @return      Oscillators that have toggled 0x400 times.
 */
rnga_osc_num rnga_osc_status(void)
{
        rnga_osc_num rnga_osc = RNGA_OSC_STATUS & 0x03;

        RNGA_DEBUG("In function %s. RNGA_OSC_STATUS = 0x%08X\n", \
                    __FUNCTION__, rnga_osc);

        return rnga_osc; 
}

#ifdef CONFIG_PM
/*!
 * This function is called to put the RNGA in a low power state. Refer to the
 * document driver-model/driver.txt in the kernel source tree for more
 * information.
 *
 * @param   dev   the device structure used to give information on RNGA
 *                to suspend.
 * @param   state the power state the device is entering.
 * @param   level the stage in device suspension process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
rnga_ret rnga_suspend(struct device *dev, ulong state, ulong level)
{
        unsigned short reg_val;
        rnga_ret ret_val = RNGA_SUCCESS;
        printk("RNGA Module: RNGA suspend function.%08lX\n", level);
        switch(level) {
        case SUSPEND_DISABLE:
                RNGA_DEBUG("RNGA Module: In suspend Disable\n");
                /* Setting suspend flag. */
                rnga_suspend_state = 1;
                break;

        case SUSPEND_SAVE_STATE:
                RNGA_DEBUG("RNGA Module: In Suspend save state.\n");
                /* Store RTIC Status register value. */
                reg_val = RNGA_STATUS;
                store_status = reg_val;
                break;

        case SUSPEND_POWER_DOWN:
                RNGA_DEBUG("RNGA Module: In suspend power down.\n");
                /* Setting suspend flag. */
                rnga_suspend_state = 1;
                break;

        default:
                ret_val = RNGA_FAILURE;
                break;
        }
        return ret_val;
}

/*!
 * This function is called to bring the RNGA back from a low power state.
 * Refer to the document driver-model/driver.txt in the kernel source tree
 * for more information.
 *
 * @param   dev   the device structure used to give information on RNGA
 *                to resume.
 * @param   level the stage in device resumption process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
rnga_ret rnga_resume(struct device *dev, ulong level)
{
        rnga_ret ret_val = RNGA_SUCCESS;
        printk("RNGA Module: RNGA resume function.%08lX\n", level);
        switch(level) {
        case RESUME_POWER_ON:
                RNGA_DEBUG("RNGA Module: Resume power on.\n");
                /* Clearing the suspend flag. */
                rnga_suspend_state = 0;
                break;
        case RESUME_RESTORE_STATE:
                RNGA_STATUS = store_status;
                break;

        case RESUME_ENABLE:
                RNGA_DEBUG("RNGA Module: Resume Enable state called.\n");
                /* Clearing the suspend flag. */
                rnga_suspend_state = 0;
                break;

        default:
                ret_val = RNGA_FAILURE;
                break;
        }
        return ret_val;
}
#else
#define mxc_rnga_suspend NULL
#define mxc_rnga_resume  NULL
#endif /* CONFIG_PM */                
