/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*!
 * @file mxc_rtic.c
 *
 * @brief APIs for RTIC Module.
 *
 * \b
 * This file provides the APIs for accessing Run-Time Integrity Checker (RTIC)
 * module. RTIC module accelerates the creation of a SHA-1 hash over
 * selected memory spaces. The RTIC has the ability to verify the memory 
 * contents during system boot and during run-time execution.
 *
 * @ingroup MXC_Security
 */

#include "mxc_rtic.h" 

#include <linux/device.h>
#ifdef CONFIG_RTIC_TEST_DEBUG
#include <linux/module.h>
#endif /* RTIC_TEST_DEBUG */

/*!
 * This variable indicates whether RTIC driver is in suspend or in resume
 * mode.
 */
static unsigned short rtic_suspend_state = 0;

/*!
 * These variables store the RTIC Module status register values when RTIC is
 * in suspend mode.
 */
#ifdef CONFIG_PM 
static unsigned long store_status;
#endif

/*!
 * This API configures the memory block (A, B, C, D) into
 * ONE_TIME Mode or RUN_TIME Mode for Hashing of data in memory.  
 * RTIC does not support enabling multiple memory blocks that aren't
 * grouped together(i.e. enabling only memory blocks A & C without memory
 * B enabled).
 *
 * @param  mode         RTIC mode of operation (ONE_TIME or RUN_TIME).
 * @param  mem_blk      Memory block to be hashed.       
 *
 * @return  RTIC_SUCCESS    Successfully hashed the data.\n
 *          RTIC_FAILURE    Error in the parameters passed.
 */
rtic_ret rtic_configure_mode(rtic_mode mode, rtic_memblk mem_blk)
{
        rtic_ret ret_val = RTIC_SUCCESS;
        /*
         * RTIC does not support enabling multiple memory blocks that aren't
         * grouped together(i.e. enabling only memory blocks A & C without
         * memory B enabled).
         */
        if (rtic_suspend_state == 1)      {
                RTIC_DEBUG("RTIC Module: RTIC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        switch (mode)  {
        case RTIC_ONE_TIME:
                switch (mem_blk) {
                case RTIC_A1:     
                        RTIC_DEBUG("RTIC Module:Memory Block A is enabled for"\
                                    "One-Time Hashing.\n");
                        RTIC_CONTROL |= RTIC_CTL_HASHONCE_MEMA_BLK_EN;
                        break;
                        
                case RTIC_B1: 
                        if ((RTIC_CONTROL & RTIC_CTL_HASHONCE_MEMA_BLK_EN) ==
                             0x10) {   
                                RTIC_DEBUG("RTIC Module:Memory Block B is " \
                                            "enabled for One_Time Hashing.\n");
                                RTIC_CONTROL |= RTIC_CTL_HASHONCE_MEMB_BLK_EN;
                        }
                        break;
                        
                case RTIC_C1: 
                        if (((RTIC_CONTROL & RTIC_CTL_HASHONCE_MEMA_BLK_EN) & 
                             (RTIC_CONTROL & RTIC_CTL_HASHONCE_MEMB_BLK_EN)) ==
                             0x20)  { 
                                RTIC_DEBUG("RTIC Module:Memory Block C is " \
                                            "enabled for One_Time Hashing.\n");
                                RTIC_CONTROL |= RTIC_CTL_HASHONCE_MEMC_BLK_EN;
                        }
                        break;
                        
                case RTIC_D1: 
                        if ((((RTIC_CONTROL & RTIC_CTL_HASHONCE_MEMA_BLK_EN) &
                             (RTIC_CONTROL & RTIC_CTL_HASHONCE_MEMB_BLK_EN)) &
                             (RTIC_CONTROL & RTIC_CTL_HASHONCE_MEMC_BLK_EN)) ==
                             0x40)  {     
                                RTIC_DEBUG("RTIC Module:Memory Block D is " \
                                            "enabled for One_Time Hashing.\n");
                                RTIC_CONTROL |= RTIC_CTL_HASHONCE_MEMD_BLK_EN;
                        }
                        break;
                        
                default:
                        ret_val = RTIC_FAILURE;
                        break;                       
                        
                }
                break;
                
        case RTIC_RUN_TIME:
                switch (mem_blk)  {
                case RTIC_A1:
                        RTIC_DEBUG("RTIC Module:Memory Block A is enabled for"\
                                    "Run_Time Hashing.\n");
                        RTIC_CONTROL |= RTIC_CTL_RUNTIME_MEMA_BLK_EN;
                        break;
                        
                case RTIC_B1: 
                        if ((RTIC_CONTROL & RTIC_CTL_RUNTIME_MEMA_BLK_EN) ==
                             0x01) {   
                                RTIC_DEBUG("RTIC Module:Memory Block B is "\
                                            "enabled for Run_Time Hashing.\n");
                                RTIC_CONTROL |= RTIC_CTL_RUNTIME_MEMB_BLK_EN;
                        }
                        break;
                        
                case RTIC_C1:
                        if (((RTIC_CONTROL & RTIC_CTL_RUNTIME_MEMA_BLK_EN) &
                            (RTIC_CONTROL & RTIC_CTL_RUNTIME_MEMB_BLK_EN)) ==
                            0x01) {
                                RTIC_DEBUG("RTIC Module:Memory Block C is "\
                                            "enabled for Run_Time Hashing.\n");
                                RTIC_CONTROL |= RTIC_CTL_RUNTIME_MEMC_BLK_EN;
                        }
                        break;
                        
                case RTIC_D1: 
                        if (((RTIC_CONTROL & RTIC_CTL_RUNTIME_MEMA_BLK_EN) &
                            (RTIC_CONTROL & RTIC_CTL_RUNTIME_MEMB_BLK_EN) &
                            (RTIC_CONTROL & RTIC_CTL_RUNTIME_MEMC_BLK_EN)) ==
                            0x01) { 
                                RTIC_DEBUG("RTIC Module:Memory Block D is "\
                                            "enabled for Run_Time Hashing.\n");
                                RTIC_CONTROL |= RTIC_CTL_RUNTIME_MEMD_BLK_EN;
                        }
                        break;
                        
                default:
                        ret_val = RTIC_FAILURE;
                        break;
                        
                }
                break;
                
        default:
                ret_val = RTIC_FAILURE;
                break;
        }
        }
        return ret_val;
}

/*!
 * This API allows to configure start address and block length of the 
 * memory content to be hashed. Start address indicates starting location from
 * where the data in the memory is to be hashed. The number of blocks that
 * needs to be hashed is loaded in the block count register. 
 * There are four memory blocks available. The user can configure any one of 
 * these four memory blocks by passing their appropriate address and block
 * length to be Hashed.
 *
 * @param    start_addr       Starting address of the memory to be hashed. 
 * @param    blk_len          Block length of data in memory.
 * @param    mem_blk          Memory block to be hashed.  
 *
 * @return  RTIC_SUCCESS    Successfully hashed the data.\n
 *          RTIC_FAILURE    Error in the parameters passed.          
 */

rtic_ret rtic_configure_mem_blk(ulong start_addr, ulong blk_len, \
                                 rtic_memblk mem_blk)
{
        rtic_ret ret_val = RTIC_SUCCESS;
        /* Validating the parameters passed. Checking for start address is 
         * word boundary aligned. Checking for block length is multiple of 
         * 4 bytes. */                
        if (rtic_suspend_state == 1)      {
                RTIC_DEBUG("RTIC Module: RTIC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        if ((!start_addr) || (blk_len > RTIC_MAX_BLOCK_LENGTH) ||
             (!((start_addr % 4) == 0)) || (!((blk_len % 4) == 0)) )  {
                return RTIC_FAILURE;
        } 
        switch (mem_blk) {
        case RTIC_A1:                
                RTIC_DEBUG("RTIC Module: Mem Block A1 start address " \
                            "0x%08lX\n", start_addr);
                RTIC_MEMAADDR1 =  start_addr;
                RTIC_DEBUG("RTIC Module: Mem Block A1 block len " \
                            "0x%08lX\n", blk_len);
                RTIC_MEMALEN1 =  blk_len;
                break;
                
        case RTIC_A2:
                RTIC_DEBUG("RTIC Module: Mem Block A2 start address" \
                            "0x%08lX\n", start_addr);
                RTIC_MEMAADDR2 =  start_addr;
                RTIC_DEBUG("RTIC Module: Mem Block A2 block len " \
                            "0x%08lX\n", blk_len);
                RTIC_MEMALEN2 =  blk_len;
                break;
                
        case RTIC_B1:
                RTIC_DEBUG("RTIC Module: Mem Block B1 start address " \
                            "0x%08lX\n", start_addr);
                RTIC_MEMBADDR1 =  start_addr;
                RTIC_DEBUG("RTIC Module: Mem Block B1 block len " \
                            "0x%08lX\n", blk_len);
                RTIC_MEMBLEN1 =  blk_len;
                break;
                
        case RTIC_B2:
                RTIC_DEBUG("RTIC Module: Mem Block B2 start address " \
                            "0x%08lX\n", start_addr);
                RTIC_MEMBADDR2 =  start_addr;
                RTIC_DEBUG("RTIC Module: Mem Block B2 block len " \
                            "0x%08lX\n", blk_len);
                RTIC_MEMBLEN2 =  blk_len;
                break;
                
        case RTIC_C1: 
                RTIC_DEBUG("RTIC Module: Mem Block C1 start address " \
                            "0x%08lX\n", start_addr);
                RTIC_MEMCADDR1 =  start_addr;
                RTIC_DEBUG("RTIC Module: Mem Block C1 block len " \
                            "0x%08lX\n", blk_len);
                RTIC_MEMCLEN1 =  blk_len;
                break;
                
        case RTIC_C2:
                RTIC_DEBUG("RTIC Module: Mem Block C2 start address " \
                            "0x%08lX\n", start_addr);
                RTIC_MEMCADDR2 =  start_addr;
                RTIC_DEBUG("RTIC Module: Mem Block C2 block len " \
                            "0x%08lX\n", blk_len);
                RTIC_MEMCLEN2 =  blk_len;
                break;
                
        case RTIC_D1:
                RTIC_DEBUG("RTIC Module: Mem Block D1 start address " \
                             "0x%08lX\n", start_addr);
                RTIC_MEMDADDR1 =  start_addr;
                RTIC_DEBUG("RTIC Module: Mem Block D1 block len " \
                            "0x%08lX\n", blk_len);
                RTIC_MEMDLEN1 =  blk_len;
                break;
                
        case RTIC_D2:
                RTIC_DEBUG("RTIC Module: Mem Block D2 start address "\
                            "0x%08lX\n", start_addr);
                RTIC_MEMDADDR2 =  start_addr;
                RTIC_DEBUG("RTIC Module: Mem Block D2 block len " \
                             "0x%08lX\n", blk_len);
                RTIC_MEMDLEN2 =  blk_len;
                break;
                
        default:
                ret_val = RTIC_FAILURE;
                break;      
        }
        }
        return ret_val;
}
                
/*!
 * This API will configure to start the Hashing of data in memory
 * either in One-Time Hash mode or Run-Time Hash mode.
 *
 * @param       mode    Memory block to be hashed.
 *
 * @return  RTIC_SUCCESS    Successfully hashed the data.\n
 *          RTIC_FAILURE    Error in the parameters passed. 
 */
rtic_ret rtic_start_hash(rtic_mode mode)
{
        rtic_ret ret_val = RTIC_SUCCESS;
        if (rtic_suspend_state == 1)      {
                RTIC_DEBUG("RTIC Module: RTIC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        switch (mode)  {
        case RTIC_ONE_TIME:
                RTIC_DEBUG("RTIC Module: Starts the One_time hashing"\
                            "process \n");       
                RTIC_COMMAND |= RTIC_CMD_HASH_ONCE;
                break;
                
        case RTIC_RUN_TIME:
                RTIC_DEBUG("RTIC Module: Starts the Run_time hashing"\
                            "process \n");
                RTIC_COMMAND |= RTIC_CMD_RUN_TIME_CHK;
                break;
                
        default:
                ret_val = RTIC_FAILURE;
                break;                        
        }
        }
        return ret_val;
}

/*!
 * This API will read the RTIC status register.
 * 
 * @return  Status of Hashing.       
 */
ulong rtic_get_status(void)
{
        if (rtic_suspend_state == 1)      {
                RTIC_DEBUG("RTIC Module: RTIC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        RTIC_DEBUG("RTIC  Module: Hashing status register value 0x%08lX\n ",
                    RTIC_STATUS);
        return RTIC_STATUS;
        }
}

/*!
 * This API will read the RTIC control register.
 *
 * @return  Control register value.
 */
ulong rtic_get_control(void)
{
        if (rtic_suspend_state == 1)      {
                RTIC_DEBUG("RTIC Module: RTIC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        RTIC_DEBUG("RTIC  Module: Hashing control register value 0x%08lX\n ",
                    RTIC_CONTROL);
        return RTIC_CONTROL;
        }
}

/*!
 * This API enables or disables interrupt for RTIC module.
 *
 * @param  irq_en    To enable or disable interrupt.   
 *
 * @return  RTIC_SUCCESS    Successfully hashed the data.\n
 *          RTIC_FAILURE    Error in the parameters passed. 
 */
rtic_ret rtic_configure_interrupt(ulong irq_en)
{       
        rtic_ret ret_val = RTIC_SUCCESS;
        if (rtic_suspend_state == 1)      {
                RTIC_DEBUG("RTIC Module: RTIC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        switch (irq_en) {
        case RTIC_INTERRUPT_ENABLE:
                RTIC_DEBUG("RTIC Module: RTIC Interrupt enabled.\n");   
                /* If in interrupt mode then, set the irq enable bit in 
                the RTIC control register */
                RTIC_CONTROL |= RTIC_CTL_IRQ_EN;
                break;
                
        case RTIC_INTERRUPT_DISABLE:
                RTIC_DEBUG("RTIC  Module: RTIC Interrupt Disabled.\n");
                /* If in polling mode, then disable the irq enable bit in 
                the RTIC Control register */ 
                RTIC_CONTROL &= ~RTIC_CTL_IRQ_EN;                 
                break;
                
        default:
                ret_val = RTIC_FAILURE;
                break;        
        }
        }
        return ret_val;
}

/*!
 * This API reads the Fault address of the RTIC module.
 *
 * @return   Fault address register value.   
 */
ulong  rtic_get_faultaddress(void)
{
        if (rtic_suspend_state == 1)      {
                RTIC_DEBUG("RTIC Module: RTIC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        RTIC_DEBUG("RTIC  Module: Hashing fault register value 0x%08lX\n ",
                    RTIC_FAULTADDR);
        return RTIC_FAULTADDR;
        }
}
 
/*!
 * This API reads 160 bit hash result from Hash result register.The data is
 * copied to the memory pointed by the input pointer.
 *
 * @param    mem_blk            Memory block to be hashed.
 * @param    hash_result_reg    Hashed value.      
 *
 * @return  RTIC_SUCCESS    Successfully hashed the data.\n
 *          RTIC_FAILURE    Error in the parameters passed. 
 */
rtic_ret rtic_hash_result(rtic_memblk mem_blk, rtic_hash_rlt *hash_result_reg)
{
        rtic_ret ret_val = RTIC_SUCCESS;
        if (rtic_suspend_state == 1)      {
                RTIC_DEBUG("RTIC Module: RTIC Module is in suspend mode.\n");
                return -EPERM;
        } else {
        switch (mem_blk) {
        case RTIC_A1:
                RTIC_DEBUG("RTIC Module: Read mem blk A hash result \n");
                hash_result_reg->hash_result[0] = RTIC_MEMAHASHRES4;
                hash_result_reg->hash_result[1] = RTIC_MEMAHASHRES3;
                hash_result_reg->hash_result[2] = RTIC_MEMAHASHRES2;
                hash_result_reg->hash_result[3] = RTIC_MEMAHASHRES1;
                hash_result_reg->hash_result[4] = RTIC_MEMAHASHRES0;
                break;
                
        case RTIC_B1:
                RTIC_DEBUG("RTIC Module: Read mem blk B hash result \n");
                hash_result_reg->hash_result[0] = RTIC_MEMBHASHRES4;
                hash_result_reg->hash_result[1] = RTIC_MEMBHASHRES3;
                hash_result_reg->hash_result[2] = RTIC_MEMBHASHRES2;
                hash_result_reg->hash_result[3] = RTIC_MEMBHASHRES1;
                hash_result_reg->hash_result[4] = RTIC_MEMBHASHRES0;
                break;
                
        case RTIC_C1:
                RTIC_DEBUG("RTIC Module: Read mem blk C hash result \n");
                hash_result_reg->hash_result[0] = RTIC_MEMCHASHRES4;
                hash_result_reg->hash_result[1] = RTIC_MEMCHASHRES3;
                hash_result_reg->hash_result[2] = RTIC_MEMCHASHRES2;
                hash_result_reg->hash_result[3] = RTIC_MEMCHASHRES1;
                hash_result_reg->hash_result[4] = RTIC_MEMCHASHRES0;
                break;
                
        case RTIC_D1:
                RTIC_DEBUG("RTIC Module: Read mem blk D hash result \n");
                hash_result_reg->hash_result[0] = RTIC_MEMDHASHRES4;
                hash_result_reg->hash_result[1] = RTIC_MEMDHASHRES3;
                hash_result_reg->hash_result[2] = RTIC_MEMDHASHRES2;
                hash_result_reg->hash_result[3] = RTIC_MEMDHASHRES1;
                hash_result_reg->hash_result[4] = RTIC_MEMDHASHRES0;
                break;
                
        default:
                ret_val = RTIC_FAILURE;
                break;                        
        }
        }
        return ret_val;
} 

#ifdef CONFIG_PM
/*!
 * This function is called to put the RTIC in a low power state. Refer to the
 * document driver-model/driver.txt in the kernel source tree for more
 * information.
 *
 * @param   dev   the device structure used to give information on RTIC
 *                to suspend.
 * @param   state the power state the device is entering.
 * @param   level the stage in device suspension process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
rtic_ret rtic_suspend(struct device *dev, ulong state, ulong level)
{
        unsigned short reg_val;
        rtic_ret ret_val = RTIC_SUCCESS;
        RTIC_DEBUG("RTIC Module: RTIC suspend function.%08lX\n", level);
        switch(level) {
        case SUSPEND_DISABLE:
                RTIC_DEBUG("RTIC Module: In suspend Disable\n");
                /* Setting suspend flag. */
                rtic_suspend_state = 1;
                break;

        case SUSPEND_SAVE_STATE:
                RTIC_DEBUG("RTIC Module: In Suspend save state.\n");
                /* Store RTIC Status register value. */
                reg_val = RTIC_STATUS;
                store_status = reg_val;
                break;

        case SUSPEND_POWER_DOWN:
                RTIC_DEBUG("RTIC Module: In suspend power down.\n");
                /* Setting suspend flag. */
                rtic_suspend_state = 1;
                break;
        
        default:
                ret_val = RTIC_FAILURE;
                break;
        }
        return ret_val;
}

/*!
 * This function is called to bring the RTIC back from a low power state.
 * Refer to the document driver-model/driver.txt in the kernel source tree
 * for more information.
 *
 * @param   dev   the device structure used to give information on RTIC
 *                to resume.
 * @param   level the stage in device resumption process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
rtic_ret rtic_resume(struct device *dev, ulong level)
{
        rtic_ret ret_val = RTIC_SUCCESS;      
        RTIC_DEBUG("RTIC Module: RTIC resume function.%08lX\n", level);
        switch(level) {
        case RESUME_POWER_ON:
                RTIC_DEBUG("RTIC Module: Resume power on.\n");
                /* Clearing the suspend flag. */
                rtic_suspend_state = 0;
                break;
                
        case RESUME_RESTORE_STATE:
                /*RTIC_DEBUG("RTIC Module: Resume restore state.\n");
                RTIC_STATUS = store_status;*/
                break;

        case RESUME_ENABLE:
                RTIC_DEBUG("RTIC Module: Resume Enable state called.\n");
                /* Clearing the suspend flag. */
                rtic_suspend_state = 0;
                break;

        default:
                ret_val = RTIC_FAILURE;
                break;
        }
        return ret_val;
}
#else
#define mxc_rtic_suspend NULL
#define mxc_rtic_resume  NULL
#endif /* CONFIG_PM */
