/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*!
 * @file mxc_scc.c
 *
 * @brief This file contains the low level functions that give access to the
 * hardware through the APIs.
 *
 * This file allows the user to use the API's for Security
 * Memory Controller(SCM), Security Monitor (SMN), Algorithm Integrity
 * Checker(AIC) and Security Timer modules. The above modules are
 * collectively called as SCC (Security Controller).
 *
 * @ingroup MXC_Security
 */

/*
 * Module header file
 */
#include "mxc_scc.h"
#include <asm/arch/mxc_security_api.h>
#include <linux/device.h> 

/*!
 * This variable indicates whether SCC module is in suspend or in resume
 * mode.
 */
static unsigned short scc_suspend_state = 0;

/*!
 * These variables store the SCC Module status registers values when SCC is
 * in suspend mode.
 */
#ifdef CONFIG_PM 
static unsigned long store_scm, store_smn;
#endif

/*!
 * This API is used to read the status of the Secure Memory module.
 *
 * @return  Status of the SCM module.
 */
ulong scc_scm_read_sts(void)
{
        SCC_DEBUG("In function %s. SCC STS = 0x%08lX\n", __FUNCTION__, \
                    SCC_SCM_STATUS);
        return SCC_SCM_STATUS;
}

/*!
 * This API is used to check if SCC is enabled or disabled.
 *
 * @return  SCC is enabled or disabled.
 */
scc_enable scc_check_enable(void)
{
        scc_enable ret_fuse_val;
        ulong scc_fuseval;

        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n", 
                        __FUNCTION__);
                return -EPERM;
        } else {
                scc_fuseval = SCC_FUSE;
                SCC_DEBUG("In function %s. Fuse = 0x%08X\n", __FUNCTION__, \
                           SCC_FUSE);

                if (scc_fuseval & SCC_DISABLED_MASK) {
                        ret_fuse_val = SCC_DISABLED;  /* SCC Disabled */
                } else {
                        ret_fuse_val = SCC_ENABLED;   /* SCC Enabled */
                }
                SCC_DEBUG("In function %s. ret fuse value = 0x%08X\n", \
                           __FUNCTION__, ret_fuse_val);
                }
                return ret_fuse_val;
}

/*!
 * This API configures Red Memory start address, Black Memory start address,
 * number of blocks of data and the init vector values of the KEM module (SCM).
 *
 * @param   red_addr       starting address of the red memory area.
 * @param   black_addr     starting address of the black memory area.
 * @param   cipher_length  number of blocks that needs to be ciphered.
 * @param   init_vect1     upper 32 bits of the initialization vector required
 *                         at the start of CBC (Cipher Block Chaining) mode of
 *                         ciphering.
 * @param   init_vect2     lower 32 bits of the initialization vector required
 *                         in CBC (Cipher Block Chaining) mode of ciphering at
 *                         the start of the processing of data.
 *
 * @return  SCC_SUCCESS  Successfully configured the KEM module.\n
 *          SCC_FAILURE  Encountered error while configuring the KEM module.
 *                       Length crosses maximum limit specified.
 */
scc_ret scc_kem_config(ulong red_addr, ulong black_addr, ulong cipher_length,
                        ulong init_vect1, ulong init_vect2)
{
        printk("SCC MODULE: Infunction scc_kem_config.\n");
        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n", 
                        __FUNCTION__);
                return -EPERM;
        } else {
                printk("In function %s. Red addr = 0x%08lX," 
                        "Black addr = 0x%08lX, Cipher length = 0x%08lX," 
                        "Init vect 1 = 0x%08lX, Init vect 2 = 0x%08lX\n", 
                        __FUNCTION__, red_addr, black_addr, cipher_length,                              init_vect1, init_vect2);
                if (cipher_length > SCC_MAX_CIPHER_LENGTH) {
                        printk("In function %s. Crosses max cipher length\n", \
                                __FUNCTION__);
                        /* Cipher length crosses the max limit specified. */
                        return SCC_FAILURE;
                } 
                SCC_RED_START_ADDR   = red_addr;
                SCC_BLACK_START_ADDR = black_addr;
                SCC_LENGTH           = cipher_length;
                SCC_INIT_VECT0       = init_vect1;
                SCC_INIT_VECT1       = init_vect2;
                printk("In function %s. Successfully configured the values.\n",\
                        __FUNCTION__);
        }
        return SCC_SUCCESS;
}

/*!
 * This API is used to load the plain data in the Red memory area and is
 * encrypted. Memory is allocated to the size that is configured in
 * the length register by the user and the pointer is passed as parameter.
 *
 * @param  encrypt_data  pointer to the data that needs to be encrypted.
 * @param  encrypt_length length (blocks) of data that needs to be encrypted.
 *
 * @return  SCC_SUCCESS  Successfully encrypted the data in the Red memory.\n
 *          SCC_FAILURE  Encountered error while encrypting data in Red memory.
 *                       Encryption length incorrect.
 */
scc_ret scc_kem_load_data2encrypt(ulong *encrypt_data, unchar encrypt_length)
{
        ulong no_of_bytes, i;
        volatile ulong *red_mem_ptr;
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Encrypt data = 0x%08lX, " \
                          "Encrypt length = 0x%02X\n", __FUNCTION__, \
                          (ulong)encrypt_data, encrypt_length);
                if (encrypt_length != SCC_LENGTH) {
                        SCC_DEBUG("In function %s.Crosses max cipher length\n",\
                                   __FUNCTION__);
                        /* Error in passing the encryption length. */
                        return SCC_FAILURE;
                }

                /*
                 * Calculating the number of bytes to be encrypted. Gives the 
                 * number of bytes that needs to be copied to the RAM area
                 */
                no_of_bytes = (SCC_LENGTH + 1) * (SCC_CONFIG & 0x7F);
                SCC_DEBUG("In function %s. Number bytes = 0x%08lX\n", \
                          __FUNCTION__, no_of_bytes);
                /*
                 * Calculating the address of the red memory area for
                 * storage of data
                 */
                red_mem_ptr = (ulong *)(SCC_RED_MEM_ADDR + \
                                ((SCC_CONFIG & 0x7F) * SCC_RED_START_ADDR));
                for (i = 0; i < no_of_bytes / 4; i++) {
                        SCC_DEBUG("In function %s. Encrypt data = 0x%08lX\n", \
                                   __FUNCTION__, encrypt_data[i]);
                        red_mem_ptr[i] = encrypt_data[i];
                }
                /* To configure the mode of Ciphering(Encryption/Decryption) */
                SCC_CONTROL |= SCC_ENCRYPT;
                /* To start the processing of Ciphering */
                SCC_CONTROL |= SCC_START_CIPHER;
                SCC_DEBUG("In function %s. Sucessfully started encryption."\
                          " Control reg = 0x%08lX\n", \
                          __FUNCTION__, SCC_CONTROL);
        }
        return SCC_SUCCESS;
}

/*!
 * This API is used to load the data to the Black memory area and decrypted.
 * The data is loaded into the Black memory area and is decrypted. Memory needs
 * to be allocated by the user to the size that is configured in the length
 * register and a pointer is passed as a parameter to this API.
 *
 * @param  decrypt_data  pointer to the data that needs to be decrypted.
 * @param  decrypt_length length (blocks) of data that needs to be encrypted.
 *
 * @return  SCC_SUCCESS  Successfully decrypted the data in the Black memory.\n
 *          SCC_FAILURE  Encountered error while decrypting data in Black
 *                       memory. Decryption length incorrect.
 */
scc_ret scc_kem_load_data2decrypt(ulong *decrypt_data, unchar decrypt_length)
{
        ushort no_of_bytes, i;
        volatile ulong *black_mem_ptr;

        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n", 
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Decrypt data = 0x%08lX, " \
                           "Decrypt length = 0x%02X\n", __FUNCTION__, \
                           (ulong) decrypt_data, decrypt_length);
                if (decrypt_length != SCC_LENGTH) {
                        SCC_DEBUG("In function %s.Crosses max cipher length\n",\
                                   __FUNCTION__);
                /* Error in passing the decryption length. */
                return SCC_FAILURE;
                }

                /*
                 * Calculating the number of bytes to be decrypted. Gives the
                 * number of bytes that needs to be copied to the RAM area
                 */
                no_of_bytes = (SCC_LENGTH + 1) * (SCC_CONFIG & 0x7F);
                SCC_DEBUG("In function %s. Number bytes = 0x%08X\n", \
                          __FUNCTION__, no_of_bytes);
                /*
                 * Calculating the address of the black memory area for
                 * storage of data
                 */
                black_mem_ptr = (ulong *)(SCC_BLACK_MEM_ADDR + \
                                ((SCC_CONFIG & 0x7F) * SCC_BLACK_START_ADDR));
                for (i = 0; i < no_of_bytes / 4; i++) {
                        SCC_DEBUG("In function %s. Decrypt data = 0x%08lX\n", \
                                   __FUNCTION__, decrypt_data[i]);
                        black_mem_ptr[i] = decrypt_data[i];
                }
                /* To configure the mode of Ciphering(Encryption/Decryption) */
                SCC_CONTROL |= SCC_DECRYPT;
                /* To start the processing of Ciphering */
                SCC_CONTROL |= SCC_START_CIPHER;
                SCC_DEBUG("In function %s. Sucessfully started decryption."\
                          "Control reg = 0x%08lX\n", __FUNCTION__, SCC_CONTROL);
        }
        return SCC_SUCCESS;
}

/*!
 * After ciphering (encryption) API is called, this API reads the data from the
 * Black memory area. Memory needs to be allocated by the caller to a size equal
 * to that as configured in the length register and passed as pointer to this
 * API.
 *
 * @param  encrypted_data  pointer where data is to be copied from the Black
 *                         Memory area.
 * @param  encrypt_length   length of data that needs to be encrypted.
 *
 * @return  SCC_SUCCESS  Successful in copying the data to the memory.\n
 *          SCC_FAILURE  Encountered error while processing the request.
 *                       Encrypt length crosses the max limit specified.
 */
scc_ret scc_kem_get_encrypt_data(ulong *encrypted_data, unchar encrypt_length)
{
        ushort no_of_bytes, i;
        volatile ulong *black_mem_ptr;

        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n", 
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Encrypt data = 0x%08lX, Decrypt"\
                          "length = 0x%02X\n", __FUNCTION__, \
                          (ulong)encrypted_data, encrypt_length);
                if (encrypt_length > SCC_MAX_CIPHER_LENGTH) {
                        SCC_DEBUG("In function %s.Crosses max cipher length\n",\
                                  __FUNCTION__);
                /* Encrypt length crosses the max limit specified. */
                return SCC_FAILURE;
                } 
                /* Calculating the number of bytes that needs to be copied. */
                no_of_bytes = (SCC_LENGTH + 1) * (SCC_CONFIG & 0x7F);
                SCC_DEBUG("In function %s. Number bytes = 0x%08X\n", \
                           __FUNCTION__, no_of_bytes);
                /*
                 * Calculating the address of the black memory area for
                 * storage of data
                 */
                black_mem_ptr = (ulong *)(SCC_BLACK_MEM_ADDR + \
                                ((SCC_CONFIG & 0x7F) * SCC_BLACK_START_ADDR));
                SCC_DEBUG("In function %s. Black mem ptr = 0x%08lX\n", \
                           __FUNCTION__, (ulong)black_mem_ptr);
                for (i = 0; i < no_of_bytes / 4; i++) {
                        SCC_DEBUG("In function %s. Black mem data =0x%08lX\n", \
                                   __FUNCTION__, black_mem_ptr[i]);
                        encrypted_data[i] = black_mem_ptr[i];
                }
        }
        return SCC_SUCCESS;
}

/*!
 * This API reads the data from the Red memory area. It is called after
 * data has been decrypted by loading data in the Black Memory area. Data is
 * read from the Red memory area and is stored in the pointer that is
 * pre-allocated to the number if bytes that is configured in the SCC length
 * register.
 *
 * @param  decrypted_data  pointer where data is to be copied from the Red
 *                          Memory area.
 * @param  decrypt_length  length of data that needs to be decrypted.
 *
 * @return  SCC_SUCCESS  Successful in copying the data to the memory.\n
 *          SCC_FAILURE  Encountered error while processing the request.
 *                       Decrypt length crosses the max limit specified.
 */
scc_ret scc_kem_get_decrypt_data(ulong *decrypted_data, unchar decrypt_length)
{
        ushort no_of_bytes, i;
        volatile ulong *red_mem_ptr;

        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n", 
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Decrypt data = 0x%08lX, Decrypt"\
                          "length = 0x%02X\n", __FUNCTION__, \
                          (ulong)decrypted_data, decrypt_length);
                if (decrypt_length > SCC_MAX_CIPHER_LENGTH) {
                        SCC_DEBUG("In function %s.Crosses max cipher length\n",\
                                   __FUNCTION__);
                        /* Decrypt length crosses the max limit specified. */
                        return SCC_FAILURE;
                } 
                /* Calculating the number of bytes that needs to be copied. */
                no_of_bytes = (SCC_LENGTH + 1) * (SCC_CONFIG & 0x7F);
                SCC_DEBUG("In function %s. Number bytes = 0x%08X\n", \
                           __FUNCTION__, no_of_bytes);
                /*
                 * Calculating the address of the red memory area for
                 * storage of data.
                 */
                red_mem_ptr = (ulong *)(SCC_RED_MEM_ADDR + \
                               ((SCC_CONFIG & 0x7F) * SCC_RED_START_ADDR));
                SCC_DEBUG("In function %s. Red mem ptr = 0x%08lX\n", \
                           __FUNCTION__, (ulong)red_mem_ptr);
                for (i = 0; i < no_of_bytes / 4; i++) {
                        SCC_DEBUG("In function %s. Decrypt data = 0x%08lX\n", \
                                   __FUNCTION__, red_mem_ptr[i]);
                        decrypted_data[i] = red_mem_ptr[i];
                }
        }
        return SCC_SUCCESS;
}

/*!
 * This API returns the status of ciphering process.
 *
 * @return  SCC_CIPHERING       KEM module still busy ciphering the data.\n
 *          SCC_CIPHERING_DONE  KEM module finished ciphering the data.
 *          SCC_CIPHERING_ERR   KEM module encountered error.
 */
scc_cip_sts scc_kem_cipher_status(void)
{
        ulong tempu32;
        scc_cip_sts cip_stat = SCC_CIPHERING_ERR;
        
        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCM Status = 0x%08lX\n", \
                           __FUNCTION__, SCC_SCM_STATUS);
                tempu32 = SCC_SCM_STATUS;
                /* Checking if ciphering and the busy bits are set */
                if ((((tempu32 & 0x0004) >> 0x02) == SCC_SET) && \
                    ((tempu32 & 0x0001)== SCC_SET)) {
                        cip_stat = SCC_CIPHERING;
                } else if ((((tempu32 & 0x0400) >> 0x0A) == SCC_SET) && \
                            ((tempu32 & 0x0001) == SCC_RESET)) {
                        /*
                         * Checking if ciphering done bit is set and the busy 
                         * bits is reset
                         */
                        cip_stat = SCC_CIPHERING_DONE;
                }
                SCC_DEBUG("In function %s. Cipher Status = 0x%08X\n", \
                           __FUNCTION__, cip_stat);
        }
        return cip_stat;
}

/*!
 * This API enables/disables the interrupt for SCM module.
 *
 * @param  config_int  enables or disables the interrupt.
 *
 * @return  SCC_SUCCESS  Successful in enabling the interrupt.\n
 *          SCC_FAILURE  Received an unexpected input.
 */
scc_ret scc_scm_config_int(scc_intr config_int)
{
        scc_ret ret_val = SCC_SUCCESS;
        
        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Interrupt configure = 0x%08X\n", \
                           __FUNCTION__, config_int);
                switch (config_int) {
                case SCC_INT_ENABLE:
                        SCC_INT_CTRL |= SCC_RESET;
                        break;

                case SCC_INT_DISABLE:
                        SCC_INT_CTRL |= SCC_SET;
                        break;

                default: 
                        ret_val = SCC_FAILURE;
                        break;                
                } 
        }
        return ret_val;
}

/*!
 * This API clears SCM interrupts that have been generated.
 *
 * @param  clr_int  clears the interrupt status.
 *
 * @return  SCC_SUCCESS  Successful in clearing the SCM interrupt.\n
 *          SCC_FAILURE  Received an unexpected input.
 */
scc_ret scc_scm_clear_int(scc_intr clr_int)
{
        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Interrupt clear = 0x%08X\n", \
                           __FUNCTION__, clr_int);
                if (clr_int != SCC_CLR_INT) {
                        return SCC_FAILURE;
                } 
        
                /* Clearing the interrupt bit from the status register. */
                SCC_INT_CTRL = SCC_SCM_INT_CLR;
        }
        return SCC_SUCCESS;
}

/*!
 * This API returns the memory size information for Secure RAM module.
 *
 * @param  scc_size  pointer to a structure to load the details pertaining to
 *                   the size information.
 */
scc_ret scc_kem_get_size_info (scc_mem_size *scc_size)
{
        ulong tempu32 = SCC_CONFIG;
        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCC Config = 0x%08lX, SCC Size = " \
                           "0x%08lX\n", __FUNCTION__, SCC_CONFIG, \
                           (ulong)scc_size);
                scc_size->scc_blk_size   = (tempu32 & 0x7F);
                scc_size->scc_red_size   = ((tempu32 & 0x01FF80) >> 0x07);
                scc_size->scc_black_size = ((tempu32 & 0x07FE0000) >> 0x11);
        }
        return SCC_SUCCESS;
}

/*!
 * This API returns the version ID of Secure Memory module.
 *
 * @return   Version ID of the Secure Memory module.
 */
ulong scc_scm_get_ver(void)
{
        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCC Config = 0x%08lX\n", \
                           __FUNCTION__, SCC_CONFIG);
        }
        return ((SCC_CONFIG & 0xF8000000) >> 0x1B);
}

/*!
 * This API returns the status of the SCM module.
 *
 * @return  Status of the SCM module.
 */
ulong scc_scm_get_status(void)
{
        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCC SCM sts = 0x%08lX\n", \
                           __FUNCTION__, SCC_SCM_STATUS);
        }
        return SCC_SCM_STATUS;
}

/*!
 * This API returns the error status of the SCM module.
 *
 * @return   Error status of the SCM module.
 */
ulong scc_scm_err_status(void)
{
        if (scc_suspend_state == 1)      {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCC ERR sts = 0x%08lX\n", \
                           __FUNCTION__, SCC_ERROR_STATUS);
        }
        return SCC_ERROR_STATUS;
}

/*!
 * This API returns the status of the SMN module.
 *
 * @return   Status of the SMN module.
 */
ulong scc_smn_get_status(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCC SMN sts = 0x%08lX\n", \
                           __FUNCTION__, SCC_SMN_STATUS);
        }
        return SCC_SMN_STATUS;
}

/*!
 * This API returns the current State of the SCC module.
 *
 * @return  State of the SCC module like Start, Zeroize RAM,
 *          Health Check, Fail, Secure, Non-Secure
 */
scc_state scc_get_state(void)
{
        scc_state scc_current_state;

        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                scc_current_state = (scc_state)(scc_smn_get_status() & \
                                     SCC_STATE);
                SCC_DEBUG("In function %s. SCC SMN sts = 0x%08lX\n", \
                           __FUNCTION__, scc_smn_get_status());
        }

        return scc_current_state;
}

/*!
 * This API returns the version ID of Secure Monitor.
 *
 * @return  Version ID of the Secure Monitor.
 */
ulong scc_smn_get_ver(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCC SMN sts = 0x%08lX\n", \
                           __FUNCTION__, scc_smn_get_status());
        }
        return ((scc_smn_get_status() >> 27) & 0x1F);
}

/*!
 * This API returns the boot state of the processor.
 *
 * @return  SCC_INTERNAL  Processor has booted internally.\n
 *          SCC_EXTERNAL  Processor has booted externally.
 */
scc_boot_state scc_check_boot(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCC SMN sts = 0x%08lX\n", \
                           __FUNCTION__, scc_smn_get_status());
        }
        return ((scc_smn_get_status() >> 1) & 0x01);
}

/*!
 * This API clears the status bits from the SMN status register. Values are
 *  considered from the smn_sts_clr data type.
 *
 * @param    smn_status_clr  to clear the status bits set in the SMN status
 *                           register.
 *
 * @return  SCC_SUCCESS      Successful in clearing the status of the requested
 *                           fields.\n
 *          SCC_FAILURE      Unsuccessful due to error in the parameter passed.
 */
scc_ret scc_smn_clear_status(scc_smn_sts_clr smn_status_clr)
{
        scc_smn_sts_clr smn_sts_chk;
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCC SMN sts clr = 0x%08X\n", \
                           __FUNCTION__, smn_status_clr);
                smn_sts_chk = SCC_ZEROIZE_FAIL_CLR | SCC_SEC_POL_ERR_CLR | \
                      SCC_ASC_ERR_CLR | SCC_BITBANK_CLR | SCC_PC_ERR_CLR | \
                      SCC_TMR_ERR_CLR | SCC_SW_ALARM_CLR | SCC_ILGAL_ACS_CLR | \
                      SCC_BAD_KEY_CLR | SCC_USR_ACS_CLR | SCC_ILGAL_ADR_CLR | \
                      SCC_BYTE_ACS_CLR | SCC_UNALGND_ACS_CLR | \
                      SCC_ILGAL_MSTR_CLR | SCC_CACHE_ACS_CLR;
                SCC_DEBUG("In function %s. SCC SMN sts chk = 0x%08X\n", \
                           __FUNCTION__, smn_sts_chk);
                /* Checking if any read-only bit position is been set. */
                if ((~(smn_sts_chk)) & smn_status_clr) {
                        return SCC_FAILURE;
                } 
        
                SCC_SMN_STATUS |= smn_status_clr;
                SCC_DEBUG("In function %s. SCC SMN sts = 0x%08lX\n", \
                           __FUNCTION__, SCC_SMN_STATUS);
        }
        return SCC_SUCCESS;
}

/*!
 * This API sets the alarm bit. Setting this bit causes the SCC module to
 * enter the Fail state. This API is called during the detection of intrusion.
 */
scc_ret scc_set_sw_alrm(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_COMMAND |= SCC_SET_ALARM;
                SCC_DEBUG("In function %s. SCC SMN sts = 0x%08lX\n", \
                           __FUNCTION__, scc_smn_get_status());
        }
        return SCC_SUCCESS;
}

/*!
 * This API clears the interrupt bit set due to an interrupt from the SMN
 * module.
 */
scc_ret scc_smn_clear_int(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_COMMAND |= SCC_SMN_INT_CLR;
                SCC_DEBUG("In function %s. SCC SMN sts = 0x%08lX\n", \
                           __FUNCTION__, scc_smn_get_status());
        }
        return SCC_SUCCESS;
}

/*!
 * This API enables or disables the interrupt for the SMN module.
 *
 * @param    config_int  interrupt configuration (ENABLE/DISABLE).
 *
 * @return   SCC_SUCCESS  Successfully configured the interrupt for SMN module.
 *           SCC_FAILURE  Error in the parameter passed.
 */
scc_ret scc_smn_config_int(scc_intr config_int)
{
        scc_ret ret_val = SCC_SUCCESS;
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Interrupt Configure = 0x%08X\n", \
                           __FUNCTION__, config_int);
                switch (config_int) {
                case SCC_INT_ENABLE:
                        SCC_COMMAND |= SCC_SMN_INT_ENABLE;
                        break;

                case SCC_INT_DISABLE:
                        SCC_COMMAND &= SCC_SMN_INT_DISABLE;
                        break;

                default:
                        ret_val = SCC_FAILURE;
                        break;
                }
        
                SCC_DEBUG("In function %s. SCC CMD = 0x%08lX\n", \
                           __FUNCTION__, SCC_COMMAND);
        }
        return ret_val;
}

/*!
 * This API clears the Red Memory & Black Memory. This API will stop any
 * ciphering that is in progress.
 *
 * @return  SCC_SUCCESS  Zeroization process was started successfully.\n
 *          SCC_FAILURE  Zeroization process cannot be started as SCC is not in
 *                       Secure or Non-Secure state.
 */
scc_ret scc_zeroize(void)
{
        scc_state cur_state;

        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                cur_state = scc_get_state();
                SCC_DEBUG("In function %s. SCC SMN sts = 0x%08lX\n", \
                           __FUNCTION__, scc_smn_get_status());
                if ((cur_state == SCC_SECURE_STATE) || \
                   (cur_state == SCC_NON_SECURE_STATE)) {
                        SCC_DEBUG("Current State 0x%02X. Setting zeroize.\n",
                                   cur_state);
                        SCC_INT_CTRL |= SCC_ZEROIZE_MEM;
                } else {
                        SCC_DEBUG("Current State 0x%02X. Not able to " \
                                   "zeroize.\n", cur_state);
                return SCC_FAILURE;
                }
        }
        return SCC_SUCCESS;
}

/*!
 * This API returns the zeroization status of the Secure RAM.
 *
 * @return     Zeroization status of the Secure RAM module.
 *             (ZEROIZING/ZEROIZE_FAIL/ZEROIZE_DONE).
 */
scc_zeroize_state scc_zeroize_status(void)
{
        scc_zeroize_state zeroize_status = 0x00;
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCC SCM sts = 0x%08lX\n", \
                           __FUNCTION__, SCC_SCM_STATUS);
                if ((SCC_SCM_STATUS & SCC_ZEROIZING) == SCC_ZEROIZING) {
                        zeroize_status = SCC_ZEROIZING;
                } else if ((SCC_SCM_STATUS & SCC_ZEROIZE_FAIL) == \
                            SCC_ZEROIZE_FAIL) {
                        zeroize_status = SCC_ZEROIZE_FAIL;
                } else if ((SCC_SCM_STATUS & SCC_ZEROIZE_DONE) == \
                            SCC_ZEROIZE_DONE) {
                        zeroize_status = SCC_ZEROIZE_DONE;
                }
        }
        return zeroize_status;
}

/*!
 * This API clears the zeroize bit set in the SMN status register caused due to
 * failure in the zeroization module.
 */
scc_ret scc_zeroize_fail_clear(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_SMN_STATUS |= SCC_SMN_ZEROIZE_FAIL_CLR;
                SCC_DEBUG("In function %s. SCC SMN sts = 0x%08lX\n", \
                           __FUNCTION__, scc_smn_get_status());
        }
        return SCC_SUCCESS;
}

/*!
 * This API puts the data sent from the upper layer to the RED memory area.
 * Memory needs to be allocated by the caller, which is equal to the size
 * requested.
 *
 * @param      redmem_data   pointer to the memory location from where data is
 *                           to be copied to the Red memory area.
 * @param      redmem_size   number of data words that needs to be copied
 *                           to the Red memory area.
 *
 * @return   SCC_SUCCESS     Data copied successfully from the buffer to the
 *                           Red memory area.\n
 *           SCC_FAILURE     Illegal or out of boundary length specified.
 */
scc_ret scc_redmem_put_data(ulong *redmem_data, ushort redmem_size)
{
        ushort i, size_calc;
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Red mem data = 0x%03lX, Red mem "\
                          "size=0x%04X\n", __FUNCTION__, (ulong)redmem_data, \
                           redmem_size);
                /* Calculating the maximum size of red memory area in words */
                size_calc = (SCC_MAX_CIPHER_LENGTH + 1) * 8 / 4;
                if (redmem_size <= size_calc) {
                        for (i = 0x00; i < redmem_size; i++) {
                                *(SCC_RED_MEM_ADDR + i) = *(redmem_data + i);
                        }
                } else {
                        return SCC_FAILURE;
                }
        }
        return SCC_SUCCESS;
}

/*!
 * This API puts the data sent from the upper layer to the BLACK memory area.
 * Memory needs to be allocated by the caller, which is equal to the size
 * requested.
 *
 * @param    blackmem_data   pointer to the memory location from where data is
 *                           to be copied to the Black memory area.
 * @param    blackmem_size   number of data words that needs to be copied
 *                           to the Black memory area.
 *
 * @return   SCC_SUCCESS     Data copied successfully from the buffer to the
 *                           Black memory area.\n
 *           SCC_FAILURE     Illegal or out of boundary length specified.
 */
scc_ret scc_blackmem_put_data(ulong *blackmem_data, ushort blackmem_size)
{
        ushort i, size_calc;

        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Black mem data = 0x%08lX,Black mem" \
                          "size = 0x%04X\n", __FUNCTION__, \
                          (ulong)blackmem_data, blackmem_size);
                /* Calculating the maximum size of red memory area in words */
                size_calc = (SCC_MAX_CIPHER_LENGTH + 1) * 8 / 4;
                if (blackmem_size <= size_calc) {
                        for (i = 0x00; i < blackmem_size; i++) {
                                *(SCC_BLACK_MEM_ADDR + i) = *(blackmem_data + i);
                        }
                } else {
                        return SCC_FAILURE;
                }
        }
        return SCC_SUCCESS;
}

/*!
 * This API gets the data requested by the upper layer from the RED memory area.
 * Memory needs to be allocated by the caller, which is equal to the size
 * requested.
 *
 * @param      redmem_data   pointer to the buffer where data from the RED
 *                           memory is to be copied.
 * @param      redmem_size   number of data words that needs to be copied
 *                           to the memory.
 *
 * @return   SCC_SUCCESS     Data copied successfully to the buffer from the
 *                           Red memory area.\n
 *           SCC_FAILURE     Illegal or out of boundary length specified.
 */
scc_ret scc_redmem_get_data(ulong *redmem_data, ushort redmem_size)
{
        ushort i, size_calc;

        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Red mem data = 0x%08lX, Red mem " \
                          "size = 0x%08X\n", __FUNCTION__, (ulong)redmem_data,\
                           redmem_size);
                /* Calculating the maximum size of red memory area in words */
                size_calc = (SCC_MAX_CIPHER_LENGTH + 1) * 8 / 4;
                if (redmem_size <= size_calc) {
                        for (i = 0; i < redmem_size; i++) {
                                *(redmem_data + i) = *(SCC_RED_MEM_ADDR + i);
                        }
                } else {
                        return SCC_FAILURE;
                }
        }
        return SCC_SUCCESS;
}

/*!
 * This API gets the data requested by the upper layer from the BLACK memory
 * area. Memory needs to be allocated by the caller, which is equal to the size
 * requested.
 *
 * @param      blackmem_data pointer to the buffer where data from the RED
 *                           memory is to be copied.
 * @param      blackmem_size number of data words that needs to be copied
 *                           to the memory.
 *
 * @return   SCC_SUCCESS     Data copied successfully to the buffer from the
 *                           Black memory area.\n
 *           SCC_FAILURE     Illegal or out of boundary length specified.
 */
scc_ret scc_blackmem_get_data(ulong *blackmem_data, ushort blackmem_size)
{
        ushort i, size_calc;

        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Black mem data = 0x%08lX, Black mem"\
                          "size =0x%04X\n", __FUNCTION__, (ulong)blackmem_data,\                          blackmem_size);
                /* Calculating the maximum size of red memory area in words */
                size_calc = (SCC_MAX_CIPHER_LENGTH + 1) * 8 / 4;
                if (blackmem_size <= size_calc) {
                        for (i = 0x00; i < blackmem_size; i++) {
                                *(blackmem_data + i) = *(SCC_BLACK_MEM_ADDR + \
                                i);
                        }
                } else {
                        return SCC_FAILURE;
                }
        }
        return SCC_SUCCESS;
}

/*!
 * This API configures the start and the end sequence number for the ASC module
 *
 * @param   asc_start_seq  starting sequence number to be configured to the ASC
 *                         module.
 * @param   asc_end_seq    ending sequence number to be configured to the ASC
 *                         module.
 *
 * @return   SCC_SUCCESS   Successfully configured the Start and End sequence
 *                         number.\n
 *           SCC_FAILURE   Illegal configuration value specified. Start or End
 *                         sequence number may have been specified as 0x00.
 */
scc_ret scc_asc_configure(ushort asc_start_seq, ushort asc_end_seq)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. ASC start seq = 0x%04X, ASC end " \
                          "seq= 0x%04X\n", __FUNCTION__, asc_start_seq, \
                          asc_end_seq);
                /* Start or end value cannot be NULL. */
                if ((asc_start_seq == 0x00) || (asc_end_seq == 0x00)) {
                        return SCC_FAILURE;
                } 
                SCC_SEQ_START = asc_start_seq;
                SCC_SEQ_END   = asc_end_seq;
        }
        return SCC_SUCCESS;
}

/*!
 * This API sets the sequence number during the Algorithm sequence checking.
 *
 * @param   asc_seq_chk  sequence number to be validated for the sequence flow.
 *
 * @return   SCC_SUCCESS   Successfully written the next sequence number.\n
 *           SCC_FAILURE   Illegal sequence check value (0x00) specified.
 */
scc_ret scc_asc_set_seq(ushort asc_seq_chk)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. ASC seq chk = 0x%04X\n", \
                          __FUNCTION__, asc_seq_chk);
                /* Sequence value cannot be NULL. */
                if (asc_seq_chk == 0x00) {
                        return SCC_FAILURE;
                } 
                SCC_SEQ_CHK = asc_seq_chk;
        }
        return SCC_SUCCESS;
}

/*!
 * This API generates the next sequence number for the given current sequence
 * number. It implements the LFSR (Linear Feedback Shift Register)
 * functionality. It uses the polynomial function:
 * F(X) = X^16 + X^15 + X^13 + X^4 + 1.
 * It can be used to generate the next sequence number wherein the current
 * sequence number is known. The 16 bits current sequence number is passed as a
 * parameter to this API and the next 16 bits sequence number is calculated and
 * returned to the caller.
 *
 * @param  seq_num   current sequence number to generate the next sequence
 *                   number.
 *
 * @return           Next sequence number calculated by the polynomial
 *                   function.
 */
ushort scc_asc_get_nxtseq(ushort seq_num)
{
        ushort count, chk_bits;
        const ushort poly_num = 0xA011;
        
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. seq num = 0x%04X\n", __FUNCTION__, \
                           seq_num);
                chk_bits = seq_num & poly_num;
                /* Counting the number of bits set */
                for (count = 0;chk_bits != 0;chk_bits &= (chk_bits-1)) {
                        count++;
                }
                seq_num >>= 1;

                if (count%2 != 0) {
                        seq_num |= 0x8000;
                }
        }
        return seq_num;
}

/*!
 * This API returns the error status of the AIC module.
 *
 * @return  Error status of the ASC or Bit Bank or PC of the AIC module.
 */
scc_aic_ret  scc_aic_bit_status(void)
{
        ulong temp_val;
        scc_aic_ret asc_sts_ret = SCC_AIC_NO_ERROR;
        
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                temp_val = scc_smn_get_status();
                SCC_DEBUG("In function %s. SMN sts = 0x%08lX\n",\
                           __FUNCTION__, temp_val);
                if(((temp_val >> 0x09) & 0x01) == SCC_SET) {
                        asc_sts_ret = SCC_ASC_ERR;
                } else if (((temp_val >> 0x0A) & 0x01) == SCC_SET) {
                        asc_sts_ret = SCC_BIT_BNK_ERR;
                } else if (((temp_val >> 0x0B) & 0x01) == SCC_SET) {
                        asc_sts_ret = SCC_PC_ERR;
                }
        }
        return asc_sts_ret;
}

/*!
 * This API clears the Bit Bank count register of the Bit Bank module.
 */
scc_ret scc_bitbnk_clear(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_COMMAND |= SCC_BIT_BANK_CLR;
                SCC_DEBUG("In function %s. Bit count = 0x%04lX\n", \
                           __FUNCTION__, SCC_BIT_CNT);
        }
        return SCC_SUCCESS;
}

/*!
 * This API returns the Bit count value of the Bit Bank module.
 *
 * @return   Bit count value of the Bit Bank module.
 */
ushort scc_rd_bitcnt(void)
{
        ushort scc_bit_cnt_ret;
        
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                scc_bit_cnt_ret = SCC_BIT_CNT;
                SCC_DEBUG("In function %s. Bit count = 0x%04lX\n", \
                           __FUNCTION__, SCC_BIT_CNT);
        }
        return scc_bit_cnt_ret;
}

/*!
 * This API returns the size value that should be compared between the
 * plain/ciphertext comparator.
 *
 * @return   Size of the bits to be used for comparison between the
 *           plain/ciphertext comparator.
 */
unchar scc_rd_bit_cmp(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Bit cmp = 0x%02lX\n", __FUNCTION__, \
                           SCC_CMP_SIZE);
        }
        return SCC_CMP_SIZE;
}

/*!
 * This API is used to read the Increment/Decrement size register of the Bit
 * Bank module.
 *
 * @param  read_reg   specify SCC_READ_INCR or SCC_READ_DECR register of the
 *                    Bit Bank module to be read.
 * @return            Value of the Increment/Decrement register of the Bit
 *                    Bank module. If the argument passed is invalid then
 *                    SCC_AIC_ERROR is returned to the caller.
 */
scc_aic_ret scc_rd_inc_dec_reg(scc_aic_param read_reg)
{
        scc_aic_ret ret_regval = SCC_AIC_ERROR;
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Read val = 0x%08X\n", __FUNCTION__, \
                           read_reg);
                if (read_reg == SCC_READ_INCR) {
                        ret_regval = SCC_BIT_INCR_SIZE & 0x7FF;
                } else if (read_reg == SCC_READ_DECR) {
                        ret_regval = SCC_BIT_DECR & 0x7FF;
                }
        }
        return ret_regval;
}

/*!
 * This API configures the Bit Bank increment register; decrement
 * register and the bit compare register of the Bit Bank module.
 *
 * @param  bit_bnk_incr  bit bank increment size.
 * @param  bit_bnk_decr  bit bank decrement size.
 * @param  bit_cmp       bit bank compare size.
 *
 * @return  SCC_SUCCESS  Successfully configured the Bit Bank module.\n
 *          SCC_FAILURE  Parameters passed has some invalid data.
 */
scc_ret scc_bitbnk_configure(ushort bit_bnk_incr, ushort bit_bnk_decr, unchar
                           bit_cmp)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Incr = 0x%08X, Decr = 0x%08X, " \
                           "Cmp = 0x%08X\n", __FUNCTION__, bit_bnk_incr, \
                          bit_bnk_decr, bit_cmp);
                /* Checking if any of the parameters are crossing the 
                 MAX limits. */
                if ((bit_bnk_incr > SCC_BIT_BNK_MAX_SIZE) || \
                    (bit_bnk_decr > SCC_BIT_BNK_MAX_SIZE) || \
                    (bit_cmp > SCC_BIT_BNK_MAX_CMP_SIZE)) {
                        return SCC_FAILURE;
                } 
                SCC_BIT_INCR_SIZE = (ulong)bit_bnk_incr;
                SCC_BIT_DECR = (ulong)bit_bnk_decr;
                SCC_CMP_SIZE = (ulong)bit_cmp;
        }
        return SCC_SUCCESS;
}

/*!
 * This API is used to write data to the Plain/Cipher text register that is used
 * to compare the plain/ciphertext data as a software assurance feature.
 *
 * @param    cipher_data   data that needs to be written to the Plaintext check
 *                         register or the Ciphertext check register.
 * @param    cipher_switch parameter to indicate where to write the sent
 *                         data (Plaintext check register/Ciphertext check
 *                         register).
 *
 * @ return  Invalid argument passed for scc_aic_param parameter.
 */
scc_ret scc_wr_pln_cip(ulong cipher_data, scc_aic_param cipher_switch)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Ciper data = 0x%08lX, Switch = " \
                          "0x%08X\n", __FUNCTION__, cipher_data, cipher_switch);
                if (cipher_switch == SCC_WRITE_PLAIN_TEXT) {
                        SCC_PLN_CHK = cipher_data;
                } else if (cipher_switch == SCC_WRITE_CIPHER_TEXT) {
                        SCC_CIP_CHK = cipher_data;
                } else {
                        return SCC_FAILURE;
                }
        }
        return SCC_SUCCESS;
}

/*
 * This API loads the Security Timer register with the value passed as
 * parameter.
 */
scc_ret scc_ldstrt_sec_tmr(unchar tmr_ctrl)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Tmr ctrl val = 0x%02X\n", \
                           __FUNCTION__, tmr_ctrl);
                SCC_TIM_CTRL = tmr_ctrl;
        }
        return SCC_SUCCESS;
}

/*!
 * This API configures the Security Timer with the new reload timer value.
 *
 * @param  reload_tmr_val  re-load timer value for the Security Timer.
 */
scc_ret scc_sec_config_tmr(ulong reload_tmr_val)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Tmr reload val = 0x%08lX\n", \
                           __FUNCTION__, reload_tmr_val);
                SCC_TIM_IV = reload_tmr_val;
        }
        return SCC_SUCCESS;
}

/*!
 * This API configures the Security timer register to Loads or Starts or
 * Loads & Starts or Stops the Security timer.
 *
 * @param   ld_strt_val  configuration values to Load or Start or Load & Start
 *                       or Stop the Security Timer.
 * @return  SCC_SUCCESS  Successfully configured the Security timer.\n
 *          SCC_FAILURE  Invalid parameter is passed.
 */
scc_ret scc_sec_ctrl_tmr(unchar ld_strt_val)
{
        unchar tmr_ctrl_val;
        scc_ret ret_val = SCC_SUCCESS;
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Tmr loadstart val = 0x%02X\n", \
                           __FUNCTION__, ld_strt_val);
                /*
                 * If there is error in passing the argument then a SCC_FAILURE
                 * is returned to the caller.
                */
                switch (ld_strt_val) {
                case SCC_LOAD_TMR:
                        tmr_ctrl_val = SCC_LOAD_TMR;
                        break;

                case SCC_START_TMR:
                        tmr_ctrl_val = SCC_START_TMR;
                        break;

                case SCC_LOADNSTART_TMR:
                        tmr_ctrl_val = SCC_LOADNSTART_TMR;
                        break;

                case SCC_STOP_TMR:
                        tmr_ctrl_val = SCC_STOP_TMR;
                        break;

                default:
                        ret_val = SCC_FAILURE;
                        break;
                }
        
                if (ret_val != SCC_FAILURE) {
                        scc_ldstrt_sec_tmr(tmr_ctrl_val);
                }
        }
        return ret_val;
}

/*!
 * This API reads the current value of the Security Timer.
 *
 * @return  Current Security Timer value.
 */
ulong scc_sec_rd_tmr(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Tmr val = 0x%08lX\n", \
                           __FUNCTION__, SCC_TIM_RD);
        }
        return SCC_TIM_RD;
}

/*!
 * This API reads the current reload value of the Security Timer.
 *
 * @return  Current Security Timer reload value.
 */
ulong scc_sec_rload_tmr(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. Tmr reload val = 0x%08lX\n", \
                           __FUNCTION__, SCC_TIM_IV);
        }
        return SCC_TIM_IV;
}

/*!
 * This API returns the status of the Security Timer error indicated in the SMN
 * Status register.
 *
 * @return  SCC_TMR_ERR      Security Timer has expired.\n
 *          SCC_NO_TMR_ERR   Security Timer has not expired.
 */
scc_tmr_sts scc_sec_rd_err_status(void)
{
        unchar tmr_err_sts;
        ulong temp_val;

        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                temp_val = scc_smn_get_status();
                SCC_DEBUG("In function %s. SMN sts = 0x%08lX\n", \
                           __FUNCTION__, temp_val);

                if (((temp_val >> 0x0C) & 0x01) == SCC_SET) {
                        tmr_err_sts = SCC_TMR_ERR;
                } else {
                tmr_err_sts = SCC_NO_TMR_ERR;
                }
        }
        return tmr_err_sts;
}

/*!
 * This API returns the Debug Detect status of the SCC module.
 *
 * @return  Status of the Debug Detect of the SCC module.
 */
ulong scc_dbdet_rd_status(void)
{
        if (scc_suspend_state == 1) {
                printk("In function %s.SCC Module is in suspend mode.\n",
                        __FUNCTION__);
                return -EPERM;
        } else {
                SCC_DEBUG("In function %s. SCC dbg sts = 0x%08lX\n", \
                           __FUNCTION__, SCC_DBG_DET_STAT);
        }
        return SCC_DBG_DET_STAT;
}

#ifdef CONFIG_PM
/*!
 * This function is called to put the SCC in a low power state. Refer to the
 * document driver-model/driver.txt in the kernel source tree for more
 * information.
 *
 * @param   dev   the device structure used to give information on HAC
 *                to suspend.
 * @param   state the power state the device is entering.
 * @param   level the stage in device suspension process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
scc_ret scc_suspend(struct device *dev, ulong state, ulong level)
{
        unsigned short reg_val;
        scc_ret ret_val = SCC_SUCCESS;
        printk(" SCC suspend function.%08lX\n", level);
        switch(level) {
        case SUSPEND_DISABLE:
                SCC_DEBUG("In function %s. In suspend Disable\n");
                /* Setting suspend flag. */
                scc_suspend_state = 1;
                break;

        case SUSPEND_SAVE_STATE:
                SCC_DEBUG(" In Suspend save state.\n");
                /* Store SCM Status register value. */
                reg_val = SCC_SCM_STATUS;
                store_scm = reg_val;
                /* Store SMN Status register value. */
                reg_val = SCC_SMN_STATUS;
                store_smn = reg_val;
                break;

       case SUSPEND_POWER_DOWN:
                SCC_DEBUG(" In suspend power down.\n");
                /* Setting suspend flag. */
                scc_suspend_state = 1;
                break;

        default:
                ret_val = SCC_FAILURE;
                break;
        }
        return ret_val;
}

/*!
 * This function is called to bring the SCC back from a low power state.
 * Refer to the document driver-model/driver.txt in the kernel source tree
 * for more information.
 *
 * @param   dev   the device structure used to give information on SCC
 *                to resume.
 * @param   level the stage in device resumption process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
scc_ret scc_resume(struct device *dev, ulong level)
{
        scc_ret ret_val = SCC_SUCCESS;
        printk(" SCC resume function.%08lX\n", level);
        switch(level) {
        case RESUME_POWER_ON:
                SCC_DEBUG(" Resume power on.\n");
                /* Clearing the suspend flag. */
                scc_suspend_state = 0;
                break;

        case RESUME_RESTORE_STATE:
                SCC_DEBUG(" Resume restore state.\n");
                SCC_DEBUG(" Get SCM status. \n");
                SCC_SCM_STATUS = store_scm;
                SCC_DEBUG(" Get SMN status. \n");
                SCC_SMN_STATUS = store_smn;
                break;

        case RESUME_ENABLE:
                SCC_DEBUG("Resume Enable state called.\n");
                /* Clearing the suspend flag. */
                scc_suspend_state = 0;
                break;

        default: 
                ret_val = SCC_FAILURE;
                break;
        }
        return ret_val;
}
#else
#define mxc_scc_suspend NULL
#define mxc_scc_resume  NULL
#endif /* CONFIG_PM */
