/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 *
 * http://www.opensource.org/licenses/gpl-licensisr_locke.html
 * http://www.gnu.org/copyleft/gpl.html
 */ 

/*!
 * @file spi-core.c
 * @brief This file contains the implementation of the SPI driver main services
 *
 *
 * @ingroup SPI
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

/*
 * Include files
 */
#include <linux/kernel.h>
#include <asm/irq.h>
#include <asm/arch/gpio.h>
#include <linux/module.h>
#include "spi-internal.h"
#include "registers.h"

extern void gpio_spi_active(int cspi_mod);
extern void gpio_spi_inactive(int cspi_mod);
extern unsigned long spi_max_bit_rate;

static int spi_hard_config(module_nb_t mod, spi_config *client_config);
static int spi_release(module_nb_t mod);
static int spi_request(module_nb_t mod, client_prio_t prio);

/*!
 * Global variable which contains the context of the SPI driver
 */
spi_mxc_port spi_port[CONFIG_SPI_NB_MAX];

/*!
 * Global variable which contains the adresses of the SPI driver
 */
spi_mxc_add spi_add[CONFIG_SPI_NB_MAX];

/*!
 * Global variable which indicates which SPI is used
 */
bool spi_present[CONFIG_SPI_NB_MAX];

/*!
 * Global array containing the name of device's IRQs
 */
static char cspi_irq_name[CONFIG_SPI_NB_MAX][11];

/*!
 * This function is called when an interrupt occurs on the SPI modules.
 * It is the interrupt handler for the SPI modules.
 *
 * @param        irq        the irq number
 * @param        dev_id     the pointer on the device
 * @param        regs       the interrupt parameters
 *
 * @return       The function returns IRQ_RETVAL(1) when handled.
 */
static irqreturn_t spi_isr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
        int handled = 0;
        int nb_bytes = 0 ;
        module_nb_t module = 0;
        spi_mxc_port *port = NULL;
        unsigned int buf_size_empty = 0;
        
        port = (spi_mxc_port *)(dev_id);
        module = port->hw_config->module_number;
        buf_size_empty = port->bytes_to_send - port->bytes_received;

        DBG_PRINTK("spi%d : spi_isr_handler : rx_fifo_data_ready \n",
                   module + 1);

        if (buf_size_empty > 0) {
               	nb_bytes = spi_get_rx_data(module, port->rx_buffer,
                                           buf_size_empty);
        }

        spi_clear_interrupt(module, SPI_STATUS_RX_FIFO_DATA_READY);
        port->rx_buffer += nb_bytes;
        port->bytes_received += nb_bytes;

        if (port->bytes_received >= port->bytes_to_send) {
                /* wakeup a synchronous call to spi_send_frame */
        	up(&(port->sync_interface));
        }

        handled = 1;

        return IRQ_RETVAL(handled);
}

/*!
 * This function frees resources when unloaded.
 * It unbinds the interrupts.
 *
 * @return       This function returns 0 if successful, -EPERM otherwise.
 */
int spi_core_exit(void)
{
        int spi = 0;
        for (spi = 0; spi < CONFIG_SPI_NB_MAX; spi++){
                if (spi_present[spi]) {
                        free_irq(spi_port[spi].int_cspi, spi_isr_handler);
                        gpio_spi_inactive(spi);
                }
        }
        return spi_release_spba();
}

/*!
 * This function gets a SPI device handler.
 * It allocates a spi_config structure to manage the SPI device configuration.
 *
 * @param        config        the SPI device configuration
 * @return       This function returns a handle or 0 if error
 */
void* spi_get_device_id(spi_config *config)
{
        void *handle;

        if (config->bit_rate > spi_max_bit_rate) {
                return 0;
        }	

        if (config->tx_delay > MAX_TX_DELAY) {
                return 0;
        }	

        if ((config->bit_count <= 0) ||
            ((config->bit_count % TRANSFER_8_BITS) != 0)) {
                return 0;
        }	
 
        handle = (void *)config;

        return handle;
}

/*!
 * This function inititializes register address and set default configuration.
 *
 * @param        spi        the module number
 * @return       This function returns 0 if successful, -EPERM otherwise.
 */
int spi_init_default_conf(module_nb_t spi) 
{
        int i = 0, j = 0;
        int error = 0;

        /* Configure SPI registers adresses */
        spi_add[spi].rx_address = spi_add[spi].base_address + 
                        OFFSET_CSPI_RXDATAREG;
        spi_add[spi].tx_address = spi_add[spi].base_address + 
                        OFFSET_CSPI_TXDATAREG;
        spi_add[spi].ctrl_address = spi_add[spi].base_address +
                        OFFSET_CSPI_CONTROLREG;
        spi_add[spi].int_address = spi_add[spi].base_address +
                        OFFSET_CSPI_INTREG;
        spi_add[spi].stat_address = spi_add[spi].base_address +
                        OFFSET_CSPI_STATREG;
        spi_add[spi].period_address = spi_add[spi].base_address +
                        OFFSET_CSPI_PERIODREG;
        spi_add[spi].test_address = spi_add[spi].base_address +
                        OFFSET_CSPI_TESTREG;

        /* Priority mutex initialization */
        for (j= LOW ; j < MAX_PRIORITY; j++) { 
                for (i = 0; i < MAX_SPI_CLIENTS; i++) {
                        init_MUTEX_LOCKED(&(spi_port[spi].pending_mutex[j][i]));
                }
                /* Priority pending_clients initialization */
                spi_port[spi].pending_clients[j] = 0;
        }

        /* Priority in_use boolean initialization */
        spi_port[spi].in_use = false;

        /* SPI current handle configuration */
        spi_port[spi].current_config_handle = NULL;

        /* SPI port access mutex initialization */
        init_MUTEX(&(spi_port[spi].access_mutex));

        /* SPI exchange mutex initialization */
        init_MUTEX_LOCKED(&(spi_port[spi].sync_interface));

        /* Software reset of the module */
        spi_disable(spi);
        /* Module enabled */
        spi_enable(spi);

        /* Disable all interrupt sources */
        error = spi_disable_interrupt(spi, 
                            SPI_STATUS_TX_FIFO_EMPTY |
                            SPI_STATUS_TX_FIFO_HALF_EMPTY |
                            SPI_STATUS_TRANSFER_COMPLETE |
                            SPI_STATUS_RX_FIFO_DATA_READY | 
                            SPI_STATUS_TX_FIFO_FULL |
                            SPI_STATUS_RX_FIFO_HALF_FULL |
                            SPI_STATUS_RX_FIFO_FULL |
                            SPI_STATUS_RX_FIFO_OVERFLOW |
                            SPI_STATUS_BIT_COUNT_OVERFLOW);
        if (error != 0 ){
                return error;
        }
        /* Request IRQ */
        sprintf(&cspi_irq_name[spi][0], "CSPI%1d_IRQ", spi + 1);
        error = request_irq(spi_port[spi].int_cspi, spi_isr_handler,
                            0, cspi_irq_name[spi], (void*)(&spi_port[spi]));
        if (error != 0 ){
                return error;
        }
        /* enable IT */
        error = spi_enable_interrupt(spi, SPI_STATUS_RX_FIFO_DATA_READY);
        if (error != 0 ){
                return error;
        }
        return 0;
}

/*!
 * This function configures the SPI module from an hardware perspective.
 *
 * @return       This function returns 0 if successful, -EPERM otherwise.
 */
int spi_hw_init(void)
{
        int spi = 0;
        int error = 0;
       /* Configure GPIO */
        for (spi = 0; spi < CONFIG_SPI_NB_MAX; spi++){
                if (spi_present[spi]) {
                        gpio_spi_active(spi);
                }
        }

        /* Configure SPI base adresses */
        spi_add[SPI1].base_address = CSPI1_BASE_ADDR;
        spi_add[SPI2].base_address = CSPI2_BASE_ADDR;
#if defined(CONFIG_ARCH_MX3)
        spi_add[SPI3].base_address = CSPI3_BASE_ADDR;
#endif

        /* Configure SPBA */
        spi_init_spba();

        /* Set interrupt number */
        spi_port[SPI1].int_cspi = INT_CSPI1;
        spi_port[SPI2].int_cspi = INT_CSPI2;
#if defined(CONFIG_ARCH_MX3)
        spi_port[SPI3].int_cspi = INT_CSPI3;
#endif

        for (spi = 0; spi < CONFIG_SPI_NB_MAX; spi++){
                if (spi_present[spi]) {
                        error = spi_init_default_conf(spi);
                        if (error != 0 ){
                                return error;
                        }
                }
        }
        return 0;
}

/*!
 * This function configures the hardware SPI for the currrent client.
 * Lock must be taken
 *
 * @param        mod              the module number
 * @param        client_config    client hardware configuration.
 * @return       This function returns 0 if successful, -EPERM otherwise.
 */
static int spi_hard_config(module_nb_t mod, spi_config *client_config) 
{
        int error=0;
        error = spi_set_baudrate(mod, client_config->bit_rate);
        if (error < 0){
                return error;
        }
        spi_set_transfer_length(mod, client_config->bit_count);
        error = spi_select_ss(mod, client_config->ss_asserted);
        if (error < 0){
                return error;
        }

        if (client_config->master_mode == true) {
                spi_set_mode(mod, SPI_MASTER);
        } else {
                spi_set_mode(mod, SPI_SLAVE);
        }

        if (client_config->active_high_ss_polarity == true) {
                spi_set_ss_polarity(mod, SPI_SS_ACTIVE_HIGH);
        } else {
                spi_set_ss_polarity(mod, SPI_SS_ACTIVE_LOW);
        }

        if (client_config->ss_low_between_bursts == true) {
                spi_set_ss_waveform(mod, SPI_LOW_BTN_BURST);
        } else {
                spi_set_ss_waveform(mod, SPI_PULSE_BTN_BURST);
        }

        if (client_config->phase == true) {
                spi_set_phase(mod, SPI_PHASE_1);
        } else {
                spi_set_phase(mod, SPI_PHASE_0); 
        }

        if (client_config->active_high_polarity == true) {
                spi_set_polarity(mod, SPI_POLARITY_ACTIVE_HIGH);
        } else {
                spi_set_polarity(mod, SPI_POLARITY_ACTIVE_LOW);
        }
        return 0;
}

/*!
 * This function releases the SPI resource.
 * Lock must be taken
 *
 * @param        module        the module number
 * @return       This function returns 0 if successful, -EPERM otherwise.
 */
static int spi_release(module_nb_t module)
{
        int result = 0, prio = 0;
        bool spi_free;
        spi_mxc_port *port;      

        port = &(spi_port[module]);

        DBG_PRINTK("spi%d : spi_release \n", module + 1);
        

        down(&(port->access_mutex));

        spi_free = true;
        
        if (port->in_use == false) {
                result = -EPERM;
        } else {
                /* If there are any client waiting for the resource */
                for ( prio = HIGH ; prio >= LOW ; prio--){
                        if ( (port->pending_clients[prio] > 0) && spi_free){
                                spi_free = false;
                                DBG_PRINTK("spi%d: nb of pending clients \
                                        : %d Prio %d", module + 1, 
                                                port->pending_clients[prio],
                                                        prio);
                                /* One client less waiting */
                                port->pending_clients[prio]--;
                                up(&(port->pending_mutex[prio]
                                        [(port->pending_clients[prio])]));
                                break;
                        }
                }
                if (spi_free) {
                        /* There's no client waiting for the resource */
                        port->in_use = false;
                }
                result = 0;
        }
        /* Release access to port struct */
        up(&(port->access_mutex));

        return result;
}

/*!
 * This function requests the SPI resource.
 *
 * @param    module         the module number
 * @param    prio           client's priority
 * @return   This function returns 0 if successful, -EBUSY otherwise.
 */
static int spi_request(module_nb_t module, client_prio_t prio)
{
        int result = 0;
        spi_mxc_port *port;      
 	struct semaphore *sema;
        
        port = &(spi_port[module]);
        DBG_PRINTK( "spi%d : spi_request \n", module + 1);
        down(&(port->access_mutex));
        
        if( port->in_use == false) {
                DBG_PRINTK( "spi%d : spi not in use \n", module + 1);
                port->in_use = true;
        } else if ( port->pending_clients[prio] < MAX_SPI_CLIENTS) {
                DBG_PRINTK( "spi%d : spi in use \n", module + 1);
                DBG_PRINTK("\nDown request prio %d, pending %d",prio,
                        port->pending_clients[prio]);
                sema = &(port->pending_mutex[prio]
                        [port->pending_clients[prio]]);
                port->pending_clients[prio]++;
                /* release port mutex */
                up(&(port->access_mutex));
                /* down the semaphore */
                down(sema);
                return result;
        } else {
                result = -EBUSY;
                DBG_PRINTK("SPI BUSY\n");
        }

        /* release port mutex */
        up(&(port->access_mutex));
        return result;
}

/*!
 * This function sends a SPI frame.
 *
 * @param        buffer                the buffer to send, this buffer must be allocated with 'nb' bytes size by user 
 * @param        nb                    the number of bytes to send, can not be more than 32 bytes.
 * @param        client_handle         the handle to identify the SPI device
 * @return       This function returns the number of bytes sent minus the number of received bytes or
 *               -EBUSY if the maximun clients pending number has been reached.
 */
ssize_t spi_send_frame(unsigned char* buffer, unsigned int nb,
                       void* client_handle)
{
        int length,error;
        int result;
        int nb_bits;
        unsigned int value_to_load;
        module_nb_t mod;
        spi_config *client_config;

        if (buffer == NULL) {
                return -1;
        }
        if (client_handle == NULL) {
                return -1;
        }
        client_config = (spi_config *)client_handle;
        /*
         * length can not safely exceed fifo length with that implementation
         * if nb > fifo length -> error
         */
        nb_bits = (client_config->bit_count * 8);
        if ((nb * 8) > nb_bits) {
                return -1;
        }
        
        mod = client_config->module_number;
        spi_port[mod].hw_config = client_config;
        result = spi_request(client_config->module_number, 
                client_config->priority);
        if (result != 0) {
                return result;
        }

        length =  nb;
        spi_port[mod].bytes_to_send = length;
        spi_port[mod].bytes_received = 0;
        value_to_load = 0;

        DBG_PRINTK("spi%d : spi_send_frame : 0x%8X \n", mod + 1,
                   ((*buffer) << 24) + ((*(buffer + 1)) << 16) +
                   ((*(buffer + 2)) << 8) + *(buffer + 3));

        DBG_PRINTK("spi%d : client handle : %d \n", mod + 1,
                   (int)client_handle);

        if ((spi_port[mod].current_config_handle != client_config) ||
            (spi_port[mod].current_config_handle == NULL)) {
                error = spi_hard_config(mod, (void*)client_config);
                if (error < 0) {
                        return error;
                }
                spi_port[mod].current_config_handle = (void*) client_config;
        }

        spi_port[mod].rx_buffer = buffer;

        spi_put_tx_data(mod, buffer, length);

        DBG_PRINTK("spi : down_int sync_interface\n");
        down(&(spi_port[mod].sync_interface));
        DBG_PRINTK("spi : down_int sync_interface done\n");

        result = spi_port[mod].bytes_to_send - 
                 spi_port[mod].bytes_received;

        spi_release(client_config->module_number);

        return result;
}

EXPORT_SYMBOL(spi_get_device_id);
EXPORT_SYMBOL(spi_send_frame);

