/*
 * Copyright 2004 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*!
 * @defgroup Framebuffer Framebuffer Driver for SDC and ADC.
 */

/*
 * @file mxcfb.c
 *
 * @brief MXC Frame buffer driver for SDC
 *
 * @ingroup Framebuffer
 */
 
/*!
 * Include files
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/fb.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/device.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/mach-types.h>
#include <asm/uaccess.h>
#include <asm/arch/platform.h>

#include "mxcfb.h"

/*
 * Debug Macros
 */
//#define MXCFB_DEBUG

#ifdef MXCFB_DEBUG

#  define DDPRINTK(fmt, args...) printk(KERN_ERR"%s :: %d :: %s - " fmt, __FILE__,__LINE__,__FUNCTION__ , ## args)
#  define DPRINTK(fmt, args...) printk("%s: " fmt, __FUNCTION__ , ## args)

#  define FUNC_START    DDPRINTK(" func start\n")
#  define FUNC_END      DDPRINTK(" func end\n")

#  define FUNC_ERR printk(KERN_ERR"%s :: %d :: %s  err= %d \n", __FILE__,__LINE__,__FUNCTION__ ,err)

#else //MXCFB_DEBUG

#define DDPRINTK(fmt, args...)  do {} while(0)
#define DPRINTK(fmt, args...)   do {} while(0)

#define FUNC_START
#define FUNC_END

#endif //MXCFB_DEBUG

struct mxcfb_data {
        struct fb_info *        fbi;
        struct fb_info *        fbi_ovl;
        volatile int32_t        vsync_flag;
        wait_queue_head_t       vsync_wq;
};

struct mxcfb_data mxcfb_drv_data;


struct panel_info sharp_qvga_panel = {
        .name                   = "Sharp QVGA Panel",
        .type                   = IPU_PANEL_SHARP_TFT,
        .pixel_fmt              = IPU_PIX_FMT_RGB666,
        .width                  = 240,
        .height                 = 320,
        .vSyncWidth             = 1,
        .vStartWidth            = 8,
        .vEndWidth              = 80,
        .hSyncWidth             = 1,
        .hStartWidth            = 3,
        .hEndWidth              = 10,
        .sig_pol.datamask_en    = false,
        .sig_pol.clkidle_en     = true,
        .sig_pol.clksel_en      = false,
        .sig_pol.Vsync_pol      = false,
        .sig_pol.enable_pol     = false,
        .sig_pol.data_pol       = true,
        .sig_pol.clk_pol        = true,
        .sig_pol.Hsync_pol      = true,
};

struct panel_info nec_vga_panel = {
        .name                   = "NEC VGA Panel",
        .type                   = IPU_PANEL_TFT,
        .pixel_fmt              = IPU_PIX_FMT_RGB666,
        .width                  = 640,
        .height                 = 480,
        .vSyncWidth             = 1,
        .vStartWidth            = 0x22,
        .vEndWidth              = 40,
        .hSyncWidth             = 1,
        .hStartWidth            = 0x90, //0x69, for Sharp
        .hEndWidth              = 0,
        .sig_pol.datamask_en    = false,
        .sig_pol.clkidle_en     = false,
        .sig_pol.clksel_en      = false,
        .sig_pol.Vsync_pol      = false,
        .sig_pol.enable_pol     = true,
        .sig_pol.data_pol       = false, //true, for Sharp
        .sig_pol.clk_pol        = false,
        .sig_pol.Hsync_pol      = false,
};

#ifdef CONFIG_FB_MXC_VGA_PANEL
struct panel_info * mxcfb_panel = &nec_vga_panel;
#else
struct panel_info * mxcfb_panel = &sharp_qvga_panel;
#endif


/*!
 * This structure contains the information such as the screen width, 
 * screen height, virtual size, and bits per pixel. This is passed into 
 * the \b fbinfo structure.
 */
static struct fb_var_screeninfo mxcfb_def_var = {
        .bits_per_pixel         = MXCFB_BPP,
        .activate               = FB_ACTIVATE_NOW,
        .height                 = -1,
        .width                  = -1,
        .vmode                  = FB_VMODE_NONINTERLACED,
};


static uint32_t bpp_to_pixfmt(int bpp)
{
        uint32_t pixfmt = 0;
        switch (bpp) {
        case 24:
                pixfmt = IPU_PIX_FMT_BGR24;
                break;
        case 32:
                pixfmt = IPU_PIX_FMT_BGR32;
                break;
        case 16:
                pixfmt = IPU_PIX_FMT_RGB565;
                break;
        }
        return pixfmt;
}

extern void gpio_lcd_active(void);
extern void gpio_lcd_inactive(void);
static irqreturn_t mxcfb_irq_handler(int irq, void *dev_id, struct pt_regs *regs);
static int mxcfb_blank(int blank, struct fb_info *info);
#ifdef CONFIG_FB_MXC_OVERLAY
static irqreturn_t mxcfb_ovl_irq_handler(int irq, void *dev_id, struct pt_regs *regs);
#endif
/*
 * Open the main framebuffer.
 * 
 * @param       fbi     framebuffer information pointer
 * 
 * @param       user    Set if opened by user or clear if opened by kernel
 */
static int 
mxcfb_open(struct fb_info *fbi, int user)
{
        struct mxcfb_info * mxc_fbi = (struct mxcfb_info *)fbi->par;

        FUNC_START;

        if (mxc_fbi->open_count == 0)
        {
                if (ipu_request_irq(IPU_IRQ_SDC_BG_EOF, mxcfb_irq_handler, 0, 
                                    MXCFB_NAME, fbi) != 0)
                {
                        printk("mxcfb: Error registering irq handler.\n");
                        return -EBUSY;
                }
                ipu_disable_irq(IPU_IRQ_SDC_BG_EOF);
                mxc_fbi->ipu_ch_irq = IPU_IRQ_SDC_BG_EOF;

                if (ipu_sdc_init_panel(mxcfb_panel->type, 60, 
                                       mxcfb_panel->width, mxcfb_panel->height,
                                       mxcfb_panel->pixel_fmt,
                                       mxcfb_panel->hStartWidth,
                                       mxcfb_panel->hSyncWidth,
                                       mxcfb_panel->hEndWidth,
                                       mxcfb_panel->vStartWidth,
                                       mxcfb_panel->vSyncWidth,
                                       mxcfb_panel->vEndWidth,
                                       mxcfb_panel->sig_pol) != 0)
                {
                        printk("mxcfb: Error initializing panel.\n");
                        return -EINVAL;
                }
                ipu_sdc_set_window_pos(MEM_SDC_BG, 0, 0);

                ipu_sdc_set_brightness(255);
                ipu_sdc_set_global_alpha(false, 0);
                ipu_sdc_set_color_key(MEM_SDC_BG, false, 0);

                ipu_init_channel(MEM_SDC_BG, NULL);
                ipu_init_channel_buffer(MEM_SDC_BG, IPU_INPUT_BUFFER, bpp_to_pixfmt(fbi->var.bits_per_pixel),
                                        fbi->var.xres, fbi->var.yres,
                                        fbi->var.xres_virtual,
                                        IPU_ROTATE_NONE,
                                        (void*)fbi->fix.smem_start, 
                                        (void*)fbi->fix.smem_start);
                down(&mxc_fbi->flip_sem);        
                mxc_fbi->cur_ipu_buf = 0;

                ipu_select_buffer(MEM_SDC_BG, IPU_INPUT_BUFFER, 0);

                mxc_fbi->ipu_ch = MEM_SDC_BG;
                ipu_enable_irq(mxc_fbi->ipu_ch_irq);
        }
        mxc_fbi->open_count++;

        mxcfb_blank(VESA_NO_BLANKING, fbi);
        return 0;
}

/*
 * Close the main framebuffer.
 * 
 * @param       fbi     framebuffer information pointer
 * 
 * @param       user    Set if opened by user or clear if opened by kernel
 */
static int 
mxcfb_release(struct fb_info *fbi, int user)
{
        struct mxcfb_info * mxc_fbi = (struct mxcfb_info *)fbi->par;
        FUNC_START;

        --mxc_fbi->open_count;
        if (mxc_fbi->open_count == 0)
        {
                mxcfb_blank(VESA_POWERDOWN, fbi);
                
                //ipu_disable_channel(MEM_SDC_BG, true);
                ipu_uninit_channel(MEM_SDC_BG);
                ipu_disable_irq(IPU_IRQ_SDC_BG_EOF);
                ipu_free_irq(IPU_IRQ_SDC_BG_EOF, fbi);
        }
        return 0;
}

/*
 * Set fixed framebuffer parameters based on variable settings.
 * 
 * @param       info     framebuffer information pointer
 */
static int 
mxcfb_set_fix(struct fb_info *info)
{
        struct fb_fix_screeninfo *fix = &info->fix;
        struct fb_var_screeninfo *var = &info->var;
        FUNC_START;

        strncpy(fix->id, MXCFB_NAME, sizeof(fix->id));

        fix->line_length = var->xres_virtual * var->bits_per_pixel / 8;

        fix->type = FB_TYPE_PACKED_PIXELS;
        fix->accel = FB_ACCEL_NONE;
        fix->visual = FB_VISUAL_TRUECOLOR;
        fix->xpanstep = 1;
        fix->ypanstep = 1;

        return 0;
}

/*
 * Set framebuffer parameters and change the operating mode.
 * 
 * @param       info     framebuffer information pointer
 */
static int
mxcfb_set_par(struct fb_info *info)
{
        struct mxcfb_info * mxc_fbi = (struct mxcfb_info *)info->par;

        FUNC_START;

        ipu_disable_irq(IPU_IRQ_SDC_BG_EOF);
        ipu_disable_channel(mxc_fbi->ipu_ch, true);
        mxcfb_set_fix(info);

        mxc_fbi->cur_ipu_buf = 0;
        sema_init(&mxc_fbi->flip_sem, 1);

        ipu_sdc_set_window_pos(mxc_fbi->ipu_ch, 0, 0);
        ipu_init_channel_buffer(mxc_fbi->ipu_ch, IPU_INPUT_BUFFER, 
                                bpp_to_pixfmt(info->var.bits_per_pixel),
                                info->var.xres, info->var.yres,
                                info->var.xres_virtual,
                                IPU_ROTATE_NONE,
                                (void*)info->fix.smem_start, 
                                (void*)info->fix.smem_start);
        
        ipu_select_buffer(mxc_fbi->ipu_ch, IPU_INPUT_BUFFER, 0);

        ipu_enable_channel(mxc_fbi->ipu_ch);

        return 0;
}

/*
 * Check framebuffer variable parameters and adjust to valid values.
 * 
 * @param       var      framebuffer variable parameters
 * 
 * @param       info     framebuffer information pointer
 */
static int 
mxcfb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
        FUNC_START;

        if (var->xres > mxcfb_panel->width)
                var->xres = mxcfb_panel->width;
        if (var->yres > mxcfb_panel->height)
                var->yres = mxcfb_panel->height;
        if (var->xres_virtual < var->xres)
                var->xres_virtual = var->xres;
        if (var->yres_virtual < var->yres)
                var->yres_virtual = var->yres;

        if ((var->bits_per_pixel != 32) && (var->bits_per_pixel != 24) &&
            (var->bits_per_pixel != 16)) {
                var->bits_per_pixel = mxcfb_def_var.bits_per_pixel;
        }

        switch (var->bits_per_pixel) {
        case 16:
                var->red.length = 5;
                var->red.offset = 11;
                var->red.msb_right = 0;

                var->green.length = 6;
                var->green.offset = 5;
                var->green.msb_right = 0;

                var->blue.length = 5;
                var->blue.offset = 0;
                var->blue.msb_right = 0;

                var->transp.length = 0;
                var->transp.offset = 0;
                var->transp.msb_right = 0;
                break;
        case 24:
                var->red.length = 8;
                var->red.offset = 16;
                var->red.msb_right = 0;

                var->green.length = 8;
                var->green.offset = 8;
                var->green.msb_right = 0;

                var->blue.length = 8;
                var->blue.offset = 0;
                var->blue.msb_right = 0;

                var->transp.length = 0;
                var->transp.offset = 0;
                var->transp.msb_right = 0;
                break;
        case 32:
                var->red.length = 8;
                var->red.offset = 16;
                var->red.msb_right = 0;

                var->green.length = 8;
                var->green.offset = 8;
                var->green.msb_right = 0;

                var->blue.length = 8;
                var->blue.offset = 0;
                var->blue.msb_right = 0;

                var->transp.length = 8;
                var->transp.offset = 24;
                var->transp.msb_right = 0;
                break;
        }

        var->height             = -1;
        var->width              = -1;
        var->grayscale          = 0;
        var->nonstd             = 0;

        var->pixclock           = -1;
        var->left_margin        = -1;
        var->right_margin       = -1;
        var->upper_margin       = -1;
        var->lower_margin       = -1;
        var->hsync_len          = -1;
        var->vsync_len          = -1;

        var->vmode              = FB_VMODE_NONINTERLACED;
        var->sync               = 0;

        return 0;
}

/*
 * Function to handle custom ioctls for MXC framebuffer.
 * 
 * @param       inode   inode struct
 * 
 * @param       file    file struct
 * 
 * @param       cmd     Ioctl command to handle
 * 
 * @param       arg     User pointer to command arguments
 * 
 * @param       fbi     framebuffer information pointer
 */
static int mxcfb_ioctl(struct inode *inode, struct file *file,
                        unsigned int cmd, unsigned long arg,
                        struct fb_info *fbi)
{
        int retval = 0;
        
        switch (cmd)
        {
        case MXCFB_SET_GBL_ALPHA:
        {
                struct mxcfb_gbl_alpha ga;
                if (copy_from_user(&ga, (void *)arg, sizeof(ga))) {
                        retval = -EFAULT;
                        break;
                }
                retval = ipu_sdc_set_global_alpha((bool)ga.enable, ga.alpha);
                DPRINTK("Set global alpha to %d\n", ga.alpha);
                break;
        }
        case MXCFB_SET_CLR_KEY:
        {
                struct mxcfb_color_key key;
                if (copy_from_user(&key, (void *)arg, sizeof(key))) {
                        retval = -EFAULT;
                        break;
                }
                retval = ipu_sdc_set_color_key(MEM_SDC_BG, key.enable, key.color_key);
                DPRINTK("Set color key to 0x%08X\n", key.color_key);
                break;
        }
        case MXCFB_WAIT_FOR_VSYNC:
        {
                mxcfb_drv_data.vsync_flag = 0;
                ipu_enable_irq(IPU_IRQ_SDC_DISP3_VSYNC);
                if (!wait_event_interruptible_timeout(mxcfb_drv_data.vsync_wq, mxcfb_drv_data.vsync_flag != 0, 1 * HZ))
                {
                        printk("MXCFB_WAIT_FOR_VSYNC: timeout\n");
                        retval = -ETIME;
                        break;
                }
                else if (signal_pending(current))
                {
                        printk("MXCFB_WAIT_FOR_VSYNC: interrupt received\n");
                        retval = -ERESTARTSYS;
                        break;
                }
                break;
        }
        case MXCFB_SET_BRIGHTNESS:
        {
                uint8_t level;
                if (copy_from_user(&level, (void *)arg, sizeof(level))) {
                        retval = -EFAULT;
                        break;
                }
                retval = ipu_sdc_set_brightness(level);
                DPRINTK("Set brightness to %d\n", level);
                break;
        }
        default:
                retval = -EINVAL;
        }
        return retval;
}

#ifdef CONFIG_FB_MXC_OVERLAY
/*
 * Open the overlay framebuffer.
 * 
 * @param       fbi     framebuffer information pointer
 * 
 * @param       user    Set if opened by user or clear if opened by kernel
 */
static int 
mxcfb_ovl_open(struct fb_info *fbi, int user)
{
        struct mxcfb_info * mxc_fbi = (struct mxcfb_info *)fbi->par;

        FUNC_START;

        if (mxc_fbi->open_count == 0)
        {
                if (ipu_request_irq(IPU_IRQ_SDC_FG_EOF, mxcfb_ovl_irq_handler, 0, 
                                    MXCFB_NAME, fbi) != 0)
                {
                        printk("mxcfb: Error registering irq handler.\n");
                        return -EBUSY;
                }
                ipu_disable_irq(IPU_IRQ_SDC_FG_EOF);
                mxc_fbi->ipu_ch_irq = IPU_IRQ_SDC_FG_EOF;

                ipu_sdc_set_window_pos(MEM_SDC_FG, fbi->var.xoffset, fbi->var.yoffset);

                ipu_init_channel(MEM_SDC_FG, NULL);
                ipu_init_channel_buffer(MEM_SDC_FG, IPU_INPUT_BUFFER, bpp_to_pixfmt(fbi->var.bits_per_pixel),
                                        fbi->var.xres, fbi->var.yres,
                                        fbi->var.xres_virtual,
                                        IPU_ROTATE_NONE,
                                        (void*)fbi->fix.smem_start, 
                                        (void*)fbi->fix.smem_start);
                down(&mxc_fbi->flip_sem);        
                mxc_fbi->cur_ipu_buf = 0;

                ipu_select_buffer(MEM_SDC_FG, IPU_INPUT_BUFFER, 0);

                mxc_fbi->ipu_ch = MEM_SDC_FG;

                ipu_enable_irq(mxc_fbi->ipu_ch_irq);
                ipu_enable_channel(MEM_SDC_FG);
        }
        mxc_fbi->open_count++;
        return 0;
}

/*
 * Close the overlay framebuffer.
 * 
 * @param       fbi     framebuffer information pointer
 * 
 * @param       user    Set if opened by user or clear if opened by kernel
 */
static int 
mxcfb_ovl_release(struct fb_info *fbi, int user)
{
        struct mxcfb_info * mxc_fbi = (struct mxcfb_info *)fbi->par;

        FUNC_START;

        --mxc_fbi->open_count;
        if (mxc_fbi->open_count == 0)
        {
                ipu_disable_channel(MEM_SDC_FG, true);
                ipu_uninit_channel(MEM_SDC_FG);
                ipu_disable_irq(IPU_IRQ_SDC_FG_EOF);
                ipu_free_irq(IPU_IRQ_SDC_FG_EOF, fbi);
        }
        return 0;
}

static irqreturn_t mxcfb_ovl_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
        struct fb_info * fbi = dev_id;
        struct mxcfb_info * mxc_fbi = fbi->par;

        up(&mxc_fbi->flip_sem);

        if (atomic_read(&(mxc_fbi->flip_sem.count)) == 2) {
                ipu_disable_irq(irq);
        }
        return IRQ_HANDLED;
}
#endif


/*
 * Formal definition of the VESA spec:
 *      On
 *              This refers to the state of the display when it is in full operation
 *      Stand-By
 *              This defines an optional operating state of minimal power reduction with
 *              the shortest recovery time
 *      Suspend
 *              This refers to a level of power management in which substantial power
 *              reduction is achieved by the display.  The display can have a longer 
 *              recovery time from this state than from the Stand-by state
 *      Off
 *              This indicates that the display is consuming the lowest level of power
 *              and is non-operational. Recovery from this state may optionally require
 *              the user to manually power on the monitor
 *
 *      Now, the fbdev driver adds an additional state, (blank), where they
 *      turn off the video (maybe by colormap tricks), but don't mess with the
 *      video itself: think of it semantically between on and Stand-By.
 *
 *      So here's what we should do in our fbdev blank routine:
 *
 *              VESA_NO_BLANKING (mode 0)       Video on,  front/back light on
 *              VESA_VSYNC_SUSPEND (mode 1)     Video on,  front/back light off
 *              VESA_HSYNC_SUSPEND (mode 2)     Video on,  front/back light off
 *              VESA_POWERDOWN (mode 3)         Video off, front/back light off
 *
 *      This will match the matrox implementation.
 */
/*
 * mxcfb_blank():
 *      Blank the display by setting all palette values to zero.  
 */
static int mxcfb_blank(int blank, struct fb_info *info)
{
        FUNC_START;

        switch (blank) {
                case VESA_POWERDOWN:
                case VESA_VSYNC_SUSPEND:
                case VESA_HSYNC_SUSPEND:
                        ipu_disable_channel(MEM_SDC_BG, true);
                        ipu_sdc_set_brightness(0);
                        gpio_lcd_inactive();
                        break;

                case VESA_NO_BLANKING:
                        ipu_enable_channel(MEM_SDC_BG);
                        ipu_sdc_set_brightness(255);
                        gpio_lcd_active();
                        break;
        }
        return 0;
}

/*
 * Pan or Wrap the Display
 *
 * This call looks only at xoffset, yoffset and the FB_VMODE_YWRAP flag
 *
 * @param               var     Variable screen buffer information
 * @param               info    Framebuffer information pointer
 */
static int 
mxcfb_pan_display(struct fb_var_screeninfo *var, struct fb_info *info)
{
        struct mxcfb_info *mxc_fbi = (struct mxcfb_info *)info->par;
        uint32_t lock_flags = 0;
        u_int y_bottom;
        unsigned long base;

        FUNC_START;

        if (var->xoffset > 0) {
                DPRINTK("x panning not supported\n");
                return -EINVAL;
        }

        if ((info->var.xoffset == var->xoffset) &&
            (info->var.yoffset == var->yoffset)) {
                return 0; // No change, do nothing
        }

        y_bottom = var->yoffset;

        if (!(var->vmode & FB_VMODE_YWRAP)) {
                y_bottom += var->yres;
        }

        if (y_bottom > info->var.yres_virtual) {
                return -EINVAL;
        }
        
        base = (var->yoffset * var->xres_virtual + var->xoffset);
        base *= (var->bits_per_pixel) / 8;
        base += info->fix.smem_start;

        
        down(&mxc_fbi->flip_sem);        

        spin_lock_irqsave(&mxc_fbi->fb_lock, lock_flags);

//        printk("Updating SDC BG buf %d address=0x%08X\n", fbi->cur_ipu_buf, base);

        mxc_fbi->cur_ipu_buf = !mxc_fbi->cur_ipu_buf;
        if (ipu_update_channel_buffer(mxc_fbi->ipu_ch, IPU_INPUT_BUFFER, 
                mxc_fbi->cur_ipu_buf, (void*)base) < 0) {
                printk("Error updating SDC buf %d to address=0x%08X\n", 
                        mxc_fbi->cur_ipu_buf, (uint32_t)base);
        }
        ipu_select_buffer(mxc_fbi->ipu_ch, IPU_INPUT_BUFFER, mxc_fbi->cur_ipu_buf);

        ipu_enable_irq(mxc_fbi->ipu_ch_irq);

        spin_unlock_irqrestore(&mxc_fbi->fb_lock, lock_flags);

//        printk("Update complete\n");
        
        info->var.xoffset = var->xoffset;
        info->var.yoffset = var->yoffset;

        if (var->vmode & FB_VMODE_YWRAP) {
                info->var.vmode |= FB_VMODE_YWRAP;
        } else {
                info->var.vmode &= ~FB_VMODE_YWRAP;
        }

        return 0;
}


/*!
 * This structure contains the pointers to the control functions that are 
 * invoked by the core framebuffer driver to perform operations like 
 * blitting, rectangle filling, copy regions and cursor definition.
 */
static struct fb_ops mxcfb_ops = {
        .owner                          = THIS_MODULE,
        .fb_open                        = mxcfb_open,
        .fb_release                     = mxcfb_release,
        .fb_set_par                     = mxcfb_set_par,
        .fb_check_var                   = mxcfb_check_var,
        .fb_pan_display                 = mxcfb_pan_display,
        .fb_ioctl                       = mxcfb_ioctl,
        .fb_fillrect                    = cfb_fillrect,
        .fb_copyarea                    = cfb_copyarea,
        .fb_imageblit                   = cfb_imageblit,
//        .fb_blank                       = mxcfb_blank,
        .fb_cursor                      = soft_cursor,
};

#ifdef CONFIG_FB_MXC_OVERLAY
static struct fb_ops mxcfb_ovl_ops = {
        .owner                          = THIS_MODULE,
        .fb_open                        = mxcfb_ovl_open,
        .fb_release                     = mxcfb_ovl_release,
        .fb_set_par                     = mxcfb_set_par,
        .fb_check_var                   = mxcfb_check_var,
        .fb_pan_display                 = mxcfb_pan_display,
        .fb_fillrect                    = cfb_fillrect,
        .fb_copyarea                    = cfb_copyarea,
        .fb_imageblit                   = cfb_imageblit,
        .fb_cursor                      = soft_cursor,
};
#endif

static irqreturn_t mxcfb_vsync_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
        struct mxcfb_data * fb_data = dev_id;

        ipu_disable_irq(irq);

        fb_data->vsync_flag = 1;
        wake_up_interruptible(&fb_data->vsync_wq);
        return IRQ_HANDLED;
}


static irqreturn_t mxcfb_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
        struct fb_info * fbi = dev_id;
        struct mxcfb_info * mxc_fbi = fbi->par;

        up(&mxc_fbi->flip_sem);

        if (atomic_read(&(mxc_fbi->flip_sem.count)) == 2) {
                ipu_disable_irq(irq);
        }
        return IRQ_HANDLED;
}


#ifdef CONFIG_PM
/*
 * Power management hooks.      Note that we won't be called from IRQ context,
 * unlike the blank functions above, so we may sleep.
 */

/*
 * Suspends the framebuffer and blanks the screen. Power management support 
 */
static int 
mxcfb_suspend(struct device *dev, u32 state, u32 level)
{
        DPRINTK("Suspending LCD\n");
        
        if (level == SUSPEND_DISABLE || level == SUSPEND_POWER_DOWN) {
                ipu_disable_channel(MEM_SDC_BG, true);
                ipu_sdc_set_brightness(0);
                gpio_lcd_inactive();
        }
        return 0;
}

/*
 * Resumes the framebuffer and unblanks the screen. Power management support 
 */
static int 
mxcfb_resume(struct device *dev, u32 level)
{
        DPRINTK("Resuming LCD\n");

        if (level == RESUME_ENABLE) {
                ipu_enable_channel(MEM_SDC_BG);
                ipu_sdc_set_brightness(255);
                gpio_lcd_active();
        }
        return 0;
}
#else
#define mxcfb_suspend   NULL
#define mxcfb_resume    NULL
#endif

/*
 * Main framebuffer functions
 */

/*!
 * Allocates the DRAM memory for the frame buffer.      This buffer is remapped 
 * into a non-cached, non-buffered, memory region to allow palette and pixel 
 * writes to occur without flushing the cache.  Once this area is remapped, 
 * all virtual memory access to the video memory should occur at the new region.
 *
 * @param       fbi     framebuffer information pointer
 * 
 * @return      Error code indicating success or failure
 */
static int
mxcfb_map_video_memory(struct fb_info *fbi)
{
        fbi->fix.smem_len = fbi->var.xres_virtual * fbi->var.yres_virtual * 4;
        
        fbi->fix.smem_start = IPU_Malloc(fbi->fix.smem_len);
        if (fbi->fix.smem_start == 0) {
                printk("MXCFB - Unable to allocate framebuffer memory\n");
                return -EBUSY;
        }
        DPRINTK("mxcfb: allocated fb @ paddr=0x%08X, size=%d.\n", 
               (uint32_t)fbi->fix.smem_start, fbi->fix.smem_len);
        
        if (!request_mem_region(fbi->fix.smem_start, fbi->fix.smem_len, "LCD")) {
                return -EBUSY;
        }

        if (!(fbi->screen_base = ioremap(fbi->fix.smem_start, fbi->fix.smem_len))) {
                release_mem_region(fbi->fix.smem_start, fbi->fix.smem_len);
                printk("MXCFB - Unable to map framebuffer memory to virtual address\n");
                return -EIO;
        }
        fbi->screen_size = fbi->fix.smem_len;


        /* Clear the screen */
        memset((char*)fbi->screen_base, 0, fbi->fix.smem_len);

        return 0;
}

/*!
 * De-allocates the DRAM memory for the frame buffer.
 *
 * @param       fbi     framebuffer information pointer
 * 
 * @return      Error code indicating success or failure
 */
static int
mxcfb_unmap_video_memory(struct fb_info *fbi)
{
        iounmap(fbi->screen_base);
        release_mem_region(fbi->fix.smem_start, fbi->fix.smem_len);
        IPU_Free(fbi->fix.smem_start);
        return 0;
}

/*!
 * Initializes the framebuffer information pointer. After allocating
 * sufficient memory for the framebuffer structure, the fields are
 * filled with custom information passed in from the configurable
 * structures.  This includes information such as bits per pixel,
 * color maps, screen width/height and RGBA offsets.
 *
 * @return      Framebuffer structure initialized with our information
 */
static struct fb_info * 
mxcfb_init_fbinfo(struct device *dev, struct fb_ops * ops)
{
        struct fb_info *fbi;
        struct mxcfb_info *mxcfbi;

        /*
         * Allocate sufficient memory for the fb structure
         */
        fbi = framebuffer_alloc(sizeof(struct mxcfb_info), dev);
        if (!fbi)
                return NULL;

        mxcfbi = (struct mxcfb_info *)fbi->par;

         /*
         * Fill in fb_info structure information
         */
        fbi->var.xres = fbi->var.xres_virtual = mxcfb_panel->width;
        fbi->var.yres = mxcfb_panel->height;
        fbi->var.yres_virtual = mxcfb_panel->height * 2;
        fbi->var.activate = FB_ACTIVATE_NOW;
        mxcfb_check_var(&fbi->var, fbi);

        mxcfb_set_fix(fbi);

        fbi->fbops                   = ops;
        fbi->flags                   = FBINFO_FLAG_DEFAULT;
        fbi->pseudo_palette          = mxcfbi->pseudo_palette;

        sema_init(&mxcfbi->flip_sem, 2);
        spin_lock_init(&mxcfbi->fb_lock);

        /*
         * Allocate colormap
         */
        fb_alloc_cmap(&fbi->cmap, 1<<fbi->var.bits_per_pixel, 0);

        return fbi;
}

/*!
 * Probe routine for the framebuffer driver. It is called during the
 * driver binding process.      The following functions are performed in
 * this routine: Framebuffer initialization, Memory allocation and
 * mapping, Framebuffer registration, IPU initialization.
 *
 * @return      Appropriate error code to the kernel common code
 */
static int mxcfb_probe(struct device *dev)
{
        struct fb_info *fbi;
#ifdef CONFIG_FB_MXC_OVERLAY
        struct fb_info *fbi_ovl;
#endif
        int ret = 0;

        FUNC_START;

        /*
         * Initialize FB structures
         */
        fbi = mxcfb_init_fbinfo(dev, &mxcfb_ops);
        if (!fbi) {
                ret = -ENOMEM;
                goto err0;
        }
        mxcfb_drv_data.fbi = fbi;

        /*
         * Allocate memory
         */
        ret = mxcfb_map_video_memory(fbi);
        if (ret < 0) {
                goto err1;
        }

        /*
         * Register framebuffer
         */
        ret = register_framebuffer(fbi);
        if (ret < 0) {
                goto err2;
        }
        

#ifdef CONFIG_FB_MXC_OVERLAY
        /*
         * Initialize Overlay FB structures
         */
        fbi_ovl = mxcfb_init_fbinfo(dev, &mxcfb_ovl_ops);
        if (!fbi_ovl) {
                ret = -ENOMEM;
                goto err3;
        }
        mxcfb_drv_data.fbi_ovl = fbi_ovl;

        ret = mxcfb_map_video_memory(fbi_ovl);
        if (ret < 0) {
                goto err4;
        }

        /*
         * Register overlay framebuffer
         */
        ret = register_framebuffer(fbi_ovl);
        if (ret < 0) {
                goto err5;
        }
#else
        mxcfb_drv_data.fbi_ovl = NULL;
#endif        
        dev_set_drvdata(dev, &mxcfb_drv_data);

        init_waitqueue_head(&mxcfb_drv_data.vsync_wq);
        if (ipu_request_irq(IPU_IRQ_SDC_DISP3_VSYNC, mxcfb_vsync_irq_handler,
                        0, MXCFB_NAME, &mxcfb_drv_data) < 0) {
                goto err6;
        }

        gpio_lcd_active();

        FUNC_END;
        return 0;
        
err6:
#ifdef CONFIG_FB_MXC_OVERLAY
        unregister_framebuffer(fbi_ovl);
err5:
        mxcfb_unmap_video_memory(fbi_ovl);
err4:
        if (&fbi_ovl->cmap)
                fb_dealloc_cmap(&fbi_ovl->cmap);
        framebuffer_release(fbi_ovl);
err3:
#endif
        unregister_framebuffer(fbi);

err2:
        mxcfb_unmap_video_memory(fbi);
err1:
        if (&fbi->cmap)
                fb_dealloc_cmap(&fbi->cmap);
        framebuffer_release(fbi);
err0:
        return ret;
}

/*!
 * This structure contains pointers to the power management callback functions.
 */
static struct device_driver mxcfb_driver = {
        .name                   = MXCFB_NAME,
        .bus                    = &platform_bus_type,
        .probe                  = mxcfb_probe,
        .suspend                = mxcfb_suspend,
        .resume                 = mxcfb_resume,
};

/*!
 * Device definition for the Framebuffer
 */
static struct platform_device mxcfb_device = {
        .name                   = MXCFB_NAME,
        .id                     = 0,
};

/*!
 * Main entry function for the framebuffer. The function registers the power
 * management callback functions with the kernel and also registers the MXCFB
 * callback functions with the core Linux framebuffer driver \b fbmem.c
 *
 * @return      Error code indicating success or failure
 */
int __init
mxcfb_init(void)
{
        int ret = 0;

        FUNC_START;

        if (fb_get_options("mxcfb", NULL))
                return -ENODEV;

        ret = driver_register(&mxcfb_driver);
        if (ret == 0) {
                ret = platform_device_register(&mxcfb_device);
                if (ret != 0) {
                        driver_unregister(&mxcfb_driver);
                }
        }

        return ret;
}

void mxcfb_exit(void)
{
        struct fb_info *fbi = mxcfb_drv_data.fbi;

        if (fbi) {
                mxcfb_unmap_video_memory(fbi);

                if (&fbi->cmap)
                        fb_dealloc_cmap(&fbi->cmap);

                unregister_framebuffer(fbi);
                framebuffer_release(fbi);
        }

        fbi = mxcfb_drv_data.fbi_ovl;
        if (fbi) {
                mxcfb_unmap_video_memory(fbi);

                if (&fbi->cmap)
                        fb_dealloc_cmap(&fbi->cmap);

                unregister_framebuffer(fbi);
                framebuffer_release(fbi);
        }

        ipu_free_irq(IPU_IRQ_SDC_DISP3_VSYNC, &mxcfb_drv_data);
        
        platform_device_unregister(&mxcfb_device);
        driver_unregister(&mxcfb_driver);
}

module_init(mxcfb_init);
module_exit(mxcfb_exit);

MODULE_AUTHOR("Freescale Semiconductor");
MODULE_DESCRIPTION("MXC framebuffer driver");
MODULE_SUPPORTED_DEVICE("fb");
