/*
 * Copyright 2004 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*!
 * @file mxcfb_epson.c
 *
 * @brief MXC Frame buffer driver for ADC
 *
 * @ingroup Framebuffer
 */
 
/*!
 * Include files
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/fb.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/device.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/mach-types.h>
#include <asm/uaccess.h>
#include <asm/arch/platform.h>
#include <asm/arch/gpio.h>
#include <asm/arch/board.h>
#include "../../mxc/ipu/ipu.h"

#include "../../mxc/atlas/module/atlas_power.h"
#include "../../mxc/atlas/module/atlas_light.h"

#include "mxcfb_epson.h"

/*
 * Debug Macros
 */
//#define MXCFB_DEBUG
#ifdef MXCFB_DEBUG

#  define DDPRINTK(fmt, args...) printk(KERN_ERR"%s :: %d :: %s - " fmt, __FILE__,__LINE__,__FUNCTION__ , ## args)
#  define DPRINTK(fmt, args...) printk("%s: " fmt, __FUNCTION__ , ## args)

#  define FUNC_START	DDPRINTK(" func start\n")
#  define FUNC_END	DDPRINTK(" func end\n")

#  define FUNC_ERR printk(KERN_ERR"%s :: %d :: %s  err= %d \n", __FILE__,__LINE__,__FUNCTION__ ,err)

#else //MXCFB_DEBUG

#define DDPRINTK(fmt, args...)  do {} while(0)
#define DPRINTK(fmt, args...)   do {} while(0)

#define FUNC_START
#define FUNC_END

#endif //MXCFB_DEBUG

#define EPSON_SCREEN_WIDTH              176
#define EPSON_SCREEN_HEIGHT             220

void mxcfb_epson_exit(void);
void slcd_gpio_config(void);

/*!
 * This structure contains the information such as the ID of the
 * framebuffer driver, scanlength, framebuffer acceleration and color
 * map. This is passed into the \b fbinfo structure.
 */
static struct fb_fix_screeninfo mxcfb_fix = {
        .id                     = MXCFB_NAME,
        .type                   = FB_TYPE_PACKED_PIXELS,
        .visual                 = FB_VISUAL_TRUECOLOR,
        .accel                  = FB_ACCEL_NONE,
        .xpanstep               = 1,
        .ypanstep               = 1,
};

/*!
 * This structure contains the information such as the screen width, 
 * screen height, virtual size, and bits per pixel. This is passed into 
 * the \b fbinfo structure.
 */
static struct fb_var_screeninfo mxcfb_var = {
        .bits_per_pixel         = MXCFB_BPP,
        .activate               = FB_ACTIVATE_NOW,
        .height                 = -1,
        .width                  = -1,
        .vmode                  = FB_VMODE_NONINTERLACED,
        .xres                   = EPSON_SCREEN_WIDTH,
        .yres                   = EPSON_SCREEN_HEIGHT,
        .xres_virtual           = EPSON_SCREEN_WIDTH,
        .yres_virtual           = EPSON_SCREEN_HEIGHT,
};

/*!
 * This structure contains the information for RGB 16bpp pixel data. 
 */
static struct mxcfb_rgb rgb_16 = {
        .pixel_fmt = IPU_PIX_FMT_RGB565,
        .red    = { .offset = 11,       .length = 5, },
        .green  = { .offset = 5,   .length = 6, },
        .blue   = { .offset = 0,  .length = 5, },
        .transp = { .offset = 0,  .length = 0, },
};

/*!
 * This structure contains the information for RGB 24bpp pixel data. 
 */
static struct mxcfb_rgb rgb_24 = {
        .pixel_fmt = IPU_PIX_FMT_BGR24,
        .red    = { .offset = 16,  .length = 8, },
        .green  = { .offset = 8,  .length = 8, },
        .blue   = { .offset = 0,  .length = 8, },
        .transp = { .offset = 0,  .length = 0, },
};

/*!
 * This structure contains the information for RGB 32bpp pixel data. 
 */
static struct mxcfb_rgb rgb_32 = {
        .pixel_fmt = IPU_PIX_FMT_BGR32,
        .red    = { .offset = 16,  .length = 8, },
        .green  = { .offset = 8,  .length = 8, },
        .blue   = { .offset = 0,  .length = 8, },
        .transp = { .offset = 24,  .length = 8, },
};

/*!
 * This function sets display region in the Epson panel
 *  
 * @param	 x1	 x-coordinate of one vertex.
 * @param	 x2	 x-coordinate of second vertex.
 * @param	 y1	 y-coordinate of one vertex.
 * @param	 y2	 y-coordinate of second vertex.
 */
void set_panel_region( uint32_t x1, uint32_t x2, uint32_t y1, uint32_t y2 )
{
        uint32_t param[8];

        memset(param, 0, sizeof(uint32_t) * 8);
        param[0] = x1;
        param[2] = x2;
        param[4] = y1;
        param[6] = y2;

        // SD_CSET
        ipu_adc_write_cmd(DISP0, CMD, SD_CSET, param, 4);
        // SD_PSET

        ipu_adc_write_cmd(DISP0, CMD, SD_PSET, &(param[4]), 4);
}

/*!
 * Function to create and initiate template command buffer for ADC. This template
 * will be written to Panel memory.
 */
static void init_channel_template(int disp)
{
        /* template command buffer for ADC is 32*/
        uint32_t tempCmd[TEMPLATE_BUF_SIZE];
        uint32_t i = 0;

        memset(tempCmd, 0, sizeof(uint32_t)*TEMPLATE_BUF_SIZE);
        /* setup update display region*/
        /* whole the screen during init*/
        /*WRITE Y COORDINATE CMND*/
        tempCmd[i++] = ipu_adc_template_gen(WR_CMND, 0, SINGLE_STEP, SD_PSET);  
        /*WRITE Y START ADDRESS CMND LSB[22:8]*/
        tempCmd[i++] = ipu_adc_template_gen(WR_YADDR, 1, SINGLE_STEP, 0x01);    
        /*WRITE Y START ADDRESS CMND MSB[22:16]*/
        tempCmd[i++] = ipu_adc_template_gen(WR_YADDR, 1, SINGLE_STEP, 0x09);    
        /*WRITE Y STOP ADDRESS CMND LSB*/
        tempCmd[i++] = ipu_adc_template_gen(WR_CMND, 1, SINGLE_STEP,\
                                   EPSON_SCREEN_HEIGHT - 1);
        /*WRITE Y STOP ADDRESS CMND MSB*/
        tempCmd[i++] = ipu_adc_template_gen(WR_CMND, 1, SINGLE_STEP, 0);        
        /*WRITE X COORDINATE CMND*/
        tempCmd[i++] = ipu_adc_template_gen(WR_CMND, 0, SINGLE_STEP, SD_CSET);  
        /*WRITE X ADDRESS CMND LSB[7:0]*/
        tempCmd[i++] = ipu_adc_template_gen(WR_XADDR, 1, SINGLE_STEP, 0x01);    
        /*WRITE X ADDRESS CMND MSB[22:8]*/
        tempCmd[i++] = ipu_adc_template_gen(WR_CMND, 1, SINGLE_STEP, 0);        
        /*WRITE X STOP ADDRESS CMND LSB*/
        tempCmd[i++] = ipu_adc_template_gen(WR_CMND, 1, SINGLE_STEP,\
                                   EPSON_SCREEN_WIDTH + 1);
        /*WRITE X STOP ADDRESS CMND MSB*/
        tempCmd[i++] = ipu_adc_template_gen(WR_CMND, 1, SINGLE_STEP, 0);        
        /*WRITE MEMORY CMND MSB*/
        tempCmd[i++] = ipu_adc_template_gen(WR_CMND, 0, SINGLE_STEP, RAMWR);    
        /*WRITE DATA CMND and STP*/
        tempCmd[i++] = ipu_adc_template_gen(WR_DATA, 1, STOP, 0);               

        ipu_adc_write_template(disp, tempCmd, true);
}

/*!
 * Function to initialize Epson panel. First it resets the panel and then initilizes 
 * panel.
 */
static void init_panel(int disp)
{
        struct t_sw3_conf sw3_conf;
        uint32_t cmd_param;
        uint32_t i;

        FUNC_START;

        slcd_gpio_config();

        sw3_conf.sw3_en = true;
        sw3_conf.sw3_op_mode = false;
        sw3_conf.sw3_ctr_stby = true;
        sw3_conf.sw3_setting = 0;
        atlas_power_set_sw3(&sw3_conf);

        atlas_light_set_bl_and_led_bais(true);
        atlas_light_set_bl_master_egde_slowing(true);
        atlas_light_bl_set_conf(BACKLIGHT_AUX, LEVEL_7, 15);
        atlas_light_set_bl_period(0);

        DPRINTK("\nEpsonPanel reset complete\n");

        // DATCTL
        // 16-bit 565 mode
        cmd_param = 0x28;
        ipu_adc_write_cmd(disp, CMD, DATCTL, &cmd_param, 1);
        // 8-bit 666 mode
//        cmd_param = 0x08;
//        ipu_adc_write_cmd(disp, CMD, DATCTL, &cmd_param, 1);


        // Sleep OUT
        ipu_adc_write_cmd(disp, CMD, SLPOUT, 0, 0);

        // Init Contrast
        cmd_param = 0x80;
        ipu_adc_write_cmd(disp, CMD, VOLCTR, &cmd_param, 1);
        cmd_param =0x0;
        ipu_adc_write_cmd(disp, CMD, VOLCTR, &cmd_param, 1);

        // Set display to white
        // Setup page and column addresses
        set_panel_region(2, 176 + 1, 0, 220 - 1);
        // Do RAM write cmd
        ipu_adc_write_cmd(disp, CMD, RAMWR, 0 , 0);
        for ( i= 0; i<(EPSON_SCREEN_WIDTH*EPSON_SCREEN_HEIGHT); i++)
                ipu_adc_write_cmd(DISP0, DAT, 0xFFFF, 0 , 0);

        // Pause 80 ms
        mdelay(80);

        // Display ON
        ipu_adc_write_cmd(disp, CMD, DISON, 0, 0);
        // Pause 200 ms
        mdelay(200);

        DPRINTK("mxcfb_epson: initialized panel\n");
}

static irqreturn_t mxcfb_sys2_eof_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
        ipu_channel_params_t params;
        struct mxcfb_info * mxc_fbi = dev_id;
        uint32_t stat[2], seg_size;
        uint32_t lsb, msb;
        uint32_t update_height, start_line, start_addr, end_line, end_addr;
        uint32_t stride_pixels = (mxc_fbi->fb.fix.line_length * 8)\
                                 /mxc_fbi->fb.var.bits_per_pixel;

        stat[0] = readl(M3IFSSR0);
        stat[1] = readl(M3IFSSR1);
        DPRINTK("epson: snoop status = 0x%08X%08X\n", stat[1], stat[0]);

        writel(0x0, M3IFSSR0);
        writel(0x0, M3IFSSR1);        

        if ( !stat[0] && !stat[1] ){
                printk("mxcfb_epson: error no bus snooping bits set\n");
                return IRQ_HANDLED;
        }
        ipu_disable_irq(IPU_IRQ_ADC_SYS2_EOF);

        lsb = ffs(stat[0]);
        if (lsb) {
                lsb--;
        }
        else {
                lsb = ffs(stat[1]);
                lsb += 32 - 1;
        }
        msb = fls(stat[1]);
        if (msb) {
                msb += 32;
        }
        else {
                msb = fls(stat[0]);
        }

        seg_size = mxc_fbi->snoop_window_size / 64;

        start_addr = lsb * seg_size; // starting address offset
        start_line = start_addr / mxc_fbi->fb.fix.line_length;
        start_addr = start_line * mxc_fbi->fb.fix.line_length;  // Address aligned to line
        start_addr += mxc_fbi->phys_addr;

        end_addr = msb * seg_size; // ending address offset
        end_line = end_addr / mxc_fbi->fb.fix.line_length;
        end_line++;

        if (end_line > mxc_fbi->fb.var.yres) {
                end_line = mxc_fbi->fb.var.yres;
        }
        
        update_height = end_line - start_line;
        DPRINTK("epson: updating rows %d to %d, start addr = 0x%08X\n", start_line, end_line, start_addr);
        
        params.adc_sys1.disp = DISP0;
        params.adc_sys1.ch_mode = WriteTemplateNonSeq;
        params.adc_sys1.x_pos = 2;
        params.adc_sys1.y_pos = start_line;
        ipu_init_channel(ADC_SYS1, &params);

        ipu_init_channel_buffer(ADC_SYS1, IPU_INPUT_BUFFER,
                mxc_fbi->cur_pixel_fmt, EPSON_SCREEN_WIDTH,
                update_height,
                stride_pixels,
                IPU_ROTATE_NONE,
                (void*)start_addr, NULL);
        ipu_enable_channel(ADC_SYS1);
        ipu_select_buffer(ADC_SYS1, IPU_INPUT_BUFFER, 0);
        ipu_enable_irq(IPU_IRQ_ADC_SYS1_EOF);

        return IRQ_HANDLED;
}


static irqreturn_t mxcfb_sys1_eof_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
        ipu_disable_irq(IPU_IRQ_ADC_SYS1_EOF);
        ipu_disable_channel(ADC_SYS1, false);
        ipu_uninit_channel(ADC_SYS1);

        ipu_enable_channel(ADC_SYS2);
        ipu_enable_irq(IPU_IRQ_ADC_SYS2_EOF);

        return IRQ_HANDLED;
}
/*!
 * Function to initialize Asynchronous Display Controller. It also initilizes the ADC
 * System1 channel. Configure ADC display 0 parallel interface for Epson panel.
 * 
 * @param       fbi     framebuffer information pointer
 */
 void 
mxcfb_init_controller(struct mxcfb_info *fbi)
{
        ipu_channel_params_t params;
        uint32_t stride_pixels = (fbi->fb.fix.line_length * 8)\
                                             / fbi->fb.var.bits_per_pixel;
        uint32_t memsize = fbi->map_size;
        
        ipu_adc_sig_cfg_t sig = {0, 0, 0, 0, 0, 0, 0, 0, 
                                 IPU_ADC_NO_RW, IPU_ADC_BURST_WCS, 
                                 IPU_ADC_IFC_MODE_SYS80_TYPE2};

        DPRINTK("Enabling LCD controller line_length=%d BPP=%d\n",fbi->fb.fix.line_length,fbi->fb.var.bits_per_pixel);

        if (ipu_request_irq(IPU_IRQ_ADC_SYS2_EOF, mxcfb_sys2_eof_irq_handler, 0, MXCFB_NAME, fbi) != 0) {
                printk("mxcfb: Error registering irq handler.\n");
                        return;
        }

        if (ipu_request_irq(IPU_IRQ_ADC_SYS1_EOF, mxcfb_sys1_eof_irq_handler, 0, MXCFB_NAME, fbi) != 0) {
                printk("mxcfb: Error registering irq handler.\n");
                        return;
        }
        ipu_enable_irq(IPU_IRQ_ADC_SYS1_EOF);

        // Init DI interface
        ipu_adc_init_parallel_ifc(DISP0, EPSON_SCREEN_WIDTH + 2, EPSON_SCREEN_HEIGHT,
                                  IPU_PIX_FMT_RGB565, stride_pixels,
                                  sig, XY, 0, VsyncInternal);

        memset(&params, 0, sizeof(params));
        params.adc_sys1.disp = DISP0;
        params.adc_sys1.ch_mode = WriteTemplateNonSeq;
        params.adc_sys1.x_pos = 2;
        params.adc_sys1.y_pos = 0;
        ipu_init_channel(ADC_SYS1, &params);

        init_panel(DISP0);
        init_channel_template(DISP0);
        if (ipu_adc_set_update_mode(ADC_SYS1, IPU_ADC_REFRESH_NONE, 60, 0, 0) < 0)
                printk("mxcfb: Error enabling auto refesh.\n");


        ipu_init_channel_buffer(ADC_SYS1, IPU_INPUT_BUFFER,
                fbi->cur_pixel_fmt, EPSON_SCREEN_WIDTH,
                EPSON_SCREEN_HEIGHT,
                stride_pixels,
                IPU_ROTATE_NONE,
                (void*)fbi->phys_addr, NULL);
        ipu_enable_channel(ADC_SYS1);

        params.adc_sys2.disp = DISP1;
        params.adc_sys2.ch_mode = WriteTemplateNonSeq;
        params.adc_sys2.x_pos = 0;
        params.adc_sys2.y_pos = 0;
        ipu_init_channel(ADC_SYS2, &params);

        ipu_adc_init_parallel_ifc(DISP1, EPSON_SCREEN_WIDTH, EPSON_SCREEN_HEIGHT,
                                  IPU_PIX_FMT_RGB565, stride_pixels,
                                  sig, XY, 0, VsyncInternal);

        init_channel_template(DISP1);

        if (ipu_adc_set_update_mode(ADC_SYS2, IPU_ADC_SNOOPING, 60, fbi->phys_addr, &memsize) < 0)
                printk("mxcfb: Error enabling auto refesh.\n");

        fbi->snoop_window_size = memsize;
       
        ipu_init_channel_buffer(ADC_SYS2, IPU_INPUT_BUFFER, fbi->cur_pixel_fmt,
                                0x1,
                                0x1,
                                stride_pixels,
                                IPU_ROTATE_NONE,
                                (void*)fbi->phys_addr, NULL);

        ipu_select_buffer(ADC_SYS1, IPU_INPUT_BUFFER, 0);

}

/*!
 * Function to disable ADC by uninitializing channels.
 * 
 * @param       fbi     framebuffer information pointer
 */
static void 
mxcfb_uninit_controller(struct mxcfb_info *fbi)
{
        printk("Disabling LCD controller\n");

        ipu_disable_channel(ADC_SYS1, true);
        ipu_uninit_channel(ADC_SYS1);
}

#ifdef CONFIG_PM
/*!
 * Power management hooks.      Note that we won't be called from IRQ context,
 * unlike the blank functions above, so we may sleep.
 */

/*!
 * Suspends the framebuffer and blanks the screen. Power management support 
 *
 * @param	dev	pointer to device structure.
 * @param	state	state of the device.
 * @param	level	level of suspend.
 *
 * @return	success
 */
static int 
mxcfb_suspend(struct device *dev, u32 state, u32 level)
{
        printk("Suspending LCD\n");
        
        if (level == SUSPEND_DISABLE || level == SUSPEND_POWER_DOWN) {
                ipu_disable_channel(ADC_SYS1, true);
//                SetContrast(0.0f);
        }
        return 0;
}

/*!
 * Resumes the framebuffer and unblanks the screen. Power management support 
 * 
 * @param       dev     pointer to device structure.
 * @param       level   level of suspend.
 *
 * @return      success
 */
static int 
mxcfb_resume(struct device *dev, u32 level)
{
        printk("Resuming LCD\n");

        if (level == RESUME_ENABLE) {
                ipu_enable_channel(ADC_SYS1);
  //              SetContrast(1.0f);
        }
        return 0;
}
#else
#define mxcfb_suspend   NULL
#define mxcfb_resume    NULL
#endif

/*
 * Main framebuffer functions
 */
/*
 * Formal definition of the VESA spec:
 *      On
 *              This refers to the state of the display when it is in full operation
 *      Stand-By
 *              This defines an optional operating state of minimal power reduction with
 *              the shortest recovery time
 *      Suspend
 *              This refers to a level of power management in which substantial power
 *              reduction is achieved by the display.  The display can have a longer 
 *              recovery time from this state than from the Stand-by state
 *      Off
 *              This indicates that the display is consuming the lowest level of power
 *              and is non-operational. Recovery from this state may optionally require
 *              the user to manually power on the monitor
 *
 *      Now, the fbdev driver adds an additional state, (blank), where they
 *      turn off the video (maybe by colormap tricks), but don't mess with the
 *      video itself: think of it semantically between on and Stand-By.
 *
 *      So here's what we should do in our fbdev blank routine:
 *
 *              VESA_NO_BLANKING (mode 0)       Video on,  front/back light on
 *              VESA_VSYNC_SUSPEND (mode 1)     Video on,  front/back light off
 *              VESA_HSYNC_SUSPEND (mode 2)     Video on,  front/back light off
 *              VESA_POWERDOWN (mode 3)         Video off, front/back light off
 *
 *      This will match the matrox implementation.
 */
/*!
 * Blank the display by setting all palette values to zero.  
 * 	
 * @param	blank	value to set blank mode.
 * @param	info	pointer to framebuffer info structure.
 *
 * @return	success
 */
static int mxcfb_blank(int blank, struct fb_info *info)
{
        printk("mxcfb_blank: blank=%d\n", blank);

        switch (blank) {
                case VESA_POWERDOWN:
                case VESA_VSYNC_SUSPEND:
                case VESA_HSYNC_SUSPEND:
                        ipu_disable_channel(ADC_SYS1, true);
          //              SetContrast(0.0f);
                        break;

                case VESA_NO_BLANKING:
                        ipu_enable_channel(ADC_SYS1);
            //            SetContrast(1.0f);
                        break;
        }
        return 0;
}

/*!
 * Allocates the DRAM memory for the frame buffer.      This buffer is remapped 
 * into a non-cached, non-buffered, memory region to allow palette and pixel 
 * writes to occur without flushing the cache.  Once this area is remapped, 
 * all virtual memory access to the video memory should occur at the new region.
 *
 * @param       fbi     framebuffer information pointer
 * 
 * @return      Error code indicating success or failure
 */
static int
mxcfb_map_video_memory(struct mxcfb_info *fbi)
{
        fbi->map_size = fbi->fb.fix.line_length * fbi->fb.var.yres_virtual;
        fbi->fb.fix.smem_len = fbi->map_size;

        fbi->phys_addr = IPU_Malloc(fbi->map_size);
        if (fbi->phys_addr == 0) {
                printk("MXCFB - Unable to allocate framebuffer memory\n");
                return -EBUSY;
        }
        DPRINTK("mxcfb: allocated fb @ paddr=0x%08X, size=%d.\n", 
               (uint32_t)fbi->phys_addr, fbi->map_size);
        
        if (!request_mem_region(fbi->phys_addr, fbi->map_size, "LCD")) {
                return -EBUSY;
        }

        if (!(fbi->virt_addr = ioremap(fbi->phys_addr, fbi->map_size))) {
                release_mem_region(fbi->phys_addr, fbi->map_size);
                return -EIO;
        }

        if (fbi->virt_addr) {
                fbi->fb.screen_base = fbi->virt_addr;
                fbi->fb.fix.smem_start = fbi->phys_addr;
                printk("MXCFB - framebuffer vaddr = 0x%08X\n",\
                                              (uint32_t)fbi->virt_addr);
        }

        /* Clear the screen */
        memset(fbi->fb.screen_base, 0, fbi->map_size);
        return 0;
}

/*!
 * This structure contains the pointers to the control functions that are 
 * invoked by the core framebuffer driver to perform operations like 
 * blitting, rectangle filling, copy regions and cursor definition.
 */
static struct fb_ops mxcfb_ops = {
        .owner                          = THIS_MODULE,
        .fb_fillrect                    = cfb_fillrect,
        .fb_copyarea                    = cfb_copyarea,
        .fb_imageblit                   = cfb_imageblit,
        .fb_blank                       = mxcfb_blank,
        .fb_cursor                      = soft_cursor,
};

/*!
 * De-allocates the DRAM memory for the frame buffer.
 *
 * @param       fbi     framebuffer information pointer
 *
 * @return      Error code indicating success or failure
 */
static int
mxcfb_unmap_video_memory(struct mxcfb_info *fbi)
{
        iounmap(fbi->virt_addr);
        release_mem_region(fbi->phys_addr, fbi->map_size);
        fbi->phys_addr = IPU_Malloc(fbi->map_size);
        return 0;
}

/*!
 * Initializes the framebuffer information pointer. After allocating
 * sufficient memory for the framebuffer structure, the fields are
 * filled with custom information passed in from the configurable
 * structures.  This includes information such as bits per pixel,
 * color maps, screen width/height and RGBA offsets.
 *
 * @return      Framebuffer structure initialized with our information
 */
static struct mxcfb_info * 
mxcfb_init_fbinfo(void)
{
        struct mxcfb_info *fbi;
        int rgbidx = 0;
        uint32_t msb;

        /*
         * Allocate sufficient memory for the fb structure
         */
        fbi = kmalloc(sizeof(struct mxcfb_info) + sizeof(u32) * 16, GFP_KERNEL);

        if (!fbi)
                return NULL;

        memset(fbi, 0, sizeof(struct mxcfb_info));

        /*
         * Fill in structure information  
         */
        fbi->fb.fix                     = mxcfb_fix;
        fbi->fb.var                     = mxcfb_var;
        fbi->fb.fbops                   = &mxcfb_ops;
        fbi->fb.flags                   = FBINFO_FLAG_DEFAULT;
        fbi->fb.pseudo_palette  = (fbi + 1);

        fbi->rgb[RGB_16] = &rgb_16;
        fbi->rgb[RGB_24] = &rgb_24;
        fbi->rgb[RGB_32] = &rgb_32;

        // Init settings based on the panel size
        // IPU XY mode requires stride pixels aligned to a power of 2
        msb = fls(EPSON_SCREEN_WIDTH);
        if (!(EPSON_SCREEN_WIDTH & ((1UL << msb) - 1)))
                msb--;       // Already aligned to power 2
        fbi->fb.fix.line_length = 1UL << msb;
        fbi->fb.fix.line_length *= fbi->fb.var.bits_per_pixel / 8;
        
        switch (fbi->fb.var.bits_per_pixel) {
        case 16:
                rgbidx = RGB_16;
                rgbidx = RGB_16;
                break;
        case 24:
                rgbidx = RGB_24;
                break;
        case 32:
                rgbidx = RGB_32;
                break;
        }

        /*
         * Copy the RGB parameters for this display
         * from the machine specific parameters.
         */
        fbi->fb.var.red    = fbi->rgb[rgbidx]->red;
        fbi->fb.var.green  = fbi->rgb[rgbidx]->green;
        fbi->fb.var.blue   = fbi->rgb[rgbidx]->blue;
        fbi->fb.var.transp = fbi->rgb[rgbidx]->transp;
        fbi->cur_pixel_fmt = fbi->rgb[rgbidx]->pixel_fmt;

        /*
         * Allocate colormap
         */
        fb_alloc_cmap(&fbi->fb.cmap, 1<<fbi->fb.var.bits_per_pixel, 0);

        return fbi;
}

/*!
 * Probe routine for the framebuffer driver. It is called during the
 * driver binding process.      The following functions are performed in
 * this routine: Framebuffer initialization, Memory allocation and
 * mapping, Framebuffer registration, IPU initialization.
 *
 * @return      Appropriate error code to the kernel common code
 */
static int mxcfb_probe(struct device *dev)
{
        struct mxcfb_info *fbi;
        int ret;

        /*
         * Initialize FB structures
         */
        fbi = mxcfb_init_fbinfo();

        if (!fbi)
                return -ENOMEM;

        /*
         * Allocate memory
         */
        ret = mxcfb_map_video_memory(fbi);

        if (ret)
                return -ENOMEM;

        dev_set_drvdata(dev, fbi);

        /*
         * Register framebuffer
         */
        if (register_framebuffer(&fbi->fb) < 0) {
                /*
                 * Unmap and free memory
                 */
                iounmap(fbi->fb.screen_base);
                release_mem_region(fbi->fb.fix.smem_start,fbi->fb.fix.smem_len);
                return -EINVAL;
        }

        
        mxcfb_init_controller(fbi);

        return 0;
}

/*!
 * This structure contains pointers to the power management callback functions.
 */
static struct device_driver mxcfb_driver = {
        .name                   = MXCFB_NAME,
        .bus                    = &platform_bus_type,
        .probe                  = mxcfb_probe,
        .suspend                = mxcfb_suspend,
        .resume                 = mxcfb_resume,
};

/*!
 * Device definition for the Framebuffer
 */
static struct platform_device mxcfb_device = {
        .name                   = MXCFB_NAME,
        .id                     = 0,
};

/*!
 * Main entry function for the framebuffer. The function registers the power
 * management callback functions with the kernel and also registers the MXCFB
 * callback functions with the core Linux framebuffer driver \b fbmem.c
 *
 * @return      Error code indicating success or failure
 */
int mxcfb_epson_init(void)
{
        int ret = 0;

        FUNC_START;

        if (fb_get_options("mxcfb_epson", NULL))
                return -ENODEV;

        ret = driver_register(&mxcfb_driver);
        if (ret == 0) {
                ret = platform_device_register(&mxcfb_device);
                if (ret != 0) {
                        driver_unregister(&mxcfb_driver);
                }
        }

        return ret;
}

void mxcfb_epson_exit(void)
{
        struct mxcfb_info *fbi = dev_get_drvdata(&mxcfb_device.dev);

        if (fbi) {
                mxcfb_unmap_video_memory(fbi);

                if (&fbi->fb.cmap)
                        fb_dealloc_cmap(&fbi->fb.cmap);

                unregister_framebuffer(&fbi->fb);
                mxcfb_uninit_controller(fbi);
                kfree(fbi);
        }

        platform_device_unregister(&mxcfb_device);
        driver_unregister(&mxcfb_driver);
}

module_init(mxcfb_epson_init);
module_exit(mxcfb_epson_exit);

MODULE_AUTHOR("Freescale Semiconductor");
MODULE_DESCRIPTION("MXC Epson framebuffer driver");
MODULE_SUPPORTED_DEVICE("fb");
