/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef __GPIO_H__
#define __GPIO_H__

/*!
 * @defgroup GPIO General Purpose Input Output (GPIO)
 * @ingroup MSL
 */

/*!
 * @file gpio.h
 * @brief This file contains the GPIO and IOMUX configuration details.
 *
 * @ingroup GPIO
 */

#include <asm/sizes.h>
#include <asm/arch/platform.h>
#include <linux/interrupt.h>

/*!
 * @name IOMUX/PAD Bit field definitions
 */

/*! @{ */

/*!
 * In order to identify pins more effectively, each mux-controlled pin's 
 * enumerated value is constructed in the following way:
 *
 * ------------------------------------------------
 * 31 | 30     24 | 23     16| 15     9 | 8     0
 * ------------------------------------------------
 *    | PAD_F     | PAD_I    | MUX_F    | MUX_I
 * ------------------------------------------------
 *
 * Bit 31 defines the MUX type (MUX_T) to be either regular mode or extended 
 * mode. ArgonPlus only has multiplexed pins in regular mode so this field 
 * define is not used. Bit 0 to 8 contains MUX_I used to identify the register
 * offset (base is IOMUX_module_base + 0xC) defined in the Section "IOMUX ctl 
 * registers description" of the ArgonPlus IC Spec. Bit 9 to 15 is MUX_F which
 * contains the offset value defined within the same register (each IOMUX 
 * control register contains four 8-bit fields for four different pins). The 
 * similar field definitions are used for the pad control register.
 */

/*!
 * Index of MUX control register
 */
#define MUX_I        0

/*!
 * Field within register for MUX control bits
 */
#define MUX_F        8

/*!
 * IOMUX type defined in bit 31
 */
#define MUX_T         31

/*!
 * Index of pad control register
 */
#define PAD_I        16

/*!
 *
 * Field within register for pad control bits
 */
#define PAD_F        24

/*! @} End IOMUX/PAD Bit field definitions */

/*!
 * This enumeration is constructed based on the Section "IOMUX ctl registers
 * description" of the ArgonPlus IC Spec. Each enumerated
 * value is constructed based on the rules described above.
 */
enum iomux_pins {
        PIN_A0_MA0             = (107 << MUX_I) | (3 << MUX_F) | (142 << PAD_I) | (2 << PAD_F),
        PIN_A1_MA1             = (107 << MUX_I) | (2 << MUX_F) | (142 << PAD_I) | (1 << PAD_F),
        PIN_A10                = (105 << MUX_I) | (1 << MUX_F) | (139 << PAD_I) | (1 << PAD_F),
        PIN_A11_MA11           = (104 << MUX_I) | (3 << MUX_F) | (138 << PAD_I) | (2 << PAD_F),
        PIN_A12_MA12           = (104 << MUX_I) | (2 << MUX_F) | (138 << PAD_I) | (1 << PAD_F),
        PIN_A13_MA13           = (104 << MUX_I) | (1 << MUX_F) | (138 << PAD_I) | (0 << PAD_F),
        PIN_A14                = (104 << MUX_I) | (0 << MUX_F) | (137 << PAD_I) | (2 << PAD_F),
        PIN_A15                = (103 << MUX_I) | (3 << MUX_F) | (137 << PAD_I) | (1 << PAD_F),
        PIN_A16                = (103 << MUX_I) | (2 << MUX_F) | (137 << PAD_I) | (0 << PAD_F),
        PIN_A17                = (103 << MUX_I) | (1 << MUX_F) | (136 << PAD_I) | (2 << PAD_F),
        PIN_A18                = (103 << MUX_I) | (0 << MUX_F) | (136 << PAD_I) | (1 << PAD_F),
        PIN_A19                = (102 << MUX_I) | (3 << MUX_F) | (136 << PAD_I) | (0 << PAD_F),
        PIN_A2_MA2             = (107 << MUX_I) | (1 << MUX_F) | (142 << PAD_I) | (0 << PAD_F),
        PIN_A20                = (102 << MUX_I) | (2 << MUX_F) | (135 << PAD_I) | (2 << PAD_F),
        PIN_A21                = (102 << MUX_I) | (1 << MUX_F) | (135 << PAD_I) | (1 << PAD_F),
        PIN_A22                = (102 << MUX_I) | (0 << MUX_F) | (135 << PAD_I) | (0 << PAD_F),
        PIN_A23                = (101 << MUX_I) | (3 << MUX_F) | (134 << PAD_I) | (2 << PAD_F),
        PIN_A24                = (101 << MUX_I) | (2 << MUX_F) | (134 << PAD_I) | (1 << PAD_F),
        PIN_A25                = (101 << MUX_I) | (1 << MUX_F) | (134 << PAD_I) | (0 << PAD_F),
        PIN_A3_MA3             = (107 << MUX_I) | (0 << MUX_F) | (141 << PAD_I) | (2 << PAD_F),
        PIN_A4_MA4             = (106 << MUX_I) | (3 << MUX_F) | (141 << PAD_I) | (1 << PAD_F),
        PIN_A5_MA5             = (106 << MUX_I) | (2 << MUX_F) | (141 << PAD_I) | (0 << PAD_F),
        PIN_A6_MA6             = (106 << MUX_I) | (1 << MUX_F) | (140 << PAD_I) | (2 << PAD_F),
        PIN_A7_MA7             = (106 << MUX_I) | (0 << MUX_F) | (140 << PAD_I) | (1 << PAD_F),
        PIN_A8_MA8             = (105 << MUX_I) | (3 << MUX_F) | (140 << PAD_I) | (0 << PAD_F),
        PIN_A9_MA9             = (105 << MUX_I) | (2 << MUX_F) | (139 << PAD_I) | (2 << PAD_F),
        PIN_ALE                = (85 << MUX_I) | (1 << MUX_F) |  (112 << PAD_I) | (2 << PAD_F),
        PIN_BATT_LINE          = (42 << MUX_I) | (1 << MUX_F) |  ( 55 << PAD_I) | (1 << PAD_F),
        PIN_BBP_RX_CLK         = (69 << MUX_I) | (3 << MUX_F) |  ( 92 << PAD_I) | (0 << PAD_F),
        PIN_BBP_RX_DATA        = (68 << MUX_I) | (3 << MUX_F) |  ( 90 << PAD_I) | (2 << PAD_F),
        PIN_BBP_RX_FRAME       = (69 << MUX_I) | (1 << MUX_F) |  ( 91 << PAD_I) | (1 << PAD_F),
        PIN_BBP_TX_CLK         = (69 << MUX_I) | (2 << MUX_F) |  ( 91 << PAD_I) | (2 << PAD_F),
        PIN_BBP_TX_DATA        = (68 << MUX_I) | (2 << MUX_F) |  ( 90 << PAD_I) | (1 << PAD_F),
        PIN_BBP_TX_FRAME       = (69 << MUX_I) | (0 << MUX_F) |  ( 91 << PAD_I) | (0 << PAD_F),
        PIN_BCLK               = (88 << MUX_I) | (2 << MUX_F) |  (117 << PAD_I) | (0 << PAD_F),
        PIN_BOOT_MODE0         = (78 << MUX_I) | (0 << MUX_F) |  (103 << PAD_I) | (0 << PAD_F),
        PIN_BOOT_MODE1         = (77 << MUX_I) | (3 << MUX_F) |  (102 << PAD_I) | (2 << PAD_F),
        PIN_CAS                = (87 << MUX_I) | (1 << MUX_F) |  (115 << PAD_I) | (1 << PAD_F),
        PIN_CKIH_REF_CLK_IN    = (79 << MUX_I) | (2 << MUX_F) |  (105 << PAD_I) | (0 << PAD_F),
        PIN_CKIL               = (80 << MUX_I) | (1 << MUX_F) |  (106 << PAD_I) | (0 << PAD_F),
        PIN_CKO1               = (79 << MUX_I) | (1 << MUX_F) |  (104 << PAD_I) | (2 << PAD_F),
        PIN_CKO2               = (79 << MUX_I) | (0 << MUX_F) |  (104 << PAD_I) | (1 << PAD_F),
        PIN_CLE                = (85 << MUX_I) | (3 << MUX_F) |  (113 << PAD_I) | (1 << PAD_F),
        PIN_CS0_B              = (90 << MUX_I) | (3 << MUX_F) |  (120 << PAD_I) | (0 << PAD_F),
        PIN_CS1_B              = (90 << MUX_I) | (2 << MUX_F) |  (119 << PAD_I) | (2 << PAD_F),
        PIN_CS2_B              = (90 << MUX_I) | (1 << MUX_F) |  (119 << PAD_I) | (1 << PAD_F),
        PIN_CS3_B              = (90 << MUX_I) | (0 << MUX_F) |  (119 << PAD_I) | (0 << PAD_F),
        PIN_CS4_B              = (89 << MUX_I) | (3 << MUX_F) |  (118 << PAD_I) | (2 << PAD_F),
        PIN_CSPI1_CK           = (40 << MUX_I) | (2 << MUX_F) |  ( 53 << PAD_I) | (0 << PAD_F),
        PIN_CSPI1_CS_0         = (39 << MUX_I) | (3 << MUX_F) |  ( 52 << PAD_I) | (0 << PAD_F),
        PIN_CSPI1_CS_1         = (39 << MUX_I) | (2 << MUX_F) |  ( 51 << PAD_I) | (2 << PAD_F),
        PIN_CSPI1_DI           = (40 << MUX_I) | (0 << MUX_F) |  ( 52 << PAD_I) | (1 << PAD_F),
        PIN_CSPI1_DO           = (40 << MUX_I) | (1 << MUX_F) |  ( 52 << PAD_I) | (2 << PAD_F),
        PIN_CSPI2_CK           = (41 << MUX_I) | (3 << MUX_F) |  ( 54 << PAD_I) | (2 << PAD_F),
        PIN_CSPI2_CS_0         = (41 << MUX_I) | (0 << MUX_F) |  ( 53 << PAD_I) | (2 << PAD_F),
        PIN_CSPI2_CS_1         = (40 << MUX_I) | (3 << MUX_F) |  ( 53 << PAD_I) | (1 << PAD_F),
        PIN_CSPI2_DI           = (41 << MUX_I) | (1 << MUX_F) |  ( 54 << PAD_I) | (0 << PAD_F),
        PIN_CSPI2_DO           = (41 << MUX_I) | (2 << MUX_F) |  ( 54 << PAD_I) | (1 << PAD_F),
        PIN_D0                 = (84 << MUX_I) | (1 << MUX_F) |  (111 << PAD_I) | (1 << PAD_F),
        PIN_D1                 = (84 << MUX_I) | (0 << MUX_F) |  (111 << PAD_I) | (0 << PAD_F),
        PIN_D10                = (81 << MUX_I) | (3 << MUX_F) |  (108 << PAD_I) | (0 << PAD_F),
        PIN_D11                = (81 << MUX_I) | (2 << MUX_F) |  (107 << PAD_I) | (2 << PAD_F),
        PIN_D12                = (81 << MUX_I) | (1 << MUX_F) |  (107 << PAD_I) | (1 << PAD_F),
        PIN_D13                = (81 << MUX_I) | (0 << MUX_F) |  (107 << PAD_I) | (0 << PAD_F),
        PIN_D14                = (80 << MUX_I) | (3 << MUX_F) |  (106 << PAD_I) | (2 << PAD_F),
        PIN_D15                = (80 << MUX_I) | (2 << MUX_F) |  (106 << PAD_I) | (1 << PAD_F),
        PIN_D2                 = (83 << MUX_I) | (3 << MUX_F) |  (110 << PAD_I) | (2 << PAD_F),
        PIN_D3                 = (83 << MUX_I) | (2 << MUX_F) |  (110 << PAD_I) | (1 << PAD_F),
        PIN_D4                 = (83 << MUX_I) | (1 << MUX_F) |  (110 << PAD_I) | (0 << PAD_F),
        PIN_D5                 = (83 << MUX_I) | (0 << MUX_F) |  (109 << PAD_I) | (2 << PAD_F),
        PIN_D6                 = (82 << MUX_I) | (3 << MUX_F) |  (109 << PAD_I) | (1 << PAD_F),
        PIN_D7                 = (82 << MUX_I) | (2 << MUX_F) |  (109 << PAD_I) | (0 << PAD_F),
        PIN_D8                 = (82 << MUX_I) | (1 << MUX_F) |  (108 << PAD_I) | (2 << PAD_F),
        PIN_D9                 = (82 << MUX_I) | (0 << MUX_F) |  (108 << PAD_I) | (1 << PAD_F),
        PIN_DAM1_RXD           = (45 << MUX_I) | (2 << MUX_F) |  ( 59 << PAD_I) | (2 << PAD_F),
        PIN_DAM1_T_CLK         = (45 << MUX_I) | (1 << MUX_F) |  ( 59 << PAD_I) | (1 << PAD_F),
        PIN_DAM1_T_FS          = (45 << MUX_I) | (0 << MUX_F) |  ( 59 << PAD_I) | (0 << PAD_F),
        PIN_DAM1_TXD           = (45 << MUX_I) | (3 << MUX_F) |  ( 60 << PAD_I) | (0 << PAD_F),
        PIN_DAM2_RXD           = (44 << MUX_I) | (2 << MUX_F) |  ( 58 << PAD_I) | (1 << PAD_F),
        PIN_DAM2_T_CLK         = (44 << MUX_I) | (1 << MUX_F) |  ( 58 << PAD_I) | (0 << PAD_F),
        PIN_DAM2_T_FS          = (44 << MUX_I) | (0 << MUX_F) |  ( 57 << PAD_I) | (2 << PAD_F),
        PIN_DAM2_TXD           = (44 << MUX_I) | (3 << MUX_F) |  ( 58 << PAD_I) | (2 << PAD_F),
        PIN_DAM3_R_CLK         = (42 << MUX_I) | (3 << MUX_F) |  ( 56 << PAD_I) | (0 << PAD_F),
        PIN_DAM3_R_FS          = (42 << MUX_I) | (2 << MUX_F) |  ( 55 << PAD_I) | (2 << PAD_F),
        PIN_DAM3_RXD           = (43 << MUX_I) | (2 << MUX_F) |  ( 57 << PAD_I) | (0 << PAD_F),
        PIN_DAM3_T_CLK         = (43 << MUX_I) | (1 << MUX_F) |  ( 56 << PAD_I) | (2 << PAD_F),
        PIN_DAM3_T_FS          = (43 << MUX_I) | (0 << MUX_F) |  ( 56 << PAD_I) | (1 << PAD_F),
        PIN_DAM3_TXD           = (43 << MUX_I) | (3 << MUX_F) |  ( 57 << PAD_I) | (1 << PAD_F),
        PIN_DBGACK             = (11 << MUX_I) | (0 << MUX_F) |  ( 13 << PAD_I) | (2 << PAD_F),
        PIN_DBGRQ              = (10 << MUX_I) | (1 << MUX_F) |  ( 12 << PAD_I) | (2 << PAD_F),
        PIN_DE_B               = (74 << MUX_I) | (2 << MUX_F) |  ( 98 << PAD_I) | (1 << PAD_F),
        PIN_DQM0               = (92 << MUX_I) | (0 << MUX_F) |  (121 << PAD_I) | (2 << PAD_F),
        PIN_DQM1               = (91 << MUX_I) | (3 << MUX_F) |  (121 << PAD_I) | (1 << PAD_F),
        PIN_DQM2               = (91 << MUX_I) | (2 << MUX_F) |  (121 << PAD_I) | (0 << PAD_F),
        PIN_DQM3               = (91 << MUX_I) | (1 << MUX_F) |  (120 << PAD_I) | (2 << PAD_F),
        PIN_DSM1_RX_ON         = (63 << MUX_I) | (1 << MUX_F) |  ( 83 << PAD_I) | (1 << PAD_F),
        PIN_DSM1_STBY          = (63 << MUX_I) | (2 << MUX_F) |  ( 83 << PAD_I) | (2 << PAD_F),
        PIN_DSM2_STBY          = (63 << MUX_I) | (0 << MUX_F) |  ( 83 << PAD_I) | (0 << PAD_F),
        PIN_DSM2_VCO_EN        = (62 << MUX_I) | (3 << MUX_F) |  ( 82 << PAD_I) | (2 << PAD_F),
        PIN_EB0_B              = (88 << MUX_I) | (0 << MUX_F) |  (116 << PAD_I) | (1 << PAD_F),
        PIN_EB3_B              = (87 << MUX_I) | (3 << MUX_F) |  (116 << PAD_I) | (0 << PAD_F),
        PIN_ECB_B              = (89 << MUX_I) | (0 << MUX_F) |  (117 << PAD_I) | (2 << PAD_F),
        PIN_EVTI_B             = (74 << MUX_I) | (1 << MUX_F) |  ( 98 << PAD_I) | (0 << PAD_F),
        PIN_EVTO               = (1 << MUX_I) | (2 << MUX_F) |   (  1 << PAD_I) | (0 << PAD_F),
        PIN_EXTRIG             = (10 << MUX_I) | (0 << MUX_F) |  ( 12 << PAD_I) | (1 << PAD_F),
        PIN_GPIO0              = (56 << MUX_I) | (1 << MUX_F) |  ( 74 << PAD_I) | (0 << PAD_F),
        PIN_GPIO1              = (56 << MUX_I) | (0 << MUX_F) |  ( 73 << PAD_I) | (2 << PAD_F),
        PIN_GPIO10             = (53 << MUX_I) | (3 << MUX_F) |  ( 70 << PAD_I) | (2 << PAD_F),
        PIN_GPIO11             = (53 << MUX_I) | (2 << MUX_F) |  ( 70 << PAD_I) | (1 << PAD_F),
        PIN_GPIO12             = (53 << MUX_I) | (1 << MUX_F) |  ( 70 << PAD_I) | (0 << PAD_F),
        PIN_GPIO13             = (53 << MUX_I) | (0 << MUX_F) |  ( 69 << PAD_I) | (2 << PAD_F),
        PIN_GPIO14             = (52 << MUX_I) | (3 << MUX_F) |  ( 69 << PAD_I) | (1 << PAD_F),
        PIN_GPIO15             = (52 << MUX_I) | (2 << MUX_F) |  ( 69 << PAD_I) | (0 << PAD_F),
        PIN_GPIO16             = (52 << MUX_I) | (1 << MUX_F) |  ( 68 << PAD_I) | (2 << PAD_F),
        PIN_GPIO17             = (52 << MUX_I) | (0 << MUX_F) |  ( 68 << PAD_I) | (1 << PAD_F),
        PIN_GPIO18             = (51 << MUX_I) | (1 << MUX_F) |  ( 67 << PAD_I) | (1 << PAD_F),
        PIN_GPIO19             = (51 << MUX_I) | (0 << MUX_F) |  ( 67 << PAD_I) | (0 << PAD_F),
        PIN_GPIO2              = (55 << MUX_I) | (3 << MUX_F) |  ( 73 << PAD_I) | (1 << PAD_F),
        PIN_GPIO20             = (50 << MUX_I) | (3 << MUX_F) |  ( 66 << PAD_I) | (2 << PAD_F),
        PIN_GPIO21             = (50 << MUX_I) | (2 << MUX_F) |  ( 66 << PAD_I) | (1 << PAD_F),
        PIN_GPIO22             = (50 << MUX_I) | (1 << MUX_F) |  ( 66 << PAD_I) | (0 << PAD_F),
        PIN_GPIO23             = (50 << MUX_I) | (0 << MUX_F) |  ( 65 << PAD_I) | (2 << PAD_F),
        PIN_GPIO24             = (49 << MUX_I) | (3 << MUX_F) |  ( 65 << PAD_I) | (1 << PAD_F),
        PIN_GPIO25             = (49 << MUX_I) | (2 << MUX_F) |  ( 65 << PAD_I) | (0 << PAD_F),
        PIN_GPIO26             = (49 << MUX_I) | (1 << MUX_F) |  ( 64 << PAD_I) | (2 << PAD_F),
        PIN_GPIO27             = (49 << MUX_I) | (0 << MUX_F) |  ( 64 << PAD_I) | (1 << PAD_F),
        PIN_GPIO28             = (68 << MUX_I) | (1 << MUX_F) |  ( 90 << PAD_I) | (0 << PAD_F),
        PIN_GPIO29             = (68 << MUX_I) | (0 << MUX_F) |  ( 89 << PAD_I) | (2 << PAD_F),
        PIN_GPIO3              = (55 << MUX_I) | (2 << MUX_F) |  ( 73 << PAD_I) | (0 << PAD_F),
        PIN_GPIO30             = (67 << MUX_I) | (3 << MUX_F) |  ( 89 << PAD_I) | (1 << PAD_F),
        PIN_GPIO31             = (67 << MUX_I) | (2 << MUX_F) |  ( 89 << PAD_I) | (0 << PAD_F),
        PIN_GPIO32             = (67 << MUX_I) | (1 << MUX_F) |  ( 88 << PAD_I) | (2 << PAD_F),
        PIN_GPIO33             = (67 << MUX_I) | (0 << MUX_F) |  ( 88 << PAD_I) | (1 << PAD_F),
        PIN_GPIO34             = (66 << MUX_I) | (3 << MUX_F) |  ( 88 << PAD_I) | (0 << PAD_F),
        PIN_GPIO35             = (66 << MUX_I) | (2 << MUX_F) |  ( 87 << PAD_I) | (2 << PAD_F),
        PIN_GPIO36             = (51 << MUX_I) | (3 << MUX_F) |  ( 68 << PAD_I) | (0 << PAD_F),
        PIN_GPIO37             = (51 << MUX_I) | (2 << MUX_F) |  ( 67 << PAD_I) | (2 << PAD_F),
        PIN_GPIO4              = (55 << MUX_I) | (1 << MUX_F) |  ( 72 << PAD_I) | (2 << PAD_F),
        PIN_GPIO5              = (55 << MUX_I) | (0 << MUX_F) |  ( 72 << PAD_I) | (1 << PAD_F),
        PIN_GPIO6              = (54 << MUX_I) | (3 << MUX_F) |  ( 72 << PAD_I) | (0 << PAD_F),
        PIN_GPIO7              = (54 << MUX_I) | (2 << MUX_F) |  ( 71 << PAD_I) | (2 << PAD_F),
        PIN_GPIO8              = (54 << MUX_I) | (1 << MUX_F) |  ( 71 << PAD_I) | (1 << PAD_F),
        PIN_GPIO9              = (54 << MUX_I) | (0 << MUX_F) |  ( 71 << PAD_I) | (0 << PAD_F),
        PIN_I2C_CLK            = (26 << MUX_I) | (3 << MUX_F) |  ( 34 << PAD_I) | (2 << PAD_F),
        PIN_I2C_DAT            = (26 << MUX_I) | (2 << MUX_F) |  ( 34 << PAD_I) | (1 << PAD_F),
        PIN_IPU_CONTRAST1      = (33 << MUX_I) | (1 << MUX_F) |  ( 43 << PAD_I) | (1 << PAD_F),
        PIN_IPU_CSI_DATA_0     = (29 << MUX_I) | (2 << MUX_F) |  ( 38 << PAD_I) | (1 << PAD_F),
        PIN_IPU_CSI_DATA_1     = (29 << MUX_I) | (1 << MUX_F) |  ( 38 << PAD_I) | (0 << PAD_F),
        PIN_IPU_CSI_DATA_2     = (29 << MUX_I) | (0 << MUX_F) |  ( 37 << PAD_I) | (2 << PAD_F),
        PIN_IPU_CSI_DATA_3     = (28 << MUX_I) | (3 << MUX_F) |  ( 37 << PAD_I) | (1 << PAD_F),
        PIN_IPU_CSI_DATA_4     = (28 << MUX_I) | (2 << MUX_F) |  ( 37 << PAD_I) | (0 << PAD_F),
        PIN_IPU_CSI_DATA_5     = (28 << MUX_I) | (1 << MUX_F) |  ( 36 << PAD_I) | (2 << PAD_F),
        PIN_IPU_CSI_DATA_6     = (28 << MUX_I) | (0 << MUX_F) |  ( 36 << PAD_I) | (1 << PAD_F),
        PIN_IPU_CSI_DATA_7     = (27 << MUX_I) | (3 << MUX_F) |  ( 36 << PAD_I) | (0 << PAD_F),
        PIN_IPU_CSI_DATA_8     = (27 << MUX_I) | (2 << MUX_F) |  ( 35 << PAD_I) | (2 << PAD_F),
        PIN_IPU_CSI_DATA_9     = (27 << MUX_I) | (1 << MUX_F) |  ( 35 << PAD_I) | (1 << PAD_F),
        PIN_IPU_CSI_HSYNC      = (30 << MUX_I) | (0 << MUX_F) |  ( 39 << PAD_I) | (0 << PAD_F),
        PIN_IPU_CSI_MCLK       = (30 << MUX_I) | (1 << MUX_F) |  ( 39 << PAD_I) | (1 << PAD_F),
        PIN_IPU_CSI_PIXCLK     = (27 << MUX_I) | (0 << MUX_F) |  ( 35 << PAD_I) | (0 << PAD_F),
        PIN_IPU_CSI_VSYNC      = (29 << MUX_I) | (3 << MUX_F) |  ( 38 << PAD_I) | (2 << PAD_F),
        PIN_IPU_D3_CLS         = (31 << MUX_I) | (0 << MUX_F) |  ( 40 << PAD_I) | (1 << PAD_F),
        PIN_IPU_D3_REV         = (30 << MUX_I) | (2 << MUX_F) |  ( 39 << PAD_I) | (2 << PAD_F),
        PIN_IPU_D3_SPL         = (30 << MUX_I) | (3 << MUX_F) |  ( 40 << PAD_I) | (0 << PAD_F),
        PIN_IPU_DRDY0          = (33 << MUX_I) | (2 << MUX_F) |  ( 43 << PAD_I) | (2 << PAD_F),
        PIN_IPU_FDAT_0         = (39 << MUX_I) | (1 << MUX_F) |  ( 51 << PAD_I) | (1 << PAD_F),
        PIN_IPU_FDAT_1         = (39 << MUX_I) | (0 << MUX_F) |  ( 51 << PAD_I) | (0 << PAD_F),
        PIN_IPU_FDAT_10        = (36 << MUX_I) | (3 << MUX_F) |  ( 48 << PAD_I) | (0 << PAD_F),
        PIN_IPU_FDAT_11        = (36 << MUX_I) | (2 << MUX_F) |  ( 47 << PAD_I) | (2 << PAD_F),
        PIN_IPU_FDAT_12        = (36 << MUX_I) | (1 << MUX_F) |  ( 47 << PAD_I) | (1 << PAD_F),
        PIN_IPU_FDAT_13        = (36 << MUX_I) | (0 << MUX_F) |  ( 47 << PAD_I) | (0 << PAD_F),
        PIN_IPU_FDAT_14        = (35 << MUX_I) | (3 << MUX_F) |  ( 46 << PAD_I) | (2 << PAD_F),
        PIN_IPU_FDAT_15        = (35 << MUX_I) | (2 << MUX_F) |  ( 46 << PAD_I) | (1 << PAD_F),
        PIN_IPU_FDAT_16        = (35 << MUX_I) | (1 << MUX_F) |  ( 46 << PAD_I) | (0 << PAD_F),
        PIN_IPU_FDAT_17        = (35 << MUX_I) | (0 << MUX_F) |  ( 45 << PAD_I) | (2 << PAD_F),
        PIN_IPU_FDAT_2         = (38 << MUX_I) | (3 << MUX_F) |  ( 50 << PAD_I) | (2 << PAD_F),
        PIN_IPU_FDAT_3         = (38 << MUX_I) | (2 << MUX_F) |  ( 50 << PAD_I) | (1 << PAD_F),
        PIN_IPU_FDAT_4         = (38 << MUX_I) | (1 << MUX_F) |  ( 50 << PAD_I) | (0 << PAD_F),
        PIN_IPU_FDAT_5         = (38 << MUX_I) | (0 << MUX_F) |  ( 49 << PAD_I) | (2 << PAD_F),
        PIN_IPU_FDAT_6         = (37 << MUX_I) | (3 << MUX_F) |  ( 49 << PAD_I) | (1 << PAD_F),
        PIN_IPU_FDAT_7         = (37 << MUX_I) | (2 << MUX_F) |  ( 49 << PAD_I) | (0 << PAD_F),
        PIN_IPU_FDAT_8         = (37 << MUX_I) | (1 << MUX_F) |  ( 48 << PAD_I) | (2 << PAD_F),
        PIN_IPU_FDAT_9         = (37 << MUX_I) | (0 << MUX_F) |  ( 48 << PAD_I) | (1 << PAD_F),
        PIN_IPU_FPSHIFT        = (33 << MUX_I) | (3 << MUX_F) |  ( 44 << PAD_I) | (0 << PAD_F),
        PIN_IPU_HSYNC          = (34 << MUX_I) | (2 << MUX_F) |  ( 45 << PAD_I) | (0 << PAD_F),
        PIN_IPU_LCS0           = (34 << MUX_I) | (0 << MUX_F) |  ( 44 << PAD_I) | (1 << PAD_F),
        PIN_IPU_LCS1           = (33 << MUX_I) | (0 << MUX_F) |  ( 43 << PAD_I) | (0 << PAD_F),
        PIN_IPU_PAR_RS         = (31 << MUX_I) | (3 << MUX_F) |  ( 41 << PAD_I) | (1 << PAD_F),
        PIN_IPU_READ           = (31 << MUX_I) | (1 << MUX_F) |  ( 40 << PAD_I) | (2 << PAD_F),
        PIN_IPU_SD_D_CLK       = (32 << MUX_I) | (1 << MUX_F) |  ( 42 << PAD_I) | (0 << PAD_F),
        PIN_IPU_SD_D_IO        = (32 << MUX_I) | (3 << MUX_F) |  ( 42 << PAD_I) | (2 << PAD_F),
        PIN_IPU_SD_D_O         = (32 << MUX_I) | (2 << MUX_F) |  ( 42 << PAD_I) | (1 << PAD_F),
        PIN_IPU_SER_RS         = (32 << MUX_I) | (0 << MUX_F) |  ( 41 << PAD_I) | (2 << PAD_F),
        PIN_IPU_VSYNC0         = (34 << MUX_I) | (3 << MUX_F) |  ( 45 << PAD_I) | (1 << PAD_F),
        PIN_IPU_VSYNC3         = (34 << MUX_I) | (1 << MUX_F) |  ( 44 << PAD_I) | (2 << PAD_F),
        PIN_IPU_WRITE          = (31 << MUX_I) | (2 << MUX_F) |  ( 41 << PAD_I) | (0 << PAD_F),
        PIN_IRDA_RX4           = (46 << MUX_I) | (0 << MUX_F) |  ( 60 << PAD_I) | (1 << PAD_F),
        PIN_IRDA_TX4           = (46 << MUX_I) | (1 << MUX_F) |  ( 60 << PAD_I) | (2 << PAD_F),
        PIN_KEY_COL0           = (24 << MUX_I) | (1 << MUX_F) |  ( 31 << PAD_I) | (1 << PAD_F),
        PIN_KEY_COL1           = (24 << MUX_I) | (0 << MUX_F) |  ( 31 << PAD_I) | (0 << PAD_F),
        PIN_KEY_COL2           = (23 << MUX_I) | (3 << MUX_F) |  ( 30 << PAD_I) | (2 << PAD_F),
        PIN_KEY_COL3           = (23 << MUX_I) | (2 << MUX_F) |  ( 30 << PAD_I) | (1 << PAD_F),
        PIN_KEY_COL4           = (23 << MUX_I) | (1 << MUX_F) |  ( 30 << PAD_I) | (0 << PAD_F),
        PIN_KEY_COL5           = (23 << MUX_I) | (0 << MUX_F) |  ( 29 << PAD_I) | (2 << PAD_F),
        PIN_KEY_COL6           = (22 << MUX_I) | (3 << MUX_F) |  ( 29 << PAD_I) | (1 << PAD_F),
        PIN_KEY_COL7           = (22 << MUX_I) | (2 << MUX_F) |  ( 29 << PAD_I) | (0 << PAD_F),
        PIN_KEY_ROW0           = (26 << MUX_I) | (1 << MUX_F) |  ( 34 << PAD_I) | (0 << PAD_F),
        PIN_KEY_ROW1           = (26 << MUX_I) | (0 << MUX_F) |  ( 33 << PAD_I) | (2 << PAD_F),
        PIN_KEY_ROW2           = (25 << MUX_I) | (3 << MUX_F) |  ( 33 << PAD_I) | (1 << PAD_F),
        PIN_KEY_ROW3           = (25 << MUX_I) | (2 << MUX_F) |  ( 33 << PAD_I) | (0 << PAD_F),
        PIN_KEY_ROW4           = (25 << MUX_I) | (1 << MUX_F) |  ( 32 << PAD_I) | (2 << PAD_F),
        PIN_KEY_ROW5           = (25 << MUX_I) | (0 << MUX_F) |  ( 32 << PAD_I) | (1 << PAD_F),
        PIN_KEY_ROW6           = (24 << MUX_I) | (3 << MUX_F) |  ( 32 << PAD_I) | (0 << PAD_F),
        PIN_KEY_ROW7           = (24 << MUX_I) | (2 << MUX_F) |  ( 31 << PAD_I) | (2 << PAD_F),
        PIN_L1T1OUT0           = (62 << MUX_I) | (2 << MUX_F) |  ( 82 << PAD_I) | (1 << PAD_F),
        PIN_L1T1OUT1           = (62 << MUX_I) | (1 << MUX_F) |  ( 82 << PAD_I) | (0 << PAD_F),
        PIN_L1T1OUT2           = (62 << MUX_I) | (0 << MUX_F) |  ( 81 << PAD_I) | (2 << PAD_F),
        PIN_L1T1OUT3           = (61 << MUX_I) | (3 << MUX_F) |  ( 81 << PAD_I) | (1 << PAD_F),
        PIN_L1T1OUT4           = (61 << MUX_I) | (2 << MUX_F) |  ( 81 << PAD_I) | (0 << PAD_F),
        PIN_L1T1OUT5           = (61 << MUX_I) | (1 << MUX_F) |  ( 80 << PAD_I) | (2 << PAD_F),
        PIN_L1T2OUT0           = (61 << MUX_I) | (0 << MUX_F) |  ( 80 << PAD_I) | (1 << PAD_F),
        PIN_L1T2OUT1           = (60 << MUX_I) | (3 << MUX_F) |  ( 80 << PAD_I) | (0 << PAD_F),
        PIN_L1T2OUT2           = (60 << MUX_I) | (2 << MUX_F) |  ( 79 << PAD_I) | (2 << PAD_F),
        PIN_L1T2OUT3           = (60 << MUX_I) | (1 << MUX_F) |  ( 79 << PAD_I) | (1 << PAD_F),
        PIN_LBA_B              = (88 << MUX_I) | (3 << MUX_F) |  (117 << PAD_I) | (1 << PAD_F),
        PIN_MA10               = (105 << MUX_I) | (0 << MUX_F) | (139 << PAD_I) | (0 << PAD_F),
        PIN_MB_REF_CLK_IN      = (79 << MUX_I) | (3 << MUX_F) |  (105 << PAD_I) | (1 << PAD_F),
        PIN_MCKO               = (1 << MUX_I) | (0 << MUX_F) |   (  0 << PAD_I) | (1 << PAD_F),
        PIN_MDO0               = (9 << MUX_I) | (3 << MUX_F) |   ( 12 << PAD_I) | (0 << PAD_F),
        PIN_MDO1               = (9 << MUX_I) | (2 << MUX_F) |   ( 11 << PAD_I) | (2 << PAD_F),
        PIN_MDO10              = (7 << MUX_I) | (1 << MUX_F) |   (  8 << PAD_I) | (2 << PAD_F),
        PIN_MDO11              = (7 << MUX_I) | (0 << MUX_F) |   (  8 << PAD_I) | (1 << PAD_F),
        PIN_MDO12              = (6 << MUX_I) | (3 << MUX_F) |   (  8 << PAD_I) | (0 << PAD_F),
        PIN_MDO13              = (6 << MUX_I) | (2 << MUX_F) |   (  7 << PAD_I) | (2 << PAD_F),
        PIN_MDO14              = (6 << MUX_I) | (1 << MUX_F) |   (  7 << PAD_I) | (1 << PAD_F),
        PIN_MDO15              = (6 << MUX_I) | (0 << MUX_F) |   (  7 << PAD_I) | (0 << PAD_F),
        PIN_MDO16              = (5 << MUX_I) | (3 << MUX_F) |   (  6 << PAD_I) | (2 << PAD_F),
        PIN_MDO17              = (5 << MUX_I) | (2 << MUX_F) |   (  6 << PAD_I) | (1 << PAD_F),
        PIN_MDO18              = (5 << MUX_I) | (1 << MUX_F) |   (  6 << PAD_I) | (0 << PAD_F),
        PIN_MDO19              = (5 << MUX_I) | (0 << MUX_F) |   (  5 << PAD_I) | (2 << PAD_F),
        PIN_MDO2               = (9 << MUX_I) | (1 << MUX_F) |   ( 11 << PAD_I) | (1 << PAD_F),
        PIN_MDO20              = (4 << MUX_I) | (3 << MUX_F) |   (  5 << PAD_I) | (1 << PAD_F),
        PIN_MDO21              = (4 << MUX_I) | (2 << MUX_F) |   (  5 << PAD_I) | (0 << PAD_F),
        PIN_MDO22              = (4 << MUX_I) | (1 << MUX_F) |   (  4 << PAD_I) | (2 << PAD_F),
        PIN_MDO23              = (4 << MUX_I) | (0 << MUX_F) |   (  4 << PAD_I) | (1 << PAD_F),
        PIN_MDO24              = (3 << MUX_I) | (3 << MUX_F) |   (  4 << PAD_I) | (0 << PAD_F),
        PIN_MDO25              = (3 << MUX_I) | (2 << MUX_F) |   (  3 << PAD_I) | (2 << PAD_F),
        PIN_MDO26              = (3 << MUX_I) | (1 << MUX_F) |   (  3 << PAD_I) | (1 << PAD_F),
        PIN_MDO27              = (3 << MUX_I) | (0 << MUX_F) |   (  3 << PAD_I) | (0 << PAD_F),
        PIN_MDO28              = (2 << MUX_I) | (3 << MUX_F) |   (  2 << PAD_I) | (2 << PAD_F),
        PIN_MDO29              = (2 << MUX_I) | (2 << MUX_F) |   (  2 << PAD_I) | (1 << PAD_F),
        PIN_MDO3               = (9 << MUX_I) | (0 << MUX_F) |   ( 11 << PAD_I) | (0 << PAD_F),
        PIN_MDO30              = (2 << MUX_I) | (1 << MUX_F) |   (  2 << PAD_I) | (0 << PAD_F),
        PIN_MDO31              = (2 << MUX_I) | (0 << MUX_F) |   (  1 << PAD_I) | (2 << PAD_F),
        PIN_MDO4               = (8 << MUX_I) | (3 << MUX_F) |   ( 10 << PAD_I) | (2 << PAD_F),
        PIN_MDO5               = (8 << MUX_I) | (2 << MUX_F) |   ( 10 << PAD_I) | (1 << PAD_F),
        PIN_MDO6               = (8 << MUX_I) | (1 << MUX_F) |   ( 10 << PAD_I) | (0 << PAD_F),
        PIN_MDO7               = (8 << MUX_I) | (0 << MUX_F) |   (  9 << PAD_I) | (2 << PAD_F),
        PIN_MDO8               = (7 << MUX_I) | (3 << MUX_F) |   (  9 << PAD_I) | (1 << PAD_F),
        PIN_MDO9               = (7 << MUX_I) | (2 << MUX_F) |   (  9 << PAD_I) | (0 << PAD_F),
        PIN_MGRNT              = (92 << MUX_I) | (2 << MUX_F) |  (122 << PAD_I) | (1 << PAD_F),
        PIN_MMC1_CLK           = (21 << MUX_I) | (0 << MUX_F) |  ( 27 << PAD_I) | (0 << PAD_F),
        PIN_MMC1_CMD           = (22 << MUX_I) | (1 << MUX_F) |  ( 28 << PAD_I) | (2 << PAD_F),
        PIN_MMC1_DATA_0        = (22 << MUX_I) | (0 << MUX_F) |  ( 28 << PAD_I) | (1 << PAD_F),
        PIN_MMC1_DATA_1        = (21 << MUX_I) | (3 << MUX_F) |  ( 28 << PAD_I) | (0 << PAD_F),
        PIN_MMC1_DATA_2        = (21 << MUX_I) | (2 << MUX_F) |  ( 27 << PAD_I) | (2 << PAD_F),
        PIN_MMC1_DATA_3        = (21 << MUX_I) | (1 << MUX_F) |  ( 27 << PAD_I) | (1 << PAD_F),
        PIN_MMC1_SDDET         = (20 << MUX_I) | (3 << MUX_F) |  ( 26 << PAD_I) | (2 << PAD_F),
        PIN_MMC2_CLK           = (19 << MUX_I) | (1 << MUX_F) |  ( 24 << PAD_I) | (2 << PAD_F),
        PIN_MMC2_CMD           = (20 << MUX_I) | (2 << MUX_F) |  ( 26 << PAD_I) | (1 << PAD_F),
        PIN_MMC2_DATA_0        = (20 << MUX_I) | (1 << MUX_F) |  ( 26 << PAD_I) | (0 << PAD_F),
        PIN_MMC2_DATA_1        = (20 << MUX_I) | (0 << MUX_F) |  ( 25 << PAD_I) | (2 << PAD_F),
        PIN_MMC2_DATA_2        = (19 << MUX_I) | (3 << MUX_F) |  ( 25 << PAD_I) | (1 << PAD_F),
        PIN_MMC2_DATA_3        = (19 << MUX_I) | (2 << MUX_F) |  ( 25 << PAD_I) | (0 << PAD_F),
        PIN_MQSPI1_CK          = (60 << MUX_I) | (0 << MUX_F) |  ( 79 << PAD_I) | (0 << PAD_F),
        PIN_MQSPI1_CS0         = (59 << MUX_I) | (1 << MUX_F) |  ( 78 << PAD_I) | (0 << PAD_F),
        PIN_MQSPI1_CS1         = (59 << MUX_I) | (0 << MUX_F) |  ( 77 << PAD_I) | (2 << PAD_F),
        PIN_MQSPI1_CS2         = (58 << MUX_I) | (3 << MUX_F) |  ( 77 << PAD_I) | (1 << PAD_F),
        PIN_MQSPI1_DI          = (59 << MUX_I) | (3 << MUX_F) |  ( 78 << PAD_I) | (2 << PAD_F),
        PIN_MQSPI1_DO          = (59 << MUX_I) | (2 << MUX_F) |  ( 78 << PAD_I) | (1 << PAD_F),
        PIN_MQSPI2_CK          = (58 << MUX_I) | (2 << MUX_F) |  ( 77 << PAD_I) | (0 << PAD_F),
        PIN_MQSPI2_CS0         = (57 << MUX_I) | (3 << MUX_F) |  ( 76 << PAD_I) | (0 << PAD_F),
        PIN_MQSPI2_CS1         = (57 << MUX_I) | (2 << MUX_F) |  ( 75 << PAD_I) | (2 << PAD_F),
        PIN_MQSPI2_DI          = (58 << MUX_I) | (1 << MUX_F) |  ( 76 << PAD_I) | (2 << PAD_F),
        PIN_MQSPI2_DO          = (58 << MUX_I) | (0 << MUX_F) |  ( 76 << PAD_I) | (1 << PAD_F),
        PIN_MRDY               = (1 << MUX_I) | (1 << MUX_F) |   (  0 << PAD_I) | (2 << PAD_F),
        PIN_MRQST              = (92 << MUX_I) | (1 << MUX_F) |  (122 << PAD_I) | (0 << PAD_F),
        PIN_MSE0               = (0 << MUX_I) | (3 << MUX_F) |   (  0 << PAD_I) | (0 << PAD_F),
        PIN_MSE1               = (1 << MUX_I) | (3 << MUX_F) |   (  1 << PAD_I) | (1 << PAD_F),
        PIN_OE_B               = (91 << MUX_I) | (0 << MUX_F) |  (120 << PAD_I) | (1 << PAD_F),
        PIN_PAD0               = (0 << MUX_I) | (2 << MUX_F),
        PIN_PAD2               = (0 << MUX_I) | (0 << MUX_F),
        PIN_PAD3               = (0 << MUX_I) | (1 << MUX_F),
        PIN_PM_INT             = (77 << MUX_I) | (1 << MUX_F) |  (102 << PAD_I) | (0 << PAD_F),
        PIN_POWER_FAIL         = (77 << MUX_I) | (0 << MUX_F) |  (101 << PAD_I) | (2 << PAD_F),
        PIN_PWM                = (42 << MUX_I) | (0 << MUX_F) |  ( 55 << PAD_I) | (0 << PAD_F),
        PIN_R_B                = (84 << MUX_I) | (2 << MUX_F) |  (111 << PAD_I) | (2 << PAD_F),
        PIN_RAS                = (87 << MUX_I) | (2 << MUX_F) |  (115 << PAD_I) | (2 << PAD_F),
        PIN_RE                 = (85 << MUX_I) | (0 << MUX_F) |  (112 << PAD_I) | (1 << PAD_F),
        PIN_REF_EN_B           = (77 << MUX_I) | (2 << MUX_F) |  (102 << PAD_I) | (1 << PAD_F),
        PIN_RESET_IN_B         = (78 << MUX_I) | (3 << MUX_F) |  (104 << PAD_I) | (0 << PAD_F),
        PIN_RESET_POR_IN_B     = (78 << MUX_I) | (2 << MUX_F) |  (103 << PAD_I) | (2 << PAD_F),
        PIN_RTCK               = (75 << MUX_I) | (2 << MUX_F) |  ( 99 << PAD_I) | (2 << PAD_F),
        PIN_RW_B               = (88 << MUX_I) | (1 << MUX_F) |  (116 << PAD_I) | (2 << PAD_F),
        PIN_SD0                = (101 << MUX_I) | (0 << MUX_F) | (133 << PAD_I) | (2 << PAD_F),
        PIN_SD1                = (100 << MUX_I) | (3 << MUX_F) | (133 << PAD_I) | (1 << PAD_F),
        PIN_SD10               = (98 << MUX_I) | (2 << MUX_F) |  (130 << PAD_I) | (1 << PAD_F),
        PIN_SD11               = (98 << MUX_I) | (1 << MUX_F) |  (130 << PAD_I) | (0 << PAD_F),
        PIN_SD12               = (98 << MUX_I) | (0 << MUX_F) |  (129 << PAD_I) | (2 << PAD_F),
        PIN_SD13               = (97 << MUX_I) | (3 << MUX_F) |  (129 << PAD_I) | (1 << PAD_F),
        PIN_SD14               = (97 << MUX_I) | (2 << MUX_F) |  (129 << PAD_I) | (0 << PAD_F),
        PIN_SD15               = (97 << MUX_I) | (1 << MUX_F) |  (128 << PAD_I) | (2 << PAD_F),
        PIN_SD16               = (97 << MUX_I) | (0 << MUX_F) |  (128 << PAD_I) | (1 << PAD_F),
        PIN_SD17               = (96 << MUX_I) | (3 << MUX_F) |  (128 << PAD_I) | (0 << PAD_F),
        PIN_SD18               = (96 << MUX_I) | (2 << MUX_F) |  (127 << PAD_I) | (2 << PAD_F),
        PIN_SD19               = (96 << MUX_I) | (1 << MUX_F) |  (127 << PAD_I) | (1 << PAD_F),
        PIN_SD2                = (100 << MUX_I) | (2 << MUX_F) | (133 << PAD_I) | (0 << PAD_F),
        PIN_SD20               = (96 << MUX_I) | (0 << MUX_F) |  (127 << PAD_I) | (0 << PAD_F),
        PIN_SD21               = (95 << MUX_I) | (3 << MUX_F) |  (126 << PAD_I) | (2 << PAD_F),
        PIN_SD22               = (95 << MUX_I) | (2 << MUX_F) |  (126 << PAD_I) | (1 << PAD_F),
        PIN_SD23               = (95 << MUX_I) | (1 << MUX_F) |  (126 << PAD_I) | (0 << PAD_F),
        PIN_SD24               = (95 << MUX_I) | (0 << MUX_F) |  (125 << PAD_I) | (2 << PAD_F),
        PIN_SD25               = (94 << MUX_I) | (3 << MUX_F) |  (125 << PAD_I) | (1 << PAD_F),
        PIN_SD26               = (94 << MUX_I) | (2 << MUX_F) |  (125 << PAD_I) | (0 << PAD_F),
        PIN_SD27               = (94 << MUX_I) | (1 << MUX_F) |  (124 << PAD_I) | (2 << PAD_F),
        PIN_SD28               = (94 << MUX_I) | (0 << MUX_F) |  (124 << PAD_I) | (1 << PAD_F),
        PIN_SD29               = (93 << MUX_I) | (3 << MUX_F) |  (124 << PAD_I) | (0 << PAD_F),
        PIN_SD3                = (100 << MUX_I) | (1 << MUX_F) | (132 << PAD_I) | (2 << PAD_F),
        PIN_SD30               = (93 << MUX_I) | (2 << MUX_F) |  (123 << PAD_I) | (2 << PAD_F),
        PIN_SD31               = (93 << MUX_I) | (1 << MUX_F) |  (123 << PAD_I) | (1 << PAD_F),
        PIN_SD4                = (100 << MUX_I) | (0 << MUX_F) | (132 << PAD_I) | (1 << PAD_F),
        PIN_SD5                = (99 << MUX_I) | (3 << MUX_F) |  (132 << PAD_I) | (0 << PAD_F),
        PIN_SD6                = (99 << MUX_I) | (2 << MUX_F) |  (131 << PAD_I) | (2 << PAD_F),
        PIN_SD7                = (99 << MUX_I) | (1 << MUX_F) |  (131 << PAD_I) | (1 << PAD_F),
        PIN_SD8                = (99 << MUX_I) | (0 << MUX_F) |  (131 << PAD_I) | (0 << PAD_F),
        PIN_SD9                = (98 << MUX_I) | (3 << MUX_F) |  (130 << PAD_I) | (2 << PAD_F),
        PIN_SDBA0              = (89 << MUX_I) | (2 << MUX_F) |  (118 << PAD_I) | (1 << PAD_F),
        PIN_SDBA1              = (89 << MUX_I) | (1 << MUX_F) |  (118 << PAD_I) | (0 << PAD_F),
        PIN_SDCKE0             = (86 << MUX_I) | (3 << MUX_F) |  (114 << PAD_I) | (2 << PAD_F),
        PIN_SDCKE1             = (86 << MUX_I) | (2 << MUX_F) |  (114 << PAD_I) | (1 << PAD_F),
        PIN_SDCLK              = (86 << MUX_I) | (1 << MUX_F) |  (114 << PAD_I) | (0 << PAD_F),
        PIN_SDCLK_B            = (86 << MUX_I) | (0 << MUX_F) |  (113 << PAD_I) | (2 << PAD_F),
        PIN_SDMA_EVNT0         = (66 << MUX_I) | (1 << MUX_F) |  ( 87 << PAD_I) | (1 << PAD_F),
        PIN_SDMA_EVNT1         = (66 << MUX_I) | (0 << MUX_F) |  ( 87 << PAD_I) | (0 << PAD_F),
        PIN_SDQS0              = (93 << MUX_I) | (0 << MUX_F) |  (123 << PAD_I) | (0 << PAD_F),
        PIN_SDQS1              = (92 << MUX_I) | (3 << MUX_F) |  (122 << PAD_I) | (2 << PAD_F),
        PIN_SDWE               = (87 << MUX_I) | (0 << MUX_F) |  (115 << PAD_I) | (0 << PAD_F),
        PIN_SIM1_CLK0          = (74 << MUX_I) | (0 << MUX_F) |  ( 97 << PAD_I) | (2 << PAD_F),
        PIN_SIM1_DATA0_RX_TX   = (73 << MUX_I) | (2 << MUX_F) |  ( 97 << PAD_I) | (0 << PAD_F),
        PIN_SIM1_RST0          = (73 << MUX_I) | (3 << MUX_F) |  ( 97 << PAD_I) | (1 << PAD_F),
        PIN_SIM1_SIMPD0        = (73 << MUX_I) | (0 << MUX_F) |  ( 96 << PAD_I) | (1 << PAD_F),
        PIN_SIM1_SVEN0         = (73 << MUX_I) | (1 << MUX_F) |  ( 96 << PAD_I) | (2 << PAD_F),
        PIN_SJC_MOD            = (74 << MUX_I) | (3 << MUX_F) |  ( 98 << PAD_I) | (2 << PAD_F),
        PIN_SRST               = (10 << MUX_I) | (2 << MUX_F) |  ( 13 << PAD_I) | (0 << PAD_F),
        PIN_TCK                = (75 << MUX_I) | (1 << MUX_F) |  ( 99 << PAD_I) | (1 << PAD_F),
        PIN_TDI                = (75 << MUX_I) | (3 << MUX_F) |  (100 << PAD_I) | (0 << PAD_F),
        PIN_TDO                = (76 << MUX_I) | (0 << MUX_F) |  (100 << PAD_I) | (1 << PAD_F),
        PIN_TEST_MODE          = (78 << MUX_I) | (1 << MUX_F) |  (103 << PAD_I) | (1 << PAD_F),
        PIN_TMS                = (75 << MUX_I) | (0 << MUX_F) |  ( 99 << PAD_I) | (0 << PAD_F),
        PIN_TRACE0             = (19 << MUX_I) | (0 << MUX_F) |  ( 24 << PAD_I) | (1 << PAD_F),
        PIN_TRACE1             = (18 << MUX_I) | (3 << MUX_F) |  ( 24 << PAD_I) | (0 << PAD_F),
        PIN_TRACE10            = (16 << MUX_I) | (2 << MUX_F) |  ( 21 << PAD_I) | (0 << PAD_F),
        PIN_TRACE11            = (16 << MUX_I) | (1 << MUX_F) |  ( 20 << PAD_I) | (2 << PAD_F),
        PIN_TRACE12            = (16 << MUX_I) | (0 << MUX_F) |  ( 20 << PAD_I) | (1 << PAD_F),
        PIN_TRACE13            = (15 << MUX_I) | (3 << MUX_F) |  ( 20 << PAD_I) | (0 << PAD_F),
        PIN_TRACE14            = (15 << MUX_I) | (2 << MUX_F) |  ( 19 << PAD_I) | (2 << PAD_F),
        PIN_TRACE15            = (15 << MUX_I) | (1 << MUX_F) |  ( 19 << PAD_I) | (1 << PAD_F),
        PIN_TRACE16            = (15 << MUX_I) | (0 << MUX_F) |  ( 19 << PAD_I) | (0 << PAD_F),
        PIN_TRACE17            = (14 << MUX_I) | (3 << MUX_F) |  ( 18 << PAD_I) | (2 << PAD_F),
        PIN_TRACE18            = (14 << MUX_I) | (2 << MUX_F) |  ( 18 << PAD_I) | (1 << PAD_F),
        PIN_TRACE19            = (14 << MUX_I) | (1 << MUX_F) |  ( 18 << PAD_I) | (0 << PAD_F),
        PIN_TRACE2             = (18 << MUX_I) | (2 << MUX_F) |  ( 23 << PAD_I) | (2 << PAD_F),
        PIN_TRACE20            = (14 << MUX_I) | (0 << MUX_F) |  ( 17 << PAD_I) | (2 << PAD_F),
        PIN_TRACE21            = (13 << MUX_I) | (3 << MUX_F) |  ( 17 << PAD_I) | (1 << PAD_F),
        PIN_TRACE22            = (13 << MUX_I) | (2 << MUX_F) |  ( 17 << PAD_I) | (0 << PAD_F),
        PIN_TRACE23            = (13 << MUX_I) | (1 << MUX_F) |  ( 16 << PAD_I) | (2 << PAD_F),
        PIN_TRACE24            = (13 << MUX_I) | (0 << MUX_F) |  ( 16 << PAD_I) | (1 << PAD_F),
        PIN_TRACE25            = (12 << MUX_I) | (3 << MUX_F) |  ( 16 << PAD_I) | (0 << PAD_F),
        PIN_TRACE26            = (12 << MUX_I) | (2 << MUX_F) |  ( 15 << PAD_I) | (2 << PAD_F),
        PIN_TRACE27            = (12 << MUX_I) | (1 << MUX_F) |  ( 15 << PAD_I) | (1 << PAD_F),
        PIN_TRACE28            = (12 << MUX_I) | (0 << MUX_F) |  ( 15 << PAD_I) | (0 << PAD_F),
        PIN_TRACE29            = (11 << MUX_I) | (3 << MUX_F) |  ( 14 << PAD_I) | (2 << PAD_F),
        PIN_TRACE3             = (18 << MUX_I) | (1 << MUX_F) |  ( 23 << PAD_I) | (1 << PAD_F),
        PIN_TRACE30            = (11 << MUX_I) | (2 << MUX_F) |  ( 14 << PAD_I) | (1 << PAD_F),
        PIN_TRACE31            = (11 << MUX_I) | (1 << MUX_F) |  ( 14 << PAD_I) | (0 << PAD_F),
        PIN_TRACE4             = (18 << MUX_I) | (0 << MUX_F) |  ( 23 << PAD_I) | (0 << PAD_F),
        PIN_TRACE5             = (17 << MUX_I) | (3 << MUX_F) |  ( 22 << PAD_I) | (2 << PAD_F),
        PIN_TRACE6             = (17 << MUX_I) | (2 << MUX_F) |  ( 22 << PAD_I) | (1 << PAD_F),
        PIN_TRACE7             = (17 << MUX_I) | (1 << MUX_F) |  ( 22 << PAD_I) | (0 << PAD_F),
        PIN_TRACE8             = (17 << MUX_I) | (0 << MUX_F) |  ( 21 << PAD_I) | (2 << PAD_F),
        PIN_TRACE9             = (16 << MUX_I) | (3 << MUX_F) |  ( 21 << PAD_I) | (1 << PAD_F),
        PIN_TRCLK              = (10 << MUX_I) | (3 << MUX_F) |  ( 13 << PAD_I) | (1 << PAD_F),
        PIN_TRSTB              = (76 << MUX_I) | (1 << MUX_F) |  (100 << PAD_I) | (2 << PAD_F),
        PIN_UART_CTS1_B        = (48 << MUX_I) | (0 << MUX_F) |  ( 63 << PAD_I) | (0 << PAD_F),
        PIN_UART_CTS2_B        = (47 << MUX_I) | (3 << MUX_F) |  ( 62 << PAD_I) | (2 << PAD_F),
        PIN_UART_CTS3_B        = (56 << MUX_I) | (2 << MUX_F) |  ( 74 << PAD_I) | (1 << PAD_F),
        PIN_UART_DCD2_B        = (47 << MUX_I) | (0 << MUX_F) |  ( 61 << PAD_I) | (2 << PAD_F),
        PIN_UART_DSR2_B        = (46 << MUX_I) | (3 << MUX_F) |  ( 61 << PAD_I) | (1 << PAD_F),
        PIN_UART_DTR2_B        = (47 << MUX_I) | (2 << MUX_F) |  ( 62 << PAD_I) | (1 << PAD_F),
        PIN_UART_MUXCTL        = (46 << MUX_I) | (2 << MUX_F) |  ( 61 << PAD_I) | (0 << PAD_F),
        PIN_UART_RI2_B         = (47 << MUX_I) | (1 << MUX_F) |  ( 62 << PAD_I) | (0 << PAD_F),
        PIN_UART_RTS1_B        = (48 << MUX_I) | (1 << MUX_F) |  ( 63 << PAD_I) | (1 << PAD_F),
        PIN_UART_RTS3_B        = (56 << MUX_I) | (3 << MUX_F) |  ( 74 << PAD_I) | (2 << PAD_F),
        PIN_UART_RXD1          = (48 << MUX_I) | (2 << MUX_F) |  ( 63 << PAD_I) | (2 << PAD_F),
        PIN_UART_RXD3          = (57 << MUX_I) | (0 << MUX_F) |  ( 75 << PAD_I) | (0 << PAD_F),
        PIN_UART_TXD1          = (48 << MUX_I) | (3 << MUX_F) |  ( 64 << PAD_I) | (0 << PAD_F),
        PIN_UART_TXD3          = (57 << MUX_I) | (1 << MUX_F) |  ( 75 << PAD_I) | (1 << PAD_F),
        PIN_USB_TXENB          = (71 << MUX_I) | (0 << MUX_F) |  ( 93 << PAD_I) | (2 << PAD_F),
        PIN_USB_VMIN           = (70 << MUX_I) | (1 << MUX_F) |  ( 92 << PAD_I) | (2 << PAD_F),
        PIN_USB_VMOUT          = (71 << MUX_I) | (1 << MUX_F) |  ( 94 << PAD_I) | (0 << PAD_F),
        PIN_USB_VPIN           = (70 << MUX_I) | (3 << MUX_F) |  ( 93 << PAD_I) | (1 << PAD_F),
        PIN_USB_VPOUT          = (70 << MUX_I) | (2 << MUX_F) |  ( 93 << PAD_I) | (0 << PAD_F),
        PIN_USB_XRXD           = (70 << MUX_I) | (0 << MUX_F) |  ( 92 << PAD_I) | (1 << PAD_F),
        PIN_USB1_TXENB         = (72 << MUX_I) | (2 << MUX_F) |  ( 95 << PAD_I) | (2 << PAD_F),
        PIN_USB1_VMIN          = (71 << MUX_I) | (3 << MUX_F) |  ( 94 << PAD_I) | (2 << PAD_F),
        PIN_USB1_VMOUT         = (72 << MUX_I) | (3 << MUX_F) |  ( 96 << PAD_I) | (0 << PAD_F),
        PIN_USB1_VPIN          = (72 << MUX_I) | (1 << MUX_F) |  ( 95 << PAD_I) | (1 << PAD_F),
        PIN_USB1_VPOUT         = (72 << MUX_I) | (0 << MUX_F) |  ( 95 << PAD_I) | (0 << PAD_F),
        PIN_USB1_XRXD          = (71 << MUX_I) | (2 << MUX_F) |  ( 94 << PAD_I) | (1 << PAD_F),
        PIN_USER_OFF           = (76 << MUX_I) | (3 << MUX_F) |  (101 << PAD_I) | (1 << PAD_F),
        PIN_WAMMO_DIV_RX_FRAME = (63 << MUX_I) | (3 << MUX_F) |  ( 84 << PAD_I) | (0 << PAD_F),
        PIN_WAMMO_DIV_RX_I_CH  = (64 << MUX_I) | (1 << MUX_F) |  ( 84 << PAD_I) | (2 << PAD_F),
        PIN_WAMMO_DIV_RX_Q_CH  = (64 << MUX_I) | (0 << MUX_F) |  ( 84 << PAD_I) | (1 << PAD_F),
        PIN_WAMMO_RX_FRAME     = (64 << MUX_I) | (2 << MUX_F) |  ( 85 << PAD_I) | (0 << PAD_F),
        PIN_WAMMO_RX_I_CH      = (65 << MUX_I) | (0 << MUX_F) |  ( 85 << PAD_I) | (2 << PAD_F),
        PIN_WAMMO_RX_Q_CH      = (65 << MUX_I) | (1 << MUX_F) |  ( 86 << PAD_I) | (0 << PAD_F),
        PIN_WAMMO_TX_FRAME     = (64 << MUX_I) | (3 << MUX_F) |  ( 85 << PAD_I) | (1 << PAD_F),
        PIN_WAMMO_TX_I_CH      = (65 << MUX_I) | (2 << MUX_F) |  ( 86 << PAD_I) | (1 << PAD_F),
        PIN_WAMMO_TX_Q_CH      = (65 << MUX_I) | (3 << MUX_F) |  ( 86 << PAD_I) | (2 << PAD_F),
        PIN_WB_REF_CLK_IN      = (80 << MUX_I) | (0 << MUX_F) |  (105 << PAD_I) | (2 << PAD_F),
        PIN_WDOG_B             = (76 << MUX_I) | (2 << MUX_F) |  (101 << PAD_I) | (0 << PAD_F),
        PIN_WE                 = (84 << MUX_I) | (3 << MUX_F) |  (112 << PAD_I) | (0 << PAD_F),
        PIN_WP                 = (85 << MUX_I) | (2 << MUX_F) |  (113 << PAD_I) | (0 << PAD_F),
};

/*
 * TODO: wait for ArgonPlus IC Spec update. Now it is not consistent.
 */
enum iopad_pins {
        DUMMYDUMMY,
};

/*!
 * various IOMUX output functions
 */
enum iomux_output_config {
        OUTPUTCONFIG_GPIO = 0,   /*!< used as GPIO */
        OUTPUTCONFIG_FUNC,       /*!< used as function */
        OUTPUTCONFIG_ALT1,       /*!< used as alternate function 1 */
        OUTPUTCONFIG_ALT2,       /*!< used as alternate function 2 */
        OUTPUTCONFIG_ALT3,       /*!< used as alternate function 3 */
        OUTPUTCONFIG_ALT4,       /*!< used as alternate function 4 */
        OUTPUTCONFIG_ALT5,       /*!< used as alternate function 5 */
        OUTPUTCONFIG_ALT6        /*!< used as alternate function 6 */
};

/*!
 * various IOMUX input functions
 */
enum iomux_input_config {
        INPUTCONFIG_NONE = 0,            /*!< not configured for input */
        INPUTCONFIG_GPIO = 1 << 0,  /*!< used as GPIO */
        INPUTCONFIG_FUNC = 1 << 1,  /*!< used as function */
        INPUTCONFIG_ALT1 = 1 << 2,  /*!< used as alternate function 1 */
        INPUTCONFIG_ALT2 = 1 << 3   /*!< used as alternate function 2 */
};

/*!
 * various IOMUX pad functions
 */
enum iomux_pad_config {
        PAD_CTL_NOLOOPBACK      = 0x0 << 8,
        PAD_CTL_LOOPBACK        = 0x1 << 8,
        PAD_CTL_PKE_NONE        = 0x0 << 7,
        PAD_CTL_PKE_ENABLE      = 0x1 << 7,
        PAD_CTL_PUE_KEEPER      = 0x0 << 6,
        PAD_CTL_PUE_PUD         = 0x1 << 6,
        PAD_CTL_100K_PD         = 0x0 << 4,
        PAD_CTL_100K_PU         = 0x1 << 4,
        PAD_CTL_47K_PU          = 0x2 << 4,
        PAD_CTL_22K_PU          = 0x3 << 4,
        PAD_CTL_HYS_CMOS        = 0x0 << 3,
        PAD_CTL_HYS_SCHMITZ     = 0x1 << 3,
        PAD_CTL_ODE_CMOS        = 0x0 << 2,
        PAD_CTL_ODE_OpenDrain   = 0x1 << 2,
        PAD_CTL_DSE0_NORMAL     = 0x0 << 1,
        PAD_CTL_DSE0_HIGH       = 0x1 << 1,
        PAD_CTL_SRE_SLOW        = 0x0 << 0,
        PAD_CTL_SRE_FAST        = 0x1 << 0
};

/*!
 * various IOMUX general purpose functions
 */
enum iomux_gp_func {
        MUX_WCSI_DIVERSITY_MODE              = 0x1 << 3,
        MUX_WCSI_PGP_REG_RX_WB_CORR_CTRL0    = 0x1 << 4,
        MUX_WCSI_PGP_REG_RX_WB_CORR_CTRL1    = 0x1 << 5,
        MUX_WCSI_PGP_REG_RX_ALT_WB_QSEL0     = 0x1 << 6,
        MUX_WCSI_PGP_REG_RX_ALT_WB_QSEL1     = 0x1 << 7,
        MUX_WCSI_PGP_REG_RX_ALT_WB_QSEL2     = 0x1 << 8,
        MUX_WCSI_PGP_REG_RX_ALT_WB_ISEL0     = 0x1 << 9,
        MUX_WCSI_PGP_REG_RX_ALT_WB_ISEL1     = 0x1 << 10,
        MUX_WCSI_PGP_REG_RX_ALT_WB_ISEL2     = 0x1 << 11,
        MUX_WCSI_PGP_REG_RX_ALT_WB_EN        = 0x1 << 12,
        MUX_WCSI_PGP_REG_BBIF_MODE           = 0x1 << 13,
        MUX_PGP_WCSI_TX_DIS_FRM_RANDOM       = 0x1 << 14,
        MUX_PGP_USB2_VPIN                    = 0x1 << 15,
        MUX_PGP_UART3_RXD                    = 0x1 << 16,
        MUX_SDCTL_SPBA_SEL                   = 0x1 << 17,
        MUX_NFC_CE_SEL                       = 0x1 << 18,
        MUX_PGP_FIRI                         = 0x1 << 19,
        MUX_SDCTL_CSD1_SEL_B                 = 0x1 << 20,
        MUX_SDCTL_CSD0_SEL_B                 = 0x1 << 21
};

typedef irqreturn_t (*gpio_irq_handler)(int, void *, struct pt_regs *);

/*!
 * This function is used to configure a pin through the IOMUX module.
 *
 * @param  pin          pin number as defined in \b #iomux_pins
 * @param  out          output function as defined in \b #iomux_output_config
 * @param  in           input function as defined in \b #iomux_input_config
 */
void iomux_config_mux(enum iomux_pins pin, 
                      enum iomux_output_config out, 
                      enum iomux_input_config in);

/*!
 * This function configures the pad value for a IOMUX pin.
 *
 * @param  pin          a pin number as defined in \b #iomux_pins
 * @param  config       ORed value of elements defined in \b #iomux_pad_config
 */
void iomux_config_pad(enum iomux_pins pin, __u32 config);

/*!
 * This function enables/disables the general purpose function for a particular
 * signal.
 *
 * @param  gp   one signal as defined in \b enum \b #iomux_gp_func
 * @param  en   \b #true to enable; \b #false to disable
 */
void iomux_config_gpr(enum iomux_gp_func gp, bool en);

/* gpio related defines */

/*!
 * There are two queues for registered GPIO ISRs. One is for high priority and
 * the other is for low priority. The ISRs in the high priority queue will be
 * called first before the low priority queue if more than one GPIO interrupt
 * occurs at the same time.
 */
enum gpio_prio {
        GPIO_HIGH_PRIO  = 0,     /*!< high priority queue */
        GPIO_LOW_PRIO            /*!< low priority queue */
};

/*!
 * This enumeration data type defines various different ways for interrupting 
 * the ARM core from GPIO signals. The way to interrupt the core is dictated 
 * by the external hardware.
 */
enum gpio_int_cfg {
        GPIO_INT_LOW_LEV        = 0x0,    /*!< low level sensitive */
        GPIO_INT_HIGH_LEV       = 0x1,    /*!< high level sensitive */
        GPIO_INT_RISE_EDGE      = 0x2,    /*!< rising edge sensitive */
        GPIO_INT_FALL_EDGE      = 0x3,    /*!< falling edge sensitive */
        GPIO_INT_NONE           = 0x4     /*!< No interrupt */
};

/*!
 * This enumeration data type defines various different ways for interrupting 
 * the ARM core from EDIO signals. The way to interrupt the core is dictated 
 * by the external hardware.
 */
enum edio_int_cfg {
        EDIO_INT_LOW_LEV        = 0x0,    /*!< low level sensitive */
        EDIO_INT_RISE_EDGE      = 0x1,    /*!< rise edge sensitive */
        EDIO_INT_FALL_EDGE      = 0x2,    /*!< falling edge sensitive */
        EDIO_INT_BOTH_EDGE      = 0x3,    /*!< both edge sensitive */
        EDIO_INT_NONE           = 0x4     /*!< No interrupt */
};

/*!
 * This enumeration data type defines various signals for interrupting the
 * ARM core from EDIO signals. 
 */
enum edio_sig_no {
        ED_INT0			= 0,		    
        ED_INT1			= 1,
        ED_INT2			= 2,
        ED_INT3			= 3,
        ED_INT4			= 4,
        ED_INT5			= 5,
        ED_INT6			= 6,
        ED_INT7			= 7,
};

/*!
 * This function configures the GPIO signal to be either input or output. For
 * input signals used for generating interrupts for the ARM core, how the
 * interrupts being triggered is also passed in via \a icr. For output signals,
 * the \a icr value doesn't matter.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  out          #true for output, #false for input
 * @param  icr          value defined in \b #gpio_int_cfg
 */
void gpio_config(__u32 port, __u32 sig_no, bool out, enum gpio_int_cfg icr);

/*!
 * This function sets a GPIO signal value.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  data         value to be set (only 0 or 1 is valid)
 */
void gpio_set_data(__u32 port, __u32 sig_no, __u32 data);

/*!
 * This function returns the value of the GPIO signal.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 *
 * @return Value of the GPIO signal
 */
__u32 gpio_get_data(__u32 port, __u32 sig_no);

/*!
 * This function returns the value of the GPIO signal pad value.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 *
 * @return Pad value of the GPIO signal
 */
__u32 gpio_get_pad(__u32 port, __u32 sig_no);

/*!
 * This function enables/disables a GPIO signal's interrupt.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  en           \b true for enabling the interrupt; \b false for disabling the interrupt
 */
void gpio_config_int_en(__u32 port, __u32 sig_no, bool en);

/*!
 * This function checks to see if the GPIO interrupt is enabled for a particular signal.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 *
 * @return \b #true if the interrupt is enabled for the GPIO signal; \b #false otherwise.
 */
bool gpio_int_enabled(__u32 port, __u32 sig_no);

/*!
 * This function clears the GPIO interrupt for a signal on a port.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based) to clear
 */
void gpio_clear_int(__u32 port, __u32 sig_no);

/*!
 * This function is responsible for registering a GPIO signal's ISR.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  prio         priority as defined in \b enum \b #gpio_prio
 * @param  handler      GPIO ISR function pointer for the GPIO signal
 * @param  irq_flags    irq flags (not used)
 * @param  devname      device name associated with the interrupt
 * @param  dev_id       some unique information for the ISR
 *
 * @return 0 if successful; non-zero otherwise.
 */
int gpio_request_irq(__u32 port, __u32 sig_no, enum gpio_prio prio,
                     gpio_irq_handler handler, __u32 irq_flags, 
                     const char * devname, void *dev_id);

/*!
 * This function un-registers an ISR with the GPIO interrupt module.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  prio         priority as defined in \b enum \b #gpio_prio
 */
void gpio_free_irq(__u32 port, __u32 sig_no, enum gpio_prio prio);

/*!
 * This function configures the EDIO signal to be either input or output. For
 * input signals used for generating interrupts for the ARM core, how the
 * interrupts being triggered is also passed in via \a icr. For output signals
 * the \a icr value doesn't matter.
 *
 * @param  sig_no       EDIO signal (0 based) as defined in \b #edio_sig_no
 * @param  out          #true for output, #false for input
 * @param  icr          value defined in \b #edio_int_cfg
 */
void edio_config(enum edio_sig_no sig_no, bool out, enum edio_int_cfg icr);

/*!
 * This function sets a EDIO signal value.
 *
 * @param  sig_no       EDIO signal (0 based) as defined in \b #edio_sig_no
 * @param  data         value to be set (only 0 or 1 is valid)
 */
void edio_set_data(enum edio_sig_no sig_no, __u32 data);

/*!
 * This function returns the value of the EDIO signal.
 *
 * @param  sig_no       EDIO signal (0 based) as defined in \b #edio_sig_no
 *
 * @return Value of the EDIO signal (either 0 or 1)
 */
__u32 edio_get_data(enum edio_sig_no sig_no);

/*!
 * This function returns the EDIO edge-trigged interrupt for a signal.
 *
 * @param  sig_no       EDIO signal (0 based) as defined in \b #edio_sig_no 
 */
__u32 edio_get_int(enum edio_sig_no sig_no);

/*!
 * This function clears the EDIO edge-trigged interrupt for a signal.
 *
 * @param  sig_no       EDIO signal (0 based) as defined in \b #edio_sig_no
 */
void edio_clear_int(enum edio_sig_no sig_no);

#endif /* __GPIO_H__ */
