/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef __GPIO_H__
#define __GPIO_H__

/*!
 * @defgroup GPIO General Purpose Input Output (GPIO)
 * @ingroup MSL
 */

/*!
 * @file gpio.h
 * @brief This file contains the GPIO and IOMUX configuration details.
 *
 * @ingroup GPIO
 */

#include <asm/sizes.h>
#include <asm/arch/platform.h>
#include <linux/interrupt.h>

/*!
 * @name IOMUX/PAD Bit field definitions
 */

/*! @{ */

/*!
 * In order to identify pins more effectively, each mux-controlled pin's 
 * enumerated value is constructed in the following way:
 *
 * ------------------------------------------------
 * 31 | 30     24 | 23     16| 15     9 | 8     0
 * ------------------------------------------------
 *    | PAD_F     | PAD_I    | MUX_F    | MUX_I
 * ------------------------------------------------
 *
 * Bit 31 defines the MUX type (MUX_T) to be either regular mode or extended 
 * mode. i.MX31 only has multiplexed pins in regular mode so this field 
 * define is not used. Bit 0 to 8 contains MUX_I used to identify the register
 * offset (base is IOMUX_module_base + 0xC) defined in the Section "sw_pad_ctl &
 * sw_mux_ctl details" of the Tortola IC Spec. Bit 9 to 15 is MUX_F which
 * contains the offset value defined within the same register (each IOMUX 
 * control register contains four 8-bit fields for four different pins). The 
 * similar field definitions are used for the pad control register.
 */

/*!
 * Index of MUX control register
 */
#define MUX_I        0

/*!
 * Field within register for MUX control bits
 */
#define MUX_F        8

/*!
 * IOMUX type defined in bit 31
 */
#define MUX_T         31

/*!
 * Index of pad control register
 */
#define PAD_I        16

/*!
 *
 * Field within register for pad control bits
 */
#define PAD_F        24

/*! @} End IOMUX/PAD Bit field definitions */

/*!
 * This enumeration is constructed based on the Section
 * "sw_pad_ctl & sw_mux_ctl details" of the Tortola IC Spec. Each enumerated
 * value is constructed based on the rules described above.
 */
enum iomux_pins {
        MX31_PIN_A0              = (0 << MUX_T) | (73 << MUX_I) | (0 << MUX_F) | (96 << PAD_I) | (0 << PAD_F),
        MX31_PIN_A1              = (0 << MUX_T) | (72 << MUX_I) | (3 << MUX_F) | (95 << PAD_I) | (2 << PAD_F),
        MX31_PIN_A10             = (0 << MUX_T) | (70 << MUX_I) | (2 << MUX_F) | (92 << PAD_I) | (2 << PAD_F),
        MX31_PIN_A11             = (0 << MUX_T) | (70 << MUX_I) | (0 << MUX_F) | (92 << PAD_I) | (0 << PAD_F),
        MX31_PIN_A12             = (0 << MUX_T) | (69 << MUX_I) | (3 << MUX_F) | (91 << PAD_I) | (2 << PAD_F),
        MX31_PIN_A13             = (0 << MUX_T) | (69 << MUX_I) | (2 << MUX_F) | (91 << PAD_I) | (1 << PAD_F),
        MX31_PIN_A14             = (0 << MUX_T) | (69 << MUX_I) | (1 << MUX_F) | (91 << PAD_I) | (0 << PAD_F),
        MX31_PIN_A15             = (0 << MUX_T) | (69 << MUX_I) | (0 << MUX_F) | (90 << PAD_I) | (2 << PAD_F),
        MX31_PIN_A16             = (0 << MUX_T) | (68 << MUX_I) | (3 << MUX_F) | (90 << PAD_I) | (1 << PAD_F),
        MX31_PIN_A17             = (0 << MUX_T) | (68 << MUX_I) | (2 << MUX_F) | (90 << PAD_I) | (0 << PAD_F),
        MX31_PIN_A18             = (0 << MUX_T) | (68 << MUX_I) | (1 << MUX_F) | (89 << PAD_I) | (2 << PAD_F),
        MX31_PIN_A19             = (0 << MUX_T) | (68 << MUX_I) | (0 << MUX_F) | (89 << PAD_I) | (1 << PAD_F),
        MX31_PIN_A2              = (0 << MUX_T) | (72 << MUX_I) | (2 << MUX_F) | (95 << PAD_I) | (1 << PAD_F),
        MX31_PIN_A20             = (0 << MUX_T) | (67 << MUX_I) | (3 << MUX_F) | (89 << PAD_I) | (0 << PAD_F),
        MX31_PIN_A21             = (0 << MUX_T) | (67 << MUX_I) | (2 << MUX_F) | (88 << PAD_I) | (2 << PAD_F),
        MX31_PIN_A22             = (0 << MUX_T) | (67 << MUX_I) | (1 << MUX_F) | (88 << PAD_I) | (1 << PAD_F),
        MX31_PIN_A23             = (0 << MUX_T) | (67 << MUX_I) | (0 << MUX_F) | (88 << PAD_I) | (0 << PAD_F),
        MX31_PIN_A24             = (0 << MUX_T) | (66 << MUX_I) | (3 << MUX_F) | (87 << PAD_I) | (2 << PAD_F),
        MX31_PIN_A25             = (0 << MUX_T) | (66 << MUX_I) | (2 << MUX_F) | (87 << PAD_I) | (1 << PAD_F),
        MX31_PIN_A3              = (0 << MUX_T) | (72 << MUX_I) | (1 << MUX_F) | (95 << PAD_I) | (0 << PAD_F),
        MX31_PIN_A4              = (0 << MUX_T) | (72 << MUX_I) | (0 << MUX_F) | (94 << PAD_I) | (2 << PAD_F),
        MX31_PIN_A5              = (0 << MUX_T) | (71 << MUX_I) | (3 << MUX_F) | (94 << PAD_I) | (1 << PAD_F),
        MX31_PIN_A6              = (0 << MUX_T) | (71 << MUX_I) | (2 << MUX_F) | (94 << PAD_I) | (0 << PAD_F),
        MX31_PIN_A7              = (0 << MUX_T) | (71 << MUX_I) | (1 << MUX_F) | (93 << PAD_I) | (2 << PAD_F),
        MX31_PIN_A8              = (0 << MUX_T) | (71 << MUX_I) | (0 << MUX_F) | (93 << PAD_I) | (1 << PAD_F),
        MX31_PIN_A9              = (0 << MUX_T) | (70 << MUX_I) | (3 << MUX_F) | (93 << PAD_I) | (0 << PAD_F),
        MX31_PIN_ATA_CS0         = (0 << MUX_T) | (4 << MUX_I) | (0 << MUX_F) | (2 << PAD_I) | (2 << PAD_F),
        MX31_PIN_ATA_CS1         = (0 << MUX_T) | (2 << MUX_I) | (3 << MUX_F) | (2 << PAD_I) | (1 << PAD_F),
        MX31_PIN_ATA_DIOR        = (0 << MUX_T) | (2 << MUX_I) | (2 << MUX_F) | (2 << PAD_I) | (0 << PAD_F),
        MX31_PIN_ATA_DIOW        = (0 << MUX_T) | (2 << MUX_I) | (1 << MUX_F) | (1 << PAD_I) | (2 << PAD_F),
        MX31_PIN_ATA_DMACK       = (0 << MUX_T) | (2 << MUX_I) | (0 << MUX_F) | (1 << PAD_I) | (1 << PAD_F),
        MX31_PIN_ATA_RESET_B     = (0 << MUX_T) | (1 << MUX_I) | (3 << MUX_F) | (1 << PAD_I) | (0 << PAD_F),
        MX31_PIN_BATT_LINE       = (0 << MUX_T) | (25 << MUX_I) | (0 << MUX_F) | (32 << PAD_I) | (0 << PAD_F),
        MX31_PIN_BCLK            = (0 << MUX_T) | (54 << MUX_I) | (0 << MUX_F) | (70 << PAD_I) | (2 << PAD_F),
        MX31_PIN_BOOT_MODE0      = (0 << MUX_T) | (76 << MUX_I) | (0 << MUX_F) | (100 << PAD_I) | (0 << PAD_F),
        MX31_PIN_BOOT_MODE1      = (0 << MUX_T) | (75 << MUX_I) | (3 << MUX_F) | (99 << PAD_I) | (2 << PAD_F),
        MX31_PIN_BOOT_MODE2      = (0 << MUX_T) | (75 << MUX_I) | (2 << MUX_F) | (99 << PAD_I) | (1 << PAD_F),
        MX31_PIN_BOOT_MODE3      = (0 << MUX_T) | (75 << MUX_I) | (1 << MUX_F) | (99 << PAD_I) | (0 << PAD_F),
        MX31_PIN_BOOT_MODE4      = (0 << MUX_T) | (75 << MUX_I) | (0 << MUX_F) | (98 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CAPTURE         = (0 << MUX_T) | (81 << MUX_I) | (3 << MUX_F) | (107 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CAS             = (0 << MUX_T) | (53 << MUX_I) | (1 << MUX_F) | (69 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CE_CONTROL      = (0 << MUX_T) | (1 << MUX_I)  | (2 << MUX_F) | (0 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CKIH            = (0 << MUX_T) | (77 << MUX_I) | (0 << MUX_F) | (101 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CKIL            = (0 << MUX_T) | (74 << MUX_I) | (3 << MUX_F) | (98 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CLKO            = (0 << MUX_T) | (76 << MUX_I) | (1 << MUX_F) | (100 << PAD_I) | (1 << PAD_F),
        MX31_PIN_COMPARE         = (0 << MUX_T) | (81 << MUX_I) | (2 << MUX_F) | (107 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CONTRAST        = (0 << MUX_T) | (5 << MUX_I)  | (2 << MUX_F) | (6 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CS0             = (0 << MUX_T) | (56 << MUX_I) | (0 << MUX_F) | (73 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CS1             = (0 << MUX_T) | (55 << MUX_I) | (3 << MUX_F) | (73 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CS2             = (0 << MUX_T) | (55 << MUX_I) | (2 << MUX_F) | (72 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CS3             = (0 << MUX_T) | (55 << MUX_I) | (1 << MUX_F) | (72 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CS4             = (0 << MUX_T) | (55 << MUX_I) | (0 << MUX_F) | (72 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CS5             = (0 << MUX_T) | (54 << MUX_I) | (3 << MUX_F) | (71 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CSI_D10         = (0 << MUX_T) | (39 << MUX_I) | (3 << MUX_F) | (51 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CSI_D11         = (0 << MUX_T) | (39 << MUX_I) | (2 << MUX_F) | (51 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CSI_D12         = (0 << MUX_T) | (39 << MUX_I) | (1 << MUX_F) | (51 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CSI_D13         = (0 << MUX_T) | (39 << MUX_I) | (0 << MUX_F) | (50 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CSI_D14         = (0 << MUX_T) | (38 << MUX_I) | (3 << MUX_F) | (50 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CSI_D15         = (0 << MUX_T) | (38 << MUX_I) | (2 << MUX_F) | (50 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CSI_D4          = (0 << MUX_T) | (41 << MUX_I) | (1 << MUX_F) | (53 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CSI_D5          = (0 << MUX_T) | (41 << MUX_I) | (0 << MUX_F) | (53 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CSI_D6          = (0 << MUX_T) | (40 << MUX_I) | (3 << MUX_F) | (53 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CSI_D7          = (0 << MUX_T) | (40 << MUX_I) | (2 << MUX_F) | (52 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CSI_D8          = (0 << MUX_T) | (40 << MUX_I) | (1 << MUX_F) | (52 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CSI_D9          = (0 << MUX_T) | (40 << MUX_I) | (0 << MUX_F) | (52 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CSI_HSYNC       = (0 << MUX_T) | (37 << MUX_I) | (3 << MUX_F) | (49 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CSI_MCLK        = (0 << MUX_T) | (38 << MUX_I) | (1 << MUX_F) | (49 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CSI_PIXCLK      = (0 << MUX_T) | (37 << MUX_I) | (2 << MUX_F) | (48 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CSI_VSYNC       = (0 << MUX_T) | (38 << MUX_I) | (0 << MUX_F) | (49 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CSPI1_MISO      = (0 << MUX_T) | (32 << MUX_I) | (2 << MUX_F) | (42 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CSPI1_MOSI      = (0 << MUX_T) | (32 << MUX_I) | (3 << MUX_F) | (42 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CSPI1_SCLK      = (0 << MUX_T) | (31 << MUX_I) | (2 << MUX_F) | (40 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CSPI1_SPI_RDY   = (0 << MUX_T) | (31 << MUX_I) | (1 << MUX_F) | (40 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CSPI1_SS0       = (0 << MUX_T) | (32 << MUX_I) | (1 << MUX_F) | (41 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CSPI1_SS1       = (0 << MUX_T) | (32 << MUX_I) | (0 << MUX_F) | (41 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CSPI1_SS2       = (0 << MUX_T) | (31 << MUX_I) | (3 << MUX_F) | (41 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CSPI2_MISO      = (0 << MUX_T) | (30 << MUX_I) | (3 << MUX_F) | (39 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CSPI2_MOSI      = (0 << MUX_T) | (31 << MUX_I) | (0 << MUX_F) | (40 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CSPI2_SCLK      = (0 << MUX_T) | (29 << MUX_I) | (3 << MUX_F) | (38 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CSPI2_SPI_RDY   = (0 << MUX_T) | (29 << MUX_I) | (2 << MUX_F) | (38 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CSPI2_SS0       = (0 << MUX_T) | (30 << MUX_I) | (2 << MUX_F) | (39 << PAD_I) | (1 << PAD_F),
        MX31_PIN_CSPI2_SS1       = (0 << MUX_T) | (30 << MUX_I) | (1 << MUX_F) | (39 << PAD_I) | (0 << PAD_F),
        MX31_PIN_CSPI2_SS2       = (0 << MUX_T) | (30 << MUX_I) | (0 << MUX_F) | (38 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CTS1            = (0 << MUX_T) | (28 << MUX_I) | (2 << MUX_F) | (36 << PAD_I) | (2 << PAD_F),
        MX31_PIN_CTS2            = (0 << MUX_T) | (25 << MUX_I) | (1 << MUX_F) | (32 << PAD_I) | (1 << PAD_F),
        MX31_PIN_D0              = (0 << MUX_T) | (45 << MUX_I) | (1 << MUX_F) | (59 << PAD_I) | (0 << PAD_F),
        MX31_PIN_D1              = (0 << MUX_T) | (45 << MUX_I) | (2 << MUX_F) | (59 << PAD_I) | (1 << PAD_F),
        MX31_PIN_D10             = (0 << MUX_T) | (47 << MUX_I) | (3 << MUX_F) | (62 << PAD_I) | (1 << PAD_F),
        MX31_PIN_D11             = (0 << MUX_T) | (48 << MUX_I) | (0 << MUX_F) | (62 << PAD_I) | (2 << PAD_F),
        MX31_PIN_D12             = (0 << MUX_T) | (48 << MUX_I) | (1 << MUX_F) | (63 << PAD_I) | (0 << PAD_F),
        MX31_PIN_D13             = (0 << MUX_T) | (48 << MUX_I) | (2 << MUX_F) | (63 << PAD_I) | (1 << PAD_F),
        MX31_PIN_D14             = (0 << MUX_T) | (48 << MUX_I) | (3 << MUX_F) | (63 << PAD_I) | (2 << PAD_F),
        MX31_PIN_D15             = (0 << MUX_T) | (49 << MUX_I) | (0 << MUX_F) | (64 << PAD_I) | (0 << PAD_F),
        MX31_PIN_D2              = (0 << MUX_T) | (45 << MUX_I) | (3 << MUX_F) | (59 << PAD_I) | (2 << PAD_F),
        MX31_PIN_D3              = (0 << MUX_T) | (46 << MUX_I) | (0 << MUX_F) | (60 << PAD_I) | (0 << PAD_F),
        MX31_PIN_D3_CLS          = (0 << MUX_T) | (5 << MUX_I)  | (0 << MUX_F) | (5 << PAD_I) | (1 << PAD_F),
        MX31_PIN_D3_REV          = (0 << MUX_T) | (5 << MUX_I)  | (1 << MUX_F) | (5 << PAD_I) | (2 << PAD_F),
        MX31_PIN_D3_SPL          = (0 << MUX_T) | (4 << MUX_I)  | (3 << MUX_F) | (5 << PAD_I) | (0 << PAD_F),
        MX31_PIN_D4              = (0 << MUX_T) | (46 << MUX_I) | (1 << MUX_F) | (60 << PAD_I) | (1 << PAD_F),
        MX31_PIN_D5              = (0 << MUX_T) | (46 << MUX_I) | (2 << MUX_F) | (60 << PAD_I) | (2 << PAD_F),
        MX31_PIN_D6              = (0 << MUX_T) | (46 << MUX_I) | (3 << MUX_F) | (61 << PAD_I) | (0 << PAD_F),
        MX31_PIN_D7              = (0 << MUX_T) | (47 << MUX_I) | (0 << MUX_F) | (61 << PAD_I) | (1 << PAD_F),
        MX31_PIN_D8              = (0 << MUX_T) | (47 << MUX_I) | (1 << MUX_F) | (61 << PAD_I) | (2 << PAD_F),
        MX31_PIN_D9              = (0 << MUX_T) | (47 << MUX_I) | (2 << MUX_F) | (62 << PAD_I) | (0 << PAD_F),
        MX31_PIN_DCD_DCE1        = (0 << MUX_T) | (27 << MUX_I) | (2 << MUX_F) | (35 << PAD_I) | (1 << PAD_F),
        MX31_PIN_DCD_DTE1        = (0 << MUX_T) | (26 << MUX_I) | (2 << MUX_F) | (34 << PAD_I) | (0 << PAD_F),
        MX31_PIN_DE_B            = (0 << MUX_T) | (19 << MUX_I) | (1 << MUX_F) | (24 << PAD_I) | (1 << PAD_F),
        MX31_PIN_DQM0            = (0 << MUX_T) | (57 << MUX_I) | (3 << MUX_F) | (75 << PAD_I) | (2 << PAD_F),
        MX31_PIN_DQM1            = (0 << MUX_T) | (57 << MUX_I) | (2 << MUX_F) | (75 << PAD_I) | (1 << PAD_F),
        MX31_PIN_DQM2            = (0 << MUX_T) | (57 << MUX_I) | (1 << MUX_F) | (75 << PAD_I) | (0 << PAD_F),
        MX31_PIN_DQM3            = (0 << MUX_T) | (57 << MUX_I) | (0 << MUX_F) | (74 << PAD_I) | (2 << PAD_F),
        MX31_PIN_DRDY0           = (0 << MUX_T) | (8 << MUX_I)  | (1 << MUX_F) | (9 << PAD_I) | (2 << PAD_F),
        MX31_PIN_DSR_DCE1        = (0 << MUX_T) | (28 << MUX_I) | (0 << MUX_F) | (36 << PAD_I) | (0 << PAD_F),
        MX31_PIN_DSR_DTE1        = (0 << MUX_T) | (27 << MUX_I) | (0 << MUX_F) | (34 << PAD_I) | (2 << PAD_F),
        MX31_PIN_DTR_DCE1        = (0 << MUX_T) | (28 << MUX_I) | (1 << MUX_F) | (36 << PAD_I) | (1 << PAD_F),
        MX31_PIN_DTR_DCE2        = (0 << MUX_T) | (26 << MUX_I) | (1 << MUX_F) | (33 << PAD_I) | (2 << PAD_F),
        MX31_PIN_DTR_DTE1        = (0 << MUX_T) | (27 << MUX_I) | (1 << MUX_F) | (35 << PAD_I) | (0 << PAD_F),
        MX31_PIN_DVFS0           = (0 << MUX_T) | (74 << MUX_I) | (0 << MUX_F) | (97 << PAD_I) | (1 << PAD_F),
        MX31_PIN_DVFS1           = (0 << MUX_T) | (73 << MUX_I) | (3 << MUX_F) | (97 << PAD_I) | (0 << PAD_F),
        MX31_PIN_EB0             = (0 << MUX_T) | (56 << MUX_I) | (3 << MUX_F) | (74 << PAD_I) | (1 << PAD_F),
        MX31_PIN_EB1             = (0 << MUX_T) | (56 << MUX_I) | (2 << MUX_F) | (74 << PAD_I) | (0 << PAD_F),
        MX31_PIN_ECB             = (0 << MUX_T) | (54 << MUX_I) | (2 << MUX_F) | (71 << PAD_I) | (1 << PAD_F),
        MX31_PIN_FPSHIFT         = (0 << MUX_T) | (8 << MUX_I)  | (2 << MUX_F) | (10 << PAD_I) | (0 << PAD_F),
        MX31_PIN_GPIO1_0         = (0 << MUX_T) | (80 << MUX_I) | (3 << MUX_F) | (106 << PAD_I) | (1 << PAD_F),
        MX31_PIN_GPIO1_1         = (0 << MUX_T) | (80 << MUX_I) | (2 << MUX_F) | (106 << PAD_I) | (0 << PAD_F),
        MX31_PIN_GPIO1_2         = (0 << MUX_T) | (80 << MUX_I) | (1 << MUX_F) | (105 << PAD_I) | (2 << PAD_F),
        MX31_PIN_GPIO1_3         = (0 << MUX_T) | (79 << MUX_I) | (0 << MUX_F) | (105 << PAD_I) | (1 << PAD_F),
        MX31_PIN_GPIO1_4         = (0 << MUX_T) | (79 << MUX_I) | (3 << MUX_F) | (105 << PAD_I) | (0 << PAD_F),
        MX31_PIN_GPIO1_5         = (0 << MUX_T) | (79 << MUX_I) | (2 << MUX_F) | (104 << PAD_I) | (2 << PAD_F),
        MX31_PIN_GPIO1_6         = (0 << MUX_T) | (79 << MUX_I) | (1 << MUX_F) | (104 << PAD_I) | (1 << PAD_F),
        MX31_PIN_GPIO3_0         = (0 << MUX_T) | (79 << MUX_I) | (0 << MUX_F) | (104 << PAD_I) | (0 << PAD_F),
        MX31_PIN_GPIO3_1         = (0 << MUX_T) | (78 << MUX_I) | (3 << MUX_F) | (103 << PAD_I) | (2 << PAD_F),
        MX31_PIN_HSYNC           = (0 << MUX_T) | (8 << MUX_I)  | (3 << MUX_F) | (10 << PAD_I) | (1 << PAD_F),
        MX31_PIN_I2C_CLK         = (0 << MUX_T) | (37 << MUX_I) | (1 << MUX_F) | (48 << PAD_I) | (1 << PAD_F),
        MX31_PIN_I2C_DAT         = (0 << MUX_T) | (37 << MUX_I) | (0 << MUX_F) | (48 << PAD_I) | (0 << PAD_F),
        MX31_PIN_IOIS16          = (0 << MUX_T) | (42 << MUX_I) | (2 << MUX_F) | (55 << PAD_I) | (1 << PAD_F),
        MX31_PIN_KEY_COL0        = (0 << MUX_T) | (22 << MUX_I) | (3 << MUX_F) | (29 << PAD_I) | (0 << PAD_F),
        MX31_PIN_KEY_COL1        = (0 << MUX_T) | (22 << MUX_I) | (2 << MUX_F) | (28 << PAD_I) | (2 << PAD_F),
        MX31_PIN_KEY_COL2        = (0 << MUX_T) | (22 << MUX_I) | (1 << MUX_F) | (28 << PAD_I) | (1 << PAD_F),
        MX31_PIN_KEY_COL3        = (0 << MUX_T) | (22 << MUX_I) | (0 << MUX_F) | (28 << PAD_I) | (0 << PAD_F),
        MX31_PIN_KEY_COL4        = (0 << MUX_T) | (21 << MUX_I) | (3 << MUX_F) | (27 << PAD_I) | (2 << PAD_F),
        MX31_PIN_KEY_COL5        = (0 << MUX_T) | (21 << MUX_I) | (2 << MUX_F) | (27 << PAD_I) | (1 << PAD_F),
        MX31_PIN_KEY_COL6        = (0 << MUX_T) | (21 << MUX_I) | (1 << MUX_F) | (27 << PAD_I) | (0 << PAD_F),
        MX31_PIN_KEY_COL7        = (0 << MUX_T) | (21 << MUX_I) | (0 << MUX_F) | (26 << PAD_I) | (2 << PAD_F),
        MX31_PIN_KEY_ROW0        = (0 << MUX_T) | (24 << MUX_I) | (3 << MUX_F) | (31 << PAD_I) | (2 << PAD_F),
        MX31_PIN_KEY_ROW1        = (0 << MUX_T) | (24 << MUX_I) | (2 << MUX_F) | (31 << PAD_I) | (1 << PAD_F),
        MX31_PIN_KEY_ROW2        = (0 << MUX_T) | (24 << MUX_I) | (1 << MUX_F) | (31 << PAD_I) | (0 << PAD_F),
        MX31_PIN_KEY_ROW3        = (0 << MUX_T) | (24 << MUX_I) | (0 << MUX_F) | (30 << PAD_I) | (2 << PAD_F),
        MX31_PIN_KEY_ROW4        = (0 << MUX_T) | (23 << MUX_I) | (3 << MUX_F) | (30 << PAD_I) | (1 << PAD_F),
        MX31_PIN_KEY_ROW5        = (0 << MUX_T) | (23 << MUX_I) | (2 << MUX_F) | (30 << PAD_I) | (0 << PAD_F),
        MX31_PIN_KEY_ROW6        = (0 << MUX_T) | (23 << MUX_I) | (1 << MUX_F) | (29 << PAD_I) | (2 << PAD_F),
        MX31_PIN_KEY_ROW7        = (0 << MUX_T) | (23 << MUX_I) | (0 << MUX_F) | (29 << PAD_I) | (1 << PAD_F),
        MX31_PIN_LBA             = (0 << MUX_T) | (54 << MUX_I) | (1 << MUX_F) | (71 << PAD_I) | (0 << PAD_F),
        MX31_PIN_LCS0            = (0 << MUX_T) | (7 << MUX_I)  | (1 << MUX_F) | (8 << PAD_I) | (1 << PAD_F),
        MX31_PIN_LCS1            = (0 << MUX_T) | (7 << MUX_I)  | (0 << MUX_F) | (8 << PAD_I) | (0 << PAD_F),
        MX31_PIN_LD0             = (0 << MUX_T) | (13 << MUX_I) | (2 << MUX_F) | (16 << PAD_I) | (2 << PAD_F),
        MX31_PIN_LD1             = (0 << MUX_T) | (13 << MUX_I) | (1 << MUX_F) | (16 << PAD_I) | (1 << PAD_F),
        MX31_PIN_LD10            = (0 << MUX_T) | (11 << MUX_I) | (0 << MUX_F) | (13 << PAD_I) | (1 << PAD_F),
        MX31_PIN_LD11            = (0 << MUX_T) | (10 << MUX_I)  | (3 << MUX_F) | (13 << PAD_I) | (0 << PAD_F),
        MX31_PIN_LD12            = (0 << MUX_T) | (10 << MUX_I)  | (2 << MUX_F) | (12 << PAD_I) | (2 << PAD_F),
        MX31_PIN_LD13            = (0 << MUX_T) | (10 << MUX_I)  | (1 << MUX_F) | (12 << PAD_I) | (1 << PAD_F),
        MX31_PIN_LD14            = (0 << MUX_T) | (10 << MUX_I)  | (0 << MUX_F) | (12 << PAD_I) | (0 << PAD_F),
        MX31_PIN_LD15            = (0 << MUX_T) | (9 << MUX_I)  | (3 << MUX_F) | (11 << PAD_I) | (2 << PAD_F),
        MX31_PIN_LD16            = (0 << MUX_T) | (9 << MUX_I)  | (2 << MUX_F) | (11 << PAD_I) | (1 << PAD_F),
        MX31_PIN_LD17            = (0 << MUX_T) | (9 << MUX_I)  | (1 << MUX_F) | (11 << PAD_I) | (0 << PAD_F),
        MX31_PIN_LD2             = (0 << MUX_T) | (13 << MUX_I) | (0 << MUX_F) | (16 << PAD_I) | (0 << PAD_F),
        MX31_PIN_LD3             = (0 << MUX_T) | (12 << MUX_I) | (3 << MUX_F) | (15 << PAD_I) | (2 << PAD_F),
        MX31_PIN_LD4             = (0 << MUX_T) | (12 << MUX_I) | (2 << MUX_F) | (15 << PAD_I) | (1 << PAD_F),
        MX31_PIN_LD5             = (0 << MUX_T) | (12 << MUX_I) | (1 << MUX_F) | (15 << PAD_I) | (0 << PAD_F),
        MX31_PIN_LD6             = (0 << MUX_T) | (12 << MUX_I) | (0 << MUX_F) | (14 << PAD_I) | (2 << PAD_F),
        MX31_PIN_LD7             = (0 << MUX_T) | (11 << MUX_I) | (3 << MUX_F) | (14 << PAD_I) | (1 << PAD_F),
        MX31_PIN_LD8             = (0 << MUX_T) | (11 << MUX_I) | (2 << MUX_F) | (14 << PAD_I) | (0 << PAD_F),
        MX31_PIN_LD9             = (0 << MUX_T) | (11 << MUX_I) | (1 << MUX_F) | (13 << PAD_I) | (2 << PAD_F),
        MX31_PIN_M_GRANT         = (0 << MUX_T) | (41 << MUX_I) | (2 << MUX_F) | (54 << PAD_I) | (0 << PAD_F),
        MX31_PIN_M_REQUEST       = (0 << MUX_T) | (41 << MUX_I) | (3 << MUX_F) | (54 << PAD_I) | (1 << PAD_F),
        MX31_PIN_MA10            = (0 << MUX_T) | (70 << MUX_I) | (1 << MUX_F) | (92 << PAD_I) | (1 << PAD_F),
        MX31_PIN_NFALE           = (0 << MUX_T) | (50 << MUX_I) | (1 << MUX_F) | (65 << PAD_I) | (2 << PAD_F),
        MX31_PIN_NFCE_B          = (0 << MUX_T) | (49 << MUX_I) | (2 << MUX_F) | (64 << PAD_I) | (2 << PAD_F),
        MX31_PIN_NFCLE           = (0 << MUX_T) | (50 << MUX_I) | (0 << MUX_F) | (65 << PAD_I) | (1 << PAD_F),
        MX31_PIN_NFRB            = (0 << MUX_T) | (49 << MUX_I) | (1 << MUX_F) | (64 << PAD_I) | (1 << PAD_F),
        MX31_PIN_NFRE_B          = (0 << MUX_T) | (50 << MUX_I) | (2 << MUX_F) | (66 << PAD_I) | (0 << PAD_F),
        MX31_PIN_NFWE_B          = (0 << MUX_T) | (50 << MUX_I) | (3 << MUX_F) | (66 << PAD_I) | (1 << PAD_F),
        MX31_PIN_NFWP_B          = (0 << MUX_T) | (49 << MUX_I) | (3 << MUX_F) | (65 << PAD_I) | (0 << PAD_F),
        MX31_PIN_OE              = (0 << MUX_T) | (56 << MUX_I) | (1 << MUX_F) | (73 << PAD_I) | (2 << PAD_F),
        MX31_PIN_PAD2            = (0 << MUX_T) | (1 << MUX_I)  | (0 << MUX_F) | (0 << PAD_I) | (0 << PAD_F),
        MX31_PIN_PAR_RS          = (0 << MUX_T) | (6 << MUX_I)  | (2 << MUX_F) | (7 << PAD_I) | (1 << PAD_F),
        MX31_PIN_PC_BVD1         = (0 << MUX_T) | (43 << MUX_I) | (1 << MUX_F) | (56 << PAD_I) | (1 << PAD_F),
        MX31_PIN_PC_BVD2         = (0 << MUX_T) | (43 << MUX_I) | (0 << MUX_F) | (56 << PAD_I) | (0 << PAD_F),
        MX31_PIN_PC_CD1_B        = (0 << MUX_T) | (45 << MUX_I) | (0 << MUX_F) | (58 << PAD_I) | (2 << PAD_F),
        MX31_PIN_PC_CD2_B        = (0 << MUX_T) | (44 << MUX_I) | (3 << MUX_F) | (58 << PAD_I) | (1 << PAD_F),
        MX31_PIN_PC_POE          = (0 << MUX_T) | (42 << MUX_I) | (0 << MUX_F) | (54 << PAD_I) | (2 << PAD_F),
        MX31_PIN_PC_PWRON        = (0 << MUX_T) | (44 << MUX_I) | (0 << MUX_F) | (57 << PAD_I) | (1 << PAD_F),
        MX31_PIN_PC_READY        = (0 << MUX_T) | (44 << MUX_I) | (1 << MUX_F) | (57 << PAD_I) | (2 << PAD_F),
        MX31_PIN_PC_RST          = (0 << MUX_T) | (42 << MUX_I) | (3 << MUX_F) | (55 << PAD_I) | (2 << PAD_F),
        MX31_PIN_PC_RW_B         = (0 << MUX_T) | (42 << MUX_I) | (1 << MUX_F) | (55 << PAD_I) | (0 << PAD_F),
        MX31_PIN_PC_VS1          = (0 << MUX_T) | (43 << MUX_I) | (3 << MUX_F) | (57 << PAD_I) | (0 << PAD_F),
        MX31_PIN_PC_VS2          = (0 << MUX_T) | (43 << MUX_I) | (2 << MUX_F) | (56 << PAD_I) | (2 << PAD_F),
        MX31_PIN_PC_WAIT_B       = (0 << MUX_T) | (44 << MUX_I) | (2 << MUX_F) | (58 << PAD_I) | (0 << PAD_F),
        MX31_PIN_POR_B           = (0 << MUX_T) | (76 << MUX_I) | (2 << MUX_F) | (100 << PAD_I) | (2 << PAD_F),
        MX31_PIN_POWER_FAIL      = (0 << MUX_T) | (74 << MUX_I) | (2 << MUX_F) | (98 << PAD_I) | (0 << PAD_F),
        MX31_PIN_PWMO            = (0 << MUX_T) | (81 << MUX_I) | (0 << MUX_F) | (106 << PAD_I) | (2 << PAD_F),
        MX31_PIN_RAS             = (0 << MUX_T) | (53 << MUX_I) | (2 << MUX_F) | (70 << PAD_I) | (0 << PAD_F),
        MX31_PIN_READ            = (0 << MUX_T) | (6 << MUX_I)  | (0 << MUX_F) | (6 << PAD_I) | (2 << PAD_F),
        MX31_PIN_RESET_IN_B      = (0 << MUX_T) | (76 << MUX_I) | (3 << MUX_F) | (101 << PAD_I) | (0 << PAD_F),
        MX31_PIN_RI_DCE1         = (0 << MUX_T) | (27 << MUX_I) | (3 << MUX_F) | (35 << PAD_I) | (2 << PAD_F),
        MX31_PIN_RI_DTE1         = (0 << MUX_T) | (26 << MUX_I) | (3 << MUX_F) | (34 << PAD_I) | (1 << PAD_F),
        MX31_PIN_RTCK            = (0 << MUX_T) | (20 << MUX_I) | (3 << MUX_F) | (26 << PAD_I) | (1 << PAD_F),
        MX31_PIN_RTS1            = (0 << MUX_T) | (28 << MUX_I) | (3 << MUX_F) | (37 << PAD_I) | (0 << PAD_F),
        MX31_PIN_RTS2            = (0 << MUX_T) | (25 << MUX_I) | (2 << MUX_F) | (32 << PAD_I) | (2 << PAD_F),
        MX31_PIN_RW              = (0 << MUX_T) | (53 << MUX_I) | (3 << MUX_F) | (70 << PAD_I) | (1 << PAD_F),
        MX31_PIN_RXD1            = (0 << MUX_T) | (29 << MUX_I) | (1 << MUX_F) | (37 << PAD_I) | (2 << PAD_F),
        MX31_PIN_RXD2            = (0 << MUX_T) | (26 << MUX_I) | (0 << MUX_F) | (33 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SCK3            = (0 << MUX_T) | (36 << MUX_I) | (1 << MUX_F) | (47 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SCK4            = (0 << MUX_T) | (35 << MUX_I) | (1 << MUX_F) | (45 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SCK5            = (0 << MUX_T) | (34 << MUX_I) | (1 << MUX_F) | (44 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SCK6            = (0 << MUX_T) | (33 << MUX_I) | (1 << MUX_F) | (43 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SCLK0           = (0 << MUX_T) | (78 << MUX_I) | (2 << MUX_F) | (103 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD_D_CLK        = (0 << MUX_T) | (7 << MUX_I)  | (2 << MUX_F) | (8 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD_D_IO         = (0 << MUX_T) | (7 << MUX_I)  | (3 << MUX_F) | (9 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD_D_O          = (0 << MUX_T) | (8 << MUX_I)  | (0 << MUX_F) | (9 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD0             = (0 << MUX_T) | (65 << MUX_I) | (3 << MUX_F) | (86 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD1             = (0 << MUX_T) | (65 << MUX_I) | (2 << MUX_F) | (86 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD1_CLK         = (0 << MUX_T) | (4 << MUX_I)  | (1 << MUX_F) | (4 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD1_CMD         = (0 << MUX_T) | (4 << MUX_I)  | (2 << MUX_F) | (4 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD1_DATA0       = (0 << MUX_T) | (4 << MUX_I)  | (0 << MUX_F) | (4 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD1_DATA1       = (0 << MUX_T) | (3 << MUX_I)  | (3 << MUX_F) | (3 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD1_DATA2       = (0 << MUX_T) | (3 << MUX_I)  | (2 << MUX_F) | (3 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD1_DATA3       = (0 << MUX_T) | (3 << MUX_I)  | (1 << MUX_F) | (3 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD10            = (0 << MUX_T) | (63 << MUX_I) | (1 << MUX_F) | (83 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD11            = (0 << MUX_T) | (63 << MUX_I) | (0 << MUX_F) | (82 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD12            = (0 << MUX_T) | (62 << MUX_I) | (3 << MUX_F) | (82 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD13            = (0 << MUX_T) | (62 << MUX_I) | (2 << MUX_F) | (82 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD14            = (0 << MUX_T) | (62 << MUX_I) | (1 << MUX_F) | (81 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD15            = (0 << MUX_T) | (62 << MUX_I) | (0 << MUX_F) | (81 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD16            = (0 << MUX_T) | (61 << MUX_I) | (3 << MUX_F) | (81 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD17            = (0 << MUX_T) | (61 << MUX_I) | (2 << MUX_F) | (80 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD18            = (0 << MUX_T) | (61 << MUX_I) | (1 << MUX_F) | (80 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD19            = (0 << MUX_T) | (61 << MUX_I) | (0 << MUX_F) | (80 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD2             = (0 << MUX_T) | (65 << MUX_I) | (1 << MUX_F) | (85 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD20            = (0 << MUX_T) | (60 << MUX_I) | (3 << MUX_F) | (79 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD21            = (0 << MUX_T) | (60 << MUX_I) | (2 << MUX_F) | (79 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD22            = (0 << MUX_T) | (60 << MUX_I) | (1 << MUX_F) | (79 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD23            = (0 << MUX_T) | (60 << MUX_I) | (0 << MUX_F) | (78 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD24            = (0 << MUX_T) | (59 << MUX_I) | (3 << MUX_F) | (78 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD25            = (0 << MUX_T) | (59 << MUX_I) | (2 << MUX_F) | (78 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD26            = (0 << MUX_T) | (59 << MUX_I) | (1 << MUX_F) | (77 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD27            = (0 << MUX_T) | (59 << MUX_I) | (0 << MUX_F) | (77 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD28            = (0 << MUX_T) | (58 << MUX_I) | (3 << MUX_F) | (77 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD29            = (0 << MUX_T) | (58 << MUX_I) | (2 << MUX_F) | (76 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD3             = (0 << MUX_T) | (65 << MUX_I) | (0 << MUX_F) | (85 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD30            = (0 << MUX_T) | (58 << MUX_I) | (1 << MUX_F) | (76 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD31            = (0 << MUX_T) | (58 << MUX_I) | (0 << MUX_F) | (76 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD4             = (0 << MUX_T) | (64 << MUX_I) | (3 << MUX_F) | (85 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD5             = (0 << MUX_T) | (64 << MUX_I) | (2 << MUX_F) | (84 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD6             = (0 << MUX_T) | (64 << MUX_I) | (1 << MUX_F) | (84 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD7             = (0 << MUX_T) | (64 << MUX_I) | (0 << MUX_F) | (84 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SD8             = (0 << MUX_T) | (63 << MUX_I) | (3 << MUX_F) | (83 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SD9             = (0 << MUX_T) | (63 << MUX_I) | (2 << MUX_F) | (83 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SDBA0           = (0 << MUX_T) | (66 << MUX_I) | (0 << MUX_F) | (86 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SDBA1           = (0 << MUX_T) | (66 << MUX_I) | (1 << MUX_F) | (87 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SDCKE0          = (0 << MUX_T) | (52 << MUX_I) | (3 << MUX_F) | (69 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SDCKE1          = (0 << MUX_T) | (52 << MUX_I) | (2 << MUX_F) | (68 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SDCLK           = (0 << MUX_T) | (52 << MUX_I) | (1 << MUX_F) | (68 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD1CLK_B        = (0 << MUX_T) | (1 << MUX_I)  | (1 << MUX_F) | (0 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SD2CLK_B        = (0 << MUX_T) | (52 << MUX_I) | (0 << MUX_F) | (68 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SDQS0           = (0 << MUX_T) | (51 << MUX_I) | (3 << MUX_F) | (67 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SDQS1           = (0 << MUX_T) | (51 << MUX_I) | (2 << MUX_F) | (67 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SDQS2           = (0 << MUX_T) | (51 << MUX_I) | (1 << MUX_F) | (67 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SDQS3           = (0 << MUX_T) | (51 << MUX_I) | (0 << MUX_F) | (66 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SDWE            = (0 << MUX_T) | (53 << MUX_I) | (0 << MUX_F) | (69 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SER_RS          = (0 << MUX_T) | (6 << MUX_I)  | (3 << MUX_F) | (7 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SFS3            = (0 << MUX_T) | (36 << MUX_I) | (0 << MUX_F) | (46 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SFS4            = (0 << MUX_T) | (35 << MUX_I) | (0 << MUX_F) | (45 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SFS5            = (0 << MUX_T) | (34 << MUX_I) | (0 << MUX_F) | (44 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SFS6            = (0 << MUX_T) | (33 << MUX_I) | (0 << MUX_F) | (42 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SIMPD0          = (0 << MUX_T) | (77 << MUX_I) | (1 << MUX_F) | (101 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SJC_MOD         = (0 << MUX_T) | (19 << MUX_I) | (0 << MUX_F) | (24 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SRST0           = (0 << MUX_T) | (78 << MUX_I) | (1 << MUX_F) | (103 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SRX0            = (0 << MUX_T) | (77 << MUX_I) | (2 << MUX_F) | (102 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SRXD3           = (0 << MUX_T) | (36 << MUX_I) | (2 << MUX_F) | (47 << PAD_I) | (1 << PAD_F),
        MX31_PIN_SRXD4           = (0 << MUX_T) | (35 << MUX_I) | (2 << MUX_F) | (46 << PAD_I) | (0 << PAD_F),
        MX31_PIN_SRXD5           = (0 << MUX_T) | (34 << MUX_I) | (2 << MUX_F) | (44 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SRXD6           = (0 << MUX_T) | (33 << MUX_I) | (2 << MUX_F) | (43 << PAD_I) | (1 << PAD_F),
        MX31_PIN_STX0            = (0 << MUX_T) | (77 << MUX_I) | (3 << MUX_F) | (102 << PAD_I) | (1 << PAD_F),
        MX31_PIN_STXD3           = (0 << MUX_T) | (36 << MUX_I) | (3 << MUX_F) | (47 << PAD_I) | (2 << PAD_F),
        MX31_PIN_STXD4           = (0 << MUX_T) | (35 << MUX_I) | (3 << MUX_F) | (46 << PAD_I) | (1 << PAD_F),
        MX31_PIN_STXD5           = (0 << MUX_T) | (34 << MUX_I) | (3 << MUX_F) | (45 << PAD_I) | (0 << PAD_F),
        MX31_PIN_STXD6           = (0 << MUX_T) | (33 << MUX_I) | (3 << MUX_F) | (43 << PAD_I) | (2 << PAD_F),
        MX31_PIN_SVEN0           = (0 << MUX_T) | (78 << MUX_I) | (0 << MUX_F) | (102 << PAD_I) | (2 << PAD_F),
        MX31_PIN_TCK             = (0 << MUX_T) | (20 << MUX_I) | (2 << MUX_F) | (26 << PAD_I) | (0 << PAD_F),
        MX31_PIN_TDI             = (0 << MUX_T) | (20 << MUX_I) | (0 << MUX_F) | (25 << PAD_I) | (1 << PAD_F),
        MX31_PIN_TDO             = (0 << MUX_T) | (19 << MUX_I) | (3 << MUX_F) | (25 << PAD_I) | (0 << PAD_F),
        MX31_PIN_TMS             = (0 << MUX_T) | (20 << MUX_I) | (1 << MUX_F) | (25 << PAD_I) | (2 << PAD_F),
        MX31_PIN_TRSTB           = (0 << MUX_T) | (19 << MUX_I) | (2 << MUX_F) | (24 << PAD_I) | (2 << PAD_F),
        MX31_PIN_TXD1            = (0 << MUX_T) | (29 << MUX_I) | (0 << MUX_F) | (37 << PAD_I) | (1 << PAD_F),
        MX31_PIN_TXD2            = (0 << MUX_T) | (25 << MUX_I) | (3 << MUX_F) | (33 << PAD_I) | (0 << PAD_F),
        MX31_PIN_USB_BYP         = (0 << MUX_T) | (18 << MUX_I) | (1 << MUX_F) | (23 << PAD_I) | (0 << PAD_F),
        MX31_PIN_USB_OC          = (0 << MUX_T) | (18 << MUX_I) | (2 << MUX_F) | (23 << PAD_I) | (1 << PAD_F),
        MX31_PIN_USB_PWR         = (0 << MUX_T) | (18 << MUX_I) | (3 << MUX_F) | (23 << PAD_I) | (2 << PAD_F),
        MX31_PIN_USBH2_CLK       = (0 << MUX_T) | (15 << MUX_I) | (0 << MUX_F) | (18 << PAD_I) | (2 << PAD_F),
        MX31_PIN_USBH2_DATA0     = (0 << MUX_T) | (14 << MUX_I) | (0 << MUX_F) | (17 << PAD_I) | (1 << PAD_F),
        MX31_PIN_USBH2_DATA1     = (0 << MUX_T) | (13 << MUX_I) | (3 << MUX_F) | (17 << PAD_I) | (0 << PAD_F),
        MX31_PIN_USBH2_DIR       = (0 << MUX_T) | (14 << MUX_I) | (3 << MUX_F) | (18 << PAD_I) | (1 << PAD_F),
        MX31_PIN_USBH2_NXT       = (0 << MUX_T) | (14 << MUX_I) | (1 << MUX_F) | (17 << PAD_I) | (2 << PAD_F),
        MX31_PIN_USBH2_STP       = (0 << MUX_T) | (14 << MUX_I) | (2 << MUX_F) | (18 << PAD_I) | (0 << PAD_F),
        MX31_PIN_USBOTG_CLK      = (0 << MUX_T) | (18 << MUX_I) | (0 << MUX_F) | (22 << PAD_I) | (2 << PAD_F),
        MX31_PIN_USBOTG_DATA0    = (0 << MUX_T) | (17 << MUX_I) | (0 << MUX_F) | (21 << PAD_I) | (1 << PAD_F),
        MX31_PIN_USBOTG_DATA1    = (0 << MUX_T) | (16 << MUX_I) | (3 << MUX_F) | (21 << PAD_I) | (0 << PAD_F),
        MX31_PIN_USBOTG_DATA2    = (0 << MUX_T) | (16 << MUX_I) | (2 << MUX_F) | (20 << PAD_I) | (2 << PAD_F),
        MX31_PIN_USBOTG_DATA3    = (0 << MUX_T) | (16 << MUX_I) | (1 << MUX_F) | (20 << PAD_I) | (1 << PAD_F),
        MX31_PIN_USBOTG_DATA4    = (0 << MUX_T) | (16 << MUX_I) | (0 << MUX_F) | (20 << PAD_I) | (0 << PAD_F),
        MX31_PIN_USBOTG_DATA5    = (0 << MUX_T) | (15 << MUX_I) | (3 << MUX_F) | (19 << PAD_I) | (2 << PAD_F),
        MX31_PIN_USBOTG_DATA6    = (0 << MUX_T) | (15 << MUX_I) | (2 << MUX_F) | (19 << PAD_I) | (1 << PAD_F),
        MX31_PIN_USBOTG_DATA7    = (0 << MUX_T) | (15 << MUX_I) | (1 << MUX_F) | (19 << PAD_I) | (0 << PAD_F),
        MX31_PIN_USBOTG_DIR      = (0 << MUX_T) | (17 << MUX_I) | (3 << MUX_F) | (22 << PAD_I) | (1 << PAD_F),
        MX31_PIN_USBOTG_NXT      = (0 << MUX_T) | (17 << MUX_I) | (1 << MUX_F) | (21 << PAD_I) | (2 << PAD_F),
        MX31_PIN_USBOTG_STP      = (0 << MUX_T) | (17 << MUX_I) | (2 << MUX_F) | (22 << PAD_I) | (0 << PAD_F),
        MX31_PIN_VPG0            = (0 << MUX_T) | (73 << MUX_I) | (2 << MUX_F) | (96 << PAD_I) | (2 << PAD_F),
        MX31_PIN_VPG1            = (0 << MUX_T) | (73 << MUX_I) | (1 << MUX_F) | (96 << PAD_I) | (1 << PAD_F),
        MX31_PIN_VSTBY           = (0 << MUX_T) | (74 << MUX_I) | (1 << MUX_F) | (97 << PAD_I) | (2 << PAD_F),
        MX31_PIN_VSYNC0          = (0 << MUX_T) | (9 << MUX_I)  | (0 << MUX_F) | (10 << PAD_I) | (2 << PAD_F),
        MX31_PIN_VSYNC3          = (0 << MUX_T) | (5 << MUX_I)  | (3 << MUX_F) | (6 << PAD_I) | (1 << PAD_F),
        MX31_PIN_WATCHDOG_RST    = (0 << MUX_T) | (79 << MUX_I) | (1 << MUX_F) | (107 << PAD_I) | (0 << PAD_F),
        MX31_PIN_WRITE           = (0 << MUX_T) | (6 << MUX_I)  | (1 << MUX_F) | (7 << PAD_I) | (0 << PAD_F)
};

/*!
 * various IOMUX output functions
 */
enum iomux_output_config {
        OUTPUTCONFIG_GPIO = 0,   /*!< used as GPIO */
        OUTPUTCONFIG_FUNC,       /*!< used as function */
        OUTPUTCONFIG_ALT1,       /*!< used as alternate function 1 */
        OUTPUTCONFIG_ALT2,       /*!< used as alternate function 2 */
        OUTPUTCONFIG_ALT3,       /*!< used as alternate function 3 */
        OUTPUTCONFIG_ALT4,       /*!< used as alternate function 4 */
        OUTPUTCONFIG_ALT5,       /*!< used as alternate function 5 */
        OUTPUTCONFIG_ALT6        /*!< used as alternate function 6 */
};

/*!
 * various IOMUX input functions
 */
enum iomux_input_config {
        INPUTCONFIG_NONE = 0,            /*!< not configured for input */
        INPUTCONFIG_GPIO = 1 << 0,  /*!< used as GPIO */
        INPUTCONFIG_FUNC = 1 << 1,  /*!< used as function */
        INPUTCONFIG_ALT1 = 1 << 2,  /*!< used as alternate function 1 */
        INPUTCONFIG_ALT2 = 1 << 3   /*!< used as alternate function 2 */
};

/*!
 * various IOMUX pad functions
 */
enum iomux_pad_config {
        PAD_CTL_NOLOOPBACK      = 0x0 << 8,
        PAD_CTL_LOOPBACK        = 0x1 << 8,
        PAD_CTL_PKE_NONE        = 0x0 << 7,
        PAD_CTL_PKE_ENABLE      = 0x1 << 7,
        PAD_CTL_PUE_KEEPER      = 0x0 << 6,
        PAD_CTL_PUE_PUD         = 0x1 << 6,
        PAD_CTL_100K_PD         = 0x0 << 4,
        PAD_CTL_100K_PU         = 0x1 << 4,
        PAD_CTL_47K_PU          = 0x2 << 4,
        PAD_CTL_22K_PU          = 0x3 << 4,
        PAD_CTL_HYS_CMOS        = 0x0 << 3,
        PAD_CTL_HYS_SCHMITZ     = 0x1 << 3,
        PAD_CTL_ODE_CMOS        = 0x0 << 2,
        PAD_CTL_ODE_OpenDrain   = 0x1 << 2,
        PAD_CTL_DSE0_NORMAL     = 0x0 << 1,
        PAD_CTL_DSE0_HIGH       = 0x1 << 1,
        PAD_CTL_SRE_SLOW        = 0x0 << 0,
        PAD_CTL_SRE_FAST        = 0x1 << 0
};

/*!
 * various IOMUX general purpose functions
 */
enum iomux_gp_func {
        MUX_PGP_FIRI            = 0x1 << 0,
        MUX_DDR_MODE            = 0x1 << 1,
        MUX_PGP_CSPI_BB         = 0x1 << 2,
        MUX_PGP_ATA_1           = 0x1 << 3,
        MUX_PGP_ATA_2           = 0x1 << 4,
        MUX_PGP_ATA_3           = 0x1 << 5,
        MUX_PGP_ATA_4           = 0x1 << 6,
        MUX_PGP_ATA_5           = 0x1 << 7,
        MUX_PGP_ATA_6           = 0x1 << 8,
        MUX_PGP_ATA_7           = 0x1 << 9,
        MUX_PGP_ATA_8           = 0x1 << 10,
        MUX_PGP_UH2             = 0x1 << 11,
        MUX_SDCTL_CSD0_SEL      = 0x1 << 12,
        MUX_SDCTL_CSD1_SEL      = 0x1 << 13,
        MUX_CSPI1_UART3         = 0x1 << 14,
        MUX_EXTDMAREQ2_MBX_SEL  = 0x1 << 15,
        MUX_TAMPER_DETECT_EN    = 0x1 << 16,
        MUX_PGP_USB_4WIRE       = 0x1 << 17,
        MUX_PGB_USB_COMMON      = 0x1 << 18,
        MUX_SDHC_MEMSTICK1      = 0x1 << 19,
        MUX_SDHC_MEMSTICK2      = 0x1 << 20,
        MUX_PGP_SPLL_BYP        = 0x1 << 21,
        MUX_PGP_UPLL_BYP        = 0x1 << 22,
        MUX_PGP_MSHC1_CLK_SEL   = 0x1 << 23,
        MUX_PGP_MSHC2_CLK_SEL   = 0x1 << 24,
        MUX_CSPI3_UART5_SEL     = 0x1 << 25
};

/*!
 * This enumeration data type defines various signals for interrupting the
 * ARM core from GPIO signals. 
 */
enum gpio_sig_no {
        GPIO_INT0		= 0,		    
        GPIO_INT1		= 1,
        GPIO_INT2		= 2,
        GPIO_INT3		= 3,
        GPIO_INT4		= 4,
};

/*!
 * This function is used to configure a pin through the IOMUX module.
 *
 * @param  pin          pin number as defined in \b #iomux_pins
 * @param  out          output function as defined in \b #iomux_output_config
 * @param  in           input function as defined in \b #iomux_input_config
 */
void iomux_config_mux(enum iomux_pins pin, 
                      enum iomux_output_config out, 
                      enum iomux_input_config in);

/*!
 * This function configures the pad value for a IOMUX pin.
 *
 * @param  pin          a pin number as defined in \b #iomux_pins
 * @param  config       ORed value of elements defined in \b #iomux_pad_config
 */
void iomux_config_pad(enum iomux_pins pin, __u32 config);

/*!
 * This function enables/disables the general purpose function for a particular
 * signal.
 *
 * @param  gp   one signal as defined in \b enum \b #iomux_gp_func
 * @param  en   \b #true to enable; \b #false to disable
 */
void iomux_config_gpr(enum iomux_gp_func gp, bool en);

/* gpio related defines */

/*!
 * There are two queues for registered GPIO ISRs. One is for high priority and
 * the other is for low priority. The ISRs in the high priority queue will be
 * called first before the low priority queue if more than one GPIO interrupt
 * occurs at the same time.
 */
enum gpio_prio {
        GPIO_HIGH_PRIO  = 0,     /*!< high priority queue */
        GPIO_LOW_PRIO            /*!< low priority queue */
};

/*!
 * This enumeration data type defines various different ways for interrupting 
 * the ARM core from GPIO signals. The way to interrupt the core is dictated 
 * by the external hardware.
 */
enum gpio_int_cfg {
        GPIO_INT_LOW_LEV        = 0x0,    /*!< low level sensitive */
        GPIO_INT_HIGH_LEV       = 0x1,    /*!< high level sensitive */
        GPIO_INT_RISE_EDGE      = 0x2,    /*!< rising edge sensitive */
        GPIO_INT_FALL_EDGE      = 0x3,    /*!< falling edge sensitive */
        GPIO_INT_NONE           = 0x4     /*!< No interrupt */
};

typedef irqreturn_t (*gpio_irq_handler)(int, void *, struct pt_regs *);

/*!
 * This function configures the GPIO signal to be either input or output. For
 * input signals used for generating interrupts for the ARM core, how the
 * interrupts being triggered is also passed in via \a icr. For output signals,
 * the \a icr value doesn't matter.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  out          #true for output, #false for input
 * @param  icr          value defined in \b #gpio_int_cfg
 */
void gpio_config(__u32 port, __u32 sig_no, bool out, enum gpio_int_cfg icr);

/*!
 * This function sets a GPIO signal value.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  data         value to be set (only 0 or 1 is valid)
 */
void gpio_set_data(__u32 port, __u32 sig_no, __u32 data);

/*!
 * This function returns the value of the GPIO signal.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 *
 * @return Value of the GPIO signal
 */
__u32 gpio_get_data(__u32 port, __u32 sig_no);

/*!
 * This function returns the value of the GPIO signal pad value.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 *
 * @return Pad value of the GPIO signal
 */
__u32 gpio_get_pad(__u32 port, __u32 sig_no);

/*!
 * This function enables/disables a GPIO signal's interrupt.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  en           \b true for enabling the interrupt; \b false for disabling the interrupt
 */
void gpio_config_int_en(__u32 port, __u32 sig_no, bool en);

/*!
 * This function checks to see if the GPIO interrupt is enabled for a particular signal.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 *
 * @return \b #true if the interrupt is enabled for the GPIO signal; \b #false otherwise.
 */
bool gpio_int_enabled(__u32 port, __u32 sig_no);

/*!
 * This function clears the GPIO interrupt for a signal on a port.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based) to clear
 */
void gpio_clear_int(__u32 port, __u32 sig_no);

/*!
 * This function is responsible for registering a GPIO signal's ISR.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  prio         priority as defined in \b enum \b #gpio_prio
 * @param  handler      GPIO ISR function pointer for the GPIO signal
 * @param  irq_flags    irq flags (not used)
 * @param  devname      device name associated with the interrupt
 * @param  dev_id       some unique information for the ISR
 *
 * @return 0 if successful; non-zero otherwise.
 */
int gpio_request_irq(__u32 port, __u32 sig_no, enum gpio_prio prio,
                     gpio_irq_handler handler, __u32 irq_flags, 
                     const char * devname, void *dev_id);

/*!
 * This function un-registers an ISR with the GPIO interrupt module.
 *
 * @param  port         specified port with 0-GPIO port 1; 1-GPIO port 2
 * @param  sig_no       specified GPIO signal (0 based)
 * @param  prio         priority as defined in \b enum \b #gpio_prio
 */
void gpio_free_irq(__u32 port, __u32 sig_no, enum gpio_prio prio);

#endif /* __GPIO_H__ */
