/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 *
 * http://www.opensource.org/licenses/gpl-license.html
 * http://www.gnu.org/copyleft/gpl.html
 */

/*!
 * @defgroup LPMD Low-Level Power Management Driver
 */

/*!
 * @file mxc_pm.h
 *
 * @brief This file contains the  chip level configuration details and
 * public API declarations for CRM_AP module
 *
 * @ingroup LPMD
 */

#ifndef MXC_PM_H

#define MXC_PM_H

#define INT_SCALING             1
#define PLL_SCALING             0

#define INT_ERROR               132

#define HI_VOLT                 16
#define MI_VOLT                 14
#define LO_VOLT                 12

#define MAX_LO_FREQ             400
#define MAX_MI_FREQ             450
#define MAX_HI_FREQ             532

#define REF_FREQ                26000000
#define AHB_FREQ                133000000
#define IPG_FREQ                66500000

#define LFDF_0                  0
#define LFDF_2                  2
#define LFDF_4                  4
#define LFDF_8                  8

#define WAIT_MODE               111
#define STOP_MODE               112
#define DSM_MODE                113
/*
 * SCMA11 Break-Point Frequency below which is low frequency and
 * above which is high frequency
 */
#define BREAKPT_FREQ            400000000

#define NUM_PLL                 3

#define CHG_VF_HI               1
#define CHG_VF_LO               2
#define CHG_FREQ_HI             3
#define CHG_FREQ_LO             4

#define PAT_REF                 5
#define NO_CLOCK                6
#define ALT_PLL                 7
#define PLL_MFN                 8

#define GATE_STOP_WAIT          9
#define GATE_STOP               10

/*
 * Configuration Options
 */
#define COUNT32_RESET           1
#define COUNT32_NOT_RESET       0

#define PLL_OPTION_SELECT       1
#define PLL_OPTION_NOT_SELECT   0

#define CONFIG_RESET_COUNT32    COUNT32_RESET

#ifdef PLL_SCALING
/* Any one of the following other than PLL_NOCLOCK can be chosen */
#ifdef SCM_PASS_2
#define CONFIG_PLL_NOCLOCK      PLL_OPTION_NOT_SELECT
#endif
/* PLL scaling using PLL_NOCLOCK is not supported on SCM Pass 1 */
#define CONFIG_PLL_NOCLOCK      0
#define CONFIG_PLL_PATREF       PLL_OPTION_SELECT
#define CONFIG_ALT_PLL          PLL_OPTION_NOT_SELECT
#define CONFIG_PLL_MFN          PLL_OPTION_NOT_SELECT

#endif

/* Software Workarounds for Hardware issues */
/* If SDRAMC issue exists then CRM issue will exist */
#define NO_SDRAMC_RATIO_ISSUE   0
#define CRM_ISSUE               1
#define CRM_ISSUE_INT           1
#define MAX_INT_FREQ            133000000

/*
 * For SCM Pass1, Integer DVFS greater than 133MHz is not allowed
 * due to the hardware issue
 */
#define INTEGER_DVFS_NOT_ALLOW  1

/*
 * If PLL freq is less than desired ARM frequency during Integer
 * DVFS, then return this error
 */
#define PLL_LESS_ARM_ERR        2

/*
 * Frequency change within the same-lo voltage is not approved.
 * Inorder to do Integer DFS, move to the high voltage range and
 * then set LFDF and move to the low voltage range
 */
#define INT_DFS_LOW_NOT_ALLOW   3

/*
 * If the desired AHB or IPG exceeds 133MHz or 66.5MHz respectively,
 * then return this error
 */
#define AHB_IPG_EXCEED_LIMIT    4

/*
 * CKOH pins configuration
 */
#define CKOH_AP_SEL             0x00001000
#define CKOH_AHB_SEL            0x00002000
#define CKOH_IP_SEL             0x00003000

/*!
 * Additional define for stop mode
 */
#define PM_SUSPEND_STOP         ((__force suspend_state_t) 2)

/*!
 * Implementing steps required to set Integer Scaling
 *
 * @param   armfreq    The desired ARM frequency. AHB and IP
 *                     frequency are changed depending on ARM
 *                     frequency and the divider values.
 * @param   ahbfreq    The desired AHB frequency
 * @param   ipfreq     The desired IP frequency
 *
 * @return             Returns 0 on success or
 *                     Returns -1 on error
 */
int mxc_pm_intscale(long armfreq, long ahbfreq, long ipfreq);

/*!
 * Implementing steps required to set PLL Scaling using patref, noclock, alt_pll
 *
 * @param   armfreq    The desired ARM frequency.AHB and IP
 *                     frequency are changed depending on ARM
 *                     frequency and the divider values.
 * @param   ahbfreq    The desired AHB frequency
 * @param   ipfreq     The desired IP frequency
 *
 * @return             Returns 0 on success or
 *                     Returns -1 on error
 */
int mxc_pm_pllscale(long armfreq, long ahbfreq, long ipfreq);

/*!
 * Implementing Level 1 CRM Gate Control. Level 2 gate control
 * is provided at module level using LPMD registers
 *
 * @param   group   The desired clock gate control register bits.
 *                  Possible values are 0 through 6
 * @param   opt     The desired option requesting clock to run during stop
 *                  and wait modes or just during the stop mode. Possible
 *                  values are GATE_STOP_WAIT and GATE_STOP.
 *
 */
void mxc_pm_clockgate(int group, int opt);

/*!
 * Implementing steps required to transition to low-power modes
 *
 * @param   mode    The desired low-power mode. Possible values are,
 *                  WAIT_MODE, STOP_MODE or DSM_MODE
 *
 */
void mxc_pm_lowpower(int mode);

/*!
 * Configuring CKOH pins to measure changed frequency
 *
 * @param   opt     The desired clock needed to measure. Possible
 *                                      values are, CKOH_AP_SEL, CKOH_AHB_SEL or CKOH_IP_SEL
 *
 */
void mxc_pm_ckohsel(unsigned int opt);

/*!
 * To find the suitable scaling techniques if user is unknown of the choice
 * between Integer scaling and PLL scaling
 *
 * @param   armfreq    The desired ARM frequency
 * @param   ahbfreq    The desired AHB frequency
 * @param   ipfreq     The desired IP frequency
 *
 * @return             Returns 0 on success
 */
int mxc_pm_dvfs(unsigned long armfreq, long ahbfreq, long ipfreq);

/*!
 * Configuring Atlas Regulators to supply 1.6V always
 */
void mxc_pm_atlas_high(void);

/*!
 * Configuring Atlas Regulators to supply 1.2V always
 */
void mxc_pm_atlas_low(void);

/*!
 * Configuring Atlas Regulators to supply 1.6V at Hi voltage and
 * 1.2V at Lo voltage
 */
void mxc_pm_atlas_hilo(void);

#endif
