/*
 * Copyright 2004 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*!
 * @defgroup MXC_Security Security Drivers
 */

/*!
 * @file mxc_security_api.h
 *
 * @brief MXC Security user header file.
 *
 * This header file provides interface for accessing the Security module
 * mentioned below. It contains declarations of datatypes and functions. The 
 * modules that it contains are:
 * SCC (Security Controller), RNGA (Random Number Generator Accelerator), HAC
 * (Hashing Controller).
 *
 * \b SCC: It is used to bring the module to Secure state where encryption and
 * decryption of data takes place. It contains sub-modules whose description are
 * as given below:
 *
 *         \b SCM (Secure Memory Controller): Controls the Secure Memory (Red
 *         Memory and Black Memory). This memory is used for encryption and
 *         decryption of data.
 *
 *         \b SMN (Secure Monitor): Monitors the status of the SCC module.
 *
 *         \b AIC (Algorithm Integrity Checker) has 3 blocks: ASC (Algorithm
 *         Sequence Checker), BitBank and PC (Plaintext/Ciphertext) comparator.
 *         These module monitors the sequence checking and functionalities
 *         related to checking the plain text and cipher text.
 *
 *         \b Security \b Timer: Timer used during initial boot up sequence and
 *         as a watchdog timer.
 *
 * \b RNGA is used to generate random numbers. It can be seeded to generate a
 * different set of random numbers.
 *
 * \b HAC is used to Hash the data stored in the memory. It can even Hash the
 * data in non-contiguous memory locations.
 *
 * \b RTIC is used to Hash the data stored in the memory during system boot
 * and during run-time execution.
 *
 * @ingroup MXC_Security
 */

#ifndef _MXC_SECURITY_API_H_
#define _MXC_SECURITY_API_H_

#ifdef __KERNEL__
#include <linux/types.h>
#else
typedef unsigned char unchar;
#endif
#include <linux/device.h>

/*!
 * @name SCC module data types.
 */

/*! @{ */

/*!
 * This parameter is used to control different operations of the Security timer.
 */
typedef enum {
        /*!
         * Stops the Security Timer.
         */
        SCC_STOP_TMR,

        /*!
         * Starts the Security Timer.
         */
        SCC_START_TMR,

        /*!
         * Loads value into Security Timer Register.
         */
        SCC_LOAD_TMR,

        /*!
         * Loads value into Security Timer Register and Starts the Timer.
         */
        SCC_LOADNSTART_TMR
} scc_tmr_ctrl;

/*!
 * These values are returned for the current Ciphering status.
 */
typedef enum {
        /*!
         * Finished Ciphering of data.
         */
        SCC_CIPHERING_DONE,

        /*!
         * Busy Ciphering the data.
         */
        SCC_CIPHERING,

        /*!
         * Error in the Ciphering module.
         */
        SCC_CIPHERING_ERR
} scc_cip_sts;

/*!
 * These values are returned for state of SCC.
 */
typedef enum {
        /*!
         * SCC module in Start state.
         */
        SCC_START_STATE        = 0x00,

        /*!
         * SCC module in Zeroize state.
         */
        SCC_ZEROIZE_RAM_STATE  = 0x05,

        /*!
         * SCC module in Health check state.
         */
        SCC_HEALTH_CHECK_STATE = 0x06,

        /*!
         * SCC module in Fail state.
         */
        SCC_FAIL_STATE         = 0x09,

        /*!
         * SCC module in Secure state.
         */
        SCC_SECURE_STATE       = 0x0A,

        /*!
         * SCC module in Non-Secure state.
         */
        SCC_NON_SECURE_STATE   = 0x0C
} scc_state;

/*!
 * These values are returned while checking the booting of the processor 
 * to be Internal or External.
 */
typedef enum {
        /*!
         * Processor has booted externally.
         */
        SCC_EXTERNAL,

        /*!
         * Processor has booted internally.
         */
        SCC_INTERNAL
} scc_boot_state;

/*!
 * This parameter is used to clear the different status bits.
 */
typedef enum {
        /*!
         * Clears Zeroize fail condition.
         */
        SCC_ZEROIZE_FAIL_CLR = 0x40,

        /*!
         * Clears Security policy error condition.
         */
        SCC_SEC_POL_ERR_CLR  = 0x100,

        /*!
         * Clears ASC error condition.
         */
        SCC_ASC_ERR_CLR      = 0x200,

        /*!
         * Clears BitBank error condition.
         */
        SCC_BITBANK_CLR      = 0x400,

        /*!
         * Clears Plaintext or Ciphertext error condition.
         */
        SCC_PC_ERR_CLR       = 0x800,

        /*!
         * Clears Timer error condition.
         */
        SCC_TMR_ERR_CLR      = 0x1000,

        /*!
         * Clears Software alarm bit set.
         */
        SCC_SW_ALARM_CLR     = 0x2000,

        /*!
         * Clears the condition for illegally accessing to SCM module.
         */
        SCC_ILGAL_ACS_CLR    = 0x10000,

        /*!
         * Clears Bad key detection by Secure memory.
         */
        SCC_BAD_KEY_CLR      = 0x20000,

        /*!
         * Clears user access bit that was set due to access by user
         * that was inaccessible to the SMN.
         */
        SCC_USR_ACS_CLR      = 0x80000,

        /*!
         * Clears bit set due to accessing the reserved address space in the
         * SMN address space.
         */
        SCC_ILGAL_ADR_CLR    = 0x100000,

        /*!
         * Clears bit that was set due to access in byte or half-word mode to
         * the SMN.
         */
        SCC_BYTE_ACS_CLR     = 0x200000,

        /*!
         * Clears bit that was set due to unaligned access to the SMN.
         */
        SCC_UNALGND_ACS_CLR  = 0x400000,

        /*!
         * Clears bit that was set due to invalid bus master access to the SMN.
         */
        SCC_ILGAL_MSTR_CLR   = 0x1000000,

        /*!
         * Clears bit that was set due to cacheable access to the SMN.
         */
        SCC_CACHE_ACS_CLR    = 0x2000000
} scc_smn_sts_clr;

/*!
 * These values are returned for the status of Zeroizing module.
 */
typedef enum {
        /*!
         * Busy zeroizing.
         */
        SCC_ZEROIZING    = 0x02,

        /*!
         * Zeroizing failed.
         */
        SCC_ZEROIZE_FAIL = 0x10,

        /*!
         * Finished zeroizing .
         */
        SCC_ZEROIZE_DONE = 0x0200
} scc_zeroize_state;

/*!
 * This parameter is used to configure the SCM or SMN module interrupt.
 */
typedef enum {
        /*!
         * Enable interrupt.
         */
        SCC_INT_ENABLE,

        /*!
         * Disable interrupt.
         */
        SCC_INT_DISABLE,

        /*!
         * Clear interrupt.
         */
        SCC_CLR_INT
} scc_intr;

/*!
 * These values are returned for the availability of SCC module.
 */
typedef enum {
        /*!
         * SCC module is enabled.
         */
        SCC_ENABLED,

        /*!
         * SCC module is disabled.
         */
        SCC_DISABLED
} scc_enable;

/*!
 * These parameters are used as modes of operation for AIC module.
 */
typedef enum {
        /*!
         * This parameter is used to read Increment size register of the 
         * BitBank module.
         */
        SCC_READ_INCR,

        /*!
         * This parameter is used to read Decrement size register of the 
         * BitBank module.
         */
        SCC_READ_DECR,

        /*!
         * This parameter is used to write data to the Plaintext register. 
         */
        SCC_WRITE_PLAIN_TEXT,

        /*!
         * This parameter is used to write data to the Ciphertext register. 
         */
        SCC_WRITE_CIPHER_TEXT,
} scc_aic_param;

/*!
 * These values are returned for status of AIC module.
 */
typedef enum {
        /*!
         * Value returned when an ASC error was detected.
         */
        SCC_ASC_ERR,

        /*!
         * Value returned when a Bit Bank Error was detected.
         */
        SCC_BIT_BNK_ERR,

        /*!
         * Value returned when Plain/Cipher text error was detected.
         */
        SCC_PC_ERR,

        /*!
         * Value returned when there is no error in the AIC module.
         */
        SCC_AIC_NO_ERROR,
        /*!
         * Return value to check for any error in the AIC module.
         */
        SCC_AIC_ERROR
}scc_aic_ret;

/*!
 * These values are returned for Security Timer status.
 */
typedef enum {
        /*!
         * Security Timer encountered error.
         */
        SCC_TMR_ERR,

        /*!
         *  No Security Timer error.
         */
        SCC_NO_TMR_ERR
} scc_tmr_sts;

/*!
 * These values are returned for SCC routines.
 */
typedef enum {
        /*!
         * Successfully finished the routine.
         */
        SCC_SUCCESS,

        /*!
         * Encountered some errors in the routine.
         */
        SCC_FAILURE
} scc_ret;

/*! @} */

/*!
 * @name RNGA module data types.
 */

/*! @{ */

/*!
 * These parameters are used for the configuring RNGA.
 */
typedef enum {
        /*!
         * High assurance bit used for configuring the RNGA module to notify
         * SCC of any security violation.
         */
        RNGA_HIGH_ASSUR = 0x02,

        /*!
         * Interrupt mask bit to configure RNGA module for masking the
         * interrupt.
         */
        RNGA_INTR_MSK = 0x04
} rnga_config;

/*!
 * These parameters are used for configuring RNGA to different testing modes.
 */
typedef enum {
        /*!
         * This parameter is used for configuring RNGA module to 
         * verification mode.
         */
        RNGA_VRFY_MODE = 0x01,

        /*!
         * This parameter is used for configuring RNGA module to Oscillator 
         * test frequency mode.
         */
        RNGA_OSC_FREQ_TEST = 0x02
} rnga_mode;

/*!
 * This parameter is used for configuring the Verify Control register.
 */
typedef enum {
        /*!
         * The shift register clocks are on.
         */
        RNGA_SHFT_CLK_ON = 0x01,

        /*!
         * The system clocks are on.
         */
        RNGA_SYS_CLK_ON = 0x02,

        /*!
         * Resets the shift registers.
         */
        RNGA_RESET_SHFT_REG = 0x04
} rnga_vrfy_ctrl;

/*!
 * This parameter is used for specifying the oscillator number.
 */
typedef enum {
        /*!
         * Oscillator 1 selected.
         */
        RNGA_OSCILLATOR1 = 0x01,

        /*!
         * Oscillator 2 selected.
         */
        RNGA_OSCILLATOR2
} rnga_osc_num;

/*!
 * This parameter is used for switching for read or write.
 */
typedef enum {
        /*!
         * Read from the register.
         */
        RNGA_READ,

        /*!
         * Write to the register.
         */
        RNGA_WRITE
} rnga_readwrite;

/*!
 * These values are returned for RNGA routines.
 */
typedef enum {
        /*!
         * Successfully finished the routine.
         */
        RNGA_SUCCESS,

        /*!
         * Encountered some errors in the routine.
         */
        RNGA_FAILURE
} rnga_ret;

/*! @} */

/*!
 * @name HAC module data types for Argonplus and SCMA11.
 */

/*! @{ */

/* As HAC module is not present on MX31 platform, these declarations will not be
 * included if it is compiled as MX31 platform.
 */
#ifndef       CONFIG_ARCH_MX3

/*!
 * These values are returned to indicate the status after the Hashing process
 * in the HAC module.
 */
typedef enum {
        /*!
         * Hashing of data is completed.
         */
        HAC_DONE,

        /*!
         * Busy hashing the data.
         */
        HAC_BUSY,

        /*!
         * Hashing was aborted due to bus error during HAC access over
         * the AHB (Advanced High-Performance Bus).
         */
        HAC_ERR,
        
        /*!
         * Hashing status unknown. This may be when the hashing process has not
         * been initiated altleast once or 'ERROR' bit or 'DONE' bits were 
         * reset after the hashing process was completed.
         */
        HAC_UNKNOWN
} hac_hash_status;

/*!
 * These parameters are used for controlling the Hashing action to be 
 * performed.
 */
typedef enum {
        /*!
         * Start the Hashing process.
         */
        HAC_START,

        /*!
         * Start the Hashing of data with last block.
         */
        HAC_START_LAST, 

        /*!
         * Continue the Hashing process.
         */
        HAC_CONTINUE,

        /*!
         * Continue the Hashing of data with last block.
         */
        HAC_LAST
} hac_hash;

/*!
 * These values are returned for HAC routines.
 */
typedef enum {
        /*!
         * Successfully finished the routine.
         */
        HAC_SUCCESS,

        /*!
         * Encountered some errors in the routine.
         */
        HAC_FAILURE,

        /*!
         * HAC module is busy in Hashing process.
         */
        HAC_HASH_BUSY
} hac_ret;

/*!
 * Gets Hashing result value from HSH Registers.
 */
typedef struct {
        /*!
         * Stores the hash result.
         */
        ulong hash_result[5];
} hac_hash_rlt;

#endif /* CONFIG_ARCH_MX3 */

/*! @} */

/*!
 * @name RTIC module data types.
 */

/*! @{ */

/*!
 * These parameters are used for selecting RTIC mode of operation.
 */  
typedef enum {
        /*!
         * RTIC in ONE_TIME Hash Mode.
         */
        RTIC_ONE_TIME,
        
        /*!
         * RTIC in RUN_TIME Hash mode.
         */ 
        RTIC_RUN_TIME 
} rtic_mode;

/*!
 * These parameter are used to enable or disable the interrupt.
 */
typedef enum {
        /*!
         * RTIC interrupt enable.
         */
        RTIC_INTERRUPT_ENABLE,

        /*!
         * RTIC interrupt disable.
         */                          
        RTIC_INTERRUPT_DISABLE
} rtic_interrupt; 

/*!
 * These parameters are used to select RTIC memory block for Hashing.
 */
typedef enum {
        /*!
         * Select Memory Block A1.
         */  
        RTIC_A1,

        /*!
         * Select Memory Block A2.
         */
        RTIC_A2,
        
        /*!
         * Select Memory Block B1.
         */
        RTIC_B1,

        /*!
         * Select Memory Block B2.
         */
        RTIC_B2,

        /*!
         * Select Memory Block C1.
         */
        RTIC_C1, 
        
        /*!
         * Select Memory Block C2.
         */
        RTIC_C2,          
        
        /*!
         * Select Memory Block D1.
         */
        RTIC_D1,

        /*!
         * Select Memory Block D2.
         */
        RTIC_D2
} rtic_memblk;

/*!
 * These values are returned when queried for the status of the RTIC module.
 */
typedef enum {
        /*!
         * Busy hashing the data.
         */
        RTIC_STAT_HASH_BUSY = 0x0,

        /*!
         * Hashing of data is completed.
         */
        RTIC_STAT_HASH_DONE = 0x02,

        /*!
         * Hashing was aborted due to corruption in memory block
         * during RUN_TIME Hashing or address/length error has
         * occurred.
         */                                 
        RTIC_STAT_HASH_ERR = 0x04,

        /*!
         * Hashing status unknown. This may happen when: 1) Hashing process has
         * not been initiated atleast once. 2) Error bit or Done bit was reset
         * after the hashing process was completed.
         */
        RTIC_UNKNOWN
} rtic_status;

/*!
 * These values are returned for RTIC routines.
 */
typedef enum {
        /*!
         * Successfully finished the routine.
         */
        RTIC_SUCCESS,

        /*!
         * Encountered some errors in the routine.
         */
        RTIC_FAILURE
} rtic_ret;

/*!
 * Gets Hashing result value from RTIC Hash Result Registers.
 */
typedef struct {
        /*!
         * Stores the hash result.
         */
        ulong hash_result[5];
} rtic_hash_rlt;

/*! @} */

/*!
 * Structure for accessing the sizes of the SCC module.
 */
typedef struct {
        /*!
         * Block size of the Red or Black Memory.
         */
        unchar scc_blk_size;

        /*!
         * Red Memory area size.
         */
        ushort scc_red_size;

        /*!
         * Black Memory area size.
         */
        ushort scc_black_size;
} scc_mem_size;

/*!
 * @name SCC module APIs.
 */

/*! @{ */

/*!
 * This API is used to read the status of the Secure Memory module.
 *
 * @return  Status of the SCM module.
 */
ulong scc_scm_read_sts(void);

/*!
 * This API is used to check if SCC is enabled or disabled.
 *
 * @return  SCC is enabled or disabled.
 */
scc_enable scc_check_enable(void);

/*!
 * This API configures Red Memory start address, Black Memory start address,
 * number of blocks of data and the init vector values of the KEM module (SCM).
 *
 * @param   red_addr       starting address of the red memory area.
 * @param   black_addr     starting address of the black memory area.
 * @param   cipher_length  number of blocks that needs to be ciphered.
 * @param   init_vect1     upper 32 bits of the initialization vector required
 *                         at the start of CBC (Cipher Block Chaining) mode of
 *                         ciphering.
 * @param   init_vect2     lower 32 bits of the initialization vector required
 *                         in CBC (Cipher Block Chaining) mode of ciphering at
 *                         the start of the processing of data.
 *
 * @return  SCC_SUCCESS  Successfully configured the KEM module.\n
 *          SCC_FAILURE  Encountered error while configuring the KEM module.
 *                       Length crosses maximum limit specified.
 */
scc_ret scc_kem_config(ulong red_addr, ulong black_addr, ulong cipher_length,
                            ulong init_vect1, ulong init_vect2);

/*!
 * This API is used to load the plain data in the Red memory area and is
 * encrypted. Memory is allocated to the size that is configured in
 * the length register by the user and the pointer is passed as parameter.
 *
 * @param  encrypt_data  pointer to the data that needs to be encrypted.
 * @param  encrypt_length length (blocks) of data that needs to be encrypted.
 *
 * @return  SCC_SUCCESS  Successfully encrypted the data in the Red memory.\n
 *          SCC_FAILURE  Encountered error while encrypting data in Red memory.
 *                       Encryption length incorrect.
 */
scc_ret scc_kem_load_data2encrypt(ulong *encrypt_data, unchar encrypt_length);

/*!
 * This API is used to load the data to the Black memory area and is decrypted.
 * The data is loaded into the Black memory area and is decrypted. Memory needs
 * to be allocated by the user to the size that is configured in the length
 * register and a pointer is passed as a parameter to this API.
 *
 * @param  decrypt_data  pointer to the data that needs to be decrypted.
 * @param  decrypt_length length (blocks) of data that needs to be decrypted.
 *
 * @return  SCC_SUCCESS  Successfully decrypted the data in the Black memory.\n
 *          SCC_FAILURE  Encountered error while decrypting data in Black
 *                       memory. Decryption length incorrect.
 */
scc_ret scc_kem_load_data2decrypt(ulong *decrypt_data, unchar decrypt_length);

/*!
 * After ciphering (encryption) API is called, this API reads the data from the
 * Black memory area. Memory needs to be allocated by the caller to a size equal
 * to that as configured in the length register and passed as pointer to this
 * API.
 *
 * @param  encrypted_data  pointer where data is to be copied from the Black
 *                         Memory area.
 * @param  encrypt_length  length(blocks)of data that needs to copied.
 *
 * @return  SCC_SUCCESS  Successful in copying the data to the memory.\n
 *          SCC_FAILURE  Encountered error while processing the request.
 *                       Encrypt length crosses the max limit specified.
 */
scc_ret scc_kem_get_encrypt_data(ulong *encrypted_data, unchar encrypt_length);

/*!
 * This API reads the data from the Red memory area. It is called after
 * data has been decrypted by loading data in the Black Memory area. Data is
 * read from the Red memory area and is stored in the pointer that is
 * pre-allocated to the number of bytes that is configured in the SCC length
 * register.
 *
 * @param  decrypted_data  pointer where data is to be copied from the Red
 *                         Memory area.
 * @param  decrypt_length  length(blocks)of data that needs to copied.
 *
 * @return  SCC_SUCCESS  Successful in copying the data to the memory.\n
 *          SCC_FAILURE  Encountered error while processing the request.
 *                       Decrypt length crosses the max limit specified.
 */
scc_ret scc_kem_get_decrypt_data(ulong *decrypted_data, unchar decrypt_length);

/*!
 * This API returns the status of ciphering process.
 *
 * @return  SCC_CIPHERING       KEM module still busy ciphering the data.\n
 *          SCC_CIPHERING_DONE  KEM module finished ciphering the data.\n
 *          SCC_CIPHERING_ERR   KEM module encountered error.
 */
scc_cip_sts scc_kem_cipher_status(void);

/*!
 * This API enables/disables the interrupt for SCM module.
 *
 * @param  config_int  enables or disables the interrupt.
 *
 * @return  SCC_SUCCESS  Successful in enabling the interrupt.\n
 *          SCC_FAILURE  Received an unexpected input.
 */
scc_ret scc_scm_config_int(scc_intr config_int);

/*!
 * This API clears SCM interrupts that have been generated.
 *
 * @param  clr_int  clears the interrupt status.
 *
 * @return  SCC_SUCCESS  Successful in clearing the SCM interrupt.\n
 *          SCC_FAILURE  Received an unexpected input.
 */
scc_ret scc_scm_clear_int(scc_intr clr_int);

/*!
 * This API returns the memory size information for Secure RAM module.
 *
 * @param  scc_size  pointer to a structure to load the details pertaining to
 *                   the size information.
 */
scc_ret scc_kem_get_size_info (scc_mem_size *scc_size);

/*!
 * This API returns the version ID of Secure Memory module.
 *
 * @return   Version ID of the Secure Memory module.
 */
ulong scc_scm_get_ver(void);

/*!
 * This API returns the status of the SCM module.
 *
 * @return  Status of the SCM module.
 */
ulong scc_scm_get_status(void);

/*!
 * This API returns the error status of the SCM module.
 *
 * @return   Error status of the SCM module.
 */
ulong scc_scm_err_status(void);

/*!
 * This API returns the status of the SMN module.
 *
 * @return   Status of the SMN module.
 */
ulong scc_smn_get_status(void);

/*!
 * This API returns the current State of the SCC module.
 *
 * @return  State of the SCC module like Start, Zeroize RAM,
 *          Health Check, Fail, Secure, Non-Secure
 */
scc_state scc_get_state(void);

/*!
 * This API returns the version ID of Secure Monitor.
 *
 * @return  Version ID of the Secure Monitor.
 */
ulong scc_smn_get_ver(void);

/*!
 * This API returns the boot state of the processor.
 *
 * @return  SCC_INTERNAL  Processor has booted internally.\n
 *          SCC_EXTERNAL  Processor has booted externally.
 */
scc_boot_state scc_check_boot(void);

/*!
 * This API clears the status bits from the SMN status register. Values are
 * considered from the smn_sts_clr data type.
 *
 * @param    smn_status_clr  to clear the status bits set in the SMN status
 *                           register.
 *
 * @return  SCC_SUCCESS      Successful in clearing the status of the requested
 *                           fields.\n
 *          SCC_FAILURE      Unsuccessful due to error in the parameter passed.
 */
scc_ret scc_smn_clear_status(scc_smn_sts_clr smn_status_clr);

/*!
 * This API sets the alarm bit. Setting this bit causes the SCC module to
 * enter the Fail state. This API is called during the detection of intrusion.
 */
scc_ret scc_set_sw_alrm(void);

/*!
 * This API clears the interrupt bit set due to an interrupt from the SMN
 * module.
 */
scc_ret scc_smn_clear_int(void);

/*!
 * This API enables or disables the interrupt for the SMN module.
 *
 * @param    config_int  interrupt configuration (ENABLE/DISABLE).
 *
 * @return   SCC_SUCCESS  Successfully configured the interrupt for SMN module.
 *           SCC_FAILURE  Error in the parameter passed.
 */
scc_ret scc_smn_config_int(scc_intr config_int);

/*!
 * This API clears the Red Memory & Black Memory. This API will stop any
 * ciphering that is in progress.
 *
 * @return  SCC_SUCCESS  Zeroization process was started successfully.\n
 *          SCC_FAILURE  Zeroization process cannot be started as SCC is not in
 *                       Secure or Non-Secure state.
 */
scc_ret scc_zeroize(void);

/*!
 * This API returns the zeroization status of the Secure RAM.
 *
 * @return     Zeroization status of the Secure RAM module.
 *             (ZEROIZING/ZEROIZE_FAIL/ZEROIZE_DONE).
 */
scc_zeroize_state scc_zeroize_status(void);

/*!
 * This API clears the zeroize bit set in the SMN status register caused due to
 * failure in the zeroization module.
 */
scc_ret scc_zeroize_fail_clear(void);

/*!
 * This API puts the data sent from the upper layer to the RED memory area.
 * Memory needs to be allocated by the caller, which is equal to the size
 * requested.
 *
 * @param      redmem_data   pointer to the memory location from where data is
 *                           to be copied to the Red memory area.
 * @param      redmem_size   number of data words that needs to be copied
 *                           to the Red memory area.
 *
 * @return   SCC_SUCCESS     Data copied successfully from the buffer to the
 *                           Red memory area.\n
 *           SCC_FAILURE     Illegal or out of boundary length specified.
 */
scc_ret scc_redmem_put_data(ulong *redmem_data, ushort redmem_size);

/*!
 * This API puts the data sent from the upper layer to the BLACK memory area.
 * Memory needs to be allocated by the caller, which is equal to the size
 * requested.
 *
 * @param    blackmem_data   pointer to the memory location from where data is
 *                           to be copied to the Black memory area.
 * @param    blackmem_size   number of data words that needs to be copied
 *                           to the Black memory area.
 *
 * @return   SCC_SUCCESS     Data copied successfully from the buffer to the
 *                           Black memory area.\n
 *           SCC_FAILURE     Illegal or out of boundary length specified.
 */
scc_ret scc_blackmem_put_data(ulong *blackmem_data, ushort blackmem_size);

/*!
 * This API gets the data requested by the upper layer from the RED memory area.
 * Memory needs to be allocated by the caller, which is equal to the size
 * requested.
 *
 * @param      redmem_data   pointer to the buffer where data from the RED
 *                           memory is to be copied.
 * @param      redmem_size   number of data words that needs to be copied
 *                           to the memory.
 *
 * @return   SCC_SUCCESS     Data copied successfully to the buffer from the
 *                           Red memory area.\n
 *           SCC_FAILURE     Illegal or out of boundary length specified.
 */
scc_ret scc_redmem_get_data(ulong *redmem_data, ushort redmem_size);

/*!
 * This API gets the data requested by the upper layer from the BLACK memory
 * area. Memory needs to be allocated by the caller, which is equal to the size
 * requested.
 *
 * @param      blackmem_data pointer to the buffer where data from the RED
 *                           memory is to be copied.
 * @param      blackmem_size number of data words that needs to be copied
 *                           to the memory.
 *
 * @return   SCC_SUCCESS     Data copied successfully to the buffer from the
 *                           Black memory area.\n
 *           SCC_FAILURE     Illegal or out of boundary length specified.
 */
scc_ret scc_blackmem_get_data(ulong *blackmem_data, ushort blackmem_size);

/*!
 * This API configures the start and the end sequence number for the ASC module
 *
 * @param   asc_start_seq  starting sequence number to be configured to the ASC
 *                         module.
 * @param   asc_end_seq    ending sequence number to be configured to the ASC
 *                         module.
 *
 * @return   SCC_SUCCESS   Successfully configured the Start and End sequence
 *                         number.\n
 *           SCC_FAILURE   Illegal configuration value specified. Start or End
 *                         sequence number may have been specified as 0x00.
 */
scc_ret scc_asc_configure(ushort asc_start_seq, ushort asc_end_seq);

/*!
 * This API sets the next sequence number. This API is used for checking the
 * sequence of the algorithm.
 *
 * @param   asc_seq_chk  sequence number to be validated for the sequence flow.
 *
 * @return   SCC_SUCCESS   Successfully written the next sequence number.\n
 *           SCC_FAILURE   Illegal sequence check value (0x00) specified.
 */
scc_ret scc_asc_set_seq(ushort asc_seq_chk);

/*!
 * This API generates the next sequence number for the given current sequence
 * number. It implements the LFSR (Linear Feedback Shift Register)
 * functionality. It uses the polynomial function:
 * F(X) = X^16 + X^15 + X^13 + X^4 + 1.
 * It can be used to generate the next sequence number wherein the current
 * sequence number is known. The 16 bits current sequence number is passed as a
 * parameter to this API and the next 16 bits sequence number is calculated and
 * returned to the caller.
 *
 * @param  seq_num   current sequence number to generate the next sequence
 *                   number.
 *
 * @return           Next sequence number calculated by the polynomial
 *                   function.
 */
ushort scc_asc_get_nxtseq(ushort seq_num);

/*!
 * This API returns the error status of the AIC module.
 *
 * @return  Error status of the ASC or Bit Bank or PC of the AIC module.
 */
scc_aic_ret scc_aic_bit_status(void);

/*!
 * This API clears the Bit Bank count register of the Bit Bank module.
 */
scc_ret scc_bitbnk_clear(void);

/*!
 * This API returns the Bit count value of the Bit Bank module.
 *
 * @return   Bit count value of the Bit Bank module.
 */
ushort scc_rd_bitcnt(void);

/*!
 * This API returns the size value that should be compared between the
 * plain/ciphertext comparator
 *
 * @return   Size of the bits to be used for comparison between the
 *           plain/ciphertext comparator.
 */
unchar scc_rd_bit_cmp(void);

/*!
 * This API is used to read the Increment/Decrement size register of the Bit
 * Bank module.
 *
 * @param  read_reg   specify SCC_READ_INCR or SCC_READ_DECR register of the
 *                    Bit Bank module to be read.
 *
 * @return            Value of the Increment/Decrement register of the Bit
 *                    Bank module. If the argument passed is invalid then
 *                    SCC_AIC_ERROR is returned to the caller.
 */
scc_aic_ret scc_rd_inc_dec_reg(scc_aic_param read_reg);

/*!
 * This API configures the Bit Bank increment register; decrement
 * register and the bit compare register of the Bit Bank module.
 *
 * @param  bit_bnk_incr  bit bank increment size.
 * @param  bit_bnk_decr  bit bank decrement size.
 * @param  bit_cmp       bit bank compare size.
 *
 * @return  SCC_SUCCESS  Successfully configured the Bit Bank module.\n
 *          SCC_FAILURE  Parameters passed has some invalid data.
 */
scc_ret scc_bitbnk_configure(ushort bit_bnk_incr, ushort bit_bnk_decr, unchar
                           bit_cmp);

/*!
 * This API is used to write data to the Plain/Cipher text register that is used
 * to compare the plain/ciphertext data as a software assurance feature.
 *
 * @param    cipher_data   data that needs to be written to the Plaintext check
 *                         register or the Ciphertext check register.
 * @param    cipher_switch parameter to indicate where to write the sent
 *                         data (Plaintext check register/Ciphertext check
 *                         register).
 *
 * @ return  Invalid argument passed for scc_aic_param parameter.
 */
scc_ret scc_wr_pln_cip(ulong cipher_data, scc_aic_param cipher_switch);

/*
 * This API loads the Security Timer register with the value passed as
 * parameter.
 */
scc_ret scc_ldstrt_sec_tmr(unchar tmr_ctrl);

/*!
 * This API configures the Security Timer with the new reload timer value.
 *
 * @param  reload_tmr_val  re-load timer value for the Security Timer.
 */
scc_ret scc_sec_config_tmr(ulong reload_tmr_val);

/*!
 * This API configures the Security timer so as to Load or Start or
 * Load and Start or Stop.
 *
 * @param   ld_strt_val  configuration values to Load or Start or Load & Start
 *                       or Stop the Security Timer.
 *
 * @return  SCC_SUCCESS  Successfully configured the Security timer.\n
 *          SCC_FAILURE  Invalid parameter is passed.
 */
scc_ret scc_sec_ctrl_tmr(unchar ld_strt_val);

/*!
 * This API reads the current value of the Security Timer.
 *
 * @return  Current Security Timer value.
 */
ulong scc_sec_rd_tmr(void);

/*!
 * This API reads the current reload value of the Security Timer.
 *
 * @return  Current Security Timer reload value.
 */
ulong scc_sec_rload_tmr(void);

/*!
 * This API returns the status of the Security Timer error indicated in the SMN
 * Status register
 *
 * @return  SCC_TMR_ERR      Security Timer has expired.\n
 *          SCC_NO_TMR_ERR   Security Timer has not expired.
 */
scc_tmr_sts scc_sec_rd_err_status(void);

/*!
 * This API returns the Debug Detect status of the SCC module.
 *
 * @return  Status of the Debug Detect of the SCC module.
 */
ulong scc_dbdet_rd_status(void);

/*!
 * This function is called to put the SCC in a low power state. Refer to the
 * document driver-model/driver.txt in the kernel source tree for more
 * information.
 *
 * @param   dev   the device structure used to give information on HAC
 *                to suspend.
 * @param   state the power state the device is entering.
 * @param   level the stage in device suspension process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
scc_ret scc_suspend(struct device *dev, ulong state, ulong level);

/*!
 * This function is called to bring the SCC back from a low power state.
 * Refer to the document driver-model/driver.txt in the kernel source tree
 * for more information.
 *
 * @param   dev   the device structure used to give information on SCC
 *                to resume.
 * @param   level the stage in device resumption process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
scc_ret scc_resume(struct device *dev, ulong level);

/*! @} */

/*!
 * @name RNGA module APIs.
 */

/*! @{ */

/*!
 * This API configures the RNGA module for interrupt enable or disable. It
 * configures the module to be configured for High assurance boot (HAB).
 * Enabling HAB will inform the SCC in case of any Security violation.
 *
 * @param   rnga_cfg   configuration value for RNGA.
 *
 * @return  RNGA_SUCCESS  Successfully configured the RNGA module.\n
 *          RNGA_FAILURE  Invalid parameter passed.
 */
rnga_ret rnga_configure(rnga_config rnga_cfg);

/*!
 * This API returns the status of the RNGA module.
 *
 * @return   Status of the RNGA module.
 */
ulong rnga_get_status(void);

/*!
 * This API returns the total count value of the random stored in the RNGA FIFO
 *
 * @return  Total count value of random numbers stored in the RNGA FIFO.
 */
unchar rnga_get_fifo_level(void);

/*!
 * Seeds the Random number generator with the new seed value.
 *
 * @param  entropy_val seed for the random number generator.
 */
rnga_ret rnga_seed_rand(ulong entropy_val);

/*!
 * This API gets the random number from the FIFO. If RNGA is in power down mode
 * (Sleep mode), no random numbers are generated. In such cases, this API checks
 * if any random numbers are available in the FIFO. It will return any available
 * random numbers from the FIFO or return zero indicating that the module is not
 * generating any random numbers.
 *
 * @return  Random number or zero (0x00) if there are no random number available
 *          in the FIFO.
 */
ulong rnga_get_rand(void);

/*!
 * This API clears the RNGA interrupt
 */
rnga_ret rnga_clear_int(void);

/*!
 * This API returns the FIFO size of the RNGA
 *
 * @return  FIFO size of the RNGA module.
 */
unchar rnga_get_fifo_size(void);

/*!
 * This API is used to set the mode of operation of RNGA module to either
 * Verification mode or oscillator frequency test mode.
 *
 * @param        rngamode  mode in which RNGA needs to be operated.
 *
 * @return   RNGA_SUCCESS  Successfully configured RNGA mode.
 *           RNGA_FAILURE  Invalid parameter passed.
 */
rnga_ret rnga_configure_mode(rnga_mode rngamode);

/*!
 * This API is brings the RNGA module to normal state from power down mode
 * (Sleep mode). This is used in scenario where the module is been put to sleep
 * mode and needs to be brought back to normal mode.
 */
rnga_ret rnga_normal(void);

/*!
 * This API puts the RNGA module to sleep mode. After placing the RNGA module
 * to sleep mode (low power mode), it stops generating the random numbers.
 */
rnga_ret rnga_sleep(void);

/*!
 * This API is used to control verification clocks. It is used reset the shift
 * registers, use system clock and to turn on the shift clock.
 *
 * @param        rnga_vrfy_control configures the RNGA verification controller.
 *
 * @return   RNGA_SUCCESS  Successfully configured the RNGA verification
 *                         controller.
 *           RNGA_FAILURE  Parameter passed has invalid data.
 */
rnga_ret rnga_config_vrfy_ctrl(rnga_vrfy_ctrl rnga_vrfy_control);

/*!
 * This API is used for the oscillator test logic to specify the length of
 * the count oscillator clock pulses.
 *
 * @param       osc_count_ptr   counter value of oscillator.
 * @param       rnga_rw         read/write data from/to oscillator counter.
 *
 * @return   RNGA_SUCCESS  Successfully preformed the operation.
 *           RNGA_FAILURE  Invalid parameter passed.
 */
rnga_ret rnga_config_osc_cntctrl(ulong *osc_count_ptr, rnga_readwrite rnga_rw);

/*!
 * This API is used to count the number of oscillator pulses received from
 * Oscillator #1/#2 starting from the time the Oscillator Counter Control
 * Register is written and ending at the time the Oscillator Counter Control
 * Register reaches zero. If the oscillator number is other than that specified,
 * it will return 0.
 *
 * @param       osc_num         oscillator number.
 *
 * @return      Count stored in the oscillator specified.
 */
ulong rnga_osc_cnt(rnga_osc_num osc_num);

/*!
 * This API gives the oscillator counter which have toggled 0x400 times. If bit
 * 0 is set then oscillator #1 and if bit 1 is set the oscillator #2 has crossed
 * 0x400 times. If both the bits are set then both the oscillators have crossed
 * 0x400 times.
 *
 * @return      Oscillators that have toggled 0x400 times.
 */
rnga_osc_num rnga_osc_status(void);

/*!
 * This function is called to put the RNGA in a low power state. Refer to the
 * document driver-model/driver.txt in the kernel source tree for more
 * information.
 *
 * @param   dev   the device structure used to give information on RNGA
 *                to suspend.
 * @param   state the power state the device is entering.
 * @param   level the stage in device suspension process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
rnga_ret rnga_suspend(struct device *dev, ulong state, ulong level);

/*!
 * This function is called to bring the RNGA back from a low power state.
 * Refer to the document driver-model/driver.txt in the kernel source tree
 * for more information.
 *
 * @param   dev   the device structure used to give information on RNGA
 *                to resume.
 * @param   level the stage in device resumption process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
rnga_ret rnga_resume(struct device *dev, ulong level);

/*! @} */

/*!
 * @name HAC module APIs.
 */

/*! @{ */

/* HAC module is not present on MX31 Platform.*/
#ifndef         CONFIG_ARCH_MX3

/*!
 * This API configures the start address and block length of the data that needs
 * to be hashed. Start address indicates starting location from where the data
 * in the flash memory is to be hashed. The number of blocks that needs to be
 * hashed is loaded in the block count register. This API also does the stating
 * of Hashing process or continue with Hashing of next block of data configured
 * in the START_ADDR, BLOCK_COUNT register depending on the hash parameter
 * passed. 
 *
 * @param   start_address   starting address of the flash memory to be hashed.
 * @param   blk_len         number of blocks to be hashed.
 * @param   hash            mode of operation like Start or Continue hashing.
 *                          Following parameters are passed:
 *                          HAC_START       : Starts the Hashing process.
 *                          HAC_LAST_START  : Starts the Hashing process with
 *                                            last block of data.
 *                          HAC_CONTINUE    : Continue the Hashing process.
 *                          HAC_LAST        : Continue the Hashing process with
 *                                            last block of data.
 *
 * @return  HAC_SUCCESS    Success on completion of configuring HAC for
 *                         Hashing.\n
 *          HAC_FAILURE    Failure in completion.
 */
hac_ret hac_hash_data (ulong start_address, ulong blk_len, hac_hash hash);

/*!
 * This API returns the status of the Hashing.
 *
 * @return      HAC_BUSY : Indicated HAC Module is busy with Hashing.\n
 *              HAC_DONE : Indicates Hashing of data is done.\n
 *              HAC_ERR  : Indicates error has occurred during Hashing.
 *              HAC_UNKNOWN: Hashing status unknown. This may be when the
 *                           hashing process has not been initiated atleast
 *                           once or 'ERROR' bit or 'DONE' bits were reset
 *                           after the hashing process was completed.
 */
hac_hash_status hac_hashing_status(void);

/*!
 * This API returns the status of the Hash module.
 *
 * @return      Value of the Hashing control register.
 */
ulong hac_get_status(void);

/*!
 * This API stops the Hashing process when the Hashing is in progress.
 */
hac_ret hac_stop(void);

/*!
 * This API reads 160 bit hash result from Hash result register. The data is
 * copied to the memory pointed by the input pointer.
 *
 * @param    hash_result_reg    structure Pointer where the hash result is
 *                              copied.
 */
hac_ret hac_hash_result(hac_hash_rlt *hash_result_reg);

/*!
 * This function is called to put the HAC in a low power state. Refer to the
 * document driver-model/driver.txt in the kernel source tree for more
 * information.
 *
 * @param   dev   the device structure used to give information on HAC
 *                to suspend.
 * @param   state the power state the device is entering.
 * @param   level the stage in device suspension process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
hac_ret hac_suspend(struct device *dev, ulong state, ulong level);

/*!
 * This function is called to bring the HAC back from a low power state.
 * Refer to the document driver-model/driver.txt in the kernel source tree
 * for more information.
 *
 * @param   dev   the device structure used to give information on HAC
 *                to resume.
 * @param   level the stage in device resumption process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
hac_ret hac_resume(struct device *dev, ulong level);

/*! @} */
#endif /* CONFIG_ARCH_MX3 */

/*!
 * @name RTIC module APIs.
 */

/*! @{ */

/*!
 * This API configures the memory block (A, B, C, D) into
 * ONE_TIME Mode or RUN_TIME Mode for Hashing of data in memory.
 * RTIC does not support enabling multiple memory blocks that aren't
 * grouped together(i.e. enabling only memory blocks A & C without memory
 * B enabled).
 *
 * @param  mode         RTIC mode of operation (ONE_TIME or RUN_TIME).
 * @param  mem_blk      Memory block to be hashed.
 *
 * @return  RTIC_SUCCESS    Successfully hashed the data.\n
 *          RTIC_FAILURE    Error in the parameters passed.
 */
rtic_ret rtic_configure_mode(rtic_mode mode, rtic_memblk mem_blk);

/*!
 * This API allows to configure start address and block length of the
 * memory content to be hashed. Start address indicates starting location from
 * where the data in the memory is to be hashed. The number of blocks that
 * needs to be hashed is loaded in the block count register.
 * There are four memory blocks available. The user can configure any one of
 * these four memory blocks by passing their appropriate address and block
 * length to be Hashed.
 *
 * @param    start_addr       Starting address of the memory to be hashed.
 * @param    blk_len          Block length of data in memory.
 * @param    mem_blk          Memory block to be hashed.
 *
 * @return  RTIC_SUCCESS    Successfully hashed the data.\n
 *          RTIC_FAILURE    Error in the parameters passed.
 */

rtic_ret rtic_configure_mem_blk(ulong start_addr, ulong blk_len, \
                                 rtic_memblk mem_blk);

/*!
 * This API will configure to start the Hashing of data in memory
 * either in One-Time Hash mode or Run-Time Hash mode.
 *
 * @param       mode    Memory block to be hashed.
 *
 * @return  RTIC_SUCCESS    Successfully hashed the data.\n
 *          RTIC_FAILURE    Error in the parameters passed.
 */
rtic_ret rtic_start_hash(rtic_mode mode);

/*!
 * This API will read the RTIC status register.
 *
 * @return  Status of Hashing.
 */
ulong rtic_get_status(void);

/*!
 * This API will read the RTIC control register.
 *
 * @return  Control register value.
 */
ulong rtic_get_control(void);

/*!
 * This API enables or disables interrupt for RTIC module.
 *
 * @param  irq_en    To enable or disable interrupt.
 *
 * @return  RTIC_SUCCESS    Successfully hashed the data.\n
 *          RTIC_FAILURE    Error in the parameters passed.
 */
rtic_ret rtic_configure_interrupt(ulong irq_en);

/*!
 * This API reads the Fault address of the RTIC module.
 *
 * @return   Fault address register value.
 */
ulong  rtic_get_faultaddress(void);

/*!
 * This API reads 160 bit hash result from Hash result register.The data is
 * copied to the memory pointed by the input pointer.
 *
 * @param    mem_blk            Memory block to be hashed.
 * @param    hash_result_reg    Hashed value.
 *
 * @return  RTIC_SUCCESS    Successfully hashed the data.\n
 *          RTIC_FAILURE    Error in the parameters passed.
 */
rtic_ret rtic_hash_result(rtic_memblk mem_blk, rtic_hash_rlt *hash_result_reg);

/*!
 * This function is called to put the RTIC in a low power state. Refer to the
 * document driver-model/driver.txt in the kernel source tree for more
 * information.
 *
 * @param   dev   the device structure used to give information on RTIC
 *                to suspend.
 * @param   state the power state the device is entering.
 * @param   level the stage in device suspension process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
rtic_ret rtic_suspend(struct device *dev, ulong state, ulong level);

/*!
 * This function is called to bring the RTIC back from a low power state.
 * Refer to the document driver-model/driver.txt in the kernel source tree
 * for more information.
 *
 * @param   dev   the device structure used to give information on RTIC
 *                to resume.
 * @param   level the stage in device resumption process that we want the
 *                device to be put in.
 *
 * @return  The function always returns 0.
 */
rtic_ret rtic_resume(struct device *dev, ulong level);

/*! @} */

#endif /* _MXC_SECURITY_API_H_ */
