/*
 * Copyright 2005 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  
 * 02111-1307, USA
 *
 */

#ifndef _PMIC_BATTERY_H
#define _PMIC_BATTERY_H

/*!
 * @defgroup PMIC_RR_BATTERY Roadrunner Battery Driver
 * @ingroup PMIC_RR_DRVRS
 */

/*!
 * @file pmic_battery.h
 * @brief This is the header of PMIC Battery driver. 
 *
 * @ingroup PMIC_RR_BATTERY
 */

#include <asm/ioctl.h>

/*!
 * @name IOCTL user space interface
 */ 

/*! @{ */
/*!
 * Enable and disable charger.
 * Argument type: pointer to t_charger_setting
 */ 
#define PMIC_BATT_CHARAGER_CONTROL              _IOW('p', 0xc0, int)
/*!
 * Set charger configuration.
 * Argument type: pointer to t_charger_setting
 */ 
#define PMIC_BATT_SET_CHARAGER                  _IOW('p', 0xc1, int)
/*!
 * Get charger configuration.
 * Argument type: pointer to t_charger_setting
 */ 
#define PMIC_BATT_GET_CHARAGER                  _IOR('p', 0xc2, int)
/*!
 * Get charger current.
 * Argument type: pointer to t_charger_setting
 */ 
#define PMIC_BATT_GET_CHARAGER_CURRENT          _IOR('p', 0xc3, int)
/*!
 * Set EOL control
 * Argument type: pointer to t_eol_setting
 */ 
#define PMIC_BATT_EOL_CONTROL                   _IOW('p', 0xc4, int)
/*!
 * Enable and disable charging LED.
 * Argument type: bool
 */ 
#define PMIC_BATT_LED_CONTROL                   _IOW('p', 0xc5, int)
/*!
 * Enable and disable reverse supply.
 * Argument type: bool
 */ 
#define PMIC_BATT_REV_SUPP_CONTROL              _IOW('p', 0xc6, int)
/*!
 * Enable and disable unregulatored charging mode.
 * Argument type: bool
 */ 
#define PMIC_BATT_UNREG_CONTROL                 _IOW('p', 0xc7, int)
/*! @} */

/*!
 * This enumeration defines battery charagers.
 */ 
typedef enum {
        BATT_MAIN_CHGR        = 0,     /*!< Main battery charger */
        BATT_CELL_CHGR,                /*!< Cell battery charger */
        BATT_TRCKLE_CHGR               /*!< Trickle charger (only available on Atlas) */
} t_batt_charger;

/*!
 * This structure is used for the following battery changer control 
 * IOCTLs:
 * - PMIC_BATT_CHARAGER_CONTROL
 * - PMIC_BATT_SET_CHARAGER
 * - PMIC_BATT_GET_CHARAGER
 */ 
typedef struct {
        /*!
         * Charger
         */
        t_batt_charger chgr;
        /*!
         * Turn on charger
         */
        bool on;
        /*!
         * Charging voltage
         */
        unsigned char  c_voltage;
        /*!
         * Charging current
         */
        unsigned char  c_current;
} t_charger_setting;

/*!
 * This structure is used for EOL setting IOCTL PMIC_BATT_EOL_CONTROL
 */ 
typedef struct {
        /*!
         * Enable EOL comparator
         */
        bool enable;
        /*!
         * c_voltage threshold
         */
        unsigned char threshold;
} t_eol_setting;


/* EXPORTED FUNCTIONS */

/*!
 * This function is used to start charging a battery. For different charger,
 * different c_voltage and current range are supported. \n
 * 
 *
 * @param      chgr        Charger as defined in /b t_batt_charger.
 * @param      c_voltage     Charging voltage.
 * @param      c_current     Charging current.
 *
 * @return     This function returns PMIC_SUCCESS if successful.
 */ 
PMIC_STATUS pmic_batt_enable_charger(t_batt_charger chgr,
                                     unsigned char c_voltage,
                                     unsigned char c_current);

/*!
 * This function turns off a charger.
 *
 * @param      chgr        Charger as defined in /b t_batt_charger.
 *
 * @return     This function returns PMIC_SUCCESS if successful.
 */ 
PMIC_STATUS pmic_batt_disable_charger(t_batt_charger chgr);

/*!
 * This function is used to change the charger setting.
 * 
 * @param      chgr        Charger as defined in /b t_batt_charger.
 * @param      c_voltage   Charging voltage.
 * @param      c_current   Charging current.
 *
 * @return     This function returns PMIC_SUCCESS if successful.
 */ 
PMIC_STATUS pmic_batt_set_charger(t_batt_charger chgr,
                                  unsigned char c_voltage,
                                  unsigned char c_current);

/*!
 * This function is used to retrive the charger setting.
 * 
 * @param      chgr        Charger as defined in /b t_batt_charger.
 * @param      c_voltage   Output parameter for charging c_voltage setting.
 * @param      c_current   Output parameter for charging current setting.
 *
 * @return     This function returns PMIC_SUCCESS if successful.
 */ 
PMIC_STATUS pmic_batt_get_charger_setting(t_batt_charger chgr,
                                          unsigned char *c_voltage,
                                          unsigned char *c_current);

/*!
 * This function is retrives the main battery charging current.
 * 
 * @param      c_current     Output parameter for charging current setting.
 *
 * @return     This function returns PMIC_SUCCESS if successful.
 */ 
PMIC_STATUS pmic_batt_get_charge_current(unsigned char *c_current);

/*!
 * This function enables End-of-Life comparator.
 * 
 * @param      threshold  End-of-Life threshold.
 *
 * @return     This function returns PMIC_SUCCESS if successful.
 */ 
PMIC_STATUS pmic_batt_enable_eol(unsigned char threshold);

/*!
 * This function disables End-of-Life comparator.
 * 
 * @return     This function returns PMIC_SUCCESS if successful.
 */ 
PMIC_STATUS pmic_batt_disable_eol(void);

/*!
 * This function controls charge LED.
 *
 * @param      on   If on is ture, LED will be turned on,
 *                  or otherwise, LED will be turned off.
 * 
 * @return     This function returns PMIC_SUCCESS if successful.
 */ 
PMIC_STATUS pmic_batt_led_control(bool on);

/*!
 * This function sets reverse supply mode. This function is not applicable
 * to Roadrunner.
 *
 * @param      enable     If enable is ture, reverse supply mode is enable,
 *                        or otherwise, reverse supply mode is disabled.
 *
 * @return     This function returns PMIC_SUCCESS if successful.
 */ 
PMIC_STATUS pmic_batt_set_reverse_supply(bool enable);

/*!
 * This function sets unregulatored charging mode on main battery. This
 * function is not aoolicable to Roadrunner.
 * 
 * @param      enable     If enable is ture, unregulated charging mode is
 *                        enable, or otherwise, disabled.
 *
 * @return     This function returns PMIC_SUCCESS if successful.
 */ 
PMIC_STATUS pmic_batt_set_unregulated(bool enable);

#endif /* _PMIC_BATTERY_H */
