/*
 * Copyright 2005 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA
 *
 */

#ifndef _PMIC_LIGHT_H
#define _PMIC_LIGHT_H

/*!
 * @defgroup PMIC_RR_LIGHT Roadrunner Light Driver
 * @ingroup PMIC_RR_DRVRS
 */

/*!
 * @file pmic_light.h
 * @brief This is the header of PMIC Light driver.
 *
 * @ingroup PMIC_RR_LIGHT
 */

#include <asm/ioctl.h>

/*!
 * @name IOCTL user space interface
 */

/*! @{ */
/*!
 * Enable Backlight.
 * Argument type: none.
 */
#define PMIC_BKLIT_ENABLE                   _IO('p', 0xe0)
/*!
 * Disable Backlight.
 * Argument type: none.
 */
#define PMIC_BKLIT_DISABLE                  _IO('p', 0xe1)
/*!
 * Set backlight configuration.
 * Argument type: pointer to t_bklit_setting_param
 */
#define PMIC_SET_BKLIT                      _IOW('p', 0xe2, int)
/*!
 * Get backlight configuration.
 * Argument type: pointer to t_bklit_setting_param
 */
#define PMIC_GET_BKLIT                      _IOR('p', 0xe3, int)
/*!
 * Ramp up configuration.
 * Argument type: t_bklit_channel
 */
#define PMIC_RAMPUP_BKLIT                   _IOW('p', 0xe4, int)
/*!
 * Ramp down configuration.
 * Argument type: t_bklit_channel
 */
#define PMIC_RAMPDOWN_BKLIT                 _IOW('p', 0xe5, int)
/*!
 * Enable Tri-color LED.
 * Argument type: t_tcled_mode
 */
#define PMIC_TCLED_ENABLE                   _IOW('p', 0xe6, int)
/*!
 * Disable Tri-color LED.
 * Argument type: none.
 */
#define PMIC_TCLED_DISABLE                  _IO('p', 0xe7)
/*!
 * Start Tri-color LED pattern.
 * Argument type: t_fun_param
 */
#define PMIC_TCLED_PATTERN                  _IOW('p', 0xe8, int)
/*! @} */

/*!
 * @enum t_bklit_mode
 * @brief Backlight modes.
 */
typedef enum {
        BACKLIGHT_CURRENT_CTRL_MODE,       /*! < Current control mode */
        BACKLIGHT_TRIODE_MODE              /*! < Triode mode */
} t_bklit_mode;

/*!
 * @enum t_bklit_channel
 * @brief Backlight channels.
 */
typedef enum {
        BACKLIGHT_LED1,                    /*! < Backlight channel 1 */
        BACKLIGHT_LED2                     /*! < Backlight channel 2 */
} t_bklit_channel;

/*!
 * @enum t_bklit_strobe_mode
 * @brief Backlight Strobe Light Pulsing modes.
 */
typedef enum {
        /*!
        * No Strobe Light Pulsing
        */
        BACKLIGHT_STROBE_NONE,
        /*!
        * Strobe Light Pulsing at 3.3% duty cycle over 300msec (Driver goes
        * into Triode Mode with pulses constrained to 10msec.)
        */
        BACKLIGHT_STROBE_FAST,
        /*!
        * Strobe Light Pulsing at 10% duty cycle over 100msec (Driver goes
        * into Triode Mode with pulses constrained to 10msec.)
        */
        BACKLIGHT_STROBE_SLOW
} t_bklit_strobe_mode;

/*!
 * @struct t_bklit_setting_param
 * @brief Backlight setting.
 */
typedef struct {
        t_bklit_channel     channel;           /*!< Channel */
        t_bklit_mode        mode;              /*!< Mode */
        t_bklit_strobe_mode strobe;            /*!< Strobe mode */
        unsigned char       current_level;     /*!< Current level */
        unsigned char       duty_cycle;        /*!< Duty cycle */
        unsigned char       cycle_time;        /*!< Cycle time */
        bool                edge_slow;         /*!< Edge Slow */
} t_bklit_setting_param;



/*!
 * @enum t_funlight_bank
 * @brief Tri-color LED fun light banks; this doesn't apply to Roadrunner.
 */
typedef enum {
    TCLED_FUN_BANK1 = 0,               /*! < Fun light bank 1 */
    TCLED_FUN_BANK2,                   /*! < Fun light bank 2 */
    TCLED_FUN_BANK3                    /*! < Fun light bank 3 */
} t_funlight_bank;

/*!
 * @enum t_tcled_mode
 * @brief Tri-color LED operation modes.
 *
 * The Tri-Color LED Driver circuitry includes 2 modes of operation. In LED
 * Indicator Mode, this circuitry operates as Red and Green LED Drivers with
 * flasher timing to indicate GSM network status. In Fun Light Mode, this
 * circuitry provides expanded capability for current control and distribution
 * that supplements the three channels.
 */
typedef enum {
    TCLED_IND_MODE = 0,                /*! < LED Indicator Mode */
    TCLED_FUN_MODE                     /*! < Fun Light Mode */
} t_tcled_mode;

/*!
 * @enum t_ind_channel
 * @brief Tri-color LED indicator mode channels.
 *
 */
typedef enum {
    TCLED_IND_RED = 0,                 /*! < Red LED */
    TCLED_IND_GREEN                    /*! < Green LED */
} t_ind_channel;


/*!
 * @enum t_funlight_channel
 * @brief Tri-color LED fun light mode channels.
 *
 */
typedef enum {
    TCLED_FUN_CHANNEL1 = 0,            /*! < Fun light channel 1 (Red) */
    TCLED_FUN_CHANNEL2,                /*! < Fun light channel 2 (Green) */
    TCLED_FUN_CHANNEL3                 /*! < Fun light channel 3 (Blue) */
} t_funlight_channel;

/*!
 * @enum t_tcled_ind_blink_pattern
 * @brief Tri-color LED Indicator Mode blinking mode.
 */
typedef enum {
    TCLED_IND_OFF = 0,                 /*! < Continuous off */
    TCLED_IND_BLINK_1,                 /*! < 0.1s on 1.9s off */
    TCLED_IND_BLINK_2,                 /*! < 0.2s on 1.8s off */
    TCLED_IND_BLINK_3,                 /*! < 0.5s on 1.5s off */
    TCLED_IND_BLINK_4,                 /*! < 0.25s on 0.75s off */
    TCLED_IND_BLINK_5,                 /*! < 0.25s on 1.75s off */
    TCLED_IND_BLINK_6,                 /*! < 0.05s on 1.95s off */
    TCLED_IND_BLINK_7,                 /*! < 0.5s on 0.5s off */
    TCLED_IND_BLINK_8,                 /*! < 0.5s off 0.5s on */
    TCLED_IND_BLINK_9,                 /*! < 0.125s on 0.5s off 0.075s on 1.5s off*/
    TCLED_IND_BLINK_10,                /*! < 0.125s on 2.075s off */
    TCLED_IND_BLINK_11,                /*! < 0.625s off 0.075s on 1.5s off*/
    TCLED_IND_ON                       /*! < Continuous on */
} t_tcled_ind_blink_pattern;

/*!
 * @enum t_tcled_cur_level
 * @brief Tri-color LED current levels.
 */
typedef enum {
    TCLED_CUR_LEVEL_1 = 0,                 /*! < Tri-Color LED current level 1 */
    TCLED_CUR_LEVEL_2,                     /*! < Tri-Color LED current level 2 */
    TCLED_CUR_LEVEL_3,                     /*! < Tri-Color LED current level 3 */
    TCLED_CUR_LEVEL_4                      /*! < Tri-Color LED current level 4 */
} t_tcled_cur_level;


/*!
 * @enum t_tcled_fun_cycle_time
 * @brief Tri-color LED fun light mode cycle time.
 */
typedef enum {
    TC_CYCLE_TIME_1 = 0,                 /*! < Tri-Color LED cycle time 1 */
    TC_CYCLE_TIME_2,                     /*! < Tri-Color LED cycle time 2 */
    TC_CYCLE_TIME_3,                     /*! < Tri-Color LED cycle time 3 */
    TC_CYCLE_TIME_4                      /*! < Tri-Color LED cycle time 4 */
} t_tcled_fun_cycle_time;


/*!
 * @enum t_tcled_fun_speed
 * @brief Tri-color LED fun light mode pattern speed.
 */
typedef enum {
    TC_OFF = 0,                 /*! < Tri-Color pattern off */
    TC_SLOW,                    /*! < Tri-Color slow pattern */
    TC_FAST                     /*! < Tri-Color fast pattern */
} t_tcled_fun_speed;


/*!
 * @enum t_tcled_fun_speed
 * @brief Tri-color LED fun light mode pattern speed.
 */
typedef enum {
    TC_STROBE_OFF = 0,          /*! < No strobe */
    TC_STROBE_SLOW,             /*! < Slow strobe pattern */
    TC_STROBE_FAST              /*! < fast strobe pattern */
} t_tcled_fun_strobe_speed;

/*!
 * @enum t_chaselight_pattern
 * @brief Tri-color LED fun light mode chasing light patterns.
 */
typedef enum {
    RGB = 0,                           /*!< R -> G -> B */
    BGR                                /*!< B -> G -> R */
} t_chaselight_pattern;


/*!
 * This enumeration of Fun Light Pattern.
 */
typedef enum {
        /*!
        * Blended ramps slow 
        */
        BLENDED_RAMPS_SLOW,
        /*!
        * Blended ramps fast 
        */
        BLENDED_RAMPS_FAST,
        /*!
        * Saw ramps slow 
        */
        SAW_RAMPS_SLOW,
        /*!
        * Saw ramps fast 
        */
        SAW_RAMPS_FAST,
        /*!
        * Blended bowtie slow 
        */
        BLENDED_BOWTIE_SLOW,
        /*!
        * Blended bowtie fast
        */
        BLENDED_BOWTIE_FAST,
        /*!
        * Strobe slow
        */
        STROBE_SLOW,
        /*!
        * Strobe fast
        */
        STROBE_FAST,
        /*!
        * Chasing Light RGB Slow
        */
        CHASING_LIGHT_RGB_SLOW,
        /*!
        * Chasing Light RGB fast
        */
        CHASING_LIGHT_RGB_FAST,
        /*!
        * Chasing Light BGR Slow
        */
        CHASING_LIGHT_BGR_SLOW,
        /*!
        * Chasing Light BGR fast
        */
        CHASING_LIGHT_BGR_FAST,
} t_fun_pattern;

/*!
 * @struct t_fun_param
 * @brief LED fun pattern IOCTL parameter
 */
typedef struct {
        t_funlight_bank    bank;            /*!< TCLED bank */
        t_funlight_channel channel;         /*!< TCLED channel */         
        t_fun_pattern      pattern;         /*!< Fun pattern */
} t_fun_param;

/*!
 * @enum t_led_channel
 * @brief LED channels including backlight and tri-color LEDs.
 */
typedef enum {
    AUDIO_LED1,                        /*! < Backlight channel 1 */
    AUDIO_LED2,                        /*! < Backlight channel 2 */
    AUDIO_LEDR,                        /*! < Fun light channel 1 (Red) */
    AUDIO_LEDG,                        /*! < Fun light channel 2 (Green) */
    AUDIO_LEDB                         /*! < Fun light channel 3 (Blue) */
} t_led_channel;

/*!
 * @enum t_aud_path
 * @brief LED audio modulation inout audio channels
 */
typedef enum {
    MIXED_RX = 0,                      /*!<  Mixed L & R Channel RX audio */     
    TX                                 /*!<  TX path */
} t_aud_path;

/*!
 * @enum t_aud_gain
 * @brief LED audio modulation inout audio channels
 */
typedef enum {
    GAIN_MINUS6DB = 0,                      /*!< -6 dB */
    GAIN_0DB,                               /*!< 0 dB */
    GAIN_6DB,                               /*!< 6 dB */
    GAIN_12DB                               /*!< 12 dB */
} t_aud_gain;


/* EXPORTED FUNCTIONS */

/*!
 * This function enables backlight.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_master_enable(void);

/*!
 * This function disables backlight.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_master_disable(void);

/*!
 * This function sets backlight current level.
 * In Roadrunner, LED1 and LED2 are designed for a nominal full scale current
 * of 84mA in 12mA steps.  The channels are not individually adjustable, hence
 * the channel parameter is ignored.
 *
 * @param        channel   Backlight channel (Ignored in Roadrunner because the
 *                         channels are not individually adjustable)
 * @param        level     Backlight current level, as the following table.
 *                         @verbatim
                               level     current
                               ------    -----------
                                 0         0 mA
                                 1         12 mA
                                 2         24 mA
                                 3         36 mA
                                 4         48 mA
                                 5         60 mA
                                 6         72 mA
                                 7         84 mA
                            @endverbatim
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_set_current(t_bklit_channel channel, unsigned char level);

/*!
 * This function retrives backlight current level.
 * In Roadrunner, LED1 and LED2 are designed for a nominal full scale current
 * of 84mA in 12mA steps.  The channels are not individually adjustable, hence
 * the channel parameter is ignored.
 *
 * @param        channel   Backlight channel (Ignored in Roadrunner because the
 *                         channels are not individually adjustable)
 * @param        level     Pointer to store backlight current level result.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_get_current(t_bklit_channel channel, unsigned char* level);

/*!
 * This function sets a backlight channel duty cycle.
 * LED perceived brightness for each zone may be individually set by setting
 * duty cycle. The default setting is for 0% duty cycle; this keeps all zone
 * drivers turned off even after the master enable command. Each LED current
 * sink can be turned on and adjusted for brightness with an independent 4 bit
 * word for a duty cycle ranging from 0% to 100% in approximately 6.7% steps.
 *
 * @param        channel   Backlight channel.
 * @param        dc        Backlight duty cycle, as the following table.
 *                         @verbatim
                                dc        Duty Cycle (% On-time over Cycle Time)
                               ------    ---------------------------------------
                                  0        0%
                                  1        6.7%
                                  2        13.3%
                                  3        20%
                                  4        26.7%
                                  5        33.3%
                                  6        40%
                                  7        46.7%
                                  8        53.3%
                                  9        60%
                                 10        66.7%
                                 11        73.3%
                                 12        80%
                                 13        86.7%
                                 14        93.3%
                                 15        100%
                             @endverbatim
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_set_dutycycle(t_bklit_channel channel, unsigned char dc);

/*!
 * This function retrives a backlight channel duty cycle.
 *
 * @param        channel   Backlight channel.
 * @param        cycle     Pointer to backlight duty cycle.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_get_dutycycle(t_bklit_channel channel, unsigned char* dc);

/*!
 * This function sets a backlight channel cycle time.
 * Cycle Time is defined as the period of a complete cycle of
 * Time_on + Time_off. The default Cycle Time is set to 0.01 seconds such that
 * the 100 Hz on-off cycling is averaged out by the eye to eliminate
 * flickering. Additionally, the Cycle Time can be programmed to intentionally
 * extend the period of on-off cycles for a visual pulsating or blinking effect.
 *
 * @param        period    Backlight cycle time, as the following table.
 *                         @verbatim
                                period      Cycle Time
                               --------    ------------
                                  0          0.01 seconds
                                  1          0.1 seconds
                                  2          0.5 seconds
                                  3          2 seconds
                             @endverbatim
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_set_cycle_time(unsigned char period);

/*!
 * This function retrives a backlight channel cycle time setting.
 *
 * @param        period    Pointer to save backlight cycle time setting result.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_get_cycle_time(unsigned char* period);

/*!
 * This function sets backlight operation mode. There are two modes of
 * operations: current control and triode mode.
 * The Duty Cycle/Cycle Time control is retained in Triode Mode. Audio
 * coupling is not available in Triode Mode.
 *
 * @param        channel   Backlight channel.
 * @param        mode      Backlight operation mode.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_set_mode(t_bklit_channel channel, t_bklit_mode mode);

/*!
 * This function starts backlight brightness ramp up function; ramp time is
 * fixed at 0.5 seconds.
 *
 * @param        channel   Backlight channel.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_rampup(t_bklit_channel channel);

/*!
 * This function starts backlight brightness ramp down function; ramp time is
 * fixed at 0.5 seconds.
 *
 * @param        channel   Backlight channel.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_rampdown(t_bklit_channel channel);

/*!
 * This function enables backlight analog edge slowing mode. Analog Edge
 * Slowing slows down the transient edges to reduce the chance of coupling LED
 * modulation activity into other circuits. Rise and fall times will be targeted
 * for approximately 50usec.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_enable_edge_slow(void);

/*!
 * This function disables backlight analog edge slowing mode. The backlight
 * drivers will default to an Instant On mode.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_disable_edge_slow(void);

/*!
 * This function sets backlight Strobe Light Pulsing mode.
 *
 * @param        channel   Backlight channel.
 * @param        mode      Strobe Light Pulsing mode.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_bklit_set_strobemode(t_bklit_channel channel, t_bklit_strobe_mode mode);

/*!
 * This function enables tri-color LED.
 *
 * @param        mode      Tri-color LED operation mode.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_enable(t_tcled_mode mode);

/*!
 * This function disables tri-color LED.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_disable(void);

/*!
 * This function retrives tri-color LED operation mode.
 *
 * @param        mode      Pointer to Tri-color LED operation mode.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_get_mode(t_tcled_mode* mode);

/*!
 * This function sets a tri-color LED channel current level in indicator mode.
 *
 * @param        channel      Tri-color LED channel.
 * @param        level        Current level.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_ind_set_current(t_ind_channel channel, t_tcled_cur_level level);

/*!
 * This function retrives a tri-color LED channel current level in indicator mode.
 *
 * @param        channel      Tri-color LED channel.
 * @param        level        Pointer to current level.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_ind_get_current(t_ind_channel channel, t_tcled_cur_level* level);

/*!
 * This function sets a tri-color LED channel blinking pattern in indication
 * mode.
 *
 * @param        channel      Tri-color LED channel.
 * @param        pattern      Blinking pattern.
 * @param        skip         If trure, skip a cycle after each cycle.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_ind_set_blink_pattern(t_ind_channel channel,
                                             t_tcled_ind_blink_pattern pattern,
                                             bool skip);

/*!
 * This function retrives a tri-color LED channel blinking pattern in
 * indication mode.
 *
 * @param        channel      Tri-color LED channel.
 * @param        pattern      Pointer to Blinking pattern.
 * @param        skip         Pointer to a boolean varible indicating if skip
 *                            a cycle after each cycle.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_ind_get_blink_pattern(t_ind_channel channel,
                                             t_tcled_ind_blink_pattern* pattern,
                                             bool* skip);

/*!
 * This function sets a tri-color LED channel current level in Fun Light mode.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        channel      Tri-color LED channel.
 * @param        level        Current level.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_set_current(t_funlight_bank bank,
                                       t_funlight_channel channel,
                                       t_tcled_cur_level level);

/*!
 * This function retrives a tri-color LED channel current level in Fun Light mode.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        channel      Tri-color LED channel.
 * @param        level        Pointer to current level.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_get_current(t_funlight_bank bank,
                                       t_funlight_channel channel,
                                       t_tcled_cur_level* level);

/*!
 * This function sets tri-color LED cycle time in Fun Light mode.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        ct           Cycle time.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_set_cycletime(t_funlight_bank bank,
                                         t_tcled_fun_cycle_time ct);

/*!
 * This function retrives tri-color LED cycle time in Fun Light mode.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        ct           Pointer to cycle time.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_get_cycletime(t_funlight_bank bank,
                                         t_tcled_fun_cycle_time* ct);

/*!
 * This function sets a tri-color LED channel duty cycle in Fun Light mode.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        channel      Tri-color LED channel.
 * @param        dc           Duty cycle.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_set_dutycycle(t_funlight_bank bank,
                                         t_funlight_channel channel,
                                         unsigned char dc);

/*!
 * This function retrives a tri-color LED channel duty cycle in Fun Light mode.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        channel      Tri-color LED channel.
 * @param        dc           Pointer to duty cycle.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_get_dutycycle(t_funlight_bank bank,
                                         t_funlight_channel channel,
                                         unsigned char* dc);

/*!
 * This function initiates Blended Ramp fun light pattern.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        speed        Speed of pattern.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_blendedramps(t_funlight_bank bank,
                                        t_tcled_fun_speed speed);

/*!
 * This function initiates Saw Ramp fun light pattern.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        speed        Speed of pattern.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_sawramps(t_funlight_bank bank,
                                    t_tcled_fun_speed speed);

/*!
 * This function initiates Blended Bowtie fun light pattern.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        speed        Speed of pattern.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_blendedbowtie(t_funlight_bank bank,
                                         t_tcled_fun_speed speed);

/*!
 * This function initiates Chasing Lights fun light pattern.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        pattern      Chasing light pattern mode.
 * @param        speed        Speed of pattern.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_chasinglightspattern(t_funlight_bank bank,
                                                t_chaselight_pattern pattern,
                                                t_tcled_fun_speed speed);

/*!
 * This function initiates Strobe Mode fun light pattern.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        channel      Tri-color LED channel.
 * @param        speed        Speed of pattern.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_strobe(t_funlight_bank bank,
                                  t_funlight_channel channel,
                                  t_tcled_fun_strobe_speed speed);

/*!
 * This function initiates Tri-color LED brightness Ramp Up function; Ramp time
 * is fixed at 1 second.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        channel      Tri-color LED channel.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_rampup(t_funlight_bank bank,
                                  t_funlight_channel channel);

/*!
 * This function initiates Tri-color LED brightness Ramp Down function; Ramp
 * time is fixed at 1 second.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        channel      Tri-color LED channel.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_rampdown(t_funlight_bank bank,
                                    t_funlight_channel channel);

/*!
 * This function enables a Tri-color channel triode mode.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        channel      Tri-color LED channel.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_triode_on(t_funlight_bank bank,
                                     t_funlight_channel channel);

/*!
 * This function disables a Tri-color LED channel triode mode.
 *
 * @param        bank         Tri-color LED bank (This parameter is ignored
 *                            for Roadrunner).
 * @param        channel      Tri-color LED channel.
 *
 * @return       This function returns PMIC_SUCCESS if successful.
 */
PMIC_STATUS pmic_tcled_fun_triode_off(t_funlight_bank bank,
                                      t_funlight_channel channel);

/*!
 * This function enables Tri-color LED edge slowing, this function does not
 * apply to Roadrunner.
 *
 * @return       This function returns PMIC_NOT_SUPPORTED.
 */
PMIC_STATUS pmic_tcled_enable_edge_slow(void);

/*!
 * This function disables Tri-color LED edge slowing, this function does not
 * apply to Roadrunner.
 *
 * @return       This function returns PMIC_NOT_SUPPORTED.
 */
PMIC_STATUS pmic_tcled_disable_edge_slow(void);

/*!
 * This function enables Tri-color LED half current mode, this function does
 * not apply to Roadrunner.
 *
 * @return       This function returns PMIC_NOT_SUPPORTED.
 */
PMIC_STATUS pmic_tcled_enable_half_current(void);

/*!
 * This function disables Tri-color LED half current mode, this function does
 * not apply to Roadrunner.
 *
 * @return       This function returns PMIC_NOT_SUPPORTED.
 */
PMIC_STATUS pmic_tcled_disable_half_current(void);

/*!
 * This function enables backlight or Tri-color LED audio modulation.
 *
 * @return       This function returns PMIC_NOT_SUPPORTED.
 */
PMIC_STATUS pmic_tcled_enable_audio_modulation(t_led_channel channel,
                                               t_aud_path path,
                                               t_aud_gain gain,
                                               bool lpf_bypass);

/*!
 * This function disables backlight or Tri-color LED audio modulation.
 *
 * @return       This function returns PMIC_NOT_SUPPORTED.
 */
PMIC_STATUS pmic_tcled_disable_audio_modulation(void);


#endif /* _PMIC_LIGHT_H */
