/*
 * Beep driver for Sony ICX1087 platform.
 *
 * Copyright 2006 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/config.h>			/* kernel configuration */
#include <linux/kernel.h>			/* often-used kernel functions */
#include <linux/fs.h>				/* file system */
#include <linux/cdev.h>				/* character device driver */
#include <linux/spinlock.h>			/* spin lock */
#include <linux/interrupt.h>		/* local_bh_enable(), local_bh_disable() */
#include <linux/timer.h>			/* kernel timer */
#include <linux/devfs_fs_kernel.h>	/* devfs */
#include <linux/device.h>			/* device and driver structure */
#include <asm/uaccess.h>			/* put_user(), get_user() */
#include <asm/arch/icx1087_audio.h>	/* ICX1087 Audio Codec API */
#include <asm/arch/icx1087_beep.h>	/* ICX1087 Beep API */
#include "cs4251.h"
//#define DEBUG
#define PrHead	"icx1087_audio: "
#include <linux/prdbg.h>

#define BEEP_AOUT_VOLUME	(uint8_t)((0)*2)	/* 0dB */

#define BEEP_DEVICE_NAME	"beep"			/* device name */
#define BEEP_DEV_ENTRY_NAME	"sound/beep"	/* /dev entry name */

/*
 *	module parameters
 */
static int	s_BeepMajor = 0; module_param(s_BeepMajor, int, S_IRUGO);
static int	s_BeepMinor = 0; module_param(s_BeepMinor, int, S_IRUGO);

/*
 *	function prototypes
 */
extern int	icx1087_mute(int mask, int value);
extern int	icx1087_restore_aout_vol(void);

/*
 *	file scope structures & variables
 */

typedef struct {
	struct cdev			cdev;				/* character device structure */
	struct class_simple	*class;				/* */
	spinlock_t			spinlock;			/* spinlock for timer */
	struct timer_list	timer;				/* kernel timer */
	int					status;				/* beep status */
	beep_play_t			*cur_param;			/* current play parameter */
	int					seq_no;				/* sequence number */
	wait_queue_head_t	sync_wq;			/* sync wait queue */
	int					beep_vol;			/* from 0 to 31 */
	uint32_t			interval_time;		/* beep interval time (ms) */
	uint32_t			last_beep_jiffies;	/* last beep time */
	int					save_mute;			/* save mute value */
} beep_ctrl_t;

static beep_ctrl_t	s_BeepCtrl = {
	.spinlock			= SPIN_LOCK_UNLOCKED,	
	.status				= BEEP_STATUS_STOP,
	.cur_param			= NULL,
	.seq_no				= 0,
	.beep_vol			= BEEP_DEFAULT_VOLUME,
	.interval_time		= BEEP_DEFAULT_INTERVAL,
	.last_beep_jiffies	= INITIAL_JIFFIES,
};

static beep_play_t	s_PlayParam;
static beep_time_t	s_BeepPattern[MAX_BEEP_PATTERN];

/* Beep Frequency Table. Fs=44.1kHz (Fs=12,24,48 or 96kHz) */
static const unsigned	s_BeepFreqTable[] = {
	 23967,	/*  239.67Hz ( 260.87Hz, Pitch:C4) */
	 47935,	/*  479.35Hz ( 521.74Hz, Pitch:C5) */
	 53781,	/*  537.81Hz ( 585.37Hz, Pitch:D5) */
	 61250,	/*  612.50Hz ( 666.67Hz, Pitch:E5) */
	 64853,	/*  648.53Hz ( 705.88Hz, Pitch:F5) */
	 71129,	/*  711.29Hz ( 774.19Hz, Pitch:G5) */
	 81667,	/*  816.67Hz ( 888.89Hz, Pitch:A5) */
	 91875,	/*  918.75Hz (1000.00Hz, Pitch:B5) */
	 95870,	/*  958.70Hz (1043.48Hz, Pitch:C6) */
	110250,	/* 1102.50Hz (1200.00Hz, Pitch:D6) */
	122500,	/* 1225.00Hz (1333.33Hz, Pitch:E6) */
	129705,	/* 1297.05Hz (1411.76Hz, Pitch:F6) */
	147000,	/* 1470.00Hz (1600.00Hz, Pitch:G6) */
	157500,	/* 1575.00Hz (1714.29Hz, Pitch:A6) */
	183750,	/* 1837.50Hz (2000.00Hz, Pitch:B6) */
	200455,	/* 2004.55Hz (2181.82Hz, Pitch:C7) */
};

#ifdef DEBUG
static inline unsigned GetUS(void)
{
	struct timeval	tm;

	do_gettimeofday(&tm);
	return (unsigned)tm.tv_usec;
}
#endif

inline void icx1087_beep_spin_lock_bh(void)
{
	spin_lock_bh(&s_BeepCtrl.spinlock);
}

inline void icx1087_beep_spin_unlock_bh(void)
{
	spin_unlock_bh(&s_BeepCtrl.spinlock);
}

static inline int	CalcBeepLevel(int vol)
{
	/* convert from volume (0 to 31) to beep level (-50 to +12) */
	return ((12 / 2) - (31 - vol)) & 0x1f;
}

static inline int SetBeepVol(int vol)
{
	int	result;

	result = cs4251_set_beep_vol(CalcBeepLevel(vol));
	if (result == 0) {
		s_BeepCtrl.beep_vol = vol;
	}
	return result;
}

/* Convert from frequency to freq table index (0 to 15) */
static inline int GetFreqIndex(unsigned hz)
{
	int	i;

	for (i = 0; i < 15; i++) {
		if (hz * 100 <= (s_BeepFreqTable[i] + s_BeepFreqTable[i+1]) / 2) {
			break;
		}
	}
	PrDebug("hz=%u, index=%d\n", hz, i);
	return i;
}

#if HZ > 1000
#error "Assume HZ <= 1000" 
#endif
static inline int StartTimer(struct timer_list	*timer, int ms)
{
	return mod_timer(timer, jiffies + (ms - 1 + (1000/HZ)) / (1000/HZ));
}

static inline int BeepStartTimer(int ms)
{
	s_BeepCtrl.timer.data = s_BeepCtrl.seq_no;
	return StartTimer(&s_BeepCtrl.timer, ms);
}

static inline int BeepStopTimer(void)
{
//	if (timer_pending(&s_BeepCtrl.timer)) {
	return del_timer_sync(&s_BeepCtrl.timer);
}

static inline int BeepOn(void)
{
	return cs4251_set_cont_beep(On);	/* beep on */
}

static inline int BeepOff(void)
{
	s_BeepCtrl.last_beep_jiffies = jiffies;
	return cs4251_set_cont_beep(Off);	/* beep off */
}

static inline void ToIntervalState(uint32_t ms)
{
	s_BeepCtrl.status = BEEP_STATUS_INTERVAL;

	/* kick interval-timer */
	BeepStartTimer(ms);
}

static inline void ToPlayState(void)
{
	s_BeepCtrl.status = BEEP_STATUS_PLAY_ON;
	cs4251_set_beep_freq(GetFreqIndex(s_BeepCtrl.cur_param->freq));
	BeepOn();

	/* kick on-timer */
	BeepStartTimer(s_BeepCtrl.cur_param->pattern->on_time);
}

static inline void ToStopState(void)
{
	s_BeepCtrl.status = BEEP_STATUS_STOP;
	s_BeepCtrl.cur_param = NULL;

	BeepStopTimer();
	icx1087_restore_aout_vol();					/* restore AOUT volume */
#ifdef USE_MUTE_ON_BEEP
	icx1087_mute(AUDIO_MUTE_ON, s_BeepCtrl.save_mute); /* restore mute state */
#endif
	wake_up_interruptible(&s_BeepCtrl.sync_wq);
}

static inline uint32_t GetIntervalExpireTime(void)
{
	uint32_t	ms = (jiffies - s_BeepCtrl.last_beep_jiffies) * (1000/HZ);
	if (ms >= s_BeepCtrl.interval_time)
		return 0;
	return s_BeepCtrl.interval_time - ms;
}

static int BeepPlayCmd(unsigned long arg)
{
	beep_play_t	param;
	beep_time_t	pattern[MAX_BEEP_PATTERN];
	uint32_t	remain_ms;
	int	result;

	PrDebug("jiffies=%lu, BeepStatus=%d\n", jiffies, s_BeepCtrl.status);

	/* get parameter and check its */
	if (copy_from_user(&param, (void *)arg, sizeof(param))) {
		return -EFAULT;
	}
	if (param.freq == 0 || param.freq > 20000 ||
		param.pri < BEEP_PRI_MIN || param.pri > BEEP_PRI_MAX ||
		param.bytes == 0 || param.bytes % sizeof(beep_time_t) ||
		param.bytes > MAX_BEEP_PATTERN_BYTES || param.pattern == NULL) {
		return -EINVAL;
	}

	/* get beep pattern parameter */
	if (copy_from_user(pattern, param.pattern, param.bytes)) {
		return -EFAULT;
	}

	/* disable software interrupt & preemption */
	icx1087_beep_spin_lock_bh();

	/* check priority. if same priority is a post-winning */
	if (s_BeepCtrl.cur_param && s_BeepCtrl.cur_param->pri > param.pri) {
		icx1087_beep_spin_unlock_bh();
		return -EBUSY;
	}

	/* save beep request */
	memcpy(&s_PlayParam, &param, sizeof(s_PlayParam));
	memcpy(s_BeepPattern, pattern, param.bytes);
	s_BeepCtrl.cur_param = &s_PlayParam;
	s_BeepCtrl.cur_param->pattern = &s_BeepPattern[0];	/* overwrite */

	/* update sequence number */
	if (++s_BeepCtrl.seq_no <= 0) {
		s_BeepCtrl.seq_no = 1;
	}
	result = s_BeepCtrl.seq_no;

	switch (s_BeepCtrl.status) {
	case BEEP_STATUS_STOP:
#ifdef USE_MUTE_ON_BEEP
		s_BeepCtrl.save_mute = icx1087_mute(AUDIO_MUTE_ON, AUDIO_MUTE_ON);
#endif
		cs4251_set_aout_vol(BEEP_AOUT_VOLUME, BEEP_AOUT_VOLUME);
		remain_ms = GetIntervalExpireTime();
		if (remain_ms > 0) {
			ToIntervalState(remain_ms);
		} else {
			ToPlayState();
		}
		break;
	case BEEP_STATUS_INTERVAL:
		/* NOP */
		break;
	case BEEP_STATUS_PLAY_ON:
		BeepOff();
		/* NO BREAK */
	case BEEP_STATUS_PLAY_OFF:
		if (s_BeepCtrl.interval_time > 0) {
			ToIntervalState(s_BeepCtrl.interval_time);
		} else {
			ToPlayState();
		}
		break;
	default:
		PrErr("driver bug! Unknown beep status\n");
		break;
	}

	icx1087_beep_spin_unlock_bh();
	return result;
}

static void BeepTimerHandler(unsigned long data)
{
	PrDebug("seq_no=%d jiffies=%lu, usec=%u, BeepStatus=%d\n",
			s_BeepCtrl.seq_no, jiffies, GetUS(), s_BeepCtrl.status);

	icx1087_beep_spin_lock_bh();

	switch (s_BeepCtrl.status) {
	case BEEP_STATUS_STOP:
		PrErr("driver bug! BEEP_STATUS_STOP is wrong status\n");
		break;
	case BEEP_STATUS_INTERVAL:
		ToPlayState();
		break;
	case BEEP_STATUS_PLAY_ON:
		BeepOff();
		s_BeepCtrl.status = BEEP_STATUS_PLAY_OFF;
		if (s_BeepCtrl.cur_param->pattern->off_time > 0) { 
			/* kick off-timer */
			BeepStartTimer(s_BeepCtrl.cur_param->pattern->off_time);
			break;
		}
		/* NO BREAK */
	case BEEP_STATUS_PLAY_OFF:
		/* next pattern exist ? */
		if (++s_BeepCtrl.cur_param->pattern <
			&s_BeepPattern[s_BeepCtrl.cur_param->bytes/sizeof(beep_time_t)]) {
			ToPlayState();
		} else {
			if (s_BeepCtrl.cur_param->repeat == 0) {
				ToStopState();	/* beep finished */
			} else {
				/* repeat beep pattern */
				if (s_BeepCtrl.cur_param->repeat > 0) {
					s_BeepCtrl.cur_param->repeat--;
				}
				/* reset pattern */
				s_BeepCtrl.cur_param->pattern = &s_BeepPattern[0];
				ToPlayState();
			}
		}
		break;
	default:
		PrErr("driver bug! Unknown beep status\n");
		break;
	}

	icx1087_beep_spin_unlock_bh();
}

static int BeepStopCmd(int arg)
{
	int	seq_no, result;

	result = get_user(seq_no, (int *)arg);
	PrDebug("seq_no=%d, BeepStatus=%d\n", seq_no, s_BeepCtrl.status);
	if (result != 0) {
		return result;
	}

	icx1087_beep_spin_lock_bh();

	if (s_BeepCtrl.status != BEEP_STATUS_STOP) {
		if (seq_no != 0 && seq_no != s_BeepCtrl.seq_no) {
			PrDebug("seq_no=%d was not found. current=%d\n",
				seq_no, s_BeepCtrl.seq_no);
			icx1087_beep_spin_unlock_bh();
			return -ENOENT;
		}

		switch (s_BeepCtrl.status) {
		case BEEP_STATUS_PLAY_ON:
			BeepOff();
			/* NO BREAK */
		case BEEP_STATUS_PLAY_OFF:
		case BEEP_STATUS_INTERVAL:
			ToStopState();	/* beep finished */
			break;
		default:
			PrErr("driver bug! Unknown beep status\n");
			break;
		}
	}

	icx1087_beep_spin_unlock_bh();
	return 0;
}

static int BeepSyncCmd(void)
{
	PrDebug("BeepStatus=%d\n", s_BeepCtrl.status);

	icx1087_beep_spin_lock_bh();
	while (s_BeepCtrl.status != BEEP_STATUS_STOP) {
		icx1087_beep_spin_unlock_bh();
		if (wait_event_interruptible(s_BeepCtrl.sync_wq,
					s_BeepCtrl.status == BEEP_STATUS_STOP)) {
			return -ERESTARTSYS;	/* signal: tell the fs layer */
		}
		icx1087_beep_spin_lock_bh();
	}
	icx1087_beep_spin_unlock_bh();
	return 0;
}

static int BeepStatusCmd(int arg)
{
	int	seq_no, status = 0, result;

	result = get_user(seq_no, (int *)arg);
	PrDebug("seq_no=%d, BeepStatus=%d\n", seq_no, s_BeepCtrl.status);
	if (result == 0) {
		icx1087_beep_spin_lock_bh();
		if (seq_no != 0 && seq_no != s_BeepCtrl.seq_no) {
			PrDebug("seq_no=%d was not found.\n", (int)seq_no);
			result = -ENOENT;
		} else {
			status = s_BeepCtrl.status;
			result = 0;
		}
		icx1087_beep_spin_unlock_bh();

		if (result == 0) {
			result = put_user(status, (int *)arg);
		}
	}
	return result;
}

static int BeepIoctl(struct inode *inode, struct file *file,
					unsigned int cmd, unsigned long arg)
{
	int val, result;

	switch (cmd) {
	case BEEP_SET_VOLUME:
		result = get_user(val, (int *)arg);
		PrDebug("BEEP_SET_VOLUME(val=%d)\n", val);
		if (result == 0) {
			if (val >= 0 && val <= 31) {
				result = SetBeepVol(val);
			} else {
				result = -EINVAL;
			}
		}
		break;
	case BEEP_GET_VOLUME:
		PrDebug("BEEP_GET_VOLUME\n");
		result = put_user(s_BeepCtrl.beep_vol, (int *)arg);
		break;
	case BEEP_PLAY:
		result = BeepPlayCmd(arg);
		break;
	case BEEP_STOP:
		result = BeepStopCmd(arg);
		break;
	case BEEP_STATUS:
		result = BeepStatusCmd(arg);
		break;
	case BEEP_SET_INTERVAL:
		result = get_user(val, (int *)arg);
		PrDebug("BEEP_SET_INTERVAL(val=%d)\n", val);
		if (result == 0) {
			if (val >= 0 && val < (10 * 1000)) {
				s_BeepCtrl.interval_time = val;
			} else {
				result = -EINVAL;
			}
		}
		break;
	case BEEP_GET_INTERVAL:
		PrDebug("BEEP_GET_INTERVAL\n");
		result = put_user(s_BeepCtrl.interval_time, (int *)arg);
		break;
	case BEEP_SYNC:
		result = BeepSyncCmd();
		break;
	default:
		PrInfo("Unsupport beep ioctl command. cmd=%x, arg=%lx\n", cmd, arg);
		result = -EINVAL;
		break;
	}
	return result;
}

static int	BeepSuspend(struct device *dev, u32 state, u32 level)
{
	PrDebug("\n");
	return 0;
}

static int	BeepResume(struct device *dev, u32 level)
{
	PrDebug("\n");
	return 0;
}

static int	BeepOpen(struct inode *inode, struct file *file)
{
	PrDebug("\n");
	return 0;
}

static int	BeepRelease(struct inode *inode, struct file *file)
{
	PrDebug("\n");
	return 0;
}

static struct file_operations s_BeepFops = {
	.owner		= THIS_MODULE,
	.ioctl		= BeepIoctl,
	.open		= BeepOpen,
	.release	= BeepRelease,
};

static struct device_driver	s_DeviceDriver = {
	.name		= BEEP_DEVICE_NAME,
	.bus		= &platform_bus_type,
	.suspend	= BeepSuspend,
	.resume		= BeepResume,
};

static inline int AllocChrDevNum(int *major, int minor, int count,
								const char *dev_name)
{
	dev_t	dev = 0;
	int	result;

	/* static or dynamic allocate character device number */
	if (*major != 0) {
		dev = MKDEV(*major, minor);
		result = register_chrdev_region(dev, count, dev_name);
	} else {
		result = alloc_chrdev_region(&dev, minor, count, dev_name);
		*major = MAJOR(dev);
	}
	return result;
}

int icx1087_beep_status(void)
{
	return s_BeepCtrl.status;
}

int icx1087_beep_init(void)
{
	dev_t	devno;		/* device number */
	int		result;

	PrInfo("Starting beep driver.\n");

	/* static or dynamic allocate device number */
	result = AllocChrDevNum(&s_BeepMajor, s_BeepMinor, 1, BEEP_DEVICE_NAME);
	if (result < 0) {
		PrErr("Can't allocate device major number (%d)\n", s_BeepMajor);
		return result;
	}
	devno = MKDEV(s_BeepMajor, s_BeepMinor);
	PrDebug("beep major device number (%d)\n", s_BeepMajor);

	/* initialize timer */
	init_timer(&s_BeepCtrl.timer);
	s_BeepCtrl.timer.data = 0;
	s_BeepCtrl.timer.function = BeepTimerHandler;

	init_waitqueue_head(&s_BeepCtrl.sync_wq);

	/* initialize character device object */
	cdev_init(&s_BeepCtrl.cdev, &s_BeepFops);
	s_BeepCtrl.cdev.owner = THIS_MODULE;

	/* add character device to kernel */
	result = cdev_add(&s_BeepCtrl.cdev, devno, 1);
	if (result < 0) {
		PrErr("cdev_add failed. (result=%d)\n", result);
		goto err3;
	}

	/* create /sys/class/beep directory */
	s_BeepCtrl.class = class_simple_create(THIS_MODULE, BEEP_DEVICE_NAME);
	if (IS_ERR(s_BeepCtrl.class)) {
		PrErr("class_simple_create failed.\n");
		result = PTR_ERR(s_BeepCtrl.class);
		goto err2;
	}

	/* create /sys/class/beep/beep/dev file */
	class_simple_device_add(s_BeepCtrl.class, devno,
					(struct device *)NULL, BEEP_DEVICE_NAME);

//	driver_register(&s_DeviceDriver);
//PrInfo("driver_register success\n");

	SetBeepVol(s_BeepCtrl.beep_vol);

#ifdef CONFIG_DEVFS_FS
	result = devfs_mk_cdev(devno, S_IFCHR | S_IRUSR | S_IWUSR,
				BEEP_DEV_ENTRY_NAME);
	if (result != 0) {
		PrErr("devfs_mk_cdev failed. (result=%d)\n", result);
		goto err1;
	}
#endif
	return 0;

err1:
//	driver_unregister(&s_DeviceDriver);

	/* remove /sys entry */
	class_simple_device_remove(devno);
	class_simple_destroy(s_BeepCtrl.class);
err2:
	/* delete character device object */
	cdev_del(&s_BeepCtrl.cdev);
err3:
	/* free character device number */
	unregister_chrdev_region(devno, 1);

	return result;
}

void icx1087_beep_exit(void)
{
	dev_t	devno = MKDEV(s_BeepMajor, s_BeepMinor);

	PrDebug("\n");

	wake_up_interruptible(&s_BeepCtrl.sync_wq);
	if (s_BeepCtrl.status != BEEP_STATUS_STOP) {
		/* Beep stop */
		del_timer_sync(&s_BeepCtrl.timer);
	}

//	driver_unregister(&s_DeviceDriver);

	/* remove /sys entry */
	class_simple_device_remove(devno);
	class_simple_destroy(s_BeepCtrl.class);

#ifdef CONFIG_DEVFS_FS
	devfs_remove(BEEP_DEV_ENTRY_NAME);	/* remove /dev entry */
#endif

	/* delete character device object */
	cdev_del(&s_BeepCtrl.cdev);

	/* free character device number */
	unregister_chrdev_region(devno, 1);
}

/* end of icx1087-beep.c */
