/*
 * OSS audio & mixer driver for Sony ICX1087 platform.
 *
 * Copyright 2006 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/config.h>			/* kernel configuration */
#include <linux/kernel.h>			/* often-used kernel functions */
#include <linux/init.h>				/* __init macro */
#include <linux/module.h>			/* kernel modules */
#include <linux/version.h>			/* kernel version */
#include <linux/device.h>			/* device and driver structure */
#include <linux/proc_fs.h>			/* proc file system */
#include <linux/interrupt.h>		/* local_irq_save(), local_irq_restore() */
#include <linux/delay.h>			/* msleep() */
#include <linux/sound.h>			/* (un)register_sound_dsp */
#include <linux/soundcard.h>		/* OSS API */
#include <asm/semaphore.h>			/* semaphore and mutex */
#include <asm/uaccess.h>			/* put_user(), get_user() */
#include <asm/io.h>					/* inl(), outl() */
#include <asm/arch/smu.h>			/* MP201 SMU */
#include <asm/arch/spi.h>			/* MP201 SPI */
#include <asm/arch/gpiocore.h>		/* gpiocore_pwc_write() */
#include <asm/arch/pmu.h>			/* MP201 PMU */
#include <asm/arch/icx1087_audio.h>	/* ICX1087 Audio Codec API */
#include <asm/arch/icx1087_beep.h>	/* ICX1087 Beep API */
#include "mp200-audio.h"			/* MP201 PCM driver */
#include "cs4251.h"					/* CS42L51 */
//#define DEBUG
#define PrHead	"icx1087_audio: "
#include <linux/prdbg.h>

//#define AUDIO_INITIALIZED_BY_DRIVER	/* see also: HowToPlayBeep.txt v1.2 */

#define CS4251_MCLK		(11289600)	/* master clock (44.1kHz * 256Fs) */
#define CS4251_POWERUP_TIME	(40)
#define CS4251_INIT_TIME	(70)

#define OSS_MAKE_VOL(lvol, rvol)	(((rvol) << 8) | (lvol))
#define OSS_GET_L_VOL(vol)			((vol) & 0xff)
#define OSS_GET_R_VOL(vol)			(((vol) >> 8) & 0xff)

/*
 * function prototypes
 */
extern int	icx1087_beep_init(void);
extern void	icx1087_beep_exit(void);
extern void	icx1087_beep_spin_lock_bh(void);
extern void	icx1087_beep_spin_unlock_bh(void);
extern int	icx1087_beep_status(void);

static void	DspHwInit(audio_stream_t *os, audio_stream_t *is);
static void	DspHwShutdown(audio_stream_t *os, audio_stream_t *is);
static int	DriverProbe(struct device *dev);
static int	DriverRemove(struct device *dev);
static int	DriverSuspend(struct device *dev, u32 state, u32 level);
static int	DriverResume(struct device *dev, u32 level);
static void DriverShutdown(struct device *dev);
static int	DspIoctl(struct inode *inode, struct file *file,
					unsigned int cmd, unsigned long arg);
static void PowerUpCodec(bool_t resume_flag);
static int	ResetHold(void);
static int	ResetRelease(void);

/* use MP201 PCM0 */
static pcm_regs_t s_Pcm0Regs = {
	.func_sel	= &PMx_FUNC_SEL(PCM0_ADDR),
	.cycle		= &PMx_CYCLE(PCM0_ADDR),
	.txrx_en	= &PMx_TXRX_EN(PCM0_ADDR),
	.txrx_dis	= &PMx_TXRX_DIS(PCM0_ADDR),
	.txrx_mon	= &PMx_TXRX_MON(PCM0_ADDR),
	.status		= &PMx_STATUS(PCM0_ADDR),
	.enset		= &PMx_ENSET(PCM0_ADDR),
	.enclr		= &PMx_ENCLR(PCM0_ADDR),
	.enmon		= &PMx_ENMON(PCM0_ADDR),
	.raw		= &PMx_RAW(PCM0_ADDR),
	.clear		= &PMx_CLEAR(PCM0_ADDR),
#ifdef MP201_SINGLE_SPECIAL
    .txq		= &PMx_TXQ(PCM0_ADDR),
    .fifo_p		= &PMx_FIFO_P(PCM0_ADDR),
#endif
	.cycle2		= &PMx_CYCLE2(PCM0_ADDR),
};

#ifdef MP201_PADDING_ON
	#define PCM_DATA_PADDING	PCM_PADDING_ON
#else
	#define PCM_DATA_PADDING	PCM_PADDING_OFF
#endif

/* arch-mp200/pcm_ioctl.h */
static pcm_ctrl_t s_PcmParams = {
	.func = {	/* function select */
		.mode_sel	= PCM_MODE_3,		/* left-justified format */
		.m_s		= PCM_SLAVE_MODE,	/* slave */
		.tx_tim		= PCM_TX_04_WORD,	/* transmission timing */
	},
	.cyc = {	/* cycle */
		.cyc_val	= 0x1f,				/* frame length = 32bit */
		.sib		= 0x0f,				/* data bit length = 16bit (PMx_SI) */
		.rx_pd		= PCM_DATA_PADDING,	/* data padding (RX) */
		.sob		= 0x0f,				/* data bit length = 16bit (PMx_SO) */
		.tx_pd		= PCM_DATA_PADDING,	/* data padding (TX) */
	},
	.cyc2 = {	/* cycle2 */
		.cyc_val2	= 0x1f,				/* frame length */
		.sib2		= 0x0f,				/* data bit length (PMx_SI) */
		.sob2		= 0x0f,				/* data bit length (PMx_SO) */
	},
	.formats_caps	= AFMT_S16_LE | AFMT_S32_LE,	/* support formats */
};

static pcm_ctrls_t s_PcmCtrls = {
	.sett	= &s_PcmParams,
	.regs	= &s_Pcm0Regs,
};

static audio_stream_t s_OutputStream = {
	.id			= "CS42L51 out",
	.buffers	= NULL,
	.dma_channel= MP200_DMAC_M2P_LCH4,	/* Memory to Peripheral (PCM0) */
	.pcm_lock	= SPIN_LOCK_UNLOCKED,
	.pcm_intr	= 0,
	.pcm_status	= 0,
	.pcm_regs	= &s_Pcm0Regs,
	.irq_bit	= PCM_INT_TX_FRE | PCM_INT_TX_URE | PCM_INT_TX_ORE,
	.cl_bit		= PCM_INT_TX_FRE | PCM_INT_TX_URE,
	.en_bit		= PCM_TX_EN,
	.phys_xq	= (MP200_PCM0_BASE + 0x24),	/* TXQ */
	.int_dma	= MP200_DMAC_INT_LENG_EN,
	.format		= AFMT_S16_LE,		/* default format. overwritable */
	.stereo		= 1,
	.type		= AUDIO_OUT,
};

static audio_stream_t s_InputStream = {
	.id			= "CS42L51 in",
	.buffers	= NULL,
	.dma_channel= MP200_DMAC_P2M_LCH4,	/* Peripheral to Memory (PCM0) */
	.pcm_lock	= SPIN_LOCK_UNLOCKED,
	.pcm_intr	= 0,
	.pcm_status	= 0,
	.pcm_regs	= &s_Pcm0Regs,
	.irq_bit	= PCM_INT_RX_FRE | PCM_INT_RX_URE | PCM_INT_RX_ORE,
	.cl_bit		= PCM_INT_RX_FRE,
	.en_bit		= PCM_RX_EN,
	.phys_xq	= (MP200_PCM0_BASE + 0x28),	/* RXQ */
	.int_dma	= MP200_DMAC_INT_LENG_EN,
	.format		= AFMT_S16_LE,		/* default format. overwritable */
	.stereo		= 1,
	.type		= AUDIO_IN,
};

static audio_state_t s_AudioState = {
	.output_stream	= &s_OutputStream,
	.input_stream	= &s_InputStream,
	.rd_ref			= 0,
	.wr_ref			= 0,
	.pcm_irq		= INT_PCM0,
	.hw_init		= DspHwInit,
	.hw_shutdown	= DspHwShutdown,
	.client_ioctl	= DspIoctl,
	.ctrls			= &s_PcmCtrls,
	.sem			= __MUTEX_INITIALIZER(s_AudioState.sem),
	.int_mode		= PCM_INT_CPU_MODE,
};

static struct device_driver s_DeviceDriver = {
	.name		= "pcm",
	.bus		= &platform_bus_type,
	.probe		= DriverProbe,
	.remove		= DriverRemove,
	.suspend	= DriverSuspend,
	.resume		= DriverResume,
	.shutdown	= DriverShutdown,
};

static unsigned	s_PwrIcVer;				/* PowerIC LSI version */
static pcm_regs_save_t s_Pcm0RegsSave;	/* MP201 PCM0 registers save area */
static int	s_MuteSave;					/* muting status save area */

#ifdef DEBUG
/* this function has 10us overhead at 248MHz */
static inline unsigned GetUS(void)
{
	struct timeval	tm;

	do_gettimeofday(&tm);
	return (unsigned)tm.tv_usec;
}
#endif /* DEBUG */

/* SPI write to CS42L51 */
int cs4251_spi_write(void *buf, size_t bytes /* not used */)
{
	static SPI_CONFIG	config = {
		.dev	= SPI_DEV_SPI0,			/* SPI-0 */
		.size	= 3,					/* 3bytes = 24bits */
		.mode	= (23 << 8) |			/* NB=23+1, CS2, Master, CPU Mode  */
			SPI_MODE_CS_SEL_CS2 | SPI_MODE_M_S_MASTER | SPI_MODE_DMA_OFF,
		.pol	= SPI_POL_VAL_SPI0, 	/* CK2_DLY:0, CK2_POL:1, CS2_POL:1 */
//		.sclk	= SPI_SCLK_1500KHZ,		/* 24bit = 16us+alpha */
		.sclk	= SPI_SCLK_6MHZ,		/* 24bit = 4us+alpha */
	};

	int	result = spi_write_without_intr(&config, buf, SPI_RESET_OFF);
	if (result != 0) {
		PrErr("spi_write_without_intr() failed. result=%d\n", result);
		return result;
	}
	result = 3;		/* 3bytes fix */

	return result;
}

/*
 * mixer interface
 */

/* Analog output volume table */
static uint8_t	s_VolTable[] = {
	0x34,	/*   0:-102.0dB */
	0x3a,	/*   1: -99.0dB */
	0x40,	/*   2: -96.0dB */
	0x46,	/*   3: -93.0dB */
	0x4c,	/*   4: -90.0dB */
	0x52,	/*   5: -87.0dB */
	0x58,	/*   6: -84.0dB */
	0x5e,	/*   7: -81.0dB */
	0x64,	/*   8: -78.0dB */
	0x6a,	/*   9: -75.0dB */
	0x70,	/*  10: -72.0dB */
	0x76,	/*  11: -69.0dB */
	0x7c,	/*  12: -66.0dB */
	0x82,	/*  13: -63.0dB */
	0x88,	/*  14: -60.0dB */
	0x8e,	/*  15: -57.0dB */
	0x94,	/*  16: -54.0dB */
	0x9a,	/*  17: -51.0dB */
	0xa0,	/*  18: -48.0dB */
	0xa2,	/*  19: -47.0dB */
	0xa4,	/*  20: -46.0dB */
	0xa6,	/*  21: -45.0dB */
	0xa8,	/*  22: -44.0dB */
	0xaa,	/*  23: -43.0dB */
	0xac,	/*  24: -42.0dB */
	0xae,	/*  25: -41.0dB */
	0xb0,	/*  26: -40.0dB */
	0xb2,	/*  27: -39.0dB */
	0xb4,	/*  29: -38.0dB */
	0xb6,	/*  29: -37.0dB */
	0xb8,	/*  30: -36.0dB */
	0xba,	/*  31: -35.0dB */
	0xbc,	/*  32: -34.0dB */
	0xbe,	/*  33: -33.0dB */
	0xc0,	/*  34: -32.0dB */
	0xc2,	/*  35: -31.0dB */
	0xc4,	/*  36: -30.0dB */
	0xc6,	/*  37: -29.0dB */
	0xc8,	/*  38: -28.0dB */
	0xca,	/*  39: -27.0dB */
	0xcc,	/*  40: -26.0dB */
	0xce,	/*  41: -25.0dB */
	0xd0,	/*  42: -24.0dB */
	0xd2,	/*  43: -23.0dB */
	0xd4,	/*  44: -22.0dB */
	0xd6,	/*  45: -21.0dB */
	0xd8,	/*  46: -20.0dB */
	0xda,	/*  47: -19.0dB */
	0xdc,	/*  48: -18.0dB */
	0xde,	/*  49: -17.0dB */
	0xe0,	/*  50: -16.0dB */
	0xe2,	/*  51: -15.0dB */
	0xe4,	/*  52: -14.0dB */
	0xe6,	/*  53: -13.0dB */
	0xe8,	/*  54: -12.0dB */
	0xea,	/*  55: -11.0dB */
	0xec,	/*  56: -10.0dB */
	0xed,	/*  57:  -9.5dB */
	0xee,	/*  58:  -9.0dB */
	0xef,	/*  59:  -8.5dB */
	0xf0,	/*  60:  -8.0dB */
	0xf1,	/*  61:  -7.5dB */
	0xf2,	/*  62:  -7.0dB */
	0xf3,	/*  63:  -6.5dB */
	0xf4,	/*  64:  -6.0dB */
	0xf5,	/*  65:  -5.5dB */
	0xf6,	/*  66:  -5.0dB */
	0xf7,	/*  67:  -4.5dB */
	0xf8,	/*  68:  -4.0dB */
	0xf9,	/*  69:  -3.5dB */
	0xfa,	/*  70:  -3.0dB */
	0xfb,	/*  71:  -2.5dB */
	0xfc,	/*  72:  -2.0dB */
	0xfd,	/*  73:  -1.5dB */
	0xfe,	/*  74:  -1.0dB */
	0xff,	/*  75:  -0.5dB */
	0x00,	/*  76:   0.0dB */
	0x01,	/*  77:  +0.5dB */
	0x02,	/*  78:  +1.0dB */
	0x03,	/*  79:  +1.5dB */
	0x04,	/*  80:  +2.0dB */
	0x05,	/*  81:  +2.5dB */
	0x06,	/*  82:  +3.0dB */
	0x07,	/*  83:  +3.5dB */
	0x08,	/*  84:  +4.0dB */
	0x09,	/*  85:  +4.5dB */
	0x0a,	/*  86:  +5.0dB */
	0x0b,	/*  87:  +5.5dB */
	0x0c,	/*  88:  +6.0dB */
	0x0d,	/*  89:  +6.5dB */
	0x0e,	/*  90:  +7.0dB */
	0x0f,	/*  91:  +7.5dB */
	0x10,	/*  92:  +8.0dB */
	0x11,	/*  93:  +8.5dB */
	0x12,	/*  94:  +9.0dB */
	0x13,	/*  95:  +9.5dB */
	0x14,	/*  96: +10.0dB */
	0x15,	/*  97: +10.5dB */
	0x16,	/*  98: +11.0dB */
	0x17,	/*  99: +11.5dB */
	0x18	/* 100: +12.0dB */
};

typedef struct {
	int			mute;			/* mute status */
	int			play_src;		/* playback source */
	int			out_dev;		/* output device */
	int			aout_vol;		/* analog output volume */
	int			bass;			/* bass volume */
	int			treble;			/* treble volume */
	int			pcm_vol;		/* PCM mixer volume */
	int			altpcm_vol;		/* ADC mixer volume */
} mixer_setting_t;

static mixer_setting_t	s_MixerSettings ={
	.mute		= AUDIO_MUTE_OFF,
	.play_src	= AUDIO_PLAYBACK_PCM,
	.out_dev	= AUDIO_OUTPUT_HP,
	.aout_vol	= OSS_MAKE_VOL(50, 50),
	.bass		= OSS_MAKE_VOL(50, 50),
	.treble		= OSS_MAKE_VOL(50, 50),
	.pcm_vol	= OSS_MAKE_VOL(76, 76),
	.altpcm_vol	= OSS_MAKE_VOL(76, 76),
};

/* only use when Power ON/OFF */
static int	ExternalMute(bool_t flag)
{
	/* Power IC GPIO12 pin: Low=mute on (10ms) High=mute off (5ms) */
	return gpiocore_pwc_write(GPIOCORE_PWC_IOOUT2,
				(flag) ? 0x00 : 0x10 /* data */, 0x10 /* mask */);
}

static inline int IsChangeBit(unsigned n1, unsigned n2, unsigned mask)
{
	return (n1 & mask) != (n2 & mask);
}

/* calling from beep function with spin_lock_bh */
/* return value: old mute value, or -1 is error */
inline int	icx1087_mute(int mask, int value)
{
	int	old_mute, new_mute, result = 0;

	old_mute = new_mute = s_MixerSettings.mute;
	new_mute &= ~mask;
	new_mute |= (value & mask);

	PrDebug("mask=0x%02x, val=%02x, old=0x%02x, new=0x%02x\n",
			mask, value, old_mute, new_mute);

	if (result == 0 && IsChangeBit(old_mute, new_mute, AUDIO_ADC_MIX_MUTE)) {
		result = cs4251_set_mute_adc_mix(new_mute & AUDIO_ADC_MIX_MUTE);
	}
	if (result == 0 && IsChangeBit(old_mute, new_mute, AUDIO_PCM_MIX_MUTE)) {
		result = cs4251_set_mute_pcm_mix(new_mute & AUDIO_PCM_MIX_MUTE);
	}
	if (result == 0 && IsChangeBit(old_mute, new_mute, AUDIO_DAC_MUTE)) {
		result = cs4251_set_dac_mute(new_mute & AUDIO_DAC_MUTE);
	}
	if (result == 0 && IsChangeBit(old_mute, new_mute, AUDIO_ADC_MUTE)) {
		result = cs4251_set_adc_mute(new_mute & AUDIO_ADC_MUTE);
	}
	if (result == 0 && IsChangeBit(old_mute, new_mute, AUDIO_EXT_MUTE)) {
		result = ExternalMute(new_mute & AUDIO_EXT_MUTE);
	}

	if (result == 0) {
		s_MixerSettings.mute = new_mute;
		result = old_mute;
	}

	return result;
}

static int DoMute(int mask, int value)
{
	int	old;

	icx1087_beep_spin_lock_bh();
	old = icx1087_mute(mask, value);
	icx1087_beep_spin_unlock_bh();
	return old;
}

static int SetPlaybackSrc(int src)
{
	int	result;

	switch (src) {
	case AUDIO_PLAYBACK_PCM:
		/* ADC powerdown */
		result = cs4251_set_powerdown_pga(On);
		result = cs4251_set_powerdown_mic(On);
		result = cs4251_set_powerdown_adc(On);
		break;
	case AUDIO_PLAYBACK_LINE:
	case AUDIO_PLAYBACK_TUNER:
		/* ADC powerup */
		result = cs4251_set_powerdown_pga(Off);
		result = cs4251_set_powerdown_mic(On);
		result = cs4251_set_powerdown_adc(Off);
		result = cs4251_set_adc_path((src == AUDIO_PLAYBACK_LINE) ?
					CS4251_AIN_MUX_PATH_AIN1 : CS4251_AIN_MUX_PATH_AIN2);
		break;
	case AUDIO_PLAYBACK_MIC:
		result = cs4251_set_powerdown_pga(Off);
		result = cs4251_set_powerdown_mic(Off);
		result = cs4251_set_powerdown_adc(Off);
//		result = cs4251_set_adc_path(CS4251_AIN_MUX_PATH_AIN3);
		result = cs4251_set_adc_path(CS4251_AIN_MUX_PATH_AIN3_PREAMP);
		break;
	default:
		result = -EINVAL;
		break;
	}

	if (result == 0) {
		s_MixerSettings.play_src = src;
	}
	return result;
}

/* set output device (HP or Cradle Line Out) */
static int SetOutputDevice(int device)
{
	unsigned	val;
	int	result;

	if (s_PwrIcVer >= 2) { /* ES1.2 or later? */
		/* Power IC GPIO11 pin: Low=Cradle Line Out, High=HP) */
		val = (device == AUDIO_OUTPUT_HP) ? 0x08 : 0x00;
	} else {
		/* Power IC GPIO11 pin: Low=HP, High=Cradle Line Out) */
		val = (device == AUDIO_OUTPUT_HP) ? 0x00 : 0x08;
	}

	/* Power IC GPIO11 pin: Low=HP, High=Cradle Line Out) */
	result = gpiocore_pwc_write(GPIOCORE_PWC_IOOUT2, val, 0x08 /* mask */);
	if (result == 0) {
		s_MixerSettings.out_dev = device;
	}
PrDebug("device=%d, lsi_ver=%d, result=%d\n", device, s_PwrIcVer, result);
	return result;
}

static inline int	OssAdjustVol(int vol)
{
	if (vol < 0) {
		vol = 0;
	} else if (vol > 100) {
		vol = 100;
	}
	return vol;
}

/* calling from beep stop function */
int	icx1087_restore_aout_vol(void)
{
	return cs4251_set_aout_vol(
				s_VolTable[OSS_GET_L_VOL(s_MixerSettings.aout_vol)],
				s_VolTable[OSS_GET_R_VOL(s_MixerSettings.aout_vol)]);
}

static int	SetAoutVol(int lch, int rch)
{
	int	result = 0;

	lch = OssAdjustVol(lch);
	rch = OssAdjustVol(rch);

	icx1087_beep_spin_lock_bh();

	if (icx1087_beep_status() != BEEP_STATUS_STOP) {
		/* Don't touch CS42L51 */
		s_MixerSettings.aout_vol = OSS_MAKE_VOL(lch, rch);
	} else {
		result = cs4251_set_aout_vol(s_VolTable[lch], s_VolTable[rch]);
		if (result == 0) {
			s_MixerSettings.aout_vol = OSS_MAKE_VOL(lch, rch);
		}
	}

	icx1087_beep_spin_unlock_bh();
	return result;
}

/* convert from OSS volume (0 to 100) to tone volume (-10.5 to +12.0dB) */
static inline int	OssVolToToneVol(int vol)
{
	/* convert from volume (0 to 100) to tone level (0 to 15) */
	int	level = 16 - (vol + (100/16) / 2) / (100/16);
	if (level < 0) {
		level = 0;
	} else if (level > 15) {
		level = 15;
	}
	return level;
}

static int	SetBassLevel(int lch, int rch)
{
	int	result;

	lch = OssAdjustVol(lch);
	rch = OssAdjustVol(rch);

	result = cs4251_set_bass(OssVolToToneVol((lch + rch) / 2));
	if (result == 0) {
		s_MixerSettings.bass = OSS_MAKE_VOL(lch, rch);

		/* enable tone control */
		result = cs4251_set_tone_enable(On);
	}
	return result;
}

static int	SetTrebleLevel(int lch, int rch)
{
	int	result;

	lch = OssAdjustVol(lch);
	rch = OssAdjustVol(rch);

	result = cs4251_set_treble(OssVolToToneVol((lch + rch) / 2));
	if (result == 0) {
		s_MixerSettings.treble = OSS_MAKE_VOL(lch, rch);

		/* enable tone control */
		result = cs4251_set_tone_enable(On);
	}
	return result;
}

/* convert from OSS volume (0 to 100) to mixer volume (-51.5 to +12.0dB) */
static inline int	OssVolToMixVol(int vol)
{
	int	n;

	if (vol <= 26) {
		n = (vol * 2) - 103;	/* -51.5 to -25.5dB, step 1.0dB */
	} else {
		n = vol - (100 - 24);	/* -24.5 to +12.0dB, step 0.5dB */
	}
	return n & 0x7f;
}

static int	SetPcmMixVol(int lch, int rch)
{
	int	result;

	lch = OssAdjustVol(lch);
	rch = OssAdjustVol(rch);

	result = cs4251_set_pcm_mix_vol(OssVolToMixVol(lch), OssVolToMixVol(rch));
	if (result == 0) {
		s_MixerSettings.pcm_vol = OSS_MAKE_VOL(lch, rch);
	}
	return result;
}

static int	SetAdcMixVol(int lch, int rch)
{
	int	result;

	lch = OssAdjustVol(lch);
	rch = OssAdjustVol(rch);

	result = cs4251_set_adc_mix_vol(OssVolToMixVol(lch), OssVolToMixVol(rch));
	if (result == 0) {
		s_MixerSettings.altpcm_vol = OSS_MAKE_VOL(lch, rch);
	}
	return result;
}

static int MixerIoctl(struct inode *inode, struct file *file,
					unsigned int cmd, unsigned long arg)
{
	int val, result = 0;

	switch (cmd) {
	case PCM_CODEC_SET:
		result = get_user(val, (int *)arg);
		PrDebug("PCM_CODEC_SET(val=%x)\n", val);
		/* b15-8: register number, b7-0: data */
//		result = cs4251_write_reg((val >> 8) & 0xff, val & 0xff);
		break;
	case SOUND_MIXER_READ_DEVMASK:	/* available device */
		PrDebug("SOUND_MIXER_READ_DEVMASK\n");
		result = put_user(SOUND_MASK_VOLUME |
					SOUND_MASK_BASS | SOUND_MASK_TREBLE |
					SOUND_MASK_PCM | SOUND_MASK_ALTPCM, (int *)arg);
		break;
	case SOUND_MIXER_READ_RECMASK:	/* recordable device */
		PrDebug("SOUND_MIXER_READ_RECMASK\n");
		result = put_user(0, (int *)arg);	/* no recordable device */
		break;
	case SOUND_MIXER_READ_STEREODEVS:	/* stereo device */
		PrDebug("SOUND_MIXER_READ_STEREODEVS\n");
		result = put_user(SOUND_MASK_VOLUME |
					SOUND_MASK_PCM | SOUND_MASK_ALTPCM, (int *)arg);
		break;
	case SOUND_MIXER_READ_RECSRC:	/* get current recoding device */
		PrDebug("SOUND_MIXER_READ_RECSRC\n");
		result = put_user(0, (int *)arg);	/* no recording device */
		break;
	case SOUND_MIXER_WRITE_RECSRC:	/* set current recording device */
		result = get_user(val, (int *)arg);
		PrDebug("SOUND_MIXER_WRITE_RECSRC(val=0x%x)\n", val);
		if (result == 0) {
			result = put_user(0, (int *)arg);	/* no recording device */
		}
		break;
	case SOUND_MIXER_WRITE_VOLUME:	/* Analog output volume */
		result = get_user(val, (int *)arg);
		PrDebug("SOUND_MIXER_WRITE_VOLUME(val=0x%x)\n", val);
		if (result == 0) {
			result = SetAoutVol(OSS_GET_L_VOL(val), OSS_GET_R_VOL(val));
			if (result == 0) {
				/* copy back */
				result = put_user(s_MixerSettings.aout_vol, (int *)arg);
			}
		}
		break;
	case SOUND_MIXER_READ_VOLUME:
		PrDebug("SOUND_MIXER_READ_VOLUME\n");
		result = put_user(s_MixerSettings.aout_vol, (int *)arg);
		break;
	case SOUND_MIXER_WRITE_BASS:
		result = get_user(val, (int *)arg);
		PrDebug("SOUND_MIXER_WRITE_BASS(val=0x%x)\n", val);
		if (result == 0) {
			result = SetBassLevel(OSS_GET_L_VOL(val), OSS_GET_R_VOL(val));
			if (result == 0) {
				/* copy back */
				result = put_user(s_MixerSettings.bass, (int *)arg);
			}
		}
		break;
	case SOUND_MIXER_READ_BASS:
		PrDebug("SOUND_MIXER_READ_BASS\n");
		result = put_user(s_MixerSettings.bass, (int *)arg);
		break;
	case SOUND_MIXER_WRITE_TREBLE:
		result = get_user(val, (int *)arg);
		PrDebug("SOUND_MIXER_WRITE_TREBLE(val=0x%x)\n", val);
		if (result == 0) {
			result = SetTrebleLevel(OSS_GET_L_VOL(val), OSS_GET_R_VOL(val));
			if (result == 0) {
				/* copy back */
				result = put_user(s_MixerSettings.treble, (int *)arg);
			}
		}
		break;
	case SOUND_MIXER_READ_TREBLE:
		PrDebug("SOUND_MIXER_READ_TREBLE\n");
		result = put_user(s_MixerSettings.treble, (int *)arg);
		break;
	case SOUND_MIXER_WRITE_PCM:		/* PCM mixer volume */
		result = get_user(val, (int *)arg);
		PrDebug("SOUND_MIXER_WRITE_PCM(val=0x%x)\n", val);
		if (result == 0) {
			result = SetPcmMixVol(OSS_GET_L_VOL(val), OSS_GET_R_VOL(val));
			if (result == 0) {
				/* copy back */
				result = put_user(s_MixerSettings.pcm_vol, (int *)arg);
			}
		}
		break;
	case SOUND_MIXER_READ_PCM:		/* PCM mixer volume */
		PrDebug("SOUND_MIXER_READ_PCM\n");
		result = put_user(s_MixerSettings.pcm_vol, (int *)arg);
		break;
	case SOUND_MIXER_WRITE_ALTPCM:	/* ADC mixer volume */
		result = get_user(val, (int *)arg);
		PrDebug("SOUND_MIXER_WRITE_ALTPCM(val=0x%x)\n", val);
		if (result == 0) {
			result = SetAdcMixVol(OSS_GET_L_VOL(val), OSS_GET_R_VOL(val));
			if (result == 0) {
				/* copy back */
				result = put_user(s_MixerSettings.altpcm_vol, (int *)arg);
			}
		}
		break;
	case SOUND_MIXER_READ_ALTPCM:	/* ADC mixer volume */
		PrDebug("SOUND_MIXER_READ_ALTPCM\n");
		result = put_user(s_MixerSettings.altpcm_vol, (int *)arg);
		break;
	case AUDIO_SET_MUTE:
		result = get_user(val, (int *)arg);
		PrDebug("AUDIO_SET_MUTE(val=%x)\n", val);
		if (result == 0) {
			DoMute(AUDIO_ALL_MUTE_ON, val & AUDIO_ALL_MUTE_ON);
		}
		break;
	case AUDIO_GET_MUTE:
		PrDebug("AUDIO_GET_MUTE\n");
		result = put_user(s_MixerSettings.mute, (int *)arg);
		break;
	case AUDIO_SET_PLAYBACK:	/* set playback source */
		result = get_user(val, (int *)arg);
		PrDebug("AUDIO_SET_PLAYBACK(val=%x)\n", val);
		if (result == 0) {
			switch (val) {
			case AUDIO_PLAYBACK_PCM:
			case AUDIO_PLAYBACK_LINE:
			case AUDIO_PLAYBACK_TUNER:
			case AUDIO_PLAYBACK_MIC:
				SetPlaybackSrc(val);
				break;
			default:
				result = -EINVAL;
				break;
			}
		}
		break;
	case AUDIO_GET_PLAYBACK:	/* get playback source */
		PrDebug("AUDIO_GET_PLAYBACK\n");
		result = put_user(s_MixerSettings.play_src, (int *)arg);
		break;
	case AUDIO_SET_OUTPUT:	/* set output device */
		result = get_user(val, (int *)arg);
		PrDebug("AUDIO_SET_OUTPUT(val=%x)\n", val);
		if (result == 0) {
			if (val == AUDIO_OUTPUT_HP || val == AUDIO_OUTPUT_LINE) {
				SetOutputDevice(val);
			} else {
				result = -EINVAL;
			}
		}
		break;
	case AUDIO_GET_OUTPUT:	/* get output device */
		PrDebug("AUDIO_GET_OUTPUT\n");
		result = put_user(s_MixerSettings.out_dev, (int *)arg);
		break;
	case AUDIO_SET_AOUT_VOL_TABLE:
	case AUDIO_GET_AOUT_VOL_TABLE:
	{
		audio_vol_table_t	param;

		PrDebug("AUDIO_{S|G}ET_VOL_TABLE\n");
		if (copy_from_user(&param, (void *)arg, sizeof(param))) {
			return -EFAULT;
		}
		if (param.offset > 100 || param.count == 0 ||
			param.offset + param.count > 101 || param.buf == NULL) {
			return -EINVAL;
		}
		if (cmd == AUDIO_SET_AOUT_VOL_TABLE) {
			if (copy_from_user(&s_VolTable[param.offset],
								param.buf, param.count)) {
				return -EFAULT;
			}
			SetAoutVol(OSS_GET_L_VOL(s_MixerSettings.aout_vol),
					  OSS_GET_R_VOL(s_MixerSettings.aout_vol));
		} else {
			if (copy_to_user(param.buf,
							&s_VolTable[param.offset], param.count)) {
				return -EFAULT;
			}
		}
		break;
	}
	case AUDIO_RESTART:	/* restart audio codec device */
		PowerUpCodec(On);
		break;
	default:
		PrInfo("Unsupport mixer ioctl command. cmd=%x, arg=%lx\n", cmd, arg);
		result = -EINVAL;
		break;
	}

	return result;
}

static int MixerOpen(struct inode *inode, struct file *file)
{
	PrDebug("\n");
	return 0;
}

static int MixerRelease(struct inode *inode, struct file *file)
{
	PrDebug("\n");
	return 0;
}

static struct file_operations s_MixerFops = {
	.open		= MixerOpen,
	.release	= MixerRelease,
	.ioctl		= MixerIoctl,
	.owner		= THIS_MODULE
};

static inline void OpenCodecClock(void)
{
	/* setup REFCLKO signal (connect to CS42L51 MCLK pin) */
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_REFCLKO);
	outl(0, SMU_REFCLKO_SEL);		/* clock source = OSC */
	outl(0, SMU_REFCLKO_DIV);		/* divide by 1 */
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_REFCLKO);
}

static inline void CloseCodecClock(void)
{
	/* REFCLKO gate close */
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_REFCLKO);
}

static void OpenPcm0Clock(void)
{
// SCLK=Audio Serial Clock?, PCLK=Peripheral Clock?
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_SCLK);
	mp200_pmu_unreset_device(MP200_RESETDEVICE_PM0);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_SCLK);
}

static void ClosePcm0Clock(void)
{
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_SCLK);
	mp200_pmu_reset_device(MP200_RESETDEVICE_PM0);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_SCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_PCLK);
}

/* Recommended Power-Up Sequence
 1. Hold RESET low until the power supplies are stable.
 2. Bring RESET high. After approximately 10 ms,
    the device will enter Hardware Mode.
 3. For Software Mode operation, set the PDN bit to '1'b under 10 ms.
    This will place the device in "standby".
 4. Load the desired register settings while keeping the PDN bit set to '1'b
 5. Start MCLK to the appropriate frequency, as discussed in Section 4.5.
 6. Set the PDN bit to '0'b.
 7. Apply LRCK, SCLK and SDIN for normal operation to begin.
 8. Bring RESET low if the analog or digital supplies drop below the recommended
    operating condition to prevent power glitch related issues.
*/
static void PowerUpCodec(bool_t resume_flag)
{
#ifndef AUDIO_INITIALIZED_BY_DRIVER
	if (resume_flag == On) {
#endif /* AUDIO_INITIALIZED_BY_DRIVER */

		/* audio power on */
		if (s_PwrIcVer >= 3) { /* ES2.02 or later? */
			pwc_regulator_on(PWC_PSCNTHP);
			PrDebug("REGHP on\n");
			msleep(CS4251_POWERUP_TIME);
		}
		CloseCodecClock();

		/* reset release and set Software Mode */
		ResetRelease();

		/* base voltage power saving */
		cs4251_write_reg(0x00, 0x99);	/* unofficial reg */
		cs4251_write_reg(0x37, 0x05);	/* unofficial reg. amount of current */
										/* 0x00(min) to 0x08(max, default) */

		/* DAC/HP AMP Queiscent power saving */
		cs4251_write_reg(0x00, 0x99);	/* unofficial register */
		cs4251_write_reg(0x38, 0xc0);	/* unofficial register */
#ifndef AUDIO_INITIALIZED_BY_DRIVER
	}
#endif /* AUDIO_INITIALIZED_BY_DRIVER */

	if (resume_flag == On) {
		cs4251_restore_regs();
		cs4251_set_powerdown_dac(Off);

		/* codec power up */
		cs4251_set_powerdown(Off);
		OpenCodecClock();
		msleep(CS4251_INIT_TIME);
		SetOutputDevice(s_MixerSettings.out_dev);
		msleep(40);
		DoMute(AUDIO_ALL_MUTE_ON, s_MuteSave);
//		msleep(5);	/* external mute delay */
		return;
	}

	/* regular configuration */
	cs4251_write_reg(0x03, 0x2f);	/* SingleSpeedMode, MCLK/2 */
	cs4251_write_reg(0x04, 0x40);	/* Master, Left Justified */
	cs4251_write_reg(0x05, 0x00);	/* */
	cs4251_write_reg(0x06, 0xaa);	/* High pass:enable, Soft Ramp:enable*/
	cs4251_write_reg(0x07, 0x00);	/* AIN1(LINE IN) -> ADC */
	cs4251_write_reg(0x08, 0x80);	/* HP-Gain:0.7099 */

	cs4251_write_reg(0x09, 0x42);	/* SPE -> DAC, Soft Ramp */

	cs4251_write_reg(0x0a, 0x0b);	/* PGA Gain A: +5.5dB */
	cs4251_write_reg(0x0b, 0x0b);	/* PGA Gain B: +5.5dB */

	cs4251_write_reg(0x0c, 0x00);	/* ADC Attenuator: 0dB */
	cs4251_write_reg(0x0d, 0x00);	/* ADC Attenuator: 0dB */

	cs4251_write_reg(0x0e, 0x00);	/* ADC Mixer Channel Mute:disable, 0dB */
	cs4251_write_reg(0x0f, 0x00);	/* ADC Mixer Channel Mute:disable, 0dB */

	cs4251_write_reg(0x10, 0x00);	/* PCM Mixer Channel Mute:disable, 0dB */
	cs4251_write_reg(0x11, 0x00);	/* PCM Mixer Channel Mute:disable, 0dB */

	cs4251_write_reg(0x12, 0xf0);	/* Freq:2004.55Hz, BeepOn:86ms */
	cs4251_write_reg(0x13, 0x00);	/* BeepOff:1.23S, Vol:0dB */
	cs4251_write_reg(0x14, 0x00);	/* Beep:disable, Tone:disable */
	cs4251_write_reg(0x15, 0x88);	/* Treble:0dB, Bass:0dB */

	/* Reg16, 17 */
	cs4251_set_aout_vol(
		s_VolTable[OSS_GET_L_VOL(s_MixerSettings.aout_vol)],
		s_VolTable[OSS_GET_R_VOL(s_MixerSettings.aout_vol)]);

	cs4251_write_reg(0x18, 0x00);	/* */
	cs4251_write_reg(0x19, 0x00);	/* */
	cs4251_write_reg(0x1a, 0x7f);	/* */
	cs4251_write_reg(0x1b, 0x00);	/* */
	cs4251_write_reg(0x1c, 0x00);	/* */
	cs4251_write_reg(0x1d, 0x3f);	/* */
	cs4251_write_reg(0x1e, 0x00);	/* */
	cs4251_write_reg(0x1f, 0x00);	/* */
	cs4251_write_reg(0x21, 0x80);	/* Charge Pump Frequency */

	/* DAC & Codec power up */
	cs4251_write_reg(0x02, 0x1e);

	/* initial source is PCM */
	SetPlaybackSrc(AUDIO_PLAYBACK_PCM);

#ifdef AUDIO_INITIALIZED_BY_DRIVER
	OpenCodecClock();
	msleep(CS4251_INIT_TIME);

	DoMute(AUDIO_EXT_MUTE, 0);	/* external mute off */
//	msleep(5);	/* external mute delay */
#endif /* AUDIO_INITIALIZED_BY_DRIVER */
}

/* Recommended Power-Down Sequence
 To minimize audible pops when turning off or placing the CODEC in standby,
 1. Mute the DAC's & ADC's.
 2. Set the PDN bit in the power control register to '1'b.
    The CODEC will not power down until it reaches a fully muted sate.
 3. Bring RESET low.
*/
static void PowerDownCodec(void)
{
	int	save_dev;
	s_MuteSave = DoMute(AUDIO_ALL_MUTE_ON, AUDIO_ALL_MUTE_ON);
	cs4251_set_powerdown_dac(On);
	cs4251_set_powerdown(On);	/* codec power down */
	msleep(5);	/* wait for external mute complete */

	save_dev = s_MixerSettings.out_dev;
	SetOutputDevice(AUDIO_OUTPUT_LINE);
	s_MixerSettings.out_dev = save_dev;

	ResetHold();

	CloseCodecClock();

	/* audio power off */
	if (s_PwrIcVer >= 3) { /* ES2.02 or later? */
		PrDebug("REGHP off\n");
		msleep(40);
		pwc_regulator_off(PWC_PSCNTHP);
	}
}

static int DriverSuspend(struct device *dev, u32 state, u32 level)
{
	/*
	 * Deep sleep: DISABLE -> SAVE_STATE -> POWER_DOWN
	 */

	switch (level) {
	case SUSPEND_NOTIFY:
		PrDebug("SUSPEND_NOTIFY (state=%x, level=%x)\n", state, level);
		break;
	case SUSPEND_DISABLE:
		PrDebug("SUSPEND_DISABLE (state=%x, level=%x)\n", state, level);
#if 0
		if ((s_AudioState.wr_ref | s_AudioState.rd_ref) & AUDIO_REF_EN) {
			return -EBUSY;
		}
#endif
		break;
	case SUSPEND_SAVE_STATE:
		PrDebug("SUSPEND_SAVE_STATE (state=%x, level=%x)\n", state, level);

		PowerDownCodec();

		/* save PCM0 registers */
		s_Pcm0RegsSave.func_sel	= *s_Pcm0Regs.func_sel;
		s_Pcm0RegsSave.cycle	= *s_Pcm0Regs.cycle;
		s_Pcm0RegsSave.cycle2	= *s_Pcm0Regs.cycle2;
		s_Pcm0RegsSave.enmon	= *s_Pcm0Regs.enmon;
		s_Pcm0RegsSave.txrx_mon	= *s_Pcm0Regs.txrx_mon;

		ClosePcm0Clock();	/* reset and power down PCM0 */
		break;
	case SUSPEND_POWER_DOWN:
		PrDebug("SUSPEND_POWER_DOWN (state=%x, level=%x)\n", state, level);
		break;
	default:
		PrDebug("Unknown level (state=%x, level=%x)\n", state, level);
		break;
	}
	return 0;
}

static int DriverResume(struct device *dev, u32 level)
{
	/*
	 * POWER_ON -> RESTORE_STATE -> ENABLE
	 */

	switch (level) {
	case RESUME_POWER_ON:
		PrDebug("RESUME_POWER_ON (level=%x)\n", level);
		break;
	case RESUME_RESTORE_STATE:
		PrDebug("RESUME_RESTORE_STATE (level=%x)\n", level);

		OpenPcm0Clock();	/* power up and unreset PCM0 */

		/* restore PCM0 registers */
		*s_Pcm0Regs.func_sel	= s_Pcm0RegsSave.func_sel;
		*s_Pcm0Regs.cycle		= s_Pcm0RegsSave.cycle;
		*s_Pcm0Regs.cycle2		= s_Pcm0RegsSave.cycle2;
		*s_Pcm0Regs.enclr		= (~s_Pcm0RegsSave.enmon);
		*s_Pcm0Regs.enset		= s_Pcm0RegsSave.enmon;
		*s_Pcm0Regs.txrx_dis	= (~s_Pcm0RegsSave.txrx_mon);
		*s_Pcm0Regs.txrx_en		= s_Pcm0RegsSave.txrx_mon;

		PowerUpCodec(On);
		break;
	case RESUME_ENABLE:
		PrDebug("RESUME_ENABLE (level=%x)\n", level);
		break;
	default:
		PrDebug("Unknown level (level=%x)\n", level);
		break;
	}
	return 0;
}

static void DriverShutdown(struct device *dev)
{
	PrDebug("\n");
	PowerDownCodec();
}

/*
 * audio (dsp) interface
 */

static void DspHwInit(audio_stream_t *os, audio_stream_t *is)
{
	PrDebug("\n");

	if (os != NULL) {
		s_AudioState.wr_ref |= AUDIO_REF_EN;	/* set open reference bit */

		/* not power on codec ? */
		if ((s_AudioState.wr_ref & AUDIO_REF_PW) == 0) {
			/* ak4642: read volume and write register */

			s_AudioState.wr_ref |= AUDIO_REF_PW;	/* set power on codec bit */
		}
	}
	if (is != NULL) {
		s_AudioState.rd_ref |= AUDIO_REF_EN;	/* set open reference bit */

		/* not power on codec ? */
		if ((s_AudioState.rd_ref & AUDIO_REF_PW) == 0) {
			/* ak4642: read volume and write register */

			s_AudioState.rd_ref |= AUDIO_REF_PW;	/* set power on codec bit */
		}
	}

//	if (error occured) {
//		s_AudioState.wr_ref |= AUDIO_REF_ERR;	/* set error bit */
//		s_AudioState.rd_ref |= AUDIO_REF_ERR;	/* set error bit */
//	}
}

static void DspHwShutdown(audio_stream_t *os, audio_stream_t *is)
{
	PrDebug("\n");

	msleep(20);

	if (os != NULL) {
		s_AudioState.wr_ref = 0;
	}
	if (is != NULL) {
		s_AudioState.rd_ref = 0;
	}
}

static int DspIoctl(struct inode *inode, struct file *file,
					unsigned int cmd, unsigned long arg)
{
	int val, result = 0;

	PrDebug("cmd=%x\n", cmd);

	switch (cmd) {
	case SNDCTL_DSP_SPEED:
		result = get_user(val, (int *)arg);
		PrDebug("SNDCTL_DSP_SPEED(%d)\n", val);
		if (result == 0) {
			if (val <= 12000) {
				val = 11025;
				outl(3, SMU_REFCLKO_DIV);		/* divide by 4 */
			} else if (val <= 24000) {
				val = 22050;
				outl(1, SMU_REFCLKO_DIV);		/* divide by 2 */
			} else {
				val = 44100;
				outl(0, SMU_REFCLKO_DIV);		/* divide by 1 */
			}
			result = put_user(val, (int *)arg);	/* copy back to user */
		}
		break;
	case SNDCTL_DSP_STEREO:
		result = get_user(val, (int *)arg);
		PrDebug("SNDCTL_DSP_STEREO(%x)\n", val);
		if (result == 0) {
			/* stereo only */
			if (val != 1) {
				val = -EINVAL;
			}
			result = put_user(val, (int *)arg);
		}
		break;
	case SNDCTL_DSP_CHANNELS:
		result = get_user(val, (int *)arg);
		PrDebug("SNDCTL_DSP_CHANNELS(%d)\n", val);
		if (result == 0) {
			/* stereo only */
			result = put_user(2, (int *)arg);
		}
		break;
	case SNDCTL_DSP_GETFMTS:
		PrDebug("SNDCTL_DSP_GETFMTS\n");
		result = put_user(s_PcmParams.formats_caps, (int *)arg);
		break;
	case SNDCTL_DSP_SETFMT:
		result = get_user(val, (int *)arg);
		PrDebug("SNDCTL_DSP_SETFMTS(0x%x)\n", val);
		if (result == 0) {
			if (val != AFMT_QUERY) {
				if (!(s_PcmParams.formats_caps & val)) {	/* unsupport ? */
					val = AFMT_S16_LE;		/* set default format */
				}

				if (file->f_mode & FMODE_READ) {
					s_InputStream.format = val;
					/* set data bit length and padding (PMx_SI) */
					s_PcmParams.cyc.sib = (val == AFMT_S32_LE) ? 0x1f : 0x0f;
#ifdef MP201_PADDING_ON
					s_PcmParams.cyc.rx_pd =
						(val == AFMT_S32_LE) ? PCM_PADDING_OFF : PCM_PADDING_ON;
#endif
				}
				if (file->f_mode & FMODE_WRITE) {
					s_OutputStream.format = val;
					/* set data bit length and padding (PMx_SO) */
					s_PcmParams.cyc.sob = (val == AFMT_S32_LE) ? 0x1f : 0x0f;
#ifdef MP201_PADDING_ON
					s_PcmParams.cyc.tx_pd =
						(val == AFMT_S32_LE) ? PCM_PADDING_OFF : PCM_PADDING_ON;
#endif
				}

				/* set PCM0 parameter */
				result = audio_pcm_set_ctrl(&s_AudioState);
				if (result != 0) {
					PrErr("audio_pcm_set_ctrl() fail. result=%d\n", result);
				}

				/* set DMAC Tx bit width */
				audio_dsp_set_ctrl(val, (audio_state_t *)file->private_data);
			} else {
				if (file->f_mode & FMODE_READ) {
					val = s_InputStream.format;
				} else {
					val = s_OutputStream.format;
				}
			}
			result = put_user(val, (int *)arg);
		}
		break;

	default:
		result = MixerIoctl(inode, file, cmd, arg);
		break;
	}

	return result;
}

static int DspOpen(struct inode *inode, struct file *file)
{
	return mp200_audio_attach(inode, file, &s_AudioState);
}

/*
 * Missing fields of this structure will be patched with calling
 * to mp200_audio_attach().
 */
static struct file_operations s_DspFops = {
	.open	= DspOpen,
	.owner	= THIS_MODULE
};

static int InitPlatform(void)
{
	PrDebug("\n");

	/* get PowerIC LSI version */
	pwc_read(PWC_LSIVER, &s_PwrIcVer);

#ifdef AUDIO_INITIALIZED_BY_DRIVER
	DoMute(AUDIO_EXT_MUTE, AUDIO_EXT_MUTE);
	SetOutputDevice(AUDIO_OUTPUT_HP);

	/* set output mode to Power IC GPIO 11 & 12 pin */
	(void)gpiocore_pwc_write(GPIOCORE_PWC_IOSEL2, 0x18, 0x18);

	/* reset CS42L51 */
	ResetHold();
#endif /* AUDIO_INITIALIZED_BY_DRIVER */
	return 0;
}

static int ResetHold(void)
{
	/* select GPO function. (PM1/GPO pin) */
	outl(0x01, SMU_GIOCTL_PM1);

	/* CS42L51 RESET pin low */
	outl(0x00200000, SMU_SMU_GPO);	/* GPO5 pin low */
	return 0;
}

static int ResetRelease(void)
{
	/* CS42L51 RESET pin high */
	outl(0x00200020, SMU_SMU_GPO);	/* GPO5 pin high */

	/* set software mode */
	cs4251_set_powerdown(On);
	return 0;
}

#define ICX1087_PROCFS_SUPPORT
#ifdef ICX1087_PROCFS_SUPPORT

static struct proc_dir_entry *s_ProcDir = NULL;
static const char	s_ProcDirName[] = "icx1087audio";

#define RO	(S_IRUSR | S_IRGRP)	/* read only */
#define WO	(S_IWUSR | S_IWGRP)	/* write only */
#define RW	(RO|WO)				/* read/write */

typedef struct icx1087_proc_entry {
	const char		*name;		/* register name */
	const mode_t	rw;			/* access type (RO, WO or RW) */
	const int		id;			/* ID for inner process */
	const char		*desc;		/* description (optional) */
	unsigned 		low_ino;	/* i-node of /proc entry */
} icx1087_proc_entry_t;

#define NUM_PROC_ENTRY	(sizeof(s_ProcEntry) / sizeof(s_ProcEntry[0]))

/* Proc Entry ID's */
enum {
	PE_ID_MAX_REG =	56,
	PE_ID_RESET = 100,
	PE_ID_PLAYSRC,
	PE_ID_OUTPUTDEV,
	PE_ID_MUTE,
	PE_ID_VOLUME,
	PE_ID_BEEP,
	PE_ID_RESTART,
};

static icx1087_proc_entry_t s_ProcEntry[] = {
  /* map to CS42L51 registers */
	{"reg00h",		RW,	 0, "Unofficial Register"},
//	{"reg01h",		RO,  1, "Chip ID & Revision (d8h)(Ignore on SPI mode)"},
	{"reg02h",		RW,  2, "Power Control-1 (00h)"},
	{"reg03h",		RW,  3, "MIC Power & Speed Control (aeh)"},
	{"reg04h",		RW,  4, "Interface Control (00h)"},
	{"reg05h",		RW,  5, "MIC Control (00h)"},
	{"reg06h",		RW,  6, "ADC Control (a0h)"},
	{"reg07h",		RW,  7, "ADC Input Select, Invert & Mute (00h)"},
	{"reg08h",		RW,  8, "DAC Output Control (60h)"},
	{"reg09h",		RW,  9, "DAC Control (06h)"},
	{"reg0ah",		RW, 10, "ALC-A & PGA-A Control (00h)"},
	{"reg0bh",		RW, 11, "ALC-B & PGA-B Control (00h)"},
	{"reg0ch",		RW, 12, "ADC-A Attenuator (00h)"},
	{"reg0dh",		RW, 13, "ADC-B Attenuator (00h)"},
	{"reg0eh",		RW, 14, "ADC-A Mixer Volume Control (80h)"},
	{"reg0fh",		RW, 15, "ADC-B Mixer Volume Control (80h)"},
	{"reg10h",		RW, 16, "PCM-A Mixer Volume Control (80h)"},
	{"reg11h",		RW, 17, "PCM-B Mixer Volume Control (80h)"},
	{"reg12h",		RW, 18, "Beep Frequency & Timing Configuration (00h)"},
	{"reg13h",		RW, 19, "Beep Off Time & Volume (00h)"},
	{"reg14h",		RW, 20, "Beep & Tone Configuration (00h)"},
	{"reg15h",		RW, 21, "Tone Control (88h)"},
	{"reg16h",		RW, 22, "AOUT-A Volume Control (00h)"},
	{"reg17h",		RW, 23, "AOUT-B Volume Control (00h)"},
	{"reg18h",		RW, 24, "ADC & PCM Channel Mixer (00h)"},
	{"reg19h",		RW, 25, "Limiter Threshold SZC Disable (00h)"},
	{"reg1ah",		RW, 26, "Limiter Release Rate (40h)"},
	{"reg1bh",		RW, 27, "Limiter Attack Rate (00h)"},
	{"reg1ch",		RW, 28, "ALC Enable & Attack Rate (00h)"},
	{"reg1dh",		RW, 29, "ALC Release Rate (00h)"},
	{"reg1eh",		RW, 30, "ALC Threshold (00h)"},
	{"reg1fh",		RW, 31, "Noise Gate Configuration & Misc (00h)"},
//	{"reg20h",		RO, 32, "Status  (00h)(Ignore on SPI mode)"},
	{"reg21h",		RW, 33, "Charge Pump Frequency (50h)"},
	{"reg2bh",		RW,	43, "Unofficial Register"},
	{"reg31h",		RW,	49, "Unofficial Register"},
	{"reg33h",		RW,	51, "Unofficial Register"},
	{"reg35h",		RW,	53, "Unofficial Register"},
	{"reg37h",		RW,	55, "Unofficial Register"},
	{"reg38h",		RW,	56, "Unofficial Register"},
  /* utility operations */
	{"reset",		RW, PE_ID_RESET, "on or off"},
	{"playsrc",		RW, PE_ID_PLAYSRC, "pcm, line, tuner or mic"},
	{"outputdev",	RW, PE_ID_OUTPUTDEV, "hp or cradle"},
	{"mute",		RW, PE_ID_MUTE, "on or off"},
	{"vol",			RW, PE_ID_VOLUME, "l-vol,r-vol"},
	{"beep",		RW, PE_ID_BEEP, "on or off"},
	{"restart",		RW, PE_ID_RESTART, "on or off"},
};

static int	GetEntryIndex(unsigned i_ino)
{
	int	i;

	for (i = 0; i < (int)NUM_PROC_ENTRY; i++) {
		if (s_ProcEntry[i].low_ino == i_ino) {
			return i;
		}
	}
	return -1;
}

static ssize_t
ProcRead(struct file *file, char *userbuf, size_t nbytes, loff_t *ppos)
{
	char	workbuf[100], convbuf[16];
	uint8_t	data;
	int		count, index;

	if (*ppos > 0) { /* Assume reading completed in previous read */
		return 0; /* EOF */
	}

	/* search same i-node entry */
	index = GetEntryIndex(file->f_dentry->d_inode->i_ino);
	if (index < 0) {
		return -EINVAL;		/* entry not found */
	}
	if (s_ProcEntry[index].rw == WO) {
		return -EACCES;
	}

	if (s_ProcEntry[index].id <= PE_ID_MAX_REG) {
		cs4251_read_reg(s_ProcEntry[index].id, &data);
		count = sprintf(workbuf, "%s: 0x%02x (%s)\n",
				s_ProcEntry[index].name, data, s_ProcEntry[index].desc);
	} else {
		char	*str;

		switch (s_ProcEntry[index].id) {
		case PE_ID_RESET:
			/* check RESET (GPO5) pin status */
			str = (inl(SMU_SMU_GPO) & 0x20) ? "off" : "on";
			break;
		case PE_ID_PLAYSRC:
			switch (s_MixerSettings.play_src) {
			case AUDIO_PLAYBACK_PCM:	str = "pcm";	break;
			case AUDIO_PLAYBACK_LINE:	str = "line";	break;
			case AUDIO_PLAYBACK_TUNER:	str = "tuner";	break;
			case AUDIO_PLAYBACK_MIC:	str = "mic";	break;
			default:					str = "unknown";break;
			}
			break;
		case PE_ID_OUTPUTDEV:
			str = (s_MixerSettings.out_dev == AUDIO_OUTPUT_HP) ? "hp": "cradle";
			break;
		case PE_ID_MUTE:
			str = (s_MixerSettings.mute == AUDIO_ALL_MUTE_ON) ? "on" : "off";
			break;
		case PE_ID_VOLUME:
			sprintf(convbuf, "%d,%d", OSS_GET_L_VOL(s_MixerSettings.aout_vol),
					OSS_GET_R_VOL(s_MixerSettings.aout_vol));
			str = convbuf;
			break;
		case PE_ID_BEEP:
			cs4251_read_reg(0x14, &data);
			str = (data & 0x40) ? "on" : "off";
			break;
		case PE_ID_RESTART:
			str = "off";
			break;
		default:
			PrWarn("driver bug. invalid size\n");
			return -EIO;
		}

		count = sprintf(workbuf, "%s: %s (%s)\n",
				s_ProcEntry[index].name, str, s_ProcEntry[index].desc);
	}

	*ppos += count;
	if (count > nbytes) { /* Assume output can be read at one time */
		return -EINVAL;
	}

	if (copy_to_user(userbuf, workbuf, count)) {
		return -EFAULT;
	}
	return count;
}

static ssize_t
ProcWrite(struct file *file, const char *userbuf, size_t nbytes, loff_t *ppos)
{
	char	workbuf[32];
	uint8_t	data;
	int		index;

	if (nbytes >= sizeof(workbuf)) {
		return -EINVAL;
	}
	if (copy_from_user(workbuf, userbuf, nbytes)) {
		return -EFAULT;
	}
	workbuf[nbytes] = '\0';

	/* search same i-node entry */
	index = GetEntryIndex(file->f_dentry->d_inode->i_ino);
PrDebug("nbytes=%dbytes, buf=%s, index=0x%02x, name=%s\n", nbytes, workbuf, index, (index < 0) ? "Unknown": s_ProcEntry[index].name);
	if (index < 0) {
		return -EINVAL;		/* entry not found */
	}
	if (s_ProcEntry[index].rw == RO) {
		return -EACCES;
	}

	if (s_ProcEntry[index].id <= PE_ID_MAX_REG) {
		data = (uint8_t)simple_strtoul(workbuf, NULL, 0);
PrDebug("%s <- 0x%02x\n", s_ProcEntry[index].name, data);
		cs4251_write_reg(s_ProcEntry[index].id, data);
	} else {
		switch (s_ProcEntry[index].id) {
		case PE_ID_RESET:
			if (strcmp(workbuf, "on") == 0) {
				ResetHold();
			} else if (strcmp(workbuf, "off") == 0) {
				ResetRelease();
			} else {
				return -EINVAL;
			}
			break;
		case PE_ID_PLAYSRC:
			if (strcmp(workbuf, "pcm") == 0) {
				SetPlaybackSrc(AUDIO_PLAYBACK_PCM);
			} else if (strcmp(workbuf, "line") == 0) {
				SetPlaybackSrc(AUDIO_PLAYBACK_LINE);
			} else if (strcmp(workbuf, "tuner") == 0) {
				SetPlaybackSrc(AUDIO_PLAYBACK_TUNER);
			} else if (strcmp(workbuf, "mic") == 0) {
				SetPlaybackSrc(AUDIO_PLAYBACK_MIC);
			} else {
				return -EINVAL;
			}
			break;
		case PE_ID_OUTPUTDEV:
			if (strcmp(workbuf, "hp") == 0) {
				SetOutputDevice(AUDIO_OUTPUT_HP);
			} else if (strcmp(workbuf, "cradle") == 0) {
				SetOutputDevice(AUDIO_OUTPUT_LINE);
			} else {
				return -EINVAL;
			}
			break;
		case PE_ID_MUTE:
			if (strcmp(workbuf, "on") == 0) {
				DoMute(AUDIO_ALL_MUTE_ON, AUDIO_ALL_MUTE_ON);
			} else if (strcmp(workbuf, "off") == 0) {
				DoMute(AUDIO_ALL_MUTE_ON, AUDIO_MUTE_OFF);
			} else {
				return -EINVAL;
			}
			break;
		case PE_ID_VOLUME: {
			unsigned	lch, rch;
			char		*l_end, *r_end;

			/* assume "lch,rch" */
			lch = simple_strtoul(workbuf, &l_end, 10);
			if (l_end == workbuf || lch > 100) {
				return -EINVAL;
			}
			rch = simple_strtoul(l_end+1, &r_end, 10);
			if (r_end == l_end+1 || rch > 100) {
				return -EINVAL;
			}
			SetAoutVol(lch, rch);
			break;
		}
		case PE_ID_BEEP:
			if (strcmp(workbuf, "on") == 0) {
				cs4251_set_cont_beep(On);
			} else if (strcmp(workbuf, "off") == 0) {
				cs4251_set_cont_beep(Off);
			} else {
				return -EINVAL;
			}
			break;
		case PE_ID_RESTART:
			if (strcmp(workbuf, "on") == 0) {
				PowerUpCodec(On);
			} else if (strcmp(workbuf, "off") == 0) {
				;
			} else {
				return -EINVAL;
			}
			break;
		default:
			PrWarn("driver bug. invalid ID\n");
			return -EIO;
		}
	}

	return nbytes;
}

static struct file_operations s_ProcFops = {
	.read	= ProcRead,
	.write	= ProcWrite
};

static int	ProcInit(void)
{
	struct proc_dir_entry	*entry;
	int	i;

	s_ProcDir = proc_mkdir(s_ProcDirName, NULL);
	if (s_ProcDir == NULL) {
		PrErr("can't create /proc/%s\n", s_ProcDirName);
		return -ENOMEM;
	}

	for (i = 0; i < (int)NUM_PROC_ENTRY; i++) {
		entry = create_proc_entry(s_ProcEntry[i].name,
						s_ProcEntry[i].rw, s_ProcDir);
		if (entry == NULL) {
			PrErr("can't create /proc/%s/%s\n",
					s_ProcDirName, s_ProcEntry[i].name);
			break;
		}
		s_ProcEntry[i].low_ino = entry->low_ino;
		entry->proc_fops = &s_ProcFops;
	}

	/* error occurred ? */
	if (i < NUM_PROC_ENTRY) {
		while (i) {
			remove_proc_entry(s_ProcEntry[--i].name, s_ProcDir);
		}
		remove_proc_entry(s_ProcDirName, NULL);
		return -ENOMEM;
	}

	return 0;
}

static void ProcCleanup(void)
{
	int i;

	if (s_ProcDir != NULL) {
		for (i = 0; i < (int)NUM_PROC_ENTRY; i++) {
			remove_proc_entry(s_ProcEntry[i].name, s_ProcDir);
		}
		remove_proc_entry(s_ProcDirName, NULL);
	}
}

#endif /* ICX1087_PROCFS_SUPPORT */

static int s_DspDev, s_MixerDev;

static int DriverProbe(struct device *dev)
{
	int	result;

	PrDebug("\n");

	InitPlatform();

	/* register devices */
	s_DspDev = register_sound_dsp(&s_DspFops, -1);
	if (s_DspDev < 0) {
		return s_DspDev;
	}

	s_MixerDev = register_sound_mixer(&s_MixerFops, -1);
	if (s_MixerDev < 0) {
		unregister_sound_dsp(s_DspDev);
		return s_MixerDev;
	}

	OpenPcm0Clock();	/* power up and unreset PCM0 */
	PowerUpCodec(Off);

#ifdef ICX1087_PROCFS_SUPPORT
	ProcInit();
#endif /* ICX1087_PROCFS_SUPPORT */

	/* set PCM0 parameter */
	result = audio_pcm_set_ctrl(&s_AudioState);
	if (result != 0) {
		PrErr("audio_pcm_set_ctrl() fail. result=%d\n", result);
		DriverRemove(dev);
	}

	/* initialize beep driver */
	result = icx1087_beep_init();
	if (result != 0) {
		DriverRemove(dev);
		return result;
	}

	return result;
}

static int DriverRemove(struct device *dev)
{
	PrDebug("\n");

	PowerDownCodec();
	ClosePcm0Clock();

#ifdef ICX1087_PROCFS_SUPPORT
	ProcCleanup();
#endif /* ICX1087_PROCFS_SUPPORT */

	/* finalize beep driver */
	icx1087_beep_exit();

	/* unregister devices */
	unregister_sound_mixer(s_MixerDev);
	unregister_sound_dsp(s_DspDev);
	return 0;
}

#if 0
/* calling from other driver */
int icx1087_set_pga_vol(int a_ch, int b_ch)
{
	return cs4251_set_pga_vol(a_ch, b_ch);
}
EXPORT_SYMBOL(icx1087_set_pga_vol);
#endif

int __init icx1087_cs4251_init(void)
{
	PrInfo("Starting CS42L51 codec.\n");
	return driver_register(&s_DeviceDriver);
}

void __exit icx1087_cs4251_exit(void)
{
	PrDebug("\n");
	driver_unregister(&s_DeviceDriver);
}

module_init(icx1087_cs4251_init);
module_exit(icx1087_cs4251_exit);

MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Sony ICX1087 Audio & Beep device driver");

/* end of icx1087-sound.c */
