/*
 *  File Name       : linux/sound/oss/mp200-ak4642.c
 *  Function        : Glue audio driver for the MP200 & AK4642 CODEC
 *  Release Version : Ver 1.00
 *  Release Date    : 2006/03/29
 *
 *  Copyright (C) NEC Electronics Corporation 2006
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/sound.h>
#include <linux/soundcard.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <asm/uaccess.h>
#include <asm/arch/gpiocore.h>
#include <asm/arch/pmu.h>

#ifdef CONFIG_MP201
#include <asm/arch/smu.h>
#include <asm/io.h>
#endif

#include "mp200-ak4642.h"
#include "mp200-audio.h"

#define MP200_AUDIO_OPEN_SUSPEND 1

/* Local prototypes */
static int i2c_ak4642_attach_adapter(struct i2c_adapter *adap);
static int i2c_ak4642_detach_client(struct i2c_client *client);
static int i2c_ak4642_command(struct i2c_client *client, unsigned int cmd,
			      void *arg);
static void mp200_ak4642_init(audio_stream_t * os, audio_stream_t * is);
static void mp200_ak4642_shutdown(audio_stream_t * os, audio_stream_t * is);
static int mp200_ak4642_ioctl(struct inode *inode, struct file *file,
			      unsigned int cmd, unsigned long arg);
static int ak4642_probe(struct device *dev);
static int ak4642_remove(struct device *dev);
static int ak4642_suspend(struct device *dev, u32 state, u32 level);
static int ak4642_resume(struct device *dev, u32 level);

static int audio_dev_id, mixer_dev_id;
static pcm_regs_save_t pcm0_regs_save;
static int ak4642_input_select = SOUND_MASK_MIC;
static unsigned char i2c_ak4642_reg[AK4642REG_MAX];

static struct i2c_driver i2c_ak4642_driver = {
	.name = "i2c for ak4642",
	.id = I2C_DRIVERID_EXP0,
	.flags = I2C_DF_NOTIFY,
	.attach_adapter = i2c_ak4642_attach_adapter,
	.detach_client = i2c_ak4642_detach_client,
	.command = i2c_ak4642_command,
};

static struct i2c_client i2c_ak4642_client = {
	.name = "(unset)",
	.id = 1,
	.flags = 0,
	.addr = -1,
	.driver = &i2c_ak4642_driver,
};

static struct device_driver mp200_ak4642_driver = {
	.name = "pcm",
	.bus = &platform_bus_type,
	.probe = ak4642_probe,
	.remove = ak4642_remove,
	.suspend = ak4642_suspend,
	.resume = ak4642_resume,
};

static struct mixer_volume ak4642_mixer_volume[] = {
	/* all values 0 -> 100 in bytes */
	{SOUND_MASK_MIC, 0x3C3C},	/* right:-0.375dB left:-0.375dB */
	{SOUND_MASK_LINE, 0x3C3C},	/* right:-0.375dB left:-0.375dB */
	{SOUND_MASK_PCM, 0x5050},	/* right:-13.5dB left:-13.5dB */
	{-1, 0}
};

static pcm_regs_t pcm0_regs = {
	.func_sel = &PMx_FUNC_SEL(PCM0_ADDR),
	.cycle = &PMx_CYCLE(PCM0_ADDR),
	.txrx_en = &PMx_TXRX_EN(PCM0_ADDR),
	.txrx_dis = &PMx_TXRX_DIS(PCM0_ADDR),
	.txrx_mon = &PMx_TXRX_MON(PCM0_ADDR),
	.status = &PMx_STATUS(PCM0_ADDR),
	.enset = &PMx_ENSET(PCM0_ADDR),
	.enclr = &PMx_ENCLR(PCM0_ADDR),
	.enmon = &PMx_ENMON(PCM0_ADDR),
	.raw = &PMx_RAW(PCM0_ADDR),
	.clear = &PMx_CLEAR(PCM0_ADDR),
#ifdef MP201_SINGLE_SPECIAL
	.txq = &PMx_TXQ(PCM0_ADDR),
	.fifo_p = &PMx_FIFO_P(PCM0_ADDR),
#endif
	.cycle2 = &PMx_CYCLE2(PCM0_ADDR),
};

static pcm_ctrl_t pcm_ak4642_sett = {
	.func = {
		 .mode_sel = PCM_MODE_3,
		 .m_s = PCM_SLAVE_MODE,
		 .tx_tim = PCM_TX_04_WORD,
		 },
	.cyc = {
#ifdef MP201_CODEC_32
		.cyc_val = 0x1f,
#else
		.cyc_val = 0x0f,
#endif
		.sib = 0x0f,
#ifdef MP201_PADDING_ON
		.rx_pd = PCM_PADDING_ON,
#else
		.rx_pd = PCM_PADDING_OFF,
#endif
		.sob = 0x0f,
#ifdef MP201_PADDING_ON
		.tx_pd = PCM_PADDING_ON,
#else
		.tx_pd = PCM_PADDING_OFF,
#endif
		},
	.cyc2 = {
#ifdef MP201_CODEC_32
		 .cyc_val2 = 0x1f,
#else
		 .cyc_val2 = 0x0f,
#endif
		 .sib2 = 0x0f,
		 .sob2 = 0x0f,
		 },
#ifdef MP201_CODEC_32
	.formats_caps = AFMT_S16_LE | AFMT_S32_LE,
#else
	.formats_caps = AFMT_S16_LE,
#endif
};

static pcm_ctrls_t pcm_ak4642_ctrls = {
	.sett = &pcm_ak4642_sett,
	.regs = &pcm0_regs,
};

static audio_stream_t output_stream_ak4642 = {
	.id = "AK4642 out",
	.buffers = NULL,
#ifdef CONFIG_MP201
	.dma_channel = MP200_DMAC_M2P_LCH4,
#else
	.dma_channel = MP200_DMAC_M2P_LCH9,
#endif
	.pcm_lock = SPIN_LOCK_UNLOCKED,
	.pcm_intr = 0,
	.pcm_status = 0,
	.pcm_regs = &pcm0_regs,
	.irq_bit = PCM_INT_TX_FRE | PCM_INT_TX_URE | PCM_INT_TX_ORE,
	.cl_bit = PCM_INT_TX_FRE | PCM_INT_TX_URE,
	.en_bit = PCM_TX_EN,
	.phys_xq = (MP200_PCM0_BASE + 0x24),
#ifdef CONFIG_MP201
	.int_dma = MP200_DMAC_INT_LENG_EN,
#else
	.int_dma = (MP200_DMAC_INT_BLOCK_EN | MP200_DMAC_INT_LENG_EN),
#endif
	.format = AFMT_S16_LE,
	.stereo = 1,
	.type = AUDIO_OUT,
};

static audio_stream_t input_stream_ak4642 = {
	.id = "AK4642 in",
	.buffers = NULL,
#ifdef CONFIG_MP201
	.dma_channel = MP200_DMAC_P2M_LCH4,
#else
	.dma_channel = MP200_DMAC_P2M_LCH9,
#endif
	.pcm_lock = SPIN_LOCK_UNLOCKED,
	.pcm_intr = 0,
	.pcm_status = 0,
	.pcm_regs = &pcm0_regs,
	.irq_bit = PCM_INT_RX_FRE | PCM_INT_RX_URE | PCM_INT_RX_ORE,
	.cl_bit = PCM_INT_RX_FRE,
	.en_bit = PCM_RX_EN,
	.phys_xq = (MP200_PCM0_BASE + 0x28),
#ifdef CONFIG_MP201
	.int_dma = MP200_DMAC_INT_LENG_EN,
#else
	.int_dma = (MP200_DMAC_INT_BLOCK_EN | MP200_DMAC_INT_LENG_EN),
#endif
	.format = AFMT_S16_LE,
	.stereo = 1,
	.type = AUDIO_IN,
};

static audio_state_t ak4642_state = {
	.output_stream = &output_stream_ak4642,
	.input_stream = &input_stream_ak4642,
	.rd_ref = 0,
	.wr_ref = 0,
	.pcm_irq = INT_PCM0,
	.hw_init = mp200_ak4642_init,
	.hw_shutdown = mp200_ak4642_shutdown,
	.client_ioctl = mp200_ak4642_ioctl,
	.ctrls = &pcm_ak4642_ctrls,
	.sem = __MUTEX_INITIALIZER(ak4642_state.sem),
#ifdef MP201_DSP_CPU
	.int_mode = PCM_INT_CPU_MODE,
#endif
};

/*
 * i2c functions
 */
static int i2c_ak4642_attach_adapter(struct i2c_adapter *adap)
{
	i2c_ak4642_client.adapter = adap;
	i2c_ak4642_client.addr = AK4642_DEVICE;
	i2c_attach_client(&i2c_ak4642_client);
	return 0;
}

static int i2c_ak4642_detach_client(struct i2c_client *client)
{
	i2c_detach_client(client);
	return 0;
}

static int i2c_ak4642_command(struct i2c_client *client, unsigned int cmd,
			      void *arg)
{
	return -1;
}

static int i2c_ak4642_cleanup(void)
{
	return i2c_del_driver(&i2c_ak4642_driver);
}

static int i2c_ak4642_inserted(void)
{
	return i2c_add_driver(&i2c_ak4642_driver);
}

static int i2c_ak4642_init(void)
{
	int res = 0;
	unsigned char reg = 0;

	res = i2c_ak4642_inserted();
	if (res == 0) {
		res = i2c_master_send(&i2c_ak4642_client, &reg, 1);
		if (res > 0) {
			res =
			    i2c_master_recv(&i2c_ak4642_client, i2c_ak4642_reg,
					    AK4642REG_MAX);
			if (res > 0) {
				res = 0;
			}
		}
		if (res != 0) {
			i2c_ak4642_cleanup();
		}
	}

	return res;
}

static int i2c_ak4642_write(unsigned char reg, unsigned char data,
			    unsigned char mask)
{
	int res = 0;
	unsigned char buf[2];

	if ((AK4642REG_MAX <= reg) || ((data & ~(mask)) != 0)) {
		return -EINVAL;
	}

	data |= (i2c_ak4642_reg[reg] & ~(mask));

	i2c_ak4642_reg[reg] = data;

	buf[0] = reg;
	buf[1] = data;

	res = i2c_master_send(&i2c_ak4642_client, buf, 2);
	if (res > 0) {
		res = 0;
	}

	return res;
}

/*
 * Audio mixer interface
 */
static int mixer_ak4642_volume_read(unsigned char mixer)
{
	int i;

	for (i = 0; ak4642_mixer_volume[i].mixer != -1; i++) {
		if (ak4642_mixer_volume[i].mixer == mixer) {
			break;
		}
	}

	return ak4642_mixer_volume[i].volume;
}

static int mixer_ak4642_volume_write(unsigned char mixer, int volume)
{
	unsigned char left;
	unsigned char right;
	int res = 0;
	int i;

	if (volume < 0) {
		volume = 0;
	}

	left = (volume & 0xff);
	right = ((volume >> 8) & 0xff);

	if (left > 100) {
		left = 100;
	}
	if (right > 100) {
		right = 100;
	}

	for (i = 0; ak4642_mixer_volume[i].mixer != -1; i++) {
		if (ak4642_mixer_volume[i].mixer == mixer) {
			ak4642_mixer_volume[i].volume = (right << 8) | left;
			break;
		}
	}

	if (input_stream_ak4642.stereo == 0) {
		right = left;
	}

	switch (mixer) {
	case SOUND_MASK_MIC:
	case SOUND_MASK_LINE:
		if (ak4642_input_select != mixer) {
			break;
		}
		res =
		    i2c_ak4642_write(AK4642REG_09H_LCH_INPUT_VOLUME_CONTROL,
				     (AK4642_INPUT_VOL_MAX * left) / 100,
				     AK4642_INPUT_VOL_MAX);
		if (res != 0) {
			break;
		}
		res =
		    i2c_ak4642_write(AK4642REG_0CH_RCH_INPUT_VOLUME_CONTROL,
				     (AK4642_INPUT_VOL_MAX * right) / 100,
				     AK4642_INPUT_VOL_MAX);
		if (res != 0) {
			break;
		}
		res =
		    i2c_ak4642_write(AK4642REG_07H_ALC_MODE_CONTROL_1, 0x00,
				     AK4642BIT_07H_20H_ALC);
		if (res != 0) {
			break;
		}
		res =
		    i2c_ak4642_write(AK4642REG_08H_ALC_MODE_CONTROL_2,
				     (AK4642_INPUT_VOL_MAX * left) / 100,
				     AK4642_INPUT_VOL_MAX);
		if (res != 0) {
			break;
		}
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(AK4642_WAIT_ALC_LOCK);
		set_current_state(TASK_RUNNING);
		res =
		    i2c_ak4642_write(AK4642REG_07H_ALC_MODE_CONTROL_1,
				     AK4642BIT_07H_20H_ALC,
				     AK4642BIT_07H_20H_ALC);
		break;

	case SOUND_MASK_PCM:
		res =
		    i2c_ak4642_write(AK4642REG_0AH_LCH_DIGITAL_VOLUME_CONTROL,
				     (AK4642_OUTPUT_VOL_MAX * (100 - left)) /
				     100, AK4642_OUTPUT_VOL_MAX);
		if (res != 0) {
			break;
		}
		res =
		    i2c_ak4642_write(AK4642REG_0DH_RCH_DIGITAL_VOLUME_CONTROL,
				     (AK4642_OUTPUT_VOL_MAX * (100 - right)) /
				     100, AK4642_OUTPUT_VOL_MAX);
		break;

	default:
		return -EINVAL;
	}

	return res;
}

static int mixer_ak4642_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int mixer_ak4642_release(struct inode *inode, struct file *file)
{
	return 0;
}

static int mixer_ak4642_ioctl(struct inode *inode, struct file *file,
			      unsigned int cmd, unsigned long arg)
{
	int res = 0;
	int val;
	int volume;
	unsigned char reg;
	unsigned char data;

	switch (cmd) {
	case PCM_CODEC_SET:
		reg = (unsigned char)(arg >> 8);
		data = (unsigned char)arg;
		res = i2c_ak4642_write(reg, data, 0xff);
		break;

	case SOUND_MIXER_READ_DEVMASK:
		res =
		    put_user(SOUND_MASK_PCM | SOUND_MASK_LINE | SOUND_MASK_MIC,
			     (int *)arg);
		break;

	case SOUND_MIXER_READ_RECMASK:
		res = put_user(SOUND_MASK_LINE | SOUND_MASK_MIC, (int *)arg);
		break;

	case SOUND_MIXER_READ_STEREODEVS:
		res =
		    put_user(SOUND_MASK_PCM | SOUND_MASK_LINE | SOUND_MASK_MIC,
			     (int *)arg);
		break;

	case SOUND_MIXER_READ_RECSRC:
		res = put_user(ak4642_input_select, (int *)arg);
		break;

	case SOUND_MIXER_WRITE_RECSRC:
		res = get_user(val, (int *)arg);
		if (res != 0) {
			break;
		}
		if (val == 0) {
			val = SOUND_MASK_MIC;
		}
		val &= (SOUND_MASK_LINE | SOUND_MASK_MIC);
		/* ak4642 choose one either */
		val &= ~(ak4642_input_select);
		if (val != 0) {
			ak4642_input_select = val;
			if (input_stream_ak4642.active == 0) {
				if ((ak4642_input_select & SOUND_MASK_LINE) !=
				    0) {
					res =
					    i2c_ak4642_write
					    (AK4642REG_10H_POWER_MANAGEMENT_3,
					     AK4642BIT_10H_04H_INR |
					     AK4642BIT_10H_02H_INL,
					     AK4642BIT_10H_06H_IN_MASK);
					if (res != 0) {
						break;
					}
					res =
					    i2c_ak4642_write
					    (AK4642REG_02H_SIGNAL_SELECT_1,
					     0x00, AK4642BIT_02H_04H_PMMP);
					if (res != 0) {
						break;
					}
				} else {
					res =
					    i2c_ak4642_write
					    (AK4642REG_10H_POWER_MANAGEMENT_3,
					     0x00, AK4642BIT_10H_06H_IN_MASK);
					if (res != 0) {
						break;
					}
					res =
					    i2c_ak4642_write
					    (AK4642REG_02H_SIGNAL_SELECT_1,
					     AK4642BIT_02H_04H_PMMP,
					     AK4642BIT_02H_04H_PMMP);
					if (res != 0) {
						break;
					}
				}
				volume =
				    mixer_ak4642_volume_read
				    (ak4642_input_select);
				res =
				    mixer_ak4642_volume_write
				    (ak4642_input_select, volume);
				if (res != 0) {
					break;
				}
			} else {
				down(&ak4642_state.sem);
				ak4642_state.rd_ref &= ~(AUDIO_REF_PW);
				up(&ak4642_state.sem);
			}
		}

		/* AK4642 is line & mic */
		res = put_user(ak4642_input_select, (int *)arg);
		break;

		/* Meaning of Volume Levels
		 * The OSS driver specifies volume levels using integer values
		 * from 0 to 100. The value 0 means minimum volume (off) and 100
		 * means maximum volume.
		 * Most mixers have anywhere from 3 to 8 bits of accuracy for
		 * controlling the volume at the hardware level. The OSS driver
		 * scales between the local (0-100) and hardware defined volume.
		 * Since this
		 * scaling is not exact, the application should be careful when using
		 * the volume returned by the ioctl calls. If the application writes the
		 * volume and then reads it back, the returned volume is usually
		 * slightly different (smaller) than the requested one. If the
		 * write-read sequence is repeated several times, the volume level
		 * slides to zero even if the application makes no changes itself. It is
		 * recommended, therefore, that the application reads the volume just during
		 * initialization and ignores the volume returned later
		 */
	case SOUND_MIXER_WRITE_MIC:
		res = get_user(val, (int *)arg);
		if (res != 0) {
			break;
		}
		res = mixer_ak4642_volume_write(SOUND_MASK_MIC, val);
		break;

	case SOUND_MIXER_WRITE_LINE:
		res = get_user(val, (int *)arg);
		if (res != 0) {
			break;
		}
		res = mixer_ak4642_volume_write(SOUND_MASK_LINE, val);
		break;

	case SOUND_MIXER_WRITE_PCM:
		res = get_user(val, (int *)arg);
		if (res != 0) {
			break;
		}
		res = mixer_ak4642_volume_write(SOUND_MASK_PCM, val);
		break;

	case SOUND_MIXER_READ_MIC:
		val = mixer_ak4642_volume_read(SOUND_MASK_MIC);
		res = put_user(val, (int *)arg);
		break;

	case SOUND_MIXER_READ_LINE:
		val = mixer_ak4642_volume_read(SOUND_MASK_LINE);
		res = put_user(val, (int *)arg);
		break;

	case SOUND_MIXER_READ_PCM:
		val = mixer_ak4642_volume_read(SOUND_MASK_PCM);
		res = put_user(val, (int *)arg);
		break;

	default:
		res = -EINVAL;
		break;
	}

	return res;
}

struct file_operations mp200_mixer_ak4642_fops = {
	.open = mixer_ak4642_open,
	.release = mixer_ak4642_release,
	.ioctl = mixer_ak4642_ioctl,
	.owner = THIS_MODULE
};

/*
 * Audio interface
 */
static int mp200_ak4642_open(struct inode *inode, struct file *file)
{
	return mp200_audio_attach(inode, file, &ak4642_state);
}

/*
 * Missing fields of this structure will be patched with calling
 * to mp200_audio_attach().
 */
struct file_operations mp200_ak4642_fops = {
	.open = mp200_ak4642_open,
	.owner = THIS_MODULE
};

static int mp200_ak4642_ioctl(struct inode *inode, struct file *file,
			      unsigned int cmd, unsigned long arg)
{
#ifdef MP201_DSP_CPU
	audio_state_t *state = file->private_data;
#endif
	audio_stream_t *os = &output_stream_ak4642;
	audio_stream_t *is = &input_stream_ak4642;
	int val;
	int res = 0;
	int volume;
	unsigned char reg;
	unsigned char data;
	unsigned char mask;

	switch (cmd) {
	case SNDCTL_DSP_SPEED:
		res = get_user(val, (int *)arg);
		if (res != 0) {
			break;
		}
		if (val < 8000) {
			val = 7350;
			data = AK4642BIT_05H_FS_7_35KHZ;
		} else if (val < 11025) {
			val = 8000;
			data = AK4642BIT_05H_FS_8KHZ;
		} else if (val < 12000) {
			val = 11025;
			data = AK4642BIT_05H_FS_11_025KHZ;
		} else if (val < 14700) {
			val = 12000;
			data = AK4642BIT_05H_FS_12KHZ;
		} else if (val < 16000) {
			val = 14700;
			data = AK4642BIT_05H_FS_14_7KHZ;
		} else if (val < 22050) {
			val = 16000;
			data = AK4642BIT_05H_FS_16KHZ;
		} else if (val < 24000) {
			val = 22050;
			data = AK4642BIT_05H_FS_22_05KHZ;
		} else if (val < 29400) {
			val = 24000;
			data = AK4642BIT_05H_FS_24KHZ;
		} else if (val < 32000) {
			val = 29400;
			data = AK4642BIT_05H_FS_29_4KHZ;
		} else if (val < 44100) {
			val = 32000;
			data = AK4642BIT_05H_FS_32KHZ;
		} else if (val < 48000) {
			val = 44100;
			data = AK4642BIT_05H_FS_44_1KHZ;
		} else {
			val = 48000;
			data = AK4642BIT_05H_FS_48KHZ;
		}

		reg = AK4642REG_05H_MODE_CONTROL_2;
		mask = AK4642BIT_05H_27H_FS_MASK;

		if ((i2c_ak4642_reg[reg] & mask) != data) {
			if ((is->active == 0) && (os->active == 0)) {
				res = i2c_ak4642_write(reg, data, mask);
				if (res != 0) {
					break;
				}
			} else {
				i2c_ak4642_reg[reg] &= ~(mask);
				i2c_ak4642_reg[reg] |= data;
				down(&ak4642_state.sem);
				ak4642_state.rd_ref &= ~(AUDIO_REF_PW);
				ak4642_state.wr_ref &= ~(AUDIO_REF_PW);
				up(&ak4642_state.sem);
			}
		}
		res = put_user(val, (int *)arg);
		break;

	case SNDCTL_DSP_STEREO:
		res = get_user(val, (int *)arg);
		if (res != 0) {
			break;
		}
		if ((val != 0) && (val != 1)) {
			val = -EINVAL;
		} else {
			if ((file->f_mode & FMODE_READ) != 0) {
				if ((is->mapped != 0) && (val == 0)) {
					val = -ENXIO;
				} else if (is->stereo != val) {
					is->stereo = val;
					if (is->active == 0) {
						volume =
						    mixer_ak4642_volume_read
						    (ak4642_input_select);
						res =
						    mixer_ak4642_volume_write
						    (ak4642_input_select,
						     volume);
						if (res != 0) {
							break;
						}
					} else {
						down(&ak4642_state.sem);
						ak4642_state.rd_ref &=
						    ~(AUDIO_REF_PW);
						up(&ak4642_state.sem);
					}
				}
			}
			if ((file->f_mode & FMODE_WRITE) != 0) {
				if ((os->mapped != 0) && (val == 0)) {
					val = -ENXIO;
				} else if (os->stereo != val) {
					os->stereo = val;
					if (os->active == 0) {
						volume =
						    mixer_ak4642_volume_read
						    (SOUND_MASK_PCM);
						res =
						    mixer_ak4642_volume_write
						    (SOUND_MASK_PCM, volume);
						if (res != 0) {
							break;
						}
					} else {
						down(&ak4642_state.sem);
						ak4642_state.wr_ref &=
						    ~(AUDIO_REF_PW);
						up(&ak4642_state.sem);
					}
				}
			}
		}
		res = put_user(val, (int *)arg);
		break;

	case SNDCTL_DSP_CHANNELS:
		res = get_user(val, (int *)arg);
		if (res != 0) {
			break;
		}
		if ((val != 1) && (val != 2)) {
			if ((file->f_mode & FMODE_READ) != 0) {
				val = is->stereo + 1;
			} else {
				val = os->stereo + 1;
			}
		} else {
			if ((file->f_mode & FMODE_READ) != 0) {
				if ((is->mapped != 0) && (val == 1)) {
					val = -ENXIO;
				} else if (is->stereo != (val - 1)) {
					is->stereo = val - 1;
					if (is->active == 0) {
						volume =
						    mixer_ak4642_volume_read
						    (ak4642_input_select);
						res =
						    mixer_ak4642_volume_write
						    (ak4642_input_select,
						     volume);
						if (res != 0) {
							break;
						}
					} else {
						down(&ak4642_state.sem);
						ak4642_state.rd_ref &=
						    ~(AUDIO_REF_PW);
						up(&ak4642_state.sem);
					}
				}
			}
			if ((file->f_mode & FMODE_WRITE) != 0) {
				if ((os->mapped != 0)
				    && (val == 1)) {
					val = -ENXIO;
				} else if (os->stereo != (val - 1)) {
					os->stereo = val - 1;
					if (os->active == 0) {
						volume =
						    mixer_ak4642_volume_read
						    (SOUND_MASK_PCM);
						res =
						    mixer_ak4642_volume_write
						    (SOUND_MASK_PCM, volume);
						if (res != 0) {
							break;
						}
					} else {
						down(&ak4642_state.sem);
						ak4642_state.wr_ref &=
						    ~(AUDIO_REF_PW);
						up(&ak4642_state.sem);
					}
				}
			}
		}
		res = put_user(val, (int *)arg);
		break;

	case SNDCTL_DSP_GETFMTS:
		res = put_user(pcm_ak4642_sett.formats_caps, (int *)arg);
		break;

	case SNDCTL_DSP_SETFMT:
		res = get_user(val, (int *)arg);
		if (res != 0) {
			break;
		}
		if (val != AFMT_QUERY) {
			if ((pcm_ak4642_sett.formats_caps & val) == 0) {
				val = AFMT_S16_LE;
			}
#ifdef MP201_CODEC_32
			if (file->f_mode & FMODE_READ) {
				is->format = val;
				pcm_ak4642_sett.cyc.sib =
				 (val == AFMT_S32_LE) ? 0x1f : 0x0f;
#ifdef MP201_PADDING_ON
				pcm_ak4642_sett.cyc.rx_pd =
				 (val == AFMT_S32_LE) ? PCM_PADDING_OFF : PCM_PADDING_ON;
#endif
			}
			if ((file->f_mode & FMODE_WRITE) != 0) {
				os->format = val;
				pcm_ak4642_sett.cyc.sob =
				 (val == AFMT_S32_LE) ? 0x1f : 0x0f;
#ifdef MP201_PADDING_ON
				pcm_ak4642_sett.cyc.tx_pd =
				 (val == AFMT_S32_LE) ? PCM_PADDING_OFF : PCM_PADDING_ON;
#endif
			}
			res = audio_pcm_set_ctrl(&ak4642_state);
			if (res != 0) {
				printk(KERN_INFO
				       "%s(): audio_pcm_set_ctrl() fail. result=%d\n",
				       __FUNCTION__, res);
			}
#else
			if ((file->f_mode & FMODE_READ) != 0) {
				is->format = val;
			}
			if ((file->f_mode & FMODE_WRITE) != 0) {
				os->format = val;
			}
#endif
#ifdef MP201_DSP_CPU
			res = audio_dsp_set_ctrl(val, state);
			if (res != 0) {
				printk(KERN_INFO
				       "%s(): audio_dsp_set_ctrl() fail."
				       " result=%d\n",
				       __FUNCTION__, res);
			}
#endif
		} else {
			if ((file->f_mode & FMODE_READ) != 0) {
				val = is->format;
			} else {
				val = os->format;
			}
		}
		res = put_user(val, (int *)arg);
		break;

	default:
		res = mixer_ak4642_ioctl(inode, file, cmd, arg);
		break;
	}

	return res;
}

static void mp200_ak4642_init(audio_stream_t * os, audio_stream_t * is)
{
	int res = 0;
	int volume;

	if (os != NULL) {
		if ((ak4642_state.wr_ref & AUDIO_REF_EN) == 0) {
			ak4642_state.wr_ref |= AUDIO_REF_EN;
		}
		if ((ak4642_state.wr_ref & AUDIO_REF_PW) == 0) {
			volume = mixer_ak4642_volume_read(SOUND_MASK_PCM);
			res = mixer_ak4642_volume_write(SOUND_MASK_PCM, volume);
			if (res != 0) {
				goto out;
			}
			res = i2c_ak4642_write(AK4642REG_05H_MODE_CONTROL_2,
					       i2c_ak4642_reg
					       [AK4642REG_05H_MODE_CONTROL_2],
					       AK4642BIT_05H_27H_FS_MASK);
			if (res != 0) {
				goto out;
			}
			ak4642_state.wr_ref |= AUDIO_REF_PW;
		}
	}
	if (is != NULL) {
		if ((ak4642_state.rd_ref & AUDIO_REF_EN) == 0) {
			ak4642_state.rd_ref |= AUDIO_REF_EN;
		}
		if ((ak4642_state.rd_ref & AUDIO_REF_PW) == 0) {
			if ((ak4642_input_select & SOUND_MASK_LINE) != 0) {
				res =
				    i2c_ak4642_write
				    (AK4642REG_10H_POWER_MANAGEMENT_3,
				     AK4642BIT_10H_04H_INR |
				     AK4642BIT_10H_02H_INL,
				     AK4642BIT_10H_06H_IN_MASK);
				if (res != 0) {
					goto out;
				}
				res =
				    i2c_ak4642_write
				    (AK4642REG_02H_SIGNAL_SELECT_1, 0x00,
				     AK4642BIT_02H_04H_PMMP);
				if (res != 0) {
					goto out;
				}
			} else {
				res =
				    i2c_ak4642_write
				    (AK4642REG_10H_POWER_MANAGEMENT_3, 0x00,
				     AK4642BIT_10H_06H_IN_MASK);
				if (res != 0) {
					goto out;
				}
				res =
				    i2c_ak4642_write
				    (AK4642REG_02H_SIGNAL_SELECT_1,
				     AK4642BIT_02H_04H_PMMP,
				     AK4642BIT_02H_04H_PMMP);
				if (res != 0) {
					goto out;
				}
			}
			volume = mixer_ak4642_volume_read(ak4642_input_select);
			res =
			    mixer_ak4642_volume_write(ak4642_input_select,
						      volume);
			if (res != 0) {
				goto out;
			}
			res =
			    i2c_ak4642_write(AK4642REG_05H_MODE_CONTROL_2,
					     i2c_ak4642_reg
					     [AK4642REG_05H_MODE_CONTROL_2],
					     AK4642BIT_05H_27H_FS_MASK);
			if (res != 0) {
				goto out;
			}
			ak4642_state.rd_ref |= AUDIO_REF_PW;
		}
	}

      out:
	if (res != 0) {
		ak4642_state.wr_ref |= AUDIO_REF_ERR;
		ak4642_state.rd_ref |= AUDIO_REF_ERR;
	}
}

static void mp200_ak4642_shutdown(audio_stream_t * os, audio_stream_t * is)
{
	/*
	 * Turn off codec after it is done.
	 * Can't do it immediately, since it may still have buffered data.
	 *
	 * Wait 20ms (arbitrary value) and then turn it off.
	 */
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(2);
	set_current_state(TASK_RUNNING);

	if (os != NULL) {
		ak4642_state.wr_ref = 0;
	}
	if (is != NULL) {
		ak4642_state.rd_ref = 0;
	}
}

static int ak4642_codec_power_on(void)
{
	int res = 0;
	int seq_no = 0;
	int volume;
	unsigned char reg;
	unsigned char data;
	unsigned char mask;

	while (seq_no != AK4642_POWER_ON_MAX_SEQ_NUM) {
		switch (seq_no) {
			/* case 0 - 4 : Clock Set Up Sequence */
		case 0:
			reg = AK4642REG_01H_POWER_MANAGEMENT_2;
			data = AK4642BIT_01H_08H_M_S;
			mask = AK4642BIT_01H_08H_M_S;
			break;
		case 1:
			reg = AK4642REG_04H_MODE_CONTROL_1;
#ifdef MP201_CODEC_32
			data =
			    AK4642BIT_04H_40H_PLL2 | AK4642BIT_04H_20H_PLL1 |
			    AK4642BIT_04H_02H_DIF1 | AK4642BIT_04H_08H_BCKO;
#else
			data =
			    AK4642BIT_04H_40H_PLL2 | AK4642BIT_04H_20H_PLL1 |
			    AK4642BIT_04H_02H_DIF1;
#endif
			mask =
			    AK4642BIT_04H_F0H_PLL_MASK | AK4642BIT_04H_08H_BCKO
			    | AK4642BIT_04H_03H_DIF_MASK;
			break;
		case 2:
			reg = AK4642REG_05H_MODE_CONTROL_2;
			data = AK4642BIT_05H_FS_8KHZ;
			mask = AK4642BIT_05H_27H_FS_MASK;
			break;
		case 3:
			reg = AK4642REG_00H_POWER_MANAGEMENT_1;
			data = AK4642BIT_00H_40H_PMVCM;
			mask = AK4642BIT_00H_40H_PMVCM;
			break;
		case 4:
			reg = AK4642REG_01H_POWER_MANAGEMENT_2;
			data = AK4642BIT_01H_01H_PMPLL;
			mask = AK4642BIT_01H_01H_PMPLL;
			break;

			/* case 5 - 13 : Stereo Lineout Sequence */
		case 5:
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(AK4642_WAIT_PLL_LOCK);
			set_current_state(TASK_RUNNING);

			reg = AK4642REG_02H_SIGNAL_SELECT_1;
			data = AK4642BIT_02H_10H_DACL;
			mask = AK4642BIT_02H_10H_DACL;
			break;
		case 6:
			reg = AK4642REG_09H_LCH_INPUT_VOLUME_CONTROL;
			data = 0x91;	/* 0dB */
			mask = AK4642_INPUT_VOL_MAX;
			break;
		case 7:
			reg = AK4642REG_0CH_RCH_INPUT_VOLUME_CONTROL;
			data = 0x91;	/* 0dB */
			mask = AK4642_INPUT_VOL_MAX;
			break;
		case 8:
			volume = mixer_ak4642_volume_read(SOUND_MASK_PCM);
			volume = (volume & 0xff);
			reg = AK4642REG_0AH_LCH_DIGITAL_VOLUME_CONTROL;
			data = (AK4642_OUTPUT_VOL_MAX * (100 - volume)) / 100;
			mask = AK4642_OUTPUT_VOL_MAX;
			break;
		case 9:
			volume = mixer_ak4642_volume_read(SOUND_MASK_PCM);
			volume = ((volume >> 8) & 0xff);
			reg = AK4642REG_0DH_RCH_DIGITAL_VOLUME_CONTROL;
			data = (AK4642_OUTPUT_VOL_MAX * (100 - volume)) / 100;
			mask = AK4642_OUTPUT_VOL_MAX;
			break;
		case 10:
			reg = AK4642REG_0EH_MODE_CONTROL_3;
			data = 0x00;
			mask = AK4642BIT_0EH_10H_DVOLC;
			break;
		case 11:
			reg = AK4642REG_03H_SIGNAL_SELECT_2;
			data = AK4642BIT_03H_40H_LOPS;
			mask = AK4642BIT_03H_40H_LOPS;
			break;
		case 12:
			reg = AK4642REG_00H_POWER_MANAGEMENT_1;
			data = AK4642BIT_00H_08H_PMLO | AK4642BIT_00H_04H_PMDAC;
			mask = AK4642BIT_00H_08H_PMLO | AK4642BIT_00H_04H_PMDAC;
			break;
		case 13:
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(AK4642_WAIT_LO_LOCK);
			set_current_state(TASK_RUNNING);

			reg = AK4642REG_03H_SIGNAL_SELECT_2;
			data = 0x00;
			mask = AK4642BIT_03H_40H_LOPS;
			break;

			/* case 14 - 24 : Mic Input Recording Sequence */
		case 14:
			reg = AK4642REG_02H_SIGNAL_SELECT_1;
			data = AK4642BIT_02H_04H_PMMP;
			mask =
			    AK4642BIT_02H_04H_PMMP | AK4642BIT_02H_01H_MGAIN0;
			break;
		case 15:
			reg = AK4642REG_10H_POWER_MANAGEMENT_3;
			data = 0x00;
			mask = AK4642BIT_10H_06H_IN_MASK;
			break;
		case 16:
			volume = mixer_ak4642_volume_read(SOUND_MASK_MIC);
			volume = (volume & 0xff);
			reg = AK4642REG_09H_LCH_INPUT_VOLUME_CONTROL;
			data = (AK4642_INPUT_VOL_MAX * volume) / 100;
			mask = AK4642_INPUT_VOL_MAX;
			break;
		case 17:
			volume = mixer_ak4642_volume_read(SOUND_MASK_MIC);
			volume = ((volume >> 8) & 0xff);
			reg = AK4642REG_0CH_RCH_INPUT_VOLUME_CONTROL;
			data = (AK4642_INPUT_VOL_MAX * volume) / 100;
			mask = AK4642_INPUT_VOL_MAX;
			break;
		case 18:
			reg = AK4642REG_0FH_MODE_CONTROL_4;
			data = 0x00;
			mask = AK4642BIT_0FH_08H_IVOLC;
			break;
		case 19:
			reg = AK4642REG_06H_TIMER_SELECT;
			data = AK4642BIT_06H_20H_ZTM1 | AK4642BIT_06H_08H_WTM1;
			mask =
			    AK4642BIT_06H_30H_ZTM_MASK |
			    AK4642BIT_06H_0CH_WTM_MASK;
			break;
		case 20:
			volume = mixer_ak4642_volume_read(SOUND_MASK_MIC);
			volume = (volume & 0xff);
			reg = AK4642REG_08H_ALC_MODE_CONTROL_2;
			data = (AK4642_INPUT_VOL_MAX * volume) / 100;
			mask = AK4642_INPUT_VOL_MAX;
			break;
		case 21:
			reg = AK4642REG_0BH_ALC_MODE_CONTROL_3;
			data = 0x00;
			mask = AK4642BIT_0BH_80H_RGAIN1 |
			    AK4642BIT_0BH_40H_LMTH1;
			break;
		case 22:
			reg = AK4642REG_07H_ALC_MODE_CONTROL_1;
			data = AK4642BIT_07H_20H_ALC | AK4642BIT_07H_01H_LMTH0;
			mask =
			    AK4642BIT_07H_20H_ALC | AK4642BIT_07H_10H_ZELMN |
			    AK4642BIT_07H_0CH_LMAT_MASK |
			    AK4642BIT_07H_02H_RGAIN0 | AK4642BIT_07H_01H_LMTH0;
			break;
		case 23:
			reg = AK4642REG_00H_POWER_MANAGEMENT_1;
			data = AK4642BIT_00H_01H_PMADL;
			mask = AK4642BIT_00H_01H_PMADL;
			break;
		case 24:
			reg = AK4642REG_10H_POWER_MANAGEMENT_3;
			data = AK4642BIT_10H_01H_PMADR;
			mask = AK4642BIT_10H_01H_PMADR;
			break;

		default:
			res = -1;
			return res;
		}

		res = i2c_ak4642_write(reg, data, mask);
		if (res != 0) {
			break;
		}

		seq_no++;
	}

	return res;
}

static int ak4642_probe(struct device *dev)
{
	int res = 0;

	/* register devices */
	audio_dev_id = register_sound_dsp(&mp200_ak4642_fops, -1);
	if (audio_dev_id < 0) {
		return audio_dev_id;
	}

	mixer_dev_id = register_sound_mixer(&mp200_mixer_ak4642_fops, -1);
	if (mixer_dev_id < 0) {
		unregister_sound_dsp(audio_dev_id);
		return mixer_dev_id;
	}

#ifdef CONFIG_MP201
	/* REFCLK sel div gate open */
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_REFCLKO);
	outl(0, SMU_REFCLKO_SEL);	/* OSC(12MHz) */
	outl(0, SMU_REFCLKO_DIV);	/* div 1 */
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_REFCLKO);
#endif

	/* PM0 MACRO unreset  */
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_SCLK);
	mp200_pmu_unreset_device(MP200_RESETDEVICE_PM0);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_SCLK);

#ifdef CONFIG_MP201
	/* ak4642 CSN/CAD0 pin "L" */
	/* SNDCDA AK4642I2Cɥ쥹 5T735:GPIO8 IOOUT2 D0 */
	gpiocore_pwc_write(GPIOCORE_PWC_IOOUT2, 0x0, 0x01);

	/* ak4642 PDN pin "L" -> "H" */
	/* SNDPDN AK4642 Reset 5T735:GPIO9 IOOUT2 D1 */
	gpiocore_pwc_write(GPIOCORE_PWC_IOOUT2, 0xff, 0x02);
#else
	/* ak4642 CSN/CAD0 pin "L" */
	gpiocore_write(GPIOCORE_GP_O, GPIOCORE_GPIO_SNDCDA, 0);

	/* ak4642 PDN pin "L" -> "H" */
	gpiocore_write(GPIOCORE_GP_O, GPIOCORE_GPIO_SNDPDN, 1);
#endif

	res = i2c_ak4642_init();
	if (res != 0) {
		goto out2;
	}

	res = ak4642_codec_power_on();
	if (res != 0) {
		goto out1;
	}

	res = audio_pcm_set_ctrl(&ak4642_state);
	if (res != 0) {
		goto out1;
	}

	return 0;

      out1:
	i2c_ak4642_cleanup();

      out2:
#ifdef CONFIG_MP201
	/* ak4642 PDN pin "L" -> "H" */
	/* SNDPDN AK4642 Reset 5T735:GPIO9 IOOUT2 D1 */
	gpiocore_pwc_write(GPIOCORE_PWC_IOOUT2, 0x0, 0x02);
#else
	gpiocore_write(GPIOCORE_GP_O, GPIOCORE_GPIO_SNDPDN, 0);
#endif

	/* PDN pin "L" needs 150ns */
	udelay(1);

	/* PM0 MACRO reset */
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_SCLK);
	mp200_pmu_reset_device(MP200_RESETDEVICE_PM0);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_SCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_PCLK);

#ifdef CONFIG_MP201
	/* REFCLK gate close */
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_REFCLKO);
#endif

	/* unregister devices */
	unregister_sound_mixer(mixer_dev_id);
	unregister_sound_dsp(audio_dev_id);

	return res;
}

static int ak4642_remove(struct device *dev)
{
	i2c_ak4642_cleanup();

#ifdef CONFIG_MP201
	/* SNDPDN AK4642 Reset 5T735:GPIO9 IOOUT2 D1 */
	gpiocore_pwc_write(GPIOCORE_PWC_IOOUT2, 0x0, 0x02);
#else
	gpiocore_write(GPIOCORE_GP_O, GPIOCORE_GPIO_SNDPDN, 0);
#endif

	/* PDN pin "L" needs 150ns */
	udelay(1);

	/* PM0 MACRO reset */
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_SCLK);
	mp200_pmu_reset_device(MP200_RESETDEVICE_PM0);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_SCLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_PCLK);

#ifdef CONFIG_MP201
	/* REFCLK gate close */
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_REFCLKO);
#endif

	/* unregister devices */
	unregister_sound_mixer(mixer_dev_id);
	unregister_sound_dsp(audio_dev_id);

	return 0;
}

static int ak4642_suspend(struct device *dev, u32 state, u32 level)
{
	int res = 0;

	switch (level) {
#ifdef MP200_AUDIO_OPEN_SUSPEND
#else
	case SUSPEND_DISABLE:
		if (((ak4642_state.wr_ref | ak4642_state.rd_ref) &
		     AUDIO_REF_EN) != 0) {
			return -EBUSY;
		}
		break;
#endif
	case SUSPEND_SAVE_STATE:
		pcm0_regs_save.func_sel = *pcm0_regs.func_sel;
		pcm0_regs_save.cycle = *pcm0_regs.cycle;
		pcm0_regs_save.cycle2 = *pcm0_regs.cycle2;
		pcm0_regs_save.enmon = *pcm0_regs.enmon;
		pcm0_regs_save.txrx_mon = *pcm0_regs.txrx_mon;

		res =
		    i2c_ak4642_write(AK4642REG_03H_SIGNAL_SELECT_2,
				     AK4642BIT_03H_40H_LOPS,
				     AK4642BIT_03H_40H_LOPS);
		if (res != 0) {
			return -EFAULT;
		}

		/* PM0 MACRO reset */
		mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_SCLK);
		mp200_pmu_reset_device(MP200_RESETDEVICE_PM0);
		mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_SCLK);
		mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_PCLK);

#ifdef CONFIG_MP201
		/* REFCLK gate close */
		mp200_pmu_close_clockgate(MP200_CLOCKGATE_REFCLKO);
#endif
		break;
	default:
		break;
	}

	return res;
}

static int ak4642_resume(struct device *dev, u32 level)
{
	int res = 0;

	switch (level) {
	case RESUME_RESTORE_STATE:
#ifdef CONFIG_MP201
		/* REFCLK sel div gate open */
		mp200_pmu_close_clockgate(MP200_CLOCKGATE_REFCLKO);
		outl(0, SMU_REFCLKO_SEL);	/* OSC(12MHz) */
		outl(0, SMU_REFCLKO_DIV);	/* div 1 */
		mp200_pmu_open_clockgate(MP200_CLOCKGATE_REFCLKO);
#endif
		/* PM0 MACRO unreset  */
		mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_PCLK);
		mp200_pmu_open_clockgate(MP200_CLOCKGATE_PM0_SCLK);
		mp200_pmu_unreset_device(MP200_RESETDEVICE_PM0);
		mp200_pmu_close_clockgate(MP200_CLOCKGATE_PM0_SCLK);

		res =
		    i2c_ak4642_write(AK4642REG_03H_SIGNAL_SELECT_2, 0x00,
				     AK4642BIT_03H_40H_LOPS);
		if (res != 0) {
			return -EFAULT;
		}

		*pcm0_regs.func_sel = pcm0_regs_save.func_sel;
		*pcm0_regs.cycle = pcm0_regs_save.cycle;
		*pcm0_regs.cycle2 = pcm0_regs_save.cycle2;
		*pcm0_regs.enclr = ~(pcm0_regs_save.enmon);
		*pcm0_regs.enset = pcm0_regs_save.enmon;
		*pcm0_regs.txrx_dis = ~(pcm0_regs_save.txrx_mon);
		*pcm0_regs.txrx_en = pcm0_regs_save.txrx_mon;
		break;
	default:
		break;
	}

	return res;
}

int __init ak4642_init(void)
{
	printk(KERN_INFO "Starting ak4642 codec.\n");

	return driver_register(&mp200_ak4642_driver);
}

void __exit ak4642_exit(void)
{
	driver_unregister(&mp200_ak4642_driver);
}

module_init(ak4642_init);
module_exit(ak4642_exit);
MODULE_LICENSE("GPL");
