/*
 *  File Name       : linux/sound/oss/mp200-audio.c
 *  Function        : Common audio handling for the MP200 processor
 *  Release Version : Ver 1.01
 *  Release Date    : 2006/03/29
 *
 *  Copyright (C) NEC Electronics Corporation 2005-2006
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/poll.h>
#include <linux/soundcard.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>

#ifdef CONFIG_MP201
#include <asm/io.h>
#endif

#include "mp200-audio.h"

/* #define AUDIO_DEBUG 1 */

#ifdef AUDIO_DEBUG
#define DPRINTK(fmt, args...)	printk(KERN_INFO "%s(): " fmt, __FUNCTION__, ## args)
#else
#define DPRINTK(fmt, args...)
#endif

/* #define AUDIO_DEBUG_LOCAL 1 */

#ifdef AUDIO_DEBUG_LOCAL
#define DLPRINTK(fmt, args...)	printk(KERN_INFO "%s(): " fmt, __FUNCTION__, ## args)
#else
#define DLPRINTK(fmt, args...)
#endif

/* #define AUDIO_DEBUG_LOG 1 */

#ifdef AUDIO_DEBUG_LOG
#define DEBUG_LOG_MAX 4096
#define DMAM2PL4_ADDR		IO_ADDRESS(MP200_DMAC_BASE+0x1400)
#define MP200_REG(ADDR, OFF) \
		(*(unsigned int volatile *)((unsigned char *)(ADDR) + OFF))
#define DMA_LENG_RCOUNT(ADDR)	MP200_REG(ADDR, 0x0044) /* R */
#define DMA_LENG_WCOUNT(ADDR)	MP200_REG(ADDR, 0x0048) /* R */
static int w_count =0;
static struct timeval w_usec;
static unsigned long f_jiff;
typedef struct {
	uint16_t pending;
	uint16_t fifo;
	uint32_t time;
	uint32_t state;
} w_log;
static w_log* w_dbg_log;
static int state0;
static int state1;
#endif

/* #define AUDIO_DEBUG_HIT 1 */

/* #define AUDIO_DEBUG_STRESS 1 */

/* #define AUDIO_DEBUG_FRE 1 */

#ifdef MP200_AUDIO_PCM_NONFRE
static void audio_stop_dma(audio_stream_t * s);
static void audio_process_dma(audio_stream_t *s);

static int mp200_dsp_speed;
static void mp200_pcm_kick(unsigned long);
DECLARE_TASKLET(mp200_pcm_tasklet, mp200_pcm_kick, 0);
static void mp200_pcm_kick(unsigned long inData)
{
	audio_state_t *state = (audio_state_t *)inData;
	audio_stream_t *os = state->output_stream;
	int wait_udelay;

	wait_udelay = (1000000/mp200_dsp_speed) + ((1000000%mp200_dsp_speed)?1:0);
	switch(state->ctrls->sett->func.mode_sel) {
	case PCM_MODE_2: case PCM_MODE_3: case PCM_MODE_4:
		break;
	default:
		wait_udelay *= 2;
	}
	udelay(wait_udelay);

	/* printk("kick"); */
	if(1 == mp200_dma_status(os->dma_channel)) {
		*os->pcm_regs->txrx_en = os->en_bit;
	}
	else {
		unsigned long   flags;

		local_irq_save(flags);
		audio_stop_dma(os);
		os->stopped = 0;
		if (!os->active) {
			audio_process_dma(os);
		}
		local_irq_restore(flags);
	}

#ifdef AUDIO_DEBUG_LOG
	if (DEBUG_LOG_MAX > w_count) {
		pcm_regs_t *pcm_regs = state->ctrls->regs;
		unsigned long jiff = jiffies;
		(w_dbg_log+w_count)->pending = (uint16_t)99;
		(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
		(w_dbg_log+w_count)->fifo = 99;
		(w_dbg_log+w_count)->state = (uint32_t)*pcm_regs->txrx_mon;
		f_jiff = jiff;
		w_count++;
	}
#endif
	return;
}
#endif

#ifdef AUDIO_DEBUG_FRE
static void debug_fre_poll(unsigned long dummy);
static struct timer_list poll_timer = TIMER_INITIALIZER(debug_fre_poll, 0, 0);
static void debug_fre_poll(unsigned long dummy)
{
	unsigned long   flags;
	audio_state_t *state = (audio_state_t *) dummy;
	pcm_regs_t *pcm_regs = state->ctrls->regs;
	audio_stream_t *os = state->output_stream;

	local_irq_save(flags);
	if (state->wr_ref && (NULL != os->buffers)) {
		/* mp200_stop_dma(os->dma_channel); */
		*pcm_regs->txrx_dis = os->en_bit;
		/* mdelay(3); */
#ifdef AUDIO_DEBUG_LOG_
		if (DEBUG_LOG_MAX > w_count) {
			unsigned long jiff = jiffies;
			(w_dbg_log+w_count)->pending = 88;
			(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
			(w_dbg_log+w_count)->fifo = 99;
			(w_dbg_log+w_count)->state = (uint32_t)*pcm_regs->txrx_mon;
			f_jiff = jiff;
			w_count++;
		}
#endif

		mp200_pcm_tasklet.data = (unsigned long)state;
		tasklet_schedule(&mp200_pcm_tasklet);
	}
	local_irq_restore(flags);

	poll_timer.expires = (3) + jiffies;
	poll_timer.data = (unsigned long)state;
	add_timer(&poll_timer);

	return;
}
#endif

/*
 * DMA processing
 */
static u_int audio_get_dma_pos(audio_stream_t * s)
{
	audio_buf_t *b = &s->buffers[s->dma_tail];
	u_int offset;

	if (b->dma_ref != 0) {
		offset = mp200_get_dma_pos(s->dma_channel) - b->dma_addr;
		if (offset >= s->fragsize) {
			offset = s->fragsize - 4;
		}
	} else if (s->pending_frags != 0) {
		offset = b->offset;
	} else {
		offset = 0;
	}

	return offset;
}

static void audio_stop_dma(audio_stream_t * s)
{
#ifdef MP200_AUDIO_SINGLE_TRANSFER_2_
#else
	audio_buf_t *b;
#endif
	unsigned long flags;
	u_int pos;

	if (s->buffers == NULL) {
		return;
	}

	local_irq_save(flags);
	if ((s->en_bit & *s->pcm_regs->txrx_mon) != 0) {
		*s->pcm_regs->txrx_dis = s->en_bit;
	}

	s->stopped = 1;
	mp200_stop_dma(s->dma_channel);
	pos = audio_get_dma_pos(s);
	mp200_clear_dma(s->dma_channel);
	local_irq_restore(flags);

	/* 1 frame wait */
	udelay(22);

#ifdef MP200_AUDIO_SINGLE_TRANSFER_3
	s->active = 0;
#endif

#ifdef MP200_AUDIO_SINGLE_TRANSFER_2_
#else
	/* back up pointers to be ready to restart from the same spot */
	while (s->dma_head != s->dma_tail) {
		b = &s->buffers[s->dma_head];
		if (b->dma_ref != 0) {
			b->dma_ref = 0;
			b->offset = 0;
		}
		s->pending_frags++;
		if (s->dma_head == 0) {
			s->dma_head = s->nbfrags;
		}
		s->dma_head--;
	}
	b = &s->buffers[s->dma_head];
	if (b->dma_ref != 0) {
		/* printk(KERN_INFO "pos : %d\n", pos); */
		b->offset = pos;
		b->dma_ref = 0;
	}
#endif
	DPRINTK("[%s] pending=%d dma_head=%d dma_tail=%d offset=%d\n",
		s->id, s->pending_frags, s->dma_head, s->dma_tail, b->offset);
}

static void audio_reset(audio_stream_t * s)
{
	if (s->buffers != NULL) {
#ifdef AUDIO_DEBUG_LOG
		if (DEBUG_LOG_MAX > w_count) {
			unsigned long jiff = jiffies;
			(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
			(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
			(w_dbg_log+w_count)->fifo = 44;
			(w_dbg_log+w_count)->state = (uint32_t)atomic_read(&s->sem.count);
			f_jiff = jiff;
			w_count++;
		}
#if 0
		if (DEBUG_LOG_MAX > w_count) {
			(w_dbg_log+w_count)->pending = (uint16_t)s->usr_head;
			(w_dbg_log+w_count)->time = (uint16_t)s->dma_head;
			(w_dbg_log+w_count)->fifo = 45;
			(w_dbg_log+w_count)->state = (uint32_t)s->dma_tail;
			w_count++;
		}
#endif
#endif
		audio_stop_dma(s);
		s->buffers[s->dma_head].offset = 0;
		s->buffers[s->usr_head].offset = 0;
#ifdef MP200_AUDIO_SINGLE_TRANSFER
		s->usr_head = s->dma_head;
#else
		s->usr_head = s->dma_head = s->dma_tail = 0;
#endif
		if (s->type == AUDIO_IN) {
			s->pending_frags = s->nbfrags;
			sema_init(&s->sem, 0);
		} else {
			s->pending_frags = 0;
			sema_init(&s->sem, s->nbfrags);
		}
	}
#ifdef MP200_AUDIO_SINGLE_TRANSFER
	s->pcm_intr = 0;
	s->pcm_status = 0;
#endif
	s->active = 0;
	s->stopped = 0;
#ifdef AUDIO_DEBUG_LOG_
	if (s->buffers != NULL) {
		if (DEBUG_LOG_MAX > w_count) {
			(w_dbg_log+w_count)->pending = (uint16_t)s->usr_head;
			(w_dbg_log+w_count)->time = (uint16_t)s->dma_head;
			(w_dbg_log+w_count)->fifo = 45;
			(w_dbg_log+w_count)->state = (uint32_t)s->dma_tail;
			w_count++;
		}
	}
#endif
}

#ifdef MP200_AUDIO_SINGLE_TRANSFER
static void audio_process_dma(audio_stream_t *s)
{
	dma_addr_t dma_ptr;
	dma_addr_t dst_ptr;
#ifdef CONFIG_MP201
	unsigned int mode_bit = MP200_DMAC_DEFMODE_16BIT;
#else
	unsigned int mode_bit = MP200_DMAC_AMODE_BIT16;
#endif
	int ret;
	audio_buf_t *b = &s->buffers[s->dma_head];
	u_int dma_size = s->fragsize - b->offset;
#ifdef MP201_SINGLE_SPECIAL
	int adj = 0;
	int timeout = 136; /* 7350Hz  FIFO 1ʬ */
	int supp = 10; /* callback 䴰 */
	int adjbyte = 0;
	unsigned int fifoval = 0;
	unsigned int r_cnt;
	unsigned int w_cnt;
	char *bdata = b->data + b->offset;
	uint32_t fifo_buf;
#endif
	if(s->stopped)
		return;

	if(1 == mp200_dma_status(s->dma_channel))
		return;

	switch (s->format) {
	case AFMT_S16_LE:
#ifdef CONFIG_MP201
#ifdef MP201_PADDING_ON
		mode_bit = (MP200_DMAC_AMODE_BIT32
			  | MP200_DMAC_ENDI_R1032
			  | MP200_DMAC_ENDI_W3210);
#ifdef MP201_SINGLE_SPECIAL
		adjbyte = 4;
#endif
#else
		mode_bit = MP200_DMAC_DEFMODE_16BIT;
#ifdef MP201_SINGLE_SPECIAL
		adjbyte = 2;
#endif
#endif
#else
		mode_bit = MP200_DMAC_AMODE_BIT16;
#endif
		break;
#ifdef MP201_CODEC_32
	case AFMT_S32_LE:
		mode_bit = MP200_DMAC_DEFMODE_32BIT;
#ifdef MP201_SINGLE_SPECIAL
		adjbyte = 4;
#endif
		break;
#endif
	default:
		break;
	}

	if (s->pending_frags) {
		if(PCM_TX_EN == s->en_bit) {
#ifdef MP201_SINGLE_SPECIAL
			/* DMA ȤƤ DMA_TDMARQ  ȡ
			 * ǥȤζ̤ĤʤΤǡǥȤ
			 * 硢ȤޤǤԤġ
			 */
			if (s->en_bit & *s->pcm_regs->txrx_mon) {
#ifdef AUDIO_DEBUG_STRESS
				{
					int i;
					for (i=0; i<timeout; i++) {
						fifoval = *s->pcm_regs->fifo_p;
						r_cnt = (fifoval & 0x0000ff00) >> 8;
						w_cnt = fifoval & 0x000000ff;
						if(w_cnt >= r_cnt) fifoval = w_cnt - r_cnt;
						else fifoval = 32 - r_cnt + w_cnt;
						if (1 > fifoval) {
							break;
						}
						udelay(5);
					}
				}
#endif
				/* FIFO ȤΤޤĴ٤
				 * ưΤᡢ1WORD θ
				 * 2ɤߡ
				 */
				fifoval = *s->pcm_regs->fifo_p;
				fifoval = *s->pcm_regs->fifo_p;

				r_cnt = (fifoval & 0x0000ff00) >> 8;
				w_cnt = fifoval & 0x000000ff;
				if(w_cnt >= r_cnt) fifoval = w_cnt - r_cnt;
				else fifoval = 32 - r_cnt + w_cnt;

#ifdef MP200_AUDIO_SINGLE_TRANSFER_2
				if (s->irq_bit & *s->pcm_regs->raw) {
#ifdef AUDIO_DEBUG_LOG
					if (DEBUG_LOG_MAX > w_count) {
						unsigned long jiff = jiffies;
						(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
						(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
						(w_dbg_log+w_count)->fifo = (uint16_t)fifoval;
						(w_dbg_log+w_count)->state = (uint32_t)7;
						f_jiff = jiff;
						w_count++;
					}
#endif
					return;
				}
#endif

#ifdef MP200_AUDIO_SINGLE_TRANSFER_2
				if (!fifoval) {
#ifdef AUDIO_DEBUG_LOG
					if (DEBUG_LOG_MAX > w_count) {
						unsigned long jiff = jiffies;
						(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
						(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
						(w_dbg_log+w_count)->fifo = (uint16_t)fifoval;
						(w_dbg_log+w_count)->state = (uint32_t)5;
						f_jiff = jiff;
						w_count++;
					}
#endif
					return;
				}
#endif

				/* FIFO  5 İʲʤ䴰롣
				 * 124MHz 44.1 kHz fifoval μ¬顣
				 */
				if (supp >= fifoval) {
					while (supp) {
						if (s->irq_bit & *s->pcm_regs->raw) {
#ifdef AUDIO_DEBUG_HIT
							printk(KERN_INFO "hit.");
#endif
#ifdef AUDIO_DEBUG_LOG
							if (DEBUG_LOG_MAX > w_count) {
								unsigned long jiff = jiffies;
								(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
								(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
								(w_dbg_log+w_count)->fifo = (uint16_t)fifoval;
								(w_dbg_log+w_count)->state = (uint32_t)4;
								f_jiff = jiff;
								w_count++;
							}
#endif
							return;
						}
						if (AFMT_S16_LE == s->format) {
							fifo_buf = *(uint32_t const *)bdata >> 16;
							fifo_buf |= *(uint32_t const *)bdata << 16;
							*s->pcm_regs->txq = fifo_buf;
						}
						else {
							*s->pcm_regs->txq = (uint32_t)*bdata;
						}
						bdata += adjbyte;
						adj += adjbyte;
						supp--;
					}
				}

				while (timeout) {
					if (PCM_INT_TX_WEN & *s->pcm_regs->raw) {
						adj += adjbyte;
						break;
					}
					udelay(1);
					timeout--;
				}
#ifdef AUDIO_DEBUG_LOG
				state1 = 0;
#endif
			}
#ifdef AUDIO_DEBUG_LOG
			else {
				state1 = 1;
			}
#endif

#ifdef AUDIO_DEBUG_LOCAL
			if (!timeout)
				DLPRINTK("***** timeout *****\n");
#endif
			dma_ptr = b->dma_addr + b->offset + adj;
			dst_ptr = s->phys_xq;
			s->dma_regs->aoff = 0;
			s->dma_regs->asize_count = 0;
			s->dma_regs->asize = dma_size - adj;
			s->dma_regs->leng = dma_size - adj;
			s->dma_regs->mode = mode_bit;
#else
			dma_ptr = b->dma_addr + b->offset;
			dst_ptr = s->phys_xq;
			s->dma_regs->aoff = 0;
			s->dma_regs->asize_count = 0;
			s->dma_regs->asize = dma_size;
			s->dma_regs->leng = dma_size;
			s->dma_regs->mode = mode_bit;
#endif
		}
		else {
			dma_ptr = s->phys_xq;
			dst_ptr = b->dma_addr + b->offset;
			s->dma_regs->boff = 0;
			s->dma_regs->bsize_count = 0;
			s->dma_regs->bsize = dma_size;
			s->dma_regs->leng = dma_size;
			s->dma_regs->mode = mode_bit;
		}

#ifdef MP201_SINGLE_SPECIAL
		/* PCM Υ顼ȯƤȤ
		 * DMA Ȥ򤫤ȥΥǤ롣
		 */
		if (s->irq_bit & *s->pcm_regs->raw) {
#ifdef AUDIO_DEBUG_HIT
			printk(KERN_INFO "hit.2");
#endif
#ifdef AUDIO_DEBUG_LOG
			if (DEBUG_LOG_MAX > w_count) {
				unsigned long jiff = jiffies;
				(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
				(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
				(w_dbg_log+w_count)->fifo = (uint16_t)fifoval;
				(w_dbg_log+w_count)->state = (uint32_t)3;
				f_jiff = jiff;
				w_count++;
			}
#endif
			return;
		}
#endif

		/* size 򥻥åȤ  mp200_start_dma() Ǥ
		 * defmode : MP200_DMAC_DEFMODE_32BIT 򥻥åȤƤޤΤǡ
		 *  dma regs 򥻥åȤƥ뤹롣
		 * MP201_SINGLE_SPECIAL : mp200_start_dma Ȥ鷺˼ DMA 򥻥å
		 *               ¿®ǽ
		 */
		ret = mp200_start_dma( s->dma_channel,
					dma_ptr, 0,
					dst_ptr, s->int_dma);

		if (ret) {
			DPRINTK("DMA start fault\n");
			return;
		}

		if(0 == (s->en_bit & *s->pcm_regs->txrx_mon)) {
			*s->pcm_regs->txrx_en = s->en_bit;
#ifdef MP200_AUDIO_SINGLE_TRANSFER_3
			s->active = 1;
#endif
#ifdef AUDIO_DEBUG_LOG
			state0 = 1;
#endif
		}
#ifdef MP201_SINGLE_SPECIAL
		else {
			if (PCM_TX_EN == s->en_bit
			&& (PCM_INT_TX_WEN & *s->pcm_regs->raw)) {
#ifdef MP201_PADDING_ON
				if (AFMT_S16_LE == s->format) {
					fifo_buf = *(uint32_t const *)bdata >> 16;
					fifo_buf |= *(uint32_t const *)bdata << 16;
					*s->pcm_regs->txq = fifo_buf;
				}
				else {
					*s->pcm_regs->txq = (uint32_t)*bdata;
				}
#else
				if (AFMT_S16_LE == s->format) {
					fifo_buf = *(uint16_t const *)bdata;
					*s->pcm_regs->txq = fifo_buf;
				}
				else {
					*s->pcm_regs->txq = (uint32_t)*bdata;
				}
#endif
#ifdef AUDIO_DEBUG_LOG
				state0 = 0;
#endif
			}
#ifdef MP200_AUDIO_SINGLE_TRANSFER_2
			else {
#ifdef AUDIO_DEBUG_LOG
				if (DEBUG_LOG_MAX > w_count) {
					unsigned long jiff = jiffies;
					(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
					(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
					(w_dbg_log+w_count)->fifo = (uint16_t)fifoval;
					(w_dbg_log+w_count)->state = (uint32_t)6;
					f_jiff = jiff;
					w_count++;
				}
#endif
				DPRINTK("DMA start fault\n");
				return;
			}
#endif
		}
#endif

#ifdef AUDIO_DEBUG_LOG
		if (state1) {
			if (DEBUG_LOG_MAX > w_count) {
				unsigned long jiff = jiffies;
				(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
				(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
				(w_dbg_log+w_count)->fifo = (uint16_t)89;
				(w_dbg_log+w_count)->state = (uint32_t)atomic_read(&s->sem.count);
				f_jiff = jiff;
				w_count++;
			}
		}
		else {
			if (10 > fifoval) {
				if (DEBUG_LOG_MAX > w_count) {
					unsigned long jiff = jiffies;
					(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
					(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
					(w_dbg_log+w_count)->fifo = (uint16_t)fifoval;
					(w_dbg_log+w_count)->state = (uint32_t)state0;
					f_jiff = jiff;
					w_count++;
				}
			}
		}
#endif

		DPRINTK("[%s] pending=%d dma_head=%d dma_tail=%d offset=%d size=%d\n",
			s->id, s->pending_frags, s->dma_head, s->dma_tail,
			s->buffers[s->dma_head].offset, dma_size);

		b->dma_ref++;
		b->offset += dma_size;
		if (b->offset >= s->fragsize) {
			s->pending_frags--;
			if (++s->dma_head >= s->nbfrags)
				s->dma_head = 0;
		}

	}

}
#else
static void audio_process_dma(audio_stream_t * s)
{
	audio_buf_t *b = &s->buffers[s->dma_head];
	u_int dma_size = s->fragsize - b->offset;

	if (s->pending_frags != 0) {
		b->dma_ref++;
		b->offset += dma_size;
		if (b->offset >= s->fragsize) {
			s->pending_frags--;
			if (++s->dma_head >= s->nbfrags) {
				s->dma_head = 0;
			}
		}
	}
}

static int audio_start_dma(audio_stream_t * s)
{
	dma_addr_t dma_ptr;
	dma_addr_t dst_ptr;
#ifdef CONFIG_MP201
	unsigned int mode_bit = MP200_DMAC_DEFMODE_16BIT;
#else
	unsigned int mode_bit = MP200_DMAC_AMODE_BIT16;
#endif
	int ret;

	if (s->stopped != 0) {
		return 0;
	}

	switch (s->format) {
	case AFMT_S16_LE:
#ifdef CONFIG_MP201
#ifdef MP201_PADDING_ON
		mode_bit = (MP200_DMAC_AMODE_BIT32
			  | MP200_DMAC_ENDI_R1032
			  | MP200_DMAC_ENDI_W1032);
#else
		mode_bit = MP200_DMAC_DEFMODE_16BIT;
#endif
#else
		mode_bit = MP200_DMAC_AMODE_BIT16;
#endif
		break;
#ifdef CONFIG_MP201
	case AFMT_S32_LE:
		mode_bit = MP200_DMAC_DEFMODE_32BIT;
		break;
#endif
	default:
		break;
	}

	if ((mp200_dma_status(s->dma_channel) == 0) && (s->pending_frags != 0)) {
		if (s->en_bit == PCM_TX_EN) {
			dma_ptr = s->dma_top;
			dst_ptr = s->phys_xq;
			s->dma_regs->aoff = 0;
			s->dma_regs->asize = s->fragsize;
			s->dma_regs->asize_count = s->nbfrags - 1;
			s->dma_regs->leng = 0;
			s->dma_regs->mode = MP200_DMAC_AMODE_REPEAT | mode_bit;
		} else {
			dma_ptr = s->phys_xq;
			dst_ptr = s->dma_top;
			s->dma_regs->boff = 0;
			s->dma_regs->bsize = s->fragsize;
			s->dma_regs->bsize_count = s->nbfrags - 1;
			s->dma_regs->leng = 0;
			s->dma_regs->mode = MP200_DMAC_BMODE_REPEAT | mode_bit;
		}

		DPRINTK("[%s] pending=%d dma_head=%d dma_tail=%d offset=%d\n",
			s->id, s->pending_frags, s->dma_head, s->dma_tail,
			s->buffers[s->dma_head].offset);

		ret =
		    mp200_start_dma(s->dma_channel, dma_ptr, 0, dst_ptr,
				    s->int_dma);
		if (ret != 0) {
			return ret;
		}

		if ((s->en_bit & *s->pcm_regs->txrx_mon) == 0) {
			*s->pcm_regs->txrx_en = s->en_bit;
		}

		s->active = 1;
		audio_process_dma(s);
	}

	return 0;
}
#endif

static void audio_dma_os_callback(void *data, int intsts, int intrawsts)
{
	audio_state_t *state = data;
	audio_stream_t *s = state->output_stream;
	audio_buf_t *b = &s->buffers[s->dma_tail];

	DPRINTK("[%s] pending=%d dma_head=%d dma_tail=%d\n",
		s->id, s->pending_frags, s->dma_head, s->dma_tail);

#ifdef AUDIO_DEBUG_LOG_
	if (DEBUG_LOG_MAX > w_count) {
		unsigned long jiff = jiffies;
		(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
		(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
		(w_dbg_log+w_count)->fifo = 98;
		(w_dbg_log+w_count)->state = (uint32_t)b->dma_ref;
		f_jiff = jiff;
		w_count++;
	}
#endif

	if (s->buffers == NULL) {
		printk(KERN_INFO
		       "%s(): received DMA IRQ for non existent buffers!\n",
		       __FUNCTION__);
		return;
	} else if ((b->dma_ref != 0) && (--b->dma_ref == 0)
		   && (b->offset >= s->fragsize)) {
		/* This fragment is done */
		b->offset = 0;
		s->bytecount += s->fragsize;
		s->fragcount++;
#ifdef MP200_AUDIO_SINGLE_TRANSFER
#else
		memset(b->data, AUDIO_INIT_DATA, s->fragsize);
#endif
		if (++s->dma_tail >= s->nbfrags) {
			s->dma_tail = 0;
		}
		if (s->mapped == 0) {
			up(&s->sem);
		} else {
			s->pending_frags++;
		}
		if (waitqueue_active(&s->wq) != 0) {
			wake_up(&s->wq);
		}
	}

#ifdef MP200_AUDIO_SINGLE_TRANSFER
	audio_process_dma(s);
#else
	if (s->pending_frags == 0) {
		DPRINTK("[%s] DMA buffers empty!\n", s->id);
		audio_stop_dma(s);
		s->stopped = 0;
	} else {
		audio_process_dma(s);
	}
#endif
}

static void audio_dma_is_callback(void *data, int intsts, int intrawsts)
{
	audio_state_t *state = data;
	audio_stream_t *s = state->input_stream;
	audio_buf_t *b = &s->buffers[s->dma_tail];

	DPRINTK("[%s] pending=%d dma_head=%d dma_tail=%d\n",
		s->id, s->pending_frags, s->dma_head, s->dma_tail);

	if (s->buffers == NULL) {
		printk(KERN_INFO
		       "%s(): received DMA IRQ for non existent buffers!\n",
		       __FUNCTION__);
		return;
	} else if ((b->dma_ref != 0) && (--b->dma_ref == 0)
		   && (b->offset >= s->fragsize)) {
		/* This fragment is done */
		b->offset = 0;
		s->bytecount += s->fragsize;
		s->fragcount++;
		if (++s->dma_tail >= s->nbfrags) {
			s->dma_tail = 0;
		}
		if (s->mapped == 0) {
			up(&s->sem);
		} else {
			s->pending_frags++;
		}
		if (waitqueue_active(&s->wq) != 0) {
			wake_up(&s->wq);
		}
#ifdef MP200_AUDIO_SINGLE_TRANSFER
#else
	} else {
		if (++s->dma_tail >= s->nbfrags) {
			s->dma_tail = 0;
		}
		if (waitqueue_active(&s->wq) != 0) {
			wake_up(&s->wq);
		}
#endif
	}

#ifdef MP200_AUDIO_SINGLE_TRANSFER
	audio_process_dma(s);
#else
	if (s->pending_frags == 0) {
		/* buffer full */
		b = &s->buffers[s->dma_head];
		b->dma_ref++;
		if (++s->dma_head >= s->nbfrags) {
			s->dma_head = 0;
		}
	} else {
		audio_process_dma(s);
	}
#endif
}

#ifdef CONFIG_MP201
static irqreturn_t
audio_pcm_callback(int irq, void *dev_id, struct pt_regs *regs)
{
	audio_state_t *state = (audio_state_t *) dev_id;
	pcm_regs_t *pcm_regs = state->ctrls->regs;
	audio_stream_t *os = state->output_stream;
	audio_stream_t *is = state->input_stream;
	static irqreturn_t ret = IRQ_NONE;

	if(*pcm_regs->status & (PCM_INT_TX_FRE | PCM_INT_TX_URE | PCM_INT_TX_ORE)) {
		if (state->wr_ref && (NULL != os->buffers)) {

#ifdef MP200_AUDIO_PCM_NONFRE
			if(1 == mp200_dma_status(os->dma_channel)) {
#ifdef AUDIO_DEBUG_LOG
				if (DEBUG_LOG_MAX > w_count) {
					unsigned long jiff = jiffies;
					(w_dbg_log+w_count)->pending = (uint16_t)os->pending_frags;
					(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
					(w_dbg_log+w_count)->fifo = 99;
					(w_dbg_log+w_count)->state = (uint32_t)*pcm_regs->raw;
					f_jiff = jiff;
					w_count++;
				}
#endif
				*pcm_regs->clear = os->cl_bit;
				*pcm_regs->txrx_dis = os->en_bit;

				mp200_pcm_tasklet.data = (unsigned long)dev_id;
				tasklet_schedule(&mp200_pcm_tasklet);

				return IRQ_HANDLED;
			}
#endif
			spin_lock(&os->pcm_lock);
			os->pcm_intr++;
			os->pcm_status = *pcm_regs->status;
			DPRINTK("[%s] irq=%d os->pcm_intr=%d os->pcm_status=%08x\n", os->id, irq,
				os->pcm_intr, os->pcm_status);
			spin_unlock(&os->pcm_lock);
#ifdef MP200_AUDIO_SINGLE_TRANSFER_2
			os->stopped = 1;
#endif

			DLPRINTK("***** pcm_status 0x%08x *****\n", os->pcm_status);
#ifdef AUDIO_DEBUG_LOG
			if (DEBUG_LOG_MAX > w_count) {
				unsigned long jiff = jiffies;
				(w_dbg_log+w_count)->pending = (uint16_t)os->pending_frags;
				(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
				(w_dbg_log+w_count)->fifo = 99;
				(w_dbg_log+w_count)->state = (uint32_t)*pcm_regs->raw;
				f_jiff = jiff;
				w_count++;
			}
#endif

			if ((os->pcm_status & os->cl_bit) != 0) {
				*pcm_regs->clear = os->pcm_status & os->cl_bit;
			}

			*pcm_regs->txrx_dis = os->en_bit;

#ifdef MP200_AUDIO_PCM_NONERR
#else
			if (os->mapped == 0) {
				up(&os->sem);
			} else {
				os->pending_frags++;
			}
#endif
			if (waitqueue_active(&os->wq) != 0) {
				wake_up(&os->wq);
			}

			ret = IRQ_HANDLED;
		}
#ifdef AUDIO_DEBUG
		else {
			if (state->wr_ref) {
				printk(KERN_INFO
				       "%s(): received DMA IRQ for non existent buffers!(out)\n",
				       __FUNCTION__);
			}
		}
#endif
	}
	if(*pcm_regs->status & (PCM_INT_RX_FRE | PCM_INT_RX_URE | PCM_INT_RX_ORE)) {
		if (state->rd_ref && (NULL != is->buffers)) {

#ifdef MP200_AUDIO_PCM_NONFRE_
			if(1 == mp200_dma_status(is->dma_channel)) {
#ifdef AUDIO_DEBUG_LOG
				if (DEBUG_LOG_MAX > w_count) {
					unsigned long jiff = jiffies;
					(w_dbg_log+w_count)->pending = (uint16_t)is->pending_frags;
					(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
					(w_dbg_log+w_count)->fifo = 99;
					(w_dbg_log+w_count)->state = (uint32_t)*pcm_regs->raw;
					f_jiff = jiff;
					w_count++;
				}
#endif
				*pcm_regs->clear = is->cl_bit;
				*pcm_regs->txrx_dis = is->en_bit;

				mp200_pcm_tasklet.data = (unsigned long)dev_id;
				tasklet_schedule(&mp200_pcm_tasklet);

				return IRQ_HANDLED;
			}
#endif
			spin_lock(&is->pcm_lock);
			is->pcm_intr++;
			is->pcm_status = *pcm_regs->status;
			DPRINTK("[%s] irq=%d os->pcm_intr=%d os->pcm_status=%08x\n", is->id, irq,
				is->pcm_intr, is->pcm_status);
			spin_unlock(&is->pcm_lock);

			if ((is->pcm_status & is->cl_bit) != 0) {
				*pcm_regs->clear = is->pcm_status & is->cl_bit;
			}
			*pcm_regs->txrx_dis = is->en_bit;

			if (is->mapped == 0) {
				up(&is->sem);
			} else {
				is->pending_frags++;
			}
			if (waitqueue_active(&is->wq) != 0) {
				wake_up(&is->wq);
			}

			ret = IRQ_HANDLED;
		}
#ifdef AUDIO_DEBUG
		else {
			if (state->rd_ref) {
				printk(KERN_INFO
				       "%s(): received DMA IRQ for non existent buffers!(in)\n",
				       __FUNCTION__);
			}
		}
#endif
	}

	return ret;
}
#else
static irqreturn_t
audio_pcm_callback(int irq, void *dev_id, struct pt_regs *regs)
{
	audio_stream_t *s = (audio_stream_t *) dev_id;

	if ((*s->pcm_regs->status & s->irq_bit) == 0) {
		return IRQ_NONE;
	}

	if (s->buffers == NULL) {
		printk(KERN_INFO
		       "%s(): received DMA IRQ for non existent buffers!\n",
		       __FUNCTION__);
		return IRQ_NONE;
	}

	spin_lock(&s->pcm_lock);
	s->pcm_intr++;
	s->pcm_status = *s->pcm_regs->status;
	DPRINTK("[%s] irq=%d s->pcm_intr=%d s->pcm_status=%08x\n", s->id, irq,
		s->pcm_intr, s->pcm_status);
	spin_unlock(&s->pcm_lock);

	if ((s->pcm_status & s->cl_bit) != 0) {
		*s->pcm_regs->clear = s->pcm_status & s->cl_bit;
	}
	*s->pcm_regs->txrx_dis = s->en_bit;
	if (s->mapped == 0) {
		up(&s->sem);
	} else {
		s->pending_frags++;
	}
	if (waitqueue_active(&s->wq) != 0) {
		wake_up(&s->wq);
	}

	return IRQ_HANDLED;
}
#endif

/*
 * Buffer creation/destruction
 */
static void audio_discard_buf(audio_stream_t * s)
{
	/* ensure DMA isn't using those buffers */
	audio_reset(s);
	if (s->buffers != NULL) {
		dma_free_coherent(NULL, s->dmabufsize, s->buffers[0].data,
				  s->buffers[0].dma_addr);
		kfree(s->buffers);
		s->buffers = NULL;
	}
}

static int audio_setup_buf(audio_stream_t * s)
{
	audio_buf_t *b;
	dma_addr_t dmaphys = 0;
	int frag;
	int dmasize = 0;
	char *dmabuf = NULL;

	if (s->buffers != NULL) {
		return -EBUSY;
	}

	s->buffers = kmalloc(sizeof(audio_buf_t) * s->nbfrags, GFP_KERNEL);
	if (s->buffers == NULL) {
		goto err;
	}
	memset(s->buffers, 0, sizeof(audio_buf_t) * s->nbfrags);

	dmasize = s->nbfrags * s->fragsize;
#ifdef CONFIG_MP201
	dmabuf = dma_alloc_coherent(NULL, dmasize, &dmaphys, GFP_KERNEL);
	if (dmabuf == NULL) {
		kfree(s->buffers);
		s->buffers = NULL;
		goto err;
	}
#else
	dmabuf = dma_alloc_coherent(NULL, dmasize, &dmaphys, 0);
	if (dmabuf == NULL) {
		goto err;
	}
#endif
	s->dmabufsize = dmasize;
	s->dma_top = dmaphys;
	memset(dmabuf, AUDIO_INIT_DATA, dmasize);

	for (frag = 0; frag < s->nbfrags; frag++) {
		b = &s->buffers[frag];
		b->data = dmabuf + s->fragsize * frag;
		b->dma_addr = dmaphys + s->fragsize * frag;
		DPRINTK
		    ("buf %d:start %p dma %#08x dmabufsize %d fragsize %d\n",
		     frag, b->data, b->dma_addr, s->dmabufsize, s->fragsize);
	}

	s->usr_head = s->dma_head = s->dma_tail = 0;
	s->bytecount = 0;
	s->fragcount = 0;

	if (s->type == AUDIO_IN) {
		s->pending_frags = s->nbfrags;
		sema_init(&s->sem, 0);
	} else {
		s->pending_frags = 0;
		sema_init(&s->sem, s->nbfrags);
	}

	return 0;

      err:
	printk(KERN_INFO "%s(): unable to allocate audio memory\n",
	       __FUNCTION__);
#ifdef CONFIG_MP201
	audio_reset(s);
#else
	audio_discard_buf(s);
#endif

	return -ENOMEM;
}

/*
 * Driver interface functions
 */
static int audio_sync(audio_state_t * state)
{
	audio_stream_t *s = state->output_stream;
	audio_buf_t *b;
	unsigned long flags;
#ifdef MP201_SINGLE_SPECIAL_2
	unsigned int pcm_intr;
	unsigned int status;
#endif
#ifdef MP200_AUDIO_SINGLE_TRANSFER
	u_int shiftval = 0;
#else
	int syncWaitCond = 0;
#endif
	int ret = 0;
	DECLARE_WAITQUEUE(wait, current);

	if ((s == NULL) || (s->buffers == NULL) || (s->mapped != 0)) {
		return 0;
	}

	b = &s->buffers[s->usr_head];

#ifdef AUDIO_DEBUG_LOG
	if (DEBUG_LOG_MAX > w_count) {
		struct timeval tv;
		do_gettimeofday(&tv);
		w_usec = tv;
		(w_dbg_log+w_count)->time = atomic_read(&s->sem.count);
		(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
		(w_dbg_log+w_count)->fifo = 67;
		(w_dbg_log+w_count)->state = (uint32_t)atomic_read(&s->sem.count);
		w_count++;
		f_jiff = jiffies;
	}
#endif

#ifdef MP200_AUDIO_SINGLE_TRANSFER
	if ((b->dma_ref == 0) && (b->offset &= ~3))
	{
#ifdef MP201_SINGLE_SPECIAL_2
		int adj = 128;
#endif
		down(&s->sem);
#ifdef MP201_SINGLE_SPECIAL_2
		/* DMA Ƶư FIFO Ƥ뤿
		 * FIFO ʬ礭ޤ䤹
		 */
		if (adj > b->offset) {
			adj -= b->offset;
			memset(b->data + b->offset, AUDIO_INIT_DATA, adj);
			b->offset += adj;
		}
#endif
#ifdef AUDIO_DEBUG_LOG
		if (DEBUG_LOG_MAX > w_count) {
			(w_dbg_log+w_count)->time = atomic_read(&s->sem.count);
			(w_dbg_log+w_count)->pending = (uint16_t)b->dma_ref;
			(w_dbg_log+w_count)->fifo = 60;
			(w_dbg_log+w_count)->state = (uint32_t)b->offset;
			w_count++;
		}
#endif
		shiftval = s->fragsize - b->offset;
		b->offset = shiftval;
		b->dma_addr -= shiftval;
		b->data -= shiftval;
		s->bytecount -= shiftval;
		if (++s->usr_head >= s->nbfrags)
			s->usr_head = 0;
		local_irq_save(flags);
		s->pending_frags++;
#ifdef MP200_AUDIO_SINGLE_TRANSFER_3
		if (!s->active) {
			audio_process_dma(s);
		}
#else
		s->active = 1;
		audio_process_dma(s);
#endif
		local_irq_restore(flags);
	}

	/* Let's wait for all buffers to complete */
	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&s->wq, &wait);
	while ((s->pending_frags || (atomic_read(&s->sem.count) < s->nbfrags))
		&& !signal_pending(current))
	{
#ifdef MP201_SINGLE_SPECIAL_2
		/* SYNC Υ ޤ
		 * ٱˤĤä֥åǤФ
		 */
		spin_lock_irq(&s->pcm_lock);
		pcm_intr = s->pcm_intr;
		status = s->pcm_status;
		s->pcm_intr = 0;
		s->pcm_status = 0;
		spin_unlock_irq(&s->pcm_lock);
		if (pcm_intr != 0) {
			if ((status & s->irq_bit) != 0) {
				audio_stop_dma(s);
				s->stopped = 0;
			}
			if (s->pending_frags) {
				local_irq_save(flags);
				audio_process_dma(s);
				local_irq_restore(flags);
			}
			else {
				break;
			}
		}
#else
		if (s->pcm_intr != 0) {
#ifdef MP201_FIFO_STOP
#else
			ret = -EFAULT;
#endif
			break;
		}
		if (0 == (s->en_bit & *s->pcm_regs->txrx_mon)) {
			break;
		}
#endif
		schedule_timeout(1);
		set_current_state(TASK_INTERRUPTIBLE);
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&s->wq, &wait);

#ifdef AUDIO_DEBUG_LOG
	if (DEBUG_LOG_MAX > w_count) {
		unsigned int t;
		struct timeval tv;
		do_gettimeofday(&tv);
		t = tv.tv_sec - w_usec.tv_sec;
		t *= 1000000;
		if (tv.tv_usec < w_usec.tv_usec)
			t -= w_usec.tv_usec - tv.tv_usec;
		else
			t += tv.tv_usec - w_usec.tv_usec;
		(w_dbg_log+w_count)->time = t;
		w_usec = tv;
		(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
		(w_dbg_log+w_count)->fifo = 61;
		(w_dbg_log+w_count)->state = (uint32_t)atomic_read(&s->sem.count);
		w_count++;
		f_jiff = jiffies;
	}
#endif

#ifdef MP201_FIFO_STOP
	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&s->wq, &wait);
	while ((s->en_bit & *s->pcm_regs->txrx_mon) && !signal_pending(current))
	{
		local_irq_save(flags);
		if (s->pcm_intr != 0) {
			if ((s->pcm_status & s->irq_bit) != 0) {
				break;
			}
		}
		local_irq_restore(flags);
		schedule_timeout(1);
		set_current_state(TASK_INTERRUPTIBLE);
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&s->wq, &wait);
#endif

	/* undo the pointer hack above */
	if (shiftval) {
		local_irq_save(flags);
		b->dma_addr += shiftval;
		b->data += shiftval;
		/* ensure sane DMA code behavior if not yet processed */
		if (b->offset != 0)
			b->offset = s->fragsize;
		local_irq_restore(flags);
	}
#else
	if (b->offset != 0) {
		/* offset  0 ǤʤȤϡ1֥åʤǡ
		 * ĤäƤ뤿ᡢХåե DMAž򤫤롣
		 * Хåե֥åζԤĤ down()
		 */
		down(&s->sem);
		b->offset = 0;
		if (++s->usr_head >= s->nbfrags) {
			s->usr_head = 0;
		}
		local_irq_save(flags);
		s->pending_frags++;
		if ((mp200_dma_status(s->dma_channel) == 0)
		    && (s->usr_head != 1)) {
			memcpy(s->buffers[0].data, b->data, s->fragsize);
			memset(b->data, AUDIO_INIT_DATA, s->fragsize);
			s->dma_head = s->dma_tail = 0;
			s->usr_head = 1;
		}
		ret = audio_start_dma(s);
		local_irq_restore(flags);
		if (ret != 0) {
			audio_reset(s);
			return -EFAULT;
		}
	}

	/* Let's wait for all buffers to complete */
	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&s->wq, &wait);
	syncWaitCond = 1;
	while (syncWaitCond != 0) {
		if (s->pcm_intr != 0) {
			if ((s->pcm_status & s->irq_bit) != 0) {
				ret = -EFAULT;
				break;
			}
		}
		set_current_state(TASK_INTERRUPTIBLE);
		syncWaitCond =
		    (((s->pending_frags != 0)
		      || (atomic_read(&s->sem.count) < s->nbfrags))
		     && (signal_pending(current) == 0));
		if (syncWaitCond != 0) {
			schedule_timeout(1000);
		}
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&s->wq, &wait);
#endif

#ifdef AUDIO_DEBUG_LOG
	if (DEBUG_LOG_MAX > w_count) {
		unsigned int t;
		struct timeval tv;
		do_gettimeofday(&tv);
		t = tv.tv_sec - w_usec.tv_sec;
		t *= 1000000;
		if (tv.tv_usec < w_usec.tv_usec)
			t -= w_usec.tv_usec - tv.tv_usec;
		else
			t += tv.tv_usec - w_usec.tv_usec;
		(w_dbg_log+w_count)->time = t;
		w_usec = tv;
		(w_dbg_log+w_count)->pending = (uint16_t)atomic_read(&s->sem.count);
		(w_dbg_log+w_count)->fifo = 68;
		(w_dbg_log+w_count)->state = (uint32_t)ret;
		w_count++;
	}
#endif

	return ret;
}

static int audio_post(struct file *file)
{
	audio_state_t *state = file->private_data;
	audio_stream_t *s = state->output_stream;
	audio_buf_t *b;
	unsigned long flags;
#ifdef MP200_AUDIO_SINGLE_TRANSFER
#else
	int ret = 0;
#endif

	if (((file->f_mode & FMODE_WRITE) == 0) || (s->buffers == NULL)
	    || (s->mapped != 0)) {
		return 0;
	}

	b = &s->buffers[s->usr_head];
	if (b->offset != 0) {
		/* offset  0 ǤʤȤϡǡ뤿ᡢ
		 * Хåե DMAž򤫤롣
		 * Хåե֥åζԤĤ down()
		 */
		down(&s->sem);
		b->offset = 0;
		if (++s->usr_head >= s->nbfrags) {
			s->usr_head = 0;
		}
		local_irq_save(flags);
		s->pending_frags++;
#ifdef MP200_AUDIO_SINGLE_TRANSFER
#ifdef MP200_AUDIO_SINGLE_TRANSFER_3
		if (!s->active) {
			audio_process_dma(s);
		}
#else
		s->active = 1;
		audio_process_dma(s);
#endif
		local_irq_restore(flags);
#else
		if ((mp200_dma_status(s->dma_channel) == 0)
		    && (s->usr_head != 1)) {
			memcpy(s->buffers[0].data, b->data, s->fragsize);
			memset(b->data, AUDIO_INIT_DATA, s->fragsize);
			s->dma_head = s->dma_tail = 0;
			s->usr_head = 1;
		}
		ret = audio_start_dma(s);
		local_irq_restore(flags);
		if (ret != 0) {
			audio_reset(s);
			return -EFAULT;
		}
#endif
	}

	return 0;
}

static int audio_prime_rx(audio_state_t * state)
{
	audio_stream_t *is = state->input_stream;
	unsigned long flags;
#ifdef MP200_AUDIO_SINGLE_TRANSFER
#else
	int ret = 0;
#endif

	local_irq_save(flags);
	is->pending_frags = is->nbfrags;
	sema_init(&is->sem, 0);
#ifdef MP200_AUDIO_SINGLE_TRANSFER
	is->active = 1;
	audio_process_dma(is);
	local_irq_restore(flags);
#else
	ret = audio_start_dma(is);
	local_irq_restore(flags);
	if (ret != 0) {
		audio_reset(is);
		return -EFAULT;
	}
#endif

	return 0;
}

static void audio_read_wait(audio_state_t * state)
{
	audio_stream_t *is = state->input_stream;
	unsigned int nblock;
	DECLARE_WAITQUEUE(wait, current);

	nblock = is->nbfrags - 1;
	if (is->nbfrags > AUDIO_NBFRAGS_READWAIT) {
		nblock = AUDIO_NBFRAGS_READWAIT;
	}

	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&is->wq, &wait);
	while ((is->pending_frags >= (is->nbfrags - nblock))
	       && (signal_pending(current) == 0)) {
		if ((is->pcm_intr != 0)
		    && ((is->pcm_status & is->irq_bit) != 0)) {
			break;
		}
		schedule();
		set_current_state(TASK_INTERRUPTIBLE);
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&is->wq, &wait);
}

static int
audio_write(struct file *file, const char *buffer, size_t count, loff_t * ppos)
{
	audio_state_t *state = file->private_data;
	audio_stream_t *s = state->output_stream;
	audio_buf_t *b;
	unsigned long flags;
	unsigned int pcm_intr;
	unsigned int status;
	uint32_t *dmabuf;
	uint16_t const *usrbuf;
	uint16_t data;
	int chunksize;
	int ret = 0;
#ifdef MP200_AUDIO_SINGLE_TRANSFER
#else
	int res = 0;
#endif
	int i;
	const char *buffer0 = buffer;

	DPRINTK("count=%d\n", count);

	if (*ppos != file->f_pos) {
#ifdef AUDIO_DEBUG_LOG
		if (DEBUG_LOG_MAX > w_count) {
			(w_dbg_log+w_count)->time = 0;
			(w_dbg_log+w_count)->pending = (uint16_t)s->usr_head;
			(w_dbg_log+w_count)->fifo = 88;
			(w_dbg_log+w_count)->state = (uint32_t)(-ESPIPE);
			w_count++;
		}
#endif
		return -ESPIPE;
	}
#ifdef MP201_DSP_CPU
	if (state->int_mode != PCM_INT_CPU_MODE){
#ifdef AUDIO_DEBUG_LOG
		if (DEBUG_LOG_MAX > w_count) {
			(w_dbg_log+w_count)->time = 0;
			(w_dbg_log+w_count)->pending = (uint16_t)s->usr_head;
			(w_dbg_log+w_count)->fifo = 88;
			(w_dbg_log+w_count)->state = (uint32_t)(-EPERM);
			w_count++;
		}
#endif
		return -EPERM;
	}
#endif
	if (s->mapped != 0) {
#ifdef AUDIO_DEBUG_LOG
		if (DEBUG_LOG_MAX > w_count) {
			(w_dbg_log+w_count)->time = 0;
			(w_dbg_log+w_count)->pending = (uint16_t)s->usr_head;
			(w_dbg_log+w_count)->fifo = 88;
			(w_dbg_log+w_count)->state = (uint32_t)(-ENXIO);
			w_count++;
		}
#endif
		return -ENXIO;
	}

	if ((state->wr_ref & AUDIO_REF_PW) == 0) {
		/* ¿ open ޤϡޥκݤ
		 * audio_sync() ¾ΤΥޥե
		 */
		down(&state->sem);
		audio_sync(state);
		up(&state->sem);
		audio_reset(s);
		if (state->hw_init != NULL) {
			/* ¿ open ޤϡޥκݤ
			 * hw_init()wr_ref ¾ΤΥޥե
			 */
			down(&state->sem);
			state->hw_init(s, NULL);
			if ((state->wr_ref & AUDIO_REF_ERR) != 0) {
				state->wr_ref &= ~(AUDIO_REF_ERR);
				up(&state->sem);
#ifdef AUDIO_DEBUG_LOG
			if (DEBUG_LOG_MAX > w_count) {
				(w_dbg_log+w_count)->time = 0;
				(w_dbg_log+w_count)->pending = (uint16_t)s->usr_head;
				(w_dbg_log+w_count)->fifo = 88;
				(w_dbg_log+w_count)->state = (uint32_t)(-ENXIO);
				w_count++;
			}
#endif
				return -ENXIO;
			}
			up(&state->sem);
		}
	}

	if ((s->buffers == NULL) && (audio_setup_buf(s) != 0)) {
#ifdef AUDIO_DEBUG_LOG
		if (DEBUG_LOG_MAX > w_count) {
			(w_dbg_log+w_count)->time = 0;
			(w_dbg_log+w_count)->pending = (uint16_t)s->usr_head;
			(w_dbg_log+w_count)->fifo = 88;
			(w_dbg_log+w_count)->state = (uint32_t)(-ENOMEM);
			w_count++;
		}
#endif
		return -ENOMEM;
	}

#ifdef AUDIO_DEBUG_LOG_
	if (DEBUG_LOG_MAX > w_count) {
		unsigned int t;
		struct timeval tv;
		do_gettimeofday(&tv);
		t = tv.tv_sec - w_usec.tv_sec;
		t *= 1000000;
		if (tv.tv_usec < w_usec.tv_usec)
			t -= w_usec.tv_usec - tv.tv_usec;
		else
			t += tv.tv_usec - w_usec.tv_usec;
		(w_dbg_log+w_count)->time = t;
		w_usec = tv;
		(w_dbg_log+w_count)->pending = (uint16_t)atomic_read(&s->sem.count);
		(w_dbg_log+w_count)->fifo = 66;
		(w_dbg_log+w_count)->state = (uint32_t)count;
		w_count++;
	}
#endif
	while (count > 0) {
		b = &s->buffers[s->usr_head];

		/* Wait for a buffer to become free */
		if ((file->f_flags & O_NONBLOCK) != 0) {
			ret = -EAGAIN;
			if (down_trylock(&s->sem) != 0) {
				break;
			}
		} else {
			ret = -ERESTARTSYS;
			if (down_interruptible(&s->sem) != 0) {
				break;
			}
		}
#ifdef AUDIO_DEBUG_LOG_
		if (DEBUG_LOG_MAX > w_count) {
			unsigned int t;
			struct timeval tv;
			do_gettimeofday(&tv);
			t = tv.tv_sec - w_usec.tv_sec;
			t *= 1000000;
			if (tv.tv_usec < w_usec.tv_usec)
				t -= w_usec.tv_usec - tv.tv_usec;
			else
				t += tv.tv_usec - w_usec.tv_usec;
			(w_dbg_log+w_count)->time = t;
			w_usec = tv;
			(w_dbg_log+w_count)->pending = (uint16_t)atomic_read(&s->sem.count);
			(w_dbg_log+w_count)->fifo = 66;
			(w_dbg_log+w_count)->state = (uint32_t)count;
			w_count++;
		}
#endif

		spin_lock_irq(&s->pcm_lock);
		pcm_intr = s->pcm_intr;
		status = s->pcm_status;
		s->pcm_intr = 0;
		s->pcm_status = 0;
		spin_unlock_irq(&s->pcm_lock);
		if (pcm_intr != 0) {
			if ((status & s->irq_bit) != 0) {
#ifdef MP200_AUDIO_SINGLE_TRANSFER
				audio_stop_dma(s);
				s->stopped = 0;
#else
				audio_reset(s);
#endif
				DLPRINTK("pcm_intr %d status %d\n",
					pcm_intr, status);
				DPRINTK("pcm_intr %d status %d\n",
					pcm_intr, status);
#ifdef AUDIO_DEBUG_LOG_
				if (DEBUG_LOG_MAX > w_count) {
					unsigned int t;
					struct timeval tv;
					do_gettimeofday(&tv);
					t = tv.tv_sec - w_usec.tv_sec;
					t *= 1000000;
					if (tv.tv_usec < w_usec.tv_usec)
						t -= w_usec.tv_usec - tv.tv_usec;
					else
						t += tv.tv_usec - w_usec.tv_usec;
					(w_dbg_log+w_count)->time = t;
					w_usec = tv;
					(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
					(w_dbg_log+w_count)->fifo = 77;
					(w_dbg_log+w_count)->state = (uint32_t)status;
					w_count++;
				}
#endif
#ifdef MP200_AUDIO_PCM_NONERR
#else
				return -EFAULT;
#endif
			}
		}

		if (s->stereo != 0) {
			/* Feed the current buffer */
#ifdef MP200_AUDIO_SINGLE_TRANSFER_2_
			local_irq_save(flags);
			chunksize = s->fragsize - b->offset;
			if (0 == chunksize /* && (s->usr_head == s->dma_head)*/) {
#ifdef AUDIO_DEBUG_HIT
				printk(KERN_INFO "hit.3");
#endif
				up(&s->sem);
				local_irq_restore(flags);
				break;
			}
			local_irq_restore(flags);
#else
			chunksize = s->fragsize - b->offset;
#endif
			if (chunksize > count) {
				chunksize = count;
			}
			DPRINTK("write %d to %d\n", chunksize, s->usr_head);
			if (copy_from_user
			    (b->data + b->offset, buffer, chunksize) != 0) {
				up(&s->sem);
#ifdef AUDIO_DEBUG_LOG_
				if (DEBUG_LOG_MAX > w_count) {
					(w_dbg_log+w_count)->time = 0;
					(w_dbg_log+w_count)->pending = (uint16_t)s->usr_head;
					(w_dbg_log+w_count)->fifo = 88;
					(w_dbg_log+w_count)->state = (uint32_t)(-EFAULT);
					w_count++;
				}
#endif
				return -EFAULT;
			}

			buffer += chunksize;
			count -= chunksize;
			b->offset += chunksize;
		} else {
			/* Feed the current buffer */
			chunksize = (s->fragsize - b->offset) / 2;
			if (chunksize > count) {
				chunksize = count;
			}
			DPRINTK("write mono %d to %d\n", chunksize, s->usr_head);
			if (access_ok(VERIFY_READ, buffer, chunksize) == 0) {
				up(&s->sem);
				return -EFAULT;
			}

			usrbuf = (uint16_t const *)buffer;
			dmabuf = (uint32_t *) (b->data + b->offset);
			for (i = 0; i < chunksize; i += 2, dmabuf++, usrbuf++) {
				__get_user(data, usrbuf);
				*dmabuf =
				    ((uint32_t) data << 16) | (uint32_t) data;
			}

			buffer += chunksize;
			count -= chunksize;
			chunksize *= 2;
			b->offset += chunksize;
		}

		if (b->offset < s->fragsize) {
			up(&s->sem);
			break;
		}

#ifdef AUDIO_DEBUG_LOG_
		if (DEBUG_LOG_MAX > w_count) {
			unsigned int t;
			struct timeval tv;
			do_gettimeofday(&tv);
			t = tv.tv_sec - w_usec.tv_sec;
			t *= 1000000;
			if (tv.tv_usec < w_usec.tv_usec)
				t -= w_usec.tv_usec - tv.tv_usec;
			else
				t += tv.tv_usec - w_usec.tv_usec;
			(w_dbg_log+w_count)->time = t;
			w_usec = tv;
			(w_dbg_log+w_count)->pending = (uint16_t)atomic_read(&s->sem.count);
			(w_dbg_log+w_count)->fifo = 65;
			(w_dbg_log+w_count)->state = (uint32_t)count;
			w_count++;
		}
#endif

#ifdef MP200_AUDIO_SINGLE_TRANSFER
		/* Update pointers and send current fragment to DMA */
		local_irq_save(flags);
		b->offset = 0;
		if (++s->usr_head >= s->nbfrags) {
			s->usr_head = 0;
		}
		s->pending_frags++;
#ifdef MP200_AUDIO_SINGLE_TRANSFER_3
		if (!s->active) {
			audio_process_dma(s);
		}
#else
		s->active = 1;
		audio_process_dma(s);
#endif
		local_irq_restore(flags);
#else
		/* Update pointers and send current fragment to DMA */
		b->offset = 0;
		if (++s->usr_head >= s->nbfrags) {
			s->usr_head = 0;
		}

		local_irq_save(flags);
		s->pending_frags++;
		if ((mp200_dma_status(s->dma_channel) == 0)
		    && (s->usr_head != 1)) {
			DPRINTK("usr_head %d -> 0\n", s->usr_head);
			memcpy(s->buffers[0].data, b->data, s->fragsize);
			memset(b->data, AUDIO_INIT_DATA, s->fragsize);
			s->dma_head = s->dma_tail = 0;
			s->usr_head = 1;
		}
		res = audio_start_dma(s);
		local_irq_restore(flags);
		if (res != 0) {
			audio_reset(s);
			return -EFAULT;
		}
#endif
	}

	if ((buffer - buffer0) != 0) {
		ret = buffer - buffer0;
	}
#ifdef AUDIO_DEBUG_LOG_
	if (DEBUG_LOG_MAX > w_count) {
		unsigned int t;
		struct timeval tv;
		do_gettimeofday(&tv);
		t = tv.tv_sec - w_usec.tv_sec;
		t *= 1000000;
		if (tv.tv_usec < w_usec.tv_usec)
			t -= w_usec.tv_usec - tv.tv_usec;
		else
			t += tv.tv_usec - w_usec.tv_usec;
		(w_dbg_log+w_count)->time = t;
		w_usec = tv;
		(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
		(w_dbg_log+w_count)->fifo = 88;
		(w_dbg_log+w_count)->state = (uint32_t)ret;
		w_count++;
	}
#endif

	DPRINTK("return=%d\n", ret);
	return ret;
}

static int
audio_read(struct file *file, char *buffer, size_t count, loff_t * ppos)
{
	audio_state_t *state = file->private_data;
	audio_stream_t *s = state->input_stream;
	audio_buf_t *b;
	unsigned long flags;
	unsigned int pcm_intr;
	unsigned int status;
	uint32_t *dmabuf;
	uint16_t const *usrbuf;
	uint16_t data;
	int chunksize;
	int ret = 0;
	int i;
	char *buffer0 = buffer;

	DPRINTK("count=%d\n", count);

	if (*ppos != file->f_pos) {
		return -ESPIPE;
	}
#ifdef MP201_DSP_CPU
	if (state->int_mode != PCM_INT_CPU_MODE){
		return -EPERM;
	}
#endif
	if (s->mapped != 0) {
		return -ENXIO;
	}

	if ((state->rd_ref & AUDIO_REF_PW) == 0) {
		if (((state->wr_ref & AUDIO_REF_EN) != 0) &&
		    ((state->wr_ref & AUDIO_REF_PW) == 0)) {
			/* ¿ open ޤϡޥκݤ
			 * audio_sync() ¾ΤΥޥե
			 */
			down(&state->sem);
			audio_sync(state);
			up(&state->sem);
		}
		audio_reset(s);
		if (state->hw_init != NULL) {
			/* ¿ open ޤϡޥκݤ
			 * hw_init()rd_ref ¾ΤΥޥե
			 */
			down(&state->sem);
			state->hw_init(NULL, s);
			if ((state->rd_ref & AUDIO_REF_ERR) != 0) {
				state->rd_ref &= ~(AUDIO_REF_ERR);
				up(&state->sem);
				return -ENXIO;
			}
			up(&state->sem);
		}
	}

	if (s->active == 0) {
		if ((s->buffers == NULL) && (audio_setup_buf(s) != 0)) {
			return -ENOMEM;
		}
		if (audio_prime_rx(state) != 0) {
			return -EFAULT;
		}
		if ((file->f_flags & O_NONBLOCK) != 0) {
			return -EAGAIN;
		}
		audio_read_wait(state);
	}

	while (count > 0) {
		b = &s->buffers[s->usr_head];

		/* Wait for a buffer to become full */
		if ((file->f_flags & O_NONBLOCK) != 0) {
			ret = -EAGAIN;
			if (down_trylock(&s->sem) != 0) {
				break;
			}
		} else {
			ret = -ERESTARTSYS;
			if (down_interruptible(&s->sem) != 0) {
				break;
			}
		}

		if ((state->rd_ref & AUDIO_REF_PW) == 0) {
			ret = 0;
			break;
		}

		spin_lock_irq(&s->pcm_lock);
		pcm_intr = s->pcm_intr;
		status = s->pcm_status;
		s->pcm_intr = 0;
		s->pcm_status = 0;
		spin_unlock_irq(&s->pcm_lock);
		if (pcm_intr != 0) {
			if ((status & s->irq_bit) != 0) {
#ifdef MP200_AUDIO_SINGLE_TRANSFER
				audio_stop_dma(s);
				s->stopped = 0;
#else
				audio_reset(s);
#endif
				DPRINTK("pcm_intr %d status %d\n",
					pcm_intr, status);
#ifdef MP200_AUDIO_PCM_NONERR
#else
				return -EFAULT;
#endif
			}
		}

		if ((s->pending_frags == 0) && (s->usr_head != s->dma_head)) {
			DPRINTK("usr_head %d -> %d\n", s->usr_head,
				s->dma_head);
			b->offset = 0;
			s->usr_head = s->dma_head;
			b = &s->buffers[s->usr_head];
		}

		if (s->stereo != 0) {
			/* Grab data from the current buffer */
			chunksize = s->fragsize - b->offset;
			if (chunksize > count) {
				chunksize = count;
			}
			DPRINTK("read %d from %d\n", chunksize, s->usr_head);
			if (copy_to_user(buffer, b->data + b->offset, chunksize)
			    != 0) {
				up(&s->sem);
				return -EFAULT;
			}
			buffer += chunksize;
			count -= chunksize;
			b->offset += chunksize;
		} else {
			/* Feed the current buffer */
			chunksize = (s->fragsize - b->offset) / 2;
			if (chunksize > count) {
				chunksize = count;
			}
			DPRINTK("read %d from %d\n", chunksize, s->usr_head);
			if (access_ok(VERIFY_READ, buffer, chunksize) == 0) {
				up(&s->sem);
				return -EFAULT;
			}

			usrbuf = (uint16_t const *)buffer;
			dmabuf = (uint32_t *) (b->data + b->offset);
			for (i = 0; i < chunksize; i += 2, dmabuf++, usrbuf++) {
				data = (uint16_t) (*dmabuf);
				__put_user(data, usrbuf);
			}

			buffer += chunksize;
			count -= chunksize;
			chunksize *= 2;
			b->offset += chunksize;
		}

		if (b->offset < s->fragsize) {
			up(&s->sem);
			break;
		}

		/* Update pointers and return current fragment to DMA */
		b->offset = 0;
		if (++s->usr_head >= s->nbfrags) {
			s->usr_head = 0;
		}
		local_irq_save(flags);
		s->pending_frags++;
		local_irq_restore(flags);
	}

	if ((buffer - buffer0) != 0) {
		ret = buffer - buffer0;
	}

	DPRINTK("return=%d\n", ret);
	return ret;
}

static int audio_mmap(struct file *file, struct vm_area_struct *vma)
{
	audio_state_t *state = file->private_data;
	audio_stream_t *s;
	audio_buf_t *buf;
	unsigned long size;
	unsigned long vma_addr;
	int ret;

#ifdef MP201_DSP_CPU
	if (state->int_mode != PCM_INT_CPU_MODE){
		return -EPERM;
	}
#endif
	if (vma->vm_pgoff != 0) {
		return -EINVAL;
	}

	if ((vma->vm_flags & VM_WRITE) != 0) {
		if (state->wr_ref == 0) {
			return -EINVAL;
		}
		s = state->output_stream;
	} else if ((vma->vm_flags & VM_READ) != 0) {
		if (state->rd_ref == 0) {
			return -EINVAL;
		}
		s = state->input_stream;
	} else {
		return -EINVAL;
	}

	if (s->stereo == 0) {
		return -ENXIO;
	}

	if (s->mapped != 0) {
		return -EINVAL;
	}
	size = vma->vm_end - vma->vm_start;
	if (size != (s->fragsize * s->nbfrags)) {
		return -EINVAL;
	}
	if ((s->buffers == NULL) && (audio_setup_buf(s) != 0)) {
		return -ENOMEM;
	}
	vma_addr = vma->vm_start;

	buf = s->buffers;

	ret =
	    remap_pfn_range(vma, vma_addr, buf->dma_addr >> PAGE_SHIFT,
			    s->dmabufsize, vma->vm_page_prot);
	if (ret != 0) {
		return ret;
	}
	vma_addr += s->dmabufsize;
	s->mapped = 1;

	return 0;
}

static unsigned int
audio_poll(struct file *file, struct poll_table_struct *wait)
{
	audio_state_t *state = file->private_data;
	audio_stream_t *is = state->input_stream;
	audio_stream_t *os = state->output_stream;
	unsigned int mask = 0;
	unsigned int pcm_intr;
	unsigned int status;

	DPRINTK("mode=%s%s\n",
		(file->f_mode & FMODE_READ) ? "r" : "",
		(file->f_mode & FMODE_WRITE) ? "w" : "");

#ifdef MP201_DSP_CPU
	if (state->int_mode != PCM_INT_CPU_MODE){
		return -EPERM;
	}
#endif

	if ((file->f_mode & FMODE_READ) != 0) {
		if (((state->rd_ref & AUDIO_REF_PW) == 0) && (is->mapped == 0)) {
			if (((state->wr_ref & AUDIO_REF_EN) != 0) &&
			    ((state->wr_ref & AUDIO_REF_PW) == 0)) {
				/* ¿ open ޤϡޥκݤ
				 * audio_sync() ¾ΤΥޥե
				 */
				down(&state->sem);
				audio_sync(state);
				up(&state->sem);
			}
			audio_reset(is);
			if (state->hw_init != NULL) {
				/* ¿ open ޤϡޥκݤ
				 * hw_init()rd_ref ¾ΤΥޥե
				 */
				down(&state->sem);
				state->hw_init(NULL, is);
				if ((state->rd_ref & AUDIO_REF_ERR) != 0) {
					state->rd_ref &= ~(AUDIO_REF_ERR);
					up(&state->sem);
					return -ENXIO;
				}
				up(&state->sem);
			}
		}

		/* Start audio input if not already active */
		if (is->active == 0) {
			if ((is->buffers == NULL) && (audio_setup_buf(is) != 0)) {
				return -ENOMEM;
			}
			if (audio_prime_rx(state) != 0) {
				return -EFAULT;
			}
			audio_read_wait(state);
		}
		poll_wait(file, &is->wq, wait);
	}

	if ((file->f_mode & FMODE_WRITE) != 0) {
		if ((os->buffers == NULL) && (audio_setup_buf(os) != 0)) {
			return -ENOMEM;
		}
		poll_wait(file, &os->wq, wait);
	}

	if ((file->f_mode & FMODE_READ) != 0) {
		spin_lock_irq(&is->pcm_lock);
		pcm_intr = is->pcm_intr;
		status = is->pcm_status;
		spin_unlock_irq(&is->pcm_lock);
		if (pcm_intr != 0) {
			if ((status & is->irq_bit) != 0) {
#ifdef MP200_AUDIO_SINGLE_TRANSFER_4
#else
				audio_reset(is);
#endif
				return POLLERR;
			}
		}
		if (((is->mapped != 0) && (is->fragcount > 0)) ||
		    ((is->mapped == 0) && (atomic_read(&is->sem.count) > 0))) {
			mask |= POLLIN | POLLRDNORM;
		}
	}

	if ((file->f_mode & FMODE_WRITE) != 0) {
		spin_lock_irq(&os->pcm_lock);
		pcm_intr = os->pcm_intr;
		status = os->pcm_status;
		spin_unlock_irq(&os->pcm_lock);
		if (pcm_intr != 0) {
			if ((status & os->irq_bit) != 0) {
#ifdef AUDIO_DEBUG_LOG
				if (DEBUG_LOG_MAX > w_count) {
					unsigned long jiff = jiffies;
					(w_dbg_log+w_count)->pending = (uint16_t)os->pending_frags;
					(w_dbg_log+w_count)->time = (uint16_t)(jiff - f_jiff);
					(w_dbg_log+w_count)->fifo = 91;
					(w_dbg_log+w_count)->state = (uint32_t)status;
					f_jiff = jiff;
					w_count++;
				}
#endif
#ifdef MP200_AUDIO_SINGLE_TRANSFER_4
#else
				audio_reset(os);
#endif
				return POLLERR;
			}
		}
		if (((os->mapped != 0) && (os->fragcount > 0)) ||
		    ((os->mapped == 0) && (atomic_read(&os->sem.count) > 0))) {
			mask |= POLLOUT | POLLWRNORM;
		}
	}

	DPRINTK("returned mask of %s%s\n",
		(mask & POLLIN) ? "r" : "", (mask & POLLOUT) ? "w" : "");

	return mask;
}

static loff_t audio_llseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}

static int audio_set_fragments(audio_stream_t * s, int val)
{
	if (s->active != 0) {
		return -EBUSY;
	}
	if (s->buffers != NULL) {
		audio_discard_buf(s);
	}
	s->nbfrags = (val >> 16) & 0x7FFF;
	val &= 0xffff;
	if (val < 4) {
		val = 4;
	}
	if (val > 15) {
		val = 15;
	}
	s->fragsize = 1 << val;
	if (s->nbfrags < 2) {
		s->nbfrags = 2;
	}
	if ((s->nbfrags * s->fragsize) > (128 * 1024)) {
		s->nbfrags = (128 * 1024) / s->fragsize;
	}
	if (audio_setup_buf(s) != 0) {
		return -ENOMEM;
	}

	return (s->nbfrags << 16) | val;
}

int audio_pcm_set_ctrl(audio_state_t * s)
{
	pcm_ctrl_t *sett = s->ctrls->sett;
	pcm_regs_t *regs = s->ctrls->regs;
	unsigned int func_val = 0;
	unsigned int func_val2 = 0;
	unsigned int cycle_val = 0;
	unsigned int cycle2_val = 0;

	if (*regs->txrx_mon != 0) {
		return -EFAULT;
	}

	/* mode_sel : Operation mode */
	switch (sett->func.mode_sel) {
	case PCM_MODE_0:	/* mode 0 */
		func_val |= PCM_FUNC_MODE0;
		break;
	case PCM_MODE_1:	/* mode 1 */
		func_val |= PCM_FUNC_MODE1;
		break;
	case PCM_MODE_2:	/* mode 2 (I2S Format) */
		func_val |= PCM_FUNC_MODE2;
		break;
	case PCM_MODE_3:	/* mode 3 (MSB First)  */
		func_val |= PCM_FUNC_MODE3;
		break;
	case PCM_MODE_4:	/* mode 4 (LSB First)  */
		func_val = PCM_FUNC_MODE4;
		break;
	case PCM_MODE_5:	/* mode 5 (Multi Channel Mode) */
		func_val |= PCM_FUNC_MODE5;
		break;
	case PCM_MODE_6:	/* mode 6 (Multi Channel Mode) */
		func_val |= PCM_FUNC_MODE6;
		break;
	default:
		return -EINVAL;
	}

	/* m_s : Master/Slave */
	switch (sett->func.m_s) {
	case PCM_MASTER_MODE:	/* Master */
		func_val2 |= PCM_FUNC_MASTER;
		break;
	case PCM_SLAVE_MODE:	/* Slave  */
		func_val2 |= PCM_FUNC_SLAVE;
		break;
	default:
		return -EINVAL;
	}

	/* tx_tim : Transmission timing */
	switch (sett->func.tx_tim) {
	case PCM_TX_30_WORD:	/* 30word */
		func_val |= PCM_FUNC_30WORD;
		break;
	case PCM_TX_16_WORD:	/* 16word */
		func_val |= PCM_FUNC_16WORD;
		break;
	case PCM_TX_08_WORD:	/*  8word */
		func_val |= PCM_FUNC_08WORD;
		break;
	case PCM_TX_04_WORD:	/*  4word */
		func_val |= PCM_FUNC_04WORD;
		break;
	default:
		return -EINVAL;
	}

	switch (sett->func.mode_sel) {
	case PCM_MODE_0:	/* mode 0 */
	case PCM_MODE_1:	/* mode 1 */
	case PCM_MODE_2:	/* mode 2 (I2S Format) */
	case PCM_MODE_3:	/* mode 3 (MSB First)  */
	case PCM_MODE_4:	/* mode 4 (LSB First)  */
		/* cyc_val : Frame length */
		if ((0x3f < sett->cyc.cyc_val) || (sett->cyc.cyc_val < 0x07)) {
			return -EINVAL;
		}
		if ((sett->cyc.cyc_val < sett->cyc.sib)
		    || (sett->cyc.cyc_val < sett->cyc.sob)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.cyc_val;

		/* sib : Data bit length (PMx_SI) */
		if ((0x1f < sett->cyc.sib) || (sett->cyc.sib < 0x07)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.sib << 8;

		/* rx_pd : Data padding (RX) */
		if ((sett->cyc.rx_pd != PCM_PADDING_OFF)
		    && (sett->cyc.rx_pd != PCM_PADDING_ON)) {
			return -EINVAL;
		}
		if (sett->cyc.rx_pd == PCM_PADDING_ON) {
			if ((sett->cyc.sib != 0x0f)
			    && (sett->cyc.sib != 0x07)) {
				return -EINVAL;
			}
			cycle_val |= sett->cyc.rx_pd << 15;
		}

		/* sob : Data bit length (PMx_SO) */
		if ((0x1f < sett->cyc.sob) || (sett->cyc.sob < 0x07)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.sob << 16;

		/* tx_pd : Data padding (TX) */
		if ((sett->cyc.tx_pd != PCM_PADDING_OFF)
		    && (sett->cyc.tx_pd != PCM_PADDING_ON)) {
			return -EINVAL;
		}
		if (sett->cyc.tx_pd == PCM_PADDING_ON) {
			if ((sett->cyc.sob != 0x0f)
			    && (sett->cyc.sob != 0x07)) {
				return -EINVAL;
			}
			cycle_val |= sett->cyc.tx_pd << 23;
		}
		break;
	case PCM_MODE_5:	/* mode 5 (Multi Channel Mode) */
	case PCM_MODE_6:	/* mode 6 (Multi Channel Mode) */
		/* cyc_val : Frame length */
		if (0x80 < sett->cyc.cyc_val) {
			return -EINVAL;
		}
		if ((sett->cyc.cyc_val == 0) && (sett->cyc2.cyc_val2 == 0)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.cyc_val;

		/* sib : Data bit length (PMx_SI) */
		if ((0x1f < sett->cyc.sib) || (sett->cyc.sib < 0x07)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.sib << 8;

		/* rx_pd : Data padding (RX) */
		if (sett->cyc.rx_pd != PCM_PADDING_OFF) {
			return -EINVAL;
		}

		/* sob : Data bit length (PMx_SO) */
		if ((0x1f < sett->cyc.sob) || (sett->cyc.sob < 0x07)) {
			return -EINVAL;
		}
		cycle_val |= sett->cyc.sob << 16;

		/* tx_pd : Data padding (TX) */
		if (sett->cyc.tx_pd != PCM_PADDING_OFF) {
			return -EINVAL;
		}

		/* cyc_val2 : Frame length */
		if (0x80 < sett->cyc2.cyc_val2) {
			return -EINVAL;
		}
		cycle2_val |= sett->cyc2.cyc_val2;

		/* sib2 : Data bit length (PMx_SI) */
		if ((0x1f < sett->cyc2.sib2) || (sett->cyc2.sib2 < 0x07)) {
			return -EINVAL;
		}
		cycle2_val |= sett->cyc2.sib2 << 8;

		/* sob2 : Data bit length (PMx_SO) */
		if ((0x1f < sett->cyc2.sob2) || (sett->cyc2.sob2 < 0x07)) {
			return -EINVAL;
		}
		cycle2_val |= sett->cyc2.sob2 << 16;
		break;
	default:
		return -EINVAL;
	}

	/* set PMx_FUNC_SEL */
	*regs->func_sel = func_val;
	*regs->func_sel |= func_val2;

	/* set PMx_CYCLE1 */
	*regs->cycle = cycle_val;

	if ((sett->func.mode_sel == PCM_MODE_5)
	    || (sett->func.mode_sel == PCM_MODE_6)) {
		/* set PMx_CYCLE2 */
		*regs->cycle2 = cycle2_val;
	}

	return 0;
}

#ifdef MP201_DSP_CPU
#define DMA_M2P_LCH4_MODE	(IO_ADDRESS (MP200_DMAC_BASE) + 0x1450)
/* DSP <-> CPU ڤؤߤ뤦 pcm int ˤ ¿ open Բ */
static unsigned int mp200_audio_busy = 0;

int audio_dsp_set_ctrl(int val, audio_state_t * s)
{
	unsigned long flags;

	if (s->int_mode == PCM_INT_DSP_MODE) {
		switch (val) {
		case AFMT_S32_LE:
			local_irq_save(flags);
			outl(inl(DMA_M2P_LCH4_MODE) & 0xffffffcf, DMA_M2P_LCH4_MODE);
			local_irq_restore(flags);
			break;
		case AFMT_S16_LE:
		default:
			local_irq_save(flags);
			outl((inl(DMA_M2P_LCH4_MODE) & 0xffffffcf) | 0x10, DMA_M2P_LCH4_MODE);
			local_irq_restore(flags);
			val = AFMT_S16_LE;
			break;
		}
	}

	return 0;
}

static int audio_free_irq(struct inode *inode, struct file *file,
				   audio_state_t * state)
{
	audio_stream_t *os =
	    (file->f_mode & FMODE_WRITE) ? state->output_stream : NULL;
	audio_stream_t *is =
	    (file->f_mode & FMODE_READ) ? state->input_stream : NULL;
	pcm_regs_t *regs = state->ctrls->regs;

	if ((file->f_mode & FMODE_READ) != 0) {
		mp200_free_dma(is->dma_channel);
		is->dma_regs = NULL;
		*regs->enclr =
		    PCM_INT_RX_FRE | PCM_INT_RX_URE | PCM_INT_RX_ORE |
		    PCM_INT_RX_REN;
		*regs->clear = is->cl_bit;
	}

	if ((file->f_mode & FMODE_WRITE) != 0) {
		mp200_free_dma(os->dma_channel);
		os->dma_regs = NULL;
		*regs->enclr =
		    PCM_INT_TX_FRE | PCM_INT_TX_URE | PCM_INT_TX_ORE |
		    PCM_INT_TX_WEN;
		*regs->clear = os->cl_bit;
	}

	if ((0 == state->wr_ref) && (0 == state->rd_ref)) {
		DPRINTK("free_irq pcm\n");
		free_irq(state->pcm_irq, (void *)state);
	}

	return 0;
}

static int audio_request_irq(struct inode *inode, struct file *file,
				   audio_state_t * state)
{
	audio_stream_t *os =
	    (file->f_mode & FMODE_WRITE) ? state->output_stream : NULL;
	audio_stream_t *is =
	    (file->f_mode & FMODE_READ) ? state->input_stream : NULL;
	pcm_regs_t *regs = state->ctrls->regs;
	int err;

	err = -EINVAL;

	/* request DMA channels */
	if ((file->f_mode & FMODE_WRITE) != 0) {
		err =
		    mp200_request_dma(os->dma_channel, os->id,
				      (dma_callback_t) audio_dma_os_callback,
				      state, &os->dma_regs);
		if (err != 0) {
			goto out_1;
		}
	}
	if ((file->f_mode & FMODE_READ) != 0) {
		err =
		    mp200_request_dma(is->dma_channel, is->id,
				      (dma_callback_t) audio_dma_is_callback,
				      state, &is->dma_regs);
		if (err != 0) {
			goto out_2;
		}
	}

	/* request PCM interrupt */
	if ((0 == state->wr_ref) && (0 == state->rd_ref)) {
		*regs->enclr =
		    PCM_INT_TX_FRE | PCM_INT_TX_URE | PCM_INT_TX_ORE |
		    PCM_INT_TX_WEN |
		    PCM_INT_RX_FRE | PCM_INT_RX_URE | PCM_INT_RX_ORE |
		    PCM_INT_RX_REN;
		*regs->clear = PCM_INT_TX_FRE | PCM_INT_TX_URE |
			       PCM_INT_RX_FRE;
		*regs->txrx_dis = PCM_TX_EN | PCM_RX_EN;
		err =
		    request_irq(state->pcm_irq, audio_pcm_callback, SA_INTERRUPT,
				"audio io", (void *)state);
		if (err != 0) {
			goto out_3;
		}
		DLPRINTK("***** request_irq *****\n");
	}
	if ((file->f_mode & FMODE_WRITE) != 0) {
		os->pcm_intr = 0;
		os->pcm_status = 0;
		*regs->enset = os->irq_bit;
	}
	if ((file->f_mode & FMODE_READ) != 0) {
		is->pcm_intr = 0;
		is->pcm_status = 0;
		*regs->enset = is->irq_bit;
	}

	return 0;

      out_3:
	if ((file->f_mode & FMODE_READ) != 0) {
		mp200_free_dma(is->dma_channel);
	}
      out_2:
	if ((file->f_mode & FMODE_WRITE) != 0) {
		mp200_free_dma(os->dma_channel);
	}
      out_1:

	return err;
}
#endif

static int
audio_ioctl(struct inode *inode, struct file *file, uint cmd, ulong arg)
{
	audio_state_t *state = file->private_data;
	audio_stream_t *os = state->output_stream;
	audio_stream_t *is = state->input_stream;
	audio_stream_t *s;
	audio_buf_t *b;
	audio_buf_info buf_inf = { 0, };
	count_info inf = { 0, };
	pcm_ctrl_t sett;
	pcm_ctrl_t sett_save;
	unsigned long flags;
	int val;
	int ret = 0;
	int bytecount;
	int offset;

	DPRINTK("cmd=0x%08x\n", cmd);

#ifdef AUDIO_DEBUG_LOG_
	if (DEBUG_LOG_MAX > w_count) {
		unsigned int t;
		struct timeval tv;
		do_gettimeofday(&tv);
		t = tv.tv_sec - w_usec.tv_sec;
		t *= 1000000;
		if (tv.tv_usec < w_usec.tv_usec)
			t -= w_usec.tv_usec - tv.tv_usec;
		else
			t += tv.tv_usec - w_usec.tv_usec;
		(w_dbg_log+w_count)->time = t;
		w_usec = tv;
		(w_dbg_log+w_count)->pending = (uint16_t)cmd;
		(w_dbg_log+w_count)->fifo = 62;
		(w_dbg_log+w_count)->state = (uint32_t)cmd;
		w_count++;
	}
#endif

	/* dispatch based on command */
	switch (cmd) {
	case OSS_GETVERSION:
		return put_user(SOUND_VERSION, (int *)arg);

	case SNDCTL_DSP_GETBLKSIZE:
		DPRINTK("SNDCTL_DSP_GETBLKSIZE cmd=0x%08x\n", cmd);
		if ((file->f_mode & FMODE_WRITE) != 0) {
			return put_user(os->fragsize, (int *)arg);
		} else {
			return put_user(is->fragsize, (int *)arg);
		}

	case SNDCTL_DSP_GETCAPS:
		DPRINTK("SNDCTL_DSP_GETCAPS cmd=0x%08x\n", cmd);
		val = DSP_CAP_REALTIME | DSP_CAP_TRIGGER | DSP_CAP_MMAP;
		if ((is != NULL) && (os != NULL)) {
			val |= DSP_CAP_DUPLEX;
		}
		return put_user(val, (int *)arg);

	case SNDCTL_DSP_SETFRAGMENT:
		DPRINTK("SNDCTL_DSP_SETFRAGMENT cmd=0x%08x\n", cmd);
		if (get_user(val, (long *)arg) != 0) {
			return -EFAULT;
		}
		if ((file->f_mode & FMODE_READ) != 0) {
			ret = audio_set_fragments(is, val);
			if (ret < 0) {
				return ret;
			}
			ret = put_user(ret, (int *)arg);
			if (ret != 0) {
				return ret;
			}
		}
		if ((file->f_mode & FMODE_WRITE) != 0) {
			ret = audio_set_fragments(os, val);
			if (ret < 0) {
				return ret;
			}
			ret = put_user(ret, (int *)arg);
			if (ret != 0) {
				return ret;
			}
		}
		return 0;

	case SNDCTL_DSP_SYNC:
		DPRINTK("SNDCTL_DSP_SYNC cmd=0x%08x\n", cmd);
		if ((file->f_mode & FMODE_WRITE) != 0) {
#ifdef MP201_DSP_CPU
			if (state->int_mode != PCM_INT_CPU_MODE){
				return -EPERM;
			}
#endif
			return audio_sync(state);
		}
		return 0;

	case SNDCTL_DSP_SETDUPLEX:
		DPRINTK("SNDCTL_DSP_SETDUPLEX cmd=0x%08x\n", cmd);
		return 0;

	case SNDCTL_DSP_POST:
		DPRINTK("SNDCTL_DSP_POST cmd=0x%08x\n", cmd);
#ifdef MP201_DSP_CPU
		if (state->int_mode != PCM_INT_CPU_MODE){
			return -EPERM;
		}
#endif
		return audio_post(file);

	case SNDCTL_DSP_GETTRIGGER:
		DPRINTK("SNDCTL_DSP_GETTRIGGER cmd=0x%08x\n", cmd);
		val = 0;
		if (((file->f_mode & FMODE_READ) != 0) && (is->active != 0)
		    && (is->stopped == 0)) {
			val |= PCM_ENABLE_INPUT;
		}
		if (((file->f_mode & FMODE_WRITE) != 0) && (os->active != 0)
		    && (os->stopped == 0)) {
			val |= PCM_ENABLE_OUTPUT;
		}
		return put_user(val, (int *)arg);

	case SNDCTL_DSP_SETTRIGGER:
		DPRINTK("SNDCTL_DSP_SETTRIGGER cmd=0x%08x\n", cmd);
#ifdef MP201_DSP_CPU
		if (state->int_mode != PCM_INT_CPU_MODE){
			return -EPERM;
		}
#endif
		if (get_user(val, (int *)arg) != 0) {
			return -EFAULT;
		}
		if ((file->f_mode & FMODE_READ) != 0) {
			if ((val & PCM_ENABLE_INPUT) != 0) {
				if ((state->rd_ref & AUDIO_REF_PW) == 0) {
					audio_reset(is);
					if (state->hw_init != NULL) {
						/* ¿ open ޤϡ
						 * ޥκݤ
						 * hw_init()rd_ref 
						 * ¾ΤΥޥե
						 */
						down(&state->sem);
						state->hw_init(NULL, is);
						if ((state->
						     rd_ref & AUDIO_REF_ERR) !=
						    0) {
							state->rd_ref &=
							    ~(AUDIO_REF_ERR);
							up(&state->sem);
							return -ENXIO;
						}
						up(&state->sem);
					}
				}
				if (is->active == 0) {
					if ((is->buffers == NULL)
					    && (audio_setup_buf(is) != 0)) {
						return -ENOMEM;
					}
					if (audio_prime_rx(state) != 0) {
						return -EFAULT;
					}
				}
				local_irq_save(flags);
				is->stopped = 0;
#ifdef MP200_AUDIO_SINGLE_TRANSFER
				audio_process_dma(is);
				local_irq_restore(flags);
#else
				ret = audio_start_dma(is);
				local_irq_restore(flags);
				if (ret != 0) {
					audio_reset(is);
					return -EFAULT;
				}
#endif
			} else {
#ifdef MP200_AUDIO_SINGLE_TRANSFER
				audio_stop_dma(is);
#else
				audio_reset(is);
#endif
			}
		}
		if ((file->f_mode & FMODE_WRITE) != 0) {
			if ((val & PCM_ENABLE_OUTPUT) != 0) {
				if ((state->wr_ref & AUDIO_REF_PW) == 0) {
					audio_reset(os);
					if (state->hw_init != NULL) {
						/* ¿ open ޤϡ
						 * ޥκݤ
						 * hw_init()wr_ref 
						 * ¾ΤΥޥե
						 */
						down(&state->sem);
						state->hw_init(os, NULL);
						if ((state->
						     wr_ref & AUDIO_REF_ERR) !=
						    0) {
							state->wr_ref &=
							    ~(AUDIO_REF_ERR);
							up(&state->sem);
							return -ENXIO;
						}
						up(&state->sem);
					}
				}
				if ((os->buffers == NULL)
				    && (audio_setup_buf(os) != 0)) {
					return -ENOMEM;
				}
				local_irq_save(flags);
				if ((os->mapped != 0)
				    && (os->pending_frags == 0)) {
					os->pending_frags = os->nbfrags;
					sema_init(&os->sem, 0);
#ifdef MP200_AUDIO_SINGLE_TRANSFER_3
#else
					os->active = 1;
#endif
				}
				os->stopped = 0;
#ifdef MP200_AUDIO_SINGLE_TRANSFER
				audio_process_dma(os);
				local_irq_restore(flags);
#else
				ret = audio_start_dma(os);
				local_irq_restore(flags);
				if (ret != 0) {
					audio_reset(os);
					return -EFAULT;
				}
#endif
			} else {
#ifdef MP200_AUDIO_SINGLE_TRANSFER
				audio_stop_dma(os);
#else
				audio_reset(os);
#endif
			}
		}
		return 0;

	case SNDCTL_DSP_GETOPTR:
	case SNDCTL_DSP_GETIPTR:
		DPRINTK("SNDCTL_DSP_GETOPTR cmd=0x%08x\n", cmd);
#ifdef MP201_DSP_CPU
		if (state->int_mode != PCM_INT_CPU_MODE){
			return -EPERM;
		}
#endif
		s = (cmd == SNDCTL_DSP_GETOPTR) ? os : is;
		if (((s == is) && ((file->f_mode & FMODE_READ) == 0))
		    || ((s == os) && ((file->f_mode & FMODE_WRITE) == 0))) {
			return -EINVAL;
		}
		if (s->active != 0) {
			local_irq_save(flags);
			offset = audio_get_dma_pos(s);
			inf.ptr = s->dma_tail * s->fragsize + offset;
			bytecount = s->bytecount + offset;
			inf.blocks = s->fragcount;
			s->fragcount = 0;
			local_irq_restore(flags);
			if (bytecount < 0) {
				bytecount = 0;
			}
			inf.bytes = bytecount;
		}
		return copy_to_user((void *)arg, &inf, sizeof(inf));

	case SNDCTL_DSP_GETOSPACE:
	case SNDCTL_DSP_GETISPACE:
		DPRINTK("SNDCTL_DSP_GETOSPACE cmd=0x%08x\n", cmd);
		s = (cmd == SNDCTL_DSP_GETOSPACE) ? os : is;
		if (((s == is) && ((file->f_mode & FMODE_READ) == 0)) ||
		    ((s == os) && ((file->f_mode & FMODE_WRITE) == 0))) {
			return -EINVAL;
		}
		if ((s->buffers == NULL) && (audio_setup_buf(s) != 0)) {
			return -ENOMEM;
		}
		local_irq_save(flags);
		b = &s->buffers[s->usr_head];
		buf_inf.fragments = atomic_read(&s->sem.count);
#ifdef MP200_AUDIO_SINGLE_TRANSFER
#ifdef MP200_AUDIO_PCM_NONERR
#else
		if(s->pcm_status & s->irq_bit)
			buf_inf.fragments -= 1;
#endif
#endif
		buf_inf.bytes = buf_inf.fragments * s->fragsize;
		buf_inf.fragsize = s->fragsize;
		buf_inf.fragstotal = s->nbfrags;
		local_irq_restore(flags);
		if ((b->offset != 0) && (b->dma_ref == 0)) {
			buf_inf.fragments -= 1;
			buf_inf.bytes -= b->offset;
		}
#ifdef AUDIO_DEBUG_LOG_
		ret = copy_to_user((void *)arg, &buf_inf, sizeof(buf_inf));
		if (DEBUG_LOG_MAX > w_count) {
			unsigned int t;
			struct timeval tv;
			do_gettimeofday(&tv);
			t = tv.tv_sec - w_usec.tv_sec;
			t *= 1000000;
			if (tv.tv_usec < w_usec.tv_usec)
				t -= w_usec.tv_usec - tv.tv_usec;
			else
				t += tv.tv_usec - w_usec.tv_usec;
			(w_dbg_log+w_count)->time = t;
			w_usec = tv;
			(w_dbg_log+w_count)->pending = (uint16_t)cmd;
			(w_dbg_log+w_count)->fifo = 62;
			(w_dbg_log+w_count)->state = (uint32_t)cmd;
			w_count++;
		}
		return ret;
#else
		return copy_to_user((void *)arg, &buf_inf, sizeof(buf_inf));
#endif

#ifdef CONFIG_MP201
	case SNDCTL_DSP_GETODELAY:
		DPRINTK("SNDCTL_DSP_GETODELAY cmd=0x%08x\n", cmd);
		if (!(file->f_mode & FMODE_WRITE))
			return -EINVAL;
		offset = 0;
		s = os;
		b = &s->buffers[s->dma_tail];
		local_irq_save(flags);
		if (b->dma_ref != 0) {
			val = mp200_get_dma_pos(s->dma_channel) - b->dma_addr;
			if (val >= s->fragsize) {
				val = s->fragsize - 4;
			}
			val = s->fragsize - val;
		} else if (s->pending_frags != 0) {
			if (b->offset) {
				val = s->fragsize - b->offset;
				offset = 1;
			}
			else {
				val = 0;
			}
		} else {
			val = 0;
		}
		val += (s->pending_frags - offset) * s->fragsize;
		b = &s->buffers[s->usr_head];
		val += b->offset;
		local_irq_restore(flags);
#ifdef AUDIO_DEBUG_LOG_
		if (DEBUG_LOG_MAX > w_count) {
			(w_dbg_log+w_count)->pending = (uint16_t)s->pending_frags;
			(w_dbg_log+w_count)->time = b->offset;
			(w_dbg_log+w_count)->fifo = (uint16_t)82;
			(w_dbg_log+w_count)->state = (uint32_t)val;
			w_count++;
		}
#endif
		return put_user(val, (int *)arg);
#endif
	case SNDCTL_DSP_NONBLOCK:
		DPRINTK("SNDCTL_DSP_NONBLOCK cmd=0x%08x\n", cmd);
		file->f_flags |= O_NONBLOCK;
		return 0;

	case SNDCTL_DSP_RESET:
		DPRINTK("SNDCTL_DSP_RESET cmd=0x%08x\n", cmd);
#ifdef MP201_DSP_CPU
		/* DSP ˤƤʤˤɬפˤʤ뤫
		 * ʤŪˤϤʤˤ⤷ʤ
		 */
		if (state->int_mode != PCM_INT_CPU_MODE){
			return 0;
		}
#endif
		if ((file->f_mode & FMODE_WRITE) != 0) {
			audio_reset(os);
			os->bytecount = 0;
			os->fragcount = 0;
		}
		if ((file->f_mode & FMODE_READ) != 0) {
			audio_reset(is);
			is->bytecount = 0;
			is->fragcount = 0;
		}
#ifdef AUDIO_DEBUG_LOG_
		if (DEBUG_LOG_MAX > w_count) {
			unsigned int t;
			struct timeval tv;
			do_gettimeofday(&tv);
			t = tv.tv_sec - w_usec.tv_sec;
			t *= 1000000;
			if (tv.tv_usec < w_usec.tv_usec)
				t -= w_usec.tv_usec - tv.tv_usec;
			else
				t += tv.tv_usec - w_usec.tv_usec;
			(w_dbg_log+w_count)->time = t;
			w_usec = tv;
			(w_dbg_log+w_count)->pending = (uint16_t)cmd;
			(w_dbg_log+w_count)->fifo = 63;
			(w_dbg_log+w_count)->state = (uint32_t)arg;
			w_count++;
		}
#endif
		return 0;

	case PCM_SET_CTRL:
		DPRINTK("PCM_SET_CTRL cmd=0x%08x\n", cmd);
		if (copy_from_user(&sett, (pcm_ctrl_t *) arg, sizeof(sett)) !=
		    0) {
			return -EFAULT;
		}
		sett_save = *state->ctrls->sett;
		*state->ctrls->sett = sett;
		ret = audio_pcm_set_ctrl(state);
		if (ret != 0) {
			*state->ctrls->sett = sett_save;
		}
		return ret;

	case PCM_GET_CTRL:
		DPRINTK("PCM_GET_CTRL cmd=0x%08x\n", cmd);
		sett = *state->ctrls->sett;
		return copy_to_user((void *)arg, &sett, sizeof(sett));
#ifdef MP201_DSP_CPU
	case PCM_DMA_INT_SET:
	{
		unsigned long flags;
		DPRINTK("PCM_DMA_INT_SET cmd=0x%08x\n", cmd);
		if (arg == state->int_mode) {
			return ret;
		}
		if (arg == PCM_INT_CPU_MODE) {
			audio_state_t dummy = *state;
			state->wr_ref = 0;
			state->rd_ref = 0;
			DLPRINTK("***** audio_request_irq *****\n");
			ret = audio_request_irq(inode, file, state);
			*state = dummy;
		}
		else if (arg == PCM_INT_DSP_MODE) {
			audio_state_t dummy = *state;
			state->wr_ref = 0;
			state->rd_ref = 0;
			audio_free_irq(inode, file, state);
			*state = dummy;
			if (file->f_mode & FMODE_WRITE) {
				if (os->dma_channel == MP200_DMAC_M2P_LCH4) {
					local_irq_save(flags);
					outl(inl(PCM_DMA_M2P_INT_SEL)
						| PCM_LCH4_INT_SEL,
						PCM_DMA_M2P_INT_SEL);
					local_irq_restore(flags);
				}
				else{
					return -EINVAL;
				}
			}
			if (file->f_mode & FMODE_READ) {
				if (is->dma_channel == MP200_DMAC_P2M_LCH4) {
					local_irq_save(flags);
					outl(inl(PCM_DMA_P2M_INT_SEL)
						| PCM_LCH4_INT_SEL,
						PCM_DMA_P2M_INT_SEL);
					local_irq_restore(flags);
				}
				else {
					return -EINVAL;
				}
			}
		}
		else {
			return -EINVAL;
		}
		state->int_mode = arg;
		return 0;
	}
#endif
	default:
#ifdef AUDIO_DEBUG_LOG_
		ret = state->client_ioctl(inode, file, cmd, arg);

		if (DEBUG_LOG_MAX > w_count) {
			unsigned int t;
			struct timeval tv;
			do_gettimeofday(&tv);
			t = tv.tv_sec - w_usec.tv_sec;
			t *= 1000000;
			if (tv.tv_usec < w_usec.tv_usec)
				t -= w_usec.tv_usec - tv.tv_usec;
			else
				t += tv.tv_usec - w_usec.tv_usec;
			(w_dbg_log+w_count)->time = t;
			w_usec = tv;
			(w_dbg_log+w_count)->pending = (uint16_t)cmd;
			(w_dbg_log+w_count)->fifo = 62;
			(w_dbg_log+w_count)->state = (uint32_t)cmd;
			w_count++;
		}
		return ret;
#else
		/*
		 * Let the client of this module handle the non generic ioctls
		 */
		DPRINTK("client_ioctl cmd=0x%08x\n", cmd);
		return state->client_ioctl(inode, file, cmd, arg);
#endif
	}

	return 0;
}

#ifdef MP201_DSP_CPU
#define DSP_INT_IT1_IDS0	(IO_ADDRESS (MP200_INTC_BASE) + MP200_INTC_ID_IDS0)
#define DSP_INT_IT1_IMN0	(IO_ADDRESS (MP200_INTC_BASE) + MP200_INTC_ID_IMN0)

#define DSP_PM0_TXRX_DIS	(IO_ADDRESS (MP200_PCM0_BASE) + 0x0008)
#define DSP_PM0_TXRX_MON	(IO_ADDRESS (MP200_PCM0_BASE) + 0x0004)
#define DSP_PM0_ENCLR		(IO_ADDRESS (MP200_PCM0_BASE) + 0x001c)
#define DSP_PM0_ENMON		(IO_ADDRESS (MP200_PCM0_BASE) + 0x0018)
#define DSP_PM0_CLEAR		(IO_ADDRESS (MP200_PCM0_BASE) + 0x0020)

#define DSP_DMA_M2P_DSP_LCH4LCH7_INT_ENABLE_CL	(IO_ADDRESS (MP200_DMAC_BASE) + 0x952C)
#define DSP_DMA_M2P_CONTSTATUS	(IO_ADDRESS (MP200_DMAC_BASE) + 0x9004)
#define DSP_DMA_M2P_END		(IO_ADDRESS (MP200_DMAC_BASE) + 0x9008)
#endif

static int audio_release(struct inode *inode, struct file *file)
{
	audio_state_t *state = file->private_data;
	audio_stream_t *os =
	    (file->f_mode & FMODE_WRITE) ? state->output_stream : NULL;
	audio_stream_t *is =
	    (file->f_mode & FMODE_READ) ? state->input_stream : NULL;
#ifdef MP201_DSP_CPU
#else
	pcm_regs_t *regs = state->ctrls->regs;
#endif

	/* ¿ open ޤϡޥκݤ open/close ¾
	 * Υޥե¾ؿѿ¾ˤѡ
	 */
	down(&state->sem);

#ifdef AUDIO_DEBUG_FRE
	del_timer_sync(&poll_timer);
#endif

#ifdef MP201_DSP_CPU
	if ((file->f_mode & FMODE_READ) != 0) {
		audio_discard_buf(is);
	}

	if ((file->f_mode & FMODE_WRITE) != 0) {
		audio_sync(state);
#ifdef AUDIO_DEBUG_LOG
		printk(KERN_INFO "write end state \n");
		printk(KERN_INFO "   usr_head = %d\n",os->usr_head);
		printk(KERN_INFO "   dma_head = %d\n",os->dma_head);
		printk(KERN_INFO "   dma_tail = %d\n",os->dma_tail);
		printk(KERN_INFO "   pending  = %d\n",os->pending_frags);
		printk(KERN_INFO "   sem      = %d\n",atomic_read(&os->sem.count));
		printk(KERN_INFO "   pcm_int  = %d\n",os->pcm_intr);
#endif
		audio_discard_buf(os);
#ifdef AUDIO_DEBUG_LOG
		printk(KERN_INFO "write reset state \n");
		printk(KERN_INFO "   usr_head = %d\n",os->usr_head);
		printk(KERN_INFO "   dma_head = %d\n",os->dma_head);
		printk(KERN_INFO "   dma_tail = %d\n",os->dma_tail);
		printk(KERN_INFO "   pending  = %d\n",os->pending_frags);
		printk(KERN_INFO "   sem      = %d\n",atomic_read(&os->sem.count));
		printk(KERN_INFO "   pcm_int  = %d\n",os->pcm_intr);
#endif
	}

	/* decide on one's own whether to shutdown */
	if (state->hw_shutdown != NULL) {
		state->hw_shutdown(os, is);
	}

	if (state->int_mode == PCM_INT_CPU_MODE){
		audio_free_irq(inode, file, state);
		DLPRINTK("cpu DSP_INT_IT1_IMN0 : 0x%08x\n", inl(DSP_INT_IT1_IMN0));
		DLPRINTK("cpu DSP_PM0_TXRX_MON : 0x%08x\n", inl(DSP_PM0_TXRX_MON));
		DLPRINTK("cpu DSP_PM0_ENMON : 0x%08x\n", inl(DSP_PM0_ENMON));
		DLPRINTK("cpu DSP_DMA_M2P_CONTSTATUS : 0x%08x\n", inl(DSP_DMA_M2P_CONTSTATUS));

	}
	else {
		/* [INTC]
		 * INT_IT1_IDS0[31:0] = 0x0040_0000
		 */
		DLPRINTK("DSP_INT_IT1_IMN0 : 0x%08x\n", inl(DSP_INT_IT1_IMN0));
		outl(0x00400000, DSP_INT_IT1_IDS0);
		DLPRINTK("DSP_INT_IT1_IMN0 : 0x%08x\n", inl(DSP_INT_IT1_IMN0));

		/* [PCM]
		 * PM0_TXRX_DIS[0] = 1
		 * PM0_CLEAR[1:0] = 0x3
		 * PM0_ENCLR[2:0] = 0x7
		 */
		DLPRINTK("DSP_PM0_TXRX_MON : 0x%08x\n", inl(DSP_PM0_TXRX_MON));
		DLPRINTK("DSP_PM0_ENMON : 0x%08x\n", inl(DSP_PM0_ENMON));
		outl(0x1, DSP_PM0_TXRX_DIS);
		outl(0x3, DSP_PM0_CLEAR);
		outl(0x7, DSP_PM0_ENCLR);
		DLPRINTK("DSP_PM0_TXRX_MON : 0x%08x\n", inl(DSP_PM0_TXRX_MON));
		DLPRINTK("DSP_PM0_ENMON : 0x%08x\n", inl(DSP_PM0_ENMON));

		/* [DMA]
		 * DMA_M2P_LCH4LCH7_INT_ENABLE_CL[2:0] = 0x5
		 * DMA_M2P_CONTSTATUS[4]  "0" ˤʤޤǥݡ󥰤
		 * 嵭ؼPM0_TXRX_DIS 򤫤ƤΤǡDMA ž
		 * ϤʤΤߤ򤫤롣
		 */
		DLPRINTK("DSP_DMA_M2P_CONTSTATUS : 0x%08x\n", inl(DSP_DMA_M2P_CONTSTATUS));
		outl(0x5, DSP_DMA_M2P_DSP_LCH4LCH7_INT_ENABLE_CL);
		outl(0x10, DSP_DMA_M2P_END);
		DLPRINTK("DSP_DMA_M2P_CONTSTATUS : 0x%08x\n", inl(DSP_DMA_M2P_CONTSTATUS));

	}
#else
	if ((file->f_mode & FMODE_READ) != 0) {
		audio_discard_buf(is);
		mp200_free_dma(is->dma_channel);
		is->dma_regs = NULL;
		*regs->enclr =
		    PCM_INT_RX_FRE | PCM_INT_RX_URE | PCM_INT_RX_ORE |
		    PCM_INT_RX_REN;
		*regs->clear = is->cl_bit;
#ifdef CONFIG_MP201
#else
		free_irq(state->pcm_irq, (void *)is);
#endif
	}

	if ((file->f_mode & FMODE_WRITE) != 0) {
		audio_sync(state);
		audio_discard_buf(os);
		mp200_free_dma(os->dma_channel);
		os->dma_regs = NULL;
		*regs->enclr =
		    PCM_INT_TX_FRE | PCM_INT_TX_URE | PCM_INT_TX_ORE |
		    PCM_INT_TX_WEN;
		*regs->clear = os->cl_bit;
#ifdef CONFIG_MP201
#else
		free_irq(state->pcm_irq, (void *)os);
#endif
	}

	/* decide on one's own whether to shutdown */
	if (state->hw_shutdown != NULL) {
		state->hw_shutdown(os, is);
	}

#ifdef CONFIG_MP201
	if ((0 == state->wr_ref) && (0 == state->rd_ref)) {
		DPRINTK("free_irq pcm\n");
		free_irq(state->pcm_irq, (void *)state);
	}
#endif
#endif

#ifdef MP201_DSP_CPU
	mp200_audio_busy = 0;
#endif

#ifdef AUDIO_DEBUG_LOG
	{
		int i;
		printk(KERN_INFO "write log output."
				" fragsize = %d : nbfrags = %d \n",
				os->fragsize, os->nbfrags);
		for (i=0; w_count>i; i++) {
			if(67 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: sync s  : pending=%2d : "
					"time=%8d us :  sem    =%2d\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(68 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: sync e  : sem    =%2d : "
					"time=%8d us : ret = %d\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(62 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: ioctl   : "
					"time=%8d us :",
					i, (w_dbg_log+i)->time);
				switch((w_dbg_log+i)->state) {
				case OSS_GETVERSION:
					printk(KERN_INFO "cmd OSS_GETVERSION  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_GETBLKSIZE:
					printk(KERN_INFO "cmd SNDCTL_DSP_GETBLKSIZE  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_GETCAPS:
					printk(KERN_INFO "cmd SNDCTL_DSP_GETCAPS  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_SETFRAGMENT:
					printk(KERN_INFO "cmd SNDCTL_DSP_SETFRAGMENT  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_SYNC:
					printk(KERN_INFO "cmd SNDCTL_DSP_SYNC  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_SETDUPLEX:
					printk(KERN_INFO "cmd SNDCTL_DSP_SETDUPLEX  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_POST:
					printk(KERN_INFO "cmd SNDCTL_DSP_POST  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_GETTRIGGER:
					printk(KERN_INFO "cmd SNDCTL_DSP_GETTRIGGER  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_SETTRIGGER:
					printk(KERN_INFO "cmd SNDCTL_DSP_SETTRIGGER  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_GETOPTR:
					printk(KERN_INFO "cmd SNDCTL_DSP_GETOPTR  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_GETIPTR:
					printk(KERN_INFO "cmd SNDCTL_DSP_GETIPTR  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_GETOSPACE:
					printk(KERN_INFO "cmd SNDCTL_DSP_GETOSPACE  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_GETODELAY:
					printk(KERN_INFO "cmd SNDCTL_DSP_GETODELAY  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_NONBLOCK:
					printk(KERN_INFO "cmd SNDCTL_DSP_NONBLOCK  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_RESET:
					printk(KERN_INFO "cmd SNDCTL_DSP_RESET  = %08x\n",(w_dbg_log+i)->state);
					break;
				case PCM_SET_CTRL:
					printk(KERN_INFO "cmd PCM_SET_CTRL  = %08x\n",(w_dbg_log+i)->state);
					break;
				case PCM_GET_CTRL:
					printk(KERN_INFO "cmd PCM_GET_CTRL  = %08x\n",(w_dbg_log+i)->state);
					break;
				case PCM_DMA_INT_SET:
					printk(KERN_INFO "cmd PCM_DMA_INT_SET  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_SPEED:
					printk(KERN_INFO "cmd SNDCTL_DSP_SPEED  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_STEREO:
					printk(KERN_INFO "cmd SNDCTL_DSP_STEREO  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_CHANNELS:
					printk(KERN_INFO "cmd SNDCTL_DSP_CHANNELS  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_GETFMTS:
					printk(KERN_INFO "cmd SNDCTL_DSP_GETFMTS  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SNDCTL_DSP_SETFMT:
					printk(KERN_INFO "cmd SNDCTL_DSP_SETFMT  = %08x\n",(w_dbg_log+i)->state);
					break;
				case PCM_CODEC_SET:
					printk(KERN_INFO "cmd PCM_CODEC_SET  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_READ_DEVMASK:
					printk(KERN_INFO "cmd SOUND_MIXER_READ_DEVMASK  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_READ_RECMASK:
					printk(KERN_INFO "cmd SOUND_MIXER_READ_RECMASK  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_READ_STEREODEVS:
					printk(KERN_INFO "cmd SOUND_MIXER_READ_STEREODEVS  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_READ_RECSRC:
					printk(KERN_INFO "cmd SOUND_MIXER_READ_RECSRC  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_WRITE_RECSRC:
					printk(KERN_INFO "cmd SOUND_MIXER_WRITE_RECSRC  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_WRITE_MIC:
					printk(KERN_INFO "cmd SOUND_MIXER_WRITE_MIC  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_WRITE_LINE:
					printk(KERN_INFO "cmd SOUND_MIXER_WRITE_LINE  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_WRITE_PCM:
					printk(KERN_INFO "cmd SOUND_MIXER_WRITE_PCM  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_READ_MIC:
					printk(KERN_INFO "cmd SOUND_MIXER_READ_MIC  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_READ_LINE:
					printk(KERN_INFO "cmd SOUND_MIXER_READ_LINE  = %08x\n",(w_dbg_log+i)->state);
					break;
				case SOUND_MIXER_READ_PCM:
					printk(KERN_INFO "cmd SOUND_MIXER_READ_PCM  = %08x\n",(w_dbg_log+i)->state);
					break;
				default:
					printk(KERN_INFO "cmd default = %08x\n",(w_dbg_log+i)->state);
				}
			}
			else if(63 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: reset   : cmd    =%d : "
					"time=%8d us : arg = %d\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(60 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: sync s1 : dma_ref=%2d : "
					"sem =%8d    : size =%8d byte\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(61 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: sync s2 : pending=%2d : "
					"time=%8d us :  sem    =%2d\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(66 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: write s : sem    =%2d : "
					"time=%8d us : size =%8d byte\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(65 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: write m : sem    =%2d : "
					"time=%8d us : size =%8d byte\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(77 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: wrt  err: head   =%2d : "
					"time=%8d us : state=%08x\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(88 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: write   : head   =%2d : "
					"time=%8d us : size =%8d byte\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(82 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: odelay  : pending=%2d : "
					"usr=%7d byte : size =%8d byte\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(91 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: pollerr : pending=%2d : "
					"time=%8d * 10 ms : state=%08x\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(99 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: pcm err : pending=%2d : "
					"time=%8d * 10 ms : state=%08x\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(98 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: dma cal : pending=%2d : "
					"time=%8d * 10 ms : ref=%08d\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(89 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: dma stt : pending=%2d : "
					"time=%8d * 10 ms :   sem    =%2d\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(44 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: reset   : pending=%2d : "
					"time=%8d * 10 ms :   sem    =%2d\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else if(45 == (w_dbg_log+i)->fifo) {
				printk(KERN_INFO "%4d: reset2  : usr    =%2d : "
					"dma h = %2d :   dma t   =%2d\n",
					i, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
			else {
				printk(KERN_INFO "%4d: fifo=%02d : pending=%2d : "
					"time=%8d * 10 ms : path %d\n",
					i, (w_dbg_log+i)->fifo, (w_dbg_log+i)->pending,
					(w_dbg_log+i)->time, (w_dbg_log+i)->state);
			}
		}
		kfree(w_dbg_log);
	}
#endif
	up(&state->sem);

	return 0;
}

int
mp200_audio_attach(struct inode *inode, struct file *file,
		   audio_state_t * state)
{
	audio_stream_t *os =
	    (file->f_mode & FMODE_WRITE) ? state->output_stream : NULL;
	audio_stream_t *is =
	    (file->f_mode & FMODE_READ) ? state->input_stream : NULL;
#ifdef MP201_DSP_CPU
#else
	pcm_regs_t *regs = state->ctrls->regs;
#endif
	int err;

	/* ¿ open ޤϡޥκݤ open/close ¾
	 * Υޥե¾ؿѿ¾ˤѡ
	 */
	down(&state->sem);

	/* access control */
	err = -ENODEV;
	if (((file->f_mode & FMODE_WRITE) != 0) && (os == NULL)) {
		goto out;
	}
	if (((file->f_mode & FMODE_READ) != 0) && (is == NULL)) {
		goto out;
	}
	err = -EBUSY;

#ifdef MP201_DSP_CPU
	if (mp200_audio_busy != 0) {
		goto out;
	}
#endif
	if (((file->f_mode & FMODE_WRITE) != 0)
	    && ((state->wr_ref & AUDIO_REF_EN) != 0)) {
		goto out;
	}
	if (((file->f_mode & FMODE_READ) != 0)
	    && ((state->rd_ref & AUDIO_REF_EN) != 0)) {
		goto out;
	}

#ifdef MP201_DSP_CPU
	err = audio_request_irq(inode, file, state);
	if (err != 0) {
		goto out;
	}
#else
	err = -EINVAL;

	/* request DMA channels */
	if ((file->f_mode & FMODE_WRITE) != 0) {
		err =
		    mp200_request_dma(os->dma_channel, os->id,
				      (dma_callback_t) audio_dma_os_callback,
				      state, &os->dma_regs);
		if (err != 0) {
			goto out;
		}
	}
	if ((file->f_mode & FMODE_READ) != 0) {
		err =
		    mp200_request_dma(is->dma_channel, is->id,
				      (dma_callback_t) audio_dma_is_callback,
				      state, &is->dma_regs);
		if (err != 0) {
			goto out_2;
		}
	}

	/* request PCM interrupt */
#ifdef CONFIG_MP201
	if ((0 == state->wr_ref) && (0 == state->rd_ref)) {
		*regs->enclr =
		    PCM_INT_TX_FRE | PCM_INT_TX_URE | PCM_INT_TX_ORE |
		    PCM_INT_TX_WEN |
		    PCM_INT_RX_FRE | PCM_INT_RX_URE | PCM_INT_RX_ORE |
		    PCM_INT_RX_REN;
		*regs->clear = PCM_INT_TX_FRE | PCM_INT_TX_URE |
			       PCM_INT_RX_FRE;
		*regs->txrx_dis = PCM_TX_EN | PCM_RX_EN;
		err =
		    request_irq(state->pcm_irq, audio_pcm_callback, SA_INTERRUPT,
				"audio io", (void *)state);
		if (err != 0) {
			goto out_3;
		}
	}
	if ((file->f_mode & FMODE_WRITE) != 0) {
		os->pcm_intr = 0;
		os->pcm_status = 0;
		*regs->enset = os->irq_bit;
	}
	if ((file->f_mode & FMODE_READ) != 0) {
		is->pcm_intr = 0;
		is->pcm_status = 0;
		*regs->enset = is->irq_bit;
	}
#else
	if ((file->f_mode & FMODE_WRITE) != 0) {
		*regs->enclr =
		    PCM_INT_TX_FRE | PCM_INT_TX_URE | PCM_INT_TX_ORE |
		    PCM_INT_TX_WEN;
		*regs->clear = os->cl_bit;
		*regs->txrx_dis = os->en_bit;
		os->pcm_intr = 0;
		os->pcm_status = 0;
		err =
		    request_irq(state->pcm_irq, audio_pcm_callback, SA_SHIRQ,
				"audio out", (void *)os);
		if (err != 0) {
			goto out_3;
		}
		*regs->enset = os->irq_bit;
	}
	if ((file->f_mode & FMODE_READ) != 0) {
		*regs->enclr =
		    PCM_INT_RX_FRE | PCM_INT_RX_URE | PCM_INT_RX_ORE |
		    PCM_INT_RX_REN;
		*regs->clear = is->cl_bit;
		*regs->txrx_dis = is->en_bit;
		is->pcm_intr = 0;
		is->pcm_status = 0;
		err =
		    request_irq(state->pcm_irq, audio_pcm_callback, SA_SHIRQ,
				"audio in", (void *)is);
		if (err != 0) {
			goto out_4;
		}
		*regs->enset = is->irq_bit;
	}
#endif
#endif

	/* now complete initialization */
	if (state->hw_init != NULL) {
		state->hw_init(os, is);
		if (((state->wr_ref | state->rd_ref) & AUDIO_REF_ERR) != 0) {
			state->wr_ref &= ~(AUDIO_REF_ERR);
			state->rd_ref &= ~(AUDIO_REF_ERR);
			err = -ENXIO;
#ifdef MP201_DSP_CPU
			goto out_1;
#else
#ifdef CONFIG_MP201
			goto out_4;
#else
			goto out_5;
#endif
#endif
		}
	}

	if ((file->f_mode & FMODE_WRITE) != 0) {
		audio_reset(os);
		os->fragsize = AUDIO_FRAGSIZE_DEFAULT;
		os->nbfrags = AUDIO_NBFRAGS_DEFAULT;
		os->mapped = 0;
		init_waitqueue_head(&os->wq);
	}
	if ((file->f_mode & FMODE_READ) != 0) {
		audio_reset(is);
		is->fragsize = AUDIO_FRAGSIZE_DEFAULT;
		is->nbfrags = AUDIO_NBFRAGS_DEFAULT;
		is->mapped = 0;
		init_waitqueue_head(&is->wq);
	}

	file->private_data = state;
	file->f_op->release = audio_release;
	file->f_op->write = audio_write;
	file->f_op->read = audio_read;
	file->f_op->mmap = audio_mmap;
	file->f_op->poll = audio_poll;
	file->f_op->ioctl = audio_ioctl;
	file->f_op->llseek = audio_llseek;

#ifdef AUDIO_DEBUG_FRE
	/* printk("dummy"); */
	poll_timer.expires = (100) + jiffies;
	poll_timer.data = (unsigned long)state;
	add_timer(&poll_timer);
#endif

#ifdef MP201_DSP_CPU
	mp200_audio_busy = 1;
	state->int_mode = PCM_INT_CPU_MODE;
#endif
#ifdef AUDIO_DEBUG_LOG
	w_count = 0;
	w_dbg_log = kmalloc(sizeof(w_log) * DEBUG_LOG_MAX, GFP_KERNEL);
	if (!w_dbg_log) {
		printk(KERN_INFO "%s(): unable to allocate debug memory\n",
		       __FUNCTION__);
		goto out_1;
	}
	memset(w_dbg_log, 0, sizeof(w_log) * DEBUG_LOG_MAX);
	f_jiff = jiffies;
	do_gettimeofday(&w_usec);
#endif

	up(&state->sem);
	return 0;

#ifdef MP201_DSP_CPU
      out_1:
	audio_free_irq(inode, file, state);
#else
#ifdef CONFIG_MP201
      out_4:
	free_irq(state->pcm_irq, (void *)state);
#else
      out_5:
	if ((file->f_mode & FMODE_READ) != 0) {
		free_irq(state->pcm_irq, (void *)is);
	}
      out_4:
	if ((file->f_mode & FMODE_WRITE) != 0) {
		free_irq(state->pcm_irq, (void *)os);
	}
#endif
      out_3:
	if ((file->f_mode & FMODE_READ) != 0) {
		mp200_free_dma(is->dma_channel);
	}
      out_2:
	if ((file->f_mode & FMODE_WRITE) != 0) {
		mp200_free_dma(os->dma_channel);
	}
#endif
      out:
	up(&state->sem);
	return err;
}

static int __init mp200_audio_init(void)
{
#ifdef MP200_AUDIO_SINGLE_TRANSFER
	printk(KERN_INFO "Starting %s. (DMA Single)\n", AUDIO_NAME);
#else
	printk(KERN_INFO "Starting %s.\n", AUDIO_NAME);
#endif
#ifdef AUDIO_DEBUG_LOG
	printk(KERN_INFO "debug %s. log output\n", AUDIO_NAME);
#endif
#ifdef AUDIO_DEBUG_STRESS
	printk(KERN_INFO "Stress check %s.\n", AUDIO_NAME);
#endif
#ifdef MP200_AUDIO_PCM_NONFRE
	mp200_dsp_speed = 44100;
#endif
	return 0;
}

static void __exit mp200_audio_exit(void)
{

}

module_init(mp200_audio_init);
module_exit(mp200_audio_exit);
EXPORT_SYMBOL(mp200_audio_attach);
EXPORT_SYMBOL(audio_pcm_set_ctrl);
#ifdef MP201_DSP_CPU
EXPORT_SYMBOL(audio_dsp_set_ctrl);
#endif
MODULE_LICENSE("GPL");
