/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Copyright (C) 2004 freescale Inc Ltd
 *
 * Modification History:
 *        19 APRIL, 2004, Creation
 *        13 MAY,   2004, Added opening policy.
 *        18 MAY,   2004, Added Atlas control IOCTLs.
 *        04/10/07      Enhancement
 */
 
 /*!
  * @file mxc-audio.c
  * @brief this file implements the mxc sound driver interface to OSS framework
  * @ingroup SOUND_DRV
  */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/dma.h>
#include <linux/fs.h>
#include <asm/system.h>        /* save/local_irq_restore */
#include <linux/preempt.h>
#include <asm/arch/clock.h>

#include <asm/arch/dma.h>        /* DMA transfers */
#include <asm/arch/platform.h>        /* register addresses */
#include <asm/arch/spba.h>      /* SPBA configuration for SSI2 */
#include <asm/arch/clock.h>
#include <ssi.h>
#include <registers.h>
#include <dam.h>
#include <atlas_audio.h>

#include "sound_config.h"          /*include many macro definiation*/
#include "dev_table.h"             /*sound_install_audiodrv()...*/
#include "mxc-defs.h"        /* driver common definitions */
#include "mxc-audio.h"        /* driver definitions */
#include "mxc-hwdep.h"

#define  MXC_PRINT_DRIVER_INFO \
         DPRINTK("Using MXC sound driver built on %s at %s\n", \
         __DATE__, __TIME__);

/*!
 Global var. that points to allocated structure (see init)
 used during clean-up time, to free the structure
*/
static mxc_card_t *card_instance;

#ifdef CONFIG_SOUND_MXC_FIFO
#define MXC_FIFO_SIZE   16 * 1024
static char snd_bufplay[2][MXC_FIFO_SIZE];
static char snd_bufrec[MXC_FIFO_SIZE];
#endif

/*! 
* This function handles the OSS command that sets the bit depth
*
* @param    dev      [in] the device's major number
* @param    bits     [in] audio sample bit number. 
*                    If 0, then not set, and return of current value.
* @return   the value set in the device
*           (may be different that the one requested)
* @see      mxc_bit_depth for the supported values
*
* Modification History:
*         19 APR,2004, Initial version
*/
static unsigned int mxc_audio_setbits(int dev, unsigned int bits)
{
        unsigned int ssi_bits;
        mxc_state_t *drv_inst = NULL;
        

        /*! Get the drv instance private data */
        if ( mxc_get_drv_data(dev, &drv_inst) ) return 0;

        switch(bits) {
        case 0:
                ssi_bits = mxc_get_bits(drv_inst);
                break;
        default:
                ssi_bits = mxc_set_bits(drv_inst, bits);
                break;
        }

        
        return ssi_bits;
}

/*! 
* This function handles the OSS command that sets the channel number
*
* @param    dev      [in] the device's major number
* @param    channels [in] the channel number to set (1 or 2)  
*                         If 0, then not set, and return of current value.
* @return   the value set in the device
*           (may be different that the one requested)
* @see      mxc_channel_number for the supported values
*
* Modification History:
*         19 APR,2004, Initial version
*/
static signed short mxc_audio_setchannel(int dev, signed short channels)
{
        signed short chan;
        mxc_state_t *drv_inst = NULL;
        

        /*! Get the drv instance private data */
        if ( mxc_get_drv_data(dev, &drv_inst) ) return 0;

        switch(channels) {
        case 0:
                chan = mxc_get_channels(drv_inst);
                break;
        default:
                chan = mxc_set_channels(drv_inst, channels);
                break;
        }

        
        return chan;
}

/*! 
* This function handles the OSS command that sets the sampling frequency
*
* @param    dev      [in] the device's major number
* @param    speed    [in] the sampling frequency. 
*                         If 0, then not set, and return of current value.
* @return   the value set in the device
*           (may be different that the one requested)
* @see      mxc_sampling_freq for the supported sampling frequencies
*
* Modification History:
*         19 APR,2004, Initial version.
*/
static int mxc_audio_setspeed(int dev, int speed)
{
        int freq;
        mxc_state_t *drv_inst = NULL;
        

        /*! Get the drv instance private data */
        if ( mxc_get_drv_data(dev, &drv_inst) ) return 0;
        
        switch(speed) {
        case 0:
                freq = mxc_get_speed(drv_inst); 
                break;
        default:
                freq = mxc_set_speed(drv_inst, speed); 
                break;
        }

        return freq;
}

/*! 
* This function handles the user commands to the mixer driver
*
* @param    dev      the device's major number
* @param    cmd      a symbol that describes the command to apply
* @param    arg      the command argument
* @return   0 if success, else error
* @see      mxc_audio_ioctl
*
* Modification History:
*       19 APR,2004, Initial version. 
*       19 MAY,2004, Corrections (get_user and put_user)
*       19 MAY,2004, Addings (new IOCTLs, new calls)
*       04/07/27: Moved some dbg IOCTL from audio to mixer
*/
static int mxc_mixer_ioctl(int dev, unsigned int cmd, void __user *arg)
{   
        int ret=1, val;
        mxc_state_t *drv_inst = NULL;

        /*! Get the drv instance private data */
        if ( mxc_get_drv_data(dev, &drv_inst) ) return -EINVAL;

        /*! Only accepts mixer (type 'M') ioctls */
        if (_IOC_TYPE(cmd)!= 'M') {
                ret = -EINVAL;
                goto _end;
        }

        switch (cmd) {
        case SOUND_MIXER_READ_DEVMASK:
                /*! Available channels */
                ret = MXC_PLAYING_SRC_MASK | MXC_RECORDING_SRC_MASK;
                put_user(ret, (int*) arg);
                DBG_INFO ("SOUND_MIXER_READ_DEVMASK 0x%.4x\n", ret);
                break;
        case SOUND_MIXER_READ_OUTMASK:
                /* Mask available playing channels */
                ret = MXC_PLAYING_SRC_MASK;
                put_user(ret, (int*) arg);
                DBG_INFO ("SOUND_MIXER_READ_OUTMASK 0x%.4x\n", ret);
                break;
        case SOUND_MIXER_READ_RECMASK:
                /* Mask available recording channels */
                ret = MXC_RECORDING_SRC_MASK;
                put_user(ret, (int*) arg);
                DBG_INFO ("SOUND_MIXER_READ_RECMASK 0x%.4x\n", ret);
                break;
        case SOUND_MIXER_READ_OUTSRC:
                /* Gets the current active playing device */
                ret = atlas_get_src_out_selected(drv_inst);
                /*! TODO : enhance : if no src selected, must return smthg */
                if ( ret == ATLAS_SRC_OUT_NONE ) ret |= SOUND_MASK_VOLUME;
                put_user(ret, (int*) arg);
                DBG_INFO ("SOUND_MIXER_READ_OUTSRC 0x%.4x\n", ret);
                break;
        case SOUND_MIXER_WRITE_OUTSRC:
                /* Sets the current active playing device */
                if (get_user(val, (int *)arg)) return -EFAULT;
                
                if ( (val == 0) || (val & MXC_PLAYING_SRC_MASK) ) {
                        /* no channel selected or valid src is selected  */
                        DBG_INFO ("SOUND_MIXER_WRITE_OUTSRC 0x%.4x\n", val);
                        ret = atlas_select_src_out(drv_inst, val);
                        put_user(ret, (int*) arg);
                }
                else /* invalid channel selection! */ {
                        DPRINTK ("SOUND_MIXER_WRITE_OUTSRC invalid channel!\n");
                        ret = -EINVAL;
                        put_user(ret, (int*) arg);
                        goto _end;
                }
                break;
        case SOUND_MIXER_READ_RECSRC:
                /* Gets the current active recording device */
                ret = atlas_get_src_in_selected(drv_inst);
                /*! TODO : enhance : if no src selected, must return smthg */
                if ( ret == ATLAS_SRC_IN_NONE ) ret |= SOUND_MASK_MIC;
                put_user(ret, (int*) arg);
                DBG_INFO ("SOUND_MIXER_READ_RECSRC 0x%.4x\n", ret);
                break;
        case SOUND_MIXER_WRITE_RECSRC:
                /* Sets the current active recording device */
                if (get_user(val, (int *)arg)) return -EFAULT;
                
                if ( (val == 0) || (val & MXC_RECORDING_SRC_MASK) ) {
                        /* no channel selected or valid src is selected  */
                        DBG_INFO ("SOUND_MIXER_WRITE_RECSRC 0x%.4x\n", val);
                        ret = atlas_select_src_in(drv_inst, val);
                        put_user(ret, (int*) arg);
                }
                else /* invalid channel selection! */ {
                        DPRINTK ("SOUND_MIXER_WRITE_RECSRC invalid channel!\n");
                        ret = -EINVAL;
                        put_user(ret, (int*) arg);
                        goto _end;
                }
                break;
        case SOUND_MIXER_READ_STEREODEVS:
                /* Mask stereo capable channels  = headset out, handset in */
                ret = MXC_STEREO_OUT_SRC_MASK | MXC_STEREO_IN_SRC_MASK;
                put_user(ret, (int*) arg);
                DBG_INFO ("SOUND_MIXER_READ_STEREODEVS 0x%.4x\n", ret);
                break;
        case SOUND_MIXER_READ_CAPS:
                /* Only one recording source at a time */
                ret = SOUND_CAP_EXCL_INPUT;
                put_user(ret, (int*) arg);
                DBG_INFO ("SOUND_MIXER_READ_CAPS 0x%.4x\n", ret);
                break;
        case SOUND_MIXER_READ_VOLUME:  /* headset */
                /* Get and return volume */
                ret = atlas_get_src_out_volume(drv_inst, 
                                               ATLAS_SRC_OUT_HEADSET);
                DBG_INFO ("SOUND_MIXER_READ_VOLUME 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_READ_PCM:   /* stereo line out */
                /* Get and return volume */
                ret = atlas_get_src_out_volume(drv_inst,
                                               ATLAS_SRC_OUT_LINE_OUT);
                DBG_INFO ("SOUND_MIXER_READ_PCM 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_READ_SPEAKER:   /* handsfree */
                /* Get and return volume */
                ret = atlas_get_src_out_volume(drv_inst, \
                                               ATLAS_SRC_OUT_HANDSFREE);
                DBG_INFO ("SOUND_MIXER_READ_SPEAKER 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_READ_PHONEOUT:   /* earpiece */
                /* Get and return volume */
                ret = atlas_get_src_out_volume(drv_inst, \
                                               ATLAS_SRC_OUT_EARPIECE);
                DBG_INFO ("SOUND_MIXER_READ_PHONEOUT 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_WRITE_VOLUME:   /* headset */
                /* Set and return new volume */
                if (get_user(val, (int *)arg)) return -EFAULT;
                ret = atlas_set_src_out_volume(drv_inst, \
                                               ATLAS_SRC_OUT_HEADSET, val);
                DBG_INFO ("SOUND_MIXER_WRITE_VOLUME 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_WRITE_PCM:  /* stereo line out */
                /* Set and return new volume */
                if (get_user(val, (int *)arg)) return -EFAULT;
                ret = atlas_set_src_out_volume(drv_inst, \
                                               ATLAS_SRC_OUT_LINE_OUT, val);
                DBG_INFO ("SOUND_MIXER_WRITE_PCM 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_WRITE_SPEAKER:  /* handset */
                /* Set and return new volume */
                if (get_user(val, (int *)arg)) return -EFAULT;
                ret = atlas_set_src_out_volume(drv_inst, \
                                               ATLAS_SRC_OUT_HANDSFREE, val);
                DBG_INFO ("SOUND_MIXER_WRITE_SPEAKER 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_WRITE_PHONEOUT:   /* earpiece */
                /* Set and return new volume */
                if (get_user(val, (int *)arg)) return -EFAULT;
                ret = atlas_set_src_out_volume(drv_inst, \
                                               ATLAS_SRC_OUT_EARPIECE, val);
                DBG_INFO ("SOUND_MIXER_WRITE_PHONEOUT 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_READ_MIC:  /*! headset */
                /* Get and return volume */
                ret = atlas_get_src_in_volume(drv_inst, \
                                              ATLAS_SRC_IN_HEADSET);
                DBG_INFO ("SOUND_MIXER_READ_MIC 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_READ_PHONEIN:  /*! handset (mono or stereo) */
                /* Get and return volume */
                ret = atlas_get_src_in_volume(drv_inst, \
                                              ATLAS_SRC_IN_HANDSET);
                DBG_INFO ("SOUND_MIXER_READ_PHONEIN 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_READ_LINE:  /*! line input */
                /* Get and return volume */
                ret = atlas_get_src_in_volume(drv_inst, \
                                              ATLAS_SRC_IN_LINE_IN);
                DBG_INFO ("SOUND_MIXER_READ_LINE 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_WRITE_MIC:  /*! headset */
                /* Set and return new volume */
                if (get_user(val, (int *)arg)) return -EFAULT;
                ret = atlas_set_src_in_volume(drv_inst, \
                                              ATLAS_SRC_IN_HEADSET, val);
                DBG_INFO ("SOUND_MIXER_WRITE_MIC 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_WRITE_PHONEIN:  /*! handset (mono or stereo) */
                /* Set and return new volume */
                if (get_user(val, (int *)arg)) return -EFAULT;
                ret = atlas_set_src_in_volume(drv_inst, \
                                              ATLAS_SRC_IN_HANDSET, val);
                DBG_INFO ("SOUND_MIXER_WRITE_PHONEIN 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        case SOUND_MIXER_WRITE_LINE:                 /*! line input */
                /* Set and return new volume */
                if (get_user(val, (int *)arg)) return -EFAULT;
                ret = atlas_set_src_in_volume(drv_inst, \
                                              ATLAS_SRC_IN_LINE_IN, val);
                DBG_INFO ("SOUND_MIXER_WRITE_LINE 0x%.4x\n", ret);
                put_user(ret, (int*) arg);
                break;
        default:
                DPRINTK ("Unknown or unsupported cmd\n");
                ret = -EINVAL;
                put_user(ret, (int*) arg);
                goto _end;
        }

_end:

        
        return ret;
}

/*! 
* This function handles the user commands to the audio driver
*
* @param    dev      the device's major number
* @param    cmd      a symbol that describes the command to apply
* @param    arg      the command argument
* @return   0 if success, else error
* @see      mxc_mixer_ioctl
*
* Modification History:
*         28 APR,2004, Initial version.
*       04/07/27: Moved some dbg IOCTL from audio to mixer
*/
static int mxc_audio_ioctl(int dev, unsigned int cmd, void __user *arg)
{   
	mxc_state_t *drv_inst = NULL;
        int val;
        
        
        if(mxc_get_drv_data(dev, &drv_inst)){
        	return -EINVAL;
        }

        switch(cmd){
        case SOUND_PCM_WRITE_RATE:
                if(get_user(val, (int *)arg)){
                	return -EFAULT;
                }
                
                val = mxc_audio_setspeed(dev, val);
                break;
                
        case SOUND_PCM_READ_RATE:
                val = mxc_audio_setspeed(dev, 0);
                break;
                
        case SOUND_PCM_WRITE_CHANNELS:
                if(get_user(val, (int *)arg)){
                	return -EFAULT;
                }
                
                val = mxc_audio_setchannel(dev, val);
                break;
                
        case SOUND_PCM_READ_CHANNELS:
                val = mxc_audio_setchannel(dev, 0);
                break;
                
        case SOUND_PCM_READ_BITS:
                val = mxc_audio_setbits(dev, 0);
                break;
                
        case SOUND_PCM_WRITE_BITS:
                if(get_user(val, (int *)arg)){
                	return -EFAULT;
                }
                
                val = mxc_audio_setbits(dev, val);
                break;
         
         /*
        case SNDCTL_ATLAS_WRITE_OUT_MIXER:
                if (get_user(val, (int *)arg)) return -EFAULT;
                if ( val < 0 ){
                	ret = atlas_set_output_mixer(drv_inst, -val, 0);
                }else{
                	ret = atlas_set_output_mixer(drv_inst, val, 1);
                put_user(ret, (int*) arg);
                break;
        case SNDCTL_ATLAS_READ_OUT_MIXER:
                ret = atlas_get_output_mixer(drv_inst);
                put_user(ret, (int*) arg);
                break;
        */
                
        case SNDCTL_ATLAS_WRITE_OUT_ADDER:
                if(get_user(val, (int *)arg)){
                	return -EFAULT;
                }
                
                val = atlas_set_output_adder(drv_inst, val);
                break;
                
        case SNDCTL_ATLAS_READ_OUT_ADDER:
                val = atlas_get_output_adder(drv_inst);
                break;
                
        case SNDCTL_ATLAS_WRITE_OUT_BALANCE:
                if(get_user(val, (int *)arg)){
                	return -EFAULT;
                }
                
                val = atlas_set_output_balance(drv_inst, val);
                break;
                
        case SNDCTL_ATLAS_READ_OUT_BALANCE:
                val = atlas_get_output_balance(drv_inst);
                break;

	/*                
        case SNDCTL_ATLAS_WRITE_IN_BIAS:
                if (get_user(val, (int *)arg)) return -EFAULT;
                ret = atlas_set_input_bias(drv_inst,val);
                put_user(ret, (int*) arg);
                break;
        case SNDCTL_ATLAS_READ_IN_BIAS:
                if (get_user(val, (int *)arg)) return -EFAULT;
                ret = atlas_get_input_bias(drv_inst);
                put_user(ret, (int*) arg);
                break;
        */
              
        case SNDCTL_ATLAS_WRITE_CODEC_FILTER:
                if(get_user(val, (int *)arg)){
                	return -EFAULT;
                }
                
                val = atlas_set_codec_filter(drv_inst, val);
                break;
                
        case SNDCTL_ATLAS_READ_CODEC_FILTER:
                if(get_user(val, (int *)arg)){
                	return -EFAULT;
                }
                
                val = atlas_get_codec_filter(drv_inst);
                break;
                
        case SNDCTL_CLK_SET_MASTER:
                if(get_user(val, (int *)arg)){
                	return -EFAULT;
                }
                
                val = mxc_setup_clk_master(drv_inst, val);
                break;
                
        case SNDCTL_DAM_SET_OUT_PORT:
                if(get_user(val, (int *)arg)){
                	return -EFAULT;
                }
                
                val = mxc_setup_dam_out_port(drv_inst, val);
                break;

        default:
                DPRINTK("IOCTL not implemented by MXC\n");
                return -EINVAL;
        }
        
        return put_user(val, (int __user *)arg);
}

/*! 
* This function resets the hardware and reset data transfers in both directions
*
* @param    dev      the device's major number
* @return   None
*
* Modification History:
*         19 APR,2004, Initial version.
*/
static void mxc_audio_reset(int dev)
{
        mxc_state_t *drv_inst = NULL;

        if ( mxc_get_drv_data(dev, &drv_inst) ) return;
        
        down(&drv_inst->card->mutex);

        /* reset ssi */
        mxc_reset_ssi(drv_inst->ssi_index);

        // disable audio chip (mute at least)
        
        ssi_interrupt_disable(drv_inst->ssi_index, ssi_tx_dma_interrupt_enable);
        ssi_tx_fifo_enable(drv_inst->ssi_index, drv_inst->ssi_fifo_nb, false);
        
        ssi_transmit_enable(drv_inst->ssi_index, false);
        drv_inst->judge[J_PLAY] = 0;
        
        
        ssi_interrupt_disable(drv_inst->ssi_index, ssi_rx_dma_interrupt_enable);
        ssi_rx_fifo_enable(drv_inst->ssi_index, drv_inst->ssi_fifo_nb, false);
        
        ssi_receive_enable(drv_inst->ssi_index, false);
        drv_inst->judge[J_REC] = 0;
        
        drv_inst->card->ssi_ongoing--;
        
  	if(drv_inst->ssi_index == SSI1){
       		atlas_audio_st_dac_conf_network_mode(t_mode_1);
       		atlas_audio_st_dac_conf_network_secondary(t_mode_0, mm_no_mixing);
       	}else{
       		atlas_audio_st_dac_conf_network_mode(t_mode_0);
       		atlas_audio_st_dac_conf_network_secondary(t_mode_1, mm_no_mixing);
       	}
        
        atlas_audio_st_dac_output_mixing_gain(mm_no_mixing);
        	        
        up(&drv_inst->card->mutex);
}

/*! 
* This function gets the hardware resources, and initializes data
* The driver can be opened multiple times, but only one time per minor-number
* If opened in WO mode, then STDAC is requested (CODEC if unavailable)
* If opened in RO mode, then CODEC is requested (if available)
* If opened in RW mode, then CODEC is requested (if available)
* The DAM configuration is set according to opening mode.
*
* @param    dev      the device's major number
* @param    mode     The parameter flag is one of 
*                    OPEN_READ, OPEN_WRITE, OPEN_READWRITE
* @return   0 if success,
* @return   ENXIO if initialisation error
* @return   EBUSY if the driver instance (minor number) already in use
* @return   EBUSY if hardware resources are not available
* @return   -1    for other errors (DMA allocation...)
* @see      sound_config.h
*
* Modification History:
*       19 APR,2004, Initial version.
*       07JUNE,2004, Changed DMA allocation condition 
*       06 JULY 2004, Moved SSI config earlier (virtio)
*       2004/10/22: Rx and Tx channels are requested, regardless opening mode
*       04/11/19: TLSbo44269: mixer rework
*       05/03/03: Compile warning removal for FIFO mode
*/
static int  mxc_audio_open(int dev, int  mode)
{
        mxc_state_t *drv_inst = NULL;
        
        
        /*! Get the drv instance private data */
        if ( mxc_get_drv_data(dev, &drv_inst) ) goto _err_priv_data;

        /*! Check if possible to open the instance (channel av. + mode) */
        down_interruptible(&drv_inst->card->open_close_sema);
        if ( !mxc_open_allowed(drv_inst, mode) ) goto _err_not_allowed;
        
#ifdef CONFIG_SOUND_MXC_SDMA
        /*! Allocate DMA channels (sound_install_audiodrv finish) */
        /* Setup TX and RX channels and interrupt */
              
        
        if ( mode & OPEN_WRITE )
                if( mxc_setup_dma(drv_inst, eDIR_TX_PLAY) ) goto _err_tx_dma2;
        if ( mode & OPEN_READ )
                 if( mxc_setup_dma(drv_inst, eDIR_RX_REC) ) goto _err_rx_dma2;
                
#endif //CONFIG_SOUND_MXC_SDMA

        drv_inst->open_mode = mode;
        drv_inst->judge[J_REC] = 0;
        drv_inst->judge[J_PLAY] = 0;
        drv_inst->dev_index = dev;
        
        up(&drv_inst->card->open_close_sema);
        
        return 0;

_err_not_allowed:
        up(&drv_inst->card->open_close_sema);
_err_priv_data:
        
        return -ENXIO;

#ifdef CONFIG_SOUND_MXC_SDMA
_err_rx_dma2:
        mxc_audio_close(dev);
_err_tx_dma2:

        up(&drv_inst->card->open_close_sema);
        mxc_audio_close(dev);

        
        return -EBUSY;
#endif
}

/*! 
* This function frees the SSI bus, and reset data transfers
*
* @param    dev      the device's major number
* @return   None
*
* Modification History:
*       19 APR,2004, Initial version.  
*       13 MAY,2004, Added opening policy.
*       06/07/04,    Changed device disconnect condition 
*       04/11/19: TLSbo44269: mixer rework
*/
static void mxc_audio_close(int dev)
{
        mxc_state_t *drv_inst = NULL;
        mxc_card_t *card;
        int tx_dma, rx_dma;
        int out_conn = ATLAS_DEVICE_OUT_NONE, in_conn = ATLAS_DEVICE_IN_NONE;
        

        /*! Get the drv instance private data */
        if ( mxc_get_drv_data(dev, &drv_inst) ) return;

        down(&drv_inst->card->mutex);

        card = drv_inst->card;

       /*! Reset audio 
        mxc_audio_reset(dev);*/

        if( drv_inst->open_mode & OPEN_WRITE ){
        	tx_dma = drv_inst->dma_playbuf;
        	
                if(tx_dma >= 0 ){
                	mxc_free_dma(tx_dma);
                        drv_inst->dma_playbuf = -1;
                        DPRINTK("DMA TX %i channel freed\n", tx_dma);
                        out_conn = atlas_enable_out_device(drv_inst, 
                                     drv_inst->card->audio_chip.out_cna, 0);
                }
        }

        if( drv_inst->open_mode & OPEN_READ ){
        	rx_dma = drv_inst->dma_recbuf;
        	
                if( rx_dma >= 0 ){
                	mxc_free_dma(rx_dma);
                        drv_inst->dma_recbuf = -1;
                        DPRINTK("DMA RX %i channel freed\n", rx_dma);
                        in_conn = atlas_enable_in_device(drv_inst, \
                                        drv_inst->card->audio_chip.in_can, 0);
                }
        }

        /*! free audio chip.
            Is a logical OR, else no disconnect when closing RDWR device */
        if ( (out_conn == ATLAS_DEVICE_OUT_NONE) && \
             (in_conn == ATLAS_DEVICE_IN_NONE) ) {
                atlas_select_src_out(drv_inst, ATLAS_SRC_OUT_NONE);
                atlas_select_src_in(drv_inst, ATLAS_SRC_IN_NONE);
                card->connected_device = ATLAS_PHYSICAL_DEVICE_NONE;
                card->audio_chip.out_cna = ATLAS_DEVICE_OUT_NONE;
                card->audio_chip.in_can = ATLAS_DEVICE_IN_NONE;
        }
        
        /*! decrement usage count */
        spin_lock(&card->drv_inst_lock);
        card->open_cnt--;

        drv_inst->device_busy = 0;
        drv_inst->ssi_busy = 0;

        if ( drv_inst->open_mode & OPEN_READ ) card->rec_cnt--;
        if ( drv_inst->open_mode & OPEN_WRITE ) card->play_cnt--;

        spin_unlock(&card->drv_inst_lock);

        up(&drv_inst->card->mutex);

        DBG_INFO("Active streams : IN[%d], OUT[%d]\n", \
                        card->rec_cnt, card->play_cnt);
        
}

/*! 
* This function arms/rearms the DMA for an output stream
*
* @param    dev      the device's major number
* @param    buf      the buffer to play
* @param    count    the buffer size in bytes
* @param    intrflag 
* @return   None
*
* Modification History:
*         19 APR,2004, Initial version.
*/
static void mxc_audio_output_block(int dev, unsigned long buf, 
                                     int count, int intrflag)
{
        mxc_state_t *drv_inst = NULL;
        dma_request_t sdma_request;
        
	struct dma_buffparms* dmap = audio_devs[dev]->dmap_out;
	dmap->flags |= DMA_NOTIMEOUT | DMA_NODMA;
	
        /*! Get the drv instance private data */
        if( mxc_get_drv_data(dev, &drv_inst) != 0 ){
        	 return;
        }

	memset(&sdma_request, 0, sizeof(dma_request_t));
        sdma_request.sourceAddr = (char*)buf;
       	sdma_request.count = count;
        mxc_dma_set_config(drv_inst->dma_playbuf, &sdma_request, 0);
        mxc_dma_start(drv_inst->dma_playbuf);
        
}

/*! 
* This function arms/rearms the DMA for an input stream
*
* @param    dev      the device's major number
* @param    buf      the buffer where to receive data
* @param    count    the buffer size in bytes
* @param    intrflag 
* @return   None
*
* Modification History:
*         19 APR,2004, Initial version.
*/
static void mxc_audio_start_input(int dev, unsigned long buf, int count, 
					int intrflag)
{
	struct dma_buffparms* dmap = audio_devs[dev]->dmap_in;
        mxc_state_t *drv_inst = NULL;
        dma_request_t sdma_request;
        
	dmap->flags |= DMA_NOTIMEOUT | DMA_NODMA;
	
        /*! Get the drv instance private data */
        if( mxc_get_drv_data(dev, &drv_inst) != 0 ){
        	 return;
        }

	memset(&sdma_request, 0, sizeof(dma_request_t));
	
        sdma_request.destAddr = (char*)buf;
       	sdma_request.count = count;
       	       	          
        mxc_dma_set_config(drv_inst->dma_recbuf, &sdma_request, 0);
        mxc_dma_start(drv_inst->dma_recbuf);
}

/*! 
* This function triggers the hardware to launch input or output
*
* @param    dev      the device's major number
* @param    state
* @return   None
*
* Modification History:
*         19 APR,2004, Initial version.
*/
static void mxc_audio_trigger(int dev, int state)
{
        mxc_state_t *drv_inst = NULL;
        

        /*! Get the drv instance private data */
        if ( mxc_get_drv_data(dev, &drv_inst) ) return;

        down(&drv_inst->card->mutex);

        state &= drv_inst->open_mode;
        if (state & PCM_ENABLE_OUTPUT) {
                if ( drv_inst->judge[J_PLAY] != 1 ) {
                        DPRINTK("Triggering on Tx\n");
                        ssi_interrupt_enable(drv_inst->ssi_index, \
                                ssi_tx_dma_interrupt_enable);
                        ssi_tx_fifo_enable(drv_inst->ssi_index,\
                                           drv_inst->ssi_fifo_nb, true);
                        ssi_transmit_enable(drv_inst->ssi_index, true);
                        drv_inst->judge[J_PLAY] = 1;
                }
        }
        else if(state & PCM_ENABLE_INPUT) {
                if ( drv_inst->judge[J_REC] != 1 ) {
                        DPRINTK("Triggering on Rx\n");
                        ssi_interrupt_enable(drv_inst->ssi_index, \
                                ssi_rx_dma_interrupt_enable);
                        ssi_rx_fifo_enable(drv_inst->ssi_index, \
                                           drv_inst->ssi_fifo_nb, true);
                        ssi_receive_enable(drv_inst->ssi_index, true);
                        drv_inst->judge[J_REC] = 1;
                }
        }

        up(&drv_inst->card->mutex);
        
}

/*! 
* This function prepares the hardware to play or record a stream.
* Is configuring the DAM according to opening mode @see mxc_audio_trigger
*
* @param    dev      the device's major number
* @param    bsize
* @param    bcount
* @return   0 if success, -1 else
*
* Modification History:
*         19 APR,2004, Initial version.
*/
static int mxc_audio_prepare_for_io(int dev, int bsize, int bcount)
{
        mxc_state_t *drv_inst = NULL;

        /*! Get the drv instance private data */
        if ( mxc_get_drv_data(dev, &drv_inst) ) return -1;

        /* Mark the used device as busy */
        down(&drv_inst->card->mutex);

        drv_inst->device_busy = 1;
        drv_inst->card->ssi_ongoing++;
        
        if(drv_inst->card->ssi_ongoing < 2){
          
        	if(drv_inst->ssi_index == SSI1){
        		atlas_audio_st_dac_conf_network_mode(t_mode_0);
        	}else{
        		atlas_audio_st_dac_conf_network_mode(t_mode_1);
        	}
        }
        
        if(drv_inst->card->ssi_ongoing == 2){
        	/* Activate mixing by setting an attenuation bit */
        	atlas_audio_st_dac_output_mixing_gain(mm_att_0dB); 
        	//atlas_audio_st_dac_conf_network_secondary(sec_ts_tx, mm_att_0dB);
        }else{
        	atlas_audio_st_dac_output_mixing_gain(mm_no_mixing);
        	
        	if(drv_inst->ssi_index == SSI1){
        		atlas_audio_st_dac_conf_network_secondary(t_mode_1, mm_no_mixing);
        	}else{
        		atlas_audio_st_dac_conf_network_secondary(t_mode_0, mm_no_mixing);
        	}
        }
        
        up(&drv_inst->card->mutex);

        return 0;
}

/*! 
* This function writes the given buffer to the SSI, in FIFO mode
*
* @param    dev              the device's major number        
* @param    buf                the buffer to write on the bus
* @param    count        the number of bytes to write
* @return   0 
*
* Modification History:
*         19 APR,2004, Initial version.
*/
#ifdef CONFIG_SOUND_MXC_FIFO
int mxc_audio_write_fifo(int dev, const char *buf, int count)
{
        mxc_state_t *drv_inst = NULL;
        ssi_mod ssi_index;
        fifo_nb fifo_index;
        int cnt, i, j, k, left, right, loops;
        FUNC_START

        /*! Get the drv instance private data */
        if ( mxc_get_drv_data(dev, &drv_inst) ) return 0;

        ssi_index = drv_inst->ssi_index;
        fifo_index = drv_inst->ssi_fifo_nb;

        ssi_transmit_enable(ssi_index, true);
        ssi_tx_fifo_enable(ssi_index, ssi_fifo_0, true);
        ssi_tx_fifo_enable(ssi_index, ssi_fifo_1, true);
        
        
#ifdef MXC_HIZ_WORKAROUND
        ssi_tx_mask_time_slot(ssi_index, ~(0x3 << drv_inst->ssi_tx_timeslot));
#endif
        cnt = count;
        k = 0;

        if ( drv_inst->in_out_channel_num == 2 ) {
                while ( cnt > 0 ) {
                        loops = ( (cnt>MXC_FIFO_SIZE) ? MXC_FIFO_SIZE : cnt);
                        DBG_INFO("Copy from user (%d)...", loops);
                        if( copy_from_user(snd_bufplay[ssi_index], 
                                           &(buf)[k], 
                                           loops) ) goto _err;
                        k += loops;
                        DBG_INFO(" completed\n");

                        for (i=0; i<loops; i+=4) {

                                left = ((snd_bufplay[ssi_index][i+1]<<8) | 
                                         snd_bufplay[ssi_index][i]);
                                right = ((snd_bufplay[ssi_index][i+3]<<8) | 
                                          snd_bufplay[ssi_index][i+2]);

                                ssi_set_data(ssi_index, fifo_index, left);
                                ssi_set_data(ssi_index, fifo_index, right);

                                j = 0;
                                do {
                                        j++;
                                        if ( j>1000000 ) {
                                                DPRINTK("Timeout -> exit \n");
                                                i = loops;
                                                cnt = -1;
                                                break;
                                        }
                                } while ( ssi_tx_fifo_counter(ssi_index, fifo_index) >= 6 );
                        }

                        cnt-=MXC_FIFO_SIZE;
                }
        }
        else {
                while ( cnt > 0 ) {
                        loops = ( (cnt>MXC_FIFO_SIZE) ? MXC_FIFO_SIZE : cnt);
                        DBG_INFO("Copy from user (%d)...", loops);
                        if( copy_from_user(snd_bufplay[ssi_index], \
                                           &(buf)[k], \
                                           loops) ) goto _err;;
                        k += loops;
                        DBG_INFO(" completed\n");

                        for (i=0; i<loops;i+=2) {
                                left = ((snd_bufplay[ssi_index][i+1]<<8) | \
                                         snd_bufplay[ssi_index][i]);
                                ssi_set_data(ssi_index, fifo_index, left);
#ifdef MXC_HIZ_WORKAROUND
                                ssi_set_data(ssi_index, fifo_index, 0);
#endif

                                j = 0;
                                do {
                                        j++;
                                        if ( j>1000000 ) {
                                                DPRINTK("Timeout -> exit \n");
                                                i = loops;
                                                cnt = -1;
                                                break;
                                        }
                                } while ( ssi_tx_fifo_counter(ssi_index, fifo_index) >= 6 );
                        }

                        cnt-=MXC_FIFO_SIZE;
                }
        }
        
        FUNC_END
        return (k);

_err:
        FUNC_END
        return -EFAULT;

}


int mxc_audio_read_fifo(int dev, char *buf, int count)
{
        mxc_state_t *drv_inst = NULL;
        ssi_mod ssi_index;
        fifo_nb fifo_index;
        int cnt, w, i, j, k, left, loops;
        

        /*! Get the drv instance private data */
        if ( mxc_get_drv_data(dev, &drv_inst) ) return 0;

        ssi_index = drv_inst->ssi_index;
        fifo_index = drv_inst->ssi_fifo_nb;

        ssi_receive_enable(ssi_index, true);
        ssi_rx_fifo_enable(ssi_index, ssi_fifo_0, true);
        ssi_rx_fifo_enable(ssi_index, ssi_fifo_1, true);
        
        w = cnt = count;
        k = 0;

        while ( cnt > 0 ) {
                loops = ( (cnt>MXC_FIFO_SIZE) ? MXC_FIFO_SIZE : cnt);

                for (i=0; i<loops; i+=2) {
                        left = ssi_get_data(ssi_index, fifo_index);
                        snd_bufrec[i+1] = left>>8;
                        snd_bufrec[i] = left&0xFF;
                        
                        j = 0;
                        do {
                                j++;
                                if ( j>1000000 ) {
                                        DPRINTK("Timeout -> exit \n");
                                        i = loops;
                                        cnt = -1;
                                        break;
                                }
                        } while ( ssi_rx_fifo_counter(ssi_index, fifo_index) <= 2 );
                }
                
                DPRINTK("Copy to user (%d)...", loops);
                if( copy_to_user( buf+k, (char*)snd_bufrec, loops) )
                        return -EFAULT;

                k += loops;
                DPRINTK(" completed (%d)\n", k);

                w -= loops;
                cnt-=MXC_FIFO_SIZE;
                if ( cnt > 0 ) DPRINTK("Remains %d bytes\n", cnt);
        }

        DPRINTK("Bytes recorded: %d\n", count - w);
        return (count-w);
}

#endif //CONFIG_SOUND_MXC_FIFO

/*! 
* This structure describes the audio operations available.
*
* Modification History:
*         19 APR,2004, Initial version.
*/
static struct audio_driver mxc_audio_driver =
{
        .owner =                  THIS_MODULE,
        .open =                   mxc_audio_open,
        .close =                  mxc_audio_close,
        .output_block =           mxc_audio_output_block,
        .start_input =            mxc_audio_start_input,
        .ioctl =                  mxc_audio_ioctl,
        .prepare_for_input =      mxc_audio_prepare_for_io,
        .prepare_for_output =     mxc_audio_prepare_for_io,
        .halt_io =                mxc_audio_reset,
        .trigger =                mxc_audio_trigger,
        .set_speed =              mxc_audio_setspeed,
        .set_bits =               mxc_audio_setbits,
        .set_channels =           mxc_audio_setchannel,
#ifdef CONFIG_SOUND_MXC_FIFO
        .write_fifo =             mxc_audio_write_fifo,
        .read_fifo =              mxc_audio_read_fifo,
#endif
};

/*! 
* This structure describes the mixer operations available.
* This is the interface to the Atlas audio chip
*
* Modification History:
*         19 APR,2004, Initial version.
*/
static struct mixer_operations mxc_mixer_operations =
{
        .owner =  THIS_MODULE,
        .id =     "MXC mixer",
        .name =   "MXC Mixer Driver",
        .ioctl =  mxc_mixer_ioctl
};

/*! 
* This function is the DMA interrupt handler for Rx
* It is common for each Tx audio devices
*
* @param    arg contains specific information: the index of the DMA channel \
*           that issued the interruption
* @return   None
* @see sdma.c
* @see sdma.h
* Modification History:
*         19 APR,2004, Initial version.
*        24 MAY,2004, Adapted to SDMA (the parameter is the DMA channel index)
*/
void mxc_dma_tx_handler(void* arg)
{
        mxc_state_t* drv_inst;
        uint32_t flags;
        
        drv_inst = (mxc_state_t*)arg;
        local_save_flags(flags);
        DMAbuf_outputintr(drv_inst->dev_index, 1);
        local_irq_restore(flags);
}

/*! 
* This function is the DMA interrupt handler for Rx
* It is common for each Rx audio devices
*
* @param    arg contains specific information:
*        the index of the DMA channel that issued the interruption
* @return   None
* @see sdma.c
* @see sdma.h
*
* Modification History:
*         19 APR,2004, Initial version.
*        24 MAY,2004, Adapted to SDMA (the parameter is the DMA channel index)
*/
void mxc_dma_rx_handler(void* arg)
{
	mxc_state_t* drv_inst;
        uint32_t flags;
        
        drv_inst = (mxc_state_t*)arg;
        local_save_flags(flags);
        DMAbuf_inputintr(drv_inst->dev_index);
        local_irq_restore(flags);
}

/*! 
* This function retrieves the driver instance specific information
*
* @param    dev [in]
* @param    drv_inst [out] contains specific information
* @return   0 if success, -1 if drv_inst is NULL.
*
* Modification History:
*         22 APR,2004, Initial version.
*/
static int mxc_get_drv_data(int dev, mxc_state_t **drv_inst)
{
        struct audio_operations *adev = audio_devs[dev];
        int ret = 0;
        

        if ( adev ) {
                *drv_inst = (mxc_state_t *) adev->devc;
                if ( !drv_inst ) ret = -1;
        }
        else {
                goto _err;
        }

        
        return ret;

_err:
        *drv_inst = NULL;
        FUNC_ERR
        return -1;
}

/*! 
* This function configures the DMA controller for Tx channel
*
* @param    drv_inst [in] contains specific information
* @param    params [out] info to fill for the tx transfer
* @return   0 if success
*
* Modification History:
*       04/07/29 Creation
*       04/11/26 Added word_size parameter setting
*/
static inline void mxc_setup_tx_dma(mxc_state_t *drv_inst, \
                                    dma_channel_params *params)
{
        

        params->watermark_level = TX_FIFO_EMPTY_WML;
        params->transfer_type = emi_2_per;
        params->callback = mxc_dma_tx_handler;
        params->arg = drv_inst;
        params->bd_number = 1;

        switch (drv_inst->in_out_bits_per_sample) {
        case AFMT_U8:
                params->word_size = TRANSFER_8BIT;
                break;
        default:    /* AFMT_S16_LE */
                params->word_size = TRANSFER_16BIT;
        }

        if ( drv_inst->ssi_fifo_nb == ssi_fifo_0 ) {
                if ( drv_inst->ssi_index == SSI1 ) {
                        /* fifo 0, SSI1 */
                        params->per_address = (SSI1_BASE_ADDR + 0); /*STX10*/
                        params->event_id = DMA_REQ_SSI1_TX1;
                        params->peripheral_type = SSI;
                }
                else {
                        /* fifo 0, SSI2 */
                        params->per_address = (SSI2_BASE_ADDR + 0); /*STX20*/
                        params->event_id = DMA_REQ_SSI2_TX1;
                        params->peripheral_type = SSI_SP;
                }
        }
        else {
                if ( drv_inst->ssi_index == SSI1 ) {
                        /* fifo 1, SSI1 */
                        params->per_address = (SSI1_BASE_ADDR + 0x4); /*STX11*/
                        params->event_id = DMA_REQ_SSI1_TX2;
                        params->peripheral_type = SSI;
                }
                else {
                        /* fifo 1, SSI2 */
                        params->per_address = (SSI2_BASE_ADDR + 0x4); /*STX21*/
                        params->event_id = DMA_REQ_SSI2_TX2;
                        params->peripheral_type = SSI_SP;
                }
        }
        
        
}

/*! 
* This function configures the DMA controller for Rx channel
*
* @param    drv_inst [in] contains specific information
* @param    params [out] info to fill for the rx transfer
* @return   0 if success
*
* Modification History:
*       04/07/29 Creation
*       04/11/26 Added word_size parameter setting
*/
static inline void mxc_setup_rx_dma(mxc_state_t *drv_inst, \
                                    dma_channel_params *params)
{
        
        params->watermark_level = RX_FIFO_FULL_WML;
        params->transfer_type = per_2_emi;
        params->callback = mxc_dma_rx_handler;
        params->arg = drv_inst;

        params->bd_number = 1;

        switch (drv_inst->in_out_bits_per_sample) {
        case AFMT_U8:
                params->word_size = TRANSFER_8BIT;
                break;
        default:    /* AFMT_S16_LE */
                params->word_size = TRANSFER_16BIT;
        }

        if ( drv_inst->ssi_fifo_nb == ssi_fifo_0 ) {
                if ( drv_inst->ssi_index == SSI1 ) {
                        /* TS 0, SSI1 */
                        params->per_address = (SSI1_BASE_ADDR + 0x8); /*SRX10*/
                        params->event_id = DMA_REQ_SSI1_RX1;
                        params->peripheral_type = SSI;
                }
                else {
                        /* TS 0, SSI2 */
                        params->per_address = (SSI2_BASE_ADDR + 0x8); /*SRX20*/
                        params->event_id = DMA_REQ_SSI2_RX1;
                        params->peripheral_type = SSI_SP;
                }
        }
        else {
                if ( drv_inst->ssi_index == SSI1 ) {
                        /* TS 1, SSI1 */
                        params->per_address = (SSI1_BASE_ADDR + 0xC); /*SRX11*/
                        params->event_id = DMA_REQ_SSI1_RX2;
                        params->peripheral_type = SSI;
                }
                else {
                        /* TS 1, SSI2 */
                        params->per_address = (SSI2_BASE_ADDR + 0xC); /*SRX21*/
                        params->event_id = DMA_REQ_SSI2_RX2;
                        params->peripheral_type = SSI_SP;
                }
        }
        
        
}

#ifdef CONFIG_SOUND_MXC_SDMA
/*! 
* This function configures the DMA controller and requests the DMA interrupt
*
* @param    drv_inst [in] contains specific information
* @param    inout [in] the direction of DMA transfer (rx/tx)
* @return   0 if success
*
* Modification History:
*       22 APR,2004, Initial version.
*       04/07/29 code split for lisibility
*       04/08/30 removed static channel allocation case
*/
static int mxc_setup_dma(mxc_state_t *drv_inst, e_transfer_dir_t inout)
{
        int dma_channel, result;
        dma_channel_params params;
        

        /*! Find a free DMA channel */
        
        dma_channel = 0;
        result = mxc_request_dma(&dma_channel, "TX SDMA OSS");
        if(result < 0){
        	goto _err;
        }
	
        if( inout == eDIR_TX_PLAY ) {
                mxc_setup_tx_dma(drv_inst, &params);
                drv_inst->dma_playbuf = dma_channel;
        }
        else {
                mxc_setup_rx_dma(drv_inst, &params);
                drv_inst->dma_recbuf = dma_channel;
        }

        /* Set the SDMA configuration */
        mxc_dma_setup_channel(dma_channel, &params);
        
        DPRINTK("SDMA channel allocated: %d\n", dma_channel);

        return 0;

_err:
        FUNC_ERR
        return -1;
}
#endif /* CONFIG_SOUND_MXC_SDMA */

/*!
 * This function checks if the driver opening is possible in read mode
*
* @param    drv_inst [in] driver instance private data
* @param    mode [in] the driver opening mode (R, W, or RW)
* @return   1 if opening allowed, else 0 
*
* Modification History:
*       04/07/29: creation
*/
static inline int mxc_open_read_allowed(mxc_state_t *drv_inst, int mode)
{
        int device;
        mxc_card_t *card=drv_inst->card;
        

        spin_lock(&drv_inst->card->drv_inst_lock);
        card->rec_cnt++;
        if ( card->rec_cnt > NR_HW_CH_RX ) goto _not_allowed;
        
        /* Look for connected device */
        if ( card->connected_device == ATLAS_PHYSICAL_DEVICE_STDAC )
                goto _not_allowed;

        spin_unlock(&drv_inst->card->drv_inst_lock);
        device = atlas_get_dev_in_selected(drv_inst);
        switch (device) {
        case(ATLAS_DEVICE_IN_CODEC):
                /* config data path, but stay connected to the same device */
                atlas_select_dev_in(drv_inst, ATLAS_DEVICE_IN_CODEC);
                DPRINTK("Codec is already connected !\n");
                break;
        case(ATLAS_DEVICE_IN_NONE):
                spin_lock(&drv_inst->card->drv_inst_lock);
                card->connected_device = ATLAS_PHYSICAL_DEVICE_CODEC;
                spin_unlock(&drv_inst->card->drv_inst_lock);

                atlas_select_dev_in(drv_inst, ATLAS_DEVICE_IN_CODEC);
#ifdef MXC_DRV_STANDALONE
                atlas_select_src_in(drv_inst, ATLAS_DEFAULT_SRC_IN);
#endif
                DPRINTK("Codec in has been connected !\n");
                break;
        default:
                DPRINTK("Unknown device already connected.\n");
                goto _not_allowed;
        }
        
        
        return 1;

_not_allowed:
        card->rec_cnt--;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_ERR
        return 0;
}

/*!
 * This function checks if the driver opening is possible in write mode
*
* @param    drv_inst [in] driver instance private data
* @param    mode [in] the driver opening mode (R, W, or RW)
* @return   1 if opening allowed, else 0 
*
* Modification History:
*       04/07/29: creation
*/
static inline int mxc_open_write_allowed(mxc_state_t *drv_inst, int mode)
{
        int device, cur_dev;
        mxc_card_t *card=drv_inst->card;
        

        spin_lock(&drv_inst->card->drv_inst_lock);
        card->play_cnt++;
        if ( card->play_cnt > NR_HW_CH_TX ) goto _not_allowed;
        spin_unlock(&drv_inst->card->drv_inst_lock);

        device = atlas_get_dev_out_selected(drv_inst);

        switch (device) {
        case(ATLAS_DEVICE_OUT_STDAC):
                /* config data path, but stay connected to the same device */
                atlas_select_dev_out(drv_inst, ATLAS_DEVICE_OUT_STDAC);
                DPRINTK("Codec is already connected !\n");
                break;
        case(ATLAS_DEVICE_OUT_CODEC):
                /* config data path, but stay connected to the same device */
                atlas_select_dev_out(drv_inst, ATLAS_DEVICE_OUT_CODEC);
                DPRINTK("StDAC is already connected !\n");
                break;
        case(ATLAS_DEVICE_OUT_NONE):
                /* Use Codec for inout transfers, or if already connected */
                spin_lock(&drv_inst->card->drv_inst_lock);
                cur_dev = card->connected_device;

                if ( (mode==OPEN_READWRITE)||(cur_dev==ATLAS_PHYSICAL_DEVICE_CODEC)){
                        card->connected_device = ATLAS_PHYSICAL_DEVICE_CODEC;
                        spin_unlock(&drv_inst->card->drv_inst_lock);

                        atlas_select_dev_out(drv_inst, ATLAS_DEVICE_OUT_CODEC);
                        DPRINTK("Codec out has been connected !\n");
                }
                else {
                        /* Use StDAC for out only transfers */
                        card->connected_device = ATLAS_PHYSICAL_DEVICE_STDAC;
                        spin_unlock(&drv_inst->card->drv_inst_lock);

                        atlas_select_dev_out(drv_inst, ATLAS_DEVICE_OUT_STDAC);
                        DPRINTK("StDAC has been connected !\n");
                }

#ifdef MXC_DRV_STANDALONE
                atlas_select_src_out(drv_inst, ATLAS_DEFAULT_SRC_OUT);
#endif

                break;
        default:
                goto _not_allowed;
        }
        
        
        return 1;

_not_allowed:
        card->play_cnt--;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_ERR
        return 0;
}

/*! 
* This function checks if the driver opening is possible (if resources free)
* If opening allowed the field "open_mode" in mxc_state_t is initialized
*
* @param    drv_inst [in] driver instance private data
* @param    mode [in] the driver opening mode (R, W, or RW)
* @return   1 if opening allowed, else 0 
*
* Modification History:
*         22 APR,2004, Initial version.
*        13 MAY,2004, Added opening policy.
*        29 JUNE 2004 Debugging the OPEN_READ case
*        04/07/29: code split for lisibility
*/
int mxc_open_allowed(mxc_state_t *drv_inst, int mode)
{
        int ret = 1;
        mxc_card_t *card=drv_inst->card;
        

        spin_lock(&drv_inst->card->drv_inst_lock);

        /*! check if this driver instance is free, if yes, lock it */
        if ( drv_inst->ssi_busy ) goto _not_allwd;

        if ( card->open_cnt >= NR_HW_CH ) {
                /*! all SSI buses are busy */
                spin_unlock(&drv_inst->card->drv_inst_lock);
                DPRINTK("Unexpected error : all SSI buses are busy\n");
                goto _end;
        }

        spin_unlock(&drv_inst->card->drv_inst_lock);

        /** Just looking at the number of used channel. Policy applied later */
        /* Test first READ mode, bcse has the most limitations */
        if ( mode & OPEN_READ ) {
                if ( !mxc_open_read_allowed(drv_inst, mode) ) goto _not_allwd;
        }

        if ( mode & OPEN_WRITE ) {
                if ( !mxc_open_write_allowed(drv_inst, mode) ) goto _not_allwd;
        }

        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->ssi_busy = 1;
        card->open_cnt++;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        drv_inst->open_mode = mode;
        goto _end;
        
_not_allwd:
        spin_unlock(&drv_inst->card->drv_inst_lock);
        DBG_INFO("Could not open stream. \n");
        ret = 0; 

_end:
        DBG_INFO("Active streams : IN[%d], OUT[%d]\n", \
                  card->rec_cnt, card->play_cnt);
        
        return ret;
}


/*! 
* This function frees resources allocated by the driver during its init
* Calls "close" function, to ensure everything is freed
*
* @param :   None
* @return:   None
*
* Modification History:
*         19 APR,2004, Initial version.
*       04/08/03: Added a check on card_instance
*/
void cleanup_mxc_audio(void)
{
        int i, tx_dma, rx_dma;
        int mode;
        mxc_state_t *drv_inst = NULL;
        

        /*! If card_instance is null, then nothing to free */
        if ( !card_instance ) goto _end;

        /*! Free of driver instance members */
        for ( i=0; i<NR_HW_CH; i++ ) {
                drv_inst = card_instance->state[i];

                /*! Free of dma channels */
                mode = drv_inst->open_mode;

                if ( mode & OPEN_WRITE ) {
                        if ( (tx_dma = drv_inst->dma_playbuf) >= 0 ) {
/*                      Done in close_dmap, in patched dmabuf.c
                                sound_close_dma(tx_dma);
                                sound_free_dma(tx_dma);*/
                                drv_inst->dma_playbuf = -1;
                                DPRINTK("DMA TX channel freed\n");
                        }
                }

                if ( mode & OPEN_READ ) {
                        if ( (rx_dma = drv_inst->dma_recbuf) >= 0 )  {
/*                      Done in close_dmap, in patched dmabuf.c
                                sound_close_dma(tx_dma);
                                sound_free_dma(tx_dma);*/
                                drv_inst->dma_recbuf = -1;
                                DPRINTK("DMA RX channel freed\n");
                        }
                }

                /*! Unload audio driver */
                if ( card_instance->state[i]->dev_audio >= 0 ) {
                        sound_unload_audiodev( \
                                card_instance->state[i]->dev_audio);
                        DPRINTK("Uninstalled audio driver at index %d\n", \
                                card_instance->state[i]->dev_audio);
                        card_instance->state[i]->dev_audio = -1;
                }

                /*! Free allocated structures */
                if ( card_instance->state[i] ) kfree(card_instance->state[i]);
                card_instance->state[i] = NULL;
        }

        /*! unload mixer driver */
        if ( card_instance->dev_mixer >= 0 ) {
                sound_unload_mixerdev(card_instance->dev_mixer);
                card_instance->dev_mixer = -1;
        }

        /*! Check if all closed as should be */
        CHECK_IF_0(card_instance->open_cnt, "open_cnt");
        CHECK_IF_0(card_instance->play_cnt, "play_cnt");
        CHECK_IF_0(card_instance->rec_cnt, "rec_cnt");
        
        kfree(card_instance);        
        card_instance = NULL;

#ifdef CONFIG_SOUND_MXC_SDMA
        /* SPBA configuration for SSI2 - SDMA is released */
        if ( spba_rel_ownership(SPBA_SSI2, SPBA_MASTER_C) )
                DPRINTK("Not able to release SSI2 ownership\n");
#endif

_end:
        DPRINTK("Bye...\n");

        
}

/*! 
* This function initializes the hardware, at driver loading
*
* @return   0 if success, -1 else
*
* Modification History:
*         19 APR,2004, Initial version.
*       04/08/03: simplified the driver cleanup if init fails
*       04/11/26, added various bit depth support
*/
int init_mxc_audio(void)
{
        int mixer_dev, audio_dev, i;
        int ret = 0;
        
        MXC_PRINT_DRIVER_INFO

        /*! Allocate and initialize card specific info */
        if ( !(card_instance = (mxc_card_t *) kmalloc(sizeof(mxc_card_t), 
                                                       GFP_KERNEL)) ) {
                ret = -ENOMEM;
                goto _fail;
        }

        memset(card_instance, 0, sizeof(mxc_card_t));
        card_instance->dev_mixer = -1;
        card_instance->audio_chip.in_out_sr = 8000;

        /* at the beginning none SSI is in use*/
        card_instance->ssi_ongoing = 0;

        /*! Allocate and initialize driver instance specific info */
        for ( i=0; i<NR_HW_CH; i++ ) {
                if ( !(card_instance->state[i] = 
                                (mxc_state_t *) kmalloc(sizeof(mxc_state_t), 
                                GFP_KERNEL)) ) {
                        ret = -ENOMEM;
                        break;
                }

                memset(card_instance->state[i], 0, sizeof(mxc_state_t));
                card_instance->state[i]->card = card_instance;

                /* first instance will use SSI1, and the second SSI2 */
#ifdef MXC_SDMA_USE_SSI2_ONLY
                card_instance->state[i]->ssi_index = 1;
#else
                card_instance->state[i]->ssi_index = i;
#endif
#ifndef DRV_NTMODE_PLUS_MIX_DISABLED
                /* use timeslots 0,1 for SSI1, and 2,3 for SSI2 (Tx) */
                /* use timeslots 0,1 for SSI1, and SSI2 (Rx)*/
                card_instance->state[i]->ssi_tx_timeslot = 2*i;
                card_instance->state[i]->ssi_rx_timeslot = 0;
#else
                /* use timeslots 0,1 for SSI1, and SSI2 (Tx)*/
                /* use timeslots 0,1 for SSI1, and SSI2 (Rx)*/
                card_instance->state[i]->ssi_tx_timeslot = 0;
                card_instance->state[i]->ssi_rx_timeslot = 0;
#endif
                card_instance->state[i]->dam_dest = DAM_DEFAULT_OUT_PORT;
                card_instance->state[i]->ssi_fifo_nb = SSI_DEFAULT_FIFO_NB;
                card_instance->state[i]->dev_audio = -1;
                
                card_instance->state[i]->dma_playbuf = -1;
                card_instance->state[i]->dma_recbuf = -1;
                
                card_instance->state[i]->in_out_bits_per_sample = AFMT_S16_LE;
                card_instance->state[i]->clock_master = CLK_DEFAULT_MASTER;                
        }
        
        /* Setup some HW configuration (optional) */
        mxc_init_dam(card_instance->state[i]);

        if ( ret != 0 ) goto _fail;

        /*! Install mixer driver */
        if ((mixer_dev = sound_install_mixer(MIXER_DRIVER_VERSION,
                                        mxc_mixer_operations.name,
                                        &mxc_mixer_operations,
                                        sizeof(struct mixer_operations),
                                        (void *) card_instance)) < 0) {
                ret = -EMFILE;
                goto _fail;
        }

        /*! Store the handle into card instance structure */
        card_instance->dev_mixer = mixer_dev;
        
        for (i = 0; i < NR_HW_CH; i++){
                /*! Install audio driver, with dummy Rx and Tx DMA channels,
                        DMA channels are allocated in open */
                if (( audio_dev = sound_install_audiodrv(AUDIO_DRIVER_VERSION,
                                "mxc low level audio driver",
                                &mxc_audio_driver,
                                sizeof(struct audio_driver),
                                DMA_DUPLEX,
#ifdef CONFIG_SOUND_MXC_FIFO
                                AFMT_S16_LE,
#else
                                AFMT_S16_LE | AFMT_U8,
#endif
                                (void *) card_instance->state[i],
                                -1, /*tx dma chann*/
                                -1  /*rx dma chann*/ )) < 0) {
                        ret = -EMFILE;
                        break;
                }
                
                /*! Store the handle into card instance structure */
                card_instance->state[i]->dev_audio = audio_dev;
                DPRINTK("Installed audio driver at index %d\n", audio_dev);

        }

        sema_init(&card_instance->mutex, 1);
        spin_lock_init(&card_instance->drv_inst_lock);
        init_MUTEX(&card_instance->open_close_sema);

#ifdef CONFIG_SOUND_MXC_SDMA
        /*
         * SPBA configuration for SSI2 - SDMA and MCU are set 
         */
        ret |= spba_take_ownership(SPBA_SSI2, SPBA_MASTER_C|SPBA_MASTER_A);
#else
        /*
         * SPBA configuration for SSI2 - MCU only is set 
         */
        ret |= spba_take_ownership(SPBA_SSI2, SPBA_MASTER_A);
#endif

#if defined(CONFIG_ARCH_SCMA11) || defined(CONFIG_ARCH_MX3)

	/* assign USBPLL to be used by SSI1 */
	mxc_set_clocks_pll(SSI1_BAUD, USBPLL);
	
	/*set divider to 2*/
	mxc_set_clocks_div(SSI1_BAUD, 2);
	
	/* enable clock*/
	mxc_clks_enable(SSI1_BAUD);
	
	/* assign USBPLL to be used by SSI2 */
	mxc_set_clocks_pll(SSI2_BAUD, USBPLL);
	
	/*set divider to 2*/
	mxc_set_clocks_div(SSI2_BAUD, 2);
	
	/* enable clock*/
	mxc_clks_enable(SSI2_BAUD);

#endif

#ifdef CONFIG_ARCH_ARGONPLUS
        /*
         * for configuration for SSI
         */
        ModifyRegister(0x18000000, 0x00000000, _reg_MCU_MCR);
        ModifyRegister(0x0003FFFF, 0x00000C06, _reg_MCU_MPDR1);
#endif

        /*! All is OK, exit */
        if ( ret == 0 ) goto _end;

_fail:
        /*! error cases, free resources */
        cleanup_mxc_audio();
        DPRINTK("!!! Failed to init sound driver !!!\n");
_end:
        
        return ret;
}

module_init(init_mxc_audio);
module_exit(cleanup_mxc_audio);

MODULE_AUTHOR("freescale");
MODULE_DESCRIPTION("OSS Audio Device Driver for MXC/Atlas");
MODULE_LICENSE("GPL");
