/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Copyright (C) 2004 freescale Inc Ltd
 */
 
 /*!
  * @file mxc-hwdep.c
  * @brief this file implements the HW dependent part of the sound driver
  * @ingroup SOUND_DRV
  */
#include "sound_config.h"  /*include many macro definition*/
#include "dev_table.h"     /*sound_install_audiodrv()...*/

#include "mxc-defs.h"
#include "mxc-hwdep.h"

/*! 
* This function converts a sound driver define to an atlas driver enum
*
* @param        val [in] the mixer target input signal
* @param        rs [out] the value converted
* @return        the current mixer configuration, -1 if error
*
* Modification History:
*         03 JUNE,2004, Initial version.
*/
int mxc_convert_receive_signal(int val, receive_signal *rs)
{
        int ret = 0;
        FUNC_START

        switch(val) {
        case ATLAS_DEVICE_OUT_STDAC : *rs = rs_st_dac; break;
        case ATLAS_DEVICE_OUT_CODEC : *rs = rs_codec; break;
        case ATLAS_DEVICE_OUT_LINE_IN : *rs = rs_line_in; break;
        default : DPRINTK("Invalid argument\n"); goto _err;
        }
        
        FUNC_END
        return ret;
_err:
        FUNC_ERR
        return -1;
}

/*! 
* This function resets the SSI bus (HW dependent)
* As recommended in datasheet
*
* @param        index : the target SSI module
* @return        None
*
* Modification History:
*         26 APR,2004, Initial version.
*/
void mxc_reset_ssi(int index)
{
        FUNC_START
        ssi_enable(index, false);
        ssi_enable(index, true);
        FUNC_END
}

/*! 
* This function configures the SSI bus (HW dependent)
* More specifically, configures the parameters that depend on used fifo
* Call this function to set the used timeslot
*
* @param        drv_inst [in] info about the current transfer
* @return       None
*
* Modification History:
*       04/08/30, Initial version.
*       04/10/04, removed some bad interrupt configuration
*/
void mxc_setup_ssi_fifo_dep(mxc_state_t *drv_inst)
{
        ssi_mod index = drv_inst->ssi_index;
        fifo_nb ssi_fifo = drv_inst->ssi_fifo_nb;
        FUNC_START

        /*! Fifos must be disabled before SDMA is configured */
        ssi_tx_fifo_enable(index, ssi_fifo, false);
        ssi_rx_fifo_enable(index, ssi_fifo, false);

        ssi_tx_fifo_empty_watermark(index, ssi_fifo_0, TX_FIFO_EMPTY_WML);
        ssi_tx_fifo_empty_watermark(index, ssi_fifo_1, TX_FIFO_EMPTY_WML);
        ssi_rx_fifo_full_watermark (index, ssi_fifo_0, RX_FIFO_FULL_WML);
        ssi_rx_fifo_full_watermark (index, ssi_fifo_1, RX_FIFO_FULL_WML);

        FUNC_END
}

/*! 
* This function configures the SSI bus for the StDac (HW dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return       None
*/
/*
* Modification History:
*       16/07/04, Initial version.
*       04/08/31, timeslot usage changed for dam mixing capability
*       04/10/15, integration of Virtio patch
*/
static inline int mxc_setup_ssi_for_stdac(mxc_state_t *drv_inst)
{
        ssi_mod ssi_index = drv_inst->ssi_index;
        int ssi_tx_timeslots = ~(0x3 << drv_inst->ssi_tx_timeslot);
        int ssi_rx_timeslots = ~(0x3 << drv_inst->ssi_rx_timeslot);
        FUNC_START

        /*! Network mode config, enable Timeslots 0 and 1 or 2 and 3 */
        DPRINTK("configure the SSI for STDAC\n");
        ssi_network_mode(ssi_index, true);
        ssi_tx_mask_time_slot(ssi_index, ssi_tx_timeslots);
        ssi_rx_mask_time_slot(ssi_index, ssi_rx_timeslots);

        drv_inst->in_out_channel_num = 2;

        FUNC_END
        return 0;
}

/*! 
* This function configures the SSI bus for the codec (HW dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return        None
*/
/*
* Modification History:
*        16/07/04, Initial version.
*       04/10/15, integration of Virtio patch
*/
static inline int mxc_setup_ssi_for_codec(mxc_state_t *drv_inst)
{
        ssi_mod ssi_index = drv_inst->ssi_index;
        int ssi_tx_timeslots = ~(0x1 << drv_inst->ssi_tx_timeslot);
        int ssi_rx_timeslots = ~(0x1 << drv_inst->ssi_rx_timeslot);
        FUNC_START

        /*! Network mode config, enable Timeslot 0 */
        DPRINTK("configure the SSI for CODEC\n");
        ssi_network_mode(ssi_index, true);
        ssi_tx_mask_time_slot(ssi_index, ssi_tx_timeslots);
        ssi_rx_mask_time_slot(ssi_index, ssi_rx_timeslots);
        
        drv_inst->in_out_channel_num = 1;

        FUNC_END
        return 0;
}

/*! 
* This function configures the SSI bus and the Atlas (HW dependent)
* Sets either the SSI as master, and atlas slave, or 
*             the atlas as master, and ssi slave.
* Default, set as MCU master
*
* @param        drv_inst [in] info about the current transfer
* @param        snd_device [in] gives the target sound chip (stdac or codec)
*                permitting a specific configuration of the SSI
* @return        None
*
* Modification History:
*       05/03/10, Initial version.
*/
static void set_clocking_mastership(mxc_state_t *drv_inst, int snd_device)
{
        ssi_mod ssi_index = drv_inst->ssi_index;
        int master = drv_inst->clock_master;
        FUNC_START

        if ( master != MXC_AUDIO_CLOCKING_ATLAS_MASTER ){
                DPRINTK("MCU MASTER...\n");

                /* Configure the SSI (MCU master) */
                ssi_tx_frame_direction(ssi_index, ssi_tx_rx_internally);
                ssi_rx_frame_direction(ssi_index, ssi_tx_rx_externally);
                ssi_tx_clock_direction(ssi_index, ssi_tx_rx_internally);
                ssi_rx_clock_direction(ssi_index, ssi_tx_rx_externally);

                /* Configure tge DAM */
                mxc_setup_dam_mcu_master(drv_inst);

                /* Configure the Atlas */
                switch ( snd_device ) {
                case ATLAS_PHYSICAL_DEVICE_STDAC:
                        TRY ( atlas_audio_st_dac_master_slave(ms_slave) )
                        TRY ( atlas_audio_st_dac_conf_pll(7) )
                        break;
                default:
                        TRY ( atlas_audio_codec_master_slave(ms_slave) )
                        TRY ( atlas_audio_codec_conf_pll(7) )
                }
        }
        else {
                /* Configure the SSI (ATLAS master) */
                DPRINTK("ATLAS MASTER...\n");

                ssi_tx_frame_direction(ssi_index, ssi_tx_rx_externally);
                ssi_rx_frame_direction(ssi_index, ssi_tx_rx_externally);
                ssi_tx_clock_direction(ssi_index, ssi_tx_rx_externally);
                ssi_rx_clock_direction(ssi_index, ssi_tx_rx_externally);

                /* Configure tge DAM */
                mxc_setup_dam_atlas_master(drv_inst);

                /* Configure the Atlas */
                switch ( snd_device ) {
                case ATLAS_PHYSICAL_DEVICE_STDAC:
                        TRY ( atlas_audio_st_dac_master_slave(ms_master) )
                        TRY ( atlas_audio_st_dac_conf_pll(4) )
                        break;
                default:
                        TRY ( atlas_audio_codec_master_slave(ms_master) )
                        TRY ( atlas_audio_codec_conf_pll(4) )
                }
        }

        FUNC_END
        return;
_err:
        FUNC_ERR
        return;
}

/*! 
* This function configures the SSI bus (HW dependent)
*
* @param        drv_inst [in] info about the current transfer
* @param        snd_device [in] gives the target sound chip (stdac or codec)
*                permitting a specific configuration of the SSI
* @return        None
*
* Modification History:
*       26 APR,2004, Initial version.
*       07/06/04        specific setup for codec or stdac
*       04/08/30        moved code that depends on used fifo
*/
void mxc_setup_ssi(mxc_state_t *drv_inst, int snd_device)
{
        ssi_mod ssi_index = drv_inst->ssi_index;
        FUNC_START

        ssi_enable(ssi_index, false);

        /* Start by disabling all interrupts */
        ssi_interrupt_disable(ssi_index, MXC_SSI_INT_DISABLE_ALL);

        /*! Set 4-wire mode, and device specific config */
        ssi_system_clock(ssi_index, false);
        ssi_i2s_mode(ssi_index, i2s_normal);
        ssi_synchronous_mode(ssi_index, true);
	ssi_clock_idle_state(ssi_index, true);
        ssi_two_channel_mode(ssi_index, false);

        /*! Connect either the StDac or Codec */
        switch ( snd_device ) {
        case ATLAS_PHYSICAL_DEVICE_STDAC:
                mxc_setup_ssi_for_stdac(drv_inst);
                break;
        default:
                mxc_setup_ssi_for_codec(drv_inst);
        }

        /*! Tell to use either fifo 0 or fifo 1 */
        mxc_setup_ssi_fifo_dep(drv_inst);

        /*! transmit config for both Codec and Stdac */
        ssi_tx_early_frame_sync(ssi_index, ssi_frame_sync_one_bit_before);
        ssi_tx_frame_sync_length(ssi_index, ssi_frame_sync_one_bit);
        ssi_tx_frame_sync_active(ssi_index, ssi_frame_sync_active_high);
        ssi_tx_clock_polarity(ssi_index, ssi_clock_on_rising_edge);
        ssi_tx_shift_direction(ssi_index, ssi_msb_first);
        ssi_tx_bit0(ssi_index, true);
        ssi_tx_flush_fifo(ssi_index);

        /*! tansmit clock control, for both Codec and Stdac */
        ssi_tx_word_length(ssi_index, ssi_16_bits);
        ssi_tx_clock_divide_by_two(ssi_index, false);
        ssi_tx_frame_rate(ssi_index, (unsigned char ) 4);
        ssi_tx_prescaler_modulus(ssi_index, (unsigned char ) 0x56); //OJO
        ssi_tx_clock_prescaler(ssi_index, false);

        /*! receive config, for both Codec and Stdac */
        ssi_rx_early_frame_sync(ssi_index, ssi_frame_sync_one_bit_before);
        ssi_rx_frame_sync_length(ssi_index, ssi_frame_sync_one_bit);
        ssi_rx_frame_sync_active(ssi_index, ssi_frame_sync_active_high);
        ssi_rx_clock_polarity(ssi_index, ssi_clock_on_rising_edge);
        ssi_rx_shift_direction(ssi_index, ssi_msb_first);
        ssi_rx_bit0(ssi_index, true);
        ssi_rx_flush_fifo(ssi_index);

        /*! receive clock control, for both Codec and Stdac */
        ssi_rx_word_length(ssi_index, ssi_16_bits);
        ssi_rx_clock_divide_by_two(ssi_index, false);
        ssi_rx_frame_rate(ssi_index, (unsigned char ) 4);
        ssi_rx_prescaler_modulus(ssi_index, (unsigned char ) 0x56);
        ssi_rx_clock_prescaler(ssi_index, false);

#ifdef CONFIG_SOUND_MXC_SDMA
        /*! Disable DMA request generation */
        ssi_interrupt_disable(ssi_index, ssi_rx_dma_interrupt_enable);
        ssi_interrupt_disable(ssi_index, ssi_tx_dma_interrupt_enable);
#endif
       
        /*! Reset SSI */
        mxc_reset_ssi(ssi_index);
        FUNC_END
}

/*! 
* This function reads the bit depth (HW dependent)
* Note: this function does not access the Atlas to get this value,
* as it is always 16. But this impacts the DMA configuration
*
* @param        drv_inst [in] info about the current transfer
* @return       returns AFMT_S16_LE or AFMT_U8 : the number of bits of 
*               an audio sample
*
* Modification History:
*       26 APR,2004, Initial version.
*       04/11/26, added various bit depth support
*/
unsigned int mxc_get_bits(mxc_state_t *drv_inst)
{
        return drv_inst->in_out_bits_per_sample;
}

/*! 
* This function writes the bit depth (HW dependent) 
* Note : as the format HW is always 16 bits, the function does not 
*        access the HW (empty function). It configures the SDMA
*        in order to make it convert the current bit depth to 16 bits
*
* @param        drv_inst [in] info about the current transfer
* @param        bits [in] the number of bits of an audio sample
* @return       returns AFMT_S16_LE or AFMT_U8 (HW setting stays AFMT_S16_LE)
*
* Modification History:
*       26 APR,2004, Initial version.
*       04/11/26, added various bit depth support
*       05/03/19, added SSI HW configuration to support 8/16 bits
*/
unsigned int mxc_set_bits(mxc_state_t *drv_inst, int bits)
{
        int ret = 0;
        
        switch (bits) {
#ifdef CONFIG_SOUND_MXC_SDMA
        case(AFMT_U8):
                ret = drv_inst->in_out_bits_per_sample = AFMT_U8;
                break;
#endif
        case(AFMT_S16_LE):
        default:
                ret = drv_inst->in_out_bits_per_sample = AFMT_S16_LE;
                break;
        }

        DPRINTK("Bit number is %d\n", ret);
        return ret;
}

/*! 
* This function reads the channel number (HW dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return        the number of bits of an audio sample
*
* Modification History:
*         26 APR,2004, Initial version.
*/
signed short mxc_get_channels(mxc_state_t *drv_inst)
{
        return drv_inst->in_out_channel_num;
}

/*! 
* This function writes the channel number (HW dependent)
*
* @param        drv_inst [in] info about the current transfer
* @param        channels [in] the number of channels of an audio stream
* @return        None
*/
/*
* Modification History:
*       26 APR,2004, Initial version.
*       29 JUNE 2004, first implementation
*       01 JULY,2004, Look at "connected_device" rather than out_can
*       04/08/10: reconfigure SSI+Atlas according to chan nb
*       04/10/15, integration of Virtio patch
*       04/11/19: TLSbo44269: mixer rework
*       FIX IT : TODO : set the SSI config for mono in/out
*/
signed short mxc_set_channels(mxc_state_t *drv_inst, signed short channels)
{
        signed short chan;
        ssi_mod ssi_index = drv_inst->ssi_index;
        int ssi_tx_timeslots, ssi_rx_timeslots;
        FUNC_START

        /*! Check the parameter validity */
        chan = channels;
        if ( channels > 2 ) chan = 2;
        if ( channels < 1 ) chan = 1;

        /*! Apply new value */
        drv_inst->in_out_channel_num = chan;
        
        if ( chan == 1 ) {
                ssi_tx_timeslots = ~(0x1 << drv_inst->ssi_tx_timeslot);
                ssi_rx_timeslots = ~(0x1 << drv_inst->ssi_rx_timeslot);

                /* mono => send 1 timeslot in the frame */
                ssi_tx_mask_time_slot(ssi_index, ssi_tx_timeslots);
                ssi_rx_mask_time_slot(ssi_index, ssi_rx_timeslots);
        }
        else {
                ssi_tx_timeslots = ~(0x3 << drv_inst->ssi_tx_timeslot);
                ssi_rx_timeslots = ~(0x3 << drv_inst->ssi_rx_timeslot);

                /* stereo => send 2 adjacent TS in the frame */
                ssi_tx_mask_time_slot(ssi_index, ssi_tx_timeslots);
                ssi_rx_mask_time_slot(ssi_index, ssi_rx_timeslots);
        }

                
        DPRINTK("Set channel number @ %d\n", chan);

        FUNC_END
        return (chan);
}

/*! 
* This function reads the sampling frequency (HW dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return        the number of bits of an audio sample
*
* Modification History:
*       26 APR,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int mxc_get_speed(mxc_state_t *drv_inst)
{
        int ret;
        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = drv_inst->card->audio_chip.in_out_sr;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        return ret;
}

/*! 
* This function writes the sampling frequency        (HW dependent)
* This shall be executed even if the speed is the same 
* (depedency against channel number)
*
* PM calculation is : PM = SystemClock/(2*BitClock)
*                     where SystemClock=MXC_SYSTEM_CLK (platform dependent)
*                     where BitClock=SR*Channels*BitsPerSample
*
* @param        drv_inst [in] info about the current transfer
* @param        speed [in] the audio stream sampling frequency
* @return        None
*
* Modification History:
*       26 APR,2004, Initial version.
*       27 MAY,2004, Integrated the SSI function calls  
*       24 JUNE,2004, Added Atlas reconfig.              
*       01 JULY,2004, Implementation of Codec cfg
*       04/07/29:       Check if an other rate already used by other SSI
*       04/10/07:       Set the nearest speed value
*       05/03/03:       TLSbo48139: Rework
*/
int mxc_set_speed(mxc_state_t *drv_inst, int speed)
{
        int (*set_sample_rate)(sample_rate);
        int i, device, ssi_pm = 100;
        unsigned long flags;
        ssi_mod ssi_index = drv_inst->ssi_index;
        FUNC_START
        
        spin_lock_irqsave(&drv_inst->card->drv_inst_lock, flags);
        device = drv_inst->card->connected_device;

        switch ( device ) {
        case ATLAS_PHYSICAL_DEVICE_STDAC : 
                set_sample_rate = &atlas_audio_st_dac_set_sample_rate;
                break;
        case ATLAS_PHYSICAL_DEVICE_CODEC : 
                set_sample_rate = &atlas_audio_codec_set_sample_rate;
                break;
        default :
                DPRINTK("Error : no physical device connected\n");
                goto CATCH;
        }

        for ( i=0; i<NR_HW_CH; i++ )
                /*! Check if an other stream already active */
                if ( (i!=ssi_index) && drv_inst->card->state[i]->device_busy ) {
                        /* If yes, then must use the same sampling frequency */
                        DPRINTK("Use SR of SSI #%d\n", i+1);
                        speed = drv_inst->card->audio_chip.in_out_sr;
                        break;
                }                        

        spin_unlock_irqrestore(&drv_inst->card->drv_inst_lock, flags);

        /*! Check for wrong values */
        if ( speed<8000 ) speed = 8000;
        if ( speed>96000 ) speed = 96000;

        if ( device == ATLAS_PHYSICAL_DEVICE_CODEC ) {
                /* Codec has only 8k and 16k available */
                if ( speed >= (8000+16000)/2 ) {
                        ssi_pm = MXC_PM_16000;
                        speed = 16000;
                        TRYOK ( set_sample_rate(sr_16k) )
                }
                else goto _default;
        } /* CODEC SPEEDS */
        else if ( device == ATLAS_PHYSICAL_DEVICE_STDAC ) {
                /* speeds from 8k to 96k */
                if ( speed >= (64000+96000)/2 ) {
                        ssi_pm = MXC_PM_96000;
                        speed = 96000;
                        TRYOK ( set_sample_rate(sr_96k) )
                }
                else if ( speed >= (48000+64000)/2 ) {
                        ssi_pm = MXC_PM_64000;
                        speed = 64000;
                        TRYOK ( set_sample_rate(sr_64k) )
                }
                else if ( speed >= (44100+48000)/2 ) {
                        ssi_pm = MXC_PM_48000;
                        speed = 48000;
                        TRYOK ( set_sample_rate(sr_48k) )
                }
                else if ( speed >= (32000+44100)/2 ) {
                        ssi_pm = MXC_PM_44100;
                        speed = 44100;
                        TRYOK ( set_sample_rate(sr_44100) )
                }
                else if ( speed >= (24000+32000)/2 ) {
                        ssi_pm = MXC_PM_32000;
                        speed = 32000;
                        TRYOK ( set_sample_rate(sr_32k) )
                }
                else if ( speed >= (22050+24000)/2 ) {
                        ssi_pm = MXC_PM_24000;
                        speed = 24000;
                        TRYOK ( set_sample_rate(sr_24k) )
                }
                else if ( speed >= (16000+22050)/2 ) {
                        ssi_pm = MXC_PM_22050;
                        speed = 22050;
                        TRYOK ( set_sample_rate(sr_22050) )
                }
                else if ( speed >= (12000+16000)/2 ) {
                        ssi_pm = MXC_PM_16000;
                        speed = 16000;
                        TRYOK ( set_sample_rate(sr_16k) )
                }
                else if ( speed >= (11025+12000)/2 ) {
                        ssi_pm = MXC_PM_12000;
                        speed = 12000;
                        TRYOK ( set_sample_rate(sr_12k) )
                }
                else if ( speed >= (8000+11025)/2 ) {
                        ssi_pm = MXC_PM_11025;
                        speed = 11025;
                        TRYOK ( set_sample_rate(sr_11025) )
                }
                else goto _default;
        } /* STDAC SPEEDS */
        
        goto _next;

_default:
        /* "default" */
        ssi_pm = MXC_PM_8000;
        speed = 8000;
        TRYOK ( set_sample_rate(sr_8k) )

_next:
        /** Configure SSI's clocking (MCLK == MXC_SYSTEM_CLK) */
        ssi_tx_prescaler_modulus(drv_inst->ssi_index, ssi_pm);
        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->card->audio_chip.in_out_sr = speed;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        DPRINTK("Set sampling frequency @ %d Hz\n", speed);
        FUNC_END
        return speed;
CATCH:
        FUNC_ERR
        return -1;
}

/*! 
* This function sets the default configuration to Stdac (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return        0 in case of success
*
* Modification History:
*         02 MAY,2004, Initial version.
*        02 JUNE,2004, Draft version

int mxc_setup_stdac(mxc_state_t *drv_inst)
{
        FUNC_START
        
        TRYEQ ( atlas_select_dev_out(drv_inst, ATLAS_DEVICE_OUT_STDAC), \
                ATLAS_DEVICE_OUT_STDAC )
        TRYEQ ( atlas_select_src_out(drv_inst, ATLAS_SRC_OUT_LINE_OUT), \
                ATLAS_SRC_OUT_LINE_OUT )

        FUNC_END
        return 0;
CATCH :
        FUNC_ERR
        return 1;
} */

/*! 
* This function sets the default configuration to codec (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return        0 in case of success
*
* Modification History:
*         02 MAY,2004, Initial version.
*        02 JUNE,2004, Draft version

int mxc_setup_codec(mxc_state_t *drv_inst)
{
        FUNC_START
        
        TRYEQ ( atlas_select_dev_out(drv_inst, ATLAS_DEVICE_OUT_CODEC), \
                ATLAS_DEVICE_OUT_CODEC )
        TRYEQ ( atlas_select_src_out(drv_inst, ATLAS_SRC_OUT_LINE_OUT), \
                ATLAS_SRC_OUT_LINE_OUT )
        TRYEQ ( atlas_select_dev_in(drv_inst, ATLAS_DEVICE_IN_CODEC), \
                ATLAS_DEVICE_IN_CODEC )
        TRYEQ ( atlas_select_src_in(drv_inst, ATLAS_SRC_IN_LINE_IN), \
                ATLAS_SRC_IN_LINE_IN )

        FUNC_END
        return 0;
CATCH :
        FUNC_ERR
        return 1;
} */

/*!
* This function retreives the timeslots used by SSI (MCU side)
* in order to configure the SSI at Atlas side
*
* @param        drv_inst [in] info about the current transfer
* @param        prim_ts [out] the first used timeslot
* @param        sec_ts_tx [out] the second used timeslot for tx
* @param        sec_ts_rx [out] the second used timeslot for rx
* @return       0 if success
*
* Modification History:
*       04/08/31        Initial version.
*       04/09/22        Fix ts assignment for codec
*       04/10/07        Fix ts assignment for stereo recording
*/
int get_ssi_timeslots(mxc_state_t *drv_inst, \
              timeslots *prim_ts, timeslots *sec_ts_tx, timeslots *sec_ts_rx)
{
        int ssi_tx_ts_1, ssi_tx_ts_2;
        int connected_device;
        unsigned long flags;

        spin_lock_irqsave(&drv_inst->card->drv_inst_lock, flags);

        /* Suppose here that only 2 SSI connected to DAM. TODO enhance */
        ssi_tx_ts_1 = drv_inst->card->state[0]->ssi_tx_timeslot;
        /* in atlas, primary record is defined by primary play */
        ssi_tx_ts_2 = drv_inst->card->state[1]->ssi_tx_timeslot;
        
        connected_device = drv_inst->card->connected_device;
        
        spin_unlock_irqrestore(&drv_inst->card->drv_inst_lock, flags);

        if ( connected_device == ATLAS_PHYSICAL_DEVICE_STDAC ) {
                /* No timeslot used for Rx, set it to 0 */
                *sec_ts_rx = t_mode_0;

                /* timeslots configuration for stdac (stereo) */
                switch(ssi_tx_ts_1){
                case(0):
                        *prim_ts = t_mode_0;    /* ts 0/1 */
                        break;
                case(2):
                        *prim_ts = t_mode_1;    /* ts 2/3 */
                        break;
                case(4):
                        *prim_ts = t_mode_2;    /* ts 4/5 */
                        break;
                case(6):
                        *prim_ts = t_mode_3;    /* ts 6/7 */
                        break;
                default:
                        *prim_ts = t_mode_0;    /* ts 0/1 */
                }

                switch(ssi_tx_ts_2){
                case(0):
                        *sec_ts_tx = t_mode_0;    /* ts 0/1 */
                        break;
                case(2):
                        *sec_ts_tx = t_mode_1;    /* ts 2/3 */
                        break;
                case(4):
                        *sec_ts_tx = t_mode_2;    /* ts 4/5 */
                        break;
                case(6):
                        *sec_ts_tx = t_mode_3;    /* ts 6/7 */
                        break;
                default:
                        *sec_ts_tx = t_mode_1;    /* ts 2/3 */
                }
        }
        else {
                /* timeslots configuration for codec (mono) */
                switch(ssi_tx_ts_2){
                case(0):
                        *sec_ts_tx = t_mode_0;    /* ts 0 */
                        break;
                case(2):
                        *sec_ts_tx = t_mode_2;    /* ts 2 */
                        break;
                case(4):
                        *sec_ts_tx = t_mode_1;    /* N/A */
                        break;
                case(6):
                        *sec_ts_tx = t_mode_3;    /* N/A */
                        break;
                default:
                        *sec_ts_tx = t_mode_2;    /* ts 2 */
                }

                /* For record, primary is defined by Tx, and secondary
                   is adjacent */
                switch(ssi_tx_ts_1){
                case(0):
                        *prim_ts = t_mode_0;    /* ts 0 */
                        *sec_ts_rx = t_mode_1;
                        break;
                case(2):
                        *prim_ts = t_mode_2;    /* ts 2 */
                        *sec_ts_rx = t_mode_3;
                        break;
                case(4):
                        *prim_ts = t_mode_1;    /* N/A */
                        *sec_ts_rx = *sec_ts_tx;
                        break;
                case(6):
                        *prim_ts = t_mode_3;    /* N/A */
                        *sec_ts_rx = *sec_ts_tx;
                        break;
                default:
                        *prim_ts = t_mode_0;    /* ts 0 */
                        *sec_ts_rx = *sec_ts_tx;
                }
        }
        
        return 0;
}

/*!
* This function configures the timeslots on the Atlas side (stdac)
* to correspond to the configuration set at the MCU side 
*
* @param        drv_inst [in] info about the current transfer
* @return       0 if success
*
* Modification History:
*       04/08/31        Initial version.
*       04/10/07        Fix ts assignment for stereo recording
*/
static inline int atlas_set_stdac_timeslots(mxc_state_t *drv_inst)
{
        FUNC_START
        TRY ( atlas_audio_st_dac_network_nb_of_timeslots(2) )   /* 4 ts */
        FUNC_END
        return 0;

CATCH:
        FUNC_ERR
        return -1;
}

/*! 
* This function connects the StDAC to the correct SSI (HW dependent)
* Configures also the SSI and DAC parameters 
*
* @param        drv_inst [in] info about the current transfer
* @return       0 if success
*
* Modification History:
*       02 JUNE,2004,   Initial version.
*       14 JUNE,2004,   Cfg addings
*       07/06/04        moved all HW configuration here
*       04/08/12        removed non-SSI related configuration
*                       dam port is no more hard-coded
*       04/08/31        rework config for secondary mixing
*/
int mxc_setup_atlas_stdac(mxc_state_t *drv_inst)
{
        FUNC_START

        TRY ( atlas_audio_st_dac_enable(1) )
        
        /* connect the chip to the right Atlas input port */
#ifdef DRV_FOR_VIRTIO_MODEL
        if ( drv_inst->dam_dest == port_4 ) {
#else
        if ( drv_inst->dam_dest == port_5 ) {
#endif
                TRY ( atlas_audio_st_dac_select_ssi(sb_ssi_1) )
                TRY ( atlas_audio_codec_select_ssi(sb_ssi_0) )
        }
        else {
                TRY ( atlas_audio_st_dac_select_ssi(sb_ssi_0) )
                TRY ( atlas_audio_codec_select_ssi(sb_ssi_1) )
        }

        TRY ( atlas_audio_output_bias_conf(1, 1) )
        TRY ( atlas_audio_st_dac_clk_en(1) )
        TRY ( atlas_audio_st_dac_select_cli(ci_cli_a) )
        TRY ( atlas_audio_st_dac_bit_clock_invert(0) )
        TRY ( atlas_audio_st_dac_frame_sync_invert(0) )
        TRY ( atlas_audio_st_dac_protocol(bp_network) )

        /* timeslots configuration */
        TRY ( atlas_set_stdac_timeslots(drv_inst) )

        FUNC_END
        return 0;

CATCH:
        FUNC_ERR
        return -1;
}

/*!
* This function configures the timeslots on the Atlas side (codec)
* to correspond to the configuration set at the MCU side 
*
* @param        drv_inst [in] info about the current transfer
* @return        0 if success
*
* Modification History:
*       04/08/31        Initial version.
*       04/10/07        Fix ts assignment for stereo recording
*/
static inline int atlas_set_codec_timeslots(mxc_state_t *drv_inst)
{
        timeslots prim_ts, sec_ts_tx, sec_ts_rx;
#ifdef DRV_NTMODE_PLUS_MIX_DISABLED
        mixing_mode mix_mode = mm_no_mixing;
#else
        mixing_mode mix_mode = mm_att_0dB;
#endif
        FUNC_START
        TRY ( get_ssi_timeslots(drv_inst, &prim_ts, &sec_ts_tx, &sec_ts_rx) )

        TRY ( atlas_audio_codec_conf_network_mode(prim_ts) )

        /* Activate mix by setting an attenuation bit. Use adjacent t.s */
        TRY ( atlas_audio_codec_output_mixing_gain(mix_mode) ) 
        TRY ( atlas_audio_codec_conf_network_secondary(sec_ts_tx, sec_ts_rx, \
                mix_mode) )

        FUNC_END
        return 0;

CATCH:
        FUNC_ERR
        return -1;
}

/*! 
* This function connects the StDAC to the correct SSI (HW dependent)
* Configures also the SSI and DAM parameters
*
* @param        drv_inst [in] info about the current transfer
* @return        0 if success
*
* Modification History:
*       02 JUNE,2004,   Initial version.
*       07/06/04        moved all HW configuration here
*       08/07/04        added frame sync length config
*       04/08/12        removed non-SSI related configuration
*                       dam port is no more hard-coded
*       04/08/31        rework config for secondary mixing
*/
int mxc_setup_atlas_codec(mxc_state_t *drv_inst)
{
        FUNC_START

        TRY ( atlas_audio_codec_enable(1) )
        
        /* connect the chip to the right Atlas input port */
#ifdef DRV_FOR_VIRTIO_MODEL
        if ( drv_inst->dam_dest == port_4 ) {
#else
        if ( drv_inst->dam_dest == port_5 ) {
#endif
                TRY ( atlas_audio_codec_select_ssi(sb_ssi_1) )
                TRY ( atlas_audio_st_dac_select_ssi(sb_ssi_0) )
        }
        else {
                TRY ( atlas_audio_codec_select_ssi(sb_ssi_0) )
                TRY ( atlas_audio_st_dac_select_ssi(sb_ssi_1) )
        }

        TRY ( atlas_audio_output_bias_conf(1, 1) )
        TRY ( atlas_audio_codec_select_cli(ci_cli_a) )
        TRY ( atlas_audio_codec_clk_en(1) )
        TRY ( atlas_audio_codec_bit_clock_invert(0) )
        TRY ( atlas_audio_codec_frame_sync_invert(0) )
        TRY ( atlas_audio_codec_protocol(bp_network) )

        /* timeslots configuration */
        TRY ( atlas_set_codec_timeslots(drv_inst) )

#ifdef DRV_FOR_VIRTIO_MODEL
        /*! Now workarounding a bug in the Virtio model */
        TRY ( atlas_audio_st_dac_set_sample_rate(sr_8k) )
#endif

        FUNC_END
        return 0;

CATCH:
        FUNC_ERR
        return -1;
}

/*! 
* This function returns which input device is selected in the audio chip
* (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return       the selected input device, 0 if no input line selected 
*               or available
*
* Modification History:
*       02 MAY,2004, Initial version.
*       02 JUNE,2004, Draft version
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_dev_in_selected(mxc_state_t *drv_inst)
{
        int ret;

        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = drv_inst->card->audio_chip.in_can;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        return ret;
}

/*! 
* This function returns which output device is selected in the audio chip
* (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return       the selected output device, 0 if no input line selected 
*               or available
*
* Modification History:
*       02 MAY,2004, Initial version.
*       02 JUNE,2004, Draft version
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_dev_out_selected(mxc_state_t *drv_inst)
{
        int ret;

        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = drv_inst->card->audio_chip.out_cna;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        return ret;
}

/*! 
* This function selects the input device in the audio chip 
* (Atlas + DAM dependent)
*
* @param        drv_inst [in] info about the current transfer
* @param        dev_in [in] the input device
* @return       the selected input device, 0 if no input line selected 
*               or available
*
* Modification History:
*       02 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_select_dev_in(mxc_state_t *drv_inst, int dev_in)
{
        FUNC_START
        
#ifdef DRV_FOR_VIRTIO_MODEL
        /* disable all chips. Only one available at a time */
        TRY ( atlas_audio_output_mixer_input(false, rs_st_dac) )
        TRY ( atlas_audio_output_mixer_input(false, rs_codec) )
#endif

        /* If the device is already connected, exit */
        spin_lock(&drv_inst->card->drv_inst_lock);
        if ( drv_inst->card->audio_chip.in_can == dev_in ) {
                spin_unlock(&drv_inst->card->drv_inst_lock);
                return dev_in;
        }
        
        switch(dev_in) {
        case(ATLAS_DEVICE_IN_CODEC) :
                DPRINTK("ATLAS_DEVICE_IN_CODEC\n");
                drv_inst->card->audio_chip.in_can = ATLAS_DEVICE_IN_CODEC;
                drv_inst->using_can = ATLAS_DEVICE_IN_CODEC;
                spin_unlock(&drv_inst->card->drv_inst_lock);

                /*! configure the Atlas */
                TRY ( mxc_setup_atlas_codec(drv_inst) )
                TRYEQ ( atlas_enable_in_device(drv_inst, \
                        ATLAS_DEVICE_IN_CODEC, 1), ATLAS_DEVICE_IN_CODEC )

                /*! configure the SSI and the DAM */
                mxc_setup_ssi(drv_inst, ATLAS_PHYSICAL_DEVICE_CODEC);
                mxc_setup_dam(drv_inst);
        
                /*! Tell the MCU or Atlas to be master */
                set_clocking_mastership(drv_inst, ATLAS_PHYSICAL_DEVICE_CODEC);
                break;
        default:
                DPRINTK("Invalid argument\n");
                spin_unlock(&drv_inst->card->drv_inst_lock);
                goto CATCH;
        }

        FUNC_END
        return dev_in;
CATCH:
        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->card->audio_chip.in_can = ATLAS_DEVICE_IN_NONE;
        drv_inst->using_can = ATLAS_DEVICE_IN_NONE;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_ERR
        return 0;
}

/*! 
* This function selects the output device in the audio chip (Atlas dependent)
* Does not change the configuration of the selector.
*
* @param        drv_inst [in] info about the current transfer
* @param        dev_out [in] the output device
* @return       the selected output device, 0 if no input line selected 
*               or available
*
* Modification History:
*       02 MAY,2004, Initial version (stub).
*       03 JUNE,2004, Implementation.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_select_dev_out(mxc_state_t *drv_inst, int dev_out)
{
        FUNC_START

#ifdef DRV_FOR_VIRTIO_MODEL
        /* disable all chips. Only one available at a time */
        TRY ( atlas_audio_output_mixer_input(false, rs_st_dac) )
        TRY ( atlas_audio_output_mixer_input(false, rs_codec) )
#endif

        switch(dev_out) {
        case(ATLAS_DEVICE_OUT_CODEC) :
                DPRINTK("ATLAS_DEVICE_OUT_CODEC\n");
                spin_lock(&drv_inst->card->drv_inst_lock);
                drv_inst->card->audio_chip.out_cna = ATLAS_DEVICE_OUT_CODEC;
                drv_inst->using_cna = ATLAS_DEVICE_OUT_CODEC;
                spin_unlock(&drv_inst->card->drv_inst_lock);
        
                /*! configure the Atlas CODEC */
                TRY ( mxc_setup_atlas_codec(drv_inst) )
                TRYEQ ( atlas_enable_out_device(drv_inst, \
                        ATLAS_DEVICE_OUT_CODEC, 1), ATLAS_DEVICE_OUT_CODEC )

                /*! configure the SSI and the DAM */
                mxc_setup_ssi(drv_inst, ATLAS_PHYSICAL_DEVICE_CODEC);
                mxc_setup_dam(drv_inst);
        
                /*! Tell the MCU or Atlas to be master */
                set_clocking_mastership(drv_inst, ATLAS_PHYSICAL_DEVICE_CODEC);
                break;
        case(ATLAS_DEVICE_OUT_STDAC) :
                DPRINTK("ATLAS_DEVICE_OUT_STDAC\n");
                spin_lock(&drv_inst->card->drv_inst_lock);
                drv_inst->card->audio_chip.out_cna = ATLAS_DEVICE_OUT_STDAC;
                drv_inst->using_cna = ATLAS_DEVICE_OUT_STDAC;
                spin_unlock(&drv_inst->card->drv_inst_lock);
        
                /*! configure the Atlas StDAC */
                TRY ( mxc_setup_atlas_stdac(drv_inst) )
                TRYEQ ( atlas_enable_out_device(drv_inst, \
                        ATLAS_DEVICE_OUT_STDAC, 1), ATLAS_DEVICE_OUT_STDAC )

                /*! configure the SSI and the DAM */
                mxc_setup_ssi(drv_inst, ATLAS_PHYSICAL_DEVICE_STDAC);
                mxc_setup_dam(drv_inst);
        
                /*! Tell the MCU or Atlas to be master */
                set_clocking_mastership(drv_inst, ATLAS_PHYSICAL_DEVICE_STDAC);
                break;
        default :
                DPRINTK("Invalid argument\n");
                goto CATCH;
        }
        
        FUNC_END
        return dev_out;
CATCH:
        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->card->audio_chip.out_cna = ATLAS_DEVICE_OUT_NONE;
        drv_inst->using_cna = ATLAS_DEVICE_OUT_NONE;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_ERR
        return 0;
}

/*! 
* This function returns which input source(s) is (are) selected in the audio 
* chip (Atlas dependent)
* One or more input source(s) may be selected
*
* @param        drv_inst [in] info about the current transfer
* @return       the selected input source(s), 0 if no input source selected 
*               or available
*
* Modification History:
*       19 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_src_in_selected(mxc_state_t *drv_inst)
{
        int ret;
        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = drv_inst->card->audio_chip.in_line;
        spin_unlock(&drv_inst->card->drv_inst_lock);

        return ret;
}

/*! 
* This function returns which output source(s) is (are) selected in 
* the audio chip (Atlas dependent)
* One or more output source(s) may be selected
*
* @param        drv_inst [in] info about the current transfer
* @return       the selected output source(s),
*               0 if no input source selected or available
*
* Modification History:
*       19 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_src_out_selected(mxc_state_t *drv_inst)
{
        int ret;
        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = drv_inst->card->audio_chip.out_line;
        spin_unlock(&drv_inst->card->drv_inst_lock);

        return ret;
}

/*! 
* This function selects the input source in the audio chip (Atlas)
* Only one input source can be selected at a time
*
* @param        drv_inst [in] info about the current transfer
* @param        src_in [in] the input source to select 
*               (if 0 then default all selected src are deselected)
* @return        the selected input source, 0 if no input source selected 
*                or available
*
* Modification History:
*       19 MAY,2004, Initial version.
*       14 JUNE,2004, Added default src selection
*       04/08/06 changed input param interpretation: 0 means disable all
*                clear all enable bits before applying enabling
*       04/08/12: added the stereo input src for handset handling
*       04/10/29: turn both amp1 and amp2 when enabling handset
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_select_src_in(mxc_state_t *drv_inst, int src_in)
{
        FUNC_START

        /*! Disable all before applying new configuration
            thus simplifying the input selection contingency */
        TRY ( atlas_audio_input_device(false, id_handset1) )
        TRY ( atlas_audio_input_device(false, id_handset2) )
        TRY ( atlas_audio_input_device(false, id_headset) )
        TRY ( atlas_audio_input_device(false, id_line_in) )

        if ( src_in == ATLAS_SRC_IN_NONE ) goto _end;

        if ( src_in & ATLAS_SRC_IN_HANDSET ) {
                /* enabling right channel, for mono or stereo */
                TRY ( atlas_audio_input_device(true, id_handset1) )
                TRY ( atlas_audio_input_device(true, id_handset2) )
        }

        /* enabling left channel, for mono only */
        if ( src_in & ATLAS_SRC_IN_HEADSET )
                TRY ( atlas_audio_input_device(true, id_headset) )
        
        /* enabling left channel, for mono only */
        if ( src_in & ATLAS_SRC_IN_LINE_IN )
                TRY ( atlas_audio_input_device(true, id_line_in) )
        
_end:
        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->card->audio_chip.in_line = src_in;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_END
        return src_in;
CATCH:
        FUNC_ERR
        return 0;
}

/*! 
* This function selects the output source in the audio chip (Atlas dependent)
* One or more output source(s) may be selected
* Data path is always taking the mixer, so that the output config. 
* is always using "os_mixer_source"
* Note: if no source out is specified (src_out == 0) then disable all
* Note: Conflicts regarding ASPSEL and ALSPSEL is not handled,
*       but leaved to atlas driver
*
* @param        drv_inst [in] info about the current transfer
* @param        src_out [in] the output source to select 
*               (if 0 then default all selected src are deselected)
* @return       the selected output source, 0 if no input source selected 
*               or available
*
* Modification History:
*       19 MAY,2004, Initial version.
*       14 JUNE,2004, Added default src selection
*       28 JUNE, 2004, Added the disabling possibility 
*       29 JUNE, 2004, Added the stereo->mono conversion
*       04/08/06 changed input param interpretation: 0 means disable all
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_select_src_out(mxc_state_t *drv_inst, int src_out)
{
        FUNC_START

/*! Code deactivated, to fit the Virtio modeling. 2 implementations
    are equivalent - the difference is in the transition state
    and the number of active sources at a time */   
#ifndef DRV_FOR_VIRTIO_MODEL

        if ( src_out & ATLAS_SRC_OUT_EARPIECE ) {
                TRY ( atlas_audio_output_set_conf(od_earpiece, os_mixer_source))
        }
        else {
                TRY ( atlas_audio_output_disable(od_earpiece) ) 
        }
        
        if ( src_out & ATLAS_SRC_OUT_HANDSFREE ) {
                TRY ( atlas_audio_output_set_conf(od_handsfree,os_mixer_source))
        }
        else {
                TRY ( atlas_audio_output_disable(od_handsfree) ) 
        }
        
        if ( src_out & ATLAS_SRC_OUT_HEADSET ) {
                TRY ( atlas_audio_output_set_conf(od_headset, os_mixer_source) )
        }
        else {
                TRY ( atlas_audio_output_disable(od_headset) ) 
        }
        
        if ( src_out & ATLAS_SRC_OUT_LINE_OUT ) {
                TRY ( atlas_audio_output_set_conf(od_line_out, os_mixer_source))
        }
        else {
                TRY ( atlas_audio_output_disable(od_line_out) ) 
        }
        
#else /* DRV_FOR_VIRTIO_MODEL */

        /*! Disable all before activating a src => only one src at a time */
        TRY ( atlas_audio_output_disable(od_earpiece) ) 
        TRY ( atlas_audio_output_disable(od_earpiece) ) 
        TRY ( atlas_audio_output_disable(od_handsfree) ) 
        TRY ( atlas_audio_output_disable(od_headset) ) 
        TRY ( atlas_audio_output_disable(od_line_out) ) 

        if ( src_out & ATLAS_SRC_OUT_EARPIECE ) {
                TRY ( atlas_audio_output_set_conf(od_earpiece, os_mixer_source))
        }
        else if ( src_out & ATLAS_SRC_OUT_HANDSFREE ) {
                TRY ( atlas_audio_output_set_conf(od_handsfree,os_mixer_source))
        }
        else if ( src_out & ATLAS_SRC_OUT_HEADSET ) {
                TRY ( atlas_audio_output_set_conf(od_headset, os_mixer_source))
        }
        else if ( src_out & ATLAS_SRC_OUT_LINE_OUT ) {
                TRY ( atlas_audio_output_set_conf(od_line_out, os_mixer_source))
        }

#endif /* DRV_FOR_VIRTIO_MODEL */

#ifdef CONFIG_SOUND_MXC_FIFO
        /* Always force the mono adding for performance reasons */
        if ( /*1*/src_out & MXC_MONO_OUT_SRC_MASK ) {
                TRYNEQ ( atlas_set_output_adder(drv_inst, \
                         ATLAS_AUDIO_ADDER_MONO), -1 )
        }
#else
        /* Force the mono adding if playing on at least one mono device */
        if ( src_out & MXC_MONO_OUT_SRC_MASK ) {
                TRYNEQ ( atlas_set_output_adder(drv_inst, \
                         ATLAS_AUDIO_ADDER_MONO), -1 )
        }
#endif
        else {
                TRYNEQ ( atlas_set_output_adder(drv_inst,
#if defined(CONFIG_ARCH_ARGONPLUS)
                         ATLAS_AUDIO_ADDER_STEREO_OPPOSITE), -1 )
#else
                         ATLAS_AUDIO_ADDER_STEREO), -1 )
#endif
        }

        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->card->audio_chip.out_line = src_out;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_END
        return src_out;
CATCH:
        FUNC_ERR
        return 0;
}

/*! 
* This function handles the OSS command that sets the clock master
*
* @param    drv_inst [in] the device's major number
* @param    clock    [in] the clock master to set 
* (MXC_AUDIO_CLOCKING_ATLAS_MASTER or MXC_AUDIO_CLOCKING_MCU_MASTER)
*
* @return   the value set in the device
*           (may be different that the one requested)
*/
int mxc_setup_clk_master(mxc_state_t *drv_inst, unsigned int clock)
{
        unsigned int clk;
        FUNC_START

        switch(clock) {
        case MXC_AUDIO_CLOCKING_MCU_MASTER:
                drv_inst->clock_master = MXC_AUDIO_CLOCKING_MCU_MASTER;
                if (drv_inst->card->connected_device == ATLAS_PHYSICAL_DEVICE_STDAC) {
                        set_clocking_mastership(drv_inst, ATLAS_PHYSICAL_DEVICE_STDAC);
                } else {
                        set_clocking_mastership(drv_inst, ATLAS_PHYSICAL_DEVICE_CODEC);
		}
                clk = clock;
                break;
        default:
                drv_inst->clock_master = MXC_AUDIO_CLOCKING_ATLAS_MASTER;
                if (drv_inst->card->connected_device == ATLAS_PHYSICAL_DEVICE_STDAC) {
                        set_clocking_mastership(drv_inst, ATLAS_PHYSICAL_DEVICE_STDAC);
                } else {
                        set_clocking_mastership(drv_inst, ATLAS_PHYSICAL_DEVICE_CODEC);
		}
                clk = clock;
                break;
        }

        FUNC_END
        return clk;
}

/*! 
* This function handles the OSS command that sets the dam default output port
* and must be called before launching the transfer
*
* @param    drv_inst [in] the device's major number
* @param    out_port [in] the dam default output number to set 
*                        (MXC_DAM_OUT_PORT_AD1:port_4 else port_5 (AD2))  
*
* @return   the value set in the device
*           (may be different that the one requested)
*/
int mxc_setup_dam_out_port(mxc_state_t *drv_inst, unsigned int out_port)
{
        unsigned int port;
        FUNC_START

        switch(out_port) {
        case MXC_DAM_OUT_PORT_AD1:
                drv_inst->dam_dest = port_4;
                port = out_port;
                break;
        default:
                drv_inst->dam_dest = port_5;
                port = out_port;
                break;
        }

        FUNC_END
        return port;
}

/*! 
* This function selects the input device in the audio chip (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer 
* @param        vol [in] the volume to set
* @return       vol the volume effectively set
*
* Modification History:
*       02 MAY,2004, Initial version.
*       04/10/04: Now returns value applied to HW
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_set_dev_in_volume(mxc_state_t *drv_inst, int vol)
{
        int vol_left_db;
        int vol_right_db;
        int vol_left = vol & 0xFF;
        int vol_right = (vol & 0xFF00)>>8;
        int ret;
        FUNC_START
        
        if ( vol_left > ATLAS_IN_VOL_CODE_MAX )
                vol_left = ATLAS_IN_VOL_CODE_MAX;
        if ( vol_left < ATLAS_IN_VOL_CODE_MIN ) 
                vol_left = ATLAS_IN_VOL_CODE_MIN;
        if ( vol_right > ATLAS_IN_VOL_CODE_MAX ) 
                vol_right = ATLAS_IN_VOL_CODE_MAX;
        if ( vol_right < ATLAS_IN_VOL_CODE_MIN ) 
                vol_right = ATLAS_IN_VOL_CODE_MIN;
        
        vol_left_db = CONVERT_IN_VOL_OSS_TO_HW(vol_left);
        vol_right_db = CONVERT_IN_VOL_OSS_TO_HW(vol_right);

        DPRINTK("--> Setting gain : left = %d, right = %d\n", \
                 vol_left_db, vol_right_db);
        TRY ( atlas_audio_input_set_gain(vol_left_db, vol_right_db) )
        
        ret = (vol_right << 8) | vol_left;
        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->card->audio_chip.in_volume = ret;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_END
        return ret;

CATCH:
        FUNC_ERR
        return -1;

}

/*! 
* This function sets the volume of the output device in the audio chip 
* (Atlas dependent)
* PGA gain range is identical for Codec (Left) and StDAC (L+R). 
* Note: Only left value is taken into account
*
* @param        drv_inst [in] info about the current transfer
* @param        vol [in] the volume to set (L+R coded in a word)
* @return       vol the volume effectively set, -1 if error
* @see          atlas_set_output_balance
*
* Modification History:
*       02 MAY,2004, Initial version.
*       04/10/04: Now returns value applied to HW
*       04/11/19: TLSbo44269: mixer rework
*       04/12/10: TLSbo : integrate bal in volume
*/
int atlas_set_dev_out_volume(mxc_state_t *drv_inst, int vol)
{
        int vol_left_db, vol_right_db, delta, attenuation;
        receive_signal cna;
        int vol_right = vol & 0xFF, vol_left = (vol&0xFF00) >> 8;
        int ret, atlas_cna;
        FUNC_START
        
        if ( vol_left < ATLAS_OUT_VOL_CODE_MIN )
                vol_left = ATLAS_OUT_VOL_CODE_MIN;
        if ( vol_left > ATLAS_OUT_VOL_CODE_MAX ) 
                vol_left = ATLAS_OUT_VOL_CODE_MAX;
        if ( vol_right < ATLAS_OUT_VOL_CODE_MIN )
                vol_right = ATLAS_OUT_VOL_CODE_MIN;
        if ( vol_right > ATLAS_OUT_VOL_CODE_MAX ) 
                vol_right = ATLAS_OUT_VOL_CODE_MAX;
        
        vol_left_db = CONVERT_OUT_VOL_OSS_TO_HW(vol_left);
        vol_right_db = CONVERT_OUT_VOL_OSS_TO_HW(vol_right);
        delta = vol_left - vol_right;

        spin_lock(&drv_inst->card->drv_inst_lock);
        atlas_cna = drv_inst->card->audio_chip.out_cna;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        TRY ( mxc_convert_receive_signal(atlas_cna, &cna) )

        if ( delta == 0 ) {
                DPRINTK("--> Setting volume : left/right = %d\n", vol_left_db);
                TRY ( atlas_audio_output_set_volume(cna, vol_left_db) )
        }
        else if ( delta > 0 ) {
                attenuation = delta * ATLAS_VOL_TO_BAL_RATIO;
                if ( attenuation > ATLAS_OUT_NO_BAL_CODE )
                        attenuation = ATLAS_OUT_NO_BAL_CODE;
                vol_right = vol_left - attenuation/ATLAS_VOL_TO_BAL_RATIO;
                attenuation = CONVERT_BAL_OSS_TO_HW(attenuation);
                DPRINTK("--> Setting volume : left = %d\n", vol_left_db);
                DPRINTK("--> Setting balance : right = %d\n", attenuation);
                TRY ( atlas_audio_output_set_volume(cna, vol_left_db) )
                TRY ( atlas_audio_output_set_balance(c_right, attenuation) )
        }
        else {
                delta = (-delta);
                attenuation = delta * ATLAS_VOL_TO_BAL_RATIO;
                if ( attenuation > ATLAS_OUT_NO_BAL_CODE )
                        attenuation = ATLAS_OUT_NO_BAL_CODE;
                vol_left = vol_right - attenuation/ATLAS_VOL_TO_BAL_RATIO;
                attenuation = CONVERT_BAL_OSS_TO_HW(attenuation);
                DPRINTK("--> Setting volume : right = %d\n", vol_right_db);
                DPRINTK("--> Setting balance : left = %d\n", attenuation);
                TRY ( atlas_audio_output_set_volume(cna, vol_right_db) )
                TRY ( atlas_audio_output_set_balance(c_left, attenuation) )
        }

        ret = (vol_left << 8) | vol_right;
        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->card->audio_chip.out_volume = ret;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_END
        return ret;

CATCH:
        FUNC_ERR
        return -1;
}

/*! 
* This function returns the input device in the audio chip (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return       the current volume
*
* Modification History:
*       02 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_dev_in_volume(mxc_state_t *drv_inst)
{
        int ret;
        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = drv_inst->card->audio_chip.in_volume;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        return ret;
}

/*! 
* This function returns the output device in the audio chip (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return        the current volume (L+R coded in a word)
*
* Modification History:
*       02 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_dev_out_volume(mxc_state_t *drv_inst)
{
        int ret;
        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = drv_inst->card->audio_chip.out_volume;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        return ret;
}

/*! 
* This function sets the input bias configuration (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @param        onoff (0 if "off", other values "on")
* @return        
*
* Modification History:
*         18 MAY,2004, Initial version.
*/
int atlas_set_input_bias(mxc_state_t *drv_inst, int onoff)
{
        return 0;
}

/*! 
* This function gets the input bias configuration (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return       onoff (0 if "off", other values "on")
*
* Modification History:
*         18 MAY,2004, Initial version.
*/
int atlas_get_input_bias(mxc_state_t *drv_inst)
{
        return 0;
}

/*! 
* This function sets the output balance configuration (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @param        val     (0 to 49 = left bal, 50=disable, 
                        51 to 100 right bal, balance from -21dB to 0)
* @return        the new output balance, -1 if error
* @see           atlas_get_output_balance
*
* Modification History:
*       18 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_set_output_balance(mxc_state_t *drv_inst, int val)
{
        int attenuation;
        channel chan;
        FUNC_START
        
        if ( val < ATLAS_OUT_BAL_CODE_MIN ) val = ATLAS_OUT_BAL_CODE_MIN;
        if ( val > ATLAS_OUT_BAL_CODE_MAX ) val = ATLAS_OUT_BAL_CODE_MAX;
        
        if ( val <= ATLAS_OUT_NO_BAL_CODE ) {
                chan = c_left;
                attenuation = CONVERT_BAL_OSS_TO_HW(ATLAS_OUT_NO_BAL_CODE-val);
        }
        else {
                chan = c_right;
                attenuation = CONVERT_BAL_OSS_TO_HW(val-ATLAS_OUT_NO_BAL_CODE);
        }
        
        DPRINTK("--> Setting balance : channel = %d, val = %d\n", \
                chan, attenuation);
        TRY ( atlas_audio_output_set_balance(chan, attenuation) )
        
        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->card->audio_chip.out_balance = val;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_END
        return val;

CATCH:
        FUNC_ERR
        return -1;
}

/*! 
* This function gets the output balance (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return       the current output balance, -1 if error
* @see          atlas_set_output_balance
*
* Modification History:
*       18 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_output_balance(mxc_state_t *drv_inst)
{
        int ret;
        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = drv_inst->card->audio_chip.out_balance;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        return ret;
}

/*! 
* This function switches on/off the device (codec or stdac) 
* TODO: do not switch off the codec if other direction still in use
*
* @param        device [in] the target device
* @param        on_off [in] to enable the device
* @return       0 if success, -1 if error
* @see          atlas_enable_out_device, atlas_enable_in_device
*
* Modification History:
*       07 JUNE,2004, Initial version.
*/
int atlas_device_switch_on_off(int device, int on_off)
{
        FUNC_START
        
        switch (device) {
        case(ATLAS_DEVICE_OUT_CODEC):
        case(ATLAS_DEVICE_IN_CODEC):
                TRY ( atlas_audio_codec_enable(on_off) )
                TRY ( atlas_audio_codec_clk_en(on_off) )
                break;
        case(ATLAS_DEVICE_OUT_STDAC):
                TRY ( atlas_audio_st_dac_enable(on_off) )
                TRY ( atlas_audio_st_dac_clk_en(on_off) )
                break;
        default:
                DPRINTK("Invalid argument\n");
                goto CATCH;
        }
        
        FUNC_END
        return 0;
CATCH:
        FUNC_ERR
        return -1;
}

/*! 
* This function checks if the Codec or StDac is still in use, 
* before disabling it
*
* @param        drv_inst [in] info about the current transfer
* @param        device [in] the input CNA to disable
* @return       0 if device can be disabled
* @see          atlas_enable_out_device atlas_enable_in_device
*
* Modification History:
*       08 JUNE,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_check_device_still_in_use(mxc_state_t *drv_inst, int device)
{
        int still_in_use = 0;
        int i, in_device, out_device;
        unsigned long flags;
        FUNC_START
        
        switch (device) {
        case(ATLAS_DEVICE_OUT_STDAC) : 
                out_device = ATLAS_DEVICE_OUT_STDAC;
                in_device = -1;        /*! or any value that is not valid */
                break;
        case(ATLAS_DEVICE_OUT_CODEC) :
        case(ATLAS_DEVICE_IN_CODEC) :
                out_device = ATLAS_DEVICE_OUT_CODEC;
                in_device = ATLAS_DEVICE_IN_CODEC;
                break;
        default:
                DPRINTK("Sound driver error: device not handled\n");
                return 0;
        }

        spin_lock_irqsave(&drv_inst->card->drv_inst_lock, flags);
        for ( i=0, still_in_use=0; i<NR_HW_CH; i++ ) {
                if ( drv_inst->card->state[i]->using_cna == \
                        out_device ) still_in_use++;
                if ( drv_inst->card->state[i]->using_can == \
                        in_device )  still_in_use++;
        }
        spin_unlock_irqrestore(&drv_inst->card->drv_inst_lock, flags);
        
        if ( still_in_use ) DPRINTK("Chip still in use\n");
        FUNC_END
        return still_in_use;
}

/*! 
* This function enables the output (HW connection). 
* Does not connect/disconnect the device to the SSI port.
*
* @param        drv_inst [in] info about the current transfer
* @param        device [in] the input CNA to enable
* @param        en_flag [in] to add/remove the input signal to/from mixer
* @return       the device enabled/disabled, -1 if error
*               If the device is not in use, then return ATLAS_DEVICE_OUT_NONE
* @see          atlas_enable_in_device
*
* Modification History:
*       18 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_enable_out_device(mxc_state_t *drv_inst, int device, int en_flag)
{
        receive_signal rs;
        int still_in_use;
        FUNC_START
        
        TRY ( mxc_convert_receive_signal(device, &rs) )  

        if ( en_flag ) {
                DPRINTK("Connect output device\n");        
                TRY ( atlas_audio_output_mixer_input(true, rs) )
                TRY ( atlas_audio_output_pga(true, rs) )
#ifdef MXC_DRV_STANDALONE
                TRYEQ ( atlas_set_dev_out_volume(drv_inst, 0x5050), 0x5050 )
                TRYEQ ( mxc_set_speed(drv_inst, 8000), 8000 )
                TRYEQ ( mxc_set_channels(drv_inst, 1), 1 )
                TRYEQ ( atlas_set_output_balance(drv_inst, 50), 50 )
#endif
                TRY ( atlas_device_switch_on_off(device, 1) )
        }
        else {
                drv_inst->using_cna = ATLAS_DEVICE_OUT_NONE;
                
                /* Check if device still used by other drv instance */
                still_in_use=atlas_check_device_still_in_use(drv_inst, device);

                if ( !still_in_use ) {
                        DPRINTK("Free output device\n");
                        TRY ( atlas_audio_output_mixer_input(false, rs) )
                        TRY ( atlas_device_switch_on_off(device, 0) )
                        FUNC_END
                        return ATLAS_DEVICE_OUT_NONE;
                }
        }

        FUNC_END
        return device;
CATCH:
        drv_inst->using_cna = ATLAS_DEVICE_OUT_NONE;
        FUNC_ERR
        return -1;
}

/*! 
* This function enables the input (HW connection). 
* Does not connect/disconnect the device to the SSI port.
*
* @param        drv_inst [in] info about the current transfer
* @param        device [in] the input CNA to enable
* @param        en_flag [in] to add/remove the input signal to/from mixer
* @return       the device enabled/disabled, -1 if error
*               If the device is not in use, then return ATLAS_DEVICE_IN_NONE
* @see          atlas_enable_out_device
*
* Modification History:
*       18 MAY,2004, Initial version.
*       06/07/04:    Correction. No functional difference
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_enable_in_device(mxc_state_t *drv_inst, int device, int en_flag)
{
        int still_in_use;
        FUNC_START

        if ( en_flag ) {
                DPRINTK("Connect input device\n");
#ifdef MXC_DRV_STANDALONE
                TRYEQ ( atlas_set_dev_in_volume(drv_inst, 0x2020), 0x2020 )
                TRYEQ ( mxc_set_speed(drv_inst, 8000), 8000 )
                TRYEQ ( mxc_set_channels(drv_inst, 1), 1 )
                TRYNEQ ( atlas_set_codec_filter(drv_inst, \
                         ATLAS_CODEC_FILTER_DISABLE), -1 )
#endif
                TRY ( atlas_device_switch_on_off(device, 1) )
        }
        else {
                drv_inst->using_can = ATLAS_DEVICE_IN_NONE;
                
                /* Check if device still used by other drv instance */
                still_in_use=atlas_check_device_still_in_use(drv_inst, device);

                if ( !still_in_use ) {
                        DPRINTK("Free input device\n");
                        TRY ( atlas_device_switch_on_off(device, 0) )
                        FUNC_END
                        return ATLAS_DEVICE_IN_NONE;
                }
        }

        FUNC_END
        return device;
CATCH:
        drv_inst->using_can = ATLAS_DEVICE_IN_NONE;
        FUNC_ERR
        return -1;
}

/*! 
* This function gets the output mixer configuration (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return        the current mixer configuration, -1 if error
* @see                atlas_set_output_mixer
*
* Modification History:
*         18 MAY,2004, Initial version.
*/
int atlas_get_output_mixer(mxc_state_t *drv_inst)
{
        return -1;
}

/*! 
* This function sets the output mixer configuration (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return        the current mixer configuration, -1 if error
* @see                atlas_get_output_mixer
*
* Modification History:
*         18 MAY,2004, Initial version.
*/
int atlas_set_output_mixer(mxc_state_t *drv_inst)
{
        return -1;
}

/*! 
* This function sets the output mono adder configuration (Atlas dependent)
* ALWAYS set to ac_stereo.
*
* @param        drv_inst [in] info about the current transfer
* @param        val the working mode of the adder
* @return        the new mono adder configuration, -1 if error
* @see                atlas_get_output_adder
*
* Modification History:
*       19 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_set_output_adder(mxc_state_t *drv_inst, int val)
{
        FUNC_START
        
        switch (val) {
        case (ATLAS_AUDIO_ADDER_STEREO) :
                TRY ( atlas_audio_output_config_mixer(ac_stereo) )
                break;
        case (ATLAS_AUDIO_ADDER_STEREO_OPPOSITE) :
                TRY ( atlas_audio_output_config_mixer(ac_stereo_opposit) )
                break;
        case (ATLAS_AUDIO_ADDER_MONO) :
                TRY ( atlas_audio_output_config_mixer(ac_mono) )
                break;
        case (ATLAS_AUDIO_ADDER_MONO_OPPOSITE) :
                TRY ( atlas_audio_output_config_mixer(ac_mono_opposit) )
                break;
        default :
                DPRINTK("Invalid parameter\n");
                goto CATCH;
        }
        
        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->card->audio_chip.out_adder_mode = val;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_END
        return val;

CATCH:
        FUNC_ERR
        return -1;
}

/*! 
* This function gets the output mixer configuration (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return        the current mono adder configuration, -1 if error
* @see                atlas_set_output_adder
*
* Modification History:
*       19 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_output_adder(mxc_state_t *drv_inst)
{
        int ret;
        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = drv_inst->card->audio_chip.out_adder_mode;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        return ret;
}

/*! 
* This function returns true if output src is connected
*
* @param        drv_inst [in] info about the current transfer
* @param        src_mask [in] the playing source
* @return       0 if not connected
*
* Modification History:
*       04/09/14, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
static inline int is_src_out_connected(mxc_state_t *drv_inst, int src_mask)
{
        int ret;
        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = (drv_inst->card->audio_chip.out_line & src_mask);
        spin_unlock(&drv_inst->card->drv_inst_lock);
        return ret;
}

/*! 
* This function returns true if input src is connected
*
* @param        drv_inst [in] info about the current transfer
* @param        src_mask [in] the recording source to set
* @return       0 if not connected
*
* Modification History:
*       04/09/14, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
static inline int is_src_in_connected(mxc_state_t *drv_inst, int src_mask)
{
        int ret;
        spin_lock(&drv_inst->card->drv_inst_lock);
        ret = (drv_inst->card->audio_chip.in_line & src_mask);
        spin_unlock(&drv_inst->card->drv_inst_lock);
        return ret;
}

/*! 
* This function sets the output volume of the given source (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @param        src_mask [in] the playing source to set
* @param        val        [in] the new output volume
* @return       the new output volume, -EINVAL if source inactive
*
* Modification History:
*       19 MAY,2004, Initial version.
*       04/09/14, the code testing if src is connected moved
*/
int atlas_set_src_out_volume(mxc_state_t *drv_inst, int src_mask, int val)
{
        /*! Check if the device is connected to our SSI */
        if ( !is_src_out_connected(drv_inst, src_mask) ) return -EINVAL;
        
        return atlas_set_dev_out_volume(drv_inst, val);
}

/*! 
* This function sets the input volume of the given source (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @param        src_mask [in] the recording source to set
* @param        val        [in] the new output volume
* @return       the new input volume, -EINVAL if source inactive
*
* Modification History:
*       19 MAY,2004, Initial version.
*       04/09/14, the code testing if src is connected moved
*/
int atlas_set_src_in_volume(mxc_state_t *drv_inst, int src_mask, int val)
{
        /*! Check if the device is connected to our SSI */
        if ( !is_src_in_connected(drv_inst, src_mask) ) return -EINVAL;

        return atlas_set_dev_in_volume(drv_inst, val);
}

/*! 
* This function gets the output volume of the given source (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @param        src_mask [in] the playing source to set
* @return        the output volume, -EINVAL if source inactive
*
* Modification History:
*       19 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_src_out_volume(mxc_state_t *drv_inst, int src_mask)
{
        /*! Check if the device is connected to our SSI */
        if ( !is_src_out_connected(drv_inst, src_mask) ) return -EINVAL;
        
        return atlas_get_dev_out_volume(drv_inst);
}

/*! 
* This function gets the input volume of the given source (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @param        src_mask [in] the recording source to set
* @return        the input volume, -EINVAL if source inactive
*
* Modification History:
*       19 MAY,2004, Initial version.
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_src_in_volume(mxc_state_t *drv_inst, int src_mask)
{
        /*! Check if the device is connected to our SSI */
        if ( !is_src_in_connected(drv_inst, src_mask) ) return -EINVAL;

        return atlas_get_dev_in_volume(drv_inst);
}

/*! 
* This function sets the input volume of the given source (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @param        mask [in] the new filter configuration
* @return       the applied config if success
*
* Modification History:
*       15 JUNE,2004, Initial version.
*       28 JUNE, 2004, Added the disabling possibility
*       04/08/03: corrected the bugged disabling case
*       04/08/19: added check on the connected device
*       04/10/04: returns value applied to HW
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_set_codec_filter(mxc_state_t *drv_inst, int mask)
{
        int l_mask;
        unsigned long flags;
        FUNC_START

        /*! Apply command only if Codec is connected */
        spin_lock_irqsave(&drv_inst->card->drv_inst_lock, flags);
        l_mask = drv_inst->card->audio_chip.in_out_filter;

        switch ( drv_inst->card->connected_device ) {
                case ATLAS_PHYSICAL_DEVICE_CODEC :
                        break;
                default:
                        spin_unlock_irqrestore(&drv_inst->card->drv_inst_lock, \
                        flags);
                        goto CATCH;
        }
                
        spin_unlock_irqrestore(&drv_inst->card->drv_inst_lock, flags);

        /*! Apply the config */
        if ( mask & ATLAS_CODEC_FILTER_HIGH_PASS_IN ) {
                TRY ( atlas_audio_codec_receiv_high_pass_filter_en(true) )
                l_mask |= ATLAS_CODEC_FILTER_HIGH_PASS_IN;
        }
        else {
                TRY ( atlas_audio_codec_receiv_high_pass_filter_en(false) )
                l_mask &= (~(int)ATLAS_CODEC_FILTER_HIGH_PASS_IN);
        }

        if ( mask & ATLAS_CODEC_FILTER_HIGH_PASS_OUT ) {
                TRY ( atlas_audio_codec_trans_high_pass_filter_en(true) )
                l_mask |= ATLAS_CODEC_FILTER_HIGH_PASS_OUT;
        }
        else {
                TRY ( atlas_audio_codec_trans_high_pass_filter_en(false) )
                l_mask &= (~(int)ATLAS_CODEC_FILTER_HIGH_PASS_OUT);
        }

        if ( mask & ATLAS_CODEC_FILTER_DITHERING ) {
                TRY ( atlas_audio_codec_dithering_en(true) )
                l_mask |= ATLAS_CODEC_FILTER_DITHERING;
        } 
        else {
                TRY ( atlas_audio_codec_dithering_en(false) )
                l_mask &= (~(int)ATLAS_CODEC_FILTER_DITHERING);
        }

        spin_lock(&drv_inst->card->drv_inst_lock);
        drv_inst->card->audio_chip.in_out_filter = l_mask;
        spin_unlock(&drv_inst->card->drv_inst_lock);
        FUNC_END
        return l_mask;
CATCH:
        FUNC_ERR
        return -1;
}

/*! 
* This function gets the input volume of the given source (Atlas dependent)
*
* @param        drv_inst [in] info about the current transfer
* @return        the filter configuration of the codec
*
* Modification History:
*       15 JUNE,2004, Initial version.
*       04/10/04: added check on the connected device 
*       04/11/19: TLSbo44269: mixer rework
*/
int atlas_get_codec_filter(mxc_state_t *drv_inst)
{
        unsigned long flags;
        int ret;

        /*! Apply command only if Codec is connected */
        spin_lock_irqsave(&drv_inst->card->drv_inst_lock, flags);

        switch ( drv_inst->card->connected_device ) {
                case ATLAS_PHYSICAL_DEVICE_CODEC :
                        break;
                default:
                        spin_unlock_irqrestore(&drv_inst->card->drv_inst_lock, \
                        flags);
                        goto CATCH;
        }

        ret = drv_inst->card->audio_chip.in_out_filter;
        spin_unlock_irqrestore(&drv_inst->card->drv_inst_lock, flags);

        return ret;
CATCH:
        FUNC_ERR
        return -1;
}

#ifdef CONFIG_ARCH_SCMA11EVB
#define SCMA11_EVB_PASS1 1
/* See scma11evb_gpio.c to complete the workaround for SSI scma11 pass1 */
#endif
/*! 
* This function sets the clocking in the DAM (DAM dependent)
*
* @param        target_port the port to configure
* @param        source_port the port to connect to
* @param        dir the data flow direction
* @return        None
*
* Modification History:
*       02 JUNE,2004, Draft version
*       05/03/10: split connection data/clock
*/
void mxc_dam_clock_connect(dam_port target_port, 
                              dam_port source_port,
                              signal_direction dir)
{
        FUNC_START

#ifdef SCMA11_EVB_PASS1

 	if (target_port == port_4){
                dam_reset_register(port_1);
                dam_reset_register(port_2);
                dam_reset_register(port_3);
                dam_reset_register(port_4);
                dam_reset_register(port_5);
                dam_reset_register(port_6);
                dam_reset_register(port_7);
 		
                dam_select_mode(source_port, normal_mode);
                dam_select_mode(port_4, normal_mode);
                
		dam_set_synchronous(source_port, true);
                dam_set_synchronous(port_4, true);
                
		dam_select_RxD_source(port_4, source_port);
                dam_select_RxD_source(source_port, port_4);
                
		dam_select_TxFS_direction(port_4, signal_out);
                dam_select_TxFS_source(port_4, false, source_port);

                dam_reset_register(port_7);
                dam_select_mode(port_7, normal_mode);
		dam_set_synchronous(port_7, false);
        	dam_select_RxClk_source(port_7, false, source_port);
                dam_select_RxClk_direction(port_7, signal_out);
                
	}else if (target_port == port_5){
                dam_reset_register(port_1);
                dam_reset_register(port_2);
                dam_reset_register(port_3);
                dam_reset_register(port_4);
                dam_reset_register(port_5);
                dam_reset_register(port_6);
                dam_reset_register(port_7);

                dam_set_synchronous(source_port, true);
		
                dam_select_RxD_source(source_port, port_5);
                dam_select_RxD_source(port_5, source_port);

                dam_select_TxFS_direction(port_5, signal_out);
                dam_select_TxFS_source(port_5, false, source_port);

		dam_set_synchronous(port_5, false);
        	dam_select_RxClk_source(port_5, false, source_port);
                dam_select_RxClk_direction(port_5, signal_out);
        }
#else
        dam_select_RxFS_source(target_port, true, source_port);
        dam_select_RxFS_direction(target_port, dir);
        dam_select_RxClk_source(target_port, true, source_port);
        dam_select_RxClk_direction(target_port, dir);

        dam_select_TxFS_source(target_port, false, source_port);
        dam_select_TxFS_direction(target_port, dir);
        dam_select_TxClk_source(target_port, false, source_port);
        dam_select_TxClk_direction(target_port, dir);

#endif //SCMA11_EVB_PASS1

        FUNC_END
}

/*! 
* This function sets the data connection in the DAM (DAM dependent)
*
* @param        target_port the port to configure
* @param        source_port the port to connect to (ignore if network mode)
* @param        network_mode 0: normal, else the network configuration
* @return        None
*
* Modification History:
*       02 JUNE,2004, Draft version
*       05/03/10: split connection data/clock
*/
void mxc_dam_data_connect(dam_port target_port, 
                              dam_port source_port,
                              int network_mode)
{
        FUNC_START

        if ( !network_mode ) {
                dam_select_RxD_source(target_port, source_port);
        }
        else {
                dam_select_mode(target_port, internal_network_mode);
                dam_set_internal_network_mode_mask(target_port, network_mode);
        }

        FUNC_END
}

/*!
* This function sets the DAM configuration when Atlas is master
* @param        drv_inst [in] info about the current transfer
* @param        target_port the port to configure
* @param        source_port the port to connect to (ignore if network mode)
* @return       0 in case of success
*
* Modification History:
*       05/03/10: Creation for handling clocking mastership (MCU or Atlas)
*/
int mxc_setup_dam_mastership(mxc_state_t *drv_inst, dam_port target_port, 
                                dam_port source_port)
{
        FUNC_START

        /*! configure the MCU-DAM-side connection */
        mxc_dam_clock_connect(source_port, target_port, signal_in);
        mxc_dam_data_connect(source_port, target_port, 0);

        /*! configure the DAM-Atlas-side connection */
        mxc_dam_clock_connect(target_port, source_port, signal_out);
        mxc_dam_data_connect(target_port, source_port, 0);

        FUNC_END
        return 0;
}

/*!
* This function sets the DAM configuration when Atlas is master
* @param        drv_inst [in] info about the current transfer
* @return       0 in case of success
*
* Modification History:
*       05/03/10: Creation for handling clocking mastership (MCU or Atlas)
*/
int mxc_setup_dam_atlas_master(mxc_state_t *drv_inst)
{
        dam_port target_port;
        dam_port source_port;
        int dest_mask;
        FUNC_START

        target_port = drv_inst->dam_dest;
        source_port = (drv_inst->ssi_index == SSI1 ? port_1 : port_2);

#ifdef DRV_NTMODE_PLUS_MIX_DISABLED
        mxc_setup_dam_mastership(drv_inst, source_port, target_port);
#else
        if ( target_port == port_4 )    dest_mask = 0xF7;
        else                            dest_mask = 0xEF;

        /*! internal connection for the mix: ports 1 and 2 to "target_port" */
        /*! Reference clock comes from "target_port" */
        mxc_dam_clock_connect(target_port, source_port, signal_in);
        mxc_dam_data_connect(target_port, 0, 0xFC);

        /*! internal connection for the mix: "target_port" to port 1 */
        mxc_dam_clock_connect(port_1, target_port, signal_out);
        mxc_dam_data_connect(port_1, 0, dest_mask);

        /*! internal connection for the mix: "target_port" to port 2 */
        mxc_dam_clock_connect(port_2, target_port, signal_out);
        mxc_dam_data_connect(port_2, 0, dest_mask);
#endif
        FUNC_END
        return 0;
}

/*!
* This function sets the DAM configuration when MCU is master
* @param        drv_inst [in] info about the current transfer
* @return       0 in case of success
*
* Modification History:
*       05/03/10: Creation for handling clocking mastership (MCU or Atlas)
*/
int mxc_setup_dam_mcu_master(mxc_state_t *drv_inst)
{
        dam_port target_port;
        dam_port source_port;
        int dest_mask;
        FUNC_START

        target_port = drv_inst->dam_dest;
        source_port = (drv_inst->ssi_index == SSI1 ? port_1 : port_2);

#ifdef DRV_NTMODE_PLUS_MIX_DISABLED
        /* inverting the src and dest */
        mxc_setup_dam_mastership(drv_inst, target_port, source_port);
#else
        if ( target_port == port_4 )    dest_mask = 0xF7;
        else                            dest_mask = 0xEF;

        /*! internal connection for the mix: ports 1 and 2 to "target_port" */
        /*! Reference clock comes from SSI1, goes to SSI2 and "target_port" */
        mxc_dam_clock_connect(target_port, source_port, signal_out);
        mxc_dam_data_connect(target_port, 0, 0xFC);

        if ( source_port == SSI1 ) {
                /*! internal connection for the mix: "target_port" to port 1 */
                mxc_dam_clock_connect(port_1, target_port, signal_in);
                mxc_dam_data_connect(port_1, 0, dest_mask);

                /*! internal connection for the mix: "target_port" to port 2 */
                mxc_dam_clock_connect(port_2, target_port, signal_out);
                mxc_dam_data_connect(port_2, 0, dest_mask);
        }
        else {
                /*! internal connection for the mix: "target_port" to port 1 */
                mxc_dam_clock_connect(port_1, target_port, signal_out);
                mxc_dam_data_connect(port_1, 0, dest_mask);

                /*! internal connection for the mix: "target_port" to port 2 */
                mxc_dam_clock_connect(port_2, target_port, signal_in);
                mxc_dam_data_connect(port_2, 0, dest_mask);
        }
#endif

        FUNC_END
        return 0;
}

/*! 
* This function sets the default configuration to the DAM (DAM dependent)
*         Connected to MCU are "port1" and "port2" (not configurable)
*         Connected to Atlas is "port4/AD1" or "port5/AD2" (not configurable)
*         Codec or StDac can be connected to port4/port5 or port5/port4
*
* @param        drv_inst [in] info about the current transfer
* @return       0 in case of success
*
* Modification History:
*       02 JUNE,2004, Draft version
*       04/08/12, added the dam destination port as drv parameter
*       04/09/21, enhancement for DRV_NTMODE_PLUS_MIX_DISABLED
*       05/03/10, added handling clocking mastership (MCU or Atlas)
*/
int mxc_setup_dam(mxc_state_t *drv_inst)
{
        FUNC_START

        dam_port target_port = drv_inst->dam_dest;
        dam_port source_port = (drv_inst->ssi_index == SSI1 ? port_1 : port_2);

        /* Set synchronous mode for up/down streams */
        dam_set_synchronous(source_port, true);
        dam_set_synchronous(target_port, true);

        /*! normal mode for the MCU and Atlas sides */
        dam_select_mode(source_port, normal_mode);
        dam_select_mode(target_port, normal_mode);

        if ( drv_inst->clock_master == MXC_AUDIO_CLOCKING_ATLAS_MASTER )
                mxc_setup_dam_atlas_master(drv_inst);
        else
                mxc_setup_dam_mcu_master(drv_inst);
                
        FUNC_END
        return 0;
}

/*! 
* This function resets the DAM configuration (DAM dependent)
* by setting port1-port6 config to be sourced by port7
*
* @param        drv_inst [in] info about the current transfer
* @return       0 in case of success
*
* Modification History:
*       05/03/19, creation
*/
void mxc_init_dam(mxc_state_t *drv_inst)
{
#ifndef SCMA11_EVB_PASS1
        dam_port port;
        FUNC_START
        
        for ( port=port_1; port<port_7; port++ ) {
                dam_select_mode(port, normal_mode);
                dam_set_synchronous(port, true);
                mxc_dam_clock_connect(port, port_7, signal_out);
                mxc_dam_data_connect(port, port_7, 0);
        }
        
        FUNC_END
#endif //SCMA11_EVB_PASS1
}
